#include "bbdlg.h"

IMPLEMENT_DYNAMIC_CLASS(BBDialog, wxDialog)
  BEGIN_EVENT_TABLE(BBDialog, wxDialog)
  EVT_BUTTON(ID_OK,BBDialog::OnOK)
  EVT_BUTTON(ID_Close,BBDialog::OnClose)
  EVT_CLOSE(BBDialog::OnCloseEvent)
  EVT_LISTBOX(-1, BBDialog::OnListSelected)
  EVT_BUTTON(ID_Reset,BBDialog::OnReset)
END_EVENT_TABLE()

extern GLCanvas * glCanvas;
extern wxListBox * imageList;
extern wxListBox * calibImageList;
extern wxListBox * segImageList;

/** OK button handler.
 * The event handler notifying successful selection of the bounding
 * box.  The okHandler object is notifiled and this object is
 * unregistered as a mouse listener from the GLCanvas
 */
void BBDialog::OnOK(wxCommandEvent & event)
{
  //do something here
  if(okHandler){
    okHandler->OnOK(GetId(),event);
  }
  images->Clear();
  Show(FALSE);
  glCanvas->UnRegisterListener(this);  
}

/** Close button handler.
 * The event handler notifying the BBDialog window has been 
 * closed.  The window is hidden and the object is
 * unregistered as a mouse listener from the GLCanvas.
 */
void BBDialog::OnClose(wxCommandEvent & event)
{
  images->Clear();
  EndModal(wxID_CANCEL);
  Show(FALSE);
  glCanvas->UnRegisterListener(this);
}

/**Called when window is closed
   Used to be commandEvent, but something changed with version upgrade
 */
void BBDialog::OnCloseEvent(wxCloseEvent & event){
  wxCommandEvent ce;
  OnClose(ce);
}

/** Completes the implementation of GLMouseListener
 * This function gets called when the bbdlg should repaint
 * the glcanvas.
 * \param w the integer width of the GLCanvas
 * \param h the integer height of the GLCanvas
*/
void BBDialog::drawGL(int w,int h)
{
  int selimage=GetSelectionSafe(calibImageList);
  

  glWidth=w;
  glHeight=h;
  int sel=GetSelectionSafe(images);
  if(sel>=0){
    ImageData * imd = (ImageData*)images->GetClientData(sel);
    if(imd){
      //NEW aspect ratio handling stuff
      double imaspect=((double)imd->image.w)/((double)imd->image.h);
      double glclipx=((double)w)/((double)h);
      double glclipy=1.0;
      if(glclipx<imaspect){
	glclipx=imaspect;
	glclipy=((double)h)/((double)w)*imaspect;
      }
      glclipx*=1.05;
      glclipy*=1.05;

      glTranslatef(w/2.0,h/2.0,0.0);
      glScalef(((double)imaspect)/glclipx,
	       ((double)1.0)/glclipy,1.0);
      glTranslatef(-w/2,-h/2,0.0);
      
      glColor3f(0,0,1);
      glDisable(GL_DEPTH_TEST);
      /*
      //Test the square
      glBegin(GL_LINE_LOOP);
      glVertex2f(0,0);
      glVertex2f(w,0);
      glVertex2f(w,h);
      glVertex2f(0,h);
      glEnd();
      printf("drawing line loop\n");
      */
      //End aspect stuff

      Matrix  pts(4,8);
      Matrix  projpts(3,8);
      pts.setAll(1);
      pts(0,0)=0;pts(1,0)=0;pts(2,0)=0;
      pts(0,1)=1;pts(1,1)=0;pts(2,1)=0;
      pts(0,2)=1;pts(1,2)=1;pts(2,2)=0;
      pts(0,3)=0;pts(1,3)=1;pts(2,3)=0;
      pts(0,4)=0;pts(1,4)=0;pts(2,4)=1;
      pts(0,5)=1;pts(1,5)=0;pts(2,5)=1;
      pts(0,6)=1;pts(1,6)=1;pts(2,6)=1;
      pts(0,7)=0;pts(1,7)=1;pts(2,7)=1;

      Matrix projsquare(3,8);
      Matrix square(4,8);
      square.setAll(1);
      square(0,0)=0;square(1,0)=0;square(2,0)=0;
      square(0,1)=1;square(1,1)=0;square(2,1)=0;
      square(0,2)=1;square(1,2)=0;square(2,2)=1;
      square(0,3)=0;square(1,3)=0;square(2,3)=1;
      square(0,4)=0.5;square(1,4)=0;square(2,4)=0.5;
      square(0,5)=0.5;square(1,5)=0;square(2,5)=0.5;
      square(0,6)=0.5;square(1,6)=1;square(2,6)=0.5;
      
      int lines[]={0,1,5,1,2,6,2,3,7,3,0,4,5,6,7,4};
      double neard=10e10;
      double fard=-10e10;

      Matrix sc = Matrix::eye(4);
      sc(0,0)=scale[0];
      sc(1,1)=scale[1];
      sc(2,2)=scale[2];
	
      Matrix tr = Matrix::eye(4);
      tr(0,3)=trans[0];
      tr(1,3)=trans[1];
      tr(2,3)=trans[2];
	
      pts=imd->extrinsic*tr*sc*pts;

      
      square=tr*sc*square;
      for(int i=0;i<5;i++)
	square(1,i)=0;
      square=imd->extrinsic*square;
      
      
      for(int i=0;i<8;i++){
	projpts(0,i)=pts(0,i)/pts(3,i);
	projpts(1,i)=pts(1,i)/pts(3,i);
	projpts(2,i)=pts(2,i)/pts(3,i);
	
	projsquare(0,i)=square(0,i)/square(3,i);
	projsquare(1,i)=square(1,i)/square(3,i);
	projsquare(2,i)=square(2,i)/square(3,i);
      }
      projpts=imd->intrinsic*projpts;
      projsquare=imd->intrinsic*projsquare;
      for(int i=0;i<8;i++){
	projpts(0,i)=projpts(0,i)/projpts(2,i);
	projpts(1,i)=projpts(1,i)/projpts(2,i);	  
	if(projpts(2,i)<neard)neard=projpts(2,i);
	if(projpts(2,i)>fard)fard=projpts(2,i);

	projsquare(0,i)=projsquare(0,i)/projsquare(2,i);
	projsquare(1,i)=projsquare(1,i)/projsquare(2,i);
      }
      
      glBegin(GL_LINE_LOOP);
      for(int i=0;i<16;i++){
	double dist=(projpts(2,lines[i])-neard)/(fard-neard);
	glColor3f(1.0-dist,1.0-dist,1.0-dist);
	glVertex2f(projpts(0,lines[i])*w/(double)imd->image.width,
		   projpts(1,lines[i])*h/(double)imd->image.height);
      }
      glEnd();
      
      glDisable(GL_DEPTH_TEST);
      glColor3f(1.0,1.0,0.0);
      glBegin(GL_LINE_LOOP);
      for(int i=0;i<4;i++){
	glVertex2f(projsquare(0,i)*w/(double)imd->image.width,
		   projsquare(1,i)*h/(double)imd->image.height);
      }
      glEnd();

      glBegin(GL_POINTS);
      glVertex2f(projsquare(0,4)*w/(double)imd->image.width,
		 projsquare(1,4)*h/(double)imd->image.height);
      glEnd();

      glBegin(GL_LINES);
      glVertex2f(projsquare(0,5)*w/(double)imd->image.width,
		 projsquare(1,5)*h/(double)imd->image.height);
      glVertex2f(projsquare(0,6)*w/(double)imd->image.width,
		 projsquare(1,6)*h/(double)imd->image.height);
      glEnd();
	
    }
  }
}

/** The image list selection callback.
 * Notifies the GLCanvas that the selected image has
 * changed.
 * \sa ImagesChanged()
 */
void BBDialog::OnListSelected(wxCommandEvent & event){
  ImagesChanged();
}

/** Redraw the screen.
 * The GLCanvas is set to draw the currently selected image 
 */
void BBDialog::ImagesChanged(){
  int sel=GetSelectionSafe(images);
  if(sel>=0){
    ImageData  * imd = (ImageData *)images->GetClientData(sel);
    glCanvas->UpdateTexture(*imd);
  }
}

/** Completes the implementation of GLMouseListener.
 * This is called from the GLCanvas whenever a mouse event occurs
 * \param event The mouse event
 * \param GLw the width of the GLCanvas
 * \param GLh the height of the GLCanvas  
 */
void BBDialog::mouse(wxMouseEvent & event,int GLw,int GLh){
  Raise();
  if(event.ButtonDown()){
    mouseDown=event.GetPosition();
  }
  if(event.ButtonUp()){
    ImagesChanged();
  }
  double dx=event.GetPosition().x-mouseCur.x;
  double dy=event.GetPosition().y-mouseCur.y;
  dx=dx/200;
  dy=dy/200;
  if(event.LeftIsDown()){
    int sel=GetSelectionSafe(images);
    ImageData * imd = (ImageData*)images->GetClientData(sel);
    if(event.ControlDown()){
      if(sel>=0 && imd){
	trans[0]-=dy*imd->extrinsic(2,0);
	//trans[1]+=dy*imd->extrinsic(2,1);
	trans[2]-=dy*imd->extrinsic(2,2);

	//Get the axis that image space x transforms to
	trans[0]+=dx*imd->extrinsic(0,0);
	//trans[1]+=dx*imd->extrinsic(0,1);
	trans[2]+=dx*imd->extrinsic(0,2);
      }
    }
    else{
      trans[1]-=(dy);//
      if(trans[1]<0){
	//trans[1]=0;
      }
    }
    
  }
  else if(event.RightIsDown()){
    //The scaling is about the center
    //of the object, so remove any translation
    //used to counter scale and and it in later
    trans[0]+=0.5*scale[0];
    trans[1]+=0.5*scale[1];
    trans[2]+=0.5*scale[2];

    int sel=GetSelectionSafe(images);
    ImageData * imd = (ImageData*)images->GetClientData(sel);
    if(event.ControlDown()){
      if(sel>=0 && imd){
	scale[0]+=dx*imd->extrinsic(0,0);
	scale[2]+=dx*imd->extrinsic(0,2);
      }
      scale[1]-=dy;
    }
    else{
      scale[0]-=dy;
      scale[1]-=dy;
      scale[2]-=dy;
    }
    trans[0]-=0.5*scale[0];
    trans[1]-=0.5*scale[1];
    trans[2]-=0.5*scale[2];
  }
  //text->Clear();
  wxString str;
  str.Printf(wxT("sc: %5.3f,%5.3f,%5.3f Trans: %5.3f %5.3f %5.3f"),
	     scale[0],scale[1],scale[2],trans[0],trans[1],trans[2]);
  //text->AppendText(str);
  //text->SetLabel(str);
  SetStatusText(str,1);
  mouseCur=event.GetPosition();
}

/** Show/Hide the BBDialog window
 * Shows/Hides the window and ensures that this object
 * is registered to recieve mouse events from the 
 * GLCanvas
 * \param show show/hide the window
 */
bool BBDialog::Show(const bool show){
  if(show){
    glCanvas->SetMode(GL_CANVAS_IMAGE);
    nb->SetSelection(0);
    if(images->GetCount()>0){
      glCanvas->RegisterListener(this);
    }
    ImagesChanged();
  }
  else{
    glCanvas->SetMode(GL_CANVAS_MODEL);
    if(glCanvas)
      glCanvas->UnRegisterListener(this);
    Lower();
  }
  wxDialog::Show(show);
  return TRUE;
}

/** Reset button callback.
 * Resets the scale and translation to the DEFAULT_SCALE
 * and DEFAULT_TRANS resp.
 */
void BBDialog::OnReset(wxCommandEvent & event){
  scale=DEFAULT_SCALE;
  trans=DEFAULT_TRANS;
}
