/**Dialog for obtaining a bounding box for the bounding box of an object.
 * 
 *  The dialog allows for selection of a bounding box that encloses the
 * 3D object.  The BBDialog also contains a command string, which may be
 * set and retrieved so that a single OKHandler, may be associated with 
 * different BBDialog 's.  
 *
 *  ImageData objects are added to the bounding box, which
 * allow the user to view the bounding box overlayed on the current the
 * selected image (using the ImageData calibration).
 * 
 * The dialog registers with the GLCanvas to recieve mouse events, which
 * are used to scale and translate the bounding box until desired.
 *
 * When the user has OK'd the bounding box, the OKHandler associated with
 * the BBDialog on instantiation is notified.
 * \sa MyFrame::OnOK()
 */
#ifndef BB_DLG_H
#define BB_DLG_H

#include <wx/wx.h>
#include <wx/notebook.h>
#include <wx/spinbutt.h>
#include <wx/spinctrl.h>
#include "guitypes.h"
#include "glcanvas.h"
#include "image.h"

class BBDialog : public wxDialog,GLMouseListener{
  //List box containing the ImageData 
  wxListBox * images;

  wxNotebook * nb;
  int glWidth;
  int glHeight;
  wxPoint mouseDown,mouseCur,mouseUp;
  wxStaticText * text;
  wxButton * reset;

  //The okHandler associated with this object
  OKHandler * okHandler;
 public:
  ///The command string
  char * command;

  ///The current scale of the bounding box
  Vec3f trans;

  ///The current translation of the bounding box
  Vec3f scale;
 public:
  DECLARE_DYNAMIC_CLASS(BBDialog)
    BBDialog(wxWindow * parent=0,wxWindowID id=-1,OKHandler * okh=0):wxDialog(parent,id,wxT("Bounding Box"),
									    wxPoint(-1,-1),
									    wxSize(230,200)){
    okHandler=okh;
    wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);
    
    nb = new wxNotebook(this,-1);
    wxPanel * colPanel = new wxPanel(nb,-1);
    wxBoxSizer * colSizer = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer * buttonSizer = new wxBoxSizer(wxHORIZONTAL);
    wxBoxSizer * horSizer;
    wxBoxSizer * verSizer;
    
    //Init the color 
    horSizer = new wxBoxSizer(wxHORIZONTAL);
    
    
    images = new wxListBox(colPanel,-1);
    horSizer->Add(images,1,wxGROW|wxEXPAND);
    /*horSizer->Add(text=new wxStaticText(colPanel,-1,"Please select the bounding box of the objec in the two provided images.",wxDefaultPosition,
				      wxDefaultSize,wxST_NO_AUTORESIZE));
				      //wxTE_MULTILINE|wxTE_READONLY),1,wxGROW|wxEXPAND);*/
    horSizer->Add(reset=new wxButton(colPanel,ID_Reset,wxT("Reset")));
    colSizer->Add(horSizer,1,wxGROW|wxEXPAND);
    colPanel->SetSizer(colSizer);

    //End init color segmentor

    nb->AddPage(colPanel,wxT("Color Seg."),TRUE);
	
    buttonSizer->Add(new wxButton(this,ID_OK,wxT("OK")),1,wxSHAPED|wxALIGN_CENTRE);
    buttonSizer->Add(new wxButton(this,ID_Close,wxT("Close")),1,wxSHAPED|wxALIGN_CENTRE);

    //wxNotebookSizer * nbSizer = new wxNotebookSizer(nb);
    wxSizer * nbSizer = new wxBoxSizer(wxVERTICAL);
    nbSizer->Add(nb,1,wxGROW);

    all->Add(nbSizer,3,wxGROW|wxEXPAND);
    all->Add(buttonSizer,1,wxGROW|wxALIGN_CENTRE);

    SetSizer(all);
    Layout();
#if wxUSE_TOOLTIPS

#endif
    scale[0]=200;
    scale[1]=200;
    scale[2]=200;
	
    trans[0]=-0.5;
    trans[1]=0;
    trans[2]=-0.5;
    CentreOnScreen();
    command=0;
  }
  /** Destructor.
   * Delete the command string.
   */
  ~BBDialog(){
    printf("in ~BBDialog()\n");
    if(command)delete [] command;
    command=0;
    printf("done ~BBDialog()\n");
  }
  /** Set the command string.
   * The old command string is deleted and is replaced
   * with the new value, d, for the command string.
   * \param d the new command string.
   */
  void SetCommand(char * d){
    if(command)
      delete [] command;
    if(d){
      command = new char[strlen(d)+1];
      strcpy(command,d);
    }
    else
      command=0;
  }
  /** Set the scale.
   * \param sc the new value for the scale.
   */
  void SetScale(Vec3f & sc){
    scale=sc;
  }
  /** Set the translation to tr.
   * \param tr the new value for the translation.
   */
  void SetTrans(Vec3f & tr){
    trans=tr;
  }
  /** Add the ImageData.image to the list. 
   * The ImageData.image is added to the listbox for use in 
   * selecting the bounding box
   */
  void AddImage(ImageData * imd){
    wxString name = wxString::Format(wxT("Image %d"), images->GetCount());
    images->Append(name,imd);
    if(images->GetSelection()<0)
      images->SetSelection(0);
  }
  void OnOK(wxCommandEvent & event);
  void OnClose(wxCommandEvent & event);
  void OnCloseEvent(wxCloseEvent & event);
  void OnReset(wxCommandEvent & event);
  virtual void drawGL(int w,int h);
  void ImagesChanged();
  void OnListSelected(wxCommandEvent & event);
  virtual void mouse(wxMouseEvent & event,int GLw,int GLh);
  bool Show(const bool show);
 protected:
  DECLARE_EVENT_TABLE();
};
#endif
