#include "calibratedlg.h"
#include "glcanvas.h"
#include "imagedata.h"
#include "guitypes.h"
#include "timer.h"

IMPLEMENT_DYNAMIC_CLASS(CalibrateDialog, wxDialog)
  BEGIN_EVENT_TABLE(CalibrateDialog, wxDialog)
  EVT_BUTTON(ID_OK,CalibrateDialog::OnOK)
  EVT_BUTTON(ID_Close,CalibrateDialog::OnClose)
  EVT_SPIN(OPT_SMALL,CalibrateDialog::OnSpinSmall)
  EVT_SPIN(OPT_BIG,CalibrateDialog::OnSpinBig)
  EVT_COMMAND_SCROLL_THUMBTRACK(OPT_SCROLL,CalibrateDialog::OnScroll)
  EVT_CLOSE(CalibrateDialog::OnCloseEvent)
  EVT_CHECKBOX(OPT_AUTO,CalibrateDialog::OnAuto)
  EVT_IDLE (CalibrateDialog::OnIdle)
  //EVT_CHECKBOX(OPT_GRAY,CalibrateDialog::OnGray)
END_EVENT_TABLE()

extern wxListBox * imageList;
extern wxListBox * calibImageList;
extern GLCanvas * glCanvas;
extern Options options;

CalibrateDialog::CalibrateDialog(wxWindow * parent,wxWindowID id)
  :wxDialog(parent,id,wxT("Options"),
	    wxPoint(-1,-1),
	    wxSize(230,200)){
  wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);
  
  wxStaticBox * threshBox = new wxStaticBox(this,
					    -1,
					    wxT("Threshold"));
  wxStaticBox * generalBox = new wxStaticBox(this,
					     -1,
					     wxT("Sizes"));
  wxBoxSizer * threshBoxSizer = new wxStaticBoxSizer(threshBox,
						     wxVERTICAL);
  wxBoxSizer * buttonSizer = new wxBoxSizer(wxHORIZONTAL);
  
  wxBoxSizer * general = new wxStaticBoxSizer(generalBox,
					      wxVERTICAL);
  /*
    gray = new wxCheckBox(this,OPT_GRAY,"Gray",
    wxPoint(-1,-1));
    red = new wxSlider(this,OPT_SCROLL,160,0,255,
    wxPoint(-1,-1),
    wxSize(-1,-1));
    green = new wxSlider(this,OPT_SCROLL,120,0,255,
    wxPoint(-1,-1),
    wxSize(-1,-1));
    blue = new wxSlider(this,OPT_SCROLL,160,0,255,
    wxPoint(-1,-1),
    wxSize(-1,-1));
  */
  channel = new wxComboBox(this,-1);
  channel->Append(wxT("Red"));
  channel->Append(wxT("Green"));
  channel->Append(wxT("Blue"));
  thresh = new wxSlider(this,OPT_SCROLL,160,0,255,
			wxPoint(-1,-1),
			wxSize(-1,-1));
  
  bigSpin = new wxSpinCtrl(this,OPT_BIG,wxT("1000"),wxDefaultPosition,
		       wxDefaultSize,
		       wxSP_ARROW_KEYS,0,10000,1000);
  smallSpin = new wxSpinCtrl(this,OPT_SMALL,wxT("15"),wxDefaultPosition,
			 wxDefaultSize,
			 wxSP_ARROW_KEYS,0,10000,15);
  
  /*wxBoxSizer * temp = new wxBoxSizer(wxHORIZONTAL);
    temp->Add(new wxStaticText(this,-1," "),0,wxALIGN_CENTER);
    temp->Add(gray,1,wxGROW);
    threshBoxSizer->Add(temp,1,wxGROW);
    
    temp = new wxBoxSizer(wxHORIZONTAL);
    temp->Add(new wxStaticText(this,-1,"R"));
    temp->Add(red,1,wxGROW);
    threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

    temp = new wxBoxSizer(wxHORIZONTAL);
    temp->Add(new wxStaticText(this,-1,"G"));
    temp->Add(green,1,wxGROW);
    threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

    temp = new wxBoxSizer(wxHORIZONTAL);
    temp->Add(new wxStaticText(this,-1,"B"));
    temp->Add(blue,1,wxGROW);
    threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);
  */
  wxBoxSizer * temp = new wxBoxSizer(wxHORIZONTAL);

  temp = new wxBoxSizer(wxHORIZONTAL);
  temp->Add(new wxStaticText(this,-1,wxT("Channel")));
  temp->Add(channel);//1,wxGROW);
  threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

  temp = new wxBoxSizer(wxHORIZONTAL);
  temp->Add(new wxStaticText(this,-1,wxT("Thresh")));
  temp->Add(thresh,1,wxGROW);
  threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

  temp = new wxBoxSizer(wxHORIZONTAL);
  temp->Add(automatic=new wxCheckBox(this,OPT_AUTO,wxT("Automatic")),1,wxGROW);
  temp->Add(fitFocalLength=new wxCheckBox(this,OPT_FOCAL_LENGTH,wxT("Fit Focal")),1,wxGROW);
  threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

  wxFlexGridSizer * grid = new  wxFlexGridSizer(1,2,2,2);

  grid->Add(new wxStaticText(this,-1,wxT("Small")),0,wxALIGN_RIGHT);
  grid->Add(smallSpin,0,wxALIGN_RIGHT);
	
  grid->Add(new wxStaticText(this,-1,wxT("Big ")),0,wxALIGN_RIGHT);
  grid->Add(bigSpin,0,wxALIGN_RIGHT);

  general->Add(grid,1,wxGROW);

  buttonSizer->Add(ok=new wxButton(this,ID_OK,wxT("OK")),1,wxSHAPED|wxALIGN_CENTRE);
  buttonSizer->Add(close=new wxButton(this,ID_Close,wxT("Close")),1,wxSHAPED|wxALIGN_CENTRE);

  all->Add(threshBoxSizer,3,wxGROW);
  all->Add(general,2,wxGROW);
  all->Add(buttonSizer,1,wxGROW|wxALIGN_CENTRE);

  SetSizer(all);
  SetAutoLayout(TRUE);
  Layout();

  calibrating=0;
  index=0;
#if wxUSE_TOOLTIPS

#endif 
  channel->SetSelection(options.calibChannel);
  thresh->SetValue(options.calibThresh);
  bigSpin->SetValue(options.calibBig);
  smallSpin->SetValue(options.calibSmall);
  automatic->SetValue(options.calibAutoThresh);

  CentreOnScreen();
}

/**Threshold slider callback.
 * The threshold is updated, and the screen is repainted.
 */
void CalibrateDialog::OnScroll(wxScrollEvent & event){
  wxObject * obj = event.GetEventObject();
  /*if(gray->GetValue()){
    int value = ((wxScrollBar *)obj)->GetValue();
    red->SetValue(value);
    green->SetValue(value);
    blue->SetValue(value);
    }*/
  SyncDisplay();
}

/**Synchronize the GLCanvas with the internal state.
 * Display the current black and white image obtained
 * from the thresholding, and the resulting calibration that
 * was obtained from that black and white image.
 */
void CalibrateDialog::SyncDisplay(){
  int sel=GetSelectionSafe(imageList);
  if(sel>=0){
    ImageData * imd=(ImageData *)imageList->GetClientData(sel);
    if(imd){
      ImageData thr;
    
      int t[3];
      double val[3];
      GetThresh(t);
      val[0]=t[0];val[1]=t[1];val[2]=t[2];
      val[channel->GetSelection()]=thresh->GetValue();
      thr.image=imd->image.threshLess(val)*255;
      thr.intrinsic=imd->intrinsic.copy();
      thr.extrinsic=imd->extrinsic.copy();
      thr.distortion=imd->distortion.copy();
      if(!(imd->flags & HAS_INTRINSIC)){
	printf("image has no calibration\n");
      }
      else{
	double residual=calibrateImage(imd->image,
				       t[0],
				       t[1],
				       t[2],
				       smallSpin->GetValue(),
				       bigSpin->GetValue(),
				       automatic->GetValue(),
				       thr.distortion,
				       thr.intrinsic,
				       thr.extrinsic,
				       fitFocalLength->GetValue());
	thr.calibInfo = new CalibInfo();
	calibrateGetInfo(*(thr.calibInfo));
	calibrateRelease();
	thr.calibInfo->residual=residual;

	glCanvas->UpdateTexture(thr);
	glCanvas->Refresh();
      }
    }
  }
}

///The "small" spinbutton callback.
void CalibrateDialog::OnSpinSmall(wxSpinEvent & event){
  glCanvas->Refresh();
}

///The "big" spinbutton callback.
void CalibrateDialog::OnSpinBig(wxSpinEvent & event){
  glCanvas->Refresh();
}

/**
 * Automatic checkbox callback.
 */
void CalibrateDialog::OnAuto(wxCommandEvent & event){
}

/**
 * Idle event handler, calibration is performed here.
 * When the CalibrateDialog is in calibration state, the
 * selected images are processed from the ImageList and
 * moved to the CalibImageList on successfull calibration.
 */
void CalibrateDialog::OnIdle(wxIdleEvent & event){
  if(calibrating){
    static int attempts=0;
    static Timer timer;
    wxArrayInt selections;
    imageList->GetSelections(selections);
    if(selections.GetCount()){
      //use the index strictly for setting the progress
      if(index==0){
	index=selections.GetCount();
	attempts=0;
	timer.start();
	printf("starting timer\n");
      }
      if(calibImageList->GetString(0)==
	 wxT(" ")){
	calibImageList->Clear();
      }
      if(imageList->GetClientData(selections[0])){
	ImageData * imd = (ImageData *) 
	  imageList->GetClientData(selections[0]);
	if(!(imd->flags & HAS_INTRINSIC)){
	  printf("image has no calibration\n");
	  imageList->Deselect(selections[0]);
	}
	else{
	  int t[3];
	  GetThresh(t);
	  double residual=calibrateImage(imd->image,
					 t[0],
					 t[1],
					 t[2],
					 smallSpin->GetValue(),
					 bigSpin->GetValue(),
					 automatic->GetValue(),
					 imd->distortion,
					 imd->intrinsic,
					 imd->extrinsic,
					 fitFocalLength->GetValue());
	  if(imd->calibInfo){		      
	    delete imd->calibInfo;
	  }
	  imd->calibInfo = new CalibInfo();
	  calibrateGetInfo(*(imd->calibInfo));
	  calibrateRelease();
	  if(residual<options.maxResidual 
	     && imd->calibInfo->nlines>2){
	    imd->set(HAS_EXTRINSIC);
	    imd->set(HAS_CALIBINFO);
	    imd->calibInfo->residual=residual; //FIXME, this should get set in calibrateGetInfo...which should be fixed also recall it is normalized by the number of points
	    
	    calibImageList->Append(imageList->GetString(selections[0]),
				   imageList->GetClientData(selections[0]));
	    imageList->Delete(selections[0]);
	    attempts=0;
	  }
	  else{
	    imd->unset(HAS_EXTRINSIC);
	    imd->unset(HAS_CALIBINFO);
	    if(imd->calibInfo){
	      delete imd->calibInfo;
	    }
	    imd->calibInfo=0;
	    imageList->Deselect(selections[0]);
	  }
	  SetChangedSinceSave(IMAGES_CHANGED);
	}
      }//imageList->GetClientData
      else{
	imageList->Deselect(selections[0]);
      }
      if(index>1)
	SetProgress((index-selections.GetCount())/(float)(index-1));
      else
	SetProgress(1.0f);
    }
    else{
      printf("stopped timer, elapsed time %g\n",timer.stop());
		     
      calibrating=0;
      ok->Enable();
      index=0;
      if(GetUncalImageCount()==0){
	SetStatusText(wxT("You can now segment the images"));
      }
      wxCommandEvent dummy;
      OnClose(dummy);
    }
    event.RequestMore(TRUE);
  }
  else{
    event.RequestMore(FALSE);
  }
}

/**OK button callback.
 * Puts the CalibrateDialog in calibrate mode.
 * \sa OnIdle
 */
void CalibrateDialog::OnOK(wxCommandEvent & event){
  index=0;
  calibrating=1;
  ok->Disable();
  //printf("on okay\n");
  //EndModal(wxID_OK);
}

/**Close button callback.
 * Close the window and update the global Options to reflect
 * the changes made in this dialog.  
 */
void CalibrateDialog::OnClose(wxCommandEvent & event){
  //printf("Close\n");
  calibrating=0;
  index=0;
  ok->Enable();
  EndModal(wxID_CANCEL);

  //Save the options
  options.calibThresh=thresh->GetValue();
  options.calibAutoThresh=automatic->GetValue();
  options.calibChannel=channel->GetSelection();
  options.calibBig=bigSpin->GetValue();
  options.calibSmall=smallSpin->GetValue();
}
void CalibrateDialog::OnCloseEvent(wxCloseEvent & event){
  wxCommandEvent dummy;
  OnClose(dummy);
}
/*
  void CalibrateDialog::OnGray(wxCommandEvent & event){
  if(gray->GetValue()){
  int value = red->GetValue()+green->GetValue()+blue->GetValue();
  red->SetValue(value/3);
  green->SetValue(value/3);
  blue->SetValue(value/3);
  glCanvas->Refresh();
  }
  }*/
