#ifndef CALBIRATE_DLG_H
#define CALIBRATE_DLG_H
#include <wx/wx.h>
#include <wx/spinbutt.h>
#include <wx/spinctrl.h>
#include "guitypes.h"

/** Dialog for calibrating the selected images in the ImageList.
 * The dialog allows for threshold and color channel selection
 * to generate a black and white image, for use in the calibration.
 *
 * When calibrating, the CalibrateDialog accesses the selected images 
 * in the ImageList, calibrates, and sends them to the CalibImageList.
 * This processing all happens in the OnIdle function.
 */
class CalibrateDialog : public wxDialog{
  enum {OPT_SCROLL,
	OPT_SMALL,OPT_BIG,OPT_GRAY,OPT_AUTO,OPT_FOCAL_LENGTH};
    
  ///checkbox for automatic threshold adjustment
  wxCheckBox * automatic;
  
  ///checkbox for whether the focal length is fit automatically
  wxCheckBox * fitFocalLength;
  
  ///combobox for color channel selection
  wxComboBox * channel;
  ///The threshold slider
  wxSlider * thresh; 
  wxSpinCtrl * smallSpin, * bigSpin;
  wxButton * ok,*close;
  int calibrating;
  int index;

  //variables for automatic calibration
  int    lastGood;
  double lastResidual;

 public:
  DECLARE_DYNAMIC_CLASS(CalibrateDialog)

  CalibrateDialog(wxWindow * parent=0,wxWindowID id=-1);
  void GetThresh(int * thresh){
    thresh[0]=thresh[1]=thresh[2]=255;
    thresh[channel->GetSelection()]=this->thresh->GetValue();
    /*
    thresh[0]=red->GetValue();
    thresh[1]=green->GetValue();
    thresh[2]=blue->GetValue();*/
  }
  int GetSmall(){
    return smallSpin->GetValue();
  }
  int GetBig(){
    return bigSpin->GetValue();
  }
  void SyncDisplay();
  void OnScroll(wxScrollEvent & event);
  void OnAuto(wxCommandEvent & event);
  //void OnGray(wxCommandEvent & event);
  void OnSpinSmall(wxSpinEvent & event);
  void OnSpinBig(wxSpinEvent & event);
  void OnOK(wxCommandEvent & event);

  void OnClose(wxCommandEvent & event);
  void OnCloseEvent(wxCloseEvent & event);

  void OnIdle(wxIdleEvent & event);
 protected:
  DECLARE_EVENT_TABLE();
};

#endif
