#include "cameradlg.h"
#include "cameraprofile.h"
#include "glcanvas.h"

extern wxListBox * imageList;
extern wxListBox * segImageList;
extern wxListBox * calibImageList;

extern GLCanvas * glCanvas;
extern Options options;
extern CameraProfiles cameraProfiles;

IMPLEMENT_DYNAMIC_CLASS(CameraDialog,wxDialog)

BEGIN_EVENT_TABLE(CameraDialog, wxDialog)
  EVT_CHECKBOX(ID_Current, CameraDialog::OnCurrent)
  EVT_BUTTON(ID_Close, CameraDialog::OnClose)
  EVT_BUTTON(ID_New, CameraDialog::OnNew)
  EVT_BUTTON(ID_OK, CameraDialog::OnOK)
  EVT_BUTTON(ID_Force,CameraDialog::OnForce)
  EVT_BUTTON(ID_Remove, CameraDialog::OnRemove)
  EVT_LISTBOX(ID_Camera,CameraDialog::OnCameraChanged)
  EVT_LISTBOX_DCLICK(ID_Camera,CameraDialog::OnCameraDoubleClick)
  EVT_TEXT(-1,CameraDialog::OnTextCtrlChanged)
  //EVT_IDLE(CameraDialog::OnIdle)
END_EVENT_TABLE()


CameraDialog::CameraDialog(wxWindow * parent,wxWindowID id):
wxDialog(parent,id,wxT("Cameras"),wxDefaultPosition,
	 wxSize(400,320)){
  wxButton * okButton, * newButton;
  wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);
  wxBoxSizer * top = new wxBoxSizer(wxHORIZONTAL);
  wxStaticBoxSizer * paramSizer
    = new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Params")),
			   wxHORIZONTAL);

  cameraList = new wxListBox(this,ID_Camera,wxDefaultPosition,
			     wxDefaultSize,0,0,wxLB_SORT|wxLB_SINGLE);
  newButton = new wxButton(this,ID_New,wxT("New"));
  removeButton = new wxButton(this,ID_Remove,wxT("Remove"));
  currentButton = new wxCheckBox(this,ID_Current,wxT("Current"));
  forceButton = new wxButton(this,ID_Force,wxT("Force"));
  wxStaticBoxSizer * cameraSizer 
    = new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Cameras")),
			   wxHORIZONTAL);
  wxBoxSizer * buttonSizer
    = new wxBoxSizer(wxVERTICAL);
  
  cameraSizer->Add(cameraList,0,wxGROW|wxEXPAND);
  buttonSizer->Add(newButton,0,wxALIGN_CENTER_VERTICAL);
  buttonSizer->Add(removeButton,0,wxALIGN_CENTER_VERTICAL);
  buttonSizer->Add(currentButton,0,wxALIGN_CENTER_VERTICAL);
  buttonSizer->Add(forceButton,0,wxALIGN_CENTER_VERTICAL);
  top->Add(cameraSizer,0,wxGROW|wxEXPAND);
  top->Add(buttonSizer);
  
  wxStaticBoxSizer * box=
    new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Intrinsic")),
			 wxVERTICAL);
  wxBoxSizer * row=new wxBoxSizer(wxHORIZONTAL);  
  row->Add(new wxStaticText(this,-1,wxT("fx:")),1,wxGROW|wxEXPAND);
  row->Add(textctrl[0]=new wxTextCtrl(this,-1),3,wxGROW|wxEXPAND);
  box->Add(row,1,wxGROW|wxEXPAND);

  row=new wxBoxSizer(wxHORIZONTAL);
  row->Add(new wxStaticText(this,-1,wxT("fy:")),1,wxGROW|wxEXPAND);
  row->Add(textctrl[1]=new wxTextCtrl(this,-1),3,wxGROW|wxEXPAND);
  box->Add(row,1,wxGROW|wxEXPAND);

  row=new wxBoxSizer(wxHORIZONTAL); 
  row->Add(new wxStaticText(this,-1,wxT("cx:")),1,wxGROW|wxEXPAND);
  row->Add(textctrl[2]=new wxTextCtrl(this,-1),3,wxGROW|wxEXPAND);
  box->Add(row,1,wxGROW|wxEXPAND);

  row=new wxBoxSizer(wxHORIZONTAL);
  row->Add(new wxStaticText(this,-1,wxT("cy:")),1,wxGROW|wxEXPAND);
  row->Add(textctrl[3]=new wxTextCtrl(this,-1),3,wxGROW|wxEXPAND);
  box->Add(row,1,wxGROW|wxEXPAND);
  box->Add(new wxStaticText(this,-1,wxT("")),0);
  
  row=new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Distortion")),
			   wxVERTICAL);
  row->Add(textctrl[4]=new wxTextCtrl(this,-1),1,wxGROW|wxEXPAND);
  row->Add(textctrl[5]=new wxTextCtrl(this,-1),1,wxGROW|wxEXPAND);
  row->Add(textctrl[6]=new wxTextCtrl(this,-1),1,wxGROW|wxEXPAND);
  row->Add(textctrl[7]=new wxTextCtrl(this,-1),1,wxGROW|wxEXPAND);
  row->Add(textctrl[8]=new wxTextCtrl(this,-1),1,wxGROW|wxEXPAND);

  paramSizer->Add(box,4,wxGROW|wxEXPAND);
  paramSizer->Add(row,3,wxGROW|wxEXPAND);

  all->Add(top,4,wxGROW|wxEXPAND);
  all->Add(paramSizer,6,wxGROW|wxEXPAND);
  all->Add(okButton=new wxButton(this,ID_OK,wxT("OK")),0,wxCENTER);
  
  SetSizer(all);
  SetAutoLayout(TRUE);

  //add all the existing camera profiles
  for(int i=0;i<cameraProfiles.nprofiles;i++){
    cameraList->Append(wxString::FromAscii(cameraProfiles[i]->name),
		       cameraProfiles[i]);
  }
  for(int i=0;i<cameraProfiles.nprofiles;){
    int where=cameraList->FindString(wxString::FromAscii(cameraProfiles[i]->name));
    if(where!=i){
      cameraProfiles.swap(i,where);
      if(where<i)i=where;
    }
    else i++;
    cameraProfiles.print();
  }

  dialogChangingText=0;
  UpdateDialog();
  CentreOnScreen();
}

void CameraDialog::OnCameraDoubleClick(wxCommandEvent & event)
{
  printf("FIXME: change name disabled, lexicographical order is not maintained\n");
  /*
  int sel=cameraList->GetSelection();
  if(sel>=0){
    wxTextEntryDialog dlg(this,"Change camera name to:"
			  ,"new camera name","");
    if(wxID_OK==dlg.ShowModal()){
      int sel=cameraList->GetSelection();
      if(sel>=0){
	int where;
	cameraProfiles[sel]->setName((const char *)dlg.GetValue().ToAscii());
	cameraList->SetString(sel,dlg.GetValue());
	where=cameraList->FindString(dlg.GetValue());
	printf("updated item is at %d, was at %d\n",where,sel);
      }
    }
    }*/
}

void CameraDialog::OnTextCtrlChanged(wxCommandEvent & event)
{
  if(dialogChangingText)return;
  int sel=cameraList->GetSelection();
  if(sel>=0){
    CameraProfile * profile=
      (CameraProfile *)cameraList->GetClientData(sel);
    if(profile){
      for(int j=0;j<9;j++){
	float d;
	printf("%d %s\n",j,(const char*)textctrl[j]->GetValue().ToAscii());
	if(!sscanf((const char *)textctrl[j]->GetValue().ToAscii(),"%f",&d)){
	  printf("invalid entry in textctrl %d\n",j);
	}
	else{
	  printf("Got value %f from %d\n",d,j);
	  if(j<2){
	    profile->intrinsic(j,j)=d;
	  }
	  else if(j<4){
	    profile->intrinsic(j-2,2)=d;
	  }
	  else{
	    profile->distortion[j-4]=d;
	  }
	}
      }
    }
  }
}

void CameraDialog::OnCameraChanged(wxCommandEvent & event)
{
  UpdateDialog();
}

//The purpose of this function is:
// -update the dialog to be consistent with the data
void CameraDialog::UpdateDialog()
{
  int sel=cameraList->GetSelection();
  if(sel<0){
    for(int i=0;i<9;i++){
      textctrl[i]->Disable();
    }
    currentButton->Disable();
    removeButton->Disable();
    forceButton->Disable();
  }
  else{
    for(int i=0;i<9;i++){
      textctrl[i]->Enable();
    }
    currentButton->Enable();
    removeButton->Enable();
    forceButton->Enable();
  }
  currentButton->SetValue(0);
  CameraProfile * profile=(CameraProfile *)cameraList->GetClientData(sel);

  dialogChangingText=1;
  if(profile){
    for(int i=0;i<9;i++){
      wxString str;
      double val;
	
      if(i<2){
	val=profile->intrinsic(i,i);
      }
      else if(i<4){
	val=profile->intrinsic(i-2,2);
      }
      else{
	val=profile->distortion[i-4];
      }
      str.Printf(wxT("%8.3f"),val);
      textctrl[i]->SetValue(str);
    }
    if(sel==cameraProfiles.currentProfile){
      currentButton->SetValue(1);
      SetStatusText(wxT("Current Camera profile"),1);
    }
    else{
      SetStatusText(wxT("Not the current camera profile"),1);
    }
  }
  else{
    for(int i=0;i<9;i++){
      textctrl[i]->SetValue(wxT("-1"));
    }
  }
  dialogChangingText=0;
}
void CameraDialog::AddNewProfile(wxString & str,
				 double fx,double fy,
				 double px,double py)
{
  CameraProfile * cp=new CameraProfile((const char *)str.ToAscii(),fx,fy,px,py);
  cameraProfiles.add(cp);
  cameraList->SetSelection(-1);
  cameraList->Append(str,cp);
  int where=cameraList->FindString(str);
  printf("Newly inserted item is at %d\n",where);
  
  //Maintain same order in the cameraprofiles...
  cameraProfiles.swap(cameraProfiles.nprofiles-1,where);
  
  cameraList->SetSelection(where);
  cameraProfiles.currentProfile=where;
  UpdateDialog();
}
void CameraDialog::OnNew(wxCommandEvent & event)
{
  wxTextEntryDialog dlg(this,wxT("Enter name to associate with camera")
			,wxT("new camera name"),wxT(""));
  if(wxID_OK==dlg.ShowModal()){
    wxString str=dlg.GetValue();
    AddNewProfile(str);
  }
}
void CameraDialog::OnForce(wxCommandEvent & event)
{
  int sel=cameraList->GetSelection();
  if(sel>=0){
    int res=wxMessageBox(wxT("This will cause all images in the uncalibrated image list to use the selected camera profile, continue?"),
			 wxT("Confirm"),
			 wxICON_QUESTION|wxYES_NO,NULL);
    if(res==wxYES){
      int i=0;
      CameraProfile * cp = cameraProfiles[sel];
      wxArrayInt selections;
      imageList->GetSelections(selections);
      printf("There are %d images selected\n",selections.GetCount());
      for(int i=0;i<selections.GetCount();i++){
	ImageData * imd=(ImageData *)imageList->GetClientData(selections[i]);
	if(imd){
	  printf("setting image data %d\n",i);
	  imd->intrinsic=cp->intrinsic;
	  imd->distortion=cp->distortion;
	  imd->flags|=(HAS_DISTORTION|HAS_INTRINSIC);
	}
      }
      SetStatusText(wxT("Forced camera parameters completed"));
    }
    else
      SetStatusText(wxT("Force camera parameters aborted"));
  }
}

void CameraDialog::OnCurrent(wxCommandEvent & event)
{
  int sel=cameraList->GetSelection();
  if(sel>=0){
    cameraProfiles.currentProfile=sel;
    UpdateDialog();
  }
}

void CameraDialog::OnRemove(wxCommandEvent & event)
{
  int sel=cameraList->GetSelection();
  if(sel>=0){
    cameraList->Delete(sel);
    cameraProfiles.remove(sel);
    UpdateDialog();
  }
}

void CameraDialog::OnOK(wxCommandEvent & event)
{

  EndModal(wxID_OK);
}

void CameraDialog::OnClose(wxCommandEvent & event)
{
  EndModal(wxID_CANCEL);
}

void CameraDialog::OnIdle(wxIdleEvent & event)
{
  event.RequestMore(FALSE);
}

