#include "capturedlg.h"
#include "cameraprofile.h"
#include "cameradlg.h"
#include "glcanvas.h"
#include <wx/thread.h>
static wxMutex s_mutex;

/**Down sample the Bayer image.
 * Currently there is no support for an upsampled
 *  bayer image.
 */
//int downSampleBayer=1;

extern wxListBox * imageList;
extern GLCanvas * glCanvas;
extern Options options;
extern CameraProfiles cameraProfiles;

IMPLEMENT_DYNAMIC_CLASS(CaptureDialog,wxDialog)

BEGIN_EVENT_TABLE(CaptureDialog, wxDialog)
  EVT_BUTTON(ID_Close, CaptureDialog::OnClose)
  EVT_BUTTON(ID_Grab, CaptureDialog::OnGrab)
  EVT_COMBOBOX(ID_Mode,CaptureDialog::OnMode)
  EVT_COMBOBOX(ID_Format,CaptureDialog::OnFormat)
  EVT_COMBOBOX(ID_Camera,CaptureDialog::OnCamera)
  EVT_IDLE(CaptureDialog::OnIdle)
  EVT_CHECKBOX(ID_BayerButton,CaptureDialog::OnBayerButton)
END_EVENT_TABLE()


#if 0 //orignal support for an emtpy dialg when no camera class supported

CaptureDialog::CaptureDialog(wxWindow * parent,wxWindowID id):
wxDialog(parent,id,wxT("Capture Options"),wxDefaultPosition,
	     wxSize(310,160)){
  new wxStaticText(this,-1,wxT("No Camera Interface"),wxDefaultPosition,
		   wxDefaultSize);
  CentreOnScreen();
}
#endif //ifdef 0


//Image formats and related properties
enum {RGB,YUV444,YUV422,YUV411,MONO,MONO16};
int nchannels[]={3,3,3,3,1,1};
int bytesPerChannel[]={1,1,1,1,1,2};
char fmt2str[][7]={"RGB","YUV444","YUV422","YUV411","MONO","MONO16"};

///The available modes for each camera format
int formatmodes[][8][4]={{{160,120,YUV444,64},
			  {320,240,YUV422,65},
			  {640,480,YUV411,66},
			  {640,480,YUV422,67},
			  {640,480,RGB,68},
			  {640,480,MONO,69},
			  {640,480,MONO16,70},
			  {-1,-1,-1,-1}},
			 {{800,600,YUV422,96},
			  {800,600,RGB,97},
			  {800,600,MONO,98},
			  {1024,768,YUV422,99},
			  {1024,768,RGB,100},
			  {1024,768,MONO,101},
			  {800,600,MONO16,102},
			  {1024,768,MONO16,103}},
			 {{1280,960,YUV422,128},
			  {1280,960,RGB,129},
			  {1280,960,MONO,130},
			  {1600,1200,YUV422,131},
			  {1600,1200,RGB,132},
			  {1600,1200,MONO,133},
			  {1280,960,MONO16,134},
			  {1600,1200,MONO16,135}},
			 {{-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1}},
			 {{-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1}},
			 {{-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1}},
			 {{-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1},
			  {-1,-1,-1,-1}},
			 //Format 7 modes are queried from camera
			 {{0,0,0,0},
			  {0,0,0,0},
			  {0,0,0,0},
			  {0,0,0,0},
			  {0,0,0,0},
			  {0,0,0,0},
			  {0,0,0,0},
			  {0,0,0,0}}};
#ifdef USING_DC1394
int modemins[]={MODE_FORMAT0_MIN,
		MODE_FORMAT1_MIN,
		MODE_FORMAT2_MIN,
		0,//MODE_FORMAT3_MIN,
		0,//MODE_FORMAT4_MIN,
		0,//MODE_FORMAT5_MIN,
		MODE_FORMAT6_MIN,
                MODE_FORMAT7_MIN};
int nummodes[]={NUM_FORMAT0_MODES,
		NUM_FORMAT1_MODES,
		NUM_FORMAT2_MODES,
		0,//NUM_FORMAT3_MODES,
		0,//NUM_FORMAT4_MODES,
		0,//NUM_FORMAT5_MODES,
		NUM_FORMAT6_MODES,
		NUM_MODE_FORMAT7};
#endif

#define YUV2RGB(y, u, v, r, g, b)\
  r = y + ((v*1436) >>10);\
  g = y - ((u*352 + v*731) >> 10);\
  b = y + ((u*1814) >> 10);\
  r = r < 0 ? 0 : r;\
  g = g < 0 ? 0 : g;\
  b = b < 0 ? 0 : b;\
  r = r > 255 ? 255 : r;\
  g = g > 255 ? 255 : g;\
  b = b > 255 ? 255 : b

void uyvy2rgb (unsigned char *src, unsigned char *dest, int NumPixels) {
  register int i = (NumPixels << 1)-1;
  register int j = NumPixels + ( NumPixels << 1 ) -1;
  register int y0, y1, u, v;
  register int r, g, b;
  
  while (i > 0)
  {
    y1 = (unsigned char) src[i--];
    v  = (unsigned char) src[i--] - 128;
    y0 = (unsigned char) src[i--];
    u  = (unsigned char) src[i--] - 128;
    YUV2RGB (y1, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
    YUV2RGB (y0, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
  }
}

void yuv411torgb (unsigned char *src, unsigned char *dest, int NumPixels) {
  register int i = (NumPixels+(NumPixels>>1))-1;
  register int j = ( NumPixels*3) -1;
  register int y0,y1,y2,y3, u, v;
  register int r, g, b;
    
  while (i > 0)
  {
    y3=(unsigned char) src[i--];
    y2=(unsigned char) src[i--];
    v =(unsigned char) src[i--] - 128;
    y1=(unsigned char) src[i--];
    y0=(unsigned char) src[i--];
    u =(unsigned char) src[i--] - 128;
	
    YUV2RGB (y3, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
    YUV2RGB (y2, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
    YUV2RGB (y1, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
    YUV2RGB (y0, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
  }
}

void yuv444torgb (unsigned char *src, unsigned char *dest, int NumPixels) {
  register int i = (NumPixels*3)-1;
  register int j = (NumPixels*3)-1;
  register int y, u, v;
  register int r, g, b;
  
  while (i > 0)
  {
      
    v  = (unsigned char) src[i--] - 128;
    y  = (unsigned char) src[i--];
    u = (unsigned char) src[i--] - 128;
           
    YUV2RGB (y, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
  }
}

Image8 bayerLinearInterp(Image8 & input,const char * mode)
{
  Image8 output(input.w,input.h,3);

  int chan_map[3]={0,1,2};
  int offs[9][2]={{0,0},
		  {-1,0},{1,0},{0,-1},{0,1},
		  {-1,-1},{-1,1},{1,-1},{1,1}};
  int starts[2][2][3];
  int ends[2][2][3];

  if(!strcmp(mode,"RGGB")||!strcmp(mode,"BGGR")){
    starts[0][0][0]=0;ends[0][0][0]=1;
    starts[1][0][0]=1;ends[1][0][0]=3;
    starts[0][1][0]=3;ends[0][1][0]=5;
    starts[1][1][0]=5;ends[1][1][0]=9;
    
    starts[0][0][1]=1;ends[0][0][1]=5;
    starts[1][0][1]=0;ends[1][0][1]=1;
    starts[0][1][1]=0;ends[0][1][1]=1;
    starts[1][1][1]=1;ends[1][1][1]=5;
    
    starts[0][0][2]=5;ends[0][0][2]=9;//was wrong with 1-5
    starts[1][0][2]=3;ends[1][0][2]=5;
    starts[0][1][2]=1;ends[0][1][2]=3;
    starts[1][1][2]=0;ends[1][1][2]=1;
    if(!strcmp(mode,"BGGR")){
      chan_map[0]=2;
      chan_map[2]=0;
    }
  }
  else if(!strcmp(mode,"GRBG")||!strcmp(mode,"GBRG")){
    starts[0][0][0]=1;ends[0][0][0]=3;
    starts[1][0][0]=0;ends[1][0][0]=1;
    starts[0][1][0]=5;ends[0][1][0]=9;//was wrong with 1-5
    starts[1][1][0]=3;ends[1][1][0]=5;
    
    starts[0][0][1]=0;ends[0][0][1]=1;
    starts[1][0][1]=1;ends[1][0][1]=5;
    starts[0][1][1]=1;ends[0][1][1]=5;
    starts[1][1][1]=0;ends[1][1][1]=1;
    
    starts[0][0][2]=3;ends[0][0][2]=5;
    starts[1][0][2]=5;ends[1][0][2]=9;
    starts[0][1][2]=0;ends[0][1][2]=1;
    starts[1][1][2]=1;ends[1][1][2]=3;
    if(!strcmp(mode,"GBRG")){
      chan_map[0]=2;
      chan_map[2]=0;
    }
  }
  for(int y=0;y<input.h;y++){
    int y2=y%2;
    for(int x=0;x<input.w;x++){
      int x2=x%2;
      for(int k=0;k<3;k++){
	int sum=0;
	int cnt=0;
	for(int h=starts[x2][y2][k];h<ends[x2][y2][k];h++){
	  int xo=x+offs[h][0];
	  int yo=y+offs[h][1];
	  if(xo>=0 && xo<input.w &&
	     yo>=0 && yo<input.h){
	    sum+=input(xo,yo);
	    cnt++;
	  }
	}
	output(x,y,chan_map[k])=sum/cnt;
      }
    }
  }
  return output;
}

Image8 bayerDownSample(Image8 & input,const char * mode)
{
  Image8 output(input.w/2,input.h/2,3);
  //default to RGGB
  int rxoff=0,gxoff1=1,gxoff2=0,bxoff=1;
  int ryoff=0,gyoff1=0,gyoff2=1,byoff=1;
  
  if(!strcmp(mode,"BGGR")){
    bxoff=0;gxoff1=1;gxoff2=0;rxoff=1;
    byoff=0;gyoff1=0;gyoff2=1;ryoff=1;
  }
  if(!strcmp(mode,"GRBG")){
    rxoff=1;gxoff1=0;gxoff2=1;bxoff=0;
    ryoff=0;gyoff1=0;gyoff2=1;byoff=1;
  }
  if(!strcmp(mode,"GBRG")){
    bxoff=1;gxoff1=0;gxoff2=1;rxoff=0;
    byoff=0;gyoff1=0;gyoff2=1;ryoff=1;
  }
  for(int y=0;y<output.h;y++){
    int y2=y*2;
    for(int x=0;x<output.w;x++){
      int x2=x*2;

      output(x,y,0)=input(x2+rxoff,y2+ryoff);
      output(x,y,1)=(((int)input(x2+gxoff1,y2+gyoff1))+
		     ((int)input(x2+gxoff2,y2+gyoff2)))/2;
      output(x,y,2)=input(x2+bxoff,y2+byoff);
    }
  }
  return output;
}

void (* func[6])(unsigned char *,unsigned char *,int)={0,yuv444torgb,uyvy2rgb,yuv411torgb,0,0};

CaptureDialog::CaptureDialog(wxWindow * parent,wxWindowID id):
  wxDialog(parent,id,wxT("Capture Options"),wxDefaultPosition,
	     wxSize(230,290)){
		
  cameraCombo=new wxComboBox(this,ID_Camera,wxT("Camera 0"));
  modeCombo=new wxComboBox(this,ID_Mode);
  formatCombo=new wxComboBox(this,ID_Format);
  nimages=new wxSpinCtrl(this,-1);
  every=new wxSpinCtrl(this,-1);
  grab=
    new wxButton(this,ID_Grab,wxT("&Grab"));
  wxButton * close=
    new wxButton(this,ID_Close,wxT("&Close"));


  bayerButton=new wxCheckBox(this,ID_BayerButton,wxT("Bayer"));//ID_BayerButton);
  bayerButton->SetValue(options.bayerEnabled);
  bayerPattern=new wxComboBox(this,ID_BayerPattern,wxT("RGGB"),wxDefaultPosition,
			      wxDefaultSize);//ID_BayerPattern);
  bayerPattern->Append(wxT("GBRG"));
  bayerPattern->Append(wxT("RGGB"));
  bayerPattern->Append(wxT("BGGR"));
  bayerPattern->Append(wxT("GRBG"));
  bayerPattern->SetSelection(options.bayerPattern);

  bayerInterpMode=new wxComboBox(this,ID_BayerInterpMode,wxT("Downsample"));
  bayerInterpMode->Append(wxT("Downsample"));
  bayerInterpMode->Append(wxT("Linear"));
  bayerInterpMode->SetSelection(options.bayerInterpolationMode);

  wxStaticBoxSizer * bayerStaticBoxSizer=new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Bayer Options")),wxVERTICAL);
  wxFlexGridSizer * bayerPanelSizer = new wxFlexGridSizer(3,2,4,4);
  bayerStaticBoxSizer->Add(bayerPanelSizer,0,wxGROW|wxEXPAND);
  
  bayerPanelSizer->Add(bayerButton,0,wxGROW);
  bayerPanelSizer->Add(new wxStaticText(this,-1,wxT("")),0,wxEXPAND);
  
  bayerPanelSizer->Add(new wxStaticText(this,-1,wxT("Pattern:"),wxDefaultPosition,
			    wxDefaultSize,wxALIGN_RIGHT),0,wxGROW);
  bayerPanelSizer->Add(bayerPattern,0,wxEXPAND);
  
  bayerPanelSizer->Add(new wxStaticText(this,-1,wxT("Interpolation:"),wxDefaultPosition,
			    wxDefaultSize,wxALIGN_RIGHT),0,wxGROW);
  bayerPanelSizer->Add(bayerInterpMode,0,wxEXPAND);
  
  nimages->SetRange(0,512);
  nimages->SetValue(options.captureImages);
    
  every->SetRange(0,10000);
  every->SetValue(options.captureDelay);
    
  wxFlexGridSizer * topsizer = new wxFlexGridSizer(3,2,4,4);
  wxFlexGridSizer * botsizer = new wxFlexGridSizer(3,2,4,4);
    
  topsizer->Add(new wxStaticText(this,-1,wxT("Camera"),wxDefaultPosition,
			      wxDefaultSize,wxALIGN_RIGHT),
	     0,wxGROW|wxALIGN_RIGHT);
  topsizer->Add(cameraCombo,0,wxEXPAND);
  topsizer->Add(new wxStaticText(this,-1,wxT("Format"),wxDefaultPosition,
			      wxDefaultSize,wxALIGN_RIGHT),
	     0,wxGROW|wxALIGN_RIGHT);
  topsizer->Add(formatCombo,0,wxEXPAND);
  topsizer->Add(new wxStaticText(this,-1,wxT("Mode "),wxDefaultPosition,
			      wxDefaultSize,wxALIGN_RIGHT),
	     0,wxGROW|wxALIGN_RIGHT);
  topsizer->Add(modeCombo,0,wxEXPAND);
  
  botsizer->Add(new wxStaticText(this,-1,wxT("#Images"),wxDefaultPosition,
			      wxDefaultSize,wxALIGN_RIGHT),
	     0,wxGROW|wxALIGN_RIGHT);
  
  botsizer->Add(nimages,0,wxEXPAND);
  botsizer->Add(new wxStaticText(this,-1,wxT("Delay"),wxDefaultPosition,
			      wxDefaultSize,wxALIGN_RIGHT)
	     ,0,wxGROW|wxALIGN_RIGHT);
  botsizer->Add(every,0,wxEXPAND);
  botsizer->Add(grab,0,wxGROW);
  botsizer->Add(close,0,wxGROW);
  botsizer->SetMinSize(330,290);
  botsizer->AddGrowableCol(1);

  wxBoxSizer * sizer=new wxBoxSizer(wxVERTICAL);
  sizer->Add(topsizer,3,wxEXPAND|wxGROW);
  sizer->Add(bayerStaticBoxSizer,3,wxEXPAND|wxGROW);
  sizer->Add(botsizer,3,wxEXPAND|wxGROW);
  SetSizer(sizer);
  //SetAutoLayout(TRUE);
  //Validate();
  started=0;
  capturing=0;
  captureCount=0;
  format7support=0;
#ifdef USING_DC1394
  startedForNode=0;
  usingDMA=1;
#endif
#ifdef USING_1394CAMERA
  //anything here?
#endif
  Init();
  CentreOnScreen();
}

/** initialize capture for current camera/format/mode.
 *  This function should be called whenever the user chanes 
 *  the camera, format or mode of capture.
 */
void CaptureDialog::CaptureDialog::initCapture(){
  wxMutexLocker lock(s_mutex);
   
  int which = cameraCombo->GetSelection();
  unsigned int format;
  unsigned int mode;
  int setDMAFailed=0;
  //avoid multiple concurrent ...
  if(which==-1)
    return;

  printf("initCapture %d/%d\n",which, cameraCombo->GetCount());

  //If the camera has been setup, shut it down
#ifdef USING_DC1394
  if(started){
    printf("Stopping transmission\n");
    dc1394_stop_iso_transmission(handle,camera_nodes[startedForNode]);
    if(usingDMA){
      dc1394_dma_release_camera(handle, &captures[startedForNode]);
    }
    else{
      dc1394_release_camera(handle, &captures[startedForNode]);
    }
    startedForNode=which;
    started=0;
  }
#endif
#ifdef USING_1394CAMERA
  if(started){
    camera.StopImageAcquisition();
    started=0;
  }
#endif  // Done shutting down the camera

  //Make sure we have a current selection, otherwise return
  if(formatCombo->GetSelection()<0){
    printf("no format selected\n");
    return;
  }

  //The format is stored in the format combo selection string
  wxString str =formatCombo->GetStringSelection();
  sscanf((const char *)str.ToAscii(),"Format %d",&format);
#ifdef USING_DC1394
  format+=FORMAT_MIN;
#endif
  //The mode is stored in the mode combo selection string
  str =modeCombo->GetStringSelection();
  sscanf((const char *)str.ToAscii(),"Mode %d",&mode);
  
#ifdef USING_DC1394
  printf("initing capture %d %d\n",format,mode);
  modeInfo=formatmodes[format-FORMAT_MIN][mode-modemins[format-FORMAT_MIN]];

  //If the format is not format7
  if(format< FORMAT_SCALABLE_IMAGE_SIZE)
  {
    //Get the supported framerates and choose the highest
    quadlet_t frates;
    int framerate;
    if(dc1394_query_supported_framerates(handle, camera_nodes[which],
					 format,mode,
					 &frates)!=DC1394_SUCCESS){
      printf("error querying supported framerates\n");
      exit(0);
    }

    for(int i=0;i<NUM_FRAMERATES;i++)
      if(frates& 1<<(31-i))
	framerate=i;
    
    printf("---------------------USING FRAMERATE %d\n",framerate);
    framerate+=FRAMERATE_MIN;

    int speed=SPEED_400;
#ifdef SPEED_800
    speed=SPEED_800;

    if(dc1394_set_operation_mode(handle, camera_nodes[which], OPERATION_MODE_1394B) != DC1394_SUCCESS){
      speed = SPEED_400;
    }
#endif

    if(usingDMA){
      if (dc1394_dma_setup_capture(handle, 
				   camera_nodes[which],  
				   which,
				   format,
				   mode,
				   speed,
				   framerate,
				   8, // number of dma buffers
				   0, // drop frames
				   strdup("/dev/video1394/0"), 
				   &captures[which])
	  != DC1394_SUCCESS){
	wxDialog * dlg = new wxMessageDialog(GetParent(),wxT("Unable to initialize DMA, falling back to RAW mode"),wxT("Error"),wxOK);
	dlg->ShowModal();
	delete dlg;
	
	setDMAFailed=1;
	usingDMA=0;
      }
    }
    if(!usingDMA ){
      if (dc1394_setup_capture(handle, 
			       camera_nodes[which],  
			       which,
			       format,
			       mode,
			       SPEED_400,
			       framerate,
			       &captures[which])
	  != DC1394_SUCCESS){
	fprintf(stderr,"Setting Camera %d failed\n", which);
	exit(1);
      }
    }
    if(dc1394_start_iso_transmission(handle,camera_nodes[which])
       != DC1394_SUCCESS)
    {
      fprintf(stderr, "Could not start ISO transmission\n");
      exit(1);
    }
    started=1;
    printf("init capture complete (camera=%d)\n",which);
  }
  //Need to handle format 7 specially
  //    else{
  //    }
#endif// USING_DC1394

#ifdef USING_1394CAMERA
  int sel;
  mode=mode-formatmodes[format][0][3];//Subtract the min
  if(sel=camera.GetNode()<0){
    return;
  }
  /*
  this has already been done.
  if(camera.InitCamera()){
    wxMessageDialog dlg(GetParent(),"Error initing camera","Init error",wxOK);
    dlg.ShowModal();
    return;
  }
  */
  //pick fastest rate for the chosen format/mode
  int rate=5;
  for (; rate>=0; rate--){
    if (camera.HasVideoFrameRate(format, mode, rate))break;
  }
  if(rate<0){
    printf("can not find rate\n");fflush(stdout);
    return;
  }
  //wxString message;
  //message.Printf("setting format:%d mode:%d rate:%d",format,mode,rate);
  //wxMessageBox(message,"Notice");
  modeInfo=formatmodes[format][mode];
  camera.SetVideoFormat(format);
  camera.SetVideoMode(mode);
  camera.SetVideoFrameRate(rate);
  camera.StartImageAcquisition();
  started=1;
#endif
}


/** Updates the modeCombo combo box based on current format and camera.
 *  Update the available modes based on the current format and camera and
 *  should be called whenever the format or camera changes.
 */
void CaptureDialog::UpdateModes(){
  //wxMutexLocker lock(s_mutex);
  int format;
  int camera=cameraCombo->GetSelection();
  modeCombo->Clear();
  if(camera==-1)return;
  
  if(camera>=cameraCombo->GetCount()){
    printf("Camera is out of bounds\n");
    return;
  }
       
  wxString fmt =formatCombo->GetStringSelection();
  sscanf((const char *)fmt.ToAscii(),"Format %d",&format);

  if(format<0)return;
#ifdef USING_DC1394
  quadlet_t modes;

  if(DC1394_SUCCESS!=
     dc1394_query_supported_modes(handle,
				  camera_nodes[camera],
				  FORMAT_MIN+format,
				  &modes)){	
    printf("error querying modes for camera %d\n",camera);
  }
  else{
    switch(format){
    case 0:
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
      for(int k=0;k<nummodes[format];k++){
	if(modes & 1<<(32-k-1)){
	  printf("%d supported\n",
		 MODE_FORMAT0_MIN+
		 k);
	  wxString name;
	  name = wxString::Format(wxT("Mode %d:%dx%d-"),
				  formatmodes[format][k][3],
				  formatmodes[format][k][0],
				  formatmodes[format][k][1])+
	    wxString::FromAscii(fmt2str[formatmodes[format][k][2]]);
	  modeCombo->Append(name);
	}
      }
      break;
    case 7:
      for(int k=0;k<nummodes[format];k++){
	if(modes & 1<<(32-k-1)){
	  printf("%d supported\n",
		 MODE_FORMAT7_MIN+
		 k);
	  wxString name;
	      
	  formatmodes[format][k][3]=MODE_FORMAT7_MIN+k;
	  unsigned int width;
	  unsigned int height;
	  if(DC1394_SUCCESS!=
	     dc1394_query_format7_image_size(handle,camera_nodes[camera],
					     MODE_FORMAT7_MIN+k,
					     &(width),
					     &(height))){
	    printf("error querying format7\n");
	  }
		     
	  formatmodes[format][k][0]=(int)width;
	  formatmodes[format][k][1]=(int)height;
	      	      
	  name = wxString::Format(wxT("Mode %d:%dx%d-"),
				  formatmodes[format][k][3],
				  formatmodes[format][k][0],
				  formatmodes[format][k][1])
	    +wxString::FromAscii(fmt2str[formatmodes[format][k][2]]);
	  modeCombo->Append(name);
	}
      }
    }
  }
#endif //USING_DC1394

#ifdef USING_1394CAMERA
  for(int mode=0;mode<8;mode++){
    if(this->camera.HasVideoMode(format, mode)){
      wxString name;
      name = wxString::Format(wxT("Mode %d:%dx%d-"),
			      formatmodes[format][mode][3],
			      formatmodes[format][mode][0],
			      formatmodes[format][mode][1])+
	wxString::FromAscii(fmt2str[formatmodes[format][mode][2]]);
      modeCombo->Append(name);
    }
  }
#endif

  if(options.cameraModeIndex<modeCombo->GetCount()){
    modeCombo->SetSelection(options.cameraModeIndex);
  }
  else{
    modeCombo->SetSelection(0);
  }
  //wxString message;
  //message.Printf("Selected mode %d",modeCombo->GetSelection());
  //wxMessageBox(message,"Message");
}
/** Update the formats that can be used with the currently selected camera.
 * Query the camera for supported formats and fill the formatCombo.
 * As as side effect the modeCombo must also be updated.
 */
void CaptureDialog::UpdateFormat(){
  {
    wxMutexLocker lock(s_mutex);
  int i=cameraCombo->GetSelection();
  formatCombo->Clear();
  format7support=0;
  if(i==-1)return;

#ifdef USING_DC1394    
  quadlet_t value;
  if(DC1394_SUCCESS!=dc1394_query_supported_formats(handle, camera_nodes[i],
						    &value)){
    printf("error querying formats for camera %d\n",i);
  }
  for(int j=0;j<=7;j++){
    if(value & 1<<(32-j-1)){
      formatCombo->Append(wxString::Format(wxT("Format %d"),j));
    }
    if(j==7){
      format7support=1;
    }
  }
#endif
#ifdef USING_1394CAMERA
  for(int j=0;j<=7;j++){
    if(camera.HasVideoFormat(j)){
      formatCombo->Append(wxString::Format(wxT("Format %d"),j));
      if(j==7)format7support=1;
    }
  }
#endif //USING_1394CAMERA

  //try to use the users default value if possible
  if(options.cameraFormatIndex<formatCombo->GetCount()){
    formatCombo->SetSelection(options.cameraFormatIndex);
  }
  else
    formatCombo->SetSelection(0);
  
  //wxString message;
  //message.Printf("Selected format %d\n",formatCombo->GetSelection());
  //wxMessageBox(message,"Message");
  
  }
  UpdateModes();
}

/** Initialize the dialog.
 */
void CaptureDialog::Init(){
  cameraCombo->Clear();
#ifdef USING_DC1394
  //Quick hack to work with all, FIXME: need to add a new combo that is the port
  //then based on the port, select all cameras on the port.
  handle = dc1394_create_handle(1);
  if(handle==NULL)
    handle = dc1394_create_handle(0);
  if (handle==NULL) {
    wxMessageDialog * dialog= new wxMessageDialog(this,
						  wxT("dc1394 handle is NULL.  Can not initialize camera capture"),
						  wxT("Error"),wxOK);
    EndModal(FALSE);
    return;
  }
  camera_nodes = dc1394_get_camera_nodes(handle, &numCameras, 1); 
  cameras=new dc1394_camerainfo[numCameras];
  captures=new dc1394_cameracapture[numCameras];
  misc_infos=new dc1394_miscinfo[numCameras];
    
  for(int i=0;i<numCameras; i++){
    dc1394_get_camera_info(handle, camera_nodes[i], &cameras[i]);
    dc1394_get_camera_misc_info(cameras[i].handle,cameras[i].id,
				&misc_infos[i]);
    dc1394_print_camera_info(&cameras[i]); 
	
    cameraCombo->Append(wxString::FromAscii(cameras[i].vendor));
    if(i==0)cameraCombo->SetSelection(0);
    //removed UpdateFormat(); from here to outside the loop.
  }
#endif

#ifdef USING_1394CAMERA
  printf("in init\n");fflush(stdout);
  if(camera.CheckLink()){
    wxMessageDialog  dlg(GetParent(),wxT("Checking Camera Link Failed"),wxT("Error"),wxOK);
    dlg.ShowModal();
    return;
  }
  int numCameras=camera.GetNumberCameras();
  if(numCameras==0){
    wxMessageDialog  dlg(GetParent(),wxT("No Firewire cameras found"),wxT("Error"),wxOK);
    dlg.ShowModal();
    return;
  }
  int selected=-1;
  for(int i=0;i<numCameras;i++){
    wxString name;
    //need to select, and maybe more to get the camera name
    if(camera.SelectCamera(i)==CAM_SUCCESS){
      camera.InitCamera();
      //These are all private now
      //camera.InquireControlRegisters();
      //camera.StatusControlRegisters();
      if(selected<0)selected=i;

      char buf[1024];
      camera.GetCameraVendor(buf, sizeof(buf));
      name.Printf("%s", buf);
    }
    else
      name.Printf("Camera %d",i);
    cameraCombo->Append(name);
    /*name.Printf("Camera %d",i);
    cameraCombo->Append(name);
    if(selected<0 && camera.SelectCamera(i)==CAM_SUCCESS){
      printf("selected camera %d\n",i);
      selected=i;
      cameraCombo->SetSelection(i);
      }*/
  }
  if(selected<0){
    wxMessageDialog dlg(GetParent(),"Error Selecting a Camera","Error",wxOK);
    dlg.ShowModal();
    return;
  }
  cameraCombo->SetSelection(selected);
  if(camera.SelectCamera(selected)!=CAM_SUCCESS){
    wxMessageDialog dlg(GetParent(),"Error Selecting Camera","Error",wxOK);
    dlg.ShowModal();
    return;
  }
  //FIXTHESE
  camera.InitCamera();
  //camera.InquireControlRegisters();
  //camera.StatusControlRegisters();
  
  //wxString str;
  //str.Printf("Selected camera %d\n",selected);
  //wxMessageBox(str,"Message");
#endif
    
  UpdateFormat();
  initCapture();
}
/** modeCombo callback.
 * Initialiazes the capture for the current mode/format/camera.
 */
void CaptureDialog::OnMode(wxCommandEvent & event){
  initCapture();
}
/** formatCombo callback.
 * The modes are updated, and capture is initialized.
 */
void CaptureDialog::OnFormat(wxCommandEvent & event){
  UpdateModes();
  initCapture();
}

/** cameraCombo callback.
 * The formats for the selected camera are updated, 
 * and capture is initialized.
 */
void CaptureDialog::OnCamera(wxCommandEvent & event){
  UpdateFormat();
  initCapture();
}

/** grab an image from the camera, and if grabbing, send to the ImageList.
 * The image is first obtained from the camera, and the converted
 * to rgb format, which is then either displayed by the GLCanvas,
 * or displayed and sent to the ImageList.
 */
void CaptureDialog::GetImageFromCamera(){
  wxMutexLocker lock(s_mutex);
  int which = cameraCombo->GetSelection();
  Image8 im(modeInfo[0],modeInfo[1],nchannels[modeInfo[2]]);

#ifdef USING_DC1394
  //printf("which is %d/%d  %d\n",which,cameraCombo->GetCount(),startedForNode);
  which = startedForNode; //This was out of bounds for some reason...
  //if(which >= cameraCombo->GetCount())return;


  if(usingDMA){
    if(dc1394_dma_single_capture(&captures[which]) != DC1394_SUCCESS){
      fprintf(stderr, "Could not capture dma frames\n");
      exit(1);
    }
    dc1394_dma_done_with_buffer(&captures[which]);	
  }
  else{	
    if(dc1394_single_capture(handle,&captures[which]) != DC1394_SUCCESS){
      fprintf(stderr, "Could not capture frame (non-dma)\n");
      exit(1);
    }
  }

  //Put it in the im data structure
  if(bytesPerChannel[modeInfo[2]]==2){
    Image16 im16(modeInfo[0],modeInfo[1],nchannels[modeInfo[2]]);
    if(func[modeInfo[2]]){
      printf("Unable to handle 16bit images with more than 1 channel\n");
    }
    else{
      memcpy(im16.data,captures[which].capture_buffer,modeInfo[0]*modeInfo[1]*
	     nchannels[modeInfo[2]]*bytesPerChannel[modeInfo[2]]);
    }
    im=im16;
  }
  else{
    if(func[modeInfo[2]]){
      func[modeInfo[2]]((unsigned char*)captures[which].capture_buffer,
			im.data, modeInfo[0]*modeInfo[1]);
    }
    else{
      memcpy(im.data,captures[which].capture_buffer,
	     modeInfo[0]*modeInfo[1]*
	     nchannels[modeInfo[2]]*bytesPerChannel[modeInfo[2]]);
    }
  }
#endif
  
#ifdef USING_1394CAMERA
  long unsigned int _width, _height;
  camera.GetVideoFrameDimensions(&_width, &_height);

  //here
  //if(im.w!=camera.m_width||im.h!=camera.m_height)
  im=Image8(_width, _height,3);
  camera.AcquireImage();
  camera.getRGB(im.data, _width*_height*3);
#endif

  if((modeInfo[2]==MONO || modeInfo[2]==MONO16)
     && bayerButton->IsChecked()){
    if(bayerInterpMode->GetStringSelection()==wxT("Downsample")){
      im=bayerDownSample(im,bayerPattern->GetStringSelection().ToAscii());
    }
    else if(bayerInterpMode->GetStringSelection()==wxT("Linear")){
      im=bayerLinearInterp(im,bayerPattern->GetStringSelection().ToAscii());
    }
  }
#if 0 //#ifdef USING_DC1394
  //Assume its a basler image
  if((modeInfo[3]==MODE_640x480_MONO ||
      modeInfo[3]==MODE_1600x1200_MONO||
      modeInfo[3]==MODE_1280x960_MONO)
     && format7support){
    if(downSampleBayer){
      //basler...
      if(modeInfo[3]==MODE_640x480_MONO){
	unsigned char * buf = (unsigned char *)captures[which].capture_buffer;
	im=Image8(modeInfo[0]/2,modeInfo[1]/2,3);
	for(int y=0;y<modeInfo[1]/2;y++){
	  int y2=y<<1;
	  
	  for(int x=0;x<modeInfo[0]/2;x++){
	    int x2=x<<1;
	    im(x,y,2)=buf[(y2+1)*modeInfo[0]+x2+1];
	    im(x,y,1)=(buf[(y2+1)*modeInfo[0]+x2]
		       +buf[(y2)*modeInfo[0]+x2+1])/2;
	    im(x,y,0)=buf[(y2)*modeInfo[0]+x2];
	  }
	}
      }
      //assume gbrg
      else{
	unsigned char * buf = (unsigned char *)captures[which].capture_buffer;
	im=Image8(modeInfo[0]/2,modeInfo[1]/2,3);
	for(int y=0;y<modeInfo[1]/2;y++){
	  int y2=y<<1;
	  
	  for(int x=0;x<modeInfo[0]/2;x++){
	    int x2=x<<1;
	    im(x,y,2)=buf[(y2)*modeInfo[0]+x2+1];
	    im(x,y,1)=(buf[(y2+1)*modeInfo[0]+x2+1]
		       +buf[(y2)*modeInfo[0]+x2])/2;
	    im(x,y,0)=buf[(y2+1)*modeInfo[0]+x2];
	  }
	}
      }
    }
    else{
      printf("No support for bayer interpolation, only downsample\n");
    }
  }
  
  else if(bytesPerChannel[modeInfo[2]]==2){
    Image16 im16(modeInfo[0],modeInfo[1],nchannels[modeInfo[2]]);
    if(func[modeInfo[2]]){
      printf("Unable to handle 16bit images with more than 1 channel\n");
    }
    else{
      memcpy(im16.data,captures[which].capture_buffer,modeInfo[0]*modeInfo[1]*
	     nchannels[modeInfo[2]]*bytesPerChannel[modeInfo[2]]);
    }
    im=im16;
  }
  else{
    if(func[modeInfo[2]]){
      func[modeInfo[2]]((unsigned char*)captures[which].capture_buffer,
			im.data, modeInfo[0]*modeInfo[1]);
    }
    else{
      memcpy(im.data,captures[which].capture_buffer,
	     modeInfo[0]*modeInfo[1]*
	     nchannels[modeInfo[2]]*bytesPerChannel[modeInfo[2]]);
    }
  }
#endif
  //set the gl windows texture
  glCanvas->UpdateTexture(im);
  glCanvas->Refresh();
	
  //Ensure that the user has added a profile for
  //the camera before continuing
  if(capturing){
    if(captureCount<nimages->GetValue()){
      static StopWatch watch;
      if(captureCount==0){
	watch.reset();
      }
      if(watch.hasElapsed(every->GetValue())||captureCount==0){
	ImageData * imd = new ImageData();
	imd->image=im;
	imd->set(HAS_IMAGE);
	wxString name;
	name.Printf(wxT("cap-%04d"),captureCount);
	if(imageList->GetString(0)==
	   wxT(" "))
	  imageList->Clear();
	imageList->Append(name,imd);
	captureCount++;
		    
	//User interface udpates
	wxString label;
	label.Printf(wxT("%d/%d"),captureCount,nimages->GetValue());
	if(nimages->GetValue()>1)
	  SetProgress(captureCount/(float)(nimages->GetValue()-1));
	else
	  SetProgress(1.0f);
	grab->SetLabel(label);
      }
    }
    else{
      capturing=0;
      captureCount=0;
      grab->Enable();
      grab->SetLabel(wxT("Grab"));
    }
  }
}

/**Grab button callback.
 * Put the captureDialog in capturing mode, and reset the
 * capture count.
 */ 
void CaptureDialog::OnGrab(wxCommandEvent & event){
  printf("about to capture\n");
  captureCount=0;

  if(captureCount==0){ //This used to be in the GetImageFromCamera function.
    static int lastw=-1;
    static int lasth=-1;
    int w=modeInfo[0];
    int h=modeInfo[1];
    
    /*
      Need to ensure that this is the same widht/height that will be used in
      GetImageFromCamera, otherwise the camera profile will be incorrectly set.
    */

#ifdef USING_1394CAMERA
    long unsigned int _width, _height;
    camera.GetVideoFrameDimensions(&_width, &_height);
    w = _width;
    h = _height;
#endif
      
    if((modeInfo[2]==MONO || modeInfo[2]==MONO16)
       && bayerButton->IsChecked()){
      if(bayerInterpMode->GetStringSelection()==wxT("Downsample")){
	w/=2;
	h/=2;
      }
    }
    printf("%d %d\n",w,h);

    wxString desiredProfile = cameraCombo->GetValue()+
      wxString::Format(wxT("-%dx%d"),w,h);
    printf("FIXME:Should also check if camera changed\n");

    if(!cameraProfiles.setCurrentProfile((const char *)desiredProfile.ToAscii())
       &&lastw!=w && lasth!=h){
      wxMessageDialog * dialog= new wxMessageDialog(0,//this,
						    wxT("Unable to find profile for camera settings.  Would you like to add the profile now?"),
						    wxT("Unknown camera"),wxYES_NO);
      printf("dhowing modal dialog\n");
      int ret=dialog->ShowModal();
      printf("dialog shown, returned %d\n", ret);
      if(ret==wxID_YES)
      {
	CameraDialog * cd =new CameraDialog(this,-1);
	cd->AddNewProfile(desiredProfile,w+h,w+h,w/2,h/2);
	cd->ShowModal();
	delete cd;
      }
      else{
	SetStatusText(wxT("Add new profile cancelled"),0);
      }
    }
    else{
      SetStatusText(desiredProfile,1);
    }
    lastw=w;
    lasth=h;
  }

  capturing=1;
  grab->Disable();
}

/**Close button callback.
 * Stop and release structures associated with the camera.
 */ 
void CaptureDialog::OnClose(wxCommandEvent & event){
#ifdef USING_DC1394
  if(started){
    dc1394_stop_iso_transmission(handle,camera_nodes[startedForNode]);
    if(usingDMA){
      dc1394_dma_release_camera(handle, &captures[startedForNode]);
    }
    else{
      dc1394_release_camera(handle, &captures[startedForNode]);
    }
    started=0;
  }
  delete camera_nodes;
  delete cameras;
  delete captures;
  delete misc_infos;
  raw1394_destroy_handle(handle);
#endif

#ifdef USING_1394CAMERA
  if(started){
    camera.StopImageAcquisition();
    started=0;
  }
#endif
  //Update the options
  options.captureImages=nimages->GetValue();
  options.captureDelay=every->GetValue();
  options.cameraFormatIndex=formatCombo->GetSelection();
  options.cameraModeIndex=modeCombo->GetSelection();
  options.bayerEnabled=bayerButton->IsChecked();
  options.bayerPattern=bayerPattern->GetSelection();
  options.bayerInterpolationMode=bayerInterpMode->GetSelection();
  
  EndModal(wxID_CANCEL);
}

/**Idle callback, grab images and display them.
 */ 
void CaptureDialog::OnIdle(wxIdleEvent & event){
  if(started){
    GetImageFromCamera();
    event.RequestMore(TRUE);
  }
}
