#include <wx/wx.h>
#include <wx/listctrl.h>
#include <wx/event.h>
#include <sys/time.h>
#include <wx/spinctrl.h>
#include "guitypes.h"
#include "image.h"

#ifndef CAPTURE_DIALOG_H
#define CAPTURE_DIALOG_H

#ifdef USING_DC1394

#include <libraw1394/raw1394.h>
#include <libdc1394/dc1394_control.h>

#endif //USING_DC1394

#ifdef USING_1394CAMERA
#include <1394Camera.h>
#endif

/** Dialog for setting up and grabbing images from a ieee1394 camera.
  Dialog for setting up and grabbing images from a ieee1394 camera.
  Currently only format 0 has been tested.

  An attempt is made to setup the capture using DMA, but will
  use raw capture if DMA setup fails.
*/
class CaptureDialog:public wxDialog{
#ifdef USING_DC1394
    raw1394handle_t handle;
    int numCameras; 
    nodeid_t* camera_nodes;
    dc1394_camerainfo *cameras;
    dc1394_cameracapture *captures;
    dc1394_miscinfo *misc_infos;
    
    ///is DMA currently being used
    int usingDMA;
    int startedForNode;
#endif
#ifdef USING_1394CAMERA
    C1394Camera camera;
#endif
    ///is format 7 supported
    int format7support;
    int * modeInfo;
    wxComboBox * cameraCombo,* modeCombo,* formatCombo;
    wxComboBox * bayerPattern;
    wxComboBox * bayerInterpMode;
    wxCheckBox * bayerButton;
    
    wxSpinCtrl * nimages,* every;
    int started;
    int capturing;
    int captureCount;
    wxButton * grab;
public:
    DECLARE_DYNAMIC_CLASS(CaptureDialog)
    CaptureDialog(wxWindow * parent=0,wxWindowID id=-1);
    void initCapture();
    void UpdateModes();
    void UpdateFormat();
    void Init();
    void OnMode(wxCommandEvent & event);
    void OnFormat(wxCommandEvent & event);
    void OnCamera(wxCommandEvent & event);
    void GetImageFromCamera();
    void OnGrab(wxCommandEvent & event);
    void OnClose(wxCommandEvent & event);
    void OnIdle(wxIdleEvent & event);
    void OnBayerButton(wxCommandEvent & event){
      int val=bayerButton->IsChecked();
      bayerPattern->Enable(val);
      bayerInterpMode->Enable(val);
    }
    DECLARE_EVENT_TABLE();
};

//empty class used for a day
#if 0 

/** Temporary, nothing when no dc1394
*/
class CaptureDialog:public wxDialog{
public:
  DECLARE_DYNAMIC_CLASS(CaptureDialog)
    CaptureDialog(wxWindow * parent=0,wxWindowID id=-1);
  void initCapture(){}
  void UpdateModes(){}
  void UpdateFormat(){}
  void Init(){}
  void OnMode(wxCommandEvent & event){}
  void OnFormat(wxCommandEvent & event){}
  void OnCamera(wxCommandEvent & event){}
  void GetImageFromCamera(){}
  void OnGrab(wxCommandEvent & event){}
  void OnClose(wxCommandEvent & event){}
  void OnIdle(wxIdleEvent & event){}
  DECLARE_EVENT_TABLE();
};
#endif

#endif
