#ifndef CHART_H
#define CHART_H

#ifdef USING_GLEE
#include <GLee.h>
#endif

#include "mc.h"
#include "image.h"
//#include "globals.h"
#include <GL/gl.h>
#include "vec.h"

template <class T>
class Array{
 public:
  T * data;
  int max;
  int n;
  int grow;
  Array(int _max=32,int _grow=8){
    n=0;
    max=_max;
    grow=_grow;
    data = new T[max];
    memset(data,0,sizeof(T)*max);
  }
  ~Array(){
    delete [] data;
    max=0;
    n=0;
  }
  int get(T & t){
    for(int i=0;i<n;i++){
      if(data[i]==t){
	return i;
      }
    }
    return -1;
  }
  void insert(T & t){
    if(n>=max){
      T * data2=new T[max+grow];
      for(int i=0;i<n;i++){
	data2[i]=data[i];
      }
      delete [] data;
      max+=grow;
      data=data2;
    }
    data[n++]=t;
  }
  int remove(T & t){
    int ind=get(t);
    if(ind==-1){
      return 0;
    }
    data[ind]=data[n-1];
    removeLast();
    return 1;
  }
  T removeLast(){
    if(n<=0){
      T t=0;
      return t;
    }
    T ret=data[--n];
    if(n<(max-2*grow)){
      T * data2=new T[max-grow];
      for(int i=0;i<n;i++){
	data2[i]=data[i];
      }
      delete [] data;
      max-=grow;
      data=data2;
    }
    return ret;
  }
  T & operator[](int index){
    if(index>=max){
      printf("index out of bounds\n");
      return data[0];
    }
    return data[index];
  }
};
//Undirected edge
class UEdge{
 public:
  unsigned i1,i2;
  UEdge(int a=0,int b=0){
    i1=a;
    i2=b;
  }
  int operator==(UEdge & other){
    return (i1==other.i1 && i2==other.i2)||
      (i2==other.i1 && i1==other.i2);
  }
};

class Chart{
 public:
  Vec3f * vert;
  _3ints * poly;
  Vec3f * norm;
  Vec2f * tc;
  int nvert;
  int npoly;
  union{
    int flags;
    void * clientData;
  };
  Chart(){
    nvert=npoly=flags=0;
    poly=0;
    tc=0;
    vert=0;
    norm=0;
  }
  Chart(int _npoly,int _nvert){
    nvert=npoly=flags=0;
    poly=0;
    tc=0;
    vert=0;
    norm=0;
    init(_npoly,_nvert);
  }
  void draw(int drawtex){
    if(drawtex && tc){
      glBegin(GL_TRIANGLES);
      for(int i=0;i<npoly;i++){
	glMultiTexCoord2fv(GL_TEXTURE0,tc[poly[i][0]].getData());
	glMultiTexCoord2fv(GL_TEXTURE1,tc[poly[i][0]].getData());
	glNormal3fv(norm[poly[i][0]].getData());
	glVertex3fv(vert[poly[i][0]].getData());
	glMultiTexCoord2fv(GL_TEXTURE0,tc[poly[i][1]].getData());
	glMultiTexCoord2fv(GL_TEXTURE1,tc[poly[i][1]].getData());
	glNormal3fv(norm[poly[i][1]].getData());
	glVertex3fv(vert[poly[i][1]].getData());	
	glMultiTexCoord2fv(GL_TEXTURE0,tc[poly[i][2]].getData());
	glMultiTexCoord2fv(GL_TEXTURE1,tc[poly[i][2]].getData());
	glNormal3fv(norm[poly[i][2]].getData());
	glVertex3fv(vert[poly[i][2]].getData());
      }
      glEnd();
    }
    else{
      
      glBegin(GL_TRIANGLES);
      for(int i=0;i<npoly;i++){
	glNormal3fv(norm[poly[i][0]].getData());
	glVertex3fv(vert[poly[i][0]].getData());
	glNormal3fv(norm[poly[i][1]].getData());
	glVertex3fv(vert[poly[i][1]].getData());
	glNormal3fv(norm[poly[i][2]].getData());
	glVertex3fv(vert[poly[i][2]].getData());
      }
      glEnd();
    }
  }
  void drawTC(){
    if(!tc)return;
    //glColor3f(1.0,1.0,1.0);
    glBegin(GL_TRIANGLES);
    for(int i=0;i<npoly;i++){
      glVertex2fv(tc[poly[i][0]].getData());
      glVertex2fv(tc[poly[i][1]].getData());
      glVertex2fv(tc[poly[i][2]].getData());
    }
    glEnd();
  }

  void setTC(Matrix & mat){
    if(npoly==0)return;
    float minx=mat.minInCol(0);
    float maxx=mat.maxInCol(0);
    float miny=mat.minInCol(1);
    float maxy=mat.maxInCol(1);

    //make sure the larger axis is the vertical axis
    if((maxx-minx)>(maxy-miny)){
      for(int i=0;i<nvert;i++){
	tc[i].y=(mat(i,0)-minx)/(maxx-minx);
	tc[i].x=(mat(i,1)-miny)/(maxy-miny);
      }
    }
    else{
      for(int i=0;i<nvert;i++){
	tc[i].x=(mat(i,0)-minx)/(maxx-minx);
	tc[i].y=(mat(i,1)-miny)/(maxy-miny);
      }
    }
  }

  double areaRatios(double & avg,double & min,double & max,
		    double & wa,double & ta){
    int usedincalc=0;
    if(!tc)return 1;
    avg=0;
    min=10e10;
    max=-10e10;
    wa=ta=0;
    for(int j=0;j<npoly;j++){
      double aw = (vert[poly[j][0]]-vert[poly[j][1]]).len();
      double bw = (vert[poly[j][0]]-vert[poly[j][2]]).len();
      double cw = (vert[poly[j][1]]-vert[poly[j][2]]).len();

      double at = (tc[poly[j][0]]-tc[poly[j][1]]).len();
      double bt = (tc[poly[j][0]]-tc[poly[j][2]]).len();
      double ct = (tc[poly[j][1]]-tc[poly[j][2]]).len();
      
      double sw=(aw+bw+cw)/2;
      double st=(at+bt+ct)/2;
      double areaw=sw*(sw-aw)*(sw-bw)*(sw-cw);
      double areat=st*(st-at)*(st-bt)*(st-ct);
      if(areaw<0)areaw=0;
      if(areat<0)areat=0;
      
      areaw=sqrt(areaw);
      areat=sqrt(areat);
      wa+=areaw;
      ta+=areat;
      if(areaw<10e-10){
	continue;
      }
      usedincalc++;
      double rat=areat/areaw;
      avg+=rat;
      if(rat<min)min=rat;
      if(rat>max)max=rat;
    }
    if(usedincalc)
      avg/=usedincalc;
    return avg;
  }
    
  void init(int np,int nv){
    release();
    npoly=np;
    nvert=nv;
    poly=new _3ints[np];
    tc=new Vec2f[nv];
    vert=new Vec3f[nv];
    norm=new Vec3f[nv];
    
    memset(poly,0,sizeof(_3ints)*np);
    memset(vert,0,sizeof(Vec3f)*nv);
    memset(norm,0,sizeof(Vec3f)*nv);
    memset(tc,0,sizeof(Vec2f)*nv);
    flags=0;
  }
  void saveobj(char * filename){
    FILE * file = fopen(filename,"wb");
    printf("you used to negate the x axis in saveobj...fix if dyna doesnt work\n");
    for(int j=0;j<nvert;j++){
      fprintf(file,"v %f %f %f\n",
	      vert[j].x,
	      vert[j].y,
	      vert[j].z);
    }
    if(tc){
      for(int j=0;j<nvert;j++){
	fprintf(file,"vt %f %f\n",
		tc[j].x,
		tc[j].y);
      }
    }
		
    int pi=1;
    for(int j=0;j<npoly;j++){
      fprintf(file,"f %d %d %d\n",
	      poly[j][0]+pi,
	      poly[j][1]+pi,
	      poly[j][2]+pi);
    }
    fclose(file);
  }
  int loadobj(char * filename){
    FILE * file=fopen(filename,"r");
    if(!file){
      printf("error opening %s\n",filename);
      return 0;
    }
    printf("opening %s\n",filename);
    int verti=0;
    int polyi=0;
    int ntvert=0;
    int tverti=0;
    char line[1024];
    while(fgets(line,1023,file)){
      if(strstr(line,"v ")==line)nvert++;
      else if(strstr(line,"vt ")==line)ntvert++;
      else if(strstr(line,"f ")==line)npoly++;
    }
    fseek(file,0,SEEK_SET);
	
    init(npoly,nvert);
    printf("reading geom %d %d\n",npoly,nvert);
    while(fgets(line,1023,file)){
      if(strstr(line,"v ")==line){
	sscanf(line+2,"%f %f %f",
	       &(vert[verti].x),
	       &(vert[verti].y),
	       &(vert[verti].z));
	verti++;
      }
      else if(strstr(line,"vt ")==line){
	sscanf(line+3,"%f %f",
	       &(tc[tverti].x),
	       &(tc[tverti].y));
		
	tverti++;
      }
      else if(strstr(line,"f ")==line){
	sscanf(line+2,"%d %d %d",
	       (poly[polyi]),
	       (poly[polyi]+1),
	       (poly[polyi]+2));
	poly[polyi][0]--;
	poly[polyi][1]--;
	poly[polyi][2]--;
	polyi++;
      }
    }
    printf("initing normals\n");
    initNormals();
    fclose(file);
    return 1;
  }
  void initNormals(){
    memset(norm,0,sizeof(Vec3f)*(nvert));
    for(int i=0;i<npoly;i++){
      Vec3f a = vert[poly[i][1]];
      Vec3f b = vert[poly[i][2]];
      a.sub(vert[poly[i][0]]);
      b.sub(vert[poly[i][0]]);
      Vec3f c=a.cross(b);
      //double len=c.normalize();

      Vec3f e0=vert[poly[i][1]]-vert[poly[i][0]];e0.normalize();
      Vec3f e1=vert[poly[i][2]]-vert[poly[i][0]];e1.normalize();
            
      double a0=acos(e0.dot(e1));
	    
      e0=vert[poly[i][0]]-vert[poly[i][1]];e0.normalize();
      e1=vert[poly[i][2]]-vert[poly[i][1]];e1.normalize();
	    
      double a1=acos(e0.dot(e1));
	    
      norm[poly[i][0]]+=(c*a0);
      norm[poly[i][1]]+=(c*a1);
      norm[poly[i][2]]+=(c*(M_PI-a0-a1));
      //c.print();
    }
    for(int i=0;i<nvert;i++){
      (norm)[i].normalize();
    }
  }
  void operator=(const Chart & c){
    release();
    init(c.npoly,c.nvert);
    memcpy(vert,c.vert,sizeof(vert[0])*c.nvert);
    memcpy(poly,c.poly,sizeof(poly[0])*c.npoly);
    memcpy(tc,c.tc,sizeof(tc[0])*c.nvert);
    memcpy(norm,c.norm,sizeof(norm[0])*c.nvert);
    clientData=c.clientData;
  }
  int getBorder(int ** border){
    Array<UEdge> array[nvert];
    Array<UEdge> bord;
    for(int i=0;i<npoly;i++){
      UEdge e1(poly[i][0],
	       poly[i][1]);
      UEdge e2(poly[i][1],
	       poly[i][2]);
      UEdge e3(poly[i][2],
	       poly[i][0]);
      array[e1.i1].insert(e1);
      array[e1.i2].insert(e1);
	
      array[e2.i1].insert(e2);
      array[e2.i2].insert(e2);

      array[e3.i1].insert(e3);
      array[e3.i2].insert(e3);
    }
    /*
      for(int i=0;i<nvert;i++){
      printf("%d:",i);
      for(int j=0;j<array[i].n;j++){
      printf("%d-%d ",array[i][j].i1,array[i][j].i2);
      }
      printf("\n");
      }*/
    for(int i=0;i<nvert;i++){
      for(int j=0;j<array[i].n;j++){
	int k;
	UEdge e1=array[i][j];
	if(i==(int)e1.i1){
	  for(k=0;k<array[i].n;k++){
	    if(k==j)continue;
	    UEdge e2=array[i][k];
	    if(e1==e2 &&
	       e1.i1==e2.i2)
	      break;
	  }
	  if(k==array[i].n)
	    bord.insert(e1);
	}
      }
    }
    //Fill up the integers
    *border = new int[bord.n*2];
    for(int i=0;i<bord.n;i++){
      UEdge e=bord[i];
      (*border)[2*i  ]=e.i1;
      (*border)[2*i+1]=e.i2;
    }
    return bord.n;
  }

  int checkForOverlap(int w,int h,int * p1=0,int * p2=0){
    
    Image8 image(w,h,1);

    glClearColor(1.0,1.0,1.0,1.0);
    glDisable(GL_DEPTH_TEST);
    glClear(GL_COLOR_BUFFER_BIT);
  
    //Draw the texture for the model first
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,1,0,1,-10,10);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearStencil(0);
    glEnable(GL_STENCIL_TEST);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
 
    glClear(GL_STENCIL_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
    glStencilFunc(GL_ALWAYS,0,~0);
    glStencilOp(GL_KEEP,GL_INCR,GL_INCR);
  
    glColor4f(0.0,0.0,0.0,0.0);
    glPushMatrix();
    double minx=10e10;
    double maxx=-10e10;
    double miny=10e10;
    double maxy=-10e10;
    for(int j=0;j<nvert;j++){
      minx=MIN(tc[j].x,minx);
      maxx=MAX(tc[j].x,maxx);
      miny=MIN(tc[j].y,miny);
      maxy=MAX(tc[j].y,maxy);
    }
    glBegin(GL_TRIANGLES);
    for(int j=0;j<npoly;j++){
      int v0=poly[j][0];
      int v1=poly[j][1];
      int v2=poly[j][2];
      glVertex2f((tc[v0].x-minx)/(maxx-minx),
		 (tc[v0].y-miny)/(maxy-miny));
      glVertex2f((tc[v1].x-minx)/(maxx-minx),
		 (tc[v1].y-miny)/(maxy-miny));
      glVertex2f((tc[v2].x-minx)/(maxx-minx),
		 (tc[v2].y-miny)/(maxy-miny));
    }
    glEnd();
    glPopMatrix();
    /*
      char name[256];
      sprintf(name,"stencil-%d.tga",i);
      image.writeTGA(name);
    */
    /*
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
      glColor4f(1.0,0.0,0.0,1.0);
      glStencilFunc(GL_LESS,1,~0);
      glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
      glBegin(GL_QUADS);
      glVertex3f(0,0,0);
      glVertex3f(1,0,0);
      glVertex3f(1,1,0);
      glVertex3f(0,1,0);
      glEnd();
      glDisable(GL_BLEND);
    */
    glReadPixels(0,0,w,h,GL_STENCIL_INDEX,GL_UNSIGNED_BYTE,image.data);
    /*char fname[64];
      sprintf(fname,"image-%d.tga",i);
      image.writeTGA(fname);
    */
    int overlap=0;
    int x,y;
    for(y=0;y<h;y++){
      for(x=0;x<w;x++){
	if(image(x,y,0)>1){
	  overlap=1;
	  break;
	}
      }
      if(x<w)break;
    }
    if(overlap && p1 && p2){
       glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

       glMatrixMode(GL_PROJECTION);
       glLoadIdentity();
       glOrtho(0,1,0,1,-10,10);
       glMatrixMode(GL_MODELVIEW);
       glLoadIdentity();

       {int err=glGetError();
       if(err)printf("%s\n",(char *)gluErrorString(err));}

       glBegin(GL_TRIANGLES);
       for(int j=0;j<npoly;j++){
	 int v0=poly[j][0];
	 int v1=poly[j][1];
	 int v2=poly[j][2];
	 float red   = ((float)(j%256))/255.0f;
	 float green = ((float)((j>>8)%256))/255.0f;
	 float blue  = ((float)((j>>16)%256))/255.0f;
	 glColor3f(red,green,blue);
	 glVertex2f((tc[v0].x-minx)/(maxx-minx),
		    (tc[v0].y-miny)/(maxy-miny));
	 glVertex2f((tc[v1].x-minx)/(maxx-minx),
		    (tc[v1].y-miny)/(maxy-miny));
	 glVertex2f((tc[v2].x-minx)/(maxx-minx),
		    (tc[v2].y-miny)/(maxy-miny));
       }
       glEnd();
       Image8 temp(w,h,3);
       glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,temp.data);
       *p1=temp(x,y,0)|(temp(x,y,1)<<8)|(temp(x,y,2)<<16);

       glBegin(GL_TRIANGLES);
       for(int j=0;j<npoly;j++){
	 if(j==*p1)continue;
	 int v0=poly[j][0];
	 int v1=poly[j][1];
	 int v2=poly[j][2];
	 float red   = ((float)(j%256))/255.0f;
	 float green = ((float)((j>>8)%256))/255.0f;
	 float blue  = ((float)((j>>16)%256))/255.0f;
	 glColor3f(red,green,blue);
	 glVertex2f((tc[v0].x-minx)/(maxx-minx),
		    (tc[v0].y-miny)/(maxy-miny));
	 glVertex2f((tc[v1].x-minx)/(maxx-minx),
		    (tc[v1].y-miny)/(maxy-miny));
	 glVertex2f((tc[v2].x-minx)/(maxx-minx),
		    (tc[v2].y-miny)/(maxy-miny));
       }
       glEnd();
       glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,temp.data);
       *p2=temp(x,y,0)|(((int)temp(x,y,1))<<8)|(((int)temp(x,y,2))<<16);
       
       if(*p1==-1 || *p2==-1){
	 printf("Chart::checkForOverlap()\n\tERROR getting overlapping polygons polygons: (%d,%d)\n",*p1,*p2);
       }
    }
    return overlap;
  }

  void release(){
      
    if(poly)delete poly;
    if(vert)delete vert;
    if(norm)delete norm;
    if(tc)delete tc;
    tc=0;
    poly=0;
    vert=0;
    norm=0;
    npoly=nvert=0;
  }
  ~Chart(){
    release();
  }
};

#endif
