#include <stdio.h>
#include <stdlib.h>
#include "clb.h"

#ifdef CLB_USE_UI_DEFINES

void writeCLB(FILE * file,ImageData * imd){
    writeCLB(file,imd->flags,
	     imd->intrinsic,
	     imd->extrinsic,
	     imd->projection,
	     imd->distortion,
	     *imd->calibInfo);
}

void writeCLB(FILE * file,int flags,
	      const Matrix & intr,
	      const Matrix & extr,
	      const Matrix & proj,
	      const Matrix & dist,
	      CalibInfo & ci){
    int len;
    int start=ftell(file);
    len=0;
    fwrite(CLB,4,1,file);
    fwrite(&len,4,1,file);

    if(flags&HAS_INTRINSIC){
	fwrite(INTR,4,1,file);len=sizeof(double)*9;
	fwrite(&len,4,1,file);
	fwrite(intr.data,sizeof(double),9,file);
    }
    if(flags&HAS_EXTRINSIC){
	fwrite(EXTR,4,1,file);len=sizeof(double)*16;
	fwrite(&len,4,1,file);
	fwrite(extr.data,sizeof(double),16,file);
    }
    if(flags&HAS_PROJECTION){
	fwrite(PROJ,4,1,file);len=sizeof(double)*12;
	fwrite(&len,4,1,file);
	fwrite(proj.data,sizeof(double),12,file);
    }
    if(flags&HAS_DISTORTION){
	fwrite(DIST,4,1,file);len=sizeof(double)*5;
	fwrite(&len,4,1,file);
	fwrite(dist.data,sizeof(double),5,file);
    }
    if(flags&HAS_CALIBINFO){
	fwrite(CINF,4,1,file);
	len=4+4;//ncenters+nlines;
	len+=(sizeof(double)*(2+2*ci.ncenters+1));
	len+=(sizeof(int)*(ci.ncenters+5*ci.nlines));
	fwrite(&len,4,1,file);
	fwrite(&(ci.ncenters),4,1,file);
	fwrite(&(ci.nlines),4,1,file);
	fwrite((ci.center),sizeof(double),2,file);
	fwrite((ci.centers),sizeof(double),
	       2*ci.ncenters,file);
	fwrite((ci.sizes),sizeof(int),
	       ci.ncenters,file);
	for(int i=0;i<ci.nlines;i++)
	    fwrite(ci.lines[i],sizeof(int),
		   4,file);
	fwrite(ci.labels,sizeof(int),
	       ci.nlines,file);
	fwrite(&ci.residual,sizeof(double),1,file);
	
    }
    int now=(unsigned int)ftell(file);
    //printf("now is %d\n",now);
    fseek(file,start+4,SEEK_SET);
    len=now-8;
    //printf("writing len %d\n",len);
    fwrite(&len,4,1,file);
    fseek(file,now,SEEK_SET);
}

int readCLB(FILE * file,ImageData * imd){
    int flags=readCLB(file,
		      imd->intrinsic,
		      imd->extrinsic,
		      imd->projection,
		      imd->distortion,
		      *(imd->calibInfo));
    return flags;
}

int readCLB(FILE * file,
	    Matrix & intr,
	    Matrix & extr,
	    Matrix & proj,
	    Matrix & dist,
	    CalibInfo & ci)
{
    char tag[5];
    unsigned int flags=0;
    unsigned int len;
    unsigned int fileSize;
    fread(tag,4,1,file);
    fread(&fileSize,4,1,file);
    tag[4]=0;
    if(!TAGEQUAL(tag,CLB)){
      //printf("not valid .clb file, first tag is %s\n",tag);
      return 0;
    }
    while(1){
	int nread=fread(tag,4,1,file);
	//printf("nread %d\n",nread);
	if(nread!=1)break;
	fread(&len,4,1,file);
	if(nread!=1)break;
	tag[4]=0;
	//printf("Got %s, len %d\n",tag,len);
	
	if(TAGEQUAL(tag,INTR)){
	    flags|=HAS_INTRINSIC;
	    intr=Matrix(3,3);
	    fread(intr.data,sizeof(double),9,file);
	}
	else if(TAGEQUAL(tag,EXTR)){
	    flags|=HAS_EXTRINSIC;
	    extr=Matrix(4,4);
	    fread(extr.data,sizeof(double),16,file);
	}
	else if(TAGEQUAL(tag,PROJ)){
	    flags|=HAS_PROJECTION;
	    proj=Matrix(3,4);
	    fread(proj.data,sizeof(double),12,file);
	}
	else if(TAGEQUAL(tag,CINF)){
	    flags|=HAS_CALIBINFO;
	    ci.release();
	    fread(&(ci.ncenters),sizeof(int),1,file);
	    fread(&(ci.nlines),sizeof(int),1,file);
	    //printf("%d %d\n",ci.ncenters,ci.nlines);
	    ci.centers = new double[2*ci.ncenters];
	    ci.lines   = new int*[ci.nlines];
	    ci.sizes   = new int[ci.ncenters];
	    ci.labels  = new int[ci.nlines];

	    fread((ci.center),sizeof(double),2,file);
	    fread((ci.centers),sizeof(double),
		  2*ci.ncenters,file);
	    fread((ci.sizes),sizeof(int),
		  ci.ncenters,file);
	    for(int i=0;i<ci.nlines;i++){
		ci.lines[i]=new int[4];
		fread(ci.lines[i],sizeof(int),
		       4,file);
	    }
	    fread(ci.labels,sizeof(int),
		   ci.nlines,file);
	    fread(&ci.residual,sizeof(double),1,file);
	}
	else if(TAGEQUAL(tag,DIST)){
	    flags|=HAS_DISTORTION;
	    dist=Matrix(5,1);
	    fread(dist.data,sizeof(double),5,file);
	}
	else {
	    printf(".....unknown tag %s\n",tag);
	}
    }
    /*if(ftell(file)-8!=fileSize){
	printf("length tag %d is no good, should be %d\n",fileSize,
	       ftell(file)-8);
	       }*/
    return flags;
}
#endif

void writeCLB(FILE * file,
	      double * intr,
	      double * extr,
	      double * proj,
	      double * dist){
    int len;
    int start=ftell(file);
    len=0;
    fwrite(CLB,4,1,file);
    fwrite(&len,4,1,file);

    if(intr){
	fwrite(INTR,4,1,file);len=sizeof(double)*9;
	fwrite(&len,4,1,file);
	fwrite(intr,sizeof(double),9,file);
    }
    if(extr){
	fwrite(EXTR,4,1,file);len=sizeof(double)*16;
	fwrite(&len,4,1,file);
	fwrite(extr,sizeof(double),16,file);
    }
    if(proj){
	fwrite(PROJ,4,1,file);len=sizeof(double)*12;
	fwrite(&len,4,1,file);
	fwrite(proj,sizeof(double),12,file);
    }
    if(dist){
	fwrite(DIST,4,1,file);len=sizeof(double)*5;
	fwrite(&len,4,1,file);
	fwrite(dist,sizeof(double),5,file);
    }
    int now=(unsigned int)ftell(file);
    //printf("now is %d\n",now);
    fseek(file,start+4,SEEK_SET);
    len=now-8;
    //printf("writing len %d\n",len);
    fwrite(&len,4,1,file);
    fseek(file,now,SEEK_SET);
}

//The return value of this is slightly different
//return flags:0x1 iff file has intrinsic and extrinsic
//             0x2 iff file has projection
//             0x4 iff file has distortion
int readCLB(FILE * file,
	    double * intr,
	    double * extr,
	    double * proj,double * dist)
{
  char tag[5];
  unsigned int len;
  unsigned int fileSize;
  int gotextr=0;
  int gotintr=0;
  unsigned int flags=0;
  fread(tag,4,1,file);
  fread(&fileSize,4,1,file);
  tag[4]=0;
  if(!TAGEQUAL(tag,CLB)){
    //printf("not valid .clb file, first tag is %s\n",tag);
    return 0;
  }
  while(1){
    int nread=fread(tag,4,1,file);
    //printf("nread %d\n",nread);
    if(nread!=1)break;
    fread(&len,4,1,file);
    if(nread!=1)break;
    tag[4]=0;
    //printf("Got %s, len %d\n",tag,len);
    
    if(TAGEQUAL(tag,INTR)){
      gotintr=1;
      fread(intr,sizeof(double),9,file);
    }
    else if(TAGEQUAL(tag,EXTR)){
      gotextr=1;
      fread(extr,sizeof(double),16,file);
    }
    else if(proj && TAGEQUAL(tag,PROJ)){
      flags|=0x2;
      fread(proj,sizeof(double),12,file);
    }
    else if(dist && TAGEQUAL(tag,DIST)){
      flags|=0x4;
      fread(dist,sizeof(double),5,file);
    }
    else {
      char * buf = new char[len];
      if(len!=fread(buf,1,len,file)){
	delete [] buf;
	return 0;
      }
      delete [] buf;
      //printf(".....unknown tag %s,%d\n",tag,len);
    }
  }
  /*if(ftell(file)-8!=fileSize){
    printf("length tag %d is no good, should be %d\n",fileSize,
    ftell(file)-8);
    }*/
  return (gotintr&&gotextr)|flags;
}

#ifdef CLB_MAIN
int main(int ac,char ** av)
{
    if(ac<2){
	return 0;
    }
    FILE * file = fopen(av[1],"rb");
    Matrix intr,extr,proj,dist;
    
    CalibInfo * ci =new CalibInfo();  
    int flags=readCLB(file,intr,
	    extr,
	    proj,
	    dist,
	    *ci);

    if(flags&HAS_INTRINSIC)
      intr.printMatlab("intrinsic");
    if(flags&HAS_EXTRINSIC)
      extr.printMatlab("extrinsic");
    if(flags&HAS_DISTORTION)
      dist.printMatlab("dist");
    if(flags&HAS_PROJECTION)
      proj.printMatlab("proj");
    
    fclose(file);
}
#endif

#ifdef CLB_MAIN2
int main(int ac,char ** av)
{
    if(ac<2){
	return 0;
    }
    FILE * file = fopen(av[1],"rb");
    double intr[9];
    double extr[16];
    
    int res=readCLB(file,intr,
		     extr);
    fclose(file);
    if(res){
      printf("Success reading %s\n",av[1]);
      printf("intrinsics = \n");
      for(int i=0;i<3;i++){
	for(int j=0;j<3;j++){
	  printf("%f ",intr[3*i+j]);
	}
	printf("\n");
      }
      printf("extrinsics = \n");
      for(int i=0;i<4;i++){
	for(int j=0;j<4;j++){
	  printf("%f ",extr[4*i+j]);
	}
	printf("\n");
      }
    }
    else{
      printf("File is either not a calibration file, or does not contain intrinsic and extrinsic calibration data\n");
    }
}
#endif
