/*
  clb.h/clb.cc
  
  Within are the functions for reading and writing
  a .clb file.  The functions read (write) 
  all the camera parameters determined by the flags. 
  
  On reading CalibInfo, the current calib info (if any) 
  released and memory is allocated for the new info.
 */
#ifndef CLB_H
#define CLB_H

#ifdef CLB_USE_UI_DEFINES
#include "imagedata.h"
#include "seg/newseg.h"
#include "matrix.h"
#endif

#define CLB  "CLB "
#define EXTR "EXTR"
#define INTR "INTR"
#define PROJ "PROJ"
#define CINF "CINF"
#define DIST "DIST"

#define TAGEQUAL(T1,T2) ((unsigned int)(*T1)==(unsigned int)(*T2))

#ifdef CLB_USE_UI_DEFINES

void writeCLB(FILE * file,ImageData * id);
void writeCLB(FILE * file,int flags,
	      const Matrix & intr,
	      const Matrix & extr,
	      const Matrix & proj,
	      const Matrix & dist,
	      CalibInfo & ci);
int readCLB(FILE * file,ImageData * id);
int readCLB(FILE * file,
	    Matrix & intr,
	    Matrix & extr,
	    Matrix & proj,
	    Matrix & dist,
	    CalibInfo & ci);
#endif

//extr must be 16 doubles
//intr must be 9 doubles
//proj either null or 12 doubles
//dist either null or 5 doubles
//both are row major
//
//The return value of this is slightly different
//returns or'ed bits:0x1 iff file has intrinsic and extrinsic
//                   0x2 iff file has projection
//                   0x4 iff file has distortion
int readCLB(FILE * file,
	    double * intr,
	    double * extr,
	    double * proj=0,
	    double * dist=0);

void writeCLB(FILE * file,
	      double * intr=0,
	      double * extr=0,
	      double * proj=0,
	      double * dist=0);

#endif
