/*
  drawareas.cc
  
  Loads the geometry from the filename in command 
  line argument 1 obtains the image used for
  stretching, and writes it to the image areas.tga
*/ 
  
#ifdef USING_GLEE
#include <GLee.h>
#endif
#include <GL/glut.h>
#include <GL/gl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vec.h"
#include "chart.h"
#include "image.h"

Chart chart;
Image8 image;

void drawAreas(Chart & m)
{
  float * areas3d;
  float * areas2d;
  float maxratio=-100000;
  float minratio=1000000;
  int i=0;
  
  areas3d = new float[m.npoly];
  areas2d = new float[m.npoly];
  
  while(i<m.npoly){
    Vec3f a,b,c;
    Vec2f a2,b2,c2;
    float la,lb,lc,s;
    a=m.vert[m.poly[i][0]]-m.vert[m.poly[i][1]];
    b=m.vert[m.poly[i][1]]-m.vert[m.poly[i][2]];
    c=m.vert[m.poly[i][0]]-m.vert[m.poly[i][2]];
    la = a.len();
    lb = b.len();
    lc = c.len();
    s=(la+lb+lc)/2;
    areas3d[i]=sqrt(s*(s-la)*(s-lb)*(s-lc));
    
    a2=m.tc[m.poly[i][0]]-m.tc[m.poly[i][1]];
    b2=m.tc[m.poly[i][1]]-m.tc[m.poly[i][2]];
    c2=m.tc[m.poly[i][0]]-m.tc[m.poly[i][2]];
    la = a2.len();
    lb = b2.len();
    lc = c2.len();
    s=(la+lb+lc)/2;
    areas2d[i]=sqrt(s*(s-la)*(s-lb)*(s-lc));
    
    if((areas2d[i]/areas3d[i])<minratio)
      minratio=areas2d[i]/areas3d[i];
    if((areas2d[i]/areas3d[i])>maxratio)
      maxratio=areas2d[i]/areas3d[i];
    i++;
  }
  
  
  glColor3f(1.0,1.0,1.0);
  
  i=0;
  
  /*
    Anything that is small in the texture but big in the real 
    world will have a small ratio, where triangles that are big 
    in the texture but small in the real world will have a large ratio.
    So anything with a large ratio should be mapped to a small number
  */
  glBegin(GL_TRIANGLES);
  while(i<m.npoly){
    float c = 1.0-(areas2d[i]/areas3d[i]-minratio)/(maxratio-minratio)*0.3;
    glColor3f(c,c,c);
    glVertex2fv(m.tc[m.poly[i][0]].data);
    glVertex2fv(m.tc[m.poly[i][1]].data);
    glVertex2fv(m.tc[m.poly[i][2]].data);
    i++;
  }
  glEnd();
  
  printf("ratios from %f to %f\n",minratio,maxratio);
  
  delete areas2d;
  delete areas3d;
}

void display()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glViewport(0,0,256,256);
  gluOrtho2D(0,1,0,1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glClearColor(0,0,0,0);
  glClear(GL_COLOR_BUFFER_BIT);
  
  drawAreas(chart);
  image=Image8(256,256,1);
  glReadPixels(0,0,256,256,GL_LUMINANCE,GL_UNSIGNED_BYTE,image.data);
  image.writeTGA("areas.tga");

  glutSwapBuffers();
  exit(0);
}
int main(int ac,char * av[]){
 
  glutInit(&ac, av);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);// | GLUT_ALPHA);
  glutInitWindowSize(256,256);
  glutCreateWindow("demo");
  glutDisplayFunc(display);

  chart.loadobj(av[1]);

  glutMainLoop();
}
