#include "glcanvas.h"
#include "dynatex.h"
#include "chart.h"
#include "vec.h"
#include <wx/process.h>

using namespace CapguiGlobals;

extern Vec3f scale;
extern Vec3f trans;
extern Chart * charts;
extern int ncharts;
extern wxListBox * imageList;
extern wxListBox * calibImageList;
extern wxListBox * segImageList;

#define DRAW_UNIQUE_COLORS

IMPLEMENT_DYNAMIC_CLASS(GLCanvas,wxGLCanvas)

//GLCanvas events
BEGIN_EVENT_TABLE(GLCanvas, wxGLCanvas)
  EVT_KEY_DOWN(GLCanvas::OnKeyDown )
  EVT_KEY_UP( GLCanvas::OnKeyUp )
  EVT_SIZE(GLCanvas::OnSize)
  EVT_PAINT(GLCanvas::OnPaint)
  EVT_ERASE_BACKGROUND(GLCanvas::OnEraseBackground)
  EVT_ENTER_WINDOW( GLCanvas::OnEnterWindow )
  EVT_MOUSE_EVENTS(GLCanvas::OnMouse)
  EVT_IDLE(GLCanvas::OnIdle)
#ifdef WIN32
  EVT_MENU(ID_DrawTexture,GLCanvas::OnDrawTexture)
  EVT_MENU(ID_ProjectTexture,GLCanvas::OnProjectTexture)
  EVT_MENU(ID_DrawGrid,GLCanvas::OnDrawGrid)
  EVT_MENU(ID_Wireframe,GLCanvas::OnWireframe)
  EVT_MENU(ID_DrawNormals,GLCanvas::OnDrawNormals)
  EVT_MENU(ID_DrawCameras,GLCanvas::OnDrawCameras)
#endif
END_EVENT_TABLE()

void GLCanvas::drawChecks()
{
  static GLuint checks=0;
  static int count=0;
  if(count==0){
    glGenTextures(1,&checks);
    Image8 checksim(2,2,3);
    
    //draw some checks
    for(int i=0;i<2;i++){
      for(int j=0;j<2;j++){
	unsigned char col=(0.7*255.0);
	if((i+j)%2==0)col=(0.8*255.0);
	checksim(i,j,0)=col;
	checksim(i,j,1)=col;
	checksim(i,j,2)=col;
      }
    }
    glBindTexture(GL_TEXTURE_2D,checks);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
    checksim.initTexture();
    count=1;
  } 
  
  float tcx=(int)(16.0*((double)imagew)/((double)imageh)+0.5);
  float tcy=16;
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,checks);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

  glColor3f(1.0,1.0,1.0);
  double imaspect=((double)imagew)/((double)imageh);
  glBegin(GL_QUADS);
  glTexCoord2f(0,0);  glVertex2f(-imaspect,-1);
  glTexCoord2f(tcx,0); glVertex2f(imaspect,-1);
  glTexCoord2f(tcx,tcy);glVertex2f(imaspect,1);
  glTexCoord2f(0,tcy); glVertex2f(-imaspect,1);
  glEnd();
  
  //FIXME: this texture never gets delted
  //glDeleteTextures(1,&checks);
}

void GLCanvas::imageModePaint(wxPaintEvent & event)
{
  glClearColor(0.6,0.6,0.6,1.0);
  glClear(GL_COLOR_BUFFER_BIT);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  double orthw=((double)w)/((double)h);
  double orthh=1.0;
  double imaspect=((double)imagew)/((double)imageh);
  if(orthw<imaspect){
    orthw=imaspect;
    orthh=orthw*((double)h)/((double)w);
  }
  orthw*=1.05;
  orthh*=1.05;
  glOrtho(-orthw,orthw,orthh,-orthh,-10000,10000);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glDisable(GL_BLEND);
  if(hasAlpha)
    drawChecks();

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tex);
  glColor3f(1.0,1.0,1.0);
  
  glBegin(GL_QUADS);
  glTexCoord2f(0,0);glVertex2f(-imaspect,-1);
  glTexCoord2f(xsize,0);glVertex2f(imaspect,-1);
  glTexCoord2f(xsize,ysize);glVertex2f(imaspect,1);
  glTexCoord2f(0,ysize);glVertex2f(-imaspect,1);
  glEnd();
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  
  if(calibInfo){
    glMatrixMode(GL_PROJECTION);
    //glLoadIdentity();
    //glOrtho(0,imagew,imageh,0,-100,100);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(-imaspect,-1,0);//center the image
    glScalef(imaspect*2.0/((double)imagew),2.0/((double)imageh),1.0);
    drawCalibInfo();

    double radii[] = {300,300,300,300};
    int which=0;
    double sr=10;
    radii[0]-=sr*12;
    radii[1]-=sr*8;
    radii[2]-=sr*4;
    radii[3]-=sr*0;
   
    Matrix worldp(4,4*14);
    worldp.setAll(1);
    for(which=0;which<14;which++){
      double angle = (which)*2.0*M_PI/14.0;
      for(int j=0;j<4;j++){
	double x=radii[j]*cos(angle);
	double y=radii[j]*sin(angle);
	worldp(0,4*which+j)=((double)x/300.0);
	worldp(1,4*which+j)=0;//-150;
	worldp(2,4*which+j)=((double)y/300.0);
      }
    }	
    /*
    printf("%dx%d %dx%d %dx%d\n",
	   lastImageData->intrinsic.m,
	   lastImageData->intrinsic.n,
	   lastImageData->projection.m,
	   lastImageData->projection.n,
	   lastImageData->extrinsic.m,
	   lastImageData->extrinsic.n);
    Matrix projpts=
      lastImageData->intrinsic*
      lastImageData->projection*
      lastImageData->extrinsic*worldp;
    glBegin(GL_POINTS);
    glColor3f(1.0,1.0,0.0);
    for(int j=0;j<projpts.n;j++){
      printf("%f %f\n",projpts(0,j)/projpts(2,j),
	     projpts(0,j)/projpts(2,j));
      glVertex2f(projpts(0,j)/projpts(2,j),
		 projpts(1,j)/projpts(2,j));
    }
    glEnd();
    */
  }
}


void GLCanvas::OnIdle(wxIdleEvent & event){
  if(mode==GL_CANVAS_MODEL){
      event.RequestMore(FALSE);
  }
  else if(mode==GL_CANVAS_GETTEX){
    event.RequestMore(TRUE);
    Refresh();
  }
  else
    event.RequestMore(FALSE);
    
}

void GLCanvas::imageModeOnMouse(wxMouseEvent & event){
}

void GLCanvas::gettexModePaint(wxPaintEvent & event){
  printf("in gettexModePaint\n");

  printErr("gettexModePaint begin");
  char * str=getenv("CAPGUI_HWWARP");
  if(str && atoi(str))
    getTexturesHW(intrinsics,projection,extrinsics,
		  images,
		  charts,ncharts,
		  scale,trans,
		  texImages,occluded,texw[0],texh[0],index,index+1,nimages);
  else
    getTextures(intrinsics,projection,extrinsics,
		images,
		charts,ncharts,
		scale,trans,
		texImages,occluded,texw[0],texh[0],index,index+1);
  index++;
  if(nimages)
    SetProgress(index/(double)nimages);
  if(index>=nimages){
    delete [] images; //Don't need these anymore

    initDynaTex(where,
		charts,ncharts,
		scale,trans,
		extrinsics,
		texImages,occluded,texw[0],texh[0],
		nimages,nreig,texw,texh,lutw,luth,flags);
    
    //delete [] texImages;
    //delete [] occluded;
    delete [] intrinsics;
    delete [] extrinsics;
    SetMode(lastMode);

#ifdef WIN32
    //wxString str = wxString::FromAscii(where);
    //str.Remove(str.Index("/data0"));
    //wxMessageBox(str, wxT("Opening Renderer with"));
    //wxProcess * renderer =
    //wxProcess::Open(wxString::Format("yuv.exe"));//+str);

    //Get rid of the "data0"
    char * copy = new char[strlen(where)+1];
    strcpy(copy,  where);
    if(strlen(where)>=6)
      copy[strlen(where)-6]=0;
    wxExecute(wxString::Format("yuv.exe \"%s\"", copy));
    delete [] copy;

    //wx-widgets says not to delete this.
#endif
  }
  Refresh();
}

void GLCanvas::setChartFlag(Chart & chart){
  int p1,p2;
  chart.flags=chart.checkForOverlap(32,32,&p1,&p2);
  if(chart.flags){
    if(p1>=0 && p2>=0){
      chart.flags=(p1<<16)|p2;
    }
    printf("polys %d %d\n",p1,p2);
  }
}


void GLCanvas::setChartFlags(){
  for(int i=0;i<ncharts;i++){
    setChartFlag(charts[i]);
  }
}

void GLCanvas::textureModePaint(wxPaintEvent & event){
  int selected=GetSelectedChart();
  printErr("textureModePaint");
  glClearColor(1.0,1.0,1.0,1.0);
  glColor3f(1.0,1.0,1.0);
  glDisable(GL_DEPTH_TEST);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  
  //Draw the texture for the model first
  glViewport(0,0,256,256);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,1,0,1,-10,10);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  int drawCheck=modelModeMenu->IsChecked(ID_DrawTexture);
  if(drawCheck){
    //Draw a checkerboard texture
    unsigned char * ct = new unsigned char[256*256*3];
    for(int y=0;y<256;y++){
      for(int x=0;x<256;x++){
	int v=(((x>>2)+(y>>2))%2==0)?228:64;
	ct[(y*256+x)*3  ]=v;
	ct[(y*256+x)*3+1]=v;
	ct[(y*256+x)*3+2]=v;
      }
    }
    glRasterPos2i(0,0);
    glDrawPixels(256,256,GL_RGB,GL_UNSIGNED_BYTE,ct);
    delete [] ct;
    //end draw checkerboard
  }
  
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glClearStencil(0);
  glClear(GL_STENCIL_BUFFER_BIT);
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS,0,~0);
  glStencilOp(GL_KEEP,GL_INCR,GL_INCR);
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  for(int i=0;i<ncharts;i++){
#ifdef DRAW_UNIQUE_COLORS
    float fi=(float)i/(float)(ncharts-1);
    glColor4f(fi,1-fi,fi,0.4*((float)(!drawCheck)));
#else
    glColor4f(0.2,0.2,0.2,0.4*((float)!drawCheck));
#endif
    glPushMatrix();
    if(selected==i){
      //#ifndef DRAW_UNIQUE_COLORS
      glColor4f(0.4,0.9,0.4,0.4);
      //#endif
      glTranslatef(xpos,ypos,0);
      glTranslatef(centroidx,centroidy,0);
      glRotatef(rot,0,0,1);
      glTranslatef(-centroidx,-centroidy,0);
    }
    charts[i].drawTC();
    glPopMatrix();
  }
  
  glEnable(GL_STENCIL_TEST);
  glColor4f(1.0,0.0,0.0,0.8);
  glStencilFunc(GL_LESS,1,~0);
  glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
  glBegin(GL_QUADS);
  glVertex3f(0,0,0);
  glVertex3f(1,0,0);
  glVertex3f(1,1,0);
  glVertex3f(0,1,0);
  glEnd();
  glDisable(GL_BLEND);
  glDisable(GL_STENCIL_TEST);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,modelTex);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB,0,0,256,256,0);
  glDisable(GL_TEXTURE_2D);
  //Done drawing texture for model

  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-0.1,1.1,-0.1,1.1,-10,10);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glColor3f(0.0,0.0,0.0);
  glBegin(GL_LINE_LOOP);
  glVertex2f(0.0,0.0);
  glVertex2f(1.0,0.0);
  glVertex2f(1.0,1.0);
  glVertex2f(0.0,1.0);
  glEnd();

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glClearStencil(0);
  glClear(GL_STENCIL_BUFFER_BIT);
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS,0,~0);
  glStencilOp(GL_KEEP,GL_INCR,GL_INCR);
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  for(int i=0;i<ncharts;i++){
#ifdef DRAW_UNIQUE_COLORS
    float fi=(float)i/(float)(ncharts-1);
    glColor4f(fi,1-fi,fi,0.4);
#else
    glColor4f(0.2,0.2,0.2,0.4);
#endif
    glPushMatrix();
    if(selected==i){
      //#ifndef DRAW_UNIQUE_COLORS
      glColor4f(0.4,0.9,0.4,0.4);
      //#endif
      glTranslatef(xpos,ypos,0);
      glTranslatef(centroidx,centroidy,0);
      glRotatef(rot,0,0,1);
      glTranslatef(-centroidx,-centroidy,0);
    }
    glEnable(GL_STENCIL_TEST);
    charts[i].drawTC();
    glDisable(GL_STENCIL_TEST);
    
    glColor3f(0.0,0.0,1.0);
    if(charts[i].flags && (charts[i].flags>>16)<charts[i].npoly &&
       (charts[i].flags&0xFFFF) < charts[i].npoly){
      glBegin(GL_TRIANGLES);
      int p1=charts[i].flags>>16;
      int p2=charts[i].flags&0xFFFF;
      glVertex2fv(charts[i].tc[charts[i].poly[p1][0]].data);
      glVertex2fv(charts[i].tc[charts[i].poly[p1][1]].data);
      glVertex2fv(charts[i].tc[charts[i].poly[p1][2]].data);
      
      glVertex2fv(charts[i].tc[charts[i].poly[p2][0]].data);
      glVertex2fv(charts[i].tc[charts[i].poly[p2][1]].data);
      glVertex2fv(charts[i].tc[charts[i].poly[p2][2]].data);
      glEnd();
    }

    glLineWidth(1);
    glColor4f(charts[i].flags,0.0,0.0,1.0);
    //#ifndef DRAW_UNIQUE_COLORS
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    charts[i].drawTC();
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    //#endif
    glLineWidth(1);
    glPopMatrix();

    if(selected==i){
      if(fabs(rot)>0.01){
	glPushMatrix();
	glColor4f(0.0,0.1,1.0,1.0);
	glTranslatef(-0.1,1.1,0);
	glScalef(1.2/(double)w,-1.2/(double)h,1.0);
	glBegin(GL_LINES);
	glVertex2f(curpt[0],curpt[1]);
	glVertex2f(selpt[0],selpt[1]);
	glEnd();

	glTranslatef( selpt[0], selpt[1],0);
	glRotatef(rot,0,0,1);
	glTranslatef(-selpt[0],-selpt[1],0);
	glBegin(GL_LINES);
	glVertex2f(curpt[0],curpt[1]);
	glVertex2f(selpt[0],selpt[1]);
	glEnd();

	glPopMatrix();
      }
    }
  }
  glEnable(GL_STENCIL_TEST);
  glColor4f(1.0,0.0,0.0,0.8);
  glStencilFunc(GL_LESS,1,~0);
  glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
  glBegin(GL_QUADS);
  glVertex3f(0,0,0);
  glVertex3f(1,0,0);
  glVertex3f(1,1,0);
  glVertex3f(0,1,0);
  glEnd();
  
  glDisable(GL_STENCIL_TEST);
  glDisable(GL_BLEND);
}

void GLCanvas::textureModeOnMouse(wxMouseEvent & event){
  static double lastAngle=0;
  curpt[0]=event.GetX();
  curpt[1]=event.GetY();
  if(event.LeftDown() ||
     event.RightDown()){
    int selected=0;
    xpos=ypos=rot=0;
    selpt[0]=event.GetX();
    selpt[1]=event.GetY();

    SetCurrent();
    GLuint selbuf[1024];
    int vp[4];

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
    glSelectBuffer(1024,
		   selbuf);	
    glRenderMode(GL_SELECT);
	
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glGetIntegerv(GL_VIEWPORT,vp);
    gluPickMatrix(event.GetX()-1,vp[3]-(event.GetY()-1),
		  2,2,vp);
    glOrtho(-0.1,1.1,-0.1,1.1,-10,10);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glInitNames();

    for(int i=0;i<ncharts;i++){
      glColor3f(0.4,0.4,0.4);
      glPushName(i);
	   
      glPushMatrix();
      glTranslatef(xpos,ypos,0);
      glRotatef(rot,0,0,1);
      charts[i].drawTC();
	   
      glPopMatrix();
      glPopName();
    }
    selected=-1;
    unsigned nsel=glRenderMode(GL_RENDER);

    GLuint * selbufptr = selbuf;
    for(int i=0;i<nsel;i++){
      unsigned int nnames=*selbufptr;
      double minz,maxz;
      unsigned int max=0xFFFFFFFF;
      selbufptr++;
      //printf("%d has nnames:%d\n",i,nnames);
      minz=((double)selbufptr[0])/((double)max);
      maxz=((double)selbufptr[1])/((double)max);
      selbufptr+=2;
      while(nnames>0){
	selected=*selbufptr;
	selbufptr++;
	nnames--;
      }
    }
    //special actions for rotate tool(right button)
    if(event.RightDown()){
      lastAngle=0;
      if(selected>=0){
	centroidx=0;
	centroidy=0;
	for(int i=0;i<charts[selected].nvert;i++){
	  centroidx+=charts[selected].tc[i].x;
	  centroidy+=charts[selected].tc[i].y;
	}
	centroidx/=charts[selected].nvert;
	centroidy/=charts[selected].nvert;
      }

    }
    SetSelectedChart(selected);
  }
  if(event.LeftIsDown() && GetSelectedChart()>=0){	
    int selected=GetSelectedChart();	
    xpos+=(event.GetX()-selpt[0])*1.2/(double)w;
    ypos-=(event.GetY()-selpt[1])*1.2/(double)h;
    selpt[0]=event.GetX();
    selpt[1]=event.GetY();
  }
  else if(event.RightIsDown() && GetSelectedChart()>=0){
    int selected=GetSelectedChart();
    double angle=-atan2(event.GetY()-selpt[1],
			event.GetX()-selpt[0])*180.0/M_PI;
    rot+=(angle-lastAngle);
    lastAngle=angle;
  }
  if(event.LeftUp()||event.RightUp()){
    int sel=GetSelectedChart();
    if(sel>=0){
      Vec2f trans1(-centroidx,
		   -centroidy);
      Vec2f trans2(centroidx,
		   centroidy);
      Vec2f trans3(xpos,ypos);
      for(int i=0;i<charts[sel].nvert;i++){
	charts[sel].tc[i]+=trans1;
	charts[sel].tc[i].rotate(rot*M_PI/180.0);
	charts[sel].tc[i]+=trans2;
	charts[sel].tc[i]+=trans3;
      }
      xpos=0;
      ypos=0;
      rot=0;
      SetChangedSinceSave(GEOMETRY_CHANGED);
    }

    //auto normalize the charts
    float minx=2,maxx=-2;
    float miny=2,maxy=-2;
    for(int i=0;i<ncharts;i++){
      for(int j=0;j<charts[i].nvert;j++){
	minx=MIN(charts[i].tc[j].x,minx);
	maxx=MAX(charts[i].tc[j].x,maxx);
	miny=MIN(charts[i].tc[j].y,miny);
	maxy=MAX(charts[i].tc[j].y,maxy);
      }
    }
    for(int i=0;i<ncharts;i++){
      for(int j=0;j<charts[i].nvert;j++){
	charts[i].tc[j].x=(254.0/256.0)*
	  (charts[i].tc[j].x-minx)/(maxx-minx)+1.0/256.0;
	charts[i].tc[j].y=(254.0/256.0)*
	  (charts[i].tc[j].y-miny)/(maxy-miny)+1.0/256.0;
      }
    }
    
  }
  Refresh();
}

void GLCanvas::modelModeMove(){
  
}

void GLCanvas::modelModeOnMouse(wxMouseEvent & event){
  static double lastx=-1;
  static double lasty=-1;
  double x=event.GetX();
  double y=event.GetY();

  if(event.Leaving()){
    memset(keys,0,256*sizeof(char));
  }
    
  if(lastx==-1){
    lastx=x;
    lasty=y;
  }
  double dx=x-lastx;
  double dy=y-lasty;

  if(event.ControlDown() && event.LeftIsDown()){
    cpos[2]-=speed[2]*dy;
  }

  //printf("model mouse %f %f %f %f %d\n",x,y,dx,dy,event.LeftIsDown());
  SetCurrent();
  if(!event.ControlDown() && event.LeftIsDown()){
    crot[1]+=speedry*dx;
    crot[0]-=speedrx*dy;
  }

  
  if(keys['a']||keys['A']){
    cpos[0]-=speed[0];
  }
  if(keys['d']||keys['D']){
    cpos[0]+=speed[0];
  }
  if(keys['w']||keys['W']){
    cpos[1]+=speed[1];
  }
  if(keys['s']||keys['S']){
    cpos[1]-=speed[1];
  }

  if(event.RightIsDown()){
    wxPoint pt=wxGetMousePosition();
    PopupMenu(modelModeMenu,(ScreenToClient(pt)));
  }

  lastx=x;
  lasty=y;
  Refresh();
}

void GLCanvas::modelModePaint(wxPaintEvent & event){
  glClearColor(0.5,0.6,0.6,1.0);
  glEnable(GL_DEPTH_TEST);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(50,(double)w/(double)h,0.1,1000);
	
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glTranslatef(cpos.x,cpos.y,cpos.z);
  glRotated(crot[0],1,0,0);
  glRotated(crot[1],0,1,0);
  glRotated(crot[2],0,0,1);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  float amb[]={0.1,0.1,0.1,1.0};
  float diff[]={0.6,0.6,0.6,1.0};
  float spec[]={0.3,0.3,0.3,1.0};

  glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,amb);
  glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,diff);
  glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec);	

  if(ncharts==0)
    ;//glutSolidTeapot(1);
  else{
    glPushMatrix();
    glEnable(GL_NORMALIZE);
    glScalef(scale.x,scale.y,scale.z);
    glTranslatef(-0.5,-0.5,-0.5);
    if(modelModeMenu->IsChecked(ID_Wireframe))
      glPolygonMode(GL_FRONT,GL_LINE);
    
    if(modelModeMenu->IsChecked(ID_DrawTexture)){
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,modelTex);
      glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    }
    else if(modelModeMenu->IsChecked(ID_ProjectTexture)){
      Matrix curCam(4,4);
      glGetDoublev(GL_MODELVIEW_MATRIX,curCam.data);
      curCam=curCam.transpose();
      Vec3f zdir(curCam(2,0),curCam(2,1),curCam(2,2));
      zdir.normalize();
      zdir*=-1;//its opposite of what we have
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,modelTex);
      int sel=-1;
      double val=-10;
      for(int i=0;i<segImageList->GetCount();i++){
	ImageData * imd=(ImageData *)segImageList->GetClientData(i);
	if(imd){
	  Vec3f z2(imd->extrinsic(2,0),imd->extrinsic(2,1),imd->extrinsic(2,2));
	  z2.normalize();
	  double v=z2.dot(zdir);
	  if(v>val){
	    val=v;
	    sel=i;
	  }
	}
      }
      double scx=1,scy=1;
      int imw=800,imh=600;
      if(sel>=0){
	ImageData * imd=(ImageData *)segImageList->GetClientData(sel);
	imd->image.initTextureNoResize(&scx,&scy);
	imw=imd->image.w;
	imh=imd->image.h;
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	double center[16]={scx*0.5,0,0,0,
			   0,scy*0.5,0,0,
			   0,0,0.5,0,
			   scx*0.5,scy*0.5,0.5,1};
	glMultMatrixd(center);
	myProjection(imd->intrinsic(0,0),imd->intrinsic(1,1),
		     imd->intrinsic(0,2),imd->intrinsic(1,2),imw,imh,1,100);
	glMultTransposeMatrixd(imd->extrinsic.data);
	glTranslatef(trans.x,trans.y,trans.z);
	glScalef(scale.x,scale.y,scale.z);
	Matrix mat(4,4);
	glGetDoublev(GL_MODELVIEW_MATRIX,mat.data);
	mat=mat.transpose();
	
	
	glPopMatrix();
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);
	glEnable(GL_TEXTURE_GEN_Q);
	glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
	glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
	glTexGeni(GL_R,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
	glTexGeni(GL_Q,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
	glTexGendv(GL_S,GL_OBJECT_PLANE,mat.data);
	glTexGendv(GL_T,GL_OBJECT_PLANE,mat.data+4);
	glTexGendv(GL_R,GL_OBJECT_PLANE,mat.data+8);
	glTexGendv(GL_Q,GL_OBJECT_PLANE,mat.data+12);
      }
    }
    for(int i=0;i<ncharts;i++){
      charts[i].draw(1);
    }
    if(modelModeMenu->IsChecked(ID_DrawTexture))
      glDisable(GL_TEXTURE_2D);
    if(modelModeMenu->IsChecked(ID_ProjectTexture)){
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
      glDisable(GL_TEXTURE_GEN_R);
      glDisable(GL_TEXTURE_GEN_Q);
      glDisable(GL_TEXTURE_2D);
    }
    if(modelModeMenu->IsChecked(ID_Wireframe)){
      glPolygonMode(GL_FRONT,GL_FILL);
    }

    glDisable(GL_LIGHTING);
    if(modelModeMenu->IsChecked(ID_DrawNormals)){
      glBegin(GL_LINES);
      for(int j=0;j<ncharts;j++){
	for(int i=0;i<charts[j].nvert;i++){
	  Vec3f v2=charts[j].norm[i];
	  v2.mult(0.03);
	  v2.add(charts[j].vert[i]);
	  glVertex3fv(charts[j].vert[i].getData());
	  glVertex3fv(v2.getData());
	}
      }
      glEnd();
    }
    if(modelModeMenu->IsChecked(ID_DrawCameras)){
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,targetTex);
      glPushMatrix();
      glColor3f(1,1,1);
      glScalef(1.0/scale.x,1.0/scale.y,1.0/scale.z);
      glTranslatef(-trans.x,-trans.y,-trans.z);
      glBegin(GL_QUADS);
      glTexCoord2f(0,0);glVertex3f(-1,0,-1);
      glTexCoord2f(1,0);glVertex3f( 1,0,-1);
      glTexCoord2f(1,1);glVertex3f( 1,0, 1);
      glTexCoord2f(0,1);glVertex3f(-1,0, 1);
      glEnd();
      glPopMatrix();
      glDisable(GL_TEXTURE_2D);

      for(int i=0;i<segImageList->GetCount();i++){
	ImageData * imd=(ImageData *)segImageList->GetClientData(i);
	if(imd){
	  Matrix tr_sc=Matrix::eye(4);
	  Matrix tr1=Matrix::eye(4);
	  /*tr1(0,3)=0.5;
	  tr1(1,3)=0.5;
	  tr1(2,3)=0.5;
	  */
	  tr_sc(0,0)=scale.x;
	  tr_sc(1,1)=scale.y;
	  tr_sc(2,2)=scale.z;
	  tr_sc(0,3)=trans.x;
	  tr_sc(1,3)=trans.y;
	  tr_sc(2,3)=trans.z;
	  Matrix tr_sc_inv=tr_sc.inverse();
	  Matrix ext=imd->extrinsic*tr_sc;
	  Matrix extinv=ext.inverse();
	  Matrix intrinv=imd->intrinsic.inverse();
	  Matrix corners(3,4);
	  Matrix corners2(4,4);
	  corners.setAll(1);
	  corners2.setAll(1);
	  corners(0,0)=0;corners(1,0)=0;
	  corners(0,1)=imd->image.width;corners(1,1)=0;
	  corners(0,2)=imd->image.width;corners(1,2)=imd->image.height;
	  corners(0,3)=0;corners(1,3)=imd->image.height;
	  corners=intrinv*corners;
	  for(int i=0;i<3;i++)
	    for(int j=0;j<4;j++)
	      corners2(i,j)=corners(i,j);
	  corners2=extinv*corners2;
	  
	  for(int j=0;j<4;j++){
	    corners2(0,j)/=corners2(3,j);
	    corners2(1,j)/=corners2(3,j);
	    corners2(2,j)/=corners2(3,j);
	  }
	    
	  Vec3d cpos(extinv(0,3),extinv(1,3),extinv(2,3));
	  Vec3d normal(extinv(0,2),extinv(1,2),extinv(2,2));
	  //cpos.normalize();
	  //cpos*=2;
	  
	  glPushMatrix();
	  glBegin(GL_LINE_LOOP);
	  glVertex3f(corners2(0,0),corners2(1,0),corners2(2,0));
	  glVertex3f(corners2(0,1),corners2(1,1),corners2(2,1));
	  glVertex3f(corners2(0,2),corners2(1,2),corners2(2,2));
	  glVertex3f(corners2(0,3),corners2(1,3),corners2(2,3));
	  glEnd();
	  glTranslatef(cpos.x,cpos.y,cpos.z);
;//glutSolidSphere(0.0001,10,10);
	  glPopMatrix();

	  glEnable(GL_BLEND);
	  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	  glLineStipple(1,0xF2F1);
	  glEnable(GL_LINE_STIPPLE);
	  glBegin(GL_LINES);

	  glColor4f(1.0,1.0,1.0,0.0);
	  //glVertex3f(tr_sc_inv(0,3),tr_sc_inv(1,3),tr_sc_inv(2,3));
	  double t=10000;
	  t=-cpos.y/normal.y;
	  if(t<0)t*=-1;
	  if(t>1000000)t=1000000;
	  Vec3d temp=cpos+normal*t;


	  glVertex3dv(temp.data);
	  glColor4f(1.0,1.0,1.0,1.0);
	  glVertex3dv(cpos.data);
	  
	  glEnd();
	  glDisable(GL_LINE_STIPPLE);
	  glDisable(GL_BLEND);
	}
      }
    }
    glPopMatrix();
  }
  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHTING);

  if(modelModeMenu->IsChecked(ID_DrawGrid)){
    glLineWidth(1);
    glColor3f(1.0,1.0,1.0);
    glBegin(GL_LINES);
    for(int i=0;i<20;i++){
      double y=i/20.0-0.5;
      glVertex3f(y,0.0,-0.5);
      glVertex3f(y,0.0, 0.5);
      glVertex3f(-0.5,0.0, y);
      glVertex3f( 0.5,0.0, y);
      
      glVertex3f(y,-0.5,0.0);
      glVertex3f(y, 0.5,0.0);
      glVertex3f(-0.5, y,0.0);
      glVertex3f( 0.5, y,0.0);
      
      glVertex3f(0.0,y,-0.5);
      glVertex3f(0.0,y, 0.5);
      glVertex3f(0.0,-0.5, y);
      glVertex3f(0.0, 0.5, y);
    }
    glEnd();
  }
  glDisable(GL_DEPTH_TEST);

  modelModeMove();   
}
