

#include "globals.h"
#include <stdarg.h>

namespace CapguiGlobals{

void log(char * fmt,...)
{
  va_list ap;
  int d;
  char c, *p, *s;
  char buffer[2048];
  va_start(ap,fmt);
  vsnprintf(buffer,2047,fmt,ap);

  int len=strlen(buffer);
  printf(buffer);

  va_end(ap);
}

void printErr(char * str)
{
  int err=0;
  if((err=glGetError())){
    printf("GLERROR (%s) : %d ",str,err);
    switch(err){
    case GL_INVALID_ENUM:
      printf("invalid enum\n");break;
    case GL_INVALID_VALUE:
      printf("invalid value\n");break;
    case GL_INVALID_OPERATION:
      printf("invalid operation\n");break;
    case GL_STACK_OVERFLOW:
      printf("stack overflow\n");break;
    case GL_STACK_UNDERFLOW:
      printf("stack underflow\n");break;
    case GL_OUT_OF_MEMORY:
      printf("out of mem\n");break;
    case GL_TABLE_TOO_LARGE:
      printf("table too large\n");break;
    }
  }
}

//project the point
void multPoint(double * res,double * mat,double * pt)
{
    res[0]=mat[0]*pt[0]+mat[4]*pt[1]+mat[8]*pt[2]+mat[12]*pt[3];
    res[1]=mat[1]*pt[0]+mat[5]*pt[1]+mat[9]*pt[2]+mat[13]*pt[3];
    res[2]=mat[2]*pt[0]+mat[6]*pt[1]+mat[10]*pt[2]+mat[14]*pt[3];
    res[3]=mat[3]*pt[0]+mat[7]*pt[1]+mat[11]*pt[2]+mat[15]*pt[3];
}

//invert the matrix, newly written somewhat more numerically
//sound than the version was using before
double invert(double * inv,double * origdata)
{
    int i, j, k;
    double data[16];
    double smallest=10e20;
    memcpy(data,origdata,sizeof(double)*16);
    memset(inv,0,sizeof(double)*16);
    inv[0]=1;
    inv[5]=1;
    inv[10]=1;
    inv[15]=1;
    

    for(j=0;j<4;j++){
	double largest=0;
	int largei=0;
	for(i=j;i<4;i++){
	    if(fabs(data[i+4*j])>largest){
		largei=i;
		largest=fabs(data[i+4*j]);
	    }
	}
	//pivot
	for(k=0;k<4;k++){
	    double temp=data[j+4*k];
	    data[j+4*k]=data[largei+4*k];
	    data[largei+4*k]=temp;

	    temp=inv[j+4*k];
	    inv[j+4*k]=inv[largei+4*k];
	    inv[largei+4*k]=temp;
	}
	if(fabs(data[j*4+j])<smallest)
	    smallest=fabs(data[j*4+j]);
	double ival=1.0/data[j+4*j];
	for(k=0;k<4;k++){
	    data[j+4*k]*=ival;
	    inv[j+4*k]*=ival;
	}
	for(i=0;i<4;i++){
	    if(i!=j){
		double scale=data[i+4*j];
		for(k=0;k<4;k++){
		    data[i+4*k]=data[i+4*k]-scale*data[j+4*k];
		    inv[i+4*k]=inv[i+4*k]-scale*inv[j+4*k];
		}
	    }
	}
    }
    return smallest;
    /*
    printf("Test invert\n");
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMultMatrixd(origdata);
    glMultMatrixd(inv);
    glGetDoublev(GL_MODELVIEW_MATRIX,data);
    
    for(i=0;i<4;i++){
	printf("%f %f %f %f\n",data[i],data[i+4],data[i+8],data[i+12]);
    }
    printf("invert \n");
    glPopMatrix();
    */
}

void myUnProject(double x,double y,double z,double * pmInv,int * view,
		 double *ox,double *oy,double *oz)
{
  double w;
  double zback=z;
  x=2.0*(x-view[0])/((double)view[2])-1.0;
  y=2.0*(y-view[1])/((double)view[3])-1.0;
  
  z=2.0*z-1.0;
  w=1.0;
  
  *ox=pmInv[0]*x+pmInv[4]*y+pmInv[8]*z+pmInv[12];
  *oy=pmInv[1]*x+pmInv[5]*y+pmInv[9]*z+pmInv[13];
  *oz=pmInv[2]*x+pmInv[6]*y+pmInv[10]*z+pmInv[14];
  w=pmInv[3]*x+pmInv[7]*y+pmInv[11]*z+pmInv[15];

  double winv=1.0/w;
  (*ox)*=winv;
  (*oy)*=winv;
  (*oz)*=winv;
}

void myProject(double x,double y,double z,double * pm,int * view,
	       double *sx,double *sy,double *sz)
{
  double w=1.0;
  double tx,ty,tz;
  tx=pm[0]*x+pm[4]*y+pm[8]*z+pm[12];
  ty=pm[1]*x+pm[5]*y+pm[9]*z+pm[13];
  tz=pm[2]*x+pm[6]*y+pm[10]*z+pm[14];
  w=pm[3]*x+pm[7]*y+pm[11]*z+pm[15];
 
  w=1.0/w;
  x=tx*w;
  y=ty*w;
  z=tz*w;

  (*sx)=view[0]+view[2]*(x+1)/2.0;
  (*sy)=view[1]+view[3]*(y+1)/2.0;
  (*sz)=(z+1.0)/2.0;
}


void UnProjectProject(double x,double y,double z,
		      double * pmInv,int * worldView,
		      double * pm,int * lightView,
		      double *sx,double *sy,double *sz){
  double w;
  double zback=z;
  double ox,oy,oz,ow;
  
  x=2.0*(x-worldView[0])/((double)worldView[2])-1.0;
  y=2.0*(y-worldView[1])/((double)worldView[3])-1.0;
  
  z=2.0*z-1.0;
  
  ox=pmInv[0]*x+pmInv[4]*y+pmInv[8]*z+pmInv[12];
  oy=pmInv[1]*x+pmInv[5]*y+pmInv[9]*z+pmInv[13];
  oz=pmInv[2]*x+pmInv[6]*y+pmInv[10]*z+pmInv[14];
  ow=pmInv[3]*x+pmInv[7]*y+pmInv[11]*z+pmInv[15];

  x=pm[0]*ox+pm[4]*oy+pm[8]*oz+pm[12]*ow;
  y=pm[1]*ox+pm[5]*oy+pm[9]*oz+pm[13]*ow;
  z=pm[2]*ox+pm[6]*oy+pm[10]*oz+pm[14]*ow;
  w=pm[3]*ox+pm[7]*oy+pm[11]*oz+pm[15]*ow;
 
  w=1.0/w;
  x=x*w;
  y=y*w;
  z=z*w;

  (*sx)=lightView[0]+lightView[2]*(x+1)/2.0;
  (*sy)=lightView[1]+lightView[3]*(y+1)/2.0;
  (*sz)=(z+1.0)/2.0;
}

}
