#include <index.html.h>
#ifdef USING_GLEE
#include <GLee.h>
#endif
#include "globals.h"

#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <wx/wx.h>
#include <wx/listctrl.h>
#include <wx/checklst.h>
#include <wx/event.h>
#include <wx/glcanvas.h>
#include <sys/time.h>
#include <wx/splitter.h>
#include <wx/spinbutt.h>
#include <wx/spinctrl.h>
#include <wx/toolbar.h> 
#include <wx/cmdline.h>
#include <wx/image.h>
#include <wx/dir.h>
#include <vector>
#include <string>
#include <wx/process.h>
#include <wx/splash.h>

#include "glcanvas.h"
#include "cameraprofile.h"
#include "cameradlg.h"
#include "capturedlg.h"
#include "calibratedlg.h"
#include "segmentdlg.h"
#include "bbdlg.h"
#include "savedlg.h"
#include "image.h"
#include "matrix.h"
#include "newseg.h"
#include "clb.h"
#include "chart.h"
#include "texgen.h"
#include "simplify.h"
#include "dynatex.h"
#include "sot.h"
#include "plugin.h"
#include "splash.xpm"

///Structure for holding the icons exported from the gimp.
typedef struct IMSTRUCT_T{
  unsigned int 	 width;
  unsigned int 	 height;
  unsigned int 	 bpp; /* 3:RGB, 4:RGBA */ 
  unsigned char	 data[32 * 32 * 4 + 1];
} imstruct;
//The images in c format
#include "images/prev.c"
#include "images/next.c"
#include "images/save.c"
#include "images/seg.c"
#include "images/saveas.c"
#include "images/open.c"
#include "images/tex.c"
#include "images/new.c"
#include "images/geometry.c"
#include "images/capture.c"

#include <wx/dnd.h>
#include <wx/dataobj.h>

#define CURRENT_PROJ_MAX 2048

//GLOBALS
Chart * charts=0;
int ncharts=0;
int attribs[]={WX_GL_RGBA,WX_GL_DOUBLEBUFFER,WX_GL_STENCIL_SIZE,8,
	       WX_GL_DEPTH_SIZE,16,0};
wxListBox * imageList;
wxListBox * calibImageList;
wxListBox * segImageList;
char currentProj[CURRENT_PROJ_MAX]="capture_project";

int changedSinceSave=0;
Options options;
CameraProfiles cameraProfiles;
Vec3f scale=DEFAULT_SCALE;
Vec3f trans=DEFAULT_TRANS;

extern char **environ;

IMPLEMENT_DYNAMIC_CLASS(StatusBar,wxStatusBar)

//Status bar events
BEGIN_EVENT_TABLE(StatusBar, wxStatusBar)
  EVT_SIZE(StatusBar::OnSize)
END_EVENT_TABLE()

void setToDefaultProj()
{
  snprintf(currentProj,CURRENT_PROJ_MAX,"%s/capture_project",(const char*)wxGetHomeDir().ToAscii());//getenv("HOME"));
}

int mergeCharts(Chart & chart,Chart * charts,int ncharts)
{
  //Put all the charts into a single chart
  int nvert=0,npoly=0;
  for(int i=0;i<ncharts;i++){
    nvert+=charts[i].nvert;
    npoly+=charts[i].npoly;
  }
  chart.init(npoly,nvert);
  nvert=0;
  npoly=0;
  for(int i=0;i<ncharts;i++){
    memcpy(chart.vert+nvert,charts[i].vert,
	   sizeof(charts[i].vert[0])*charts[i].nvert);
    memcpy(chart.tc+nvert,charts[i].tc,
	   sizeof(charts[i].tc[0])*charts[i].nvert);
    for(int j=0;j<charts[i].npoly;j++){
      chart.poly[j+npoly][0]=charts[i].poly[j][0]+nvert;
      chart.poly[j+npoly][1]=charts[i].poly[j][1]+nvert;
      chart.poly[j+npoly][2]=charts[i].poly[j][2]+nvert;
    }
    nvert+=charts[i].nvert;
    npoly+=charts[i].npoly;
    printf("%d %d\n",npoly,nvert);
  }
  chart.initNormals();
  printf("combined geometry %d %d\n",npoly,nvert);
}

void writeConfig(){
  wxString homedir_s=wxGetHomeDir();
  char homedir[2048];
  strncpy(homedir,(const char *)homedir_s.ToAscii(),2048);
  if(!homedir_s.Len()){
    printf("no home environment variable...\n");
    return;
  }
  char configfile[2048];
  char configdir[2048];
  snprintf(configdir,2048,"%s/.capsys",homedir);
  snprintf(configfile,2048,"%s/.capsys/config",homedir);

  FILE * file = fopen(configfile,"w");
  if(!file){
    printf("cannot open config file for writing\n");
    return;
  }
  fprintf(file,"# This is a config file for the capture ui\n");
  fprintf(file,"# Lines starting with \'#\' are ignored\n");
  fprintf(file,"playdelay       %d\n",options.playDelay);
  fprintf(file,"capturedelay    %d\n",options.captureDelay);
  fprintf(file,"captureimages   %d\n",options.captureImages);
  fprintf(file,"cameraformatindex %d\n",options.cameraFormatIndex);
  fprintf(file,"cameramodeindex %d\n",options.cameraModeIndex);
  fprintf(file,"mainframesize %d %d\n",options.mainFrameWidth,
	  options.mainFrameHeight);
  fprintf(file,"maxresidual %lf\n",options.maxResidual);
  fprintf(file,"currentcameraprofile %d\n",cameraProfiles.currentProfile);
  fprintf(file,"calibthresh %d\n",options.calibThresh);
  fprintf(file,"calibautothresh %d\n",options.calibAutoThresh);
  fprintf(file,"calibchannel %d\n",options.calibChannel);
  fprintf(file,"calibsmall %d\n",options.calibSmall);
  fprintf(file,"calibbig %d\n",options.calibBig);
  fprintf(file,"bayerenabled %d\n",options.bayerEnabled);
  fprintf(file,"bayerpattern %d\n",options.bayerPattern);
  fprintf(file,"bayerinterpmode %d\n",options.bayerInterpolationMode);
  
  fclose(file);

  //Load the camera profiles
  cameraProfiles.save(configdir);
}

void readConfig()
{
  wxString wtfS(_T("wtf"));
  wxString homedir_s=wxGetHomeDir();
  char configdir[2048];
  char configfile[2048];
  char homedir[2048];
  
  strncpy(homedir,(const char*)homedir_s.ToAscii(),2048);
  
  struct stat buf;
  if(!homedir_s.Len()){
    printf("no home environment variable\n");
  }
  snprintf(configdir,2048,"%s/.capsys",homedir);
  snprintf(configfile,2048,"%s/config",configdir);
  printf("home dir is \"%s\"\n", homedir);
  printf("loading config from %s\n",configfile);

  if(stat(configdir,&buf)<0){
    printf("creating config directory %s\n",configdir);
    if(wxMkdir(wxString::FromAscii(configdir))==0){
      printf("error creating config directory...\n");
      return;
    }
    printf("config directory created\n");
    wxString message;
    message =  wxT("The config directory :")+
      wxString::FromAscii(homedir)+ 
      wxT("\nDid not exist.  You must install the plugins into this directory in order to have a working interface");
    wxMessageBox(message,wxT("Error:Missing Config directory"));
  }
  if(stat(configfile,&buf)<0){
    printf("no config file, writing default values\n");
    writeConfig();
  }
  
  wxDir dir(wxString::FromAscii(configdir));
  /*
  wxString plugs;
  plugs = wxT("*")+wxString::FromAscii(PLUG_EXTENSION);
  if(!dir.HasFiles(plugs)){
    wxString message;
    wxArrayString names;
    wxDir::GetAllFiles(wxString::FromAscii(homedir), &names);
    message = wxT("Config directory ")
      + wxString::FromAscii(configdir) + 
      wxString::Format(wxT("\nHas %d files but no plugins of type")+
		       names.Count())+
      plugs;
    wxMessageBox(message,wxT("Error"));
  }
  */
  
  FILE * file=fopen(configfile,"r");
  char line[1024];
  while(NULL!=fgets(line,1023,file)){
    if(line[0]!='#'){
      char option[256];
      if(sscanf(line,"%s",option)!=1){
	continue;
      }
      int len=strlen(option);
      char * params=line+len;
      if(!strcmp(option,"playdelay")){
	int p;
	sscanf(params,"%d",&p);
	options.playDelay=p;
	printf("play delay is %d\n",p);
      }
      else if(!strcmp(option,"capturedelay")){
	int p;
	sscanf(params,"%d",&p);
	options.captureDelay=p;
	printf("capture delay is %d\n",p);
      }
      else if(!strcmp(option,"captureimages")){
	int p;
	sscanf(params,"%d",&p);
	options.captureImages=p;
	printf("capture images is %d\n",p);
      }
      else if(!strcmp(option,"cameraformatindex")){
	int p;
	sscanf(params,"%d",&p);
	options.cameraFormatIndex=p;
	printf("cam format is %d\n",p);
      }
      else if(!strcmp(option,"cameramodeindex")){
	int p;
	sscanf(params,"%d",&p);
	options.cameraModeIndex=p;
	printf("cam mode index is %d\n",p);
      }
      else if(!strcmp(option,"mainframesize")){
	int p1,p2;
	if(2==sscanf(params,"%d %d",&p1,&p2)){
	  options.mainFrameWidth=p1;
	  options.mainFrameHeight=p2;
	}
      }
      else if(!strcmp(option,"maxresidual")){
	double p;
	if(1==sscanf(params,"%lf",&p)){
	  options.maxResidual=p;
	}
      }
      else if(!strcmp(option,"currentcameraprofile")){
	int p;
	sscanf(params,"%d",&p);
	cameraProfiles.currentProfile=p;
	printf("current camera profile %d\n",p);
      }
      else if(!strcmp(option,"calibthresh")){
	int p;
	sscanf(params,"%d",&p);
	options.calibThresh=p;
      }
      else if(!strcmp(option,"calibautothresh")){
	int p;
	sscanf(params,"%d",&p);
	options.calibAutoThresh=p;
      }
      else if(!strcmp(option,"calibchannel")){
	int p;
	sscanf(params,"%d",&p);
	options.calibChannel=p;
      }
      else if(!strcmp(option,"calibbig")){
	int p;
	sscanf(params,"%d",&p);
	options.calibBig=p;
      }
      else if(!strcmp(option,"calibsmall")){
	int p;
	sscanf(params,"%d",&p);
	options.calibSmall=p;
      }
      else if(!strcmp(option,"bayerenabled")){
	int p;
	sscanf(params,"%d",&p);
	options.bayerEnabled=p;
      }
      else if(!strcmp(option,"bayerpattern")){
	int p;
	sscanf(params,"%d",&p);
	options.bayerPattern=p;
      }
      else if(!strcmp(option,"bayerinterpmode")){
	int p;
	sscanf(params,"%d",&p);
	options.bayerInterpolationMode=p;
      }
      else{
	printf("unknown option %s\n",option);
      }
    }
  }
  fclose(file);

  //Load the camera profiles
  cameraProfiles.load(configdir);
}




void save(ImageData * imd,const char * fname,
	  int savecalib=0,
	  int savecamera=0,
	  int savemask=0,
	  int overwrite=0){
  struct stat buf;
  int write=1;
  if(stat(fname,&buf)==0 && !overwrite){
    int answer = wxMessageBox(wxT("File exists, overwrite?"), 
			      wxT("Confirm"),
			      wxICON_QUESTION|wxYES_NO,NULL);
    if(answer==wxNO){
      write=0;
      printf("write cancelled\n");
    }
  }
  if(write){
    if(imd){
      if(savemask && (imd->flags & HAS_MASK)){
	if(!(strstr(fname,".tga")||strstr(fname,".png"))){
	  printf("can only save tga's or png's with mask");
	}
	else if(imd->image.nchannels==1){
	  printf("can not save grayscale tga's with mask");
	}
	else {
	  Image8 im(imd->image.width,imd->image.height,
		    4);
	  for(int y=0;y<im.height;y++){
	    for(int x=0;x<im.width;x++){
	      im(x,y,0)=imd->image(x,y,0);
	      im(x,y,1)=imd->image(x,y,1);
	      im(x,y,2)=imd->image(x,y,2);
	      im(x,y,3)=imd->mask(x,y)*255;
	    }
	  }
	  im.save(fname);
	}
      }
      else{
	if(savemask && !(imd->flags&HAS_MASK)){
	  printf("no mask to save\n");
	}
	imd->image.save((char *)fname);
      }
	  
      if(savecamera){
	if(!(imd->flags & HAS_EXTRINSIC) ||
	   !(imd->flags & HAS_INTRINSIC)){
	  printf("no calibration to save, extr:%d intr:%d\n",
		 (imd->flags&HAS_EXTRINSIC)>0,
		 (imd->flags&HAS_INTRINSIC)>0);
	}
	else{
	  Matrix proj; //fixme, handle if needed
	  char * cname = new char[strlen(fname)+5];
	  sprintf(cname,"%s.clb",fname);
	  FILE * file = fopen(cname,"wb");
	  writeCLB(file,imd);
	  fclose(file);
	}
      }	  
    }
    else{
      printf("nothing to save\n");
    }
  }
}
GLCanvas * glCanvas;

/*
void saveProject(char * directory,int saveas=0,ProgressDialog *caller=0);
//helper class to show a progress indicator while saving
class SaveProjectDialog:public ProgressDialog{
public:
  char * directory;
  int saveas;
  SaveProjectDialog(char * dir,int sas):
    ProgressDialog(GetMainFrame(),-1,"Saving Project"){
    directory=dir;
    saveas=sas;
  }
  void OnIdle(wxIdleEvent & event){
    saveProject(directory,saveas,this);
    event.RequestMore(TRUE);
  }
  };*/


void saveProject(const char * directory,int saveas=0)
{
  printf("saving project to %s\n",directory);
  int imi=0;
  int nimages=0;
  int worktodo;
  char filename[2048];
  int overwrite=0;
  imi=0;
  nimages=0;
  
  
  if(!wxDirExists(wxString::FromAscii(directory))){
    printf("making directory\n");
    if(wxMkdir(wxString::FromAscii(directory))==0){
      wxString message;
      message = wxT("Error creating directory: \"")+
	wxString::FromAscii(directory)+wxT("\"");
      wxMessageBox(message,wxT("Save Failed"),
		   wxICON_ERROR|wxOK,NULL);
      return;
    }
  }
  else{
    wxString message;
	
    DIR * dir = opendir(directory);
    if(!dir){
      wxMessageBox(wxT("Directory exists, but is not readable"),
		   wxT("Save Failed"),
		   wxICON_ERROR|wxOK,NULL);
      return;
    }
    struct dirent * de;
    int _ncalib=0;
    int _nimages=0;
    int _ncharts=0;
    int _nfiles=0;
    while(de=readdir(dir)){
      int lastdot=strlen(de->d_name)-4;
      snprintf(filename,2048,"%s/%s",directory,de->d_name);
      if((strstr(de->d_name,".tga")-de->d_name)==lastdot){
	_nimages++;
      }
      else if((strstr(de->d_name,".obj")-de->d_name)==lastdot){
	_ncharts++;
      }
      else if((strstr(de->d_name,".clb")-de->d_name)==lastdot){
	_ncalib++;
      }
      _nfiles++;
    }
    closedir(dir);
    message = wxT("Overwrite directory :")
      + wxString::FromAscii(directory) 
      + wxString::Format(wxT(" (%d/%d calib. images, %d charts)"),
			 _ncalib,_nimages,_ncharts);
      
    if(_nfiles==2 && _nimages==0 && _ncharts==0
       && _ncalib==0){
      printf("its okay, its an empty directory\n");
    }
    else{
      int answer = wxMessageBox(message, wxT("Confirm"),
				wxICON_QUESTION|wxYES_NO,NULL);
      if(answer==wxYES){
	printf("overwrite it\n");
	overwrite=1;
	printf("FIXME:Check to make sure that there is more files in the current save...else remove the old ones\n");
	  
	dir = opendir(directory);
	while(de=readdir(dir)){
	  int lastdot=strlen(de->d_name)-4;
	  snprintf(filename,2048,"%s/%s",directory,de->d_name);
	  if(GetChangedSinceSave()&IMAGES_CHANGED){
	    if((strstr(de->d_name,".tga")-de->d_name)==lastdot){
	      printf("Conflict in directory, removing: %s\n",filename);
	      remove(filename);
	    }
	    else if((strstr(de->d_name,".clb")-de->d_name)==lastdot){
	      printf("Conflict in directory, removing: %s\n",filename);
	      remove(filename);
	    }
	  }
	  else if(GetChangedSinceSave()&GEOMETRY_CHANGED){
	    if((strstr(de->d_name,".obj")-de->d_name)==lastdot &&
	       strstr(de->d_name,"chart")){
	      printf("Conflict in directory, removing: %s\n",filename);
	      remove(filename);
	    }
	  }
	}
	closedir(dir);
      }
      else
	return;
    }
  }

  wxBeginBusyCursor();
  DisableAll();
 
  nimages=GetAllImageCount();
  worktodo=nimages*(GetChangedSinceSave()&IMAGES_CHANGED)
    +ncharts*(GetChangedSinceSave()&GEOMETRY_CHANGED);
  
  if((GetChangedSinceSave()&IMAGES_CHANGED) || saveas){
    int nu=GetUncalImageCount();
    int nc=GetCalImageCount();
    int ns=GetSegImageCount();

    for(int i=0;i<imageList->GetCount();i++){
      ImageData * imd = (ImageData *)imageList->GetClientData(i);
      snprintf(filename,2048,"%s/image-%04d.tga",directory,imi);
      printf("saving image %d\n",imi);
      if(imd){
	save(imd,filename,1,1,1,overwrite);
	imi++;
	SetStatusText(wxT("Saving uncalib. images"));
	SetProgress(((float)imi)/((float)worktodo));
	GetMainFrame()->Refresh();
	GetApp()->Yield();
      }
    }
        
    for(int i=0;i<calibImageList->GetCount();i++){
      ImageData * imd = (ImageData *)calibImageList->GetClientData(i);
      snprintf(filename,2048,"%s/image-%04d.tga",directory,imi);
      if(imd){
	save(imd,filename,1,1,1,overwrite);
	imi++;
	SetProgress(((float)imi)/((float)worktodo));
	SetStatusText(wxT("Saving calib. images"));
	GetMainFrame()->Refresh();
	GetApp()->Yield();
      }
    }
    
    for(int i=0;i<segImageList->GetCount();i++){
      ImageData * imd = (ImageData *)segImageList->GetClientData(i);
      snprintf(filename,2048,"%s/image-%04d.tga",directory,imi);
      printf("saving %d\n",imi);
      if(imd){
	save(imd,filename,1,1,1,overwrite);
	imi++;
	SetProgress(((float)imi)/((float)worktodo));
	SetStatusText(wxT("Saving segmented images"));
	GetMainFrame()->Refresh();
	GetApp()->Yield();
      }
    }
    printf("done saving images\n");
  }
  if((GetChangedSinceSave()&GEOMETRY_CHANGED )|| saveas){
    SetStatusText(wxT("Saving geometry"));
    for(int i=0;i<ncharts;i++){
      snprintf(filename,2048,"%s/chart-%04d.obj",directory,i);
      charts[i].saveobj(filename);
      
      SetProgress(((float)imi+i)/((float)worktodo));
      GetMainFrame()->Refresh();
      GetApp()->Yield();
    }
  }
  snprintf(filename,2048,"%s/scale_and_translation",directory);
  FILE * file = fopen(filename,"wb");
  fwrite(scale.data,sizeof(scale.data),1,file);
  fwrite(trans.data,sizeof(trans.data),1,file);
  fclose(file);
  
  wxString status;
  SetStatusText(status);
  //save other shit here
  UpdateTitle();
  strncpy(currentProj,directory,CURRENT_PROJ_MAX);
  printf("changed currentProj <%s>=<%s>\n",currentProj,directory);
  
  wxEndBusyCursor();
  EnableAll();
  SetChangedSinceSave(0);

}

int loadProject(const char * directory)
{
  char imext[5]=".tga";
  char filename[2048];
  int imi=0;
  int _ncharts=0;
  DIR * dir = opendir(directory);
  if(!dir){
    return 0;
  }
  DisableAll();
  wxBeginBusyCursor();
  struct dirent * de;
  int nimages=0;
  int imagessofar=0;
  std::vector<std::string> imageNames;

  while(de=readdir(dir)){
    int lastdot=strlen(de->d_name)-4;
    if((strstr(de->d_name,".tga")-de->d_name)==lastdot ||
       (strstr(de->d_name,".jpg")-de->d_name)==lastdot ||
       (strstr(de->d_name,".png")-de->d_name)==lastdot ||
       (strstr(de->d_name,".ppm")-de->d_name)==lastdot){
      imageNames.push_back(std::string(de->d_name));
      nimages++;
    }
    else if((strstr(de->d_name,".obj")-de->d_name)==lastdot &&
	    (strstr(de->d_name,"chart-"))){
      _ncharts++;
    }
  }
  std::sort(imageNames.begin(), imageNames.end());

  //rewinddir(dir);
  //while(de=readdir(dir)){
  //int lastdot=strlen(de->d_name)-4;
  /*
    Used to use readdir to look for files, but redhat put them out
    of order and it was super annoying.
   */
  for(int i=0; i<imageNames.size(); i++){
    struct stat buf;
    std::string nm = imageNames[i];
    snprintf(filename,2048,"%s/%s",directory, imageNames[i].c_str());
    if(stat(filename, &buf)>=0){
      ImageData * imd = new ImageData();
      imd->calibInfo=0;
      if(imd->image.read(filename)){
	imd->flags|=HAS_IMAGE;
      }
      if(imd->image.nchannels==4){
	Image8 all=imd->image;
	imd->flags|=HAS_MASK;

	imd->image=Image8(all.width,all.height,all.nchannels-1);
	imd->mask =Image8(all.width,all.height,1);
	//FIXME: move into the image class?
	for(int y=0;y<all.height;y++){
	  for(int x=0;x<all.width;x++){
	    imd->image(x,y,0)=all(x,y,0);
	    imd->image(x,y,1)=all(x,y,1);
	    imd->image(x,y,2)=all(x,y,2);
	    imd->mask(x,y)=all(x,y,3)>0;
	  }
	}
      }
      sprintf(filename,"%s/%s.clb",directory, nm.c_str());
      struct stat buf;
      if(stat(filename,&buf)==0 &&
	 S_ISREG(buf.st_mode)){
	imd->calibInfo = new CalibInfo();
	FILE * f = fopen(filename,"rb");
	imd->flags|=readCLB(f,imd);
	fclose(f);
      }
      if(imd->flags&HAS_MASK){
	if(segImageList->GetString(0)==wxT(" "))
	  segImageList->Clear();
	segImageList->Append(wxString::FromAscii(nm.c_str()),imd);
      }
      else if(imd->flags&HAS_EXTRINSIC){
	if(calibImageList->GetString(0)==wxT(" "))
	  calibImageList->Clear();
	calibImageList->Append(wxString::FromAscii(nm.c_str()),imd);
      }
      else{
	if(imageList->GetString(0)==wxT(" "))
	  imageList->Clear();
	imageList->Append(wxString::FromAscii(nm.c_str()),imd);
      }
      SetStatusText(wxString::FromAscii(nm.c_str()));
      GetMainFrame()->Refresh();
      GetApp()->Yield();
      imagessofar++;
      SetProgress(((float)(imagessofar-1))/(float)nimages);
    }
    //printf("%s\n",de->d_name);
  }
    
  if(ncharts && charts){
    delete [] charts;
  }

  ncharts=_ncharts;
  charts=new Chart[ncharts];
  for(int i=0;i<ncharts;i++){
    snprintf(filename,2048,"%s/chart-%04d.obj",directory,i);
    charts[i].loadobj(filename);
  }
  snprintf(filename,2048,"%s/scale_and_translation",directory);
  FILE * file=fopen(filename,"rb");
  if(!file){
    scale=DEFAULT_SCALE;
    trans=DEFAULT_TRANS;
  }
  else{
    fread(scale.data,sizeof(scale.data),1,file);
    fread(trans.data,sizeof(trans.data),1,file);
    fclose(file);
  }
  snprintf(currentProj,CURRENT_PROJ_MAX,directory);
  printf("FIXME: SetChangedSinceSave needs to be called, but should it be with a 0 or 1\n");
  SetChangedSinceSave(0);
  UpdateTitle();
  
  SetStatusText(wxT("Done loading project"));
  SetProgress(1.0f);
  wxEndBusyCursor();
  EnableAll();
}

/** The listbox containing the uncalibrated (and unsegmented) images.
 * The listbox contains all the information (see ImageData) about the
 * uncalibrated images in the capture project.  Operations are performed
 * on selected items using a popup menu which is active on right clicks.
 *
 * Items removed from the ImageList are removed from the project. 
 */
class ImageList:public wxListBox{
  wxMenu * menu;
  int index;

  //current calibration parameters
  int thresh[3];
  int small;
  int big;

  enum {IDLE=0,CALIBRATING,PLAYING} state;
public:
  DECLARE_DYNAMIC_CLASS(ImageList)
    ImageList():wxListBox(){
  }
  ImageList(wxWindow * parent):
    wxListBox(parent,ID_ImageList,wxDefaultPosition,
	      wxDefaultSize,0,0,wxLB_EXTENDED){
    menu = new wxMenu();
    menu->Append(ID_SelectAll,wxT("Select All"),wxT("Select All"),wxITEM_NORMAL);
    menu->AppendSeparator();
    menu->Append(ID_Remove,wxT("Remove Selected"),wxT("Delete Selected"),wxITEM_NORMAL);
    //menu->Append(ID_RemoveAll,wxT("Remove All"),wxT("Delete Selected"),wxITEM_NORMAL);
    menu->AppendSeparator();
    menu->Append(ID_Play,wxT("Play"),wxT("Play"),wxITEM_NORMAL);
    menu->AppendSeparator();
    menu->Append(ID_Load,wxT("Load"),wxT("Load Images"));
    menu->Append(ID_Capture,wxT("Capture"),wxT("Capture Images"));
    menu->AppendSeparator();
    menu->Append(ID_Save,wxT("Save"),wxT("Save Selected"),wxITEM_NORMAL);
    menu->Append(ID_Calibrate,wxT("Calibrate"),wxT("Calibrate"),wxITEM_NORMAL);
    menu->AppendSeparator();
    menu->Append(ID_Properties,wxT("Properties"),wxT("Properties"));
    state=IDLE;
  }
  void OnProperties(wxCommandEvent & event){
    int sel=GetSelectionSafe(this);
    if(sel>=0){
      ImageData * imd = (ImageData *)GetClientData(sel);
      if(imd){
	imd->showProperties();
      }
    }
  }
  void OnIdle(wxIdleEvent & event){
    if(state==PLAYING){
      if(index>=GetCount()){
	state=IDLE;
	index=0;
      }
      else{
	static StopWatch watch;
	static int called=0;
	if(index==1&&!called){
	  watch.reset();
	  called=1;
	}
	if(watch.hasElapsed(options.playDelay)){
	  wxArrayInt selections;
	  GetSelections(selections);
	  for(int i=selections.GetCount()-1;i>=0;i--){
	    Deselect(selections[i]);
	  }
	  SetSelection(index);
	  wxCommandEvent dummy;
	  OnListSelected(dummy);

	  index++;
	  called=0;
	}
	event.RequestMore(TRUE);
      }
    }    
    else if(GetCount()==0){
      Append(wxT(" "));
      Append(wxT(" "));
      Append(wxT(" "));
      Append(wxT(" "));
      Append(wxT(" "));
      Append(wxT(" "));
      Append(wxT(" "));
      event.RequestMore(FALSE);
    }
    else{
      event.RequestMore(FALSE);
    }
  }
  void OnPlay(wxCommandEvent & event){
    for(int i=0;i<GetCount();i++){
      Deselect(i);
    }
    Select(0);
    state=PLAYING;
    index=1;
    Refresh();
  }
  void OnCalibrate(wxCommandEvent & event){
    CalibrateDialog * cd=new CalibrateDialog(GetMainFrame(),0);
    if(cd->ShowModal()==wxID_OK){
      /*index=0;
      state=CALIBRATING;

      cd->GetThresh(thresh);
      big=cd->GetBig();
      small=cd->GetSmall();
      printf("got thresh %d %d %d\n",thresh[0],thresh[1],thresh[2]);
      printf("got small %d and big %d\n",small,big);*/
    }
    cd->Destroy();
  }
  void OnCapture(wxCommandEvent & event){
    int before=GetUncalImageCount();
    CaptureDialog * cd =new CaptureDialog(GetMainFrame(),0);
    cd->ShowModal();
    cd->Destroy();
    int after=GetUncalImageCount();
    if(after>before){
      SetChangedSinceSave(IMAGES_CHANGED);
    }
  }
  void OnLoad(wxCommandEvent & event){
    wxFileDialog fd(GetParent(),wxT("Select file to load"),wxT(""),wxT(""),wxT("Images|*.ppm;*.pgm;*.tga;*.jpg;*.png"),wxOPEN|wxMULTIPLE);
    if(wxID_OK==fd.ShowModal()){
      wxArrayString names;
      wxArrayString paths;
      fd.GetFilenames(names);
      fd.GetPaths(paths);

      if(GetString(0)==wxT(" ")){
	Clear();
      }

      for(int i=0;i<names.GetCount();i++){
	ImageData * imd = new ImageData();
	imd->image.read((const char *)paths[i].ToAscii());
	imd->set(HAS_IMAGE);
	Append(names[i],imd);
      }
      SetChangedSinceSave(IMAGES_CHANGED);
    }
  }
  //Just saves images
  void OnSave(wxCommandEvent & event){
    wxArrayInt sel;
    GetSelections(sel);
    if(sel.GetCount()==1){
      wxFileDialog fd(GetParent(),wxT("Select filename to save"),wxT(""),wxT("image.tga"),wxT("*.*"),wxSAVE);
      if(wxID_OK==fd.ShowModal()){
	wxString path=fd.GetPath();
	save((ImageData *)GetClientData(GetSelectionSafe(this)),(const char *)path.ToAscii());
      }
    }
    else {
      wxFileDialog fd(GetParent(),wxT("Select directory and enter printf command"),wxT(""),wxT("image-%04d.tga"),wxT("*.*"),wxSAVE);
      
      if(wxID_OK==fd.ShowModal()){
	wxString paths=fd.GetPath();
	wxString path;
	for(int i=0;i<sel.GetCount();i++){
	  path.sprintf(paths,i);
	  printf("%s\n",(const char*)paths.ToAscii());
	  printf("%s\n",(const char*)path.ToAscii());
	  save((ImageData *)GetClientData(sel[i]),(const char *)path.ToAscii());
	}
      }
      else{
	printf("fd.ShowModal returned error\n");
      }
    }
  }
#ifdef OLD_SEL
  void OnListSelected(wxCommandEvent & event){    
    int i=GetSelection(this);
    
    //So texture doesn't get updated for multi selections
    wxArrayInt sel;
    GetSelections(sel);
    int n=sel.GetCount();
    if(n>1)return;
    
    if(i>=0 && GetClientData(i))
      glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
  } 
#else
  void OnListSelected(wxCommandEvent & event){
    static wxArrayInt selLast;
    int i;
    wxArrayInt sel;
    GetSelections(sel);
    
    for(i=0;i<sel.GetCount();i++){
      if(i<selLast.GetCount()){
	if(sel[i]!=selLast[i])
	  break;
      }else break;
    }
    if(i>=sel.GetCount())
      i=GetSelection();
    else
      i=sel[i];
    if(i>=0 && GetClientData(i)){
      glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
    }
    selLast=sel;
  }
#endif
  void OnRightDown(wxMouseEvent & event){
    wxPoint pt=wxGetMousePosition();
    PopupMenu(menu,(ScreenToClient(pt)));
    //if(state==PLAYING)state=IDLE;
  }
  void OnRemove(wxCommandEvent & event){
    wxArrayInt selections;
    GetSelections(selections);
    for(int i=selections.GetCount()-1;i>=0;i--){
      if(GetClientData(selections[i])){
	delete ((ImageData *)GetClientData(selections[i]));
      }
      Delete(selections[i]);
    }
    SetChangedSinceSave(IMAGES_CHANGED);
  }
  void OnRemoveAll(wxCommandEvent & event){
    for(int i=0;i<GetCount();i++){
      if(GetClientData(i)){
	delete ((ImageData *)GetClientData(i));
      }
    }
    Clear();
    SetChangedSinceSave(IMAGES_CHANGED);
  }
  void OnSelectAll(wxCommandEvent & event){
    for(int i=0;i<GetCount();i++){
      Select(i);
    }
  }
  DECLARE_EVENT_TABLE();
};

IMPLEMENT_DYNAMIC_CLASS(ImageList,wxListBox)

BEGIN_EVENT_TABLE(ImageList, wxListBox)
  EVT_RIGHT_DOWN(ImageList::OnRightDown)
  EVT_MENU    (ID_Remove,ImageList::OnRemove)
  EVT_MENU    (ID_SelectAll,ImageList::OnSelectAll)
  EVT_MENU    (ID_RemoveAll,ImageList::OnRemoveAll)
  EVT_MENU    (ID_Load,ImageList::OnLoad)
  EVT_MENU    (ID_Play,ImageList::OnPlay)
  EVT_MENU    (ID_Properties,ImageList::OnProperties)
  EVT_MENU    (ID_Save,ImageList::OnSave)
  EVT_MENU    (ID_Capture,ImageList::OnCapture)
  EVT_MENU    (ID_Calibrate,ImageList::OnCalibrate)
  EVT_LISTBOX(-1, ImageList::OnListSelected)
  EVT_IDLE (ImageList::OnIdle)
END_EVENT_TABLE()



/***************
 * FIXME: Move this stuff somewhere
 ******************************************************/
namespace FocalLengthOptimize{
  int nimages;
  Matrix * worldps;
  Matrix * imageps;
  ImageData ** imds;
  int npts;
  void focal_length_opt(const int & m,
			const int & n,
			double * x,
			double * fvec,
			int & iflag)
  {
    static int which=0;
    int pti=0;
    Matrix intr=imds[0]->intrinsic.copy();
    intr(0,0)=x[0];
    intr(1,1)=x[0];

    double err=0;
    for(int i=0;i<nimages;i++){
      /*
      Matrix rot(3,3);
      Matrix trans(3,1);
      for(int ii=0;ii<3;ii++){
	trans[ii]=imds[i]->extrinsic(ii,3);
	for(int jj=0;jj<3;jj++){
	  rot(ii,jj)=imds[i]->extrinsic(ii,jj);
	}
	}*/
      Matrix rod(6,1);
      for(int k=0;k<6;k++)rod[k]=x[1+k+6*i];
      Matrix origPoints=projectPoints(imds[i]->distortion,intr,rod,worldps[i]);
      for(int j=0;j<worldps[i].n;j++){
	fvec[pti+j]=origPoints(0,j)-imageps[i](0,j);
	fvec[npts+pti+j]=origPoints(1,j)-imageps[i](1,j);

	err+=fvec[pti+j]*fvec[pti+j];
	err+=fvec[npts+pti+j]*fvec[npts+pti+j];
      }
      pti+=worldps[i].n;
    }
    if(which%100==0)
      printf("%f\n",sqrt(err/npts));
    which++;
  }
};
/***************
 * FIXME: Move this stuff too
 ******************************************************/
double getFocalLength(ImageData ** imds,int nci)
{
  int ignore[128];//should only be 14 anyhow
  memset(ignore,0,sizeof(ignore));

  FocalLengthOptimize::worldps=new Matrix[nci];
  FocalLengthOptimize::imageps=new Matrix[nci];
  FocalLengthOptimize::nimages=nci;
  FocalLengthOptimize::imds=imds;

  int npts=0;
  double estimate=0;
  for(int j=0;j<nci;j++){
    CalibInfo * ci=imds[j]->calibInfo;
    groupToPoint(ci->centers,ci->sizes,ci->ncenters,
		 ci->lines,ci->nlines,ci->labels,ignore,
		 FocalLengthOptimize::worldps[j],
		 FocalLengthOptimize::imageps[j]);
    FocalLengthOptimize::worldps[j]=FocalLengthOptimize::worldps[j].transpose();
    FocalLengthOptimize::imageps[j]=FocalLengthOptimize::imageps[j].transpose();
    npts+=FocalLengthOptimize::worldps[j].n;
    estimate+=(imds[j]->intrinsic(0,0)+imds[j]->intrinsic(1,1))/2.0;
  }
  estimate=estimate/nci;

  FocalLengthOptimize::npts=npts;

  Matrix fvec(npts,2);
  Matrix x(1+6*nci,1);
  x[0]=estimate;

  for(int i=0;i<nci;i++){
    Matrix rot(3,3);
    Matrix trans(3,1);
    for(int ii=0;ii<3;ii++){
      trans[ii]=imds[i]->extrinsic(ii,3);
      for(int jj=0;jj<3;jj++){
	rot(ii,jj)=imds[i]->extrinsic(ii,jj);
      }
    }
    Matrix rod=rot.matrix2rod();
    x[1+6*i  ]=rod[0];
    x[1+6*i+1]=rod[1];
    x[1+6*i+2]=rod[2];
    x[1+6*i+3]=trans[0];
    x[1+6*i+4]=trans[1];
    x[1+6*i+5]=trans[2];
  }
  Matrix::lsqnonlin(FocalLengthOptimize::focal_length_opt,x,fvec);
  
  printf("refined focal length is %f (old was %f)\n",x[0],estimate);

  wxString str;
  str.Printf(wxT("Focal Length %f (was:%f)"),x[0],estimate); 
  SetStatusText(str);
  
  for(int i=0;i<nci;i++){
    imds[i]->intrinsic(0,0)=x[0];
    imds[i]->intrinsic(1,1)=x[0];
    
    Matrix rod(3,1);

    for(int ii=0;ii<3;ii++){
      rod[ii]=x[1+6*i+ii];
      imds[i]->extrinsic(ii,3)=x[1+6*i+3+ii];
    }
    Matrix R=rod.rod2matrix();
    for(int ii=0;ii<3;ii++){
      for(int jj=0;jj<3;jj++){
	imds[i]->extrinsic(ii,jj)=R(ii,jj);
      }
    }
  }

  delete [] FocalLengthOptimize::worldps;
  delete [] FocalLengthOptimize::imageps;  
}

/** The listbox containing the calibrated (but unsegmented) images.
 * The CalibImageList behaves similarly to the ImageList, in that
 * all operations are performed by first selecting items and
 * using the right mouse button to select an operation.  The CalibImageList
 * appears in two of the phases of capture (calibration and segmentatino) and
 * thus has two modes.  The only difference between the two modes are
 * the available operations in the popup menu (see CalibImageList::SetCaptureMode(), CalibImageList::SetSegmentMode())). 
 *
 * Items that are removed from the list box are placed in the ImageList.  
 */
class CalibImageList:public wxListBox{
  wxMenu * menu;
  int playing;
  wxString message;
  wxMenuItem * segmentMenuItem;
  SegmentDialog * sd;
public:
  DECLARE_DYNAMIC_CLASS(CalibImageList)
    CalibImageList():wxListBox(){
  }
  CalibImageList(wxWindow * parent):
    wxListBox(parent,ID_CalibImageList,wxDefaultPosition,
	      wxDefaultSize,0,0,wxLB_EXTENDED){

    menu = new wxMenu();
    menu->Append(ID_SelectAll,wxT("Select All"),wxT("Select All"),wxITEM_NORMAL);
    menu->AppendSeparator();
    menu->Append(ID_Remove,wxT("Remove Selected"),wxT("Delete Selected"),wxITEM_NORMAL);
    //menu->Append(ID_RemoveAll,wxT("Remove All"),wxT("Delete Selected"),wxITEM_NORMAL);
    menu->AppendSeparator();
    menu->Append(ID_Play,wxT("Play"),wxT("Play"),wxITEM_NORMAL);
    
#ifdef PARTIAL_LOAD_FROM_LIST
    menu->AppendSeparator();
    menu->Append(ID_Load,wxT("Load"),wxT("Load Images"));
#endif
    menu->AppendSeparator();
#ifdef PARTIAL_SAVE_FROM_LIST
    menu->Append(ID_Save,wxT("Save"),wxT("Save Selected"),wxITEM_NORMAL);
#endif
    menu->Append(6666,wxT("Focal Length"),wxT("Get Focal Length"),wxITEM_NORMAL);
    
    segmentMenuItem = new wxMenuItem(menu,ID_Segment,wxT("Segment"));
	
      sd = new SegmentDialog(this,0);
      sd->Show(FALSE);
      playing=0;
    }
    void SetCaptureMode(){
    
      menu->Remove(segmentMenuItem);
    }
    void SetSegmentMode(){
      //just remove it to avoid having multiple 
      menu->Remove(segmentMenuItem);
      menu->Append(segmentMenuItem);
    }
    void OnIdle(wxIdleEvent & event){
      if(playing){
	if(playing>=GetCount()){
	  playing=0;
	}
	else{
	  static StopWatch watch;
	  static int called=0;
	  if(playing==1&&!called){
	    watch.reset();
	    called=1;
	  }
	  if(watch.hasElapsed(options.playDelay)){
	    wxArrayInt selections;
	    GetSelections(selections);
	    for(int i=selections.GetCount()-1;i>=0;i--){
	      Deselect(selections[i]);
	    }
	    SetSelection(playing);
	    wxCommandEvent dummy;
	    OnListSelected(dummy);
	    
	    playing++;
	    called=0;
	  }
	  event.RequestMore(TRUE);
	}
	/*
	  static struct timeval tv;
	  static int called=0;
	  if(playing==1&&!called){
	    gettimeofday(&tv,NULL);
	    called=1;
	  }
	  static struct timeval now;
	  gettimeofday(&now,NULL);
	
	  unsigned long long span = (now.tv_sec-tv.tv_sec)*1000+
	    (now.tv_usec-tv.tv_usec)/1000;
		
	  if(span>10){
	    wxArrayInt selections;
	    GetSelections(selections);
	    for(int i=selections.GetCount()-1;i>=0;i--){
	      Deselect(selections[i]);
	    }
	    gettimeofday(&tv,NULL);
	    SetSelection(playing);
	    OnListSelected(wxCommandEvent());

	    playing++;
	    called=0;
	  }
	  event.RequestMore(TRUE);
	  }*/
      }
      else if(GetCount()==0){
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	event.RequestMore(FALSE);
      }
      else{
	event.RequestMore(FALSE);
      }
    }
    void OnPlay(wxCommandEvent & event){
      for(int i=0;i<GetCount();i++){
	Deselect(i);
      }
      Select(0);
      playing=1;
      Refresh();
    }
#ifdef PARTIAL_LOAD_FROM_LIST
    void OnLoad(wxMouseEvent & event){
      wxFileDialog fd(GetParent(),wxT("Select file to load"),wxT(""),wxT(""),wxT("Images|*.ppm;*.pgm;*.tga;.*.png"),wxOPEN|wxMULTIPLE);
      if(wxID_OK==fd.ShowModal()){
	wxArrayString names;
	wxArrayString paths;
	fd.GetFilenames(names);
	fd.GetPaths(paths);
	if(GetString(0)==wxT(" ")){
	  Clear();
	}
	for(int i=0;i<names.GetCount();i++){
	  ImageData * imd = new ImageData();
	  imd->image.read((const char *)paths[i].ToAscii());
	  Append(names[i],imd);
	}
      }
      SetChangedSinceSave(IMAGES_CHANGED);
    }
#endif
#ifdef PARTIAL_SAVE_FROM_LIST
    void OnSave(wxCommandEvent & event){
      wxArrayInt sel;
      GetSelections(sel);
      if(sel.GetCount()==1){
	wxFileDialog fd(GetParent(),"Select filename to save", "","image.tga","*.*",wxSAVE);
	if(wxID_OK==fd.ShowModal()){
	  wxString path=fd.GetPath();
	  save((ImageData *)GetClientData(GetSelectionSafe(this)),(const char *)path.ToAscii());
	}
      }
      else {
	wxFileDialog fd(GetParent(),"Select directory and enter printf command", "","image-%04d.tga","*.*",wxSAVE);
      
	if(wxID_OK==fd.ShowModal()){
	  wxString paths=fd.GetPath();
	  wxString path;
	  for(int i=0;i<sel.GetCount();i++){
	    path.sprintf(paths,i);
	    printf("%s\n",(const char*)paths.ToAscii());
	    printf("%s\n",(const char*)path.ToAscii());
	    save((ImageData *)GetClientData(sel[i]),(const char *)path.ToAscii(),1,1,1);
	  }
	}
	else{
	  printf("fd.ShowModal returned error\n");
	}
      }
      SetChangedSinceSave(IMAGES_CHANGED);
    }
#endif

#ifdef OLD_SEL
  void OnListSelected(wxCommandEvent & event){
    
    int i=GetSelectionSafe(this);
    
    //So texture doesn't get updated for multi selections
    wxArrayInt sel;
    GetSelections(sel);
    int n=sel.GetCount();
    if(n>1)return;
    
    if(i>=0 && GetClientData(i))
      glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
  } 
#else
  void OnListSelected(wxCommandEvent & event){
    static wxArrayInt selLast;
    int i;
    wxArrayInt sel;
    GetSelections(sel);
    
    for(i=0;i<sel.GetCount();i++){
      if(i<selLast.GetCount()){
	if(sel[i]!=selLast[i])
	  break;
      }else break;
    }
    if(i>=sel.GetCount())
      i=GetSelection();
    else
      i=sel[i];
    if(i>=0 && GetClientData(i)){
      glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
    }
    selLast=sel;
  }
#endif
    void OnRightDown(wxMouseEvent & event){
      wxPoint pt=wxGetMousePosition();
      PopupMenu(menu,(ScreenToClient(pt)));
      //if(playing)playing=0;
    }
    void SendToImageList(int i){
      if(imageList->GetString(0)==wxT(" "))
	imageList->Clear();
      imageList->Append(GetString(i),
			GetClientData(i));
    }
    void OnRemove(wxCommandEvent & event){
      wxArrayInt selections;
      GetSelections(selections);
      for(int i=0;i<selections.GetCount();i++){
	SendToImageList(selections[i]);
      }
      for(int i=selections.GetCount()-1;i>=0;i--){
	Delete(selections[i]);
      }
      SetChangedSinceSave(IMAGES_CHANGED);
    }
    void OnRemoveAll(wxCommandEvent & event){
      for(int i=0;i<GetCount();i++){
	SendToImageList(i);
      }
      Clear();
      SetChangedSinceSave(IMAGES_CHANGED);
    }
    void OnSelectAll(wxCommandEvent & event){
      for(int i=0;i<GetCount();i++){
        Select(i);
      }
    }

  void OnFocalLength(wxCommandEvent & event){
    printf("getting the focal length\n");
    wxArrayInt sels;
    GetSelections(sels);
    ImageData ** imds=new ImageData *[sels.GetCount()];
    for(int i=0;i<sels.GetCount();i++){
      imds[i]=(ImageData *)GetClientData(sels[i]);
    }
    getFocalLength(imds,sels.GetCount());
    delete [] imds;
  }
    void OnSegment(wxCommandEvent & event){
      printf("showing\n");
      int sel=GetSelectionSafe(this);
      if(sel>=0 && GetClientData(sel)){
	sd->Show(TRUE);
      }
      printf("done showing\n");
    }
    DECLARE_EVENT_TABLE();
  };
IMPLEMENT_DYNAMIC_CLASS(CalibImageList,wxListBox)

  BEGIN_EVENT_TABLE(CalibImageList, wxListBox)
  EVT_RIGHT_DOWN(CalibImageList::OnRightDown)
  EVT_MENU    (ID_Remove,CalibImageList::OnRemove)
  EVT_MENU    (ID_RemoveAll,CalibImageList::OnRemoveAll)
  EVT_MENU    (ID_SelectAll,CalibImageList::OnSelectAll)
#ifdef PARTIAL_LOAD_FROM_LIST
  EVT_MENU    (ID_Load,CalibImageList::OnLoad)
#endif
  EVT_MENU    (ID_Play,CalibImageList::OnPlay)
#ifdef PARTIAL_SAVE_FROM_LIST
  EVT_MENU    (ID_Save,CalibImageList::OnSave)
#endif
  EVT_MENU    (6666,CalibImageList::OnFocalLength)
  EVT_MENU    (ID_Segment,CalibImageList::OnSegment)
  EVT_LISTBOX(-1, CalibImageList::OnListSelected)
  EVT_IDLE (CalibImageList::OnIdle)
END_EVENT_TABLE()

/** The listbox containing the calibrated and segmented images.
 * The behavior of the SegImageList is similar to that of the ImageList
 * and CalibImageList classes.  
 *
 * Removed items are sent to the CalibImageList box.  
 */
class SegImageList:public wxListBox{
    wxMenu * menu;
    int playing;
    wxString message;
  public:
    DECLARE_DYNAMIC_CLASS(SegImageList)
      SegImageList():wxListBox(){
    }
    SegImageList(wxWindow * parent):
      wxListBox(parent,ID_CalibImageList,wxDefaultPosition,
		wxDefaultSize,0,0,wxLB_EXTENDED){
      menu = new wxMenu();
      menu->Append(ID_SelectAll,wxT("Select All"),wxT("Select All"),wxITEM_NORMAL);
      menu->Append(ID_Remove,wxT("Remove Selected"),wxT("Delete Selected"),wxITEM_NORMAL);
      menu->Append(ID_RemoveAll,wxT("Remove All"),wxT("Delete Selected"),wxITEM_NORMAL);
      menu->AppendSeparator();
      menu->Append(ID_Play,wxT("Play"),wxT("Play"),wxITEM_NORMAL);
      menu->AppendSeparator();
      menu->Append(6666,wxT("Focal Length"),wxT("Get the focal length"),wxITEM_NORMAL);
      //menu->Append(ID_Play,wxT("BW Tool"),wxT("BW Tool"),wxITEM_NORMAL);
      //menu->AppendSeparator();
      playing=0;
    }
    void OnIdle(wxIdleEvent & event){
      if(playing){
	if(playing>=GetCount()){
	  playing=0;
	}
	else{
	  static StopWatch watch;
	  static int called=0;
	  if(playing==1&&!called){
	    watch.reset();
	    called=1;
	  }
	  if(watch.hasElapsed(options.playDelay)){
	    wxArrayInt selections;
	    GetSelections(selections);
	    for(int i=selections.GetCount()-1;i>=0;i--){
	      Deselect(selections[i]);
	    }
	    SetSelection(playing);
	    wxCommandEvent dummy;
	    OnListSelected(dummy);
	    
	    playing++;
	    called=0;
	  }
	  event.RequestMore(TRUE);
	}
      }
      else if(GetCount()==0){
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	Append(wxT(" "));
	event.RequestMore(FALSE);
      }
      else{
	event.RequestMore(FALSE);
      }
    }
    void OnPlay(wxCommandEvent & event){
      for(int i=0;i<GetCount();i++){
	Deselect(i);
      }
      Select(0);
      playing=1;
      Refresh();
    }
#ifdef OLD_SEL
  void OnListSelected(wxCommandEvent & event){
    
    int i=GetSelectionSafe(this);
    
    //So texture doesn't get updated for multi selections
    wxArrayInt sel;
    GetSelections(sel);
    int n=sel.GetCount();
    if(n>1)return;
    
    if(i>=0 && GetClientData(i))
      glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
  } 
#else
  void OnListSelected(wxCommandEvent & event){
    static wxArrayInt selLast;
    int i;
    wxArrayInt sel;
    GetSelections(sel);
    
    for(i=0;i<sel.GetCount();i++){
      if(i<selLast.GetCount()){
	if(sel[i]!=selLast[i])
	  break;
      }else break;
    }
    if(i>=sel.GetCount())
      i=GetSelectionSafe(this);
    else
      i=sel[i];
    if(i>=0 && GetClientData(i)){
      glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
    }
    selLast=sel;
  }
#endif
    
    void OnRightDown(wxMouseEvent & event){
      wxPoint pt=wxGetMousePosition();
      PopupMenu(menu,(ScreenToClient(pt)));
      //if(playing)playing=0;
    }
    void SendToCalibImageList(int i){
      if(calibImageList->GetString(0)==wxT(" "))
	calibImageList->Clear();
      ImageData * imd = (ImageData *)GetClientData(i);
      if(imd)imd->unset(HAS_MASK);
      calibImageList->Append(GetString(i),
			     imd);
    }
    void OnRemove(wxCommandEvent & event){
      wxArrayInt selections;
      GetSelections(selections);
      for(int i=0;i<selections.GetCount();i++){
	SendToCalibImageList(selections[i]);
      }
      for(int i=selections.GetCount()-1;i>=0;i--){
	Delete(selections[i]);
      }
      SetChangedSinceSave(IMAGES_CHANGED);
    }
    void OnRemoveAll(wxCommandEvent & event){
      for(int i=0;i<GetCount();i++){
	SendToCalibImageList(i);
      }
      Clear();
      SetChangedSinceSave(IMAGES_CHANGED);
    }
    void OnSelectAll(wxCommandEvent & event){
      for(int i=0;i<GetCount();i++){
	Select(i);
      }
    }
  void OnFocalLength(wxCommandEvent & event){
    printf("getting the focal length\n");
    wxArrayInt sels;
    GetSelections(sels);
    ImageData ** imds=new ImageData *[sels.GetCount()];
    for(int i=0;i<sels.GetCount();i++){
      imds[i]=(ImageData *)GetClientData(sels[i]);
    }
    getFocalLength(imds,sels.GetCount());
    delete [] imds;
  }
    DECLARE_EVENT_TABLE();
  };
IMPLEMENT_DYNAMIC_CLASS(SegImageList,wxListBox)

BEGIN_EVENT_TABLE(SegImageList, wxListBox)
  EVT_RIGHT_DOWN(SegImageList::OnRightDown)
  EVT_MENU    (ID_SelectAll,SegImageList::OnSelectAll)
  EVT_MENU    (ID_Remove,SegImageList::OnRemove)
  EVT_MENU    (ID_RemoveAll,SegImageList::OnRemoveAll)
  EVT_MENU    (ID_Play,SegImageList::OnPlay)
  EVT_MENU    (6666,SegImageList::OnFocalLength)
  EVT_LISTBOX(-1, SegImageList::OnListSelected)
  EVT_IDLE (SegImageList::OnIdle)
END_EVENT_TABLE()


typedef enum {CAPTURE,SEGMENTATION,MODEL,
		TEXTURE,NUM_STAGES} Stage;

/** Main window of the application.
 *
 */
class MyFrame: public wxFrame,OKHandler
{
  enum {TB_Capture=110,TB_Segment,TB_Geometry,TB_Texture};

  int changingStage;
  wxPanel * page[4];
  wxSplitterWindow * vertSplit;
  wxToolBar * toolBar;
  wxMenuBar * menuBar;
  
  Stage stage;
  wxSplitterWindow * leftSplit;
  wxSplitterWindow * leftPanelModel;
  wxPanel * leftPanelTexture;
  StatusBar * statusBar;
  wxCheckBox * ksfscb;

  //Controls for the model panel
  wxNotebook * triNb,* geomNb;
  wxTextCtrl * geomCommand;
  wxSpinCtrl * nrays;
  //wxSpinCtrl * percentRemove;

  BBDialog * bbd;

  //Controls for the texture panel
  wxChoice * splitChoice, * mapChoice;
  wxSplitterWindow  * splitSplitter,* mapSplitter;
  wxButton * mapButton,*splitButton;
  wxCheckBox * normalsCheck,*distanceCheck;
  wxRadioButton * lscmCheck,* mdsCheck, *svdCheck;
  wxSpinCtrl * coneAngle,* maxFaces;
  wxListBox  * selCharts;
  wxTextCtrl * selChartInfo;
  wxButton * pack,* stretch,* spread, * removeOverlap;
  
  //Plugin stuff
  GeometryPlugin * geomPlugsCreate, * geomPlugsOp;
  int nGeomPlugsCreate,nGeomPlugsOp;
  TexturePlugin * texPlugsSplit,* texPlugsMap,* texPlugsOther;
  int nTexPlugsSplit,nTexPlugsMap,nTexPlugsOther;
public:
  void loadGeomPlugs(){
    char plugDir[2048];

    printf("Looking in %s\n",PLUGDIR);    
    snprintf(plugDir,2048,"%s/.capsys/%s",(const char *)wxGetHomeDir().ToAscii(),ARCH);//getenv("HOME"));
    printf("  and also %s\n", plugDir);
    wxArrayString plugs=
      findPlugins(plugDir,"_geom");
    WX_APPEND_ARRAY(plugs,
		    findPlugins(PLUGDIR,"_geom"));
    printf("got potential geom plugs %d\n",plugs.Count());

    nGeomPlugsOp=0;
    nGeomPlugsCreate=0;
    for(int j=0;j<plugs.Count();j++){
      GeometryPlugin p;

      printf("Inspecting: %s\n",
	     (const char *)plugs[j].ToAscii());

      if(!p.open((const char *)plugs[j].ToAscii()) ||
	 !p.load()){
	wxString str;
	str = wxT("unable to load plugin ")+
	  wxString(plugs[j])+
	  wxString::Format(wxT(" (opened:%d)\n"),p.loaded());
	wxMessageBox(str,wxT("Notice"));
      }
      else{
	int type=p.getType();
	nGeomPlugsCreate+=(type&GEOMPLUGIN_CREATE_TYPE);
	nGeomPlugsOp+=(type&GEOMPLUGIN_OPERATION_TYPE);
      }	
      p.close();
    }
    geomPlugsCreate = new GeometryPlugin[nGeomPlugsCreate];
    geomPlugsOp = new GeometryPlugin[nGeomPlugsOp];
    nGeomPlugsCreate=nGeomPlugsOp=0;
    for(int j=0;j<plugs.Count();j++){
      GeometryPlugin p;
      //just to be safe, do it the shitty way
      if(p.open((const char *)plugs[j].ToAscii()) &&
	 p.load()){
	int type=p.getType();
	p.close();
	switch(type){
	case GEOMPLUGIN_CREATE_TYPE:
	  geomPlugsCreate[nGeomPlugsCreate].open((const char *)plugs[j].ToAscii());
	  geomPlugsCreate[nGeomPlugsCreate++].load();
	  break;
	case GEOMPLUGIN_OPERATION_TYPE:
	  geomPlugsOp[nGeomPlugsOp].open((const char *)plugs[j].ToAscii());
	  geomPlugsOp[nGeomPlugsOp++].load();
	  break;
	}
      }
    }
    printf("Loaded %d %d Geometry plugs\n",nGeomPlugsCreate,
	   nGeomPlugsOp);
  }

  void loadTexPlugs(){
    char plugDir[2048];
    snprintf(plugDir,2048,"%s/.capsys/%s", (const char*)wxGetHomeDir().ToAscii(),ARCH);//getenv("HOME"));
    printf("Looking for plugins in %s\n",PLUGDIR);
    wxArrayString plugs=
      findPlugins(plugDir,"_tex");
    WX_APPEND_ARRAY(plugs,
		    findPlugins(PLUGDIR,"_tex"));
    printf("got potential tex plugs %d\n",plugs.Count());
   
    nTexPlugsSplit=0;
    nTexPlugsMap=0;
    nTexPlugsOther=0;
    for(int j=0;j<plugs.Count();j++){
      TexturePlugin p;
      printf("Inspecting: %s\n",
	     (const char *)plugs[j].ToAscii());

      if(!p.open((const char *)plugs[j].ToAscii()) ||
	 !p.load())
	;
      else{
	int type=p.getType();
	nTexPlugsSplit+=(type==TEXPLUGIN_SPLIT_TYPE);
	nTexPlugsMap+=(type==TEXPLUGIN_MAP_TYPE);
	nTexPlugsOther+=(type==TEXPLUGIN_OTHER_TYPE);
      }
      p.close();
    }
    texPlugsSplit=new TexturePlugin[nTexPlugsSplit];
    texPlugsMap=new TexturePlugin[nTexPlugsMap];
    texPlugsOther=new TexturePlugin[nTexPlugsOther];
    nTexPlugsSplit=nTexPlugsMap=nTexPlugsOther=0;
    for(int j=0;j<plugs.Count();j++){
      TexturePlugin p;
      //just to be safe, do it the shitty way
      if(p.open((const char *)plugs[j].ToAscii()) &&
	 p.load()){
	int type=p.getType();
	p.close();
	
	switch(type){
	case TEXPLUGIN_SPLIT_TYPE:
	  texPlugsSplit[nTexPlugsSplit].open((const char *)plugs[j].ToAscii());
	  texPlugsSplit[nTexPlugsSplit++].load();
	  break;
	case TEXPLUGIN_MAP_TYPE:
	  texPlugsMap[nTexPlugsMap].open((const char *)plugs[j].ToAscii());
	  texPlugsMap[nTexPlugsMap++].load();
	  break;
	case TEXPLUGIN_OTHER_TYPE:
	  texPlugsOther[nTexPlugsOther].open((const char *)plugs[j].ToAscii());
	  texPlugsOther[nTexPlugsOther++].load();
	  break;
	}
      }
    }
    printf("Loaded %d %d %d texture plugs\n",
	   nTexPlugsSplit,nTexPlugsMap,nTexPlugsOther);
  }

  wxBitmap c2bitmap(const imstruct & im){
    unsigned char * data = new unsigned char[im.width*im.height*3];
    unsigned char * mask = new unsigned char[im.width*im.height*3];
    for(int i=0;i<im.width*im.height;i++){
      data[3*i]=im.data[4*i];
      data[3*i+1]=im.data[4*i+1];
      data[3*i+2]=im.data[4*i+2];
      mask[3*i  ]=im.data[4*i+3]>128;
      mask[3*i+1]=im.data[4*i+3]>128;
      mask[3*i+2]=im.data[4*i+3]>128;
    }
    wxImage maskImage(im.width,im.height,mask,(bool)FALSE);

    wxImage image(im.width,im.height,
		  data,(bool)FALSE);
    image.SetMaskFromImage(maskImage,0,0,0);
    return wxBitmap(image);
  }
  MyFrame(const wxString& title,
	  const wxPoint& pos, const wxSize& size)
    : wxFrame((wxFrame *)NULL, -1, title, pos, size){
    //First things first, load the plugins
    loadGeomPlugs();
    loadTexPlugs();

    wxMenu *menuFile = new wxMenu;
    wxMenu *menuOptions = new wxMenu;
    wxMenu *menuHelp = new wxMenu;
    menuFile->Append(ID_New,wxT("&New"));
    menuFile->Append(ID_Load,wxT("&Load"));
    menuFile->AppendSeparator();
    menuFile->Append(ID_Save,wxT("&Save"));
    menuFile->Append(ID_SaveAs,wxT("Save &As"));
    menuFile->Append(ID_Export,wxT("&Export"));
    menuFile->AppendSeparator();
    menuFile->Append( ID_Quit, wxT("E&xit") );
    
    menuOptions->Append(ID_Settings,wxT("&Settings"));
    menuOptions->Append(ID_Camera,wxT("&Cameras"));

    menuHelp->Append(ID_About, wxT("&About"));

    menuBar = new wxMenuBar;
    menuBar->Append( menuFile, wxT("&File") );
    menuBar->Append( menuOptions,wxT("&Options"));
    menuBar->Append( menuHelp,wxT("&Help"));

    SetMenuBar( menuBar );
       
    statusBar = new StatusBar(this);
    SetStatusBar(statusBar);
   
    imageList = new ImageList(this);
    calibImageList = new CalibImageList(this);
    segImageList = new SegImageList(this);
    
    leftSplit=0;
    vertSplit = new wxSplitterWindow(this,-1,wxDefaultPosition,
				     wxDefaultSize,
				     wxSP_NOBORDER|wxSP_3DSASH);
    
    leftPanelModel = makeLeftPanelModel();
    leftPanelTexture = makeLeftPanelTexture();
    
    //wxSplitterWindow * split = new wxSplitterWindow(vertSplit,-1);

    //wxPanel *topPanel=new wxPanel(split,ID_TopPanel);
    //wxPanel *botPanel=new wxPanel(split,ID_BotPanel);

    glCanvas = new GLCanvas(vertSplit,attribs);
    vertSplit->SetMinimumPaneSize(50);

    wxBoxSizer * boxsizer = new wxBoxSizer(wxHORIZONTAL);
    boxsizer->Add(vertSplit,1,wxGROW);
    SetSizer(boxsizer);
    
    toolBar=new wxToolBar(this,-1,wxDefaultPosition,wxDefaultSize,
			  wxTB_HORIZONTAL | wxNO_BORDER | wxTB_DOCKABLE |wxTB_FLAT); 

    toolBar->SetToolBitmapSize(wxSize(32,32));
    toolBar->AddTool(ID_New,wxT("New"),
		     c2bitmap(new_image),
		     wxT("New project"),wxITEM_NORMAL);
    toolBar->AddTool(ID_Load,wxT("Open"),
		     c2bitmap(open_image),
		     wxT("Open project"),wxITEM_NORMAL);
    toolBar->AddTool(ID_Save,wxT("Save"),
		     c2bitmap(save_image),
		     wxT("Save project"),wxITEM_NORMAL);
    toolBar->AddTool(ID_SaveAs,wxT("Save As"),
		     c2bitmap(saveas_image),
		     wxT("Save project As"),wxITEM_NORMAL);
    toolBar->AddSeparator();
    toolBar->AddSeparator();
    toolBar->AddSeparator();
    toolBar->AddSeparator();
    toolBar->AddSeparator();
    toolBar->AddSeparator();
    
    toolBar->AddTool(TB_Capture,wxT("Cap"),
		     c2bitmap(capture_image),
		     wxT("Capture and calibrate images"),wxITEM_CHECK);
    toolBar->AddTool(TB_Segment,wxT("Seg"),
		     c2bitmap(seg_image),
		     wxT("Segment images"),wxITEM_CHECK);
    toolBar->AddTool(TB_Geometry,wxT("Geo"),
		     c2bitmap(geometry_image),
		     wxT("Obtain geometry"),wxITEM_CHECK);
    toolBar->AddTool(TB_Texture,wxT("Tex"),
		     c2bitmap(tex_image),
		     wxT("Get texture coordinates"),wxITEM_CHECK);
    toolBar->Realize();
        
    //toolBar->AddTool(ID_Next,"Tool",wxBitmap(wxImage("next.bmp")));
    SetToolBar(toolBar);
    
    setStage(CAPTURE);
    bbd = new BBDialog(NULL,0,this);
    UpdateTitle();
  }
  void UpdateTitle(){
    wxString title;
    int nuncal=GetUncalImageCount();
    int ncal=GetCalImageCount();
    int nseg=GetSegImageCount();

    title = wxT("Project: ")+wxString::FromAscii(currentProj)+
      wxString::Format(wxT("(%d,%d,%d,%d)"),
		       nuncal,ncal,nseg,ncharts);
    SetTitle(title);
  }
  wxPanel * makeLeftPanelTexture(){
    wxPanel * panel = new wxPanel(this,-1,wxDefaultPosition,
				  wxDefaultSize,
				  wxSP_NOBORDER);
    
    wxBoxSizer * panelSizer = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer * sizer = new wxStaticBoxSizer(new wxStaticBox(panel,-1,wxT("Current Chart")),
					      wxVERTICAL);
    wxNotebook * nbbot = new wxNotebook(panel,-1);
    //wxNotebookSizer * nbbotsizer = new wxNotebookSizer(nbbot);
    wxPanel * splitPage = new wxPanel(nbbot,-1);
    wxPanel * mapPage = new wxPanel(nbbot,-1);
    wxPanel * otherButtonsPage = new wxPanel(nbbot,-1);

    selCharts = new wxListBox(panel,ID_SelCharts
			      ,wxDefaultPosition,wxDefaultSize,0,0,wxLB_EXTENDED);

    sizer->Add(selCharts,1,wxGROW);
    sizer->Add(selChartInfo=new wxTextCtrl(panel,
					   ID_ChartInfo,
					   wxT("Chart Info"),
					   wxDefaultPosition,
					   wxDefaultSize,
					   wxTE_MULTILINE|wxTE_READONLY),
	       1,wxGROW);
    panelSizer->Add(sizer,1,wxGROW);
    
    //Bottom half, page0=split operations
    sizer = new wxStaticBoxSizer(new wxStaticBox(panel,-1,wxT("Operations")),
				 wxVERTICAL);
	 
    wxBoxSizer * sizer1 = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer * sizer2 = new wxStaticBoxSizer(new wxStaticBox(splitPage,-1,wxT("Method")),
					       wxVERTICAL);
    wxBoxSizer * sizer3;
    splitSplitter = new wxSplitterWindow(splitPage,-1,
				      wxDefaultPosition,
				      wxDefaultSize,
				      wxSP_NOBORDER);
    splitChoice=new wxChoice(splitPage,ID_SplitChoice,
			     wxDefaultPosition,
			     wxDefaultSize);
    printf("getting normals options panel\n");
    
    wxPanel * normalsOptionPanel = new wxPanel(splitSplitter);
    wxBoxSizer * normalsOPSizer 
      = new wxStaticBoxSizer(new wxStaticBox(normalsOptionPanel,
					     -1,wxT("Options")),
			     wxVERTICAL);
    sizer3 = new wxBoxSizer(wxHORIZONTAL);
    sizer3->Add(new wxStaticText(normalsOptionPanel,
				 -1,wxT("cone angle")));
    sizer3->Add(coneAngle=new wxSpinCtrl(normalsOptionPanel,-1),
		1,wxGROW);
    coneAngle->SetRange(2,90);
    coneAngle->SetValue(45);
    normalsOPSizer->Add(sizer3,0,wxGROW);

    sizer3 = new wxBoxSizer(wxHORIZONTAL);
    sizer3->Add(new wxStaticText(normalsOptionPanel,-1,wxT("max faces")));
    sizer3->Add(maxFaces=new wxSpinCtrl(normalsOptionPanel,-1),1,wxGROW);
    normalsOPSizer->Add(sizer3,0,wxGROW);
    normalsOptionPanel->SetSizer(normalsOPSizer);
    maxFaces->SetRange(1,100000);
    maxFaces->SetValue(10000);
    splitChoice->Append(wxT("Normals"),normalsOptionPanel);
    
    /*printf("getting distance option panel\n");
    wxPanel * distanceOptionPanel=new wxPanel(splitSplitter);
    sizer3 = new wxBoxSizer(wxHORIZONTAL);
    wxStaticText * text=new wxStaticText(distanceOptionPanel,-1,
					 "No Options");
    sizer3->Add(text,0,wxGROW|wxLEFT|wxRIGHT|wxALIGN_CENTER,3);
    distanceOptionPanel->SetSizer(sizer3);
    splitChoice->Append(wxString("Distance"),distanceOptionPanel);
    */

    for(int i=0;i<nTexPlugsSplit;i++){
      wxPanel * panel=texPlugsSplit[i].getPanel(this);//splitSplitter);
      //Make a default panel, that says No Options
      if(!panel){
	panel=new wxPanel(this);//splitSplitter);
	sizer3 = new wxBoxSizer(wxHORIZONTAL);
	wxStaticText * text=new wxStaticText(panel,
					     -1,
					     wxT("No Options"));
	sizer3->Add(text,0,wxGROW|wxLEFT|wxRIGHT|wxALIGN_CENTER,3);
	panel->SetSizer(sizer3);
      }
      panel->Show(FALSE);
      texPlugsSplit[i].setUserData(panel);
      splitChoice->Append(wxString::FromAscii(texPlugsSplit[i].getName()),panel);
    }
    splitChoice->SetSelection(0);
    {wxCommandEvent ev;OnSplitChoice(ev);}//Force the event
    sizer2->Add(splitChoice,0,wxGROW);
    sizer1->Add(sizer2,0,wxGROW);
    /*
    sizer2->Add(normalsCheck=new wxCheckBox(splitPage,ID_Normals,"normals"),1,wxGROW);
   
    sizer3 = new wxBoxSizer(wxHORIZONTAL);
    sizer3->Add(new wxStaticText(splitPage,-1,"cone angle"));
    sizer3->Add(coneAngle=new wxSpinCtrl(splitPage,-1),1,wxGROW);
    sizer2->Add(sizer3,0,wxGROW);
    coneAngle->SetRange(2,90);
    coneAngle->SetValue(45);

    sizer3 = new wxBoxSizer(wxHORIZONTAL);
    sizer3->Add(new wxStaticText(splitPage,-1,"max faces"));
    sizer3->Add(maxFaces=new wxSpinCtrl(splitPage,-1),1,wxGROW);
    sizer2->Add(sizer3,0,wxGROW);
    maxFaces->SetRange(1,100000);
    maxFaces->SetValue(10000);
    */
    /*
    sizer1->Add(sizer2,0,wxGROW);
    sizer2=new wxStaticBoxSizer(new wxStaticBox(splitPage,-1,wxT("")),
				wxVERTICAL);
    sizer2->Add(distanceCheck=new wxCheckBox(splitPage,ID_Distance,"distance"));
    sizer1->Add(sizer2,0,wxGROW); */

    sizer1->Add(splitSplitter,1,wxALIGN_CENTER|wxGROW|wxEXPAND);
    sizer1->Add(splitButton=new wxButton(splitPage,ID_Split,wxT("Split")),0,wxALIGN_CENTER|wxTOP|wxBOTTOM,4);
    splitPage->SetSizer(sizer1);

    //Bottom half, page1=Map operations
    sizer1 = new wxBoxSizer(wxVERTICAL);
    sizer2 = new wxStaticBoxSizer(new wxStaticBox(mapPage,-1,wxT("Method")),
				  wxVERTICAL);

    /*
    //NOTE, if the panel returns something, it could be handled
    //in a similar manner as the splitting was.
    for(int j=0;j<nTexPlugsMap;j++){
      wxRadioButton * radical
	=new wxRadioButton(mapPage,0,texPlugsMap[j].getName());
      radical->SetToolTip(texPlugsMap[j].getInfo());
      texPlugsMap[j].setUserData(radical);
      sizer2->Add(radical);
      radical->SetValue(j==0);//the first button is selected
    }
    */
    mapSplitter = new wxSplitterWindow(mapPage,-1,
				      wxDefaultPosition,
				      wxDefaultSize,
				      wxSP_NOBORDER);
    mapChoice=new wxChoice(mapPage,ID_MapChoice,
			   wxDefaultPosition,
			   wxDefaultSize);

    for(int i=0;i<nTexPlugsMap;i++){
      wxPanel * panel=texPlugsMap[i].getPanel(this);//mapSplitter);
      //Make a default panel, that says No Options
      if(!panel){
	panel=new wxPanel(this);//mapSplitter);
	sizer3 = new wxBoxSizer(wxHORIZONTAL);
	wxStaticText * text=new wxStaticText(panel,
					     -1,
					     wxT("No Options"));
	sizer3->Add(text,0,wxGROW|wxLEFT|wxRIGHT|wxALIGN_CENTER,3);
	panel->SetSizer(sizer3);
      }
      panel->Show(FALSE);
      //panel->SetClientData(texPlugsMap+i);
      texPlugsMap[i].setUserData(panel);
      mapChoice->Append(wxString::FromAscii(texPlugsMap[i].getName()),panel);
    }
    mapChoice->SetSelection(0);
    {wxCommandEvent ev;OnMapChoice(ev);}//Force the event
    sizer2->Add(mapChoice,0,wxGROW);
    sizer1->Add(sizer2,0,wxGROW);

    //sizer1->Add(sizer2,0,wxGROW);
    sizer1->Add(mapSplitter,1,wxGROW|wxEXPAND);
    sizer1->Add(mapButton=new wxButton(mapPage,ID_Map,wxT("Map")),0,wxALIGN_CENTER|wxTOP|wxBOTTOM,4);
    mapPage->SetSizer(sizer1);

    sizer1 = new wxStaticBoxSizer(new wxStaticBox(otherButtonsPage,-1,wxT("")),
				  wxVERTICAL);
    
    pack = new wxButton(otherButtonsPage,ID_Pack,wxT("Pack"));
    stretch = new wxButton(otherButtonsPage,ID_Stretch,wxT("Stretch"));
    spread = new wxButton(otherButtonsPage,ID_Spread,wxT("Spread"));
    removeOverlap = new wxButton(otherButtonsPage,ID_RemoveOverlap,wxT("Remove Overlap"));
    sizer1->Add(pack,0,wxGROW);
    sizer1->Add(stretch,0,wxGROW);
    sizer1->Add(spread,0,wxGROW);
    sizer1->Add(removeOverlap,0,wxGROW);

    printf("adding other tex plugins\n");
    for(int i=0;i<nTexPlugsOther;i++){
      wxPanel * page = new wxPanel(nbbot);
      wxPanel * panel = texPlugsOther[i]
	.getPanel(page);
      if(panel){
	printf("adding a page\n");
	wxButton * apply = new wxButton(page,ID_TexApply,wxT("Apply"));
	wxSizer * sizer = new wxBoxSizer(wxVERTICAL);
	panel->SetToolTip(wxString::FromAscii(texPlugsOther[i].getInfo()));
	apply->SetToolTip(wxT("Apply the ")+
			  wxString::FromAscii(texPlugsOther[i].getName())+
			  wxT(" operation"));
	printf("adding to sizer\n");
	sizer->Add(panel,1,wxGROW|wxALIGN_CENTER);
	sizer->Add(apply,0,wxALIGN_CENTER);
	page->SetSizer(sizer);
	nbbot->AddPage(page,wxString::FromAscii(texPlugsOther[i].getName()));
	texPlugsOther[i].setUserData(apply);
	printf("done adding\n");
      }
      else{
	wxButton * button=
	  new wxButton(otherButtonsPage,ID_TexApply,
		       wxString::FromAscii(texPlugsOther[i].getName()));
	button->SetToolTip(wxString::FromAscii(texPlugsOther[i].getInfo()));
	texPlugsOther[i].setUserData(button);
	sizer1->Add(button,0,wxGROW);
	//don't need the page
	delete page;
      }
    }

    otherButtonsPage->SetSizer(sizer1);//0,wxLEFT|wxRIGHT|wxTOP|wxBOTTOM,5);

    nbbot->InsertPage(0,splitPage,wxT("Split"),TRUE);
    nbbot->InsertPage(1,mapPage,wxT("Map"),FALSE);
    nbbot->AddPage(otherButtonsPage,wxT("Other"),FALSE);
    
    sizer->Add(nbbot,1,wxGROW);
    panelSizer->Add(sizer,2,wxGROW);
	 
    panel->SetSizer(panelSizer);

    //normalsCheck->SetValue(1);
    //distanceCheck->SetValue(0);

    //lscmCheck->SetValue(1);
    //mdsCheck->SetValue(0);
    //svdCheck->SetValue(0);
    return panel;
  }
  wxSplitterWindow * makeLeftPanelModel(){
    wxSplitterWindow * panel = new wxSplitterWindow(this,-1,wxDefaultPosition,
						    wxDefaultSize,
						    wxSP_NOBORDER);
    wxPanel * topPanel = new wxPanel(panel,-1);
    wxPanel * botPanel = new wxPanel(panel,-1);

    wxBoxSizer * topSizer = new wxStaticBoxSizer(new wxStaticBox(topPanel,-1,wxT("Geometry")),
						 wxVERTICAL);
    wxBoxSizer * botSizer = new wxStaticBoxSizer(new wxStaticBox(botPanel,-1,wxT("Operations")),
						 wxVERTICAL);
    wxNotebook * nbtop = new wxNotebook(topPanel,-1);
    wxNotebook * nbbot = new wxNotebook(botPanel,-1);
    wxPanel * toppage0 = new wxPanel(nbtop,-1);
    //wxPanel * botpage0 = new wxPanel(nbbot,-1);
    wxPanel * botpage1 = new wxPanel(nbbot,-1);

    //wxNotebookSizer * nbtopsizer = new wxNotebookSizer(nbtop);
    //wxNotebookSizer * nbbotsizer = new wxNotebookSizer(nbbot);

    //Geometry notebook, SFS(MI)
    wxBoxSizer * sizer; 
    //OLD, used to have options//=new wxStaticBoxSizer(new wxStaticBox(toppage0,-1,"Options"),
    //wxVERTICAL);
    for(int i=0;i<nGeomPlugsCreate;i++){
      wxPanel * panel = geomPlugsCreate[i].getPanel(nbtop);
      if(!panel){
	wxMessageBox(wxT("Invalid plugin:")
		     +wxString::FromAscii(geomPlugsCreate[i].getName()) 
		     ,wxT("Error"),
		     wxICON_ERROR|wxOK,NULL);
	panel = new wxPanel(nbtop);
      }
      panel->SetToolTip(wxString::FromAscii(geomPlugsCreate[i].getInfo()));
      nbtop->AddPage(panel,wxString::FromAscii(geomPlugsCreate[i].getName()),
		     (i==0)?TRUE:FALSE);
    }
    /*
    wxBoxSizer * row = new wxBoxSizer(wxHORIZONTAL);
    nrays = new wxSpinCtrl(toppage0,ID_Nrays);
    nrays->SetRange(0,512);
    nrays->SetValue(32);
    row->Add(new wxStaticText(toppage0,-1,"# rays"),0);
    row->Add(nrays);
    sizer->Add(row);
    sizer->Add(ksfscb=new wxCheckBox(toppage0,-1,"ksfs"),0);
    toppage0->SetSizer(sizer);
    ksfscb->SetValue(1);
    */

    /*Operations Notebook Simplify
    sizer = new wxStaticBoxSizer(new wxStaticBox(botpage0,-1,""),
				 wxHORIZONTAL);
    sizer->Add(new wxStaticText(botpage0,-1,"% remove"),0);
    sizer->Add(percentRemove=new wxSpinCtrl(botpage0,-1));
    percentRemove->SetRange(1,99);
    percentRemove->SetValue(70);
    botpage0->SetSizer(sizer);
    */

    //Triangulation Notebook, Other
    //These are the plugins with no panels
    sizer = new wxStaticBoxSizer(new wxStaticBox(botpage1,-1,wxT("")),
				 wxVERTICAL);

    //Load any of the geom plugins with no panel, into
    //the same place
    for(int j=0;j<nGeomPlugsOp;j++){
      wxPanel * panel = geomPlugsOp[j].getPanel(nbbot);
      if(!panel){
	wxButton * button=new wxButton(botpage1,
				       ID_OperationsApply,
				       wxString::FromAscii(geomPlugsOp[j].getName()));
	button->SetToolTip(wxString::FromAscii(geomPlugsOp[j].getInfo()));
	geomPlugsOp[j].setUserData(button);
	sizer->Add(button,0,wxALIGN_CENTRE);
      }else{
	printf("got a panel, adding the page\n");
	//panel->Reparent(nbbot);
	panel->SetToolTip(wxString::FromAscii(geomPlugsOp[j].getInfo()));
      	nbbot->AddPage(panel,wxString::FromAscii(geomPlugsOp[j].getName()),TRUE);
      }
    }
    //sizer->Add(restore=new wxButton(botpage1,ID_Restore,"Restore"),0,
    //       wxALIGN_CENTRE);
    botpage1->SetSizer(sizer);
	 

    //Add the other page to the bottom notebook
    nbbot->AddPage(botpage1,wxT("Other"),FALSE);
	 
    topSizer->Add(nbtop,1,wxEXPAND);
    topSizer->Add(new wxButton(topPanel,ID_GeomApply,wxT("Apply")),0,wxEXPAND);
	 	 
    botSizer->Add(nbbot,1,wxEXPAND);
    botSizer->Add(new wxButton(botPanel,ID_OperationsApply,wxT("Apply")),0,wxEXPAND);

    topPanel->SetSizer(topSizer);
    botPanel->SetSizer(botSizer);

    panel->SetMinimumPaneSize(50);
    panel->SplitHorizontally(topPanel,botPanel);
#if WX_2_6
    panel->SetSashGravity(0.3);
#endif
    geomNb=nbtop;
    triNb=nbbot;
    return panel;
  }
  wxSplitterWindow * makeLeftPanel(wxWindow * bottom,const wxString botName,
				   wxWindow * top,const wxString topName){
    wxSplitterWindow * leftPanel = new wxSplitterWindow(this,-1,wxDefaultPosition,
							wxDefaultSize,
							wxSP_NOBORDER);
    wxPanel * topPanel  = new wxPanel(leftPanel);
    wxPanel * botPanel  = new wxPanel(leftPanel);	
    wxStaticBox * topLabel = new wxStaticBox(topPanel,-1,topName);
    wxStaticBox * botLabel = new wxStaticBox(botPanel,-1,botName);
    wxStaticBoxSizer * topSizer = new wxStaticBoxSizer(topLabel,wxVERTICAL);
    wxStaticBoxSizer * botSizer = new wxStaticBoxSizer(botLabel,wxVERTICAL);
    wxBoxSizer * leftSizer = new wxBoxSizer(wxVERTICAL);

    top->Reparent(topPanel);
    bottom->Reparent(botPanel);
    topSizer->Add(top,1,wxGROW|wxEXPAND);
    botSizer->Add(bottom,1,wxGROW|wxEXPAND);
    topPanel->SetSizer(topSizer);
    botPanel->SetSizer(botSizer);
	
    leftPanel->SetSashPosition(50,TRUE);	
    leftPanel->SetMinimumPaneSize(50);
    leftPanel->SplitHorizontally(botPanel,topPanel);
    leftPanel->SetSashPosition(200,TRUE);
    //leftSizer->Add(topPanel,1,wxGROW|wxEXPAND);
    //leftSizer->Add(botPanel,1,wxGROW|wxEXPAND);
    //leftPanel->SetSizer(leftSizer);
    return leftPanel;
  }
 
  int nextStageValid(){
    if(stage+1<NUM_STAGES)return 1;
    return 0;
  }
  int prevStageValid(){
    if(stage>0)return 1;
    return 0;
  }
  void setNextStage(){
    if(nextStageValid()){
      setStage((Stage)(stage+1));
    }
  }
  void setPrevStage(){
    if(prevStageValid()){
      setStage((Stage)(stage-1));
    }
  }
  void OnTool(wxCommandEvent & event){
    setStage((Stage)(event.GetId()-TB_Capture));
  }
  void setStage(Stage s){
    wxWindow * back;

    for(int i=0;i<4;i++)
      toolBar->ToggleTool(TB_Capture+i,FALSE);
    toolBar->ToggleTool(TB_Capture+(int)s,TRUE);

    if(leftSplit)
      leftSplit->Show(FALSE);
    leftPanelTexture->Show(FALSE);
    leftPanelModel->Show(FALSE);
    
    segImageList->Show(FALSE);
    calibImageList->Show(FALSE);
    imageList->Show(FALSE);

    switch(s){
    case CAPTURE:
      back=leftSplit;

      //This is necessary so that these don't get deleted below
      imageList->Reparent(this);
      calibImageList->Reparent(this);
      segImageList->Reparent(this);

      imageList->Show(TRUE);
      calibImageList->Show(TRUE);
      segImageList->Show(FALSE);

      leftSplit=makeLeftPanel(imageList,wxT("Images"),
			      calibImageList,wxT("Calibrated Images"));
      leftSplit->Reparent(vertSplit);
      if(back)delete back;

      if(vertSplit->GetWindow1()){
	vertSplit->ReplaceWindow(vertSplit->GetWindow1(),
				 leftSplit);
	vertSplit->ReplaceWindow(vertSplit->GetWindow2(),
				 glCanvas);
      }
      else{
	vertSplit->SplitVertically(leftSplit,glCanvas);
      }
#ifdef WX_2_6
      vertSplit->UpdateSize();
#endif 
      ((CalibImageList *)calibImageList)->SetCaptureMode();
      glCanvas->SetMode(GL_CANVAS_IMAGE);
      break;
    case SEGMENTATION:
      back=leftSplit;
      
      //This is necessary so that these don't get deleted below
      imageList->Reparent(this);
      calibImageList->Reparent(this);
      segImageList->Reparent(this);
      
      imageList->Show(FALSE);
      calibImageList->Show(TRUE);
      segImageList->Show(TRUE);
      
      leftSplit=makeLeftPanel(calibImageList,wxT("Calibrated Images"),
			      segImageList,wxT("Segmented Images"));
      leftSplit->Reparent(vertSplit);
      
      if(back)delete back;

      if(vertSplit->GetWindow1()){
	vertSplit->ReplaceWindow(vertSplit->GetWindow1(),
				 leftSplit);
	vertSplit->ReplaceWindow(vertSplit->GetWindow2(),
				 glCanvas);
      }
      else{
	vertSplit->SplitVertically(leftSplit,glCanvas);
      }
#ifdef WX_2_6
      vertSplit->UpdateSize();
#endif
      ((CalibImageList *)calibImageList)->SetSegmentMode();
      glCanvas->SetMode(GL_CANVAS_IMAGE);
      break;
    case MODEL:
      leftPanelModel->Reparent(vertSplit);//this for windows
      vertSplit->ReplaceWindow(vertSplit->GetWindow1(),
			       leftPanelModel);
      leftPanelModel->Show(TRUE);
#ifdef WX_2_6
      vertSplit->UpdateSize();
#endif            
      glCanvas->SetMode(GL_CANVAS_MODEL);
      break;
    case TEXTURE:
      selCharts->Clear();
      for(int i=0;i<ncharts;i++){
	wxString str;
	str.Printf(wxT("Chart %d"),i);
	selCharts->Append(str);
      }
      SetSelectedChart(0);
      leftPanelTexture->Reparent(vertSplit);
      vertSplit->ReplaceWindow(vertSplit->GetWindow1(),
			       leftPanelTexture);
      leftPanelTexture->Show(TRUE);
#ifdef WX_2_6
      vertSplit->UpdateSize();
#endif
      glCanvas->SetMode(GL_CANVAS_TEXTURE);
      break;
    }	
    stage=s;
    changingStage=0;
  }
  void OnActivate(wxActivateEvent & event){
    //for some reason, this doesnt work in OnInit
    vertSplit->SetSashPosition(150);
#ifdef USING_GLEE
    GLeeInit(); //This was failing in the OnInit, need a window to be open
#endif
  }
  void OnExit(wxCommandEvent& event){
    Close();
  }
  void OnAbout(wxCommandEvent & event){
    wxDialog dialog(this, 0, "About", wxDefaultPosition, wxSize(360,240));
    wxBoxSizer * boxsizer = new wxBoxSizer(wxVERTICAL);
    
    wxStaticText * textCtrl = new wxStaticText(&dialog, 0, wxString::Format("Contributors: Neil Birkbeck, Martin Jagersand, Keith Yerex, Dana Cobzas, Adam Rachmielowski\nHome-page: http://www.cs.ualberta.ca/~vis/ibmr\n\nArch: %s\nVersion: %s\nBuild Date: %s", ARCH, VERSION, BUILD_DATE), 
					       wxDefaultPosition,
					       wxDefaultSize,wxALIGN_CENTRE);
    //, wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE | wxTE_READONLY | wxTE_RICH | wxTE_AUTO_URL );
    
    wxStaticText * title = new wxStaticText(&dialog, 0, 
					    wxString::Format("Capture System"), 
					    wxDefaultPosition,
					    wxDefaultSize,
					    wxALIGN_CENTRE);
    title->SetFont(wxFont(18, wxFONTFAMILY_DECORATIVE, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD));

    wxStaticText * subTitle = new wxStaticText(&dialog, 0, 
					       wxString::Format("Image-based capture and rendering"),
					       wxDefaultPosition,
					       wxDefaultSize,
					       wxALIGN_CENTRE);
    subTitle->SetFont(wxFont(14, wxFONTFAMILY_DECORATIVE, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD));
    
    
    boxsizer->Add(title,0, wxGROW | wxTOP | wxBOTTOM | wxLEFT | wxRIGHT,4);
    boxsizer->Add(subTitle,0, wxGROW | wxTOP | wxBOTTOM | wxLEFT | wxRIGHT,4);
    boxsizer->Add(textCtrl,1,wxGROW | wxTOP | wxBOTTOM | wxLEFT | wxRIGHT,10);
    dialog.SetSizer(boxsizer);

    

    dialog.ShowModal();
    
  }

  void OnLoad(wxCommandEvent & event){
    wxDirDialog dd(this,wxT("Choose a directory"),wxT(""), wxDD_NEW_DIR_BUTTON);
    if(wxID_OK==dd.ShowModal()){
      wxString path=dd.GetPath();
      printf("FIXME: the current project needs to get cleared first\n");
      loadProject((const char *)path.ToAscii());
    }
  }
     
  void OnExport(wxCommandEvent & event){
    SaveDialog sd;

    if(wxID_OK==sd.ShowModal()){
      wxDirDialog dd(this,wxT("Choose a directory"),wxT(""),
		     wxDD_NEW_DIR_BUTTON);
      sd.GetParams(glCanvas->nreig,glCanvas->texw,glCanvas->texh,
		   glCanvas->lutw,glCanvas->luth,glCanvas->flags);
      char s[1024];
      snprintf(s,1024,"%s/output",currentProj);

      fprintf(stdout, "Directory %s\n", s);
      
      if(!wxDirExists(wxString::FromAscii(s)))wxMkdir(wxString::FromAscii(s));
      snprintf(glCanvas->where,1024,"%s/data0",s);
      if(!wxDirExists(wxString::FromAscii(glCanvas->where)))
	wxMkdir(wxString::FromAscii(glCanvas->where));

      glCanvas->intrinsics = new double[segImageList->GetCount()*16];
      glCanvas->extrinsics = new double[segImageList->GetCount()*16];
      double projection[]={1,0,0,0,
			   0,1,0,0,
			   0,0,1,1,
			   0,0,0,0};
      memcpy(glCanvas->projection,projection,16*sizeof(double));
	   
      glCanvas->images = new Image8[segImageList->GetCount()];
      glCanvas->texImages = new unsigned char[glCanvas->texw[0]*
					      glCanvas->texh[0]*
					      segImageList->GetCount()*3];
      glCanvas->occluded = new unsigned char[glCanvas->texw[0]*
					     glCanvas->texh[0]*
					     segImageList->GetCount()];
	   
      glCanvas->nimages=segImageList->GetCount();
      
      for(int i=0;i<segImageList->GetCount();i++){
	double * extr=glCanvas->extrinsics+16*i;
	double * intr=glCanvas->intrinsics+16*i;
	ImageData * imd=(ImageData *) segImageList->GetClientData(i);
	if(!imd){
	  printf("no image data for entry in segmented image list\n");
	  printf("save failed\n");
	  return;
	}
	//Good thing this isn't a copy...
	glCanvas->images[i]=imd->image;
	memset(intr,0,sizeof(double)*16);
	//FIXME:assumes 0 skew
	intr[15]=1;
	intr[10]=1;
	intr[0]=imd->intrinsic(0,0);
	intr[5]=imd->intrinsic(1,1);
	intr[12]=imd->intrinsic(0,2);
	intr[13]=imd->intrinsic(1,2);

	for(int k=0;k<4;k++){
	  for(int j=0;j<4;j++){
	    extr[k+4*j]=imd->extrinsic(k,j);		 
	  }
	}
      }
      glCanvas->SetMode(GL_CANVAS_GETTEX);
    }
    else{
      //wxMessageBox(wxT("Export has failed"), wxT("Error"));
    }
  }
  void OnSave(wxCommandEvent & event){
    if(GetChangedSinceSave()){
      saveProject(currentProj);
    }
    else{
      SetStatusText(wxT("Save not necessary"));
    }
  }
  void OnSaveAs(wxCommandEvent & event){
    wxDirDialog dd(this,wxT("Choose a directory"),wxT(""), wxDD_NEW_DIR_BUTTON);
    if(wxID_OK==dd.ShowModal()){
      wxString path=dd.GetPath();
      saveProject((const char *)path.ToAscii(),1);
    }
  }
  void OnNew(wxCommandEvent & event){
    int clear=1;
    if(GetChangedSinceSave()){
      int res=wxMessageBox(wxT("Project changed since save, continue?"),
			   wxT("Confirm"),
			   wxICON_QUESTION|wxYES_NO,NULL);
      if(res!=wxYES){
	clear=0;
	SetStatusText(wxT("New aborted"));
      }
    }
    if(clear){
      //Any clean up to do here
      setToDefaultProj();
      imageList->Clear();
      calibImageList->Clear();
      segImageList->Clear();
      if(ncharts){
	delete [] charts;
      }
      charts=0;
      ncharts=0;
      SetChangedSinceSave(0);

      SetStatusText(wxT("New project created"));
      setStage(CAPTURE);
    }
  }
  void OnSettings(wxCommandEvent & event){
   
  }
  void OnCameras(wxCommandEvent & event){
    CameraDialog * cd =new CameraDialog(GetMainFrame(),-1);
    cd->ShowModal();
    delete cd;
  }
  void OnButton3(wxCommandEvent & event){
    printf("button 3\n");
  }
  void OnNext(wxCommandEvent & next){
    setNextStage();
  }
  void OnPrev(wxCommandEvent & prev){
    setPrevStage();
  }
  void SetStatusText(const wxString & text,int i=0){
    statusBar->SetStatusText(text,i);
  }
  void SetProgress(float f){
    statusBar->SetProgress(f);
  }
  void SetSelectedChart(int sel,int clear=1){
    if(clear){
      //need to turn them all off now
      //SetSelection actually toggles another one on,
      //so in a multi box, they will keep getting turned on.
      for(int i=0;i<selCharts->GetCount();i++)
	selCharts->SetSelection(i,0);
    }
    selCharts->SetSelection(sel);
    if(sel>=0){
      selChartInfo->Clear();
      wxString str;
      double avg,min,max,worldarea,texarea;
      charts[sel].areaRatios(avg,min,max,worldarea,texarea);
      str.Printf(wxT("chart %d\nnpoly %d\nnvert %d\nflags %d\nratios\nmin:%f\nmax:%f\n"),
		  sel,
		  charts[sel].npoly,
		  charts[sel].nvert,
		  charts[sel].flags,min,max);
      selChartInfo->WriteText(str);
    }
    else{
      wxString str = wxT("No selection");
    }
    glCanvas->Refresh();
  }
  int GetSelectedChart(){
    return GetSelectionSafe(selCharts);
  }
  int GetSelectedCharts(wxArrayInt & sel){
    selCharts->GetSelections(sel);
    return sel.GetCount();
  }
  //callback
  void OnOK(int id,wxCommandEvent & event){
    printf("Ok okay %d\n",id);
    switch(id){
      //change if any others need this function;
    case 0:
      {
	char dir[CURRENT_PROJ_MAX+256];
	char geomfilename[CURRENT_PROJ_MAX+256];
	char cmdline[CURRENT_PROJ_MAX+256];
	int fid=0;
	scale=bbd->scale;
	trans=bbd->trans;
	
	snprintf(dir,CURRENT_PROJ_MAX+256,
		 "%s",currentProj);
	snprintf(geomfilename,CURRENT_PROJ_MAX+256,
		 "%s/geometry.obj",currentProj);

	//snprintf(dir,1024,"%s/captureui-XXXXXX",P_tmpdir);
	//snprintf(geomfilename,1024,"%s/captureui-geom-XXXXXX",P_tmpdir);
	//fid=mkstemp(dir);
	//if(fid)close(fid);
	//fid=mkstemp(geomfilename);
	//if(fid)close(fid);

	if(GetChangedSinceSave()){
	  saveProject(dir);
	}
	//printf("making dir %s\n",dir);
	//saveProject(dir);
	
	/*BEFORE plugins
	printf("calling ray\n");
	snprintf(cmdline,1024,"%s -dir %s -file image-%%04d.tga -nrays %d -nimages %d -sx %8.4f -sy %8.4f -sz %8.4f -tx %8.4f -ty %8.4f -tz %8.4f -ofile %s",
		 bbd->command,
		 dir,nrays->GetValue(),
		 segImageList->GetCount(),//FIXME: proper n images
		 scale.x,scale.y,scale.z,
		 trans.x,trans.y,trans.z,geomfilename);
	printf("running %s\n",cmdline);
	FILE * file= popen(cmdline,"r");
	char c;
	while(EOF!=(c=fgetc(file))){
	  putchar(c);
	}
	pclose(file);
	*/
	
	int selPlug=getSelectedGeomPlugsCreate();
	if(selPlug>=0)
	  geomPlugsCreate[selPlug].create(dir,"image-%04d.tga",
				   segImageList->GetCount(),
				   geomfilename,
				   scale.x,scale.y,scale.z,
				   trans.x,trans.y,trans.z);
	if(charts && ncharts){
	  delete [] charts;
	}
	charts=new Chart[1];
	memset(charts,0,sizeof(Chart));
	printf("loading %s\n",geomfilename);
	
	if(!charts[0].loadobj(geomfilename)){
	  wxMessageBox(wxT("Error loading the .obj"),wxT("Error"));
	}
	/*
	else{
	  for(int i=0;i<charts[0].nvert;i++){
	    charts[0].vert[i].x=scale.x*charts[0].vert[i].x+trans.x;
	    charts[0].vert[i].y=scale.y*charts[0].vert[i].y+trans.y;
	    charts[0].vert[i].z=scale.z*charts[0].vert[i].z+trans.z;	    
	  }
	}
	charts[0].saveobj(geomfilename);*/

	ncharts=1;
	SetChangedSinceSave(GEOMETRY_CHANGED);
	glCanvas->Refresh();
	
	printf("loaded\n");
      }
      break;
		
    }
  }
  int getSelectedGeomPlugsCreate(){
    int sel=geomNb->GetSelection();
    for(int i=0;i<nGeomPlugsCreate;i++){
      if(wxString::FromAscii(geomPlugsCreate[i].getName())==
	 geomNb->GetPageText(sel)){
	return i;
      }
    }
    return -1;
  }
  void OnGeomApply(wxCommandEvent & event){
    int sel=geomNb->GetSelection();
    int i=0;
    int nuncal=GetUncalImageCount();
    int ncal=GetCalImageCount();
    if(sel<0)return;
    if(nuncal+ncal>0){
      wxString str,uncalstr,calstr,connector;
      if(nuncal)uncalstr.Printf(wxT("%d uncalibrated "),nuncal);
      if(ncal)calstr.Printf(wxT("%d unsegmented "),ncal);
      if(ncal>0 && nuncal>0)connector = wxT("and ");

      str = wxT("There are still ")+uncalstr + connector + calstr + wxT("image(s).\nPlease remove these before continuing");

      wxMessageDialog * msg= new wxMessageDialog(this,str,wxT("Error"),wxICON_ERROR);
      msg->ShowModal();
      delete msg;
      return;
    }

    int needsBB=0;
    int selPlug=getSelectedGeomPlugsCreate();
    if(selPlug<0)return;
    needsBB=(geomPlugsCreate[selPlug].create(0,0,0,0,0,0,0,0,0,0)&
	     GEOMPLUGIN_NEEDS_BOUNDING_BOX_BIT)>0;
    
    if(needsBB){
      if(!bbd->IsShown()){
	for(i=0;i<segImageList->GetCount();i++){
	  bbd->AddImage((ImageData*)segImageList->GetClientData(i));
	}
	bbd->Show(TRUE);
	bbd->SetScale(scale);
	bbd->SetTrans(trans);
      }
      else{
	bbd->Show(TRUE);
	bbd->Raise();
      }
    }
    else{
      wxCommandEvent ev;
      OnOK(0,ev);
    }

    /*
    switch(sel){
    case 0:{
      if(!bbd->IsShown()){
	char command[1024];
        if(ksfscb->GetValue())
	  sprintf(command,"sfs");
        else
	  sprintf(command,"ray");
	bbd->SetCommand(command);
	for(i=0;i<segImageList->GetCount();i++){
	  bbd->AddImage((ImageData*)segImageList->GetClientData(i));
	}
	bbd->Show(TRUE);
	bbd->SetScale(scale);
	bbd->SetTrans(trans);
      }
      else{
	bbd->Show(TRUE);
	bbd->Raise();
      }
      break;
    }
    case 1:
      break;
    case 2:
      bbd->SetCommand((char *)geomCommand->GetLineText(0).ToAscii());
      
      for(i=0;i<segImageList->GetCount();i++){
	bbd->AddImage((ImageData*)segImageList->GetClientData(i));
      }
      bbd->Show(TRUE);
      bbd->SetScale(scale);
      bbd->SetTrans(trans);
      break;
      }*/
  }
  //make sure geometry exists
  void OnOperationsApply(wxCommandEvent & event){
    int sel=triNb->GetSelection();
    wxString cmdName=triNb->GetPageText(sel);
    //If its the last page, then it is the page
    //of buttons with no parameters
    printf("sel %d\n",sel);
    if(sel==triNb->GetPageCount()-1){
      wxButton * but=(wxButton*)event.GetEventObject();
      cmdName=but->GetLabel();
    }
    for(int j=0;j<nGeomPlugsOp;j++){
      //printf("%s\n",geomPlugsOp[j].getName());
      if(cmdName==wxString::FromAscii(geomPlugsOp[j].getName())){
	int ret=geomPlugsOp[j].apply(&charts,&ncharts);
	
	if(ret & GEOMPLUGIN_RESTORE_BIT){
	  OnRestore(event);
	}
	if(ret & GEOMPLUGIN_MERGE_BIT) 
	  OnMerge(event);
	if(ret & GEOMPLUGIN_APPLY_AGAIN_BIT) //keep doing this?
	  ret=geomPlugsOp[j].apply(&charts,&ncharts);
      }
    }
    SetChangedSinceSave(GEOMETRY_CHANGED);
    glCanvas->Refresh();
    return;
    
    /*
    switch(sel){
      //Simplification
    case 0:
      if(charts && ncharts){
	if(ncharts>1){
	  OnRestore(event);
	  //Chart chart;
	  //mergeCharts(chart,charts,ncharts);
	  //delete [] charts;
	  //charts = new Chart[1];
	  //ncharts=1;
	  //charts[0]=chart;
	}
	simplify(charts[0],percentRemove->GetValue());
      }
      SetChangedSinceSave(GEOMETRY_CHANGED);
      break;
    case 1:
    //wxMessageDialog * msg=
    //new wxMessageDialog(this,triCommand->GetLineText(0),"Command");
    //msg->ShowModal();
    //delete msg;
    //SetChangedSinceSave(GEOMETRY_CHANGED);
    //break;
    }*/
  
  }
  void OnMerge(wxCommandEvent & event){
    if(ncharts>1){
      int res=wxMessageBox(wxT("This operation will reset any texture coordinates.  Continue?"),wxT("Proceed"),
			   wxICON_QUESTION|wxYES_NO,NULL);
      if(res==wxYES){
	Chart chart;
	mergeCharts(chart,charts,ncharts);
	delete [] charts;
	charts = new Chart[1];
	ncharts=1;
	charts[0]=chart;
      }
    }
  }
  void OnRestore(wxCommandEvent & event){
    FILE * file;
    char fname[CURRENT_PROJ_MAX+256];
    snprintf(fname,CURRENT_PROJ_MAX+256,
	     "%s/geometry.obj",currentProj);
    file=fopen(fname,"r");
    if(file){
      int res=wxMessageBox(wxT("This operation will reset any texture coordinates.  Continue?"),wxT("Proceed"),
			   wxICON_QUESTION|wxYES_NO,NULL);
      if(res==wxYES){
	fclose(file);
	delete [] charts;
	ncharts=1;
	charts = new Chart[1];
	charts[0].loadobj(fname);
	//Texture coordinates are reset, so don't draw texture
	glCanvas->modelModeMenuCheck(ID_DrawTexture,0);
      }
    }
    else{
      wxMessageBox(wxT("No geometry to restore from"),wxT("Save Failed"),
		   wxICON_ERROR|wxOK,NULL);
    }
  }

  void OnSelCharts(wxCommandEvent & event){
    int sel=GetSelectionSafe(selCharts);
    if(sel>=0)SetSelectedChart(sel,0);//don't clear em
  }
  void OnSpread(wxCommandEvent & event){
    if(ncharts){
      for(int i=0;i<ncharts;i++){
	double meanx=0;
	double meany=0;
	for(int j=0;j<charts[i].nvert;j++){
	  meanx+=charts[i].tc[j].x;
	  meany+=charts[i].tc[j].y;
	}
	meanx/=charts[i].nvert;
	meany/=charts[i].nvert;
	for(int j=0;j<charts[i].nvert;j++){
	  charts[i].tc[j].x=(charts[i].tc[j].x-meanx)*255.0/256.0+meanx;
	  charts[i].tc[j].y=(charts[i].tc[j].y-meany)*255.0/256.0+meany;
	}
      }
      glCanvas->Refresh();
    }
  }
  void OnStretch(wxCommandEvent & event){
    if(ncharts){
      if(charts){
	Chart chart;
	int nvert=0;
	//optimize the single chart
	mergeCharts(chart,charts,ncharts);
	sot(chart);
	nvert=0;
	//and unpack the texture coordinates into the multiple charts
	for(int i=0;i<ncharts;i++){
	  memcpy(charts[i].tc,chart.tc+nvert,
		 sizeof(charts[i].tc[0])*charts[i].nvert);
	  nvert+=charts[i].nvert;
	}
      }
      SetChangedSinceSave(GEOMETRY_CHANGED);
    }
  }
  void OnTexApply(wxCommandEvent & event){
    for(int i=0;i<nTexPlugsOther;i++){
      if(event.GetEventObject()==texPlugsOther[i].getUserData()){
	texPlugsOther[i].apply(&charts,&ncharts);
	break;
      }
    }
    SetChangedSinceSave(GEOMETRY_CHANGED);
    glCanvas->Refresh();
  }
  void OnPack(wxCommandEvent & event){
    if(ncharts){
      packChartsNew(charts,ncharts,512,512);
      glCanvas->Refresh();
      SetChangedSinceSave(GEOMETRY_CHANGED);
    }
  }
  void OnRemoveOverlap(wxCommandEvent & event){
    int changed=0;
    while(1){
      int nChartsNeeded=0;
      for(int i=0;i<ncharts;i++){
	int p1,p2;
	int split=charts[i].checkForOverlap(32,32,&p1,&p2);
	printf("chart[%d].split=%d\n",i,split);
	if(split && p1>=0 && p2>=0 && p1!=p2){
	  nChartsNeeded++;
	}
      }
      if(nChartsNeeded==0)break;
      changed=1;
      printf("REMOVING OVERLAP\n");
      printf("%d charts have overlap\n",nChartsNeeded);
      Chart * newcharts=new Chart[ncharts+nChartsNeeded];
      
      int newcharti=0;
      for(int i=0;i<ncharts;i++){
	int p1,p2;
	int split=charts[i].checkForOverlap(32,32,&p1,&p2);
	if(split && p1>=0 && p2>=0 && p1!=p2){
	  int shouldbe2=0;
	  Chart * twocharts=splitDistance(charts+i,shouldbe2,p1,p2);
	  /*FIXME: should re-compute the texture coordinates here?*/
	  /*if(lscmCheck->GetValue()){
	    getTexCoordsLSCM(twocharts[0]);
	    getTexCoordsLSCM(twocharts[1]);
	  }
	  else if(mdsCheck->GetValue()){
	    getTexCoordsMDS(twocharts[0]);
	    getTexCoordsMDS(twocharts[1]);
	  }
	  else {
	    getTexCoordsSVD(twocharts[0]);
	    getTexCoordsSVD(twocharts[1]);
	    }*/
	  getTexCoords(twocharts+0);
	  getTexCoords(twocharts+1);
	  
	  newcharts[newcharti++]=twocharts[0];
	  newcharts[newcharti++]=twocharts[1];
	}
	else{
	  newcharts[newcharti++]=charts[i];
	}
      }
      printf("%d %d\n",newcharti,nChartsNeeded+ncharts);
      delete [] charts;
      charts=newcharts;
      ncharts=ncharts+nChartsNeeded;
    }
    if(changed){
      selCharts->Clear();
      for(int i=0;i<ncharts;i++){
	wxString str;
	str.Printf(wxT("Chart %d"),i);
	selCharts->Append(str);
      }
      OnPack(event);
    }
  }
  void OnMapChoice(wxCommandEvent & event){
    int sel=mapChoice->GetSelection();
    if(sel<0)return;
    for(int i=0;i<mapChoice->GetCount();i++){
      wxPanel * win=(wxPanel *)mapChoice->GetClientData(i);
      win->Show(FALSE);
      win->Reparent(this);
    }
    wxPanel * selectedPanel=
      ((wxPanel *)mapChoice->GetClientData(sel));

    selectedPanel->Show(TRUE);
    selectedPanel->Reparent(mapSplitter);
    mapSplitter->Initialize(selectedPanel);    
    
    mapSplitter->Layout();
    mapSplitter->GetParent()->Layout();
    
    mapSplitter->Update();
#ifdef WX_2_6
    mapSplitter->UpdateSize();
#endif
  }
  void OnSplitChoice(wxCommandEvent & event){
    int sel=splitChoice->GetSelection();
    if(sel<0)return;
    printf("OnSplitChoice %d\n",sel);
    for(int i=0;i<splitChoice->GetCount();i++){
      printf("%d %p\n",i,splitChoice->GetClientData(i));
      wxPanel * win=(wxPanel *)splitChoice->GetClientData(i);
      win->Show(FALSE);
      win->Reparent(this);
    }
    wxPanel * selectedPanel=
      ((wxPanel *)splitChoice->GetClientData(sel));
    selectedPanel->Show(TRUE);
    selectedPanel->Reparent(splitSplitter);
    splitSplitter->Initialize(selectedPanel);    
    
    splitSplitter->Layout();
    splitSplitter->GetParent()->Layout();

    splitSplitter->Update();
#ifdef WX_2_6
    splitSplitter->UpdateSize();
#endif
    printf("split choice %d\n",sel);
  }

  void OnSplit(wxCommandEvent & event){
    int newncharts=0;
    Chart * newcharts=0;
    int sel;
    int texgenAndPack=(ncharts==1 && GetSelectedChart()==0);
    if((sel=GetSelectedChart())==-1)return;
    int choice=splitChoice->GetSelection();
    if(choice<0)return;
    printf("sel chart %d splitChoice %d\n",sel,choice);
    if(splitChoice->GetString(choice)==wxT("Normals")){
      newcharts=split(charts+sel,newncharts,
		      coneAngle->GetValue()*M_PI/180.0);
    }
    else{
      printf("plugin split\n");
      wxPanel * selPanel=(wxPanel *)splitChoice->GetClientData(choice);
      for(int i=0;i<nTexPlugsSplit;i++){
	if(((wxPanel *)texPlugsSplit[i].getUserData())==selPanel){
	  printf("OnSplit %s\n",texPlugsSplit[i].getName());
	  newcharts = new Chart[1];
	  newcharts[0]=charts[sel];
	  newncharts=1;
	  texPlugsSplit[i].apply(&newcharts,&newncharts);
	  printf("done %d\n",newncharts);
	  break;
	}
      }
    }
    //printf("%d %d\n",newcharts[0].nvert,newcharts[1].nvert);
    if(newcharts&&newncharts){
      Chart * allcharts=new Chart[newncharts+ncharts-1];
      int aci=0;
      
      for(int i=0;i<sel;i++){
	allcharts[aci]=charts[i];
       	aci++;
      }
      for(int i=0;i<newncharts;i++){
	allcharts[aci]=newcharts[i];
	aci++;
      }
      for(int i=sel+1;i<ncharts;i++){
	allcharts[aci]=charts[i];
	aci++;
      }
      delete [] charts;
      delete [] newcharts;
      charts=allcharts;
      ncharts=newncharts+ncharts-1;
    }


    if(texgenAndPack){
      wxBeginBusyCursor();
      SetStatusText(wxT("Mapping the charts"));
      for(int i=sel;i<sel+newncharts;i++){
	SetProgress(((float) i-sel)/((float)newncharts));
	GetMainFrame()->Refresh();
	GetApp()->Yield();
	getTexCoords(charts+i);
      }
      wxEndBusyCursor();
      /*if(lscmCheck->GetValue()){
	for(int i=sel;i<sel+newncharts;i++)
	  getTexCoordsLSCM(charts[i]);
      }
      else if(mdsCheck->GetValue()){
	for(int i=sel;i<sel+newncharts;i++)
	  getTexCoordsMDS(charts[i]);
      }
      else {
	for(int i=sel;i<sel+newncharts;i++)
	  getTexCoordsSVD(charts[i]);
	  }*/
    }

    selCharts->Clear();
    for(int i=0;i<ncharts;i++){
      wxString str;
      str.Printf(wxT("Chart %d"),i);
      selCharts->Append(str);
    }
    wxBeginBusyCursor();
    SetStatusText(wxT("Packing the charts"));
    if(texgenAndPack){
      if(newncharts==ncharts)
	packChartsNew(charts,ncharts,512,512);
      else
	scaleCharts(charts,ncharts,1);
    }
    wxEndBusyCursor();
    glCanvas->Refresh();
    glCanvas->setChartFlags();
    SetChangedSinceSave(GEOMETRY_CHANGED);
  }
  //use the currently selected method to get the texture coordinates
  void getTexCoords(Chart * chart){
    int sel = mapChoice->GetSelection();
    if(sel<0)return;
    wxPanel * selPanel = (wxPanel *)mapChoice->GetClientData(sel);
    for(int i=0;i<nTexPlugsMap;i++){
      if(((wxPanel *)texPlugsMap[i].getUserData())==selPanel){
	int nc=1;
	texPlugsMap[i].apply(&(chart),&nc);
	if(nc!=1){
	  wxMessageBox(wxT("Texture Plugin")+
		       wxString::FromAscii(texPlugsMap[i].getName())+wxT("is ill behaved")
		       ,wxT("Error"),wxOK);
	}
	break;
      }
    }
  }
  void OnMap(wxCommandEvent & event){
    wxArrayInt sels;
    GetSelectedCharts(sels);//
    if(!sels.GetCount())return;
    double minx=10e10,maxx=-10e10;
    double miny=10e10,maxy=-10e10;
    for(int s=0;s<sels.GetCount();s++){
      int sel=sels[s];
      for(int i=0;i<charts[sel].nvert;i++){
	if(charts[sel].tc[i].x>maxx)maxx=charts[sel].tc[i].x;
	if(charts[sel].tc[i].y>maxy)maxy=charts[sel].tc[i].y;
	if(charts[sel].tc[i].x<minx)minx=charts[sel].tc[i].x;
	if(charts[sel].tc[i].y<miny)miny=charts[sel].tc[i].y;
      }
      getTexCoords(charts+sel);
    
      /*
	if(lscmCheck->GetValue()){
	getTexCoordsLSCM(charts[sel]);
	}
	else if(mdsCheck->GetValue()){
	getTexCoordsMDS(charts[sel]);
	}
	else{
	getTexCoordsSVD(charts[sel]);
	}*/
      //scaleCharts(charts,ncharts);
      
      double sc=MAX(maxx-minx,maxy-miny);
      for(int i=0;i<charts[sel].nvert;i++){
	charts[sel].tc[i].x=charts[sel].tc[i].x*sc+minx;//*(maxx-minx)+minx;
	charts[sel].tc[i].y=charts[sel].tc[i].y*sc+miny;//*(maxy-miny)+miny;
      }
    }
    glCanvas->setChartFlags();
    glCanvas->Refresh();
    SetChangedSinceSave(GEOMETRY_CHANGED);
  }
  void OnDistance(wxCommandEvent & event){
    normalsCheck->SetValue(!distanceCheck->GetValue());
  }
  void OnNormals(wxCommandEvent & event){
    distanceCheck->SetValue(!normalsCheck->GetValue());
  }
  /*void OnLSCM(wxCommandEvent & event){
    mdsCheck->SetValue(!lscmCheck->GetValue());
  }
  void OnMDS(wxCommandEvent & event){
    lscmCheck->SetValue(!mdsCheck->GetValue());
    }*/
  void OnClose(wxCloseEvent & event){
    GetSize(&options.mainFrameWidth,
	    &options.mainFrameHeight);
    if(GetChangedSinceSave()){
      int res=wxMessageBox(wxT("Project changed since last save.  Save before exit?"),wxT("Confirm"),
			   wxICON_QUESTION|wxYES_NO,NULL);
      if(res==wxYES){
	saveProject(currentProj);
      }
    }
    writeConfig();
    
    delete bbd;
    this->Destroy();
  }
  void SetSaveEnabled(bool enable){
    toolBar->EnableTool(ID_Save,enable);
    menuBar->Enable(ID_Save,enable);
  }
  void SetSaveAsEnabled(bool enable){
    toolBar->EnableTool(ID_SaveAs,enable);
    menuBar->Enable(ID_SaveAs,enable);
  }
  void SetExportEnabled(bool enable){
    toolBar->EnableTool(ID_Export,enable);
    menuBar->Enable(ID_Export,enable);
  }
  void SetLoadEnabled(bool enable){
    toolBar->EnableTool(ID_Load,enable);
    menuBar->Enable(ID_Load,enable);
  }
  void SetNewEnabled(bool enable){
    toolBar->EnableTool(ID_New,enable);
    menuBar->Enable(ID_New,enable);
  }

  void SetSegmentEnabled(bool enable){
    toolBar->EnableTool(TB_Segment,enable);
  }
  void SetGeometryEnabled(bool enable){
    toolBar->EnableTool(TB_Geometry,enable);
  }
  void SetTextureEnabled(bool enable){
    toolBar->EnableTool(TB_Texture,enable);
  }
  ~MyFrame(){
    delete [] geomPlugsCreate;
    delete [] geomPlugsOp;
    delete [] texPlugsMap;
    delete [] texPlugsSplit;
    delete [] texPlugsOther;
  }
  DECLARE_EVENT_TABLE();
};
BEGIN_EVENT_TABLE(MyFrame, wxFrame)
  EVT_MENU    (ID_About,  MyFrame::OnAbout)
  EVT_MENU    (ID_Quit,   MyFrame::OnExit)
  EVT_MENU    (ID_Load,  MyFrame::OnLoad)
  EVT_MENU    (ID_Save,   MyFrame::OnSave)
  EVT_MENU    (ID_SaveAs, MyFrame::OnSaveAs)
  EVT_MENU    (ID_New  ,  MyFrame::OnNew)
  EVT_MENU    (ID_Export,MyFrame::OnExport)
  EVT_MENU(ID_Camera,MyFrame::OnCameras)
  EVT_MENU(ID_Settings,MyFrame::OnSettings)
  EVT_BUTTON(ID_Next,MyFrame::OnNext)
  EVT_BUTTON(ID_Prev,MyFrame::OnPrev)
  EVT_MENU(TB_Segment,MyFrame::OnTool)
  EVT_MENU(TB_Capture,MyFrame::OnTool)
  EVT_MENU(TB_Texture,MyFrame::OnTool)
  EVT_MENU(TB_Geometry,MyFrame::OnTool)
  EVT_BUTTON(ID_OperationsApply,MyFrame::OnOperationsApply)
  EVT_BUTTON(ID_GeomApply,MyFrame::OnGeomApply)
  EVT_ACTIVATE(MyFrame::OnActivate)
  EVT_LISTBOX(ID_SelCharts,MyFrame::OnSelCharts)
  EVT_CHECKBOX(ID_Distance,MyFrame::OnDistance)
  EVT_CHECKBOX(ID_Normals,MyFrame::OnNormals)
  EVT_BUTTON(ID_TexApply,MyFrame::OnTexApply)
  EVT_BUTTON(ID_Split,MyFrame::OnSplit)
  EVT_BUTTON(ID_Map,MyFrame::OnMap)
  EVT_BUTTON(ID_Pack,MyFrame::OnPack)
  EVT_BUTTON(ID_Spread,MyFrame::OnSpread)
  EVT_BUTTON(ID_Stretch,MyFrame::OnStretch)
  EVT_BUTTON(ID_RemoveOverlap,MyFrame::OnRemoveOverlap)
  EVT_CHOICE(ID_SplitChoice,MyFrame::OnSplitChoice)
  EVT_CHOICE(ID_MapChoice,MyFrame::OnMapChoice)
  EVT_CLOSE(MyFrame::OnClose)
  
  //EVT_CHECKBOX(ID_LSCM,MyFrame::OnLSCM)
  //EVT_CHECKBOX(ID_MDS,MyFrame::OnMDS)
END_EVENT_TABLE()

MyFrame * frame;
wxApp * app;

wxApp * GetApp(){
  return app;
}

wxFrame * GetMainFrame(){
  return frame;
}

int GetSelectionSafe(wxListBox * listbox)
{
  int sel;
  if((sel=listbox->GetSelection())>=0)return sel;
  wxArrayInt sels;
  listbox->GetSelections(sels);
  if(sels.GetCount()==0)return -1;
  return sels[0];
}
void UpdateTitle(){
  frame->UpdateTitle();
}

void SetStatusText(const wxString & text,int i){
  frame->SetStatusText(text,i);
}
void SetProgress(float f){
  frame->SetProgress(f);
}
int GetSelectedCharts(wxArrayInt & sel)
{
  return frame->GetSelectedCharts(sel);
}
int GetSelectedChart(){
  return frame->GetSelectedChart();
}
void SetSelectedChart(int i){
  frame->SetSelectedChart(i);
}

int GetUncalImageCount(){
  int nuncal=0;
  for(int i=0;i<imageList->GetCount();i++){
    nuncal+=(imageList->GetClientData(i)!=NULL);
  }
  return nuncal;
}
int GetCalImageCount(){
  int ncal=0;
  for(int i=0;i<calibImageList->GetCount();i++){
    ncal+=(calibImageList->GetClientData(i)!=NULL);
  }
  return ncal;
}
int GetSegImageCount(){
  int nseg=0;
  for(int i=0;i<segImageList->GetCount();i++){
    nseg+=(segImageList->GetClientData(i)!=NULL);
  }
  return nseg;
}
int GetAllImageCount(){
  return GetUncalImageCount()+
    GetCalImageCount()+
    GetSegImageCount();
}

int GetChangedSinceSave(){
  return changedSinceSave;
}

void DisableAll(){
  frame->SetSaveEnabled(FALSE);
  frame->SetSaveAsEnabled(FALSE);
  frame->SetNewEnabled(FALSE);
  frame->SetLoadEnabled(FALSE);
  frame->SetExportEnabled(FALSE);
  
  frame->SetSegmentEnabled(0);
  frame->SetGeometryEnabled(0);
  frame->SetTextureEnabled(0);
  frame->SetExportEnabled(0);
}

void EnableAll(){
  int calcnt=GetCalImageCount();
  int segcnt=GetSegImageCount();
  

  frame->SetSegmentEnabled((calcnt+segcnt)>0);
  frame->SetGeometryEnabled((segcnt+ncharts)>0);
  frame->SetTextureEnabled(ncharts);
  frame->SetExportEnabled(ncharts);
  
  if(GetChangedSinceSave())
    frame->SetSaveEnabled(TRUE);
  else
    frame->SetSaveEnabled(FALSE);
  frame->SetSaveAsEnabled(TRUE);
  frame->SetNewEnabled(TRUE);
  frame->SetLoadEnabled(TRUE);
}
void SetChangedSinceSave(int flags){
  if(flags&&!changedSinceSave){
    SetStatusText(wxT("Save your work often"),STATUS_INFO_FIELD);
  }
  else if(!flags && changedSinceSave){
    SetStatusText(wxT("Copy on disk is in sync"),STATUS_INFO_FIELD);
  }
  if(flags){
    changedSinceSave|=flags;
  }
  else
    changedSinceSave=0;
  EnableAll();
  UpdateTitle();
}


///The application entry point
class MyApp: public wxApp
{
  wxSplashScreen * splash;
public:
  MyApp(){
    printf("constructor\n");
    wxString s = wxT("Hi");
    printf("a string is %s\n", (const char*)s.mb_str(wxConvUTF8));
  }
  void ShowSplash(){
    //Splash screen
    char blah[10240];
    wxInitAllImageHandlers();
    wxBitmap bitmap(splash_xpm);

    //if (bitmap.LoadFile("splash.gif", wxBITMAP_TYPE_GIF)){
    splash = new wxSplashScreen(bitmap,wxSPLASH_CENTRE_ON_SCREEN |wxSPLASH_TIMEOUT,4000,
						 frame,0,wxDefaultPosition,wxDefaultSize,
						 wxSIMPLE_BORDER);//|wxSTAY_ON_TOP);
    wxYield();
  }
  bool OnInit(){
    printf("in oninit\n");
    app=this;
    wxApp::OnInit();
    wxInitAllImageHandlers();
    setToDefaultProj();
    readConfig();
    
    frame = new MyFrame(wxT("Frame"),wxDefaultPosition,
			wxSize(options.mainFrameWidth,options.mainFrameHeight));
    frame->Show(TRUE);
    SetTopWindow( frame );
    ShowSplash();

    if(argc>1){
      printf("loading project %s\n",argv[1]);
      loadProject((const char*)wxString(argv[1]).ToAscii());
    }
    SetExitOnFrameDelete(TRUE);
    SetChangedSinceSave(0);
    
#ifdef USING_GLEE
    //This is fucked, it only fails when you "don't" call messagebox befor initting.
    //Windows must not show or something.
    //wxMessageBox("About to init GLee\n");
    int res = GLeeInit();
    if(!res){
      //wxMessageBox("Initting GLee Failed\n");
      printf("Initting glee failed\n");
    }
#endif
    return TRUE;
  }
  //Add any switches and command line options here
  void OnInitCmdLine(wxCmdLineParser& parser){
    parser.AddParam(wxT("filename or http link"),
		    wxCMD_LINE_VAL_STRING,
		    wxCMD_LINE_PARAM_OPTIONAL);
  }
  bool OnCmdLineParsed(wxCmdLineParser& parser){
    return TRUE;
  }
  bool OnCmdLineError(wxCmdLineParser& parser){
    return TRUE;
  }
  bool OnCmdLineHelp(wxCmdLineParser& parser){
    return TRUE;
  }
  int OnExit(){
  }
};
IMPLEMENT_APP(MyApp)




