#ifndef GUI_TYPES_H
#define GUI_TYPES_H

#include <wx/wx.h>
#include <wx/listctrl.h>
#include <sys/time.h>
#include "image.h"
#include "matrix.h"
#include "newseg.h"
#include <wx/glcanvas.h>

enum {ID_About,ID_Quit,ID_ImageList,
      ID_CalibImageList,
      ID_Calibrate,ID_Remove,ID_RemoveAll,ID_Play,
      ID_New,ID_Save,ID_SaveAs,
      ID_Load,ID_Capture,
      ID_Grab,ID_Close,ID_OK,ID_Mode,ID_Format,
      ID_Camera,ID_TopPanel,ID_BotPanel,
      ID_Next,ID_Prev,ID_Segment,ID_SubtractSeg,
      ID_ColorSeg,ID_Thresh,ID_Add,
      ID_GeomApply,ID_OperationsApply,
      ID_Nrays,ID_Command,ID_SelCharts,ID_ChartInfo,
      ID_Normals,ID_Distance,ID_LSCM,ID_MDS,ID_Split,
      ID_Map,ID_Pack,ID_Export,ID_DrawNormals,
      ID_DrawGrid,ID_Wireframe,ID_DrawCameras,
      ID_Current,ID_Properties,ID_Force,ID_DrawTexture,
      ID_Stretch,ID_SelectAll,ID_Restore,ID_Spread,ID_Reset,
      ID_RemoveOverlap,ID_Settings,
      ID_TexApply,ID_SplitChoice,ID_MapChoice,ID_BayerButton,
      ID_BayerPattern,ID_BayerInterpMode,ID_ProjectTexture
};

enum {IMAGES_CHANGED=1,GEOMETRY_CHANGED=2};

//FIXME: the latest scaling and translation
#define DEFAULT_SCALE Vec3f(1,1,1); //was 200
#define DEFAULT_TRANS Vec3f(-0.5,0,-0.5) //was -100,0,-100


///The application options
struct Options{
  int playDelay;
  int captureDelay;
  int captureImages;
  unsigned char cameraFormatIndex;
  unsigned char cameraModeIndex;
  int mainFrameWidth;
  int mainFrameHeight;
  double maxResidual;
  int currentCameraProfile;
  int calibThresh;
  int calibAutoThresh;
  int calibChannel;
  int calibSmall;
  int calibBig;
  int bayerEnabled;
  int bayerPattern;
  int bayerInterpolationMode;
  Options(){
    playDelay=100;
    captureDelay=30;
    captureImages=32;
    cameraFormatIndex=0;
    cameraModeIndex=1;
    mainFrameWidth=640;
    mainFrameHeight=480;
    maxResidual=1.0;
    currentCameraProfile=0;
    calibThresh=60;
    calibAutoThresh=1;
    calibChannel=1;
    calibSmall=15;
    calibBig=1000;
    bayerEnabled=0;
    bayerPattern=0;
    bayerInterpolationMode=0;
  }
};

///Very simple StopWatch class
class StopWatch{
 public:
#ifdef WIN32
  DWORD start;
#else
  struct timeval start;
#endif
  StopWatch(){
    reset();
  }
  void reset(){
#ifdef WIN32
    start=GetTickCount();
#else
    gettimeofday(&start,NULL);
#endif
  }
  int hasElapsed(unsigned long millis){
#ifdef WIN32
    DWORD now=GetTickCount();
    if((now-start)>millis){
      //start=now-((DWORD)millis);//set it to the time that the event happened
      start=now;
      return 1;
    }
#else
    struct timeval now;
    gettimeofday(&now,NULL);
    unsigned long long span = (now.tv_sec-start.tv_sec)*1000+
      (now.tv_usec-start.tv_usec)/1000;
    if(span>millis){
      start=now;
      start.tv_sec-=(span-millis)/1000;
      start.tv_usec-=(span-millis)%1000;
      return 1;
    }
#endif
    return 0;
  }
};


///Status bar with two info fields, and a progress bar.
class StatusBar:public wxStatusBar
{
  wxGauge * progress;
  wxStaticText * text;
 public:
  DECLARE_DYNAMIC_CLASS(StatusBar)

    StatusBar(wxWindow * parent=0):
    wxStatusBar(parent,-1){
    SetFieldsCount(3);
    //text=new wxStaticText(this,STATUS_TEXT,"status");
    progress=new wxGauge(this,-1,
			 100,wxDefaultPosition, wxDefaultSize,
			 wxBU_EXACTFIT);
    SetProgress(0);
    wxSizeEvent dummy;
    OnSize(dummy);
  }
  void SetProgress(float f){
    progress->SetValue((int)(f*100));
  }
  void OnSize(wxSizeEvent & event){
    if ( !progress)
      return;
	
    wxRect rect;
    GetFieldRect(2, rect);
    wxSize size = progress->GetSize();
	
    progress->Move(rect.x,
		   rect.y+1);
    progress->SetSize(rect.width,
		      rect.height-1);
    event.Skip();
  }
  DECLARE_EVENT_TABLE();
};

///Abstract class allowing transmission of an OnOK message. 
class OKHandler{
 public:
  virtual void OnOK(int id,wxCommandEvent & event)=0; 
};

#define STATUS_MESSAGE_FIELD 0
#define STATUS_INFO_FIELD    1

void SetStatusText(const wxString & text,int i=0);
void SetProgress(float f);
int GetUncalImageCount();
int GetCalImageCount();
int GetSegImageCount();
int GetSelectedChart();
int GetSelectedCharts(wxArrayInt & sel);
int GetAllImageCount();
int GetChangedSinceSave();
void SetChangedSinceSave(int val);
void SetSelectedChart(int i);
void UpdateTitle();
void DisableAll();
void EnableAll();
wxFrame * GetMainFrame();
wxApp * GetApp();
int GetSelectionSafe(wxListBox * listbox);
#endif
