

#include "heap.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

//#define CHECK
/*
  post: allocate a heap with the number of elements
*/
Heap::Heap(int n)
{
  data = (HeapElement *) new HeapElement[n];
  memset(data,0,sizeof(HeapElement)*(n));
  indices = new int[n];
  memset(indices,0xff,sizeof(int)*n);
  this->n = 0;
  this->max = n;
}

/*
  post: free the memory from h
*/
Heap::~Heap()
{
  delete indices;
  if(data)
    delete [] data;
  data=0;
  n=0;
  max=0;
}

/*
  post: remove all the elements from the heap
*/
void Heap::removeAll()
{
  delete indices;
  indices = new int[max];
  memset(indices,0xff,sizeof(int)*max);
  memset(data,0,sizeof(HeapElement)*(max));
  n=0;
}

/*
  post: sift the element i up as far as it can go
*/
void Heap::siftup(int i)
{
  int j=(i-1)/2;
  HeapElement temp;
  while(j>=0){
    if(data[j].value>data[i].value){
      temp = data[j];
      data[j]=data[i];
      data[i]=temp;
      indices[data[j].node]=j;
      indices[data[i].node]=i;
    } 
    else break;
    i=j;
    j = (i-1)/2;
  }
}

/*
  post: sift element i down until its children are
        larger than it
*/
void Heap::siftDown(int i)
{
  int child;
  HeapElement top = data[n-1];
 
  while((child=(2*i+1))<n){
    if((child+1)<n && data[child].value > data[child+1].value){
      child=child+1;
    }
    if(data[child].value < top.value){//Here, was <
      data[i]=data[child];
      indices[data[i].node]=i;
      i=child;
    }
    else {
      break;
    }
  }
  data[i]=top;
  indices[data[i].node]=i;
}

void Heap::siftDown2(int i)
{
  int child;

  while((child=(2*i+1))<n){
    if((child+1)<n && data[child].value > data[child+1].value){
      child=child+1;
    }
    if(data[child].value < data[i].value){//Here, was <
      HeapElement temp = data[i];
      data[i]=data[child];
      indices[data[i].node]=i;
      data[child]=temp;
      indices[data[child].node]=child;
      i=child;
    }
    else {
      break;
    }
  }
}

int Heap::exists(int n)
{
    return indices[n]!=-1;
}

double Heap::get(int node)
{
    return data[indices[node]].value;
}

/*
  post: adjust the weight for the given node to the 
        given value
*/
void Heap::update(int node,double val,int usemax)
{
  int index = indices[node];
  if(index==-1){
    printf("problem in update, -1 index for %d,%f",node,val);
    exit(0);
  }
  if(index>=n){
      printf("problem in update, node=(%d,%f), %d>=%d\n",node,val,index,n);
      return;
  }
  if(data[index].node!=node){
      printf("error in update %d!=%d",data[index].node,node);
      return;
  }
  else{
      //printf("update okay!!!\n");
  }
#ifdef CHECK
  check("Before sift down\n");
#endif
  //printf("sifting down %d\n",index);
  int better=val<data[index].value;
  if(better&&usemax){
      return;
  }
  data[index].value=val;
  
  if(better)
    siftup(index);
  else 
    siftDown2(index);
  //printf("done siff\n");
#ifdef CHECK
  check("After sift down\n");
  checkHeapProperty("update");
#endif
  
  // int i=h->index[node];
  // h->data[i].value=value;
  // heapSiftup(h,i);
}

/*
  post: return the key of the first element of the heap
        return 0 if the heap is empty
*/
int Heap::peek()
{
  if(n>0)
    return data[0].node;
  else return -1;
}

/*
  post: remove the smallest valued element and return
        the node number.
	return -1 if the heap is empty.
*/
int Heap::remove(double * v)
{
    int node=-1;
    *v=-1;
    //check("before removed");
    if(n>0){
	node = data[0].node;
	*v=data[0].value;
	indices[node]=-1;
	//check("before sift down (remove)\n");
	siftDown(0);
	//    printf("removing %d from heap\n",node->id);
	n--;
	//check("after sift down (remove)\n");
    }
#ifdef CHECK
  check("removing");
#endif
  return node;
}  

void Heap::checkHeapProperty(char * s)
{
  int i;
  int nerr=0;
  for(i=0;i<n;i++){
    int child1=2*i+1;
    int child2=2*i+2;
    if(child1<n && data[child1].value<data[i].value){
      printf("-----------------------\n");
      printf("heap property is not valid (%d,%f) < (%d,%f)\n",
	     child1,data[child1].value,i,data[i].value);
      nerr++;
    }
    if(child2<n && data[child2].value<data[i].value){
      printf("-----------------------\n");
      printf("heap property is not valid (%d,%f) < (%d,%f)\n",
	     child1,data[child2].value,i,data[i].value);
      nerr++;
    }
  }
  if(nerr){
    printf(s);
    exit(0);
  }
}

void Heap::check(char * s)
{
    int nerr=0;
    static char lasts[1024];
    for(int i=0;i<n;i++){
	if(indices[data[i].node]!=i){
	    printf(s);
	    printf("[%d]=%d thinks its at %d not %d\n",i,data[i].node,indices[data[i].node],i);
	    nerr++;
	}
    }
    
    if(nerr){
	printf("happened at: %s\n",s);
	printf("prevopis:%s\n",lasts);
	exit(0);
    }else{
	sprintf(lasts,s);
    }
}

/*
  post: insert the node value pair into the heap.
        This is done by inserting at the end and
	moving the element up using siftup
*/
int Heap::insert(int node,double value)
{
#ifdef CHECK
  check("top of insert\n");
#endif
  
  if(indices[node]!=-1){
      printf("its already in heap\n");
      return 1;
  }
  data[n].node=node;
  data[n].value=value;
  indices[node]=n;
  n++;
#ifdef CHECK
  check("Before siftup in insert");
#endif 
  siftup(n-1);
#ifdef CHECK
  check("After siftup in insert");
#endif
  return 0;
}

/*void Heap::remove(int node)
{
    
    HeapElement node = (HeapElement)v;
    if(data[node->index]==node){
	data[node->index]=data[n-1];
	data[node->index]->index=node->index;
	n--;
	printf("%f / %f\n",node.value,data[node->index].value);
	indices->remove(node);
	check("siftDown2 in remove (void * v)\n");
	if(node.value<=data[node->index].value){
	    siftDown2(node->index);
	}
	else
	    siftup(node->index);
	
	checkHeapProperty("in remove(void*v)\n");
	check("after siftDown2 remove (void * v)\n");
	
    }
    else{
	printf("Heap::remove node is not consistent\n");
    }
}*/

/*
  post: print the heap
*/
void Heap::print()
{
  int i=0;
  while(i<n){
      printf("(%d,%2.0f),",data[i].node,data[i].value);
    i++;
  }
  printf("\n");
  //  printf("\n");
}

#ifdef HEAP_MAIN
#include "SlidingTile.h"
int main()
{

}
#endif
