#include "image.h"

extern "C"{
  #include <png.h>
}
extern "C"{
  #include <jpeglib.h>
}
template <class T>
Image<T>::Image(int _w,int _h,int _nchannels)
{
  data=0;
  refcount=0;
  width=_w;
  height=_h;
  nchannels=_nchannels;
  if(_w!=0&&_h!=0){
    refcount = new int;
    *refcount=1;
    data = new T[width*height*nchannels];
    //printf("allocing %p (%dx%dx%d)\n",data,width,height,nchannels);
  }
  else{
    //printf("null image\n");
  }
}

template <class T>
Image<T>::Image(const Image<T> & image)
{
  width=image.width;
  height=image.height;
  nchannels=image.nchannels;
  refcount=image.refcount;
  if(refcount)(*refcount)++;//it is possible for refcount to be null
  data=image.data;
}

template <class T>
void Image<T>::release(){
  if(refcount){
    (*refcount)--;
    if(*refcount==0){
      //printf("deleting %p (%dx%dx%d)\n",data,width,height,nchannels);
      delete refcount;
      delete [] data;
      data=0;
      refcount=0;
    }
  }
}

template <class T>
Image<T>::~Image(){
  release();
  width=0;
  height=0;
  nchannels=0;
}

template <class T>
Image<T> Image<T>::convolve(const Image<T> & filter){
  Image<T> fim(w,h,nchannels);
  int fh=2*(filter.h/2);
  int fw=2*(filter.w/2);
  if(filter.h-fh!=1 && filter.w-fw!=1){
    printf("filter is not proper dimension %d %d",fw,fh);
    fim=0;
    return fim;
  }
  fw/=2;
  fh/=2;
  fim=0;
  for(int i=fh;i<fim.h-fh;i++){
    for(int j=fw;j<fim.w-fw;j++){
      T vals[nchannels];
      for(int k=0;k<nchannels;k++)vals[k]=0;
      for(int ii=0;ii<filter.h;ii++){
	for(int jj=0;jj<filter.w;jj++){
	  for(int k=0;k<nchannels;k++){
	    vals[k]+=operator()(jj+j-fw,ii+i-fh,k)*filter(jj,ii);
	  }
	}
      }
      for(int k=0;k<nchannels;k++)fim(j,i,k)=vals[k];
    }
  }
  return fim;
}

template <class T>
Image<T> Image<T>::map(const Image<unsigned int> & indices){
  Image<T> image(w,h,nchannels);
  if(indices.w!=w || indices.h!=h){
    printf("Image::map inconsistent sizes for map %dx%d!=%dx%d\n",
	   w,h,indices.w,indices.h);
    return image;
  }
  for(int i=0;i<width*height;i++){
    for(int j=0;j<nchannels;j++){
      /*if(indices[i]>=width*height){
	printf("bad index in map %d>=%dx%d=%d\n",
	    indices[i],width*height,width,height);
	    }
	    else if(indices[i]<0){
	    printf("bad index in map %d<0\n",
		   indices[i]);
		   }*/
      image[nchannels*i+j]=(data[nchannels*indices[i]+j]);
    }
  }
  return image;
}

template <class T>
void Image<T>::checkSize(int resize,int * newWidth,int * newHeight){
  int w=0,h=0;
  while((1<<w)<width)
    w=w+1;
  while((1<<h)<height)
    h=h+1;
  if(newWidth){
    *newWidth=(1<<w);
    *newHeight=(1<<h);
  }
  if((1<<w)!=width ||
     (1<<h)!=height){
    if(resize){
      int format=channelToGL();
      
      Image<T> im2(1<<w,1<<h,nchannels);
      
      glPixelStorei(GL_UNPACK_ALIGNMENT,1);
      gluScaleImage(format,
		    width,height,
		    typeToGL(),
		    getDataPtr(),
		    1<<w,
		    1<<h,
		    typeToGL(),
		    im2.getDataPtr());
      //printf("resizing\n");
      operator=(im2);
    }    
  }
}

template <class T>
void Image<T>::initTexture(int mode){
  //the case when w==0 and h==0 was causing a problem
  //so just initialize a checker pattern
  if(w==0 && h==0){
    Image8 im8(8,8,3);
    for(int y=0;y<im8.w;y++){
      for(int x=0;x<im8.h;x++){
	im8(x,y,0)=(x+y)%2;
	im8(x,y,1)=(x+y+1)%2;
	im8(x,y,2)=(x+y*x%2)%2;
      }
    }
    im8.initTexture();
    return;
  }
  checkSize();

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,mode);
  
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  glTexImage2D(GL_TEXTURE_2D,0,channelToGL(),width,height,
	       0,channelToGL(),
	       typeToGL(),data);
}

template <class T>
void Image<T>::initTextureNoResize(double * sizex,double * sizey,
			 int mode){
  int neww,newh;
  checkSize(0,&neww,&newh);
  *sizex=(double)width/(double)neww;
  *sizey=(double)height/(double)newh;
  
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,mode);
  
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  
  int curtexw=0;
  int curtexh=0;
  int texcomp=0;
  glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_WIDTH,&curtexw);
  glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_HEIGHT,&curtexh);
  glGetTexLevelParameteriv(GL_TEXTURE_2D,0,GL_TEXTURE_COMPONENTS,&texcomp);

  if(curtexw!=neww||curtexh!=newh||texcomp!=channelToGL()){
    Image<T> newIm(neww,newh,nchannels);
    glTexImage2D(GL_TEXTURE_2D,0,channelToGL(),neww,newh,
		 0,channelToGL(),
		 typeToGL(),newIm.data);
  }
  glTexSubImage2D(GL_TEXTURE_2D,0,
		  0,0,width,height,
		  channelToGL(),typeToGL(),data);
}
template <>
int Image<unsigned char>::visit(Image<unsigned char> & vis,int x,int y,int on){
  int cnt=0;
  int * xs=new int[width*height];
  int * ys=new int[width*height];
  int head=0;
  int tail=0;
  xs[tail]=x;
  ys[tail++]=y;

      while(head!=tail){
	x=xs[head];
	y=ys[head];
	if(!vis(x,y)){
	  cnt++;
	  vis(x,y)=1;
	
	  if(x>0 && operator()(x-1,y,0)==on && !vis(x-1,y)){
	    xs[tail]=x-1;
	    ys[tail]=y;
	    tail++;
	    if(tail>=width*height)
	      tail=0;
	  }
	  if(y>0 && operator()(x,y-1,0)==on && !vis(x,y-1)){
	    xs[tail]=x;
	    ys[tail]=y-1;
	    tail++;
	    if(tail>=width*height)
	      tail=0;
	  }
	  if(x<width-1 && operator()(x+1,y,0)==on && !vis(x+1,y)){
	    xs[tail]=x+1;
	    ys[tail]=y;
	    tail++;
	    if(tail>=width*height)
	      tail=0;
	  }
	  if(y<height-1 && operator()(x,y+1,0)==on && !vis(x,y+1)){
	    xs[tail]=x;
	    ys[tail]=y+1;
	    tail++;
	    if(tail>=width*height)
	      tail=0;
	  }      
	}
	head++;
	if(head>=width*height)
	  head=0;
	
      }
      delete xs;
      delete ys;
      return cnt;
    }
template <>
void Image<unsigned char>
::fill(int x,int y,int from,int to,int depth,int & howmany){
      if(depth>width)return;
      if(operator()(x,y,0)==from){
	operator()(x,y,0)=to;
	howmany--;
	if(howmany<=0)return;
	if(x>0){
	  fill(x-1,y,from,to,depth+1,howmany);
	}
	if(y>0){
	  fill(x,y-1,from,to,depth+1,howmany);
	}
	if(x<width-1){
	  fill(x+1,y,from,to,depth+1,howmany);
	}
	if(y<height-1){
	  fill(x,y+1,from,to,depth+1,howmany);
	}
      }
    }
template <>
void Image<unsigned char>::clean(int size,int on){
  Image<unsigned char> visited(width,height,1);
  visited=0;
  printf("cleaning....................\n");
  
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      if(operator()(x,y,0)==on && !visited(x,y)){
	int thissize=visit(visited,x,y,on);
	
	if(thissize<size){
	  fill(x,y,(int)operator()(x,y,0),!on,0,thissize);
	}
      }
    }
  }
}

template <class T>
void Image<T>::spur(){
      for(int y=1;y<height-1;y++){
	for(int x=1;x<width-1;x++){
	  int up=(int)operator()(x,y+1,0);
	  int down=(int)operator()(x,y-1,0);
	  int left=(int)operator()(x-1,y,0);
	  int right=(int)operator()(x+1,y,0);
	  if(left && right){
	    operator()(x,y,0)=1;
	  }
	  if(!left && !right){
	    operator()(x,y,0)=0;
	  }
	  if(up && down){
	    operator()(x,y,0)=1;
	  }
	  if(!up && !down){
	    operator()(x,y,0)=0;
	  }

	}
      }
    }

template <class T>
Image<T> Image<T>::getChannel(int which){
  if(nchannels==0)return Image<T>(0,0,0);
  Image<T> res(width,height,1);
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      res(x,y)
	=operator()(x,y,which);
    }
  }
  return res;
}
template <class T>
void Image<T>::setToChannel(int which){
  if(nchannels==0)return;
  
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      for(int k=0;k<nchannels;k++){
	operator()(x,y,k)
	  =operator()(x,y,which);
      }
    }
  }
}

template <class T>
void Image<T>::operatorEquals(const Image & other){
  if(&other!=this){
    release();
    width=other.width;
    height=other.height;
    nchannels=other.nchannels;
    refcount=other.refcount;
    if(refcount)(*refcount)++;//possible for refcount to be null of w=h=nchannels=data=0
    data=other.data;
  }
}

//overloaded operator= for each type
template <>
void Imagef::operator=(const Image8 & image){
    operatorEquals(Image<float>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(float)(image[i])/255.0f;
    }
}
template <>
void Imagef::operator=(const Imagef & image){
    operatorEquals(image);
}
template <>
void Imagef::operator=(const Image16 & image){
    operatorEquals(Image<float>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(float)(image[i])/65536.0f;
    }
}
template <>
void Imagef::operator=(const Image32 & image){
  operatorEquals(Image<float>(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(float)(image[i])/(float)(1<<31);
  }
}
template <>
void Image8::operator=(const Image8 & image){
    operatorEquals(image);
}
template <>
void Image8::operator=(const Imagef & image){
    operator=(Image<unsigned char>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(unsigned char)(image[i]*255.0f);
    }
}
template <>
void Image8::operator=(const Image16 & image){
    operator=(Image<unsigned char>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(unsigned char)(((double)image[i]*255.0)/65535.0);
    }
}
template <>
void Image8::operator=(const Image32 & image){
  operator=(Image<unsigned char>(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned char)(((double)image[i]*255.0)/(double)(1<<31));
  }
}
template <>
void Image16::operator=(const Image16 & image){
  operatorEquals(image);
}
template <>
void Image16::operator=(const Image8 & image){
    operator=(Image<unsigned short>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
      data[i]=(unsigned short)(image[i]<<8);
    }
}
template <>
void Image16::operator=(const Imagef & image){
    operator=(Image<unsigned short>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(unsigned short)(image[i]*65535.0f);
    }
}
template <>
void Image16::operator=(const Image32 & image){
  operator=(Image<unsigned short>(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned short)(image[i]/65536);
  }
}
template <>
void Image32::operator=(const Image32 & image){
  operatorEquals(image);
}
template <>
void Image32::operator=(const Image8 & image){
  operator=(Image32(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned int)(image[i]<<24);
  }
}
template <>
void Image32::operator=(const Imagef & image){
  operator=(Image32(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned int)(image[i]*(1<<31));
  }
}
template <>
void Image32::operator=(const Image16 & image){  
  operator=(Image32(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned int)(image[i]<<16);
  }
}


typedef struct {
  unsigned char id_length;
  unsigned char color_map_type;
  unsigned char image_type;
  unsigned char color_map_spec[5];
  
  unsigned short x0;
  unsigned short y0;
  unsigned short w;
  unsigned short h;
  unsigned char bpp;
  unsigned char image_descriptor;
  
}TgaHeader;

template <class T>
int Image<T>::readTGA(const char * filename){
  Image8 im;
  FILE * f = fopen(filename,"rb");
  if(!f)return 0;
  
  TgaHeader header;
  //FIXME:need to fix up for endianness
  if(1!=fread(&header,sizeof(header),1,f)){
    printf("can not read TGA\n");
    return 0;
  }
  if(header.x0 ||header.y0){
    printf("ignoring non-zero origin\n");
  }
  if(header.id_length){
    printf("don't know how to read in the header\n");
    printf("can not read TGA\n");
    return 0;
  }
  if(header.color_map_type){
    printf("don't know how to read in color mapped images\n");
    printf("can not read TGA\n");
    return 0;
  }
  if(!(header.image_type==2||header.image_type==3
       ||header.image_type==0x0A||header.image_type==0x0B)){
    printf("unknown image type\n");
    printf("can not read tga\n");
    return 0;
  }
  header.bpp/=8;
  printf("reading image %d %d %d\n",header.w,header.h,header.bpp);
  im=Image8(header.w,header.h,header.bpp);
  int nread;
  /*
  //Read in one pixel at a time
  for(int y=0;y<h;y++){
    for(int x=0;x<w;x++){
      if(bpp!=(nread=fread(im.data+(y*w+x)*bpp,1,bpp,f))){
	printf("nread is %d not %d\n",nread,bpp);
	fclose(f);
	return 0;
      }
    }
    }*/
  //Read in the image
  if(header.image_type==2||
     header.image_type==3){
    if(header.w*header.h*header.bpp!=(nread=fread(im.data,1,header.w*header.h*header.bpp,f)))
      return 0;
  }
  else{
    printf("got RLE image %dx%d\n",im.w,im.h);
    //Image type is 0x0A or 0x0B
    //process packets until out of file
    char buffer[256*4];
    int x=0;
    int y=0;
    while(y<im.h){
      unsigned char rep;
      if(fread(&rep,sizeof(char),1,f)!=1){
	fclose(f);
	return 0;
      }
      //if first bit set, then it is a RLE packet
      if(rep&0x80){
	unsigned char buffer[4];
	rep=(rep&0x7F)+1;
	if(fread(buffer,header.bpp,1,f)!=1){
	  printf("problem reading in RLE tga at %d %d\n",x,y);
	  fclose(f);
	  return 0;
	}
	for(int j=0;j<rep;j++){
	  for(int k=0;k<im.nchannels;k++)
	    im(x,y,k)=buffer[k];
	  x++;
	  if(x>=im.w){
	    x=0;
	    y++;
	    if(j!=rep-1){
	      printf("j=%d, rep=%d\n",j,rep);
	      printf("warning: RLE packets are not supposed to carry over scanlines\n");
	      break;
	    }
	  }
	}
      }
      //not RLE
      else{
	rep=(rep&0x7F)+1;
	int nread=fread(im.data+y*im.w*im.nchannels+x*im.nchannels,header.bpp,rep,f);
	if(nread!=rep){
	  printf("problem reading in RLE tga at %dx%d\n",x,y);
	  printf("read %d, should have read %d\n",nread,rep);
	  fclose(f);
	  return 0;
	}
	x+=rep;
	//This allows non-RLE packets to go over the
	//scanline, set x=0 if you don't want this
	if(x>=im.w){
	  x=x-im.w;
	  y++;
	}
      }
    }
  }
  fclose(f);

  if(header.bpp==3 || header.bpp==4) // flip red and blue
    for(int i=0;i<header.w*header.h*header.bpp;i+=header.bpp)
    {
      char tmp=im.data[i];
      im.data[i]=im.data[i+2];
      im.data[i+2]=tmp;
    }
  
  //left to right ordering
  if(!(header.image_descriptor&0x10)){
    //printf("maybe this is just for the offset\n");
  }
  if(!(header.image_descriptor&0x20)){
    printf("flipping y\n");
    
    for(int y=0;y<im.h/2;y++){
      for(int x=0;x<im.w;x++){
	for(int k=0;k<im.nchannels;k++){
	  unsigned char back=im(x,y,k);
	  im(x,y,k)=im(x,im.h-y-1,k);
	  im(x,im.h-y-1,k)=back;
	}
      }
    }
  }
  operator=(im);
  //return im;
}

template <class T>
int Image<T>::readPNG(const char * filename){
  printf("%% reading png %s\n",filename);
  png_bytep * row_pointers=0;
  unsigned char header[8];
  FILE *file = fopen(filename, "rb");
  if (!file) return 0;
  fread(header,1,8,file);
  int is_png = !png_sig_cmp(header,0,8);
  if (!is_png) {
    fclose(file);
    return 0;
  }
  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,0,0,0);
  if(!png_ptr){
    fclose(file);
    return 0;
  }
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    png_destroy_read_struct(&png_ptr,
			    (png_infopp)NULL, (png_infopp)NULL);
    fclose(file);
    return 0;
  }
  png_infop end_info = png_create_info_struct(png_ptr);
  if (!end_info) {
    png_destroy_read_struct(&png_ptr, &info_ptr,
			    (png_infopp)NULL);
    fclose(file);
    return 0;
  }
  if(setjmp(png_ptr->jmpbuf)){
    png_destroy_read_struct(&png_ptr, &info_ptr,&end_info);
    fclose(file);
    return 0;
  }
  printf("%% setting io\n");
  png_init_io(png_ptr, file);
  png_set_sig_bytes(png_ptr,8);
  png_read_info(png_ptr, info_ptr);
  png_uint_32 w,h;
  int bit_depth,color_type,interlace_type,compression_type,filter_type;
  png_get_IHDR(png_ptr, info_ptr, &w, &h,
	       &bit_depth, &color_type, &interlace_type,
	       &compression_type, &filter_type);
  printf("%% width:%d height:%d\n",w,h);
  if (color_type == PNG_COLOR_TYPE_PALETTE &&
      bit_depth <= 8) png_set_expand(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY &&
      bit_depth < 8) png_set_expand(png_ptr);
  if (png_get_valid(png_ptr, info_ptr,
		    PNG_INFO_tRNS)) png_set_expand(png_ptr);
 
  if (bit_depth == 16){
    //png_set_strip_16(png_ptr);//used to just set strip_16 and then continue
    printf("%% reading 16 bit image (used to strip it)\n");
    unsigned int bigend=1;
    char * cstr=(char *)&bigend;
    if(*cstr==1)png_set_swap(png_ptr);
  }
  if (bit_depth < 8)
    png_set_packing(png_ptr);
  /*need to do this if care about endian
  if (bit_depth == 16)
        png_set_swap(png_ptr);
  */
  png_read_update_info(png_ptr, info_ptr);
  int nch=1;
  if(color_type==PNG_COLOR_TYPE_RGB)nch=3;
  else if(color_type==PNG_COLOR_TYPE_RGBA)nch=4;
  
  row_pointers = new png_bytep[h];
  if(bit_depth==16){
    Image16 im(w,h,nch);
    for(int i=0;i<h;i++)
      row_pointers[i]=(unsigned char *)(im.data+i*w*nch);
    
    png_read_image(png_ptr, row_pointers);
    delete [] row_pointers;
    png_read_end(png_ptr, end_info);
    png_destroy_read_struct(&png_ptr, &info_ptr,&end_info);
    fclose(file);
    operator=(im);
    return 1;
  }
  Image8 im(w,h,nch);
  for(int i=0;i<h;i++)
    row_pointers[i]=im.data+i*w*nch;
  
  png_read_image(png_ptr, row_pointers);
  delete [] row_pointers;
  png_read_end(png_ptr, end_info);
  png_destroy_read_struct(&png_ptr, &info_ptr,&end_info);
  fclose(file);
  operator=(im);
  return 1;
}

template <class T>
int Image<T>::readJPG(const char * filename){
  FILE * ifile = fopen(filename,"rb");
  if(!ifile)return 0;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPLE * jsp;
  
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, ifile);
  
  jpeg_read_header(&cinfo, TRUE);
  jpeg_start_decompress(&cinfo); 

  Image8 image(cinfo.output_width,cinfo.output_height,cinfo.output_components);
  
  jsp = (JSAMPLE *) new char[cinfo.output_width*cinfo.output_components];
  
  int i=0;
  while(i<cinfo.output_height){
    int ret = jpeg_read_scanlines(&cinfo,&jsp,1);
    if(ret==0){
      fprintf(stderr,"Error reading scanlines\n");
      return 0;
    }
    memcpy(image.data+i*image.w*image.nchannels,jsp,image.w*image.nchannels);
    i++;
  }

  delete jsp;
  
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo); 
  fclose(ifile);

  operator=(image);
}

template <class T>
int Image<T>::readPPM(const char * filename){
  FILE * f = fopen(filename,"rb");
  int width,height,number,max,i=0;
  int nread=0;
  Image<unsigned char> image;
  if(!f){
    printf("Error opening %s\n",filename);
    return 0;
  }
  char buffer[512];
  if(NULL==fgets(buffer,512,f)){
    fclose(f);
    return 0;
  }
  
  if(buffer[0]!='P'){
    fclose(f);
    return 0;
  }
  number=buffer[1]-'0';
  
  while(1){
    if(NULL==fgets(buffer,512,f)){
      fclose(f);
      return 0;
    }
    if(buffer[0]!='#')
      break;
  }
  
  sscanf(buffer,"%d %d",&width,&height);
  if(NULL==fgets(buffer,512,f)){
    fclose(f);
    return 0;
  }
  //P1 is the bitmap, no max
  if(number!=1){
    sscanf(buffer,"%d",&max);
    if(max>65535){
      fclose(f);
      return 0;
    }
  }
  
  if(max>255){
    printf("ppm with max>255 not handled\n");
    exit(-1);
  }
  
  printf("reading image P%d, max=%d,width=%d,height=%d\n",
	 number,max,width,height);
  
  switch(number){
  case 1:
    printf("not handled yet\n");
    fclose(f);
    return 0;
    break;
  case 2:
    image=Image<unsigned char>(width,height,1);  
    for(i=0;i<width*height;i++){
      int n;
      
      if(!fscanf(f,"%d",&n)){
	fclose(f);
	return 0;
      }
      if(i==0)
	printf("[%d]\n",n);
      image[i]=n;
    }
    break;
  case 3:
    image=Image<unsigned char>(width,height,3);
    for(i=0;i<width*height*3;i++){
      int n;
      if(!fscanf(f,"%d",&n)){
	fclose(f);
	return 0;
      }
      image[i]=n;
    }
    break;
		
  case 5:
    //This assumes same endian in machine for max>255
    image=Image<unsigned char>(width,height,1);
    if(((unsigned int)width*height)!=
       fread((unsigned char *)image.getDataPtr(),(max>255)+1,width*height,f)){
      fclose(f);
      return 0;
    }
    break;
  case 6:
    image=Image<unsigned char>(width,height,3);
    //This assumes same endian in machine for max>255
		
    if(width*height*3!=
       (nread=fread(image.getDataPtr(),(max>255)+1,width*height*3,f))){
      fclose(f);
      printf("unexpected EOF %d!=%d\n",nread,width*height*3);
      return 0;
    }
    break;
  default: 
    printf("unknown magic number %d",number);
    fclose(f);
    return 0;
  }
	
  operator=(image);
	
  fclose(f);
  return 1;
}
template <class T>
int Image<T>::writePNG(const char * filename){
  Image8 im;
  im=(*this);
  FILE *f = fopen(filename, "wb");
  if(!f) return 0;
  png_structp png_ptr = png_create_write_struct
    (PNG_LIBPNG_VER_STRING,0,0,0);
  if(!png_ptr){
    fclose(f);
    return 0;
  }
  
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if(!info_ptr) {
    png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
    fclose(f);
    return 0;
  }
  if(setjmp(png_ptr->jmpbuf)) {
    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(f);
    return 0;
  }
  png_init_io(png_ptr,f);
  
  //Defaults from the libpng manual
  png_set_compression_level(png_ptr,Z_BEST_COMPRESSION);
  /* set other zlib parameters */
  /*png_set_compression_mem_level(png_ptr, 8);
  png_set_compression_strategy(png_ptr,Z_DEFAULT_STRATEGY);
  png_set_compression_window_bits(png_ptr,15);
  png_set_compression_method(png_ptr,8);*/
 
  png_uint_32 w=width,h=height;
  int bit_depth=8,color_type=PNG_COLOR_TYPE_GRAY;
  int interlace_type=PNG_INTERLACE_NONE;
  int compression_type=PNG_COMPRESSION_TYPE_DEFAULT;
  int filter_type=PNG_FILTER_TYPE_DEFAULT;
  if(nchannels==3)color_type=PNG_COLOR_TYPE_RGB;
  else if(nchannels==4)color_type=PNG_COLOR_TYPE_RGBA;
  png_set_IHDR(png_ptr, info_ptr, width, height,
	       bit_depth, color_type, interlace_type,
	       compression_type, filter_type);
  png_write_info(png_ptr, info_ptr);

  /* Endian stuff, want little endian
  if (bit_depth > 8)
       png_set_swap(png_ptr);
  */
  png_bytep * row_pointers=new png_bytep[height];
  for(int i=0;i<height;i++)
    row_pointers[i]=im.data+i*im.width*im.nchannels;
  
  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, info_ptr);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  delete [] row_pointers;
  fclose(f);
}

template <>
int Image<unsigned char>::writeTGA(const char * filename){
  FILE * f = fopen(filename,"wb");
  Image8 im;
  im=*this;
  if(!f)return 0;
  unsigned short w=width;
  unsigned short h=height;
  unsigned char bpp=nchannels;
  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  char tga_header_gray[12] = {0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
  if(bpp==3 || bpp==4)fwrite(tga_header_rgb,sizeof(tga_header_rgb),1,f);
  else fwrite(tga_header_gray,sizeof(tga_header_gray),1,f);
  fwrite(&w,2,1,f);
  fwrite(&h,2,1,f);
  fputc(bpp*8,f);
  if(bpp==4)fputc(32|8,f);
  else fputc(32,f); //flipping

  if(bpp==4)
  for(int i=0;i<w*h*bpp;i+=bpp)
  {
    fputc(im.data[i+2],f);
    fputc(im.data[i+1],f);
    fputc(im.data[i],f);
    fputc(im.data[i+3],f);
  }
  else if(bpp==3)
  for(int i=0;i<w*h*bpp;i+=bpp)
  {
    fputc(im.data[i+2],f);
    fputc(im.data[i+1],f);
    fputc(im.data[i],f);
  }

  else
    fwrite(im.data,w*h*bpp,1,f);
  fclose(f);
  return 18+w*h*bpp;
}

/*
int Image<unsigned char>::writeJPG(char * filename){
{
  FILE * ofile=fopen(filename,"wb");
  if(!ofile)return 0;
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPLE * jsp;

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, ofile);
  
  if(nchannels==1)
    cinfo.input_components=JCS_GRAYSCALE;
  else {
    printf("unable to write color jpegs\n");
    return 0;
  }
  
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo,30,1); 
  jpeg_start_compress(&cinfo,TRUE);
    
  jsp = (JSAMPLE *)new char[cinfo.image_width];

  int i=0;
  while(cinfo.next_scanline<cinfo.image_height){
    
    int nread=fread(jsp,1,cinfo.image_width,ifile);
    int ret = jpeg_write_scanlines(&cinfo,&jsp,1);
    if(ret==0){
      fprintf(stderr,"Error writing scanlines\n");
      return(0);
      break;
    }
  }
  delete jsp;
  
  fflush(ifile);
  jpeg_finish_compress(&cinfo);
  fclose(ofile);
  jpeg_destroy_compress(&cinfo); 
  fclose(ifile);
  return 1;
}
 */
template <class T>
int Image<T>::writePPM(const char * filename){
  Image8 im;
  im=*this;
  ::writePPM((unsigned char *)im.data,im.w,im.h,filename,im.nchannels);
  return 1;
}

void writePPM(unsigned char * image,int w,int h,const char * filename,int nchannels,int flip)
{
    FILE * file = fopen(filename,"wb");
    int i=0;

    if(nchannels==3){
	fprintf(file,"P6\n%d %d\n%d\n",w,
		h,
		255);
	if(flip)
	  for(i=0;i<h;i++){
	    fwrite(image+(h-i-1)*w*3,1,w*3,file);
	  }
	else
	  fwrite(image,1,w*h*3,file);
	  
    }
    else if(nchannels==1){
	fprintf(file,"P5\n%d %d\n%d\n",w,
		h,
		255);
	if(flip)
	  for(i=0;i<h;i++){
	    fwrite(image+(h-i-1)*w,1,w,file);
	  }
	else
	  fwrite(image,1,w*h,file);
    }
    fclose(file);    
}

template <class T>
int Image<T>::read(const char * filename){
  if((strstr(filename,".ppm")==
      filename+strlen(filename)-4) ||
     (strstr(filename,".pgm")==
      filename+strlen(filename)-4)){
    return readPPM(filename);
  }
  else if((strstr(filename,".tga")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".TGA")==
	   filename+strlen(filename)-4)){
    return readTGA(filename);
  }
  else if((strstr(filename,".jpg")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".JPG")==
	   filename+strlen(filename)-4)){
    return readJPG(filename);
  }
  else if((strstr(filename,".png")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".PNG")==
	   filename+strlen(filename)-4)){
    return readPNG(filename);
  }
  return 0;
}
template <>
int Image<unsigned char>::save(const char * filename){
  if((strstr(filename,".ppm")==
      filename+strlen(filename)-4) ||
     (strstr(filename,".pgm")==
      filename+strlen(filename)-4)){
    return writePPM(filename);
  }
  else if((strstr(filename,".tga")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".TGA")==
	   filename+strlen(filename)-4)){
    return writeTGA(filename);
  }
  return 0;
}


template <class T> 
Image<T> Image<T>::resize(int w,int h)
{
  Image<T> result(w,h,nchannels);
  double xd;
  double yd;
  int xdi,ydi;
  int x;
  int y;

  if(width<w && height < h){
    printf("warning, interpolation is not correct for resize %dx%d-->%dx%d\n",
	   width,height,w,h);
    double nxd=(double)width/(double)w;
    double nyd=(double)height/(double)h;
    for(int y=0;y<h;y++){
      for(int x=0;x<w;x++){
	for(int k=0;k<nchannels;k++){
	  result(x,y,k)=operator()((int)(nxd*x),(int)(nyd*y),k);
	}
      }
    }
    return result;
  }
  else if(width<w || height < h){
    printf("general image resizes are not handled properly %dx%d-->%dx%d\n",
	   width,height,w,h);
    exit(0);
  }

  double nxd=(double)width/(double)w;
  double nyd=(double)height/(double)h;
  int nx=(int)ceil(nxd);
  int ny=(int)ceil(nyd);
  double a,b;
  double weight=0;
  double tmp[nchannels];
  printf("%f(%d) %f(%d)\n",nxd,nx,nyd,ny);
  result=0;
  for(y=0;y<h;y++){
    int ystart;
    int yend;
    yd=(double)(y*height)/((double)(h));
    ystart=(int)yd;
    yend=ystart+ny;
    for(x=0;x<w;x++){
      int xstart;
      int xend;
      xd=(double)(x*width)/((double)(w));
      xstart=(int)xd;
      xend=xstart+nx;
      weight=0;
      memset(tmp,0,sizeof(tmp));
      for(ydi=ystart;ydi<yend;ydi++){
	b=1;
	if(ydi==ystart)
	  b=1.0-(yd-ystart);
	else if(ydi==yend-1)
	  b=(yd+nyd-ydi);
      	for(xdi=xstart;xdi<xend;xdi++){
	  a=1;
	  if(xdi==xstart)
	    a=1.0-(xd-xstart);
	  else if(xdi==xend-1)
	    a=(xd+nxd-(xend-1));
	  for(int k=0;k<nchannels;k++){
	    double x1y1=operator()(xdi,ydi,k);
	    tmp[k]+=(a*b*x1y1);//FIXME: cast a problem for templates
	  }
	  weight+=a*b;
	}
      }
      if(::fabs(weight-nxd*nyd)>0.01)
	printf("weight %f %f\n",weight,nxd*nyd);
      for(int k=0;k<nchannels;k++){
       	result(x,y,k)=(T)(tmp[k]/(nxd*nyd));//FIXME: cast a problem for template types
	if(result(x,y,k)>1){
	  //printf("clamping %d,%d,%d\n",x,y,k);
	  //result(x,y,k)=1;
	  //exit(0);
	}
      }
    }
  }
  return result;
}




template class Image<float>;
template class Image<unsigned char>;
template class Image<unsigned short>;
template class Image<unsigned int>;
