/*
  This great Image class is now independent of the rest of the capture
  system.  This should have been done a long time ago!

  Many of the old routines have been moved into image_utils.h/cc.
*/
#include <typeinfo>
#include <stdio.h>
#include <stdlib.h>
#ifdef USING_GLEE
#include <GLee.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>

//Expose some functionality to python
#ifdef __PY_PLUS_PLUS
#include <boost/python/tuple.hpp>
#include <boost/python/extract.hpp>
#include <boost/python/detail/api_placeholder.hpp>
using boost::python::tuple;
using boost::python::extract;
#endif

#ifndef IMAGE_H
#define IMAGE_H

#define imMAX(a,b)((a)>(b)?(a):(b))
#define imMIN(a,b)((a)<(b)?(a):(b))

//Hopefully this doesn't mess with anyone,
//these two should have pretty standard meanings anyhow
#ifndef MAX
#define MAX imMAX
#endif
#ifndef MIN
#define MIN imMIN
#endif

template <class T>
class Image{
 protected:
    int * refcount;
 public:
    ///static int gcount; //<<this was unused and undefined
    T * data;
    union{
      int width;
      int w;
    };
    union{
      int height;
      int h;
    };
    int nchannels;
  
    Image(int _w=0,int _h=0,int _nchannels=0);
    Image(const Image<T> & image);

    //accessors
    int getWidth() const{
      return width;
    }
    int getHeight() const{
      return height;
    }
    void release();
    virtual ~Image();
    
    Image map(const Image<unsigned int> & indices);
    void operator=(int what){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]=(T)what;
    }
    void operator=(double what){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]=(T)what;
    }
    void operatorEquals(const Image & other);

    void operator=(const Image<float>  & image);
    void operator=(const Image<unsigned char>  & image);
    void operator=(const Image<unsigned short>  & image);
    void operator=(const Image<unsigned int>  & image);
    
    Image operator+(int add){
	Image<T> image(width,height,nchannels);
	for(int i=0;i<width*height*nchannels;i++)
	    image.data[i]=(data[i]+add);
	return image;
    }
    void operator+=(int add){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]+=add;
    }
    Image operator+(double add){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=(T)(data[i]+add);
      return image;
    }
    void operator+=(double add){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]=(T)(data[i]+add);
    }
    Image operator-(int add){
	Image<T> image(width,height,nchannels);
	for(int i=0;i<width*height*nchannels;i++)
	    image.data[i]=data[i]-add;
	return image;
    }
    Image operator-(double  add){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=(T)(data[i]-add);
      return image;
    }
    void operator-=(int add){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]-=add;
    }
    void operator-=(double add){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]=(T)(data[i]-add);
    }
    Image operator*(const int & mult){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=(T)(data[i]*mult);
      return image;
    }
    Image operator*(const double mult){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=(T)(data[i]*mult);
      return image;
    }
    void operator*=(const int & mult){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]*=mult;
    }
    void operator*=(const double & mult){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]=(T)(data[i]*mult);
    }
    Image operator/(const double div){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=(T)(((double)data[i])/div);
      return image;
    }
    void operator/=(const double & d){
      for(int i=0;i<width*height*nchannels;i++)
	data[i]=(T)(((double)data[i])/d);
    }
    Image operator+(const Image & other){
      if(other.width!=width||
	 other.height!=height||
	 other.nchannels!=nchannels){
	printf("invalid sizes for operator+\n");
	exit(1);
      }
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=data[i]+other.data[i];
      
      return image;
    }
    void operator+=(const Image<T> & other){
      if(other.width!=width||
	 other.height!=height||
	 other.nchannels!=nchannels){
	printf("invalid sizes for operator+=\n");
	exit(1);
      }
      for(int i=0;i<width*height*nchannels;i++)
	data[i]+=other.data[i];
    }
    Image<T> operator-(const Image<T> & other){
      if(other.width!=width||
	 other.height!=height||
	 other.nchannels!=nchannels){
	printf("invalid sizes for operator-\n");
	exit(1);
      }
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++)
	image.data[i]=data[i]-other.data[i];
      return image;
    }
    void operator-=(const Image<T> & other){
      if(other.width!=width||
	 other.height!=height||
	 other.nchannels!=nchannels){
	printf("invalid sizes for operator-=\n");
	exit(1);
      }
      for(int i=0;i<width*height*nchannels;i++)
	data[i]-=other.data[i];
    }
    Image operator*(const Image<T> & other){
      if(other.width!=width||
	 other.height!=height){
	printf("invalid sizes for operator*\n(%d,%d,%d),(%d,%d,%d)\n",
	       width,height,nchannels,other.width,
	       other.height,other.nchannels);
	exit(1);
      }
      Image<T> image(width,height,nchannels);
      if(other.nchannels==1){
	int i=0;
	for(int j=0;j<width*height;j++){
	  int top=i+nchannels;
	  for(;i<top;i++){
	    image.data[i]=(data[i])*(other.data[j]);
	  }
	}
      }
      else{
	for(int i=0;i<width*height*nchannels;i++){
	  image.data[i]=(data[i])*(other.data[i]);
	}
      }
      return image;
    }
    Image operator|(const Image & other){
      if(other.width!=width||
	 other.height!=height||
	 other.nchannels!=nchannels){
	printf("invalid sizes for operator|\n");
	exit(1);
      }
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image.data[i]=((unsigned int)data[i])|((unsigned int)other.data[i]);
      }
      return image;
    }
    void operator&=(const Image & other){
      if(other.width!=width||
	 other.height!=height||
	 other.nchannels!=nchannels){
	printf("invalid sizes for operator&\n");
	exit(1);
      }
      for(int i=0;i<width*height*nchannels;i++){
	data[i]=(data[i]!=0)&(other.data[i]!=0);
      }
    }
    Image operator&(const Image & other){
	if(other.width!=width||
	   other.height!=height||
	   other.nchannels!=nchannels){
	    printf("invalid sizes for operator&\n");
	    exit(1);
	}
	Image<T> image(width,height,nchannels);
	for(int i=0;i<width*height*nchannels;i++){
	  image.data[i]=(data[i]!=0)&(other.data[i]!=0);
	}
	return image;
    }
    Image threshLess(double * a){
      Image<T> image(width,height,1);
      for(int y=0;y<height;y++){
	for(int x=0;x<width;x++){
	  int set=1;
		for(int k=0;k<nchannels;k++){
		  set&=(operator()(x,y,k)<a[k]);
		}
		image(x,y)=set;
	}
      }
      return image;
    }
    Image operator>(const double & t){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=data[i]>t;
      }
      return image;
    }
    Image operator>(const Image & o){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=data[i]>o.data[i];
      }
      return image;
    }
    Image operator>=(const double & t){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=data[i]>=t;
      }
      return image;
    }
    Image operator>=(const Image & o){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=data[i]>=o.data[i];
      }
      return image;
    }
    Image operator<(const double & t){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=data[i]<t;
      }
      return image;
    }
    Image operator<=(const double & t){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=data[i]<=t;
      }
      return image;
    }
    Image operator!=(const double & t){
      Image<T> image(width,height,nchannels);
      for(int i=0;i<width*height*nchannels;i++){
	image[i]=(data[i]!=t);
      }
      return image;
    }
    Image maximum(){
      Image<T> image(width,height,1);
      for(int y=0;y<height;y++){
	for(int x=0;x<width;x++){
	  T m=0;
	  for(int k=0;k<nchannels;k++){
	    T what=operator()(x,y,k);
	    m=(m>what)?m:what;
	  }
	  image(x,y)=m;
	}
      }
      return image;
    }
    Image<T> len(){
      Image<T> image(width,height,1);
      for(int y=0;y<height;y++){
	for(int x=0;x<width;x++){
	  double l=0;
	  for(int k=0;k<nchannels;k++){
	    double what=operator()(x,y,k);
	    l+=what*what;
	  }
	  image(x,y)=(T)::sqrt(l);
	}
      }
      return image;
    }
    void setSubimage(int x1,int y1,Image sim){
      if(nchannels!=sim.nchannels){
	printf("illegal setSubimage call, nchannels not equal\n");
	return;
      }
      if(x1<0)x1=0;
      if(y1<0)y1=0;
      if(x1+sim.w>w){
	printf("setSubimage is larger than image itself\n");
	return;
      }
      if(y1+sim.h>h){
	printf("setSubimage is larger than image itself\n");
	return;
      }
      for(int y=0;y<sim.h;y++){
	for(int x=0;x<sim.w;x++){
	  for(int k=0;k<sim.nchannels;k++){
	    operator()(x+x1,y+y1,k)=sim(x,y,k);
	  }
	}
      }
    }
    Image subimage(int x1,int y1,int w,int h){
      if(w<=0)w=1;
      if(h<=0)h=1;
      if(x1<0)x1=0;
      if(y1<0)y1=0;
      if(y1>height-h)y1=height-h;
      if(x1>width-w)x1=width-w;
      Image image(w,h,nchannels);
      for(int y=0;y<h;y++){
	for(int x=0;x<w;x++){
	  for(int k=0;k<nchannels;k++){
	    image(x,y,k)=operator()(x+x1,y+y1,k);
	  }
	}
      }
      return image;
    }
    T & operator[](int i) const{
	return data[i];
    }
    T & operator()(int x,int y,int channel=0) const{
	return data[y*width*nchannels+x*nchannels+channel];
    }
    
    void print(char * fmt){
	for(int ci=0;ci<nchannels;ci++){
	    printf("channel %d\n",ci);
	    for(int y=0;y<height;y++){
		for(int x=0;x<width;x++){
		    printf(fmt,data[y*width*nchannels+x*nchannels+ci]);
		}
		printf("\n");
	    }
	}
    }
    int channelToGL(){
	GLenum m[5]={0,GL_LUMINANCE,
		     GL_LUMINANCE_ALPHA,
		     GL_RGB,GL_RGBA};
	return m[nchannels];
    }
    int typeToGL(){
	if(typeid(T)==typeid(unsigned char)){
	    return GL_UNSIGNED_BYTE;
	}
	else if(typeid(T)==typeid(float)){
	    return GL_FLOAT;
	}
	else if(typeid(T)==typeid(unsigned short)){
	  return GL_UNSIGNED_SHORT;
	}
	else if(typeid(T)==typeid(unsigned int)){
	  return GL_UNSIGNED_INT;
	}
	else{
	    printf("unhandled type\n");

	}
    }

    void checkSize(int resize=1,int * newWidth=0,int * newHeight=0);
    void abs(){
	for(int i=0;i<width*height*nchannels;i++){
	  data[i]=(T)::abs((int)data[i]);
	}
    }
    void fabs(){
      for(int i=0;i<width*height*nchannels;i++){
	data[i]=(T)::fabs((double)data[i]);
      }
    }

    int visit(Image<unsigned char> & vis,
				    int x,int y,int on);
    void fill(int x,int y,int from,int to,int depth,int & howmany);
    void clean(int size,int on=1);
    
    Image getChannel(int which);
    void setToChannel(int which);
    void spur();
    void initTexture(int mode=GL_MODULATE);
    void initTextureNoResize(double * sizex,double * sizey,
			     int mode=GL_MODULATE);
    virtual T  * getDataPtr(){
	return data;
    }
    void getRange(double & min,double & max){
      min=10e10;
      max=-10e10;
      for(int i=0;i<w*h*nchannels;i++){
	min=imMIN(min,data[i]);
	max=imMAX(max,data[i]);
      }
    }
    int readPNG(const char * filename);
    int readJPG(const char * filename);
    int readTGA(const char * filename);
    int writePNG(const char * filename);
    int writeTGA(const char * filename);
    int readPPM(const char * filename);
    int writePPM(const char * filename);
    int read(const char * filename);
    int save(const char * filename);
    //NOTE: this function has been moved to ImageUtils.
    //Image removeDistortion(Matrix & ,Matrix & );
    Image resize(int w,int h);
    Image<T> convolve(const Image<T> & f);
    Image<T> medianFilter(int size){
      int halfsize=size/2;
      int s_2=size*size;
      Image<T> res(w,h,nchannels);
      T * pts = new T[3*s_2];
      if(nchannels>1){
	printf("median filter not implemented on nchannels>%d\n",
	       nchannels);
      }
      for(int y=0;y<h;y++){
	int yb=imMAX(0,y-halfsize);
	int yt=imMIN(y+halfsize,h-1);
	for(int x=0;x<w;x++){
	  int npts=0;
	  int xb=imMAX(0,x-halfsize);
	  int xt=imMIN(x+halfsize,w-1);
	  for(int yy=yb;yy<=yt;yy++){
	    for(int xx=xb;xx<=xt;xx++){
	      T val=operator()(xx,yy);
	      int k=0;
	      for(k=0;k<npts;k++){
		if(pts[k]>val)break;
	      }
	      //Finish this
	      for(int k2=npts;k2>k;k2--){
		pts[k2]=pts[k2-1];
	      }
	      pts[k]=val;
	      npts++;
	    }
	  }
	  res(x,y)=pts[npts/2];
	}
      }
      delete [] pts;
      return res;
    }

#ifdef __PY_PLUS_PLUS
    void python_set(const tuple & tup,const T & d){
      if(boost::python::len(tup)>=3)operator()(extract<int>(tup[0]),extract<int>(tup[1]),extract<int>(tup[2]))=d;
      else operator()(extract<int>(tup[0]),extract<int>(tup[1]))=d;
    }
    T & operator[](const tuple & tup){
      if(boost::python::len(tup)>=3)return operator()(extract<int>(tup[0]),extract<int>(tup[1]),extract<int>(tup[2]));
      return operator()(extract<int>(tup[0]),extract<int>(tup[1]));
    }
#endif
};

typedef Image<unsigned char> Image8;
typedef Image<unsigned short> Image16;
typedef Image<unsigned int> Image32;
typedef Image<float> Imagef;

void writePPM(unsigned char * buf,int w,int h,const char * filename,int nchannels=3,int flip=0);

//Just the classes we want to expose in Python
#ifdef __PY_PLUS_PLUS
#ifdef INSTANTIATE_IMAGE
template class Image<unsigned char>;
template class Image<float>;
#endif
#endif

#endif
