#include "image_utils.h"
#include "timer.h"
#include "vec.h"

Matrix  distort(Matrix & pt,Matrix & dist)
{
  Matrix ret(3,1);
  ret[2]=1;
  double r2 = pt[0]*pt[0]+pt[1]*pt[1];
  double r4=r2*r2;
  double r6=r2*r4;
  double dx_x=2*dist[2]*pt[0]*pt[1]+dist[3]*(r2+2*pt[0]*pt[0]);
  double dx_y=dist[2]*(r2+2*pt[1]*pt[1])+2*dist[3]*pt[0]*pt[1];
  double f=(1+dist[0]*r2+dist[1]*r4+dist[4]*r6);
  ret[0]=f*pt[0]+dx_x;
  ret[1]=f*pt[1]+dx_y;
  return ret;
}
Matrix  undistort(Matrix & xd,Matrix & dist)
{
  Matrix J(2,2);
  Matrix xn(2,1);
  int maxits=100;
  xn[0]=xd[0];
  xn[1]=xd[1];
  for(int i=0;i<maxits;i++){
    double r2=xn[0]*xn[0]+xn[1]*xn[1];
    double r4=r2*r2;
    double r6=r4*r2;
    double ddx_dx_x=dist[2]*2*xn[1]+dist[1]*4*r2*xn[0]+dist[4]*6*r4*xn[0];	
    double ddx_dx_y=dist[2]*2*xn[0]+2*dist[3]*xn[1];
    double ddx_dy_x=dist[2]*2*xn[0]+dist[3]*2*xn[1];
    double ddx_dy_y=dist[2]*4*xn[1]+dist[3]*2*xn[0];
    double f=(1+dist[0]*r2+dist[1]*r4+dist[4]*r6);
    double df_dx=dist[0]*2*xn[0]+dist[1]*4*r2*xn[0]+dist[4]*6*r4*xn[0];
    double df_dy=dist[0]*2*xn[1]+dist[1]*4*r2*xn[1]+dist[4]*6*r4*xn[1];
    
    J(0,0)=f+df_dx*xn[0]+ddx_dx_x;
    J(1,0)=df_dx*xn[1]+ddx_dx_y;
    J(0,1)=df_dy*xn[0]+ddx_dy_x;
    J(1,1)=f+df_dy*xn[1]+ddx_dy_y;

    Matrix cur=xd-distort(xn,dist);
    Matrix update=Matrix::linLeastSq(J,cur);
    xn+=update;
  }
  
  return xn;
}
template <class T>
Image<T> removeDistortion(Image<T> & image,Matrix & intr,Matrix & dist)
{
  Image<T> undist(image.w,image.h,image.nchannels);
  Matrix intrInv = intr.inverse();
  Matrix pt(3,1);
  for(int y=0;y<undist.h;y++){
    for(int x=0;x<undist.w;x++){
      pt[0]=x;
      pt[1]=y;
      pt[2]=1;
      pt=intrInv*pt;
      pt=intr*distort(pt,dist);
      int xd=(int)pt[0];
      int yd=(int)pt[1];
      double a=pt[0]-xd;
      double b=pt[1]-yd;

      for(int k=0;k<undist.nchannels;k++){
	if(xd>=0 && xd<image.w &&
	   yd>=0 && yd<image.h){
	  undist(x,y,k)=
	    (T)((1-a)*(1-b)*((double)image(xd,yd,k))+
	    (1-a)*(b)*((double)image(xd,yd+1,k))+
	    (a)*(b)*((double)image(xd+1,yd+1,k))+
	    (a)*(1-b)*((double)image(xd+1,yd,k)));
	}
	else{
	  undist(x,y,0)=0;
	  undist(x,y,1)=0;
	  undist(x,y,2)=255;
	}
      }
    }
  }
  return undist;
}

Imagef gaussFilt(int hs,double sdx,double sdy)
{
  printf("gf %d\n",hs);
  Imagef filt(hs*2+1,hs*2+1,1);
  double sum=0;
  for(int i=0;i<filt.h;i++){
    for(int j=0;j<filt.w;j++){
      filt(j,i)=exp(-((j-hs)*(j-hs)/(2*sdx*sdx)+
		      (i-hs)*(i-hs)/(2*sdy*sdy)));
      //sum=MAX(sum,filt(j,i));
      sum+=filt(j,i);
    }
  }
  for(int i=0;i<filt.w*filt.h;i++)
    filt[i]/=sum;
  return filt;
}

void suppress(Imagef & im,int hs)
{
  for(int i=hs;i<im.h-hs;i++)
    for(int j=hs;j<im.w-hs;j++){
      float cur=im(j,i);
      for(int ii=0;ii<2*hs+1;ii++){
	int f=0;
	for(int jj=0;jj<2*hs+1;jj++){
	  if(im(jj+j-hs,ii+i-hs)>cur){
	    im(j,i)=0;f=1;break;
	  }
	}
	if(f)break;
      }
    }
}

/** Gaussian kernel in x direction.
 */
Imagef gaussx(int hs,double sigma)
{
  Imagef filtx(hs*2+1,1,1);
  double sqrt2pi=sqrt(2*M_PI);
  double twos2=2*sigma*sigma;
  for(int i=0;i<filtx.w;i++){
    double xc=i-hs;
    double val=1.0/sqrt2pi*exp(-(xc*xc)/twos2);
    filtx(i,0)=val;
  }
  return filtx;
}

/** Gaussian kernel in x direction.
 */
Imagef gaussy(int hs,double sigma)
{
  Imagef filty(1,hs*2+1,1);
  double sqrt2pi=sqrt(2*M_PI);
  double twos2=2*sigma*sigma;
  double max;
  for(int i=0;i<filty.h;i++){
    double yc=i-hs;
    double val=1.0/sqrt2pi*exp(-(yc*yc)/twos2);
    filty(0,i)=val;
  }
  return filty;
}

/** gaussian differential? in x direction
 * not sure how accurate this is
 */
Imagef dogx(int hs,double sigma)
{
  Imagef filtxt(hs*2+1,1,1);
  Imagef filtx(hs*2+1,1,1);
  double sqrt2pi=sqrt(2*M_PI);
  double twos2=2*sigma*sigma;
  double max=0;
  for(int i=0;i<filtx.w;i++){
    double xc=i-hs;
    double val=1.0/sqrt2pi*exp(-(xc*xc)/twos2);
    filtxt(i,0)=val;
  }
  for(int i=1;i<filtx.w-1;i++){
    filtx(i,0)=(filtxt(i+1,0)-filtxt(i-1,0))/2;
    max=MAX(max,filtx(i,0));
  }

  filtx(0,0)=filtxt(1,0)-filtxt(0,0);
  filtx(filtx.w-1,0)=filtxt(filtx.w-1,0)-filtxt(filtx.w-2,0);
  max=MAX(max,filtx(0,0));
  max=MAX(max,filtx(filtx.w-1,0));
  filtx*=1.0/max;
  return filtx;
}

/** gaussian differential? in y direction
 * not sure how accurate this is
 */
Imagef dogy(int hs,double sigma)
{
  Imagef filtyt(1,hs*2+1,1);
  Imagef filty(1,hs*2+1,1);
  double sqrt2pi=sqrt(2*M_PI);
  double twos2=2*sigma*sigma;
  double max;
  for(int i=0;i<filty.h;i++){
    double yc=i-hs;
    double val=1.0/sqrt2pi*exp(-(yc*yc)/twos2);
    filtyt(0,i)=val;
  }
  for(int i=1;i<filty.h-1;i++){
    filty(0,i)=(filtyt(0,i+1)-filtyt(0,i-1))/2;
    max=MAX(max,filty(0,i));
  }

  filty(0,0)=filtyt(0,1)-filtyt(0,0);
  filty(0,filty.h-1)=filtyt(0,filty.h-1)-filtyt(0,filty.h-2);
  max=MAX(max,filty(0,0));
  max=MAX(max,filty(0,filty.h-1));
  filty*=1.0/max;
  return filty;
}

/**
 * Harris corner detector, sigma is gaussian blur.
 * The implementation should allow for the multi-resolution
 * rotation (maybe affine invariant) extraction of features.
 * An attempt at the feature descriptor in :
 *   Matching Images with Different Resolutions (Dufournaud et al.)
 * incomplete in that sense, but valid as a harris corner detector.
 * \sa suppress
 */
Imagef harris(Imagef & im,double sigma)
{
  double sc=0.04;
  int hs=(int)sigma*3;
  double sigmabar=sigma;
  Imagef resp(im.w,im.h,1);
  Imagef dxdx(im.w,im.h,im.nchannels);
  Imagef dydy(im.w,im.h,im.nchannels);
  Imagef dxdy(im.w,im.h,im.nchannels);
  Imagef gfiltx=gaussx(hs,sigma);
  Imagef gfilty=gaussy(hs,sigma);
  
  Imagef filtx=dogx(hs,sigmabar);
  Imagef filty=dogy(hs,sigmabar);
  
  dxdx=im.convolve(filtx);
  dydy=im.convolve(filty);
  dxdy=dxdx*dydy;
  dxdx=dxdx*dxdx;
  dydy=dydy*dydy;
  
  dxdx=dxdx.convolve(gfiltx).convolve(gfilty);
  dxdy=dxdy.convolve(gfiltx).convolve(gfilty);
  dydy=dydy.convolve(gfiltx).convolve(gfilty);
  for(int i=0;i<im.h;i++){
    for(int j=0;j<im.w;j++){
      double r=0;
      double maxr=0;
      for(int k=0;k<3;k++){
	double tr=dxdx(j,i,k)+dydy(j,i,k);
	r=dxdx(j,i,k)*dydy(j,i,k)-dxdy(j,i,k)*dxdy(j,i,k)
	    -sc*tr*tr;
	//if(r<0)r=0;
	if(r>maxr)maxr=r;
      }
      resp(j,i,0)=maxr;
      //if(resp(j,i,0)<0)resp(j,i,0)=0;
    }
  }
  double max=0;
  for(int i=0;i<im.w*im.h;i++){
    max=MAX(resp[i],max);
  }
  for(int i=0;i<im.w*im.h;i++){
    resp[i]/=max;
  }
  return resp;
}

Imagef getCorners(Imagef & im,int hs,double thresh)
{
  double sc=0.04;
  Imagef resp(im.w,im.h,1);
  Imagef dxdx(im.w,im.h,im.nchannels);
  Imagef dydy(im.w,im.h,im.nchannels);
  Imagef dxdy(im.w,im.h,im.nchannels);
  Imagef filter=gaussFilt(2,2,2);
  filter.writePPM("filt.ppm");
  for(int i=1;i<im.h-1;i++){
    for(int j=1;j<im.w-1;j++){
      for(int k=0;k<im.nchannels;k++){
	dxdx(j,i,k)=im(j+1,i,k)-im(j-1,i,k);
	dydy(j,i,k)=im(j,i+1,k)-im(j,i-1,k);
	dxdy(j,i,k)=dxdx(j,i,k)*dydy(j,i,k);
	dxdx(j,i,k)*=dxdx(j,i,k);
	dydy(j,i,k)*=dydy(j,i,k);
      }
    }
  }
  dxdx=dxdx.convolve(filter);
  dxdy=dxdy.convolve(filter);
  dydy=dydy.convolve(filter);
  for(int i=0;i<im.h;i++){
    for(int j=0;j<im.w;j++){
      double r=0;
      double maxr=0;
      for(int k=0;k<3;k++){
	double tr=dxdx(j,i,k)+dydy(j,i,k);
	r=dxdx(j,i,k)*dydy(j,i,k)-dxdy(j,i,k)*dxdy(j,i,k)
	    -sc*tr*tr;
	if(r<0)r=0;
	if(r>maxr)maxr=r;
      }
      resp(j,i,0)=maxr;
      //if(resp(j,i,0)<0)resp(j,i,0)=0;
    }
  }
  double max=0;
  for(int i=0;i<im.w*im.h;i++){
    max=MAX(resp[i],max);
  }
  for(int i=0;i<im.w*im.h;i++){
    resp[i]/=max;
  }
  return resp;
}

/*
  Rectification algorithm obtained from Fusiello:
  http://homepages.inf.ed.ac.uk/rbf/CVonline/LOCAL_COPIES/FUSIELLO/node18.html
*/
int rectify(Matrix & Po1,Matrix & Po2,
	    Matrix & T1,Matrix & T2,Matrix & Pn1,Matrix & Pn2){
  Pn1=Matrix(3,4);
  Pn2=Matrix(3,4);
  printf("Getting t's\n");
  Vec3d t1(Po1(0,0),Po1(0,1),Po1(0,2));
  Vec3d t2(Po1(1,0),Po1(1,1),Po1(1,2));
  Vec3d t3(Po1(2,0),Po1(2,1),Po1(2,2));
  double au=(t1.cross(t3)).len();
  double av=(t2.cross(t3)).len();
  int first3[]={0,1,2};
  Matrix c1= (Po1.getCols(first3,3).inverse()*Po1.getColumn(3))*-1;
  Matrix c2= (Po2.getCols(first3,3).inverse()*Po2.getColumn(3))*-1;
  Vec3d fl(Po1(2,0),Po1(2,1),Po1(2,2));
  Vec3d fr(Po2(2,0),Po2(2,1),Po2(2,2));
  Vec3d nn = fl.cross(fr);

  printf("au:%f av:%f\n",au,av);
  c1.printMatlab("c1");
  c2.printMatlab("c2");
  printf("fl =[ %f %f %f]\n",fl.x,fl.y,fl.z);
  printf("fr =[ %f %f %f]\n",fr.x,fr.y,fr.z);
  printf("nn =[ %f %f %f]\n",nn.x,nn.y,nn.z);

  printf("getting first A\n");
  Matrix A=Matrix(3,4);
  Matrix U,S,V;
  A(0,0)=c1[0];A(0,1)=c1[1];A(0,2)=c1[2];A(0,3)=1;
  A(1,0)=c2[0];A(1,1)=c2[1];A(1,2)=c2[2];A(1,3)=1;
  A(2,0)=nn[0];A(2,1)=nn[1];A(2,2)=nn[2];A(2,3)=0;
  A.Lsvd(U,S,V);
  Vec3d lastCol(V(0,3),V(1,3),V(2,3));
  double r = 1.0/(lastCol.len());
  Matrix a3 = V.getColumn(3)*r;
  a3.printMatlab("a3");
  printf("r %f\n",r);
  printf("getting second A\n");
  A(2,0)=a3[0];A(2,1)=a3[1];A(2,2)=a3[2];A(2,3)=0;
  A.Lsvd(U,S,V);
 
  lastCol=Vec3d(V(0,3),V(1,3),V(2,3));
  r = (av)/(lastCol.len());
  Matrix a2 = V.getColumn(3)*r;

  V.printMatlab("V");
  a2.printMatlab("a2");
  printf("r: %f\n",r);
  printf("getting third A\n");
  A(1,0)=a2[0];A(1,1)=a2[1];A(1,2)=a2[2];A(1,3)=0;
  A.Lsvd(U,S,V);
  A.printMatlab("A");
  lastCol=Vec3d(V(0,3),V(1,3),V(2,3));
  r = (au)/(lastCol.len());
  Matrix a1 = V.getColumn(3)*r;
  
  V.printMatlab("V");
  a1.printMatlab("a1");
  printf("r: %f\n",r);
  
  printf("getting fourth A\n");
  A(0,0)=c2[0];A(0,1)=c2[1];A(0,2)=c2[2];A(0,3)=1;
  A.Lsvd(U,S,V);
  lastCol=Vec3d(V(0,3),V(1,3),V(2,3));
  r = (au)/(lastCol.len());
  Matrix b1 = V.getColumn(3)*r;

  Pn1(0,0)=a1[0];Pn1(0,1)=a1[1];Pn1(0,2)=a1[2];Pn1(0,3)=a1[3];
  Pn1(1,0)=a2[0];Pn1(1,1)=a2[1];Pn1(1,2)=a2[2];Pn1(1,3)=a2[3];
  Pn1(2,0)=a3[0];Pn1(2,1)=a3[1];Pn1(2,2)=a3[2];Pn1(2,3)=a3[3];

  Pn2(0,0)=b1[0];Pn2(0,1)=b1[1];Pn2(0,2)=b1[2];Pn2(0,3)=b1[3];
  Pn2(1,0)=a2[0];Pn2(1,1)=a2[1];Pn2(1,2)=a2[2];Pn2(1,3)=a2[3];
  Pn2(2,0)=a3[0];Pn2(2,1)=a3[1];Pn2(2,2)=a3[2];Pn2(2,3)=a3[3];
  
  Matrix H = Matrix::eye(3);

  Pn1 = H * Pn1;
  Pn2 = H * Pn2;
  
  T1 = Pn1.getCols(first3,3)* (Po1.getCols(first3,3).inverse());
  T2 = Pn2.getCols(first3,3)* (Po2.getCols(first3,3).inverse());
}

/*
  From:
  A compact algorithm for rectification of stereo pairs.
  Fusiello et al. 2000.
  function art in the matlab code
*/
int rectify(Matrix & intr1,Matrix & extr1,Matrix & intr2,Matrix & extr2,
	    Matrix & T1,Matrix & T2,Matrix & Pn1,Matrix & Pn2){
  int first3[]={0,1,2};
  Matrix rot1(3,3);
  Matrix rot2(3,3);
  Matrix t1(3,1),t2(3,1);
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      rot1(i,j)=extr1(i,j);
      rot2(i,j)=extr2(i,j);
    }
    t1[i]=extr1(i,3);
    t2[i]=extr2(i,3);
  }
  Matrix proj(3,4);
  proj.setAll(0);proj(0,0)=1;proj(1,1)=1;proj(2,2)=1;
  Matrix Po1=intr1*proj*extr1;
  Matrix Po2=intr2*proj*extr2;
  //Matrix c1= (Po1.getCols(first3,3).inverse()*Po1.getColumn(3))*-1;
  //Matrix c2= (Po2.getCols(first3,3).inverse()*Po2.getColumn(3))*-1;

  Matrix c1=(rot1.transpose())*t1*-1;
  Matrix c2=(rot2.transpose())*t2*-1;

  Vec3d oldx(rot1(0,0),rot1(0,1),rot1(0,2));
  Vec3d oldy(rot1(1,0),rot1(1,1),rot1(1,2));
  Vec3d oldz(rot1(2,0),rot1(2,1),rot1(2,2));
  
  Vec3d v1(c1[0]-c2[0],c1[1]-c2[1],c1[2]-c2[2]);
  //Make sure it points in the same direction..
  if(oldx.dot(v1)<0)v1*=-1;
  Vec3d v2=oldz.cross(v1);
  //Make sure y points in same direction
  if(oldy.dot(v2)<0)v2*=-1;
  Vec3d v3=v1.cross(v2);
  Matrix R(3,3);
  v1.normalize();
  v2.normalize();
  v3.normalize();
  R(0,0)=v1.x;R(0,1)=v1.y;R(0,2)=v1.z;
  R(1,0)=v2.x;R(1,1)=v2.y;R(1,2)=v2.z;
  R(2,0)=v3.x;R(2,1)=v3.y;R(2,2)=v3.z;
  Matrix newintr=(intr1+intr2);
  newintr*=0.5;
  newintr(0,1)=0;//remove skew

  Matrix t1n=(R*c1)*-1;
  Matrix t2n=(R*c2)*-1;

  Matrix extr1n=Matrix(3,4);
  Matrix extr2n=Matrix(3,4);
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      extr1n(i,j)=R(i,j);
      extr2n(i,j)=R(i,j);
    }
  }
  for(int i=0;i<3;i++){
    extr1n(i,3)=t1n[i];
    extr2n(i,3)=t2n[i];
  }
  //instead could just change extrinsics....
  Pn1=newintr*extr1n;
  Pn2=newintr*extr2n;
  
  T1 = Pn1.getCols(first3,3)* (Po1.getCols(first3,3).inverse());
  T2 = Pn2.getCols(first3,3)* (Po2.getCols(first3,3).inverse());
}

namespace im_tri_opt{
  Matrix xhat(3,1);
  Matrix yhat(3,1);
  void quadratic_func_r(const int & m,const int & n,
		      double * x,double * fvec,int & iflag)
  {
    Matrix Sr=Matrix::eye(3,3);
    Sr(0,0)=x[0];
    Sr(1,0)=1.0-x[0];
    Matrix res=(Sr*xhat).transpose()*(Sr*yhat);
    fvec[0]=res[0]*res[0];
  }
  void quadratic_func_t(const int & m,const int & n,
			double * x,double * fvec,int & iflag)
  {
    Matrix St=Matrix::eye(3,3);
    St(1,1)=   x[0];
    St(0,1)=1.0-x[0];
    Matrix res=(St*xhat).transpose()*(St*yhat);
    fvec[0]=res[0]*res[0];
  }
}
/*An implementation of A linear Trinocular Rectification Method for Accurate Stereoscopic matching (zhang,Cech, Sara, Wu, Hu)
 
Fbr is the fundamental matrix transfering points st:
x'^t Fbr x = 0 where x is in image b, and x' is image r.
The others are similar.

The Hb,Hr and Ht are the rectifying homographies.

The canonical F matrices should be
Fbr ~= [0 0 1;0 0 0; -1 0 0];
Fbt ~= [0 0 0;0 0 -1; 0 1 0];
Frt ~= [0 0 1;0 0 1;-1 -1 0];
We want to obtain the homographies for each image Hb, Hr and Ht
The epipoles are used to find the bottom row of a valid Homography (Hb_star,...).
Then the alpha1 and alpha2 that minimize the skew are found.  A uniform scale factor alpha3 attempts to keep the area in the bottom image constant through the homography.  Finally the translations (s1,s2 and s3) as appropriate bounding boxes.

Hb=[1,0,s1;0,1,s2;0,0,1]*diag([alpha3,alpha3,1])*diag([alpha1,alpha2,1])*Hb_star;
Hr=[1,0,s1;0,1,s2+s3;0,0,1]*diag([alpha3,alpha3,1])*Sr*Hr_star;
Ht=[1,0,s1+s3;0,1,s2;0,0,1]*diag([alpha3,alpha3,1])*St*Ht_star;
Where Sr=[alpha1,0,0;1-alpha1,1,Fbr(3,3)*(alpha1-1);0,0,1];
 and  St=[1,1-alpha2,Fbr(3,3)*(alpha1-1);0,alpha2,0;0,0,1];

The above formulas are actually so that the epipolar lines in the b and r images are vertical... the authors must have been using (i,j) indices or something.  Anyhow, the swap at the end of this ensures that it is x and y.
 */
int tri_rectify(int w,int h,
		Matrix & Fbr,
		Matrix & Fbt,
		Matrix & Frt,
		Matrix & Hb,
		Matrix & Hr,
		Matrix & Ht)
{
  Matrix U,S,V;
  Fbr.Lsvd(U,S,V);
  //Fbr is the transfer from image-b to image-r
  //so the last column of U is erb
  Vec3f erb(U(0,2),U(1,2),U(2,2));
  Vec3f ebr(V(0,2),V(1,2),V(2,2));
  
  //Get the epoles, erb is the epipole in the right image
  //of camera b.
  Matrix x(1,3);
  x[0]=erb.x;
  x[1]=erb.y;
  x[2]=erb.z;
  x=x*Fbr;
  x.printMatlab("x");
  x=Matrix(3,1);
  x[0]=ebr.x;
  x[1]=ebr.y;
  x[2]=ebr.z;
  x=Fbr*x;
  x.printMatlab("x");

  Fbt.Lsvd(U,S,V);
  Vec3f etb(U(0,2),U(1,2),U(2,2));
  Vec3f ebt(V(0,2),V(1,2),V(2,2));

  Frt.Lsvd(U,S,V);
  Vec3f etr(U(0,2),U(1,2),U(2,2));
  Vec3f ert(V(0,2),V(1,2),V(2,2));
  
  Vec3f wb=ebr.cross(ebt);
  Vec3f wr=erb.cross(ert);
  Vec3f wt=etb.cross(etr);
  wb/=wb.z;
  wr/=wr.z;
  wt/=wt.z;
  printf("ebr.dot(wb)=%f\n",ebr.dot(wb));
  Matrix Hb_star(3,3);
  Matrix Hr_star(3,3);
  Matrix Ht_star(3,3);
  
  //Get valid, but not distortion prserving homos
  Hb_star(0,0)=Fbr(2,2)*wb.x-Fbr(2,0);
  Hb_star(1,0)=Fbt(2,0);
  Hb_star(2,0)=wb.x;
  Hb_star(0,1)=Fbr(2,2)*wb.y-Fbr(2,1);
  Hb_star(1,1)=Fbt(2,1);
  Hb_star(2,1)=wb.y;
  Hb_star(0,2)=0;
  Hb_star(1,2)=Fbt(2,2);
  Hb_star(2,2)=1;
  
  Hr_star(0,0)=Fbr(0,2);
  Hr_star(1,0)=(Frt(2,2)+Fbr(2,2))*wr.x-Frt(2,0)-Fbr(0,2);
  Hr_star(2,0)=wr.x;
  Hr_star(0,1)=Fbr(1,2);
  Hr_star(1,1)=(Frt(2,2)+Fbr(2,2))*wr.y-Frt(2,1)-Fbr(1,2);
  Hr_star(2,1)=wr.y;
  Hr_star(0,2)=Fbr(2,2);
  Hr_star(1,2)=0;
  Hr_star(2,2)=1;

  Ht_star(0,0)=(Fbr(2,2)-Fbt(2,2))*wt.x+Frt(0,2)+Fbt(0,2);
  Ht_star(1,0)=Fbt(2,2)*wt.x-Fbt(0,2);
  Ht_star(2,0)=wt.x;
  Ht_star(0,1)=(Fbr(2,2)-Fbt(2,2))*wt.y+Frt(1,2)+Fbt(1,2);
  Ht_star(1,1)=Fbt(2,2)*wt.y-Fbt(1,2);
  Ht_star(2,1)=wt.y;
  Ht_star(0,2)=Fbr(2,2)+Frt(2,2);
  Ht_star(1,2)=0;
  Ht_star(2,2)=1;

#ifdef TRI_RECT_CHECK
  //Make sure the condition (3) of the paper holds
  Matrix FbrI(3,3);
  FbrI.setAll(0);
  FbrI(2,0)=-1;
  FbrI(0,2)= 1;

  Matrix FbtI(3,3);
  FbtI.setAll(0);
  FbtI(1,2)=-1;
  FbtI(2,1)= 1;

  Matrix FrtI(3,3);
  FrtI.setAll(0);
  FrtI(2,0)=FrtI(2,1)=-1;
  FrtI(0,2)=FrtI(1,2)=1;
  
  Matrix F=Hr_star.transpose()*FbrI*Hb_star;
  F.printMatlab("FbrI");
  Fbr.printMatlab("Fbr");
  F=Ht_star.transpose()*FbtI*Hb_star;
  F.printMatlab("FbtI");
  Fbt.printMatlab("Fbt");
  F=Ht_star.transpose()*FrtI*Hr_star;
  F.printMatlab("FrtI");
  Frt.printMatlab("Frt");
  Matrix nebr(3,1);
  nebr[0]=ebr.x;
  nebr[1]=ebr.y;
  nebr[2]=ebr.z;
#endif
  //Use the two lines through the middle of the
  //image to avoid skew.
  Matrix a(3,1);
  Matrix b(3,1);
  Matrix c(3,1);
  Matrix d(3,1);
  Matrix ar,br,cr,dr,at,bt,ct,dt;
  a[0]=0;
  a[1]=h/2;
  a[2]=1;
  b[0]=w;
  b[1]=h/2;
  b[2]=1;
  c[0]=w/2;
  c[1]=0;
  c[2]=1;
  d[0]=w/2;
  d[1]=h;
  d[2]=1;
  ar=Hr_star*a;
  br=Hr_star*b;
  cr=Hr_star*c;
  dr=Hr_star*d;
  at=Ht_star*a;
  bt=Ht_star*b;
  ct=Ht_star*c;
  dt=Ht_star*d;

  double alpha1=0;
  double alpha2=0;
  double alpha3=0;

  Matrix xr=ar-br;
  Matrix yr=cr-dr;
  Matrix xt=at-bt;
  Matrix yt=ct-dt;
  //Solve the quadratic equations
  {  
    double x1=xr[0];
    double x2=xr[1];
    double x3=xr[2];
    double y1=yr[0];
    double y2=yr[1];
    double y3=yr[2];
    double a=2*x1*y1;
    double b=-(2*x1*y1+x1*y2+x2*y1);
    double c=x1*y1+x2*y1+x1*y2+x2*y2+x3*y3;
    //double a=2*x1*y1;
    //double b=-2*x1*y1+x2*y1+y2*x1;
    //double c=y1*x1-y1*x2-x1*y2+x2*y2+x3*y3;
    if(fabs(a)<10e-10){
      alpha1=-c/b;
    }
    else{
      double root1=(-b-sqrt(b*b-4*a*c))/(2*a);
      double root2=(-b+sqrt(b*b-4*a*c))/(2*a);
      if(b*b-4*a*c<0){
	printf("determinant less than 0 (a=%lf,b=%lf,c=%lf,det=%lf)...bailing out (%s,%d)\n",
	       a,b,c,
	       b*b-4.0*a*c,
	       __FILE__,__LINE__);
	//exit(1);
	alpha1=1.0;
      }
      else if(fabs(root1)<fabs(root2))
	alpha1=root1;
      else alpha1=root2;
    }
  }
  {  
    double x1=xt[0];
    double x2=xt[1];
    double x3=xt[2];
    double y1=yt[0];
    double y2=yt[1];
    double y3=yt[2];
    double a=2*x2*y2;
    double b=-(y1*x2+y2*x2+y2*x1+y2*x2);
    double c=y3*x3+y1*x1+y1*x2+y2*x1+y2*x2;
    if(fabs(a)<10e-10){
      alpha2=-c/b;
    }
    else{
      double root1=(-b-sqrt(b*b-4*a*c))/(2*a);
      double root2=(-b+sqrt(b*b-4*a*c))/(2*a);
      if(b*b-4*a*c<0){
	printf("determinant less than 0...bailing out (%s,%d)\n",
	       __FILE__,__LINE__);
	//exit(1);
	alpha2=1;
      }
      else if(fabs(root1)<fabs(root2))
	alpha2=root1;
      else alpha2=root2;
    }
  }
#ifdef TRI_RECT_CHECK
  printf("got alphas\n");
  Matrix a1(1,1),a2(1,1),e1(1,1),e2(1,1);
  a1[0]=alpha1;
  a2[0]=alpha2;
  im_tri_opt::xhat=xr;
  im_tri_opt::yhat=yr;
  Matrix::lsqnonlin(im_tri_opt::quadratic_func_r,a1,e1);
  im_tri_opt::xhat=xt;
  im_tri_opt::yhat=yt;
  Matrix::lsqnonlin(im_tri_opt::quadratic_func_t,a2,e2);
  printf("%f %f\n",alpha1,alpha2);
  printf("%f %f  (error: %f %f)\n",a1[0],a2[0],e1[0],e2[0]);
#endif
  Matrix Sb=Matrix::eye(3,3);
  Matrix Sr=Matrix::eye(3,3);
  Matrix St=Matrix::eye(3,3);
  Matrix Suniform=Matrix::eye(3,3);

  Sb(0,0)=alpha1;
  Sb(1,1)=alpha2;
  
  Sr(0,0)=alpha1;
  Sr(1,0)=1.0-alpha1;
  Sr(1,2)=Fbr(2,2)*(alpha1-1);
  
  St(0,1)=1.0-alpha2;
  St(1,1)=alpha2;
  St(0,2)=Fbr(2,2)*(alpha1-1);

  //Choose alpha3 so that areas are equivalent.
  Matrix corners(3,4);
  Matrix corners2;
  corners.setAll(1);
  corners(0,0)=0;
  corners(1,0)=0;
  corners(0,1)=w;
  corners(1,1)=0;
  corners(0,2)=w;
  corners(1,2)=h;
  corners(0,3)=0;
  corners(1,3)=h;
  
  corners2=Sb*Hb_star*corners;
  Vec2f corn[4];
  for(int i=0;i<4;i++){
    corners2(0,i)/=corners2(2,i);
    corners2(1,i)/=corners2(2,i);
    corners2(2,i)=1;
    corn[i]=Vec2f(corners2(0,i),corners2(1,i));
  }
  corners2.printMatlab("corners2");
  double ab=(corn[1]-corn[0]).len();
  double bc=(corn[2]-corn[1]).len();
  double ca=(corn[2]-corn[0]).len();
  double cd=(corn[3]-corn[2]).len();
  double da=(corn[0]-corn[3]).len();
  double s_abc=(ab+bc+ca)/2;
  double s_cda=(cd+da+ca)/2;
  
  //herons formula * 2
  double area=sqrt(s_abc*(s_abc-ab)*(s_abc-bc)*(s_abc-cd))+
    sqrt(s_cda*(s_cda-cd)*(s_cda-da)*(s_cda-ca));
  alpha3=sqrt((w*h)/area);
  printf("area=%f, alpha3=%f\n",area,alpha3);
  Suniform(0,0)=alpha3;
  Suniform(1,1)=alpha3;

  //Find s1,s2,s3
  Hb=Suniform*Sb*Hb_star;
  Hr=Suniform*Sr*Hr_star;
  Ht=Suniform*St*Ht_star;

  //Make sure that the images didn't flip;
  /*Matrix fb=Matrix::eye(3,3);
  Matrix fr=Matrix::eye(3,3);
  Matrix ft=Matrix::eye(3,3);
  //fb(0,0)=(((Hb(0,0)*Hb(2,2))<0)?-1:1);
  //fb(1,1)=(((Hb(1,1)*Hb(2,2))<0)?-1:1);
  //fr(0,0)=(((Hr(0,0)*Hr(2,2))<0)?-1:1);
  //fr(0,0)=(((Hr(1,1)*Hr(2,2))<0)?-1:1);
  //ft(0,0)=(((Ht(0,0)*Ht(2,2))<0)?-1:1);
  //ft(1,1)=(((Ht(1,1)*Ht(2,2))<0)?-1:1);
  
  Hb=fb*Hb;
  Hr=fb*Hr;
  Ht=fb*Ht;
  */
  double s1=10e10;
  double s2=10e10;
  double s3=0;
  corners2=Hb*corners;
  for(int i=0;i<4;i++){
    corners2(0,i)/=corners2(2,i);
    corners2(1,i)/=corners2(2,i);
    s1=MIN(s1,corners2(0,i));
    s2=MIN(s2,corners2(1,i));
  }
  s1*=-1;//subtract minx
  s2*=-1;//subtract miny

  corners2=Hr*corners;
  for(int i=0;i<4;i++){
    corners2(1,i)/=corners2(2,i);
    s3=MIN(s3,corners2(1,i));
  }
  s3+=s2;
  s3=-s3;
  
  Matrix Tb=Matrix::eye(3,3);
  Matrix Tr=Matrix::eye(3,3);
  Matrix Tt=Matrix::eye(3,3);
  Tb(0,2)=s1;
  Tb(1,2)=s2;
  Tr(0,2)=s1;
  Tr(1,2)=s2+s3;
  Tt(0,2)=s1+s3;
  Tt(1,2)=s2;

  Hb=Tb*Hb;
  Hr=Tr*Hr;
  Ht=Tt*Ht;

#ifdef TRI_RECT_CHECK
  F=Ht.transpose()*FrtI*Hr;
  F*=(1.0/F(2,2));
  (Frt*(1.0/Frt(2,2))).printMatlab("FrtI");
  F.printMatlab("F");
#endif
  Matrix swap(3,3);
  swap.setAll(0);
  swap(0,1)=1;
  swap(1,0)=1;
  swap(2,2)=1;
  Hb=swap*Hb;
  Hr=swap*Hr;
  Ht=swap*Ht;
  
#ifdef TRI_RECT_CHECK
  corners2=Hb*corners;
  for(int i=0;i<4;i++){
    corners2(0,i)/=corners2(2,i);
    corners2(1,i)/=corners2(2,i);
    corners2(2,i)=1;
  }
  corners2.printMatlab("Hb*c");

  corners2=Hr*corners;
  for(int i=0;i<4;i++){
    corners2(0,i)/=corners2(2,i);
    corners2(1,i)/=corners2(2,i);
    corners2(2,i)=1;
  }
  corners2.printMatlab("Hr*c");

  corners2=Ht*corners;
  for(int i=0;i<4;i++){
    corners2(0,i)/=corners2(2,i);
    corners2(1,i)/=corners2(2,i);
    corners2(2,i)=1;
  }
  corners2.printMatlab("Ht*c");
#endif
  //F.printMatlab("FrtI");
  //Frt.printMatlab("Frt");
}

Matrix eightPoint(Matrix & _pts1,Matrix & _pts2)
{
  Matrix pts1=_pts1.copy();
  Matrix pts2=_pts2.copy();

  if(pts1.n>=8 && (pts1.m==2||pts1.m==3)){
    pts1=pts1.transpose();
    pts2=pts2.transpose();
  }
  else if(pts1.m>=8 && (pts1.n==2||pts1.n==3)){
    //this is what we want
  }
  else{
    fprintf(stderr,"Unexpected matrix for eightPoint, pts1 is %dx%d\n",pts1.m,pts1.n);
    exit(1);
  }
  
  Matrix m1(2,1),m2(2,1);
  Matrix s1(2,1),s2(2,1);
  m1.setAll(0);
  m2.setAll(0);
  s1.setAll(0);
  s2.setAll(0);
  for(int i=0;i<pts1.m;i++){
    m1[0]+=pts1(i,0);
    m1[1]+=pts1(i,1);
    m2[0]+=pts2(i,0);
    m2[1]+=pts2(i,1);
  }
  m1*=(1.0/(double)pts1.m);
  m2*=(1.0/(double)pts1.m);

  for(int i=0;i<pts1.m;i++){
    pts1(i,0)-=m1[0];
    pts1(i,1)-=m1[1];
    pts2(i,0)-=m2[0];
    pts2(i,1)-=m2[1];
    s1[0]+=(pts1(i,0)*pts1(i,0));
    s1[1]+=(pts1(i,1)*pts1(i,1));
    s2[0]+=(pts2(i,0)*pts2(i,0));
    s2[1]+=(pts2(i,1)*pts2(i,1));
  }
  //Uniform scaling
  s1[0]=1.0/sqrt((s1[0]+s1[1])/(2*pts1.m));
  s1[1]=s1[0];
  s2[0]=1.0/sqrt((s2[0]+s2[1])/(2*pts1.m));
  s2[1]=s2[0];
  /*
  s1*=(1.0/(double)pts1.m);
  s2*=(1.0/(double)pts1.m);
  s1[0]=1.0/sqrt(s1[0]);
  s1[1]=1.0/sqrt(s1[1]);
  s2[0]=1.0/sqrt(s2[0]);
  s2[1]=1.0/sqrt(s2[1]);
  */

  Matrix T1=Matrix::eye(3,3);
  Matrix S1=Matrix::eye(3,3);
  Matrix T2=Matrix::eye(3,3);
  Matrix S2=Matrix::eye(3,3);
  T1(0,2)=-m1[0];
  T1(1,2)=-m1[1];
  S1(0,0)=s1[0];
  S1(1,1)=s1[1];
  
  T2(0,2)=-m2[0];
  T2(1,2)=-m2[1];
  S2(0,0)=s2[0];
  S2(1,1)=s2[1];
  
  for(int i=0;i<pts1.m;i++){
    pts1(i,0)*=s1[0];
    pts1(i,1)*=s1[1];
    pts2(i,0)*=s2[0];
    pts2(i,1)*=s2[1];
  }
  /*double avgdist1=0;
  double avgdist2=0;
  for(int i=0;i<pts1.m;i++){
    avgdist1+=sqrt(pts1(i,0)*pts1(i,0)+(pts1(i,1)*pts1(i,1)));
    avgdist2+=sqrt(pts2(i,0)*pts2(i,0)+(pts2(i,1)*pts2(i,1)));
  }
  printf("avgdist1 %f, avgdist2 %f\n",avgdist1/pts1.m,avgdist2/pts2.m);*/

  
  //(pts1-(S1*T1*_pts1).transpose()).printMatlab("p1diff");
  //(pts2-(S2*T2*_pts2).transpose()).printMatlab("p2diff");
  //printf("doing eight point\n");
  Matrix A(pts1.m,9);
  for(int i=0;i<pts1.m;i++){
    A(i,0)=pts1(i,0)*pts2(i,0);
    A(i,1)=pts1(i,1)*pts2(i,0);
    A(i,2)=          pts2(i,0);
    A(i,3)=pts1(i,0)*pts2(i,1);
    A(i,4)=pts1(i,1)*pts2(i,1);
    A(i,5)=          pts2(i,1);
    A(i,6)=pts1(i,0);
    A(i,7)=pts1(i,1);
    A(i,8)=1;
  }

  Matrix U,S,V;
  A.Lsvd(U,S,V);
  Matrix F=V.getColumn(8);
  F.m=3;
  F.n=3;
  F.Lsvd(U,S,V);
  Matrix sing=Matrix::eye(3,3);
  //printf("%f cond:%f\n",S[2],S[0]/S[2]);
  S[2]=0;
  sing(0,0)=S[0];
  sing(1,1)=S[1];
  sing(2,2)=S[2];
  F=U*sing*V.transpose();//rank 2 approximation
  
  F=(S2*T2).transpose()*F*S1*T1;
  //printf("F = %dx%d\n",_pts1.m,pts1.n);
  //Matrix res=_pts2.transpose()*F*_pts1;
  //res.printMatlab("res8pt");
  //  F=F.transpose();
  //F.printMatlab("F");
  return F;
}

//There must be a better way to do this...
Matrix fundamentalMatrix(Matrix & P1,Matrix & P2)
{
  
  //From hartley and zisserman
  int first3[]={0,1,2};
  Matrix P1_3x3=P1.getCols(first3,3);
  Matrix P2_3x3=P2.getCols(first3,3);
  Matrix neg_P1_lastcol=P1.getColumn(3)*-1;
  Matrix cc1=Matrix::LlinLeastSq(P1_3x3,neg_P1_lastcol);
  Matrix cc1_4(4,1);
  cc1_4[0]=cc1[0];
  cc1_4[1]=cc1[1];
  cc1_4[2]=cc1[2];
  cc1_4[3]=1;
  Matrix e2=P2*cc1_4;
  Matrix e2ss(3,3);
  e2ss(0,0)=0;e2ss(0,1)=-e2[2];e2ss(0,2)= e2[1];
  e2ss(1,0)= e2[2];e2ss(1,1)=0;e2ss(1,2)=-e2[0];
  e2ss(2,0)=-e2[1];e2ss(2,1)= e2[0];e2ss(2,2)=0;

  Matrix F=e2ss*P2_3x3*(P1_3x3.inverse());
  return F;
  
  /* 
     Old stupid way
  Matrix pts=Matrix::random(4,8);
  for(int j=0;j<8;j++)pts(3,j)=1;
  Matrix pts1=P1*pts;
  Matrix pts2=P2*pts;

  for(int j=0;j<8;j++){
    pts1(0,j)/=pts1(2,j);
    pts1(1,j)/=pts1(2,j);
    
    pts2(0,j)/=pts2(2,j);
    pts2(1,j)/=pts2(2,j);
  }
  pts1=pts1.transpose();
  pts2=pts2.transpose();
  //Matrix F=eightPoint(pts1,pts2);
  //return F;
  for(int i=0;i<8;i++){
    Matrix p1(3,1);
    p1.setAll(1);
    Matrix p2(3,1);
    p2.setAll(1);
    p1[0]=pts1(i,0);
    p1[1]=pts1(i,1);
    p2[0]=pts2(i,0);
    p2[1]=pts2(i,1);
    Matrix err=(p2.transpose())*F*p1;
    printf("%f\n",err.dot(err));
  }
  */
  return F;
}


/*Image8 colorSeg(Image8 & image,Image8 & color)
{
    Imagef imf;
    imf=image;
    Imagef cof;
    cof=color;
    return colorSeg(imf,cof);
}*/
template <class T>
Image8 colorSeg(Image<T> & inputim,Image<T> & inputcolor)
{
  Timer t;
  t.start();

  //Used to use floating point images
  //Imagef image;
  //Imagef color;
  //image.operator=(inputim);
  //color.operator=(inputcolor);

  //Matrix imflat(3,image.getWidth()*image.getHeight());
  Matrix sample(3,inputcolor.getWidth()*inputcolor.getHeight());
  for(int y=0;y<inputcolor.getHeight();y++){
    for(int x=0;x<inputcolor.getWidth();x++){
      sample(0,y*inputcolor.getWidth()+x)=(double)inputcolor(x,y,0);
      sample(1,y*inputcolor.getWidth()+x)=(double)inputcolor(x,y,1);
      sample(2,y*inputcolor.getWidth()+x)=(double)inputcolor(x,y,2);
    }
  }
  
  /*
  for(int y=0;y<image.getHeight();y++){
    for(int x=0;x<image.getWidth();x++){
      imflat(0,y*image.getWidth()+x)=image(x,y,0);
      imflat(1,y*image.getWidth()+x)=image(x,y,1);
      imflat(2,y*image.getWidth()+x)=image(x,y,2);
    }
  }
  */
  Matrix mean(3,1);
  mean.setAll(0);
  for(int j=0;j<sample.n;j++){
    mean[0]+=sample(0,j);
    mean[1]+=sample(1,j);
    mean[2]+=sample(2,j);
  }
  mean[0]/=sample.n;
  mean[1]/=sample.n;
  mean[2]/=sample.n;
  
  for(int j=0;j<sample.n;j++){
    sample(0,j)-=mean[0];
    sample(1,j)-=mean[1];
    sample(2,j)-=mean[2];
  }
  /*for(int j=0;j<imflat.n;j++){
    imflat(0,j)-=mean[0];
    imflat(1,j)-=mean[1];
    imflat(2,j)-=mean[2];
    }*/
  printf("getting the mean and subtracting took %f\n",t.stop());
  t.start();
  Matrix covar=sample*(sample.transpose());
  Matrix evec,eval,d;
  covar.Lsvd(evec,eval,d);
  printf("covar and svd took %f\n",t.stop());

  //Matrix u2,s2,v2;
  //sample.transpose().svd(u2,s2,v2);
  /* My original
  Matrix imtrans=evec.transpose()*imflat;
  Matrix sampletrans=evec.transpose()*sample;
  Matrix stdv(3,1);
  Matrix mean2(3,1);
  stdv.setAll(0);
  for(int j=0;j<sampletrans.n;j++){
    double dx=sampletrans(0,j);
    double dy=sampletrans(1,j);
    double dz=sampletrans(2,j);
    stdv[0]+=(dx*dx);
    stdv[1]+=(dy*dy);
    stdv[2]+=(dz*dz);
  }
  stdv*=(1.0/sampletrans.n);
  double sigma=sqrt(stdv[0]+stdv[1]+stdv[2]);
  Image8 mask(image.getWidth(),image.getHeight(),1);

  for(int y=0;y<image.getHeight();y++){
    for(int x=0;x<image.getWidth();x++){
      double dx=
	imtrans(0,y*image.getWidth()+x);
      double dy=
	imtrans(1,y*image.getWidth()+x);
      double dz=
	imtrans(2,y*image.getWidth()+x);
      double len=sqrt(dx*dx+dy*dy+dz*dz);
      if(len>3.0*sigma){
	  mask(x,y,0)=255;
      }
      else{
	  mask(x,y,0)=0;
      }
    }
    }*/
  //Keiths 
  
  //v2.printMatlab("v2");
  //evec.printMatlab("evec");
  if(sample.n>1)
    eval*=1.0/(sample.n-1); //make eigenvalues the variance
  if(eval.m*eval.n!=3){
    printf("svd doesn't return that...\n");
    exit(0);
  }

  Matrix scale(3,3);
  Matrix iscale(3,3);
  iscale.setAll(0);
  scale.setAll(0);

  for(int j=0;j<3;j++){
    eval[j]=sqrt(eval[j]);
    if(fabs(eval[j])<10e-6)eval[j]=10e-6;
    scale(j,j)=eval[j];
    iscale(j,j)=1.0/eval[j];
  }
  //The new eigenvectors are:
  //evec*scale so, to project into space we need
  //inv(evec*scale)=iscale*evec' .. which is the same as
  //just scaling the columns of evec by 1/eval and transposing
  Matrix sampletrans=iscale*evec.transpose()*sample;
  //Matrix imtrans=iscale*evec.transpose()*imflat;
  Image8 mask(inputim.getWidth(),inputim.getHeight(),1);
  /*
    //Old check to see that the variance was correct
  Matrix var3(3,1);
  var3.setAll(0);
  double var=0;
  double maxdist=0;
  for(int i=0;i<sampletrans.n;i++){
    double len=(sampletrans(0,i)*sampletrans(0,i)+
		sampletrans(1,i)*sampletrans(1,i)+
		sampletrans(2,i)*sampletrans(2,i));
    var+=len;
    var3[0]+=sampletrans(0,i)*sampletrans(0,i);
    var3[1]+=sampletrans(1,i)*sampletrans(1,i);
    var3[2]+=sampletrans(2,i)*sampletrans(2,i);
    if(sqrt(len)>maxdist)maxdist=sqrt(len);
  }
  var3*=(1.0/(sampletrans.n-1));
  //var3.printMatlab("var3");
  if(sampletrans.n!=1)
    var*=(1.0/(sampletrans.n-1));
  double stdev=sqrt(var);
  printf("maxdist %f, 3*stdev: %f 2*stdev:%f \n",maxdist,3*stdev,2*stdev);
  printf("hmmm, stdev should be %f, and is %f\n",sqrt(3),stdev);
  //Imagef tformed(image.w,image.h,1);
  */
  float stdev=sqrt(3);
  float thresh=(4.0f*stdev)*(4.0f*stdev);
  
  t.start();
  //transposes of the evec, so the columns
  Vec3f r1(evec(0,0),evec(1,0),evec(2,0));
  Vec3f r2(evec(0,1),evec(1,1),evec(2,1));
  Vec3f r3(evec(0,2),evec(1,2),evec(2,2));
  Vec3f isc(iscale(0,0),iscale(1,1),iscale(2,2));
  r1*=isc.x;
  r2*=isc.y;
  r3*=isc.z;
  for(int y=0;y<inputim.getHeight();y++){
    for(int x=0;x<inputim.getWidth();x++){
      /*      double dx=
	imtrans(0,y*image.getWidth()+x);
      double dy=
	imtrans(1,y*image.getWidth()+x);
      double dz=
      imtrans(2,y*image.getWidth()+x);*/
      Vec3f col(((float)inputim(x,y,0))-mean[0],
		((float)inputim(x,y,1))-mean[1],
		((float)inputim(x,y,2))-mean[2]);
      float dx=r1.dot(col);
      float dy=r2.dot(col);
      float dz=r3.dot(col);
            
      float len_sq=(dx*dx+dy*dy+dz*dz);
      if(len_sq>thresh){
	mask(x,y,0)=255;
      }
      else{
	mask(x,y,0)=0;
      }
      //tformed(x,y,0)=len;
    }
  }
  printf("transforming the image took %f\n",t.stop());
  /*
  double min,max;
  tformed.getRange(min,max);
  tformed=(tformed-min)*(1.0/(max-min));
  tformed.writePPM("tformed.ppm");
  */
  return mask;
}

void derivativeXY(const Imagef & im,Matrix & pd)
{
  int xx=0,yy=0;
  int sizex=im.w;
  int sizey=im.h;
  pd(0,0)=im(1,0)-im(0,0);
  pd(0,1)=im(0,1)-im(0,0);
  for(xx=1;xx<sizex-1;xx++){
    pd(xx,0)=(im(xx+1,0)-im(xx-1,0))/2.0;
    pd(xx,1)=(im(xx,1)-im(xx,0));
  }
  pd(xx,0)=im(xx,0)-im(xx-1,0);
  pd(xx,1)=im(xx,1)-im(xx,0);
  for(yy=1;yy<sizey-1;yy++){
    xx=0;
    pd(yy*sizex+xx,0)=im(xx+1,yy)-im(xx,yy);
    pd(yy*sizex+xx,1)=(im(xx,yy+1)-im(xx,yy-1))/2.0;
    for(xx=1;xx<sizex-1;xx++){
      pd(yy*sizex+xx,0)=(im(xx+1,yy)-im(xx-1,yy))/2.0;
      pd(yy*sizex+xx,1)=(im(xx,yy+1)-im(xx,yy-1))/2.0;
    }
    pd(yy*sizex+xx,0)=im(xx,yy)-im(xx-1,yy);
    pd(yy*sizex+xx,1)=(im(xx,yy+1)-im(xx,yy-1))/2.0;
  }
  xx=0;
  pd(yy*sizex+xx,0)=im(xx+1,yy)-im(xx,yy);
  pd(yy*sizex+xx,1)=im(xx,yy)-im(xx,yy-1);
  for(xx=1;xx<sizex-1;xx++){
    pd(yy*sizex+xx,0)=(im(xx+1,yy)-im(xx-1,yy))/2.0;
    pd(yy*sizex+xx,1)=(im(xx,yy)-im(xx,yy-1));
  }
  pd(yy*sizex+xx,0)=im(xx,yy)-im(xx-1,yy);
  pd(yy*sizex+xx,1)=im(xx,yy)-im(xx,yy-1);
}

#include "dgels.h"

int initMultiResXY(Imagef & region,Imagef ** regions,
		    Matrix ** pds)
{
  int nres=0;
  int x=region.w;
  int y=region.h;
  while(x>2 && y>2){
    x/=2;
    y/=2;
    nres++;
  }
  (*regions) = new Imagef[nres];
  (*pds) = new Matrix[nres];
  (*regions)[0]=region;
  (*pds)[0]=Matrix(region.w*region.h,2);
  derivativeXY((*regions)[0],(*pds)[0]);
  x=region.w;
  y=region.h;
  for(int i=1;i<nres;i++){
    x/=2;
    y/=2;
    (*regions)[i]=region.resize(x,y);
    //char fname[64];
    //sprintf(fname,"region%d.ppm",i);
    //(*regions)[i].writePPM(fname);
    (*pds)[i]=Matrix(x*y,2);
    derivativeXY((*regions)[i],(*pds)[i]);
  }
  return nres;
}

void trackXYMultiRes(Imagef & image,Imagef * regions,
		     Matrix * pds,int nres,
		     double & x,double & y)
{
  int xint=(int)x;
  int yint=(int)y;
  
  for(int i=0;i>=0;i--){
    //Clamp region to image
    if(xint<1){
      x=1;
      xint=1;
    }
    if(yint<1){
      y=1;
      yint=1;
    }
    if(xint+regions[0].w>=image.w){
      xint=regions[0].w-image.w-1;
      x=xint;
    }
    if(yint+regions[0].h>=image.h){
      yint=regions[0].h-image.h-1;
      y=yint;
    }
    Imagef subim=image.subimage(xint,yint,regions[0].w,regions[0].h);
    if(i!=0){
      subim=subim.resize(regions[i].w,regions[i].h);
    }
    Matrix diff(regions[i].w*regions[i].h,1);
    for(int j=0;j<regions[i].w*regions[i].h;j++)
      diff[j]=subim[j]-regions[i][j];
    
    Matrix difft=diff.transpose();
    Matrix pdt=pds[i].transpose();
    Matrix work(pdt.m*pdt.m,1);
    int info;
    dgels_('N',pds[i].m,pds[i].n,1,pdt.data,pds[i].m,
	   difft.data,pds[i].m,work.data,work.m,info);
    printf("%f %f\n",difft[0],difft[1]);
    //FIXME: the movement should be with respect to the resolution...
    x-=difft[0];//*((double)regions[0].w)/((double)regions[i].w);
    y-=difft[1];//*((double)regions[0].h)/((double)regions[i].h);
  }
}
void trackXY(Imagef & image,Imagef & region,
	     Matrix & pd,double & x,double & y)
{
  int xint=(int)x;
  int yint=(int)y;

  //Clamp region to image
  if(xint<1){
    x=1;
    xint=1;
  }
  if(yint<1){
    y=1;
    yint=1;
  }
  if(xint+region.w>=image.w){
    xint=region.w-image.w-1;
    x=xint;
  }
  if(yint+region.h>=image.h){
    yint=region.h-image.h-1;
    y=yint;
  }
  Imagef subim=image.subimage(xint,yint,region.w,region.h);
  Matrix diff(region.w*region.h,1);
  for(int i=0;i<region.w*region.h;i++)
    diff[i]=subim[i]-region[i];
    
  Matrix difft=diff.transpose();
  Matrix pdt=pd.transpose();
  Matrix work(pdt.m*pdt.m,1);
  int info;
  dgels_('N',pd.m,pd.n,1,pdt.data,pd.m,
	 difft.data,pd.m,work.data,work.m,info);
  x-=difft[0];
  y-=difft[1];
}

void motionEstimation(Imagef & image1,Imagef & image2,int size,
		      Matrix & xmov,
		      Matrix & ymov)
{
  int nw=image1.w/size;
  int nh=image1.h/size;

  if(nw*size!=image1.w ||
     nh*size!=image1.h){
    printf("image not divisible by size\n");
  }
  xmov=Matrix(nw,nh);
  ymov=Matrix(nw,nh);
  Matrix pd(size*size,2);
  Imagef diff=image2-image1;
  for(int y=0;y<image1.h;y+=size){
    for(int x=0;x<image1.w;x+=size){
      Imagef subim1=image1.subimage(x,y,size,size);
      Imagef diffim=diff.subimage(x,y,size,size);
      Matrix diff(size*size,1);
      
      derivativeXY(subim1,pd);
      
      for(int i=0;i<size*size;i++)
	diff[i]=diffim[i];

      Matrix difft=diff.transpose();
      Matrix pdt=pd.transpose();
      Matrix work(pdt.m*pdt.m,1);
      int info;
      dgels_('N',pd.m,pd.n,1,pdt.data,pd.m,
	     difft.data,pd.m,work.data,work.m,info);
      //printf("info %d\n",info);
      //Matrix coeff=Matrix::linLeastSq(pd,diff);
      //coeff.printMatlab("coeff");
      xmov(x/size,y/size)=difft[0];
      ymov(x/size,y/size)=difft[1];
    }
    
  }
}

template <class T>
Image<T> applyHomography(Image<T> im,
			 Matrix & H,
			 double minx,
			 double maxx,
			 double miny,
			 double maxy)
{
  Matrix Hinv=H.inverse();

  int width=(int)(maxx-minx); //ceil?
  int height=(int)(maxy-miny);
  Image<T> res(width,height,im.nchannels);
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      Matrix pt(3,1);
      pt[0]=x+minx;
      pt[1]=y+miny;
      pt[2]=1;
      pt=Hinv*pt;
      pt[0]/=pt[2];
      pt[1]/=pt[2];
      int x0=(int)pt[0];
      int y0=(int)pt[1];
      double a=pt[0]-x0;
      double b=pt[1]-y0;
      if(x0>=0 && x0<im.w &&
	 y0>=0 && y0<im.h){
	int x1=x0+1;
	int y1=y0+1;
	if(x1==im.w)x1=x0;
	if(y1==im.h)y1=y0;
	for(int k=0;k<im.nchannels;k++){
	  res(x,y,k)=(T)((1.0-a)*(1.0-b)*((double)im(x0,y0,k))+
			 (a)*(1.0-b)*((double)im(x1,y0,k))+
			 (a)*(b)*((double)im(x1,y1,k))+
			 (1.0-a)*(b)*((double)im(x0,y1,k)));
	}
      }
      else{
	for(int k=0;k<im.nchannels;k++){
	  res(x,y,k)=(T)0.0;
	}
      }
    }
  }  
  return res;
}

template Imagef removeDistortion(Imagef & image,Matrix & intr,Matrix & dist);
template Image8 removeDistortion(Image8 & image,Matrix & intr,Matrix & dist);

template Image8 colorSeg(Image<unsigned char> & im,
			 Image <unsigned char> & col);
template Imagef applyHomography(Image<float> ,Matrix &,
				double,double,double,double);
template Image8 applyHomography(Image8,Matrix &,
				double,double,double,double);



#ifdef IMAGE_MAIN
#include <GL/glut.h>

double sx,sy;
int hasdisplayed=0;
int ww=640;
int wh=480;
int bdown=0;
GLuint tex;
Image8 image;
char ** imageNames=0;
int nimages=0;
int imagei=0;

double rect[4];
int gotregion=0;

void imageChanged();

void resize(int x,int y)
{
  ww=x;
  wh=y;
  glViewport(0,0,ww,wh);
}

void mouse(GLint button,GLint state,int x,int y)
{
  if(state==GLUT_DOWN){
    rect[0]=(double)x;
    rect[1]=(double)y;
    bdown=button+1;
  }
  else {
    bdown=0;
    rect[2]=(double)x;
    rect[3]=(double)y;
    rect[2]-=rect[0];
    rect[3]-=rect[1];
    gotregion=1;
  }
  glutPostRedisplay();
}
void motion(int x,int y)
{
  if(bdown){
    rect[2]=(double)x;
    rect[3]=(double)y;
    rect[2]-=rect[0];
    rect[3]-=rect[1];
    glutPostRedisplay();
  }
}

void keys(unsigned char c,int x,int y)
{
  switch(c){
  case '+':
    ww*=2;
    wh*=2;
    glutReshapeWindow(ww,wh);
    break;
  case '-':
    ww/=2;
    wh/=2;
    glutReshapeWindow(ww,wh);
    break;
  case ' ':
    if(hasdisplayed){
      imagei++;
      imageChanged();
      hasdisplayed=0;
    }
    break;
  }
  glutPostRedisplay();
}


void drawChecks()
{
  int vp[4];
  glGetIntegerv(GL_VIEWPORT,vp);

  GLuint checks;
  glGenTextures(1,&checks);
  glViewport(0,0,32,32);
  glBegin(GL_QUADS);
  //draw some checks
  for(int i=0;i<2;i++){
    for(int j=0;j<2;j++){
      glColor3f(0.6,0.6,0.6);
      if((i+j)%2==0)
	glColor3f(0.7,0.7,0.7);
      glVertex3f(i/(float)2.0f,j/(float)2.0f,0);
      glVertex3f((i+1)/(float)2.0f,j/(float)2.0f,0);
      glVertex3f((i+1)/(float)2.0f,(j+1)/(float)2.0f,0);
      glVertex3f(i/(float)2.0f,(j+1)/(float)2.0f,0);
    }
  }
  glEnd();
  glBindTexture(GL_TEXTURE_2D,checks);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
  glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB,
		   0,0,32,32,0);
 
  glViewport(vp[0],vp[1],vp[2],vp[3]);
  
  float tcx=(sx)*image.w/16.0f;
  float tcy=(sx)*image.h/16.0f;
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,checks);
  glBegin(GL_QUADS);
  glTexCoord2f(0,0);  glVertex2f(0,0);
  glTexCoord2f(tcx,0); glVertex2f(1.0,0);
  glTexCoord2f(tcx,tcy);glVertex2f(1.0,1.0);
  glTexCoord2f(0,tcy); glVertex2f(0,1.0);
  glEnd();
  glDeleteTextures(1,&checks);
}

void display()
{
  glClearColor(0.0,0.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  
  hasdisplayed=1;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,1,1,0,-1,1);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  drawChecks();
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tex);
  
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glColor3f(1.0,1.0,1.0);
  glBegin(GL_QUADS);
   glTexCoord2f(0,0);  glVertex2f(0,0);
   glTexCoord2f(sx,0); glVertex2f(1.0,0);
   glTexCoord2f(sx,sy);glVertex2f(1.0,1.0);
   glTexCoord2f(0,sy); glVertex2f(0,1.0);
  glEnd();
  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);



  Imagef a,b;
  a.readTGA(imageNames[imagei]);
  /*Imagef res=a.resize((int)(320.0/7.0),
		      (int)(240.0/7.0));
  res.writePPM("res.ppm");
  */
  
  a=a.getChannel(0);
  
  b.readTGA(imageNames[(imagei+1)%nimages]);
  b=b.getChannel(0);
  if(a.w==b.w && a.h==b.h){
    Matrix x,y;
    int sz=16;
    int sz2=8;
    motionEstimation(a,b,sz,x,y);
     
    glLoadIdentity();
    glScalef(1.0/((double)b.w),1.0/((double)b.h),1.0);
    for(int i=sz2;i<b.h;i+=sz){
      for(int j=sz2;j<b.w;j+=sz){
	Vec2d dir(x((j-sz2)/sz,(i-sz2)/sz),
		  y((j-sz2)/sz,(i-sz2)/sz));
	dir*=-1;
	if(dir.x>1)dir.x=1;
	if(dir.y>1)dir.y=1;
	if(dir.x<-1)dir.x=-1;
	if(dir.y<-1)dir.y=-1;
	dir*=sz2;
      
	glBegin(GL_LINES);
	glVertex2f(j,i);
	glVertex2f(j+dir.x,i+dir.y);
	glEnd();
      }
    }
  }  
  static int cnt=0;
  static Imagef region;
  static int nregions=0;
  static Imagef * regions=0;
  static Matrix * pds=0;
  static Matrix pd;
  
  if(rect[2]>1 && rect[3]>1){
    if(gotregion){
      /*region=a.subimage((int)rect[0],(int)rect[1],
			(int)(rect[2]),(int)(rect[3]));
      pd=Matrix(region.w*region.h,2);
      printf("getting derivative, regionsize: %d %d\n",region.w,region.h);
      derivativeXY(region,pd);*/
      
      region=a.subimage((int)rect[0],(int)rect[1],
			(int)(rect[2]),(int)(rect[3]));
      if(regions)delete [] regions;
      if(pds)delete [] pds;
      printf("initing multi res\n");
      nregions=initMultiResXY(region,&regions,&pds);
      printf("inited multi res\n");
      gotregion=0;
      
    }
    if(region.w>0 && region.h>0){
      /*for(int i=0;i<10;i++)
	trackXY(a,region,pd,rect[0],rect[1]);*/
      for(int i=0;i<10;i++)
	trackXYMultiRes(a,regions,pds,nregions, rect[0],rect[1]);
    }
  }
  glLoadIdentity();
  glScalef(1.0/((double)a.w),1.0/((double)a.h),1.0);
  glBegin(GL_LINE_LOOP);
  glVertex2f(rect[0],rect[1]);
  glVertex2f(rect[0]+rect[2],rect[1]);
  glVertex2f(rect[0]+rect[2],rect[1]+rect[3]);
  glVertex2f(rect[0],rect[1]+rect[3]);
  glEnd();

  if(nregions && regions[0].w){
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glTexCoord2f(sx*rect[0]/((double)a.w),sy*rect[1]/((double)a.h));
    glVertex2f(0,0);
    glTexCoord2f(sx*(rect[0]+rect[2])/((double)a.w),
		 sy*rect[1]/((double)a.h));
    glVertex2f(rect[2],0);
    glTexCoord2f(sx*(rect[0]+rect[2])/((double)a.w),
		 sy*(rect[1]+rect[3])/((double)a.h));
    glVertex2f(rect[2],rect[3]);
    glTexCoord2f(sx*rect[0]/((double)a.w),
		 sy*(rect[1]+rect[3])/((double)a.h));
    glVertex2f(0,rect[3]);
    glEnd();
    
    //glTranslatef(rect[2],0,0);
    GLuint tempTex;
    glGenTextures(1,&tempTex);
    
    glBindTexture(GL_TEXTURE_2D,tempTex);
    Imagef temp=regions[0]*1;
    temp.initTexture();

    glBegin(GL_QUADS);
    glTexCoord2f(0,0);glVertex2f(0,0);
    glTexCoord2f(1,0);glVertex2f(rect[2],0);
    glTexCoord2f(1,1);glVertex2f(rect[2],rect[3]);
    glTexCoord2f(0,1);glVertex2f(0,rect[3]);
    glEnd();

    temp=a.subimage((int)rect[0],(int)rect[1],
		    (int)(rect[2]),(int)(rect[3]));
    printf("%dx%dx%d %dx%dx%d\n",temp.w,temp.h,temp.nchannels,
	   regions[0].w,regions[0].h,regions[0].nchannels);
    temp=temp.getChannel(0)-regions[0];
   
    temp.fabs();
    temp*=4;
    temp.initTexture();
    //glTranslatef(rect[2],0,0);
    glBegin(GL_QUADS);
    glTexCoord2f(0,0);glVertex2f(0,0);
    glTexCoord2f(1,0);glVertex2f(rect[2]*2,0);
    glTexCoord2f(1,1);glVertex2f(rect[2]*2,rect[3]*2);
    glTexCoord2f(0,1);glVertex2f(0,rect[3]*2);
    glEnd();

    glLoadIdentity();
    glDeleteTextures(1,&tempTex);


  }
  glutSwapBuffers();

}

void imageChanged()
{
  if(imagei>=nimages)imagei=0;
  image.read(imageNames[imagei]);
  glBindTexture(GL_TEXTURE_2D,tex);
  
  image.initTextureNoResize(&sx,&sy,GL_REPLACE);
  ww=image.w;
  wh=image.h;
  glutReshapeWindow(ww,wh);
  glutSetWindowTitle(imageNames[imagei]);
  glViewport(0,0,ww,wh);
}
void init()
{
  memset(rect,0,sizeof(double)*4);
  glGenTextures(1,&tex);
  imageChanged();
}

int main(int ac,char ** av){
  /*Image<unsigned char> a(2,2,1);
  Image<unsigned char> b(2,2,1);
  Image<unsigned char> c(2,2,1);
  a=1;
  a(1,1)=2;
  a.print("%d ");
  b=a+2;
  b.print("%d ");
  c=a&b;
  c.print("%d ");
  c.initTexture();
  printf("done addition\n");
  */
  if(ac==0){
    return 0;
  }
  else{
    imagei=0;
    nimages=ac-1;
    imageNames = new char *[nimages];
    for(int i=0;i<nimages;i++){
      imageNames[i]=new char[strlen(av[i+1])+1];
      strcpy(imageNames[i],av[i+1]);
    }
  }
  glutInit(&ac,av);
  glutInitDisplayMode(GLUT_RGB|GLUT_DEPTH|GLUT_DOUBLE);
  glutInitWindowSize(ww,wh);
  glutInitWindowPosition(0,0);
  glutCreateWindow("win");
  glutDisplayFunc(display);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keys);
  init();
  glutMainLoop();
  
  return 0;
}
#endif


void findBounds(Image8 & im1,Matrix & H1,
		double & minx,double & maxx)
{
  Matrix pts(3,4);
  pts.setAll(1);
  pts(0,0)=0;pts(1,0)=0;
  pts(0,1)=im1.w;pts(1,1)=0;
  pts(0,2)=im1.w;pts(1,2)=im1.h;
  pts(0,3)=0;pts(1,3)=im1.h; 

  pts=H1*pts;
  for(int j=0;j<pts.n;j++){
    pts(0,j)/=pts(2,j);
    pts(1,j)/=pts(2,j);
  }

  minx=pts.minInRow(0);
  maxx=pts.maxInRow(0);
}

void findBounds(Image8 & im1,Matrix & H1,
		Image8 & im2,Matrix & H2,
		double & minx,double & maxx,
		double & miny,double & maxy)
{
  Matrix pts(3,4);
  pts.setAll(1);
  pts(0,0)=0;pts(1,0)=0;
  pts(0,1)=im1.w;pts(1,1)=0;
  pts(0,2)=im1.w;pts(1,2)=im1.h;
  pts(0,3)=0;pts(1,3)=im1.h; 

  pts=H1*pts;
  for(int j=0;j<pts.n;j++){
    pts(0,j)/=pts(2,j);
    pts(1,j)/=pts(2,j);
  }

  minx=pts.minInRow(0);
  miny=pts.minInRow(1);
  maxx=pts.maxInRow(0);
  maxy=pts.maxInRow(1);

  pts.setAll(1);
  pts(0,0)=0;pts(1,0)=0;
  pts(0,1)=im2.w;pts(1,1)=0;
  pts(0,2)=im2.w;pts(1,2)=im2.h;
  pts(0,3)=0;pts(1,3)=im2.h; 

  pts=H2*pts;
  for(int j=0;j<pts.n;j++){
    pts(0,j)/=pts(2,j);
    pts(1,j)/=pts(2,j);
  }
  minx=MIN(pts.minInRow(0),minx);
  miny=MIN(pts.minInRow(1),miny);
  maxx=MAX(pts.maxInRow(0),maxx);
  maxy=MAX(pts.maxInRow(1),maxy);
}
#ifdef IMAGE_MAIN2
#include <GL/glut.h>

#include "newseg.h"
#include "clb.h"

Image8 im1,im2,Him1,Him2;
Matrix F,Fn;
Matrix Pn1,Pn2;
Matrix P1,P2;
GLuint win;
Matrix intr1;
Matrix extr1;
Matrix intr2;
Matrix extr2;
Matrix dist1;
Matrix dist2;

void texturedQuad()
{
  glBegin(GL_QUADS);
  glTexCoord2f(0,0);glVertex2f(0,0);
  glTexCoord2f(1,0);glVertex2f(1,0);
  glTexCoord2f(1,1);glVertex2f(1,1);
  glTexCoord2f(0,1);glVertex2f(0,1);
  glEnd();
}

void drawEpipolarLine(Matrix & P1,Matrix & P2,Matrix & F,
		      Matrix & pt)
{
  Matrix pt1=P1*pt;
  Matrix pt2=P2*pt;

  pt1[0]/=pt1[2];
  pt1[1]/=pt1[2];
  pt1[2]/=pt1[2];

  pt2[0]/=pt2[2];
  pt2[1]/=pt2[2];
  pt2[2]/=pt2[2];

  Matrix err=pt2.transpose()*F*pt1;
  
  Matrix tmp[2];
  tmp[0]=F*pt1;
  tmp[1]=pt2.transpose()*F;
  Vec3d line[2];
  line[1]=Vec3d(tmp[0][0],tmp[0][1],tmp[0][2]);
  line[0]=Vec3d(tmp[1][0],tmp[1][1],tmp[1][2]);
  Vec3d x0y0(0,0,1);
  Vec3d x0y1(0,im1.h,1);
  Vec3d x1y1(im1.w,im1.h,1);
  Vec3d x1y0(im1.w,0,1);
  Vec3d isect[2][4];
  Vec3d border[4];
  border[0]=x0y1.cross(x0y0);
  border[1]=x1y0.cross(x0y0);
  border[2]=x1y1.cross(x1y0);
  border[3]=x1y1.cross(x0y1);
  for(int i=0;i<2;i++){
    for(int j=0;j<4;j++){
      isect[i][j]=line[i].cross(border[j]);
      isect[i][j]/=isect[i][j].z;
    }
  }
  int got[2]={0,0};
  Vec3d startEnd[2][4];
  double eps=1;

  for(int i=0;i<2;i++){
    for(int j=0;j<4;j++){
      if(isect[i][j].x>=-eps && isect[i][j].y>=-eps &&
	 isect[i][j].x<im1.w+eps && isect[i][j].y<im1.h+eps){
	startEnd[i][got[i]++]=isect[i][j];
      }
    }
  }
  if(got[0]>=2){
    glColor3f(1.0,1.0,1.0);
    glPushMatrix();
    glScalef(1.0/(double)im1.w,1.0/(double)im1.h,1);

    glBegin(GL_LINES);
    glVertex2dv(startEnd[0][0].data);
    glVertex2dv(startEnd[0][1].data);
    glEnd();
    glPopMatrix();
  }

  if(got[1]>=2){
    glColor3f(1.0,1.0,1.0);
    glPushMatrix();
    glTranslatef(1,0,0);
    glScalef(1.0/(double)im1.w,1.0/(double)im1.h,1);
    glBegin(GL_LINES);
    glVertex2dv(startEnd[1][0].data);
    glVertex2dv(startEnd[1][1].data);
    glEnd();
    glPopMatrix();
  }
}

void drawEpipolarLines(Matrix & P1,Matrix & P2,Matrix & F)
{
  Matrix pt(4,1);

  pt[0]=0;pt[1]=0;pt[2]=0;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
  pt[0]=1;pt[1]=0;pt[2]=0;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
  pt[0]=0;pt[1]=1;pt[2]=0;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
  pt[0]=0;pt[1]=0;pt[2]=1;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
  pt[0]=-1;pt[1]=0;pt[2]=0;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
  pt[0]=0;pt[1]=-1;pt[2]=0;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
  pt[0]=0;pt[1]=0;pt[2]=-1;pt[3]=1;
  drawEpipolarLine(P1,P2,F,pt);
}
void display()
{
  GLuint tex;
  glGenTextures(1,&tex);

  glClear(GL_COLOR_BUFFER_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,2,2,0,-1,1);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tex);
  im1.initTexture();
  texturedQuad();

  glTranslatef(1,0,0);
  im2.initTexture();
  texturedQuad();
  glTranslatef(-1,0,0);

  glDisable(GL_TEXTURE_2D);
  drawEpipolarLines(P1,P2,F);
  glEnable(GL_TEXTURE_2D);

  glTranslatef(0,1,0);
  Him1.initTexture();
  texturedQuad();
  
  glTranslatef(1,0,0);
  Him2.initTexture();
  texturedQuad();
  glTranslatef(-1,0,0);

  glDisable(GL_TEXTURE_2D);
  drawEpipolarLines(Pn1,Pn2,Fn);
  glEnable(GL_TEXTURE_2D);

  glDeleteTextures(1,&tex);
  glutSwapBuffers();
}
void keys(unsigned char c,int x,int y)
{
  switch(c){
  case 27:
    exit(0);
    break;
  }
}

int main(int ac,char * av[])
{
  glutInit(&ac,av);
  glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE);
  win=glutCreateWindow("win");
  glutDisplayFunc(display);
  glutKeyboardFunc(keys);

  if(ac<=1){
    printf("rect image.tga");
    printf("results in an undistored image, undist.tga being written");
    printf("\n");
    printf("rect r image1.tga image2.tga");
    printf("rectifies the images so that epipolar lines are horizontal");
    exit(0);
  }
  if(av[1][0]=='r'){
    if(ac<4){
      printf("not enough arguments for rectification\n");
      exit(0);
    }
    Matrix proj1;
    Matrix proj2;
    Matrix proj=Matrix(3,4);
    CalibInfo ci;
    proj.setAll(0);
    proj(0,0)=1;
    proj(1,1)=1;
    proj(2,2)=1;

    char fname1[1024];
    char fname2[1024];

    im1.read(av[2]);
    im2.read(av[3]);

    sprintf(fname1,"%s.clb",av[2]);
    sprintf(fname2,"%s.clb",av[3]);
    
    FILE * f1,* f2;
    f1=fopen(fname1,"rb");
    f2=fopen(fname2,"rb");
    
    if(!f1 || !f2){
      printf("error opening calib files\n");
      return 0;
    }
    
    readCLB(f1,intr1,extr1,proj1,dist1,ci);
    readCLB(f2,intr2,extr2,proj2,dist2,ci);

    fclose(f1);fclose(f2);

    P1=intr1*proj*extr1;
    P2=intr2*proj*extr2;
    
    Matrix T1,T2;
    if(av[1][1]=='0' || 
       av[1][1]==' '){
      printf("rectify 0\n");
      rectify(P1,P2,T1,T2,Pn1,Pn2);
    }
    else {
      printf("rectify 1\n");
      rectify(intr1,extr1,intr2,extr2,T1,T2,Pn1,Pn2);
    }
    double minx,maxx,miny,maxy;
    //Use the miny and maxy bounds on y,
    //but use the individual image boudns on x
    findBounds(im1,T1,im2,T2,minx,maxx,miny,maxy);
    findBounds(im1,T1,minx,maxx);
    printf("got rangex=[%f,%f]\n",minx,maxx);
    Him1=applyHomography(im1,T1,minx,maxx,miny,maxy);
    findBounds(im2,T2,minx,maxx);
    printf("got rangex=[%f,%f]\n",minx,maxx);
    Him2=applyHomography(im2,T2,minx,maxx,miny,maxy);
    Him1.writeTGA("him1.tga");
    Him2.writeTGA("him2.tga");
    Matrix zero(3,1);
    zero[0]=0;
    zero[1]=0;
    zero[2]=1;
    Matrix in2=T1*zero;
    in2*=(1.0/in2[2]);
    printf("In2 is %f %f\n",in2[0],in2[1]);
    F=fundamentalMatrix(P1,P2);
    Fn=fundamentalMatrix(Pn1,Pn2);
    glutMainLoop();
    return 0;
  }
  Image8 im;
  im.read(av[1]);
  printf("done reading %d %d %d\n",im.w,im.h,im.nchannels);
  Matrix intr=Matrix::eye(3);
  Matrix dist(5,1);
  intr(0,0)=363.0017;
  intr(1,1)=363.5376;
  intr(0,2)=152.3608;
  intr(1,2)=113.1121;
  dist[0]=-0.4074;
  dist[1]=0.5339;
  dist[2]=0.0004;
  dist[3]=-0.0038;
  dist[4]=0;
  Image8 im2=removeDistortion(im,intr,dist);
  
  im2.writeTGA("undist.tga");
  
  return 0;
}

#endif
