#include "imagedata.h"
#include "cameraprofile.h"

extern CameraProfiles cameraProfiles;

///Dialog for displaying ImageData properties.
class PropDialog:public wxDialog{
public:
  PropDialog(ImageData * imd):wxDialog(NULL,-1,wxT("Selection"),
				       wxDefaultPosition,
				       wxSize(260,240)){
    printf("showing properties\n");
    wxString profileName;
    printf("finding camera\n");
    CameraProfile * cp=cameraProfiles.findCameraProfile(imd->intrinsic,
							imd->distortion);
    if(cp){
      profileName = wxString::FromAscii(cp->name);
      printf("found profile, name %s\n",cp->name);
    }
    else
      profileName = wxT("N/A");

    wxString intr;
    wxString dist;
    if((imd->flags & HAS_INTRINSIC)>0){
      intr.Printf(wxT("fx:%6.2f fy:%6.2f cx:%6.2f cy:%6.2f"),
		   imd->intrinsic(0,0),
		   imd->intrinsic(1,1),
		   imd->intrinsic(0,2),
		   imd->intrinsic(1,2));
    }
    else
      intr = wxT("");

    if((imd->flags & HAS_DISTORTION)>0){
      dist.Printf(wxT("d=[%6.2f %6.2f %6.2f %6.2f %6.2f]"),
		   imd->distortion[0],
		   imd->distortion[1],
		   imd->distortion[2],
		   imd->distortion[3],
		   imd->distortion[4]);
    }
    else
      dist = wxT("");

    wxString message;
    message = wxString::Format(wxT("HAS_EXTRINSIC:%d\n"	\
		       "HAS_INTRINSIC:%d\n"	\
		       "HAS_PROJECTION:%d\n"	\
		       "HAS_DISTORTION:%d\n"	\
		       "HAS_CALIBINFO:%d\n"	\
		       "HAS_IMAGE:%d\n"		\
		       "HAS_MASK:%d\n"		\
		       "-------------\n"	\
		       "Camera profile:"),
		   (imd->flags&HAS_EXTRINSIC)>0,
		   (imd->flags&HAS_INTRINSIC)>0,
		   (imd->flags&HAS_PROJECTION)>0,
		   (imd->flags&HAS_DISTORTION)>0,
		   (imd->flags&HAS_CALIBINFO)>0,
		   (imd->flags&HAS_IMAGE)>0,
		   (imd->flags&HAS_MASK)>0)
      +profileName+wxT("\n")+intr+wxT("\n")+dist;

    printf("Showing properties %s\n",(const char*)message.ToAscii());
    wxStaticBoxSizer * box 
      = new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Properties")),wxHORIZONTAL);
    wxStaticText * text = new wxStaticText(this,-1,message);
    box->Add(text,0,wxGROW|wxEXPAND|wxLEFT|wxRIGHT|wxTOP|wxBOTTOM,8);
    SetSizer(box);
    SetAutoLayout(TRUE);
  }
};

ImageData::ImageData(){
  distortion=Matrix(5,1);
  intrinsic=Matrix::eye(3);
  extrinsic=Matrix::eye(4);
  projection=Matrix(3,4);
  projection.setAll(0);
  projection(0,0)=projection(1,1)=projection(2,2)=1;
  distortion.setAll(0);
  
  if(cameraProfiles.getCurrentParameters(intrinsic,distortion)){
    printf("FIXME(%s,%d): should imagedata be initialized with INTRINSIC params and distortion even if they are defaults\n",__FILE__,__LINE__);
    flags=HAS_INTRINSIC|HAS_DISTORTION;
  }
  else
    flags=0;

  calibInfo=0;
}


void ImageData::showProperties(int x,int y)
{
  wxDialog * dialog = new PropDialog(this);
  dialog->ShowModal();
  delete dialog;
}


