
#ifndef IMAGE_DATA_H
#define IMAGE_DATA_H

#include "image.h"
#include "matrix.h"
#include "seg/newseg.h"

///possible bitmask for flags in ImageData
enum {HAS_EXTRINSIC=1,HAS_INTRINSIC=2,HAS_PROJECTION=4,
      HAS_DISTORTION=8,HAS_CALIBINFO=16,
      HAS_IMAGE=32,HAS_MASK=64};

/** Container for image, calibration, etc.
 * This class is simply a container to hold all
 * the info (intrinsic, extrnisic calibration,
 * calibration information,image, etc...) about
 * an image currently in the system
 */
class ImageData{
 public:
  Image8 image;
  Image8 mask;
  int flags;
  Matrix projection;//not really necessary,but here anyhow
  Matrix extrinsic;
  Matrix intrinsic;
  Matrix distortion;
  CalibInfo * calibInfo;
  ImageData();
  void setMask(const Image8 & mask){
    this->mask=mask;
    set(HAS_MASK);
  }
  void set(int f){
    flags|=f;
  }
  void unset(int f){
    flags&=(~f);
  }
  void showProperties(int x=0,int y=0);
  ~ImageData(){
    delete calibInfo;
  }
};

#endif
