/*! \mainpage Main
 * \section usage Usage
 * The capture process consists of four main steps: capture/calibration, segmentation, geometry and texture coordinate generation.  These four steps are accessible using the toolbar buttons.
 * 
 * <img src="toolbar.jpg"><br>
 * <table align="left" border=0 padding=20><tr><td><img src="imagelist.jpg"></td></tr></table>
 * A capture project consists of a set of image files, the associated calibration (in a filename.ext.clb file, see clb.h), the geometry files (*.obj) and a single file containing the most recent scale and translation.  Such a project can be loaded from the command line, or from the load menu option.  When using the capture interface, there is always an active project directory, which is $HOME/capture_project by default, or the most recently loaded, or saved directory path.  Throughout the capture process, the system may need to save the project.  At this time, the project will be saved to the curently active project directory (also, any manual saves will be to this directory).  For this reason, and since the project files can get quite large it is a good idea to start the capture system up using a path to a local disk, which has plenty of free space.
 *
 * Before continuing with the overview of the individual steps, you should famialirze yourself with the interface.  During the capture process the images and their associated calibration and segmentation information are modified by the capture stages.  The images (and associated information) are accessible in the interface through three list boxes.  These are the ImageList (uncalibrated images), the CalibImageList (calibrated images, and the SegImageList(calibrated and segmented images).  The images in the project appear in only one of these list boxes at any given time and in a typical capture process the images will start in the ImageList, and then after calibration will move to the CalibImageList and finally after segmentation will appear in the SegImageList.    Operations are performed by selecting images (use ctrl/shift for multiselection) and then right clicking on the listbox to select an operation.  Images from the CalibImageList and SegImageList can be sent to the previous stage by using the remove operation.  Removing images from the ImageList, removes them from the project and frees the memory used by them. <br> 
 *  
 * \section capture Capture/Calibration
 * The capture/calibration stage can be accessed through the toolbar, and is the default stage for new projects.  The first step in capturing a model, is to obtain the images.  This can be done by right clicking on the uncalibirated imagelistbox (recall that this is the one on the upper left side of the dialog, see ImageList) and either loading images from disk, or capturing images from the camera.  
 * 
 *  Once the images have been obtained, they must be calibrated.  This step requires that each image in the ImageList has intrinsic calibration parameters.  The capture dialog will try to select these parameters automatically from the available CameraProfiles, and will prompt you to enter a new CameraProfile if one cannot be found.  Upon loading files from disk, the parameters from the currently active CameraProfile will be used (Note: this can be changed by issuing a force from the CameraDialog).  Select the images you wish to calbirate, and then select the calibration option from the popup menu to calibrate the selected image(s).
 * 
 * The CalibrateDialog allows a few options when it comes to calibration.  Option 1: The color channel and threshold can be used to generate a black and white image, where the white regions are detected as the circles on the calibration pattern.  Option 2: The automatic checkbox can be pressed, and the system will attempt to find the circles without using the provided threshold.  The small and big SpinCtrl's can be used to ensure that any objects outside that range will not be used.  When the parameters are as desired, select OK and wait for the calibration to finish.  Image are only considered to be successfully calibrated if the residual in the calibration is below a threshold (which is in the options file) and the minimum number of lines has been used to obtain this calibration.  Successfully calibrated images move into the CalibImageList, while unsuccessful images remain in the ImageList.  
 * 
 * \section segment Segmentation
 * <img src="segmentdlg.jpg"><br>
 * Now that all the images have been calibrated, you can move onto the Segmentation stage by choosing the button on the toolbar.  Segmentation proceeds in the same manner as the calibration, in that you first select the desired images, and then right-click and choose the segment option.  The segment option in the popup-menu will open a SegmentDialog, which is used in the segmentation process.  The SegmentDialog has support for color segmentation and background subtraction.  
 * \subsection colorseg Color Segmentation
 * The color segmentation proceeds by adding color regions using the Add button, and selecting the appropriate background color in the images.  This can be repeated until the desired segmentation has been reached.
 *
 * \subsection background Background Subtraction
 * Background subtraction is accomplished by loading the background image, and adjusting the threshold until a desired segmentation has been reached.
 *
 * Choosing okay from the SegmentDialog will segment all the images.  The images then move into the SegImageList.  Removing an image from the SegImageList sends it back to the CalibImageList, where it can be resegmented.
 *
 * \section geom Geometry
 * <table align=center border=0><tr><td><img src=bb.jpg></td><td><img src="bbdlg.jpg"></td></tr></table><br>
 * The geometry is now created using the available methods.  Pressing the Apply button in the upper notebook will call the selected algorithm.  The algorithm may require you to select a bounding box in the images.  When necessary a window will popup (a BBDialog) allowing you to scale and translate the bounding box. The left mouse button allows for translation in the y direction (or also in the x-z direction when ctrl is held down).  The right mouse button allows for scale in x,y,z directions (or when ctrl is held, a non-uniform scaling, with y-translation maps to y-scale, and x-translation maps to x-z scale).  Once the appropriate bounding box has been found, selecting Okay will call the geometry algorithm.  Running the algorithm may take some time, so be patient.  When it is finished, you will see the resulting geometry.  
 *
 * The resulting geometry can be modified using the operations in the bottom notebook(FIXME). 
 *
 * \section tcg Texture Coordinate Generation
 * <img src="texture.jpg"><br>
 * The texture coordinate generation works as follows: First select the chart from either the listbox or by clicking on the chart.  Then you may apply one of the operations.  Typically you will first have to split the original geometry (only one chart) with one of the split options.  The cone base split will group together triangles that have normals fitting inside a cone with the provided angle.  The distance based split will grow two charts from the furthest two vertices (more on this later).  Once the splitting has been done, if it is the first split, then the individual charts are mapped to 2D using the currently selected option for mapping. These charts are then packed into the 2D texture rectangle. 
 * 
 * The individual charts can then be further split by selecting and choosing a split option.  By default, these charts will not be re-mapped to 2D, so you will have to select the charts and choose the 2D mapping method.  If all is well, the charts should be gray in color, with outlines of the triangles (with the exception of the selected chart which is highlighted).  Any overlap in the texture coordinates is outlined in red.  Any charts that have self overlap have red highlights and a blue triangles denoting one place where two or more triangles intersect.  Removing overlap can be done by selecting the pack option, in the others tab, or by manually moving the charts.  Charts can be translated and scaled by holding down the left button and right button respectively.  Self overlap can be removed manually by splitting the chart with the distance method (which grows two charts from overlapping triangles) until there is no overlap.  There is also an automatic removal of self overlap, which can be found on the others tab also.  
 *
 * The current texture can be viewed on the model by navigating to the geometry mode, right clicking on the openGL window and selecting the Draw Texture checkbox.  If all is well, the model is now ready for export.
 *
 * \section export Export
 *  <img src=export.jpg> <br>
 * Once all the stages have been completed, you can now output the geometry with a dynamic texture.  Select the File->Export option to popup the export dialog (SaveDialog).   Here you can select the number of eigenvectors used for each channel, as well as the texture size for each channel.  The texture size of the U and V channels are forced to be smaller than that chosen for the Y channel.  The number of eigenvectors is forced to be a multiple of 4, since this is the lowest multiple that the software renderer can render.  The number of eigenvectors should be chosen to be a multiple of 16 for full hardware compatibility.  
 *
 * The export dialog provides several options for generating the lookup tables.  The original lookup-table (original lut) uses the euler angles of the camera to index a 2D lookup-table, which is generated with one of the provided methods.  The user can enter the desired lookup table dimensions, which should be chosen to be a power of 2 for forwards compatibility.  Currently the linear original lut is the best choice for objects captured in one revolution (Note the lut size is updated to be widthx2 for one dimensional luts).  The triangle original (triangulate the 2d (x-y euler angles) points and blend the coefficients linearly) lut is the most reliable for models captured with varying x rotation.  
 * 
 *  The newer cubemap lut uses the bounding box of the object as a cube map, which is indexed by the camera position relative to the object.  The lut size is the size of each face of the cube, so the actual lookup table is 6*lutwidth*lutheight.  The options for the cubemap lut are global, nearest and triangle.  The global uses each pixel in each face of the cube to generate a synthetic camera position, the coefficients of all the true camera positions are then weighted based on their similarity to the synthetic camera position.  Nearest is selects the nearest camera coefficients for the pixel.  Triangle cubemap lut uses qhull to triangulate the normalized camera positions (points on a unit sphere).  Each point on the cube is then projected onto the triangles, and the point on the triangle is used to spherically interpolate the coefficients of the vertices of the triangle.          
 *  
 * Once you have selected the desired options, choose okay to start the export procedure.  The final result is saved in current_project_name/output.
 */
