#include<math.h>
#include<stdio.h>
#include"general.h"

#define sign(x) ((x)<0?-1:1)

void clip_line(double *ln)
{
  double t;
  int i,j;
 
//printf("%f,%f,%f,%f -- %f,%f,%f,%f : ",ln[0],ln[1],ln[2],ln[3],ln[4],ln[5],ln[6],ln[7],ln[7]);
  
  for(j=0;j<3;j++)
  {
    t=(ln[j]-ln[3])/(ln[7]-ln[3]-ln[4+j]+ln[j]);
    if(t>0 && t<1)
    {
      if(ln[j]>ln[3])
	for(i=0;i<4;i++) ln[i]=ln[i]+t*(ln[i+4]-ln[i]);
      else if(ln[4+j]>ln[7])
	for(i=0;i<4;i++) ln[i+4]=ln[i]+t*(ln[i+4]-ln[i]);
    }
    t=-(ln[j]+ln[3])/(ln[7]-ln[3]+ln[4+j]-ln[j]);
    if(t>0 && t<1)
    {
      if(ln[j]<-ln[3])
	for(i=0;i<4;i++) ln[i]=ln[i]+t*(ln[i+4]-ln[i]);
      else if(ln[4+j]<-ln[7])
	for(i=0;i<4;i++) ln[i+4]=ln[i]+t*(ln[i+4]-ln[i]);
    }
  }

//printf(" %f,%f,%f,%f -- %f,%f,%f,%f\n",ln[0],ln[1],ln[2],ln[3],ln[4],ln[5],ln[6],ln[7],ln[7]);
}

void print_matrix(const double a[4][4],const char *s,const char *q)
{
  printf("%s[\n",s);
  for(int i=0;i<4;i++)
  {
    for(int j=0;j<4;j++)
      printf("%f ",a[i][j]);
    printf(";\n");
  }
  printf("]%s\n",q);
}

void normalize(double a[3])
{
  double len=magnitude(a);
  a[0]/=len;
  a[1]/=len;
  a[2]/=len;
}

void invert3(double b[4][4],const double a[4][4])
{
  double oodet=1.0/determinant3(a);
  b[0][0]=oodet*(a[2][2]*a[1][1]-a[2][1]*a[1][2]);
  b[0][1]=oodet*(a[2][1]*a[0][2]-a[2][2]*a[0][1]);
  b[0][2]=oodet*(a[1][2]*a[0][1]-a[1][1]*a[0][2]);
  b[1][0]=oodet*(a[2][0]*a[1][2]-a[2][2]*a[1][0]);
  b[1][1]=oodet*(a[2][2]*a[0][0]-a[2][0]*a[0][2]);
  b[1][2]=oodet*(a[1][0]*a[0][2]-a[1][2]*a[0][0]);
  b[2][0]=oodet*(a[2][1]*a[1][0]-a[2][0]*a[1][1]);
  b[2][1]=oodet*(a[2][0]*a[0][1]-a[2][1]*a[0][0]);
  b[2][2]=oodet*(a[1][1]*a[0][0]-a[1][0]*a[0][1]);

  b[3][0]=0; b[3][1]=0; b[3][2]=0;
  b[0][3]=0; b[1][3]=0; b[2][3]=0; b[3][3]=1;
}

void perspective1(double cam[4][4],double resx,double resy,
                   double dx,double dy,double dz,double zp)
{
  double oodz=1/dz;
  double ooQdz=oodz/sqrt(dx*dx+dy*dy+dz*dz);
  cam[0][0]=resx;
  cam[0][1]=0;
  cam[0][2]=-resx*dx*oodz;
  cam[0][3]=resx*zp*dx*oodz;
  cam[1][0]=0;
  cam[1][1]=resy;
  cam[1][2]=-resy*dy*oodz;
  cam[1][3]=resy*zp*dy*oodz;
  cam[2][0]=0;
  cam[2][1]=0;
  cam[2][2]=-zp*ooQdz;
  cam[2][3]=zp*zp*ooQdz+zp;
  cam[3][0]=0;
  cam[3][1]=0;
  cam[3][2]=-ooQdz;
  cam[3][3]=zp*ooQdz+1;
}

void perspective(double cam[4][4],double res,double dz,double zp)
{
  double oodz=1/dz;
  double ooQdz=oodz*oodz;
  cam[0][0]=res; cam[0][1]=0; cam[0][2]=0;        cam[0][3]=0;
  cam[1][0]=0; cam[1][1]=-res; cam[1][2]=0;        cam[1][3]=0;
  cam[2][0]=0; cam[2][1]=0; cam[2][2]=-zp*ooQdz;cam[2][3]=zp*zp*ooQdz+zp;
  cam[3][0]=0; cam[3][1]=0; cam[3][2]=-ooQdz;   cam[3][3]=zp*ooQdz+1;
}

void rotate_X(double a,double m_out[4][4],double m_in[4][4])
{
  static double m[4][4]={{1,0,0,0}, {0,1,0,0}, {0,0,1,0},{0,0,0,1}};

  m[2][2]=m[1][1]=cos(a);
  m[1][2]=-(m[2][1]=sin(a));

  mat_mult(m_out,m_in,m);
}

void rotate_Y(double a,double m_out[4][4],double m_in[4][4])
{
  static double m[4][4]={{1,0,0,0}, {0,1,0,0}, {0,0,1,0},{0,0,0,1}};

  m[2][2]=m[0][0]=cos(a);
  m[0][2]=-(m[2][0]=sin(a));

  mat_mult(m_out,m_in,m);
}

void rotate_Z(double a,double m_out[4][4],double m_in[4][4])
{
  static double m[4][4]={{1,0,0,0}, {0,1,0,0}, {0,0,1,0},{0,0,0,1}};

  m[1][1]=m[0][0]=cos(a);
  m[0][1]=-(m[1][0]=sin(a));

  mat_mult(m_out,m_in,m);
}

void mat_mult(double a[4][4],const double b[4][4],const double c[4][4])
{
  double temp[4][4];

  temp[0][0]=b[0][0]*c[0][0]+b[0][1]*c[1][0]+
             b[0][2]*c[2][0]+b[0][3]*c[3][0];
  temp[0][1]=b[0][0]*c[0][1]+b[0][1]*c[1][1]+
             b[0][2]*c[2][1]+b[0][3]*c[3][1];
  temp[0][2]=b[0][0]*c[0][2]+b[0][1]*c[1][2]+
             b[0][2]*c[2][2]+b[0][3]*c[3][2];
  temp[0][3]=b[0][0]*c[0][3]+b[0][1]*c[1][3]+
             b[0][2]*c[2][3]+b[0][3]*c[3][3];

  temp[1][0]=b[1][0]*c[0][0]+b[1][1]*c[1][0]+
             b[1][2]*c[2][0]+b[1][3]*c[3][0];
  temp[1][1]=b[1][0]*c[0][1]+b[1][1]*c[1][1]+
             b[1][2]*c[2][1]+b[1][3]*c[3][1];
  temp[1][2]=b[1][0]*c[0][2]+b[1][1]*c[1][2]+
             b[1][2]*c[2][2]+b[1][3]*c[3][2];
  temp[1][3]=b[1][0]*c[0][3]+b[1][1]*c[1][3]+
             b[1][2]*c[2][3]+b[1][3]*c[3][3];

  temp[2][0]=b[2][0]*c[0][0]+b[2][1]*c[1][0]+
             b[2][2]*c[2][0]+b[2][3]*c[3][0];
  temp[2][1]=b[2][0]*c[0][1]+b[2][1]*c[1][1]+
             b[2][2]*c[2][1]+b[2][3]*c[3][1];
  temp[2][2]=b[2][0]*c[0][2]+b[2][1]*c[1][2]+
             b[2][2]*c[2][2]+b[2][3]*c[3][2];
  temp[2][3]=b[2][0]*c[0][3]+b[2][1]*c[1][3]+
             b[2][2]*c[2][3]+b[2][3]*c[3][3];

  temp[3][0]=b[3][0]*c[0][0]+b[3][1]*c[1][0]+
             b[3][2]*c[2][0]+b[3][3]*c[3][0];
  temp[3][1]=b[3][0]*c[0][1]+b[3][1]*c[1][1]+
             b[3][2]*c[2][1]+b[3][3]*c[3][1];
  temp[3][2]=b[3][0]*c[0][2]+b[3][1]*c[1][2]+
             b[3][2]*c[2][2]+b[3][3]*c[3][2];
  temp[3][3]=b[3][0]*c[0][3]+b[3][1]*c[1][3]+
             b[3][2]*c[2][3]+b[3][3]*c[3][3];

  memcpy(a,temp,16*sizeof(double));
}

void threshold(bitmap *bm,unsigned char t)
{
  unsigned char *p=bm->data;
  for(int i=0;i<bm->w*bm->h;i++,p++)
  {
    if(*p>=t)*p=255;
    else *p=0;
  }
}

int loadtga(bitmap *im,char *filename)
{
  FILE *f=fopen(filename,"rb");
  if(!f)return 0;

  unsigned char bpp;
  
  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  char file_header[12];

  fread(file_header,sizeof(file_header),1,f);

  if (memcmp(file_header, tga_header_rgb, 12)) {
    fclose(f);
    return 0;
  }

  fread(&im->w,sizeof(im->w),1,f);
  fread(&im->h,sizeof(im->h),1,f);
  fread(&bpp,sizeof(bpp),1,f);
  bpp/=8;
  fgetc(f);
  if(bpp!=4){
    printf("tga file not RGBA %d\n",bpp);
    fclose(f);return 0;
  }

  unsigned char *data = new unsigned char[im->w*im->h*4];

  fread(data,im->w*im->h*4,1,f);
  im->data=new unsigned char[im->w*im->h];
  for(int i=0;i<im->w*im->h;i++) im->data[i]=data[i*4+3];
  delete data;

  fclose(f);
  return 1;
}


int loadpgm(bitmap *img,char *filename)
{
  char c;
  FILE *f=fopen(filename,"r");
  if(!f)return 0;

  // check magic number
  if(fgetc(f)!='P' || fgetc(f)!='5' || fgetc(f)!='\n')
  {
    fclose(f);
    return 0;
  }

  // ignore comments
  while((c=fgetc(f))=='#') while(fgetc(f)!='\n');
  ungetc(c,f);
  
  fscanf(f,"%hd %hd %*d\n",&(img->w),&(img->h));
  img->data=new unsigned char[img->w*img->h];

  // ignore comments
  while((c=fgetc(f))=='#') while(fgetc(f)!='\n');
  ungetc(c,f);
  
  fread(img->data,1,img->w*img->h,f);
  
  fclose(f);
  return 1;
}

int savepgm(bitmap *img,char *filename)
{
  FILE *f=fopen(filename,"w");
  if(!f)return 0;
  fprintf(f,"P5\n%d %d 255\n",img->w,img->h);
  
  fwrite(img->data,1,img->w*img->h,f);
  fclose(f);
  
  return 1;
}

void transform_point(double a[4],const double A[4],const double cam[4][4])
{
  a[0] = A[0]*cam[0][0] + A[1]*cam[0][1] + A[2]*cam[0][2] + A[3]*cam[0][3];
  a[1] = A[0]*cam[1][0] + A[1]*cam[1][1] + A[2]*cam[1][2] + A[3]*cam[1][3];
  a[2] = A[0]*cam[2][0] + A[1]*cam[2][1] + A[2]*cam[2][2] + A[3]*cam[2][3];
  a[3] = A[0]*cam[3][0] + A[1]*cam[3][1] + A[2]*cam[3][2] + A[3]*cam[3][3];
}

void project_point(double a[4],double b[4])
{
  a[0]=b[0]/b[3]; a[1]=b[1]/b[3]; a[2]=b[2]/b[3]; a[3]=b[3]/a[3];
  
  // viewport transform hack
//  b[0]=717/2*b[0]+717/2;
//  b[1]=705/2*b[1]+705/2;
}

#define check_pixel2 \
      if(x>=0 && y>=0 && x<img.w && y<img.h) img.data[y*img.w+x]=120;


#define check_pixel1 \
      if(x>=0 && y>=0 && x<img.w && y<img.h) {\
	if(img.data[y*img.w+x]!=in) { \
	  in=img.data[y*img.w+x]; \
	  cuts.push_back(cut2(in,((double)x-cx)/sx,((double)y-cy)/sy)); \
	}\
      } else {\
	if(in){\
	  in=0; \
	  cuts.push_back(cut2(in,((double)x-cx)/sx,((double)y-cy)/sy)); \
	}\
      }
#define check_pixel if(dbg){check_pixel2}else{check_pixel1}

void line::intersect(bitmap img,int dbg)
{
  int x,y, dx, dy, balance, incx, incy, in;

  //printf("%f,%f,%f,%f\n",x_1,y_1,x_2,y_2);

  int sx=1;
  int sy=1;
  int cx=0;
  int cy=0;

  int x1=(int)(x_1*sx+cx);
  int y1=(int)(y_1*sy+cy);
  int x2=(int)(x_2*sx+cx);
  int y2=(int)(y_2*sy+cy);


  if (x2 >= x1) { dx = x2 - x1; incx = 1; }
  else { dx = x1 - x2; incx = -1; }

  if (y2 >= y1) { dy = y2 - y1; incy = 1; }
  else { dy = y1 - y2; incy = -1; }

  if(isnan(x_1)|| isnan(x_2) || isnan(y_1) || isnan(y_2)
      || dx>1000 || dx<0 || dy>1000 || dy<0)return;

  x = x1;
  y = y1;

 
  if (dx >= dy)
  {
    dy <<= 1;
    balance = dy - dx;
    dx <<= 1;
    
    if(x != x2)
    {
      if (balance >= 0)
      {
	y += incy;
	balance -= dx;
      }
      balance += dy;
      x += incx;
      if(x>=0 && y>=0 && x<img.w && y<img.h) in=img.data[y*img.w+x];
      else in=0;
    }

    while (x != x2)
    {
      check_pixel;
      if (balance >= 0)
      {
	y += incy;
	balance -= dx;
      }
      balance += dy;
      x += incx;
    }// check_pixel;
  }
  else
  {
    dx <<= 1;
    balance = dx - dy;
    dy <<= 1;

    if (y != y2)
    {
      if (balance >= 0)
      {
	x += incx;
	balance -= dy;
      }
      balance += dx;
      y += incy;
      if(x>=0 && y>=0 && x<img.w && y<img.h) in=img.data[y*img.w+x];
      else in=0;
    } 

    while (y != y2)
    {
      check_pixel;
      if (balance >= 0)
      {
	x += incx;
	balance -= dy;
      }
      balance += dx;
      y += incy;
    }// check_pixel;
  }
  if(!in && cuts.empty() && !dbg)
  {
    cuts.push_back(cut2(1,0,0));
  }
}
