#ifndef GENERAL_H
#define GENERAL_H

#include<list>
using namespace std;
#include<math.h>

#define dot3(a,b) ((a)[0]*(b)[0] + (a)[1]*(b)[1] + (a)[2]*(b)[2])
#define magnitude(a)(sqrt((a)[0]*(a)[0]+(a)[1]*(a)[1]+(a)[2]*(a)[2]))

#define determinant3(a) (a[0][0]*(a[2][2]*a[1][1]-a[2][1]*a[1][2])\
                        -a[1][0]*(a[2][2]*a[0][1]-a[2][1]*a[0][2])\
		        +a[2][0]*(a[1][2]*a[0][1]-a[1][1]*a[0][2]))



void clip_line(double *);
void print_matrix(const double a[4][4],const char *s="",const char *q="");
void mat_mult(double a[4][4],const double b[4][4],const double c[4][4]);
void transform_point(double a[4],const double A[4],const double cam[4][4]);
void invert3(double b[4][4],const double a[4][4]);
void normalize(double a[3]);
void project_point(double a[4],double b[4]);
void rotate_X(double a,double m_out[4][4],double m_in[4][4]);
void rotate_Y(double a,double m_out[4][4],double m_in[4][4]);
void rotate_Z(double a,double m_out[4][4],double m_in[4][4]);
void perspective1(double cam[4][4],double resx,double resy,
                   double dx,double dy,double dz,double zp);
void perspective(double cam[4][4],double res,double dz,double zp);



struct bitmap
{
  unsigned char * data;
  unsigned char pixel(int x,int y)
  {
    if(x>=w || x<0 || y>=h || y<0) return 0;
    else return data[y*w+x];
  }
  short w;
  short h;
};

void threshold(bitmap *bm,unsigned char t);

struct cut2
{
  cut2(char i,double X,double Y){in=i;x=X;y=Y;};
  unsigned char in;
  double x;
  double y;
};

struct segment
{
  segment(double i=0,double o=0){in=i;out=o;};
  double in;
  double out;

  int disjoint(const segment &s) const
  {
    if(s.in>out || s.out<in)return true;
    return false;
  }

  segment intersect(const segment &s)
  {
    segment seg;
    seg.in=max(s.in,in);
    seg.out=min(s.out,out);
    return seg;
  }

  bool inside(double p)
  {
    return (p>in && p<out);
  }

  double distance(double p)
  {
    return min(fabs(p-in),fabs(p-out));
  }

  double intersect_point(double a,double b)
  {
    if(a<in && in<b)return in;
    if(a<out && out<b)return out;
    return (in+out)/2.0;
  }
};

int loadtga(bitmap *,char *);
int loadpgm(bitmap *,char *);
int savepgm(bitmap *,char *);

struct line
{
  double x_1;
  double y_1;
  double x_2;
  double y_2;
  double x_3;
  double y_3;

  void intersect(bitmap,int dbg=0);
  list<cut2> cuts;
};

#endif
