#include<math.h>
#include<stdio.h>
#include"mi.h"

line mi_ray::project(const double cam[4][4])
{
  line ln;
  double A[4],a[4];
  double B[4],b[4];
  double C[4],c[4];
  A[3]=1;
  B[3]=1;
  C[3]=1;
  switch(axis)
  {
    case 0: A[0]=-1; A[1]=origin[0]; A[2]=origin[1];
	    B[0]=1; B[1]=A[1]; B[2]=A[2];
	    C[0]=0; C[1]=A[1]; C[2]=A[2];
	    break;
    case 1: A[0]=origin[0]; A[1]=-1; A[2]=origin[1];
	    B[0]=A[0]; B[1]=1; B[2]=A[2];
	    C[0]=A[0]; C[1]=0; C[2]=A[2];
	    break;
    case 2: A[0]=origin[0]; A[1]=origin[1]; A[2]=-1; 
	    B[0]=A[0]; B[1]=A[1]; B[2]=1;
	    C[0]=A[0]; C[1]=A[1]; C[2]=0;
	    break;
  }

  transform_point(a,A,cam);
  transform_point(b,B,cam);
  transform_point(c,C,cam);
  project_point(a,a);
  project_point(b,b);
  project_point(c,c);
  
  ln.x_1=a[0];
  ln.y_1=a[1];
  ln.x_2=b[0];
  ln.y_2=b[1];
  ln.x_3=c[0];
  ln.y_3=c[1];
  
  return ln;
}

void mi_ray::unproject(const double cam[4][4],const line &l,int res)
{
  double a,b,c,d,cr,p;
  list<cut2>::const_iterator it,it2;
  list<segment>::iterator s1,s2;
  list<segment> seglist,segments_copy(segments);
  segment seg(-1,1);
  
  if(l.cuts.begin()->in==1 || segments.empty())
  {
    segments.clear();
    return;
  }
  if(l.cuts.empty())return;

  for(it=l.cuts.begin();it!=l.cuts.end();it++)
  {
    a=hypot(it->x - l.x_1, it->y - l.y_1);
    b=hypot(l.x_2 - l.x_3, l.y_2 - l.y_3);
    c=hypot(l.x_1 - l.x_3, l.y_1 - l.y_3);
    d=hypot(it->x - l.x_2, it->y - l.y_2);
    
    cr=(a*b)/(c*d);
    p=(cr-1)/(1+cr);

    if(it->in)seg.in=p;
    else seg.out=p;
    
    it2=it;
    it2++;
    if(it2==l.cuts.end() || !it->in) 
    {
      seglist.push_back(seg);
      seg.in=-1;
      seg.out=1;
    }
  }

  segments.clear();
  for(s2=seglist.begin();s2!=seglist.end();s2++)
  {
    s1=segments_copy.begin();
    while(s1!=segments_copy.end())
    {
      if(!s1->disjoint(*s2)) segments.push_back(s1->intersect(*s2));
      s1++;
    }
  }
  
  check(res);
}

void mi_ray::check(int res)
{
  list<segment>::iterator s1,s2,s3;

  s1=segments.begin();
  s2=segments.begin();
  s2++;
  while(s2!=segments.end())
  {
    if(int((s1->out+1)*(res-1)/2)==int((s2->in+1)*(res-1)/2))
    {
      s2->in=s1->in;
      s3=s1;
      s1++;
      s2++;
      segments.erase(s3);
    }
    else s1++,s2++;
  }

  
  s1=segments.begin();
  while(s1!=segments.end())
  {
    if(int((s1->in+1)*(res-1)/2)==int((s1->out+1)*(res-1)/2))
    {
      s3=s1;
      s1++;
      segments.erase(s3);
    }
    else s1++;
  }
}

double mi_ray::intersect_point(double a,double b)
{
  list<segment>::iterator it;
  for(it=segments.begin();it!=segments.end();it++)
  {
    if(it->inside(a)!=it->inside(b))
      return it->intersect_point(a,b);
  }  
  return (a+b)/2;
}

bool mi_ray::inside(double a)
{
  list<segment>::iterator it;
  for(it=segments.begin();it!=segments.end();it++)
  {
    if(it->inside(a))return true;
  } 
  return false;
}

segment * mi_ray::nearest(double p)
{
  list<segment>::iterator s1,s2;
  double dist,d;
  
  s1=segments.begin();
  s2=segments.begin();
  s1++;
  dist=s2->distance(p);

  while(s1!=segments.end())
  {
    if(s1->inside(p))return &(*s1);
    if((d=s1->distance(p))<dist)
    {
      s2=s1;
      dist=d;
    }
    s1++;
  }
  return &(*s2);
}

mi_grid::mi_grid(int r)
{
  res=r;
 
  int i,j,k;

  for(i=0;i<3;i++)
  {
    rays[i]=new mi_ray[res*res];
    for(j=0;j<res;j++)
    for(k=0;k<res;k++)
    {
      rays[i][j*res+k].axis=i;
      rays[i][j*res+k].origin[0]=2.0*j/((double)res-1)-1;
      rays[i][j*res+k].origin[1]=2.0*k/((double)res-1)-1;
      rays[i][j*res+k].segments.push_back(segment(-1,1));
    }
  }
}

void mi_grid::intersect(bitmap img,const double cam[4][4])
{
  int i,j,k;
  line ln;

  for(i=0;i<3;i++) for(j=0;j<res;j++) for(k=0;k<res;k++)
  {
 // printf("%d,%d,%d\n",i,j,k);
    if(rays[i][j*res+k].segments.empty())continue;
    ln=rays[i][j*res+k].project(cam);
    ln.intersect(img);
    rays[i][j*res+k].unproject(cam,ln,res);
  }

  //printf("-----------------------------------------------\n");
}

void mi_grid::render(void triangle(double *,double *,double *))
{
  int i,j,k;
	
  for(i=-1;i<res-1;i++)
  for(j=-1;j<res-1;j++)
  for(k=-1;k<res-1;k++)
  {
    mi_cube cb(this,i,j,k);
    cb.render(triangle);
  }
  
}

void mi_grid::fixpoint(int x,int y,int z)
{
  double p[3];
  mi_ray *r[3];
  int in[3],incount=0;
  int i;
  segment *s;
  double EPS=0.0001;

  r[0]=&(rays[0][y*res+z]);
  r[1]=&(rays[1][x*res+z]);
  r[2]=&(rays[2][x*res+y]);

  p[0]=2.0*double(x)/(res-1)-1.0;
  p[1]=2.0*double(y)/(res-1)-1.0;
  p[2]=2.0*double(z)/(res-1)-1.0;

  for(i=0;i<3;i++)
  {
    in[i]=r[i]->inside(p[i]);
    if(in[i])incount++;
  }
  if(incount==1) // move one intersection out
  {
    bad_incount1++;
    i=0;while(!in[i])i++;
    s=r[i]->nearest(p[i]);
    if(s->out-p[i] < p[i]-s->in)s->out=p[i]-EPS;
    else s->in=p[i]+EPS;
  }
  else if(incount==2) // move one intersection in
  {
    bad_incount2++;
    i=0;while(in[i])i++;
    s=r[i]->nearest(p[i]);
    if(p[i] > s->out)s->out=p[i]+EPS;
    else s->in=p[i]-EPS;
  }
}

void mi_grid::cleanup()
{
  int i,j,k;

  bad_incount1=0;
  bad_incount2=0;
  
  for(i=0;i<res;i++)
  for(j=0;j<res;j++)
  for(k=0;k<res;k++)
  {
    fixpoint(i,j,k);
  }

  printf("fixed: %d,%d\n",bad_incount1,bad_incount2);
}

mi_cube::mi_cube(mi_grid *g,int x,int y,int z)
{
  X0=2.0*double(x)/(g->res-1)-1;
  Y0=2.0*double(y)/(g->res-1)-1;
  Z0=2.0*double(z)/(g->res-1)-1;
  X1=2.0*double(x+1)/(g->res-1)-1;
  Y1=2.0*double(y+1)/(g->res-1)-1;
  Z1=2.0*double(z+1)/(g->res-1)-1;
  double *p;
  static int edgeorder[12]={8,9,10,11,2,3,6,7,0,1,4,5};
  static int vertordery[8]={0,1,4,5,2,3,6,7};
  static int vertorderz[8]={0,2,4,6,1,3,5,7};


  intersect[0][0]=X0;         intersect[0][1]=Y0;        intersect[0][2]=(Z0+Z1)/2;
  intersect[1][0]=X0;         intersect[1][1]=Y1;        intersect[1][2]=(Z0+Z1)/2;
  intersect[2][0]=X0;         intersect[2][1]=(Y0+Y1)/2; intersect[2][2]=Z0;
  intersect[3][0]=X0;         intersect[3][1]=(Y0+Y1)/2; intersect[3][2]=Z1;
  intersect[4][0]=X1;         intersect[4][1]=Y0;        intersect[4][2]=(Z0+Z1)/2;
  intersect[5][0]=X1;         intersect[5][1]=Y1;        intersect[5][2]=(Z0+Z1)/2;
  intersect[6][0]=X1;         intersect[6][1]=(Y0+Y1)/2; intersect[6][2]=Z0;
  intersect[7][0]=X1;         intersect[7][1]=(Y0+Y1)/2; intersect[7][2]=Z1;
  intersect[8][0]=(X0+X1)/2;  intersect[8][1]=Y0;        intersect[8][2]=Z0;
  intersect[9][0]=(X0+X1)/2;  intersect[9][1]=Y0;        intersect[9][2]=Z1;
  intersect[10][0]=(X0+X1)/2; intersect[10][1]=Y1;       intersect[10][2]=Z0;
  intersect[11][0]=(X0+X1)/2; intersect[11][1]=Y1;       intersect[11][2]=Z1;

  mi_ray *r;
  int i,j;

  for(i=0;i<2;i++)
  for(j=0;j<2;j++)
  {
    if(i+y<0 || j+z<0)
    {
      in[i*2+j]=0;
      in[i*2+j+4]=0;
    }
    else
    {
      r=&(g->rays[0][(i+y)*g->res+j+z]);
      
      in[i*2+j]=r->inside(X0);
      in[i*2+j+4]=r->inside(X1);

      if(in[i*2+j]!=in[i*2+j+4])
      {
	p=intersect[edgeorder[i*2+j]];
	p[0]=r->intersect_point(X0,X1);
	p[1]=r->origin[0];
	p[2]=r->origin[1];
      }  
    }

  }

  for(i=0;i<2;i++)
  for(j=0;j<2;j++)
  {
    if(i+x>=0 && j+z>=0)
    {
      r=&(g->rays[1][(i+x)*g->res+j+z]);
     
      if(in[vertordery[i*2+j]]!=in[vertordery[i*2+j+4]])
      {
        p=intersect[edgeorder[i*2+j+4]];
        p[0]=r->origin[0];
        p[1]=r->intersect_point(Y0,Y1);
        p[2]=r->origin[1];
      }  
    }
  }

  for(i=0;i<2;i++)
  for(j=0;j<2;j++)
  {
    if(i+x>=0 && j+y>=0)
    {
      r=&(g->rays[2][(i+x)*g->res+j+y]);
      
      if(in[vertorderz[i*2+j]]!=in[vertorderz[i*2+j+4]])
      {
	p=intersect[edgeorder[i*2+j+8]];
	p[0]=r->origin[0];
	p[1]=r->origin[1];
	p[2]=r->intersect_point(Z0,Z1);
      }  
    }
  }
}

static list<list<int> > cubetable[256];

void mi_cube::render(void triangle(double *,double *,double *))
{
  int index=0;

  for(int i=0;i<8;i++)if(in[i])index+=1<<i;
  list<list<int> > polys=cubetable[index];

  for(list<list<int> >::iterator pi=polys.begin();pi!=polys.end();pi++)
  {
    int a = -1, b = -1, count = 0;
    for(list<int>::iterator vi=pi->begin();vi!=pi->end();vi++)
    {
      int c = *vi;
      if (++count > 2) triangle(intersect[a],intersect[b],intersect[c]);
      if (count < 3) a = b;
      b = c;
    }
  }
}

//cube faces:
#define L	0  /* left direction:	-x, -i */
#define R	1  /* right direction:	+x, +i */
#define B	2  /* bottom direction: -y, -j */
#define T	3  /* top direction:	+y, +j */
#define N	4  /* near direction:	-z, -k */
#define F	5  /* far direction:	+z, +k */
//cube verts
#define LBN	0  /* left bottom near corner  */
#define LBF	1  /* left bottom far corner   */
#define LTN	2  /* left top near corner     */
#define LTF	3  /* left top far corner      */
#define RBN	4  /* right bottom near corner */
#define RBF	5  /* right bottom far corner  */
#define RTN	6  /* right top near corner    */
#define RTF	7  /* right top far corner     */
//cube edges
#define LB	0  /* left bottom edge	*/
#define LT	1  /* left top edge	*/
#define LN	2  /* left near edge	*/
#define LF	3  /* left far edge	*/
#define RB	4  /* right bottom edge */
#define RT	5  /* right top edge	*/
#define RN	6  /* right near edge	*/
#define RF	7  /* right far edge	*/
#define BN	8  /* bottom near edge	*/
#define BF	9  /* bottom far edge	*/
#define TN	10 /* top near edge	*/
#define TF	11 /* top far edge	*/

#define BIT(i, bit) (((i)>>(bit))&1)

/*			edge: LB, LT, LN, LF, RB, RT, RN, RF, BN, BF, TN, TF */
static int corner1[12]	   = {LBN,LTN,LBN,LBF,RBN,RTN,RBN,RBF,LBN,LBF,LTN,LTF};
static int corner2[12]	   = {LBF,LTF,LTN,LTF,RBF,RTF,RTN,RTF,RBN,RBF,RTN,RTF};
static int leftface[12]	   = {B,  L,  L,  F,  R,  T,  N,  R,  N,  B,  T,  F};
			     /* face on left when going corner1 to corner2 */
static int rightface[12]   = {L,  T,  N,  L,  B,  R,  R,  F,  B,  F,  N,  T};
			     /* face on right when going corner1 to corner2 */

int nextcwedge (int edge,int face)
{
    switch (edge) {
	case LB: return (face == L)? LF : BN;
	case LT: return (face == L)? LN : TF;
	case LN: return (face == L)? LB : TN;
	case LF: return (face == L)? LT : BF;
	case RB: return (face == R)? RN : BF;
	case RT: return (face == R)? RF : TN;
	case RN: return (face == R)? RT : BN;
	case RF: return (face == R)? RB : TF;
	case BN: return (face == B)? RB : LN;
	case BF: return (face == B)? LB : RF;
	case TN: return (face == T)? LT : RN;
	case TF: return (face == T)? RT : LF;
	default: return 0;
    }
}

/* otherface: return face adjoining edge that is not the given face */

int otherface (int edge, int face)
{
    int other = leftface[edge];
    return face == other? rightface[edge] : other;
}

/* makecubetable: create the 256 entry table for cubical polygonization */

void makecubetable ()
{
    int i, e, c, done[12], in[8];
    for (i = 0; i < 256; i++) {
	for (e = 0; e < 12; e++) done[e] = 0;
	for (c = 0; c < 8; c++) in[c] = BIT(i, c);
	for (e = 0; e < 12; e++)
	    if (!done[e] && (in[corner1[e]] != in[corner2[e]])) {
		list<int> poly;
		int start = e, edge = e;
		/* get face that is to right of edge from in to out corner: */
		int face = in[corner1[e]]? rightface[e] : leftface[e];
		while (1) {
		    edge = nextcwedge(edge, face);
		    done[edge] = 1;
		    if (in[corner1[edge]] != in[corner2[edge]]) {
			poly.push_front(edge);
			if (edge == start) break;
			face = otherface(edge, face);
		    }
		}
		cubetable[i].push_front(poly);
	    }
    }
}
