#include<stdio.h>
#include<math.h>
#ifdef WIN32
#include <GLee.h>
#endif
#include<GL/gl.h>
#include<GL/glut.h>
#include<stdarg.h>
#include<getopt.h>
#include<ext/hash_map>
//#include <hash_map>
#include"mi.h"
#include"../clb.h"

using namespace std;
using namespace ::__gnu_cxx;


int wwidth=640,wheight=480;
int mousebutton=0;
int mousepos[2];
mi_grid *mi;
double cam[4][4];
double sc[3]={1,1,1};
double tr[3]={0,0,0};
bitmap bm;
int render_mode=1;
char outfile[1024]="geom.obj";
char extfile[1024];
char imfile[1024];
char s[1024];
int nimages=0;

struct Vertex
{
  Vertex(double *v){x=v[0];y=v[1];z=v[2];};
  double x;
  double y;
  double z;
  size_t hash() const
  {
    return int((x+1)*100000)+int((y+1)*10000)+int((z+1)*1000);
  }
  bool operator ==(const Vertex &v)const
  {
    return fabs(x-v.x)+fabs(y-v.y)+fabs(z-v.z)<0.00000001;
  };
  bool operator <(const Vertex &v)const
  {
    if(v==*this) return 0;
    else return x+y+z<v.x+v.y+v.z;
  };
};

struct hashvertex{
  size_t operator()(const Vertex &v)const{return v.hash();};
};

struct comparevertex{
  bool operator()(const Vertex &v1,const Vertex &v2)const{return v1==v2;};
};



struct Triangle
{
  int a;
  int b;
  int c;
};

Vertex *vertex_array=NULL;
int vertex_array_size=0;
int vertex_count=0;
Triangle *polygon_array=NULL;
int polygon_array_size=0;
int polygon_count=0;
typedef hash_map<Vertex,int,hashvertex,comparevertex> vhash_map;
vhash_map vertex_table;

void triangle(double *a,double *b,double *c)
{
  Vertex A(a),B(b),C(c);
  Triangle t;
 
  if(vertex_count+3>vertex_array_size)
    vertex_array=(Vertex *)realloc(vertex_array,sizeof(Vertex)*(vertex_array_size+=100));

  if(polygon_count+1>polygon_array_size)
    polygon_array=(Triangle *)realloc(polygon_array,sizeof(Triangle)*(polygon_array_size+=100));

  vhash_map::iterator it;
  if((it=vertex_table.find(A))==vertex_table.end())
  {
    t.a=vertex_count++;
    vertex_table[A]=t.a;
    vertex_array[t.a]=A;
  }else t.a=it->second;
  if((it=vertex_table.find(B))==vertex_table.end())
  {
    t.b=vertex_count++;
    vertex_table[B]=t.b;
    vertex_array[t.b]=B;
  }
  else t.b=it->second;
  if((it=vertex_table.find(C))==vertex_table.end())
  {
    t.c=vertex_count++;
    vertex_table[C]=t.c;
    vertex_array[t.c]=C;
  }else t.c=it->second;

  polygon_array[polygon_count++]=t;
}



void processCommandLine(int ac,char ** av)
{
    static struct option long_opt[] = {
                   {"sx", 1, 0, 0},
                   {"sy", 1, 0, 0},
                   {"sz", 1, 0, 0},
                   {"tx", 1, 0, 0},
                   {"ty", 1, 0, 0},
		   {"tz", 1, 0, 0},
                   {"dir", 1, 0, 0},
		   {"file", 1, 0, 0},
		   {"ofile",1,0,0},
		   {"nrays", 1, 0, 0}, 
		   {"nimages", 1, 0, 0},
                   {0, 0, 0, 0}
               };
    int nrays=32;
    char dir[1024]=".";
    char file[256]="image%.4.tga";

    while(1){
	int opt_index=0;
	char c=getopt_long(ac,av,"",long_opt,&opt_index);
	printf("%d %d\n",c,opt_index);
	if(c==-1)break;
	switch(opt_index){
	    case 0:
	    case 1:
	    case 2:
		sscanf(optarg,"%lf",&(sc[opt_index]));
		break;
	    case 3:
	    case 4:
	    case 5:
		sscanf(optarg,"%lf",&(tr[opt_index-3]));
		break;	
	    case 6:
		snprintf(dir,1024,"%s",optarg);
		break;
	    case 7:
		snprintf(file,256,"%s",optarg);
		break;
	    case 8:
		snprintf(outfile,1024,"%s",optarg);
		break;
	    case 9:
		sscanf(optarg,"%d",&nrays);
		break;	
	    case 10:
		sscanf(optarg,"%d",&nimages);
		break;
	}
    }
    printf("%s/%s nrays=%d\n",dir,file,nrays);
    printf("Scale %f %f %f\n",sc[0],sc[0],sc[0]);
    printf("Trans %f %f %f\n",tr[0],tr[0],tr[0]);

    mi=new mi_grid(nrays);

    snprintf(extfile,1024,"%s/%s.clb",dir,file);
    snprintf(imfile,1024,"%s/%s",dir,file);
    printf("extfile %s\nimfile %s\n",extfile,imfile);
}


void display_lines()
{
  float col[3]={0,0,1};
  list<segment>::iterator s1;
  int i,j;

  glDisable(GL_LIGHTING);

  glBegin(GL_LINES);

  for(j=0;j<3;j++)
  {
      col[0]=0; col[1]=0;col[2]=0;col[j]=1;
      glColor3fv(col);
  for(i=mi->res/2;i<mi->res*mi->res;i+=mi->res) //draw only one slice
  for(i=0;i<mi->res*mi->res;i++)
  {
    double A[4];
    double B[4];
    A[3]=1;
    B[3]=1;
    
    s1=mi->rays[j][i].segments.begin();
    while(s1!=mi->rays[j][i].segments.end())
    {
      switch(mi->rays[j][i].axis)
      {
	case 0: A[0]=s1->in;
		A[1]=mi->rays[0][i].origin[0];
		A[2]=mi->rays[0][i].origin[1];
		B[0]=s1->out; B[1]=A[1]; B[2]=A[2];
		break;
	case 1: A[1]=s1->in;
		A[0]=mi->rays[1][i].origin[0];
		A[2]=mi->rays[1][i].origin[1];
		B[0]=A[0];
		B[1]=s1->out;
		B[2]=A[2];
		break;
	case 2: A[2]=s1->in;
		A[0]=mi->rays[2][i].origin[0];
		A[1]=mi->rays[2][i].origin[1];
		B[0]=A[0]; B[1]=A[1];
		B[2]=s1->out;
		break;
      }
      glVertex3dv(A);
      glVertex3dv(B);

      s1++;
    }
  }
  }
  
  glEnd();

}

void draw_triangle(double *a,double *b,double *c)
{
   glNormal3dv(a);
   glTexCoord2d((atan2(a[2],a[0])+M_PI)/(2*M_PI),a[1]/2.0+0.5);
   glVertex3dv(a);
   glNormal3dv(b);
   glTexCoord2d((atan2(b[2],b[0])+M_PI)/(2*M_PI),b[1]/2.0+0.5);
   glVertex3dv(b);
   glNormal3dv(c);
   glTexCoord2d((atan2(c[2],c[0])+M_PI)/(2*M_PI),c[1]/2.0+0.5);
   glVertex3dv(c);
}

void display_polys()
{
  glBegin(GL_TRIANGLES);
  for(int i=0;i<polygon_count;i++)
  {
    glVertex3dv((double *)&vertex_array[polygon_array[i].a]);
    glVertex3dv((double *)&vertex_array[polygon_array[i].b]);
    glVertex3dv((double *)&vertex_array[polygon_array[i].c]);
  }
  glEnd();
}

void writeobj()
{
  FILE *f=fopen(outfile,"w");
  if(!f)return;
  
  for(int i=0;i<vertex_count;i++)
    fprintf(f,"v %lf %lf %lf\n",vertex_array[i].x*0.5+0.5,vertex_array[i].y*0.5+0.5,vertex_array[i].z*0.5+0.5);
  for(int i=0;i<polygon_count;i++)
    fprintf(f,"f %d %d %d\n",polygon_array[i].a+1,polygon_array[i].b+1,polygon_array[i].c+1);
  fclose(f);
}

void display()
{
  glClearColor(0.5,0.5,0.5,1);
  glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
  
  glEnable(GL_DEPTH_TEST);

  //glDrawPixels(bm.w,bm.h,GL_LUMINANCE,GL_UNSIGNED_BYTE,bm.data);

  glCallList(render_mode+1);
  glutSwapBuffers();
}

void mouse(int x,int y)
{
  if(mousebutton)
  {
    glRotatef(5,mousepos[1]-y,mousepos[0]-x,0);
    glutPostRedisplay();
  }
  mousepos[0]=x;
  mousepos[1]=y;
}

void clicks(int button,int state,int x,int y)
{
  if(state==GLUT_DOWN)mousebutton=button+1;
  else
  {
    mousebutton=0;
  }
}

void keys(unsigned char key,int,int)
{
  switch(key)
  {
    case 'l': render_mode=(render_mode+1)%2;break;
  }
  glutPostRedisplay();
}

void reshape(int w,int h)
{
  wwidth=w;
  wheight=h;
  glViewport(0,0,w,h);
}


int main(int ac,char **av)
{
  char s[1024];
  bitmap tex;
  perspective(cam,1.3,2,1);

  processCommandLine(ac,av);

  printf("intersecting\n");
  makecubetable();

  double viewport_mat[4][4]={{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
  double bbox_mat[4][4]={{sc[0],0,0,tr[0]},{0,sc[1],0,tr[1]},{0,0,sc[2],tr[2]},{0,0,0,1}};
  double scale_bias_mat[4][4]={{0.5,0,0,0.5},{0,0.5,0,0.5},{0,0,0.5,0.5},{0,0,0,1}};
  double projection_mat[4][4]={{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,1,0}};

  for(int i=0;i<nimages;i++)
  {
    printf("message: Doing image %d/%d\n",i+1,nimages);
    printf("progress:%d\n",i);
    
   sprintf(s,imfile,i);
   if(!loadtga(&bm,s)){printf("opening %s failed!\n",s);exit(1);}
   sprintf(s,extfile,i);
   double intr[3][3];
   double extr[4][4];
   FILE *f=fopen(s,"rb");
   readCLB(f,intr[0],extr[0]);
   fclose(f);
   viewport_mat[0][3]=intr[0][2];
   viewport_mat[1][3]=intr[1][2];
   projection_mat[3][2]=2.0/(intr[0][0]+intr[1][1]);
   mat_mult(cam,projection_mat,extr);
   mat_mult(cam,viewport_mat,cam);
   mat_mult(cam,cam,bbox_mat);
   mat_mult(cam,cam,scale_bias_mat);
   
   threshold(&bm,5);
   printf(".");
   mi->intersect(bm,cam);
   sprintf(s,"/usr/scratch/keith/testoutput/im%.4d.pgm",i);
   savepgm(&bm,s);
   delete(bm.data);
   fflush(stdout);
  }
  mi->cleanup();
  mi->render(triangle);
  writeobj();
  //exit otherwise
#ifndef KSFS_MAIN
  exit(0);
#endif

  //savepgm(&bm,"test.pgm");

  glutInit(&ac, av);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize(wwidth,wheight);
  glutCreateWindow("demo");
  glutDisplayFunc(display);
 // glutIdleFunc(display);
  glutMotionFunc(mouse);
  glutPassiveMotionFunc(mouse);
  glutMouseFunc(clicks);
  glutKeyboardFunc(keys);
  glutReshapeFunc(reshape);
  glEnable(GL_DEPTH_TEST);
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  glPointSize(3);
  glLineWidth(3);
//  glEnable(GL_LIGHTING);
//  glEnable(GL_LIGHT0);
//  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
  glLineWidth(1);
//  glEnable(GL_CULL_FACE);

  glMatrixMode(GL_PROJECTION);
  gluPerspective(15,4.0/3.0,1,1000);
  glTranslatef(0,0,-10);
  glMatrixMode(GL_MODELVIEW);

  loadpgm(&tex,"/compsci/gibbons4/guest/keith/demos/capsys/ksfs/tex.pgm");
  
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,1);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  glTexImage2D(GL_TEXTURE_2D,0,GL_LUMINANCE,256,256,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,tex.data);

  glNewList(1,GL_COMPILE);
  display_lines();
  glEndList();
  glNewList(2,GL_COMPILE);
  display_polys();
  glEndList();


  glutMainLoop();
  
  return 0;
}
