#include "matrix.h"
#include "vec.h"

int Matrix::gcount = 0;
int Matrix::acount = 0;


Matrix Matrix::matrix2rod(){
  if(m!=3||n!=3){
    printf("error in matrix2rod\n");
    return Matrix(0,0);
  }
  Matrix axis(3,1);
  double angle;
  double ct;
  double st;
  double t1;
  double t2;
  double t3;
  ct=(get(0,0)+get(1,1)+get(2,2)-1)/2;
  t1=get(2,1)-get(1,2);
  t2=get(0,2)-get(2,0);
  t3=get(1,0)-get(0,1);
  st=0.5*::sqrt(t1*t1+t2*t2+t3*t3);

  angle=atan2(st,ct);
  double temp=1.0/(2*st);
  axis[0]=temp*t1;
  axis[1]=temp*t2;
  axis[2]=temp*t3;
    
  axis*=angle;
    
#ifdef PD
  axis.printMatlab("axis");
  printMatlab("matrix");
#endif
  return axis;
}

Matrix Matrix::rod2matrix(){
#ifdef MATRIX_CHECK
  if(m!=3||n!=1){
    printf("error \n");
  }
#endif
  Vec3d axis(get(0,0),get(1,0),get(2,0));
  double angle = axis.normalize();
  Matrix mat = Matrix::eye(3);
  Matrix S(3,3);
  S(0,0)=0;
  S(0,1)=-axis.z;
  S(0,2)=axis.y;
  S(1,0)=axis.z;
  S(1,1)=0;
  S(1,2)=-axis.x;
  S(2,0)=-axis.y;
  S(2,1)=axis.x;
  S(2,2)=0;
  Matrix S2=S*S;
  S*=(sin(angle));
  S2*=(1-cos(angle));
  
  mat+=S;
  mat+=S2;
  return mat;
}

//Implmentation from the matlab rot2yxz
Matrix Matrix::rot2zxy(){
  double ax,ay,az;
  double sy,cy;
  double eps=10e-16;
  if(fabs((*this)(2,0)<eps) && fabs((*this)(2,2))<eps){
    ay = 0;
    az = atan2((*this)(1,0),(*this)(0,0));
    ax = atan2((*this)(2,1),(*this)(2,2));
  }
  else{
    ay = atan2(-(*this)(2,0),(*this)(2,2));
    sy = sin(ay);
    cy = cos(ay);
    ax = atan2((*this)(2,1),-sy*(*this)(2,0)+cy*(*this)(2,2));
    az = atan2(cy*(*this)(1,0)+sy*(*this)(1,2),cy*(*this)(0,0)+sy*(*this)(0,2));
  }
  Matrix ret(3,1);
  ret[0]=az;ret[1]=ax;ret[2]=ay;
  return ret;
}
Matrix Matrix::rot2xyz(){
  if(m!=3 || n!=3){
    return Matrix(0,0);
  }
  Matrix xyz(3,1);
  double thetaX,thetaZ,thetaY = asin(get(0,2));
  if(thetaY < M_PI/2){
    if(thetaY > -M_PI/2){
      thetaX=atan2(-get(1,2),get(2,2));
      thetaZ=atan2(-get(0,1),get(0,0));
    }
    else{
      thetaX=-atan2(get(1,0),get(1,1));
      thetaZ=0;
    }
  }
  else{
    thetaX=atan2(get(1,0),get(1,1));
    thetaZ=0;
  }
  xyz[0]=thetaX;
  xyz[1]=thetaY;
  xyz[2]=thetaZ;
  return xyz;
}
//Numerical stuff
int Matrix::eig(Matrix & eigVectors,Matrix & eigenValues,int toget)
{
  Matrix A=copy();
  eigVectors=Matrix::eye(m);
  int i=0;
  double lastEig;
  Matrix ident;
  eigenValues=Matrix(m,1);
  int eigi=m-1;
  int size=m;
  int got=0;
    
  //Vector qr=QR();
    
  ident = Matrix::eye(size);
  i=0;
    
  while(size>=1){
    i=0;
    lastEig = A(size-1,size-1);
      
    while(i<40){
      Matrix p1;
      Matrix p2;

      ident*=A(size-1,size-1);
      A-=ident;
      A.QR(p1,p2);
	
      eigVectors=eigVectors*p1;
	
      A=p2*p1;
      A+=ident;
	
      int j=0;
      while(j<ident.m){
	ident(j,j)=1;
	j++;
      }
      if(fabs(lastEig-A(size-1,size-1))<0.000000001){
	eigenValues(eigi,0)=A(size-1,size-1);
	eigi--;
	size--;
	j=0;
	while(j<A.m){
	  A(j,size)=0;
	  A(size,j)=0;
	  j++;
	}
	break;
      }
      lastEig = A(size-1,size-1);
      i++;
    }
    if(i==40){
      printf("Matrix::eig() possible failure to converge (%s,%d)\n",
	     __FILE__,__LINE__);
      return 0;
    }
    got++;
    if(got==toget)break;
  }
  return 1;
}

void Matrix::Lsvd(Matrix & a,Matrix & w,Matrix & v){
  Matrix t=transpose();
  w=Matrix(MIN(m,n),1);
  Matrix U(m,m);
  v=Matrix(n,n);
  int worksize=2*MAX(3*MIN(m,n)+MAX(m,n),5*MIN(m,n));
  Matrix work(worksize,1);
  int info=0;
  dgesvd_('A','A',m,n,t.data,m,w.data,U.data,U.m,v.data,v.m,
	  work.data,worksize,info);
  if(info!=0)fprintf(stderr,"error in Lsvd (returned %d)\n",info);
  a=U.transpose();
}
void Matrix::LsvdLarge(Matrix & a,Matrix & w,Matrix & v){
  Matrix t=transpose();
  w=Matrix(MIN(m,n),1);
  Matrix U(m,m);
  v=Matrix(n,n);
  int worksize=MIN(m,n)*MIN(m,n) +
    MAX(MAX(m,n),4*MAX(m,n)*MAX(m,n)+4*MAX(m,n));
  int * iwork=new int[8*MIN(m,n)];
  worksize*=2;
  Matrix work(worksize,1);
  int info=0;
  dgesdd_('A',m,n,t.data,m,w.data,U.data,U.m,v.data,v.m,work.data,worksize,iwork,info);
  if(info!=0)fprintf(stderr,"error in LsvdLarge (returned %d)\n",info);
  a=U.transpose();
  delete [] iwork;
}

//numerical recipes in c
void Matrix::svd(Matrix &a,Matrix &w,Matrix &v){
  int flag,i,its,j,jj,k,l,nm;
  double anorm,c,f,g,h,s,scale,x,y,z;
  double *rv1;
  rv1 = new double[n];
  g=scale=anorm=0;

  //printf("doing svd\n");
  a=copy();
  w=Matrix(n,1);
  v=Matrix(n,n);

  for(i=0;i<n;i++){
    l=i+1;
    rv1[i]=scale*g;
    g=s=scale=0.0;
    if(i<m){
      for(k=i;k<m;k++)scale+=fabs(get(k,i));
      if(scale){
	for(k=i;k<m;k++){
	  set(k,i,get(k,i)/scale);
	  s+=get(k,i)*get(k,i);
	}
	f=get(i,i);
	g=-SIGN(::sqrt(s),f);
	h=f*g-s;
	set(i,i,f-g);
	for(j=l;j<n;j++){
	  for(s=0.0,k=i;k<m;k++)s+=get(k,i)*get(k,j);
	  f=s/h;
	  for(k=i;k<m;k++)set(k,j,get(k,j)+f*get(k,i));
	}
	for(k=i;k<m;k++)set(k,i,get(k,i)*scale);
      }
    }
    w.set(i,0,scale*g);
    g=s=scale=0.0;
    if(i<m&&i!=(n-1)){
      for(k=l;k<n;k++)scale+=fabs(get(i,k));
      if(scale){
	for(k=l;k<n;k++){
	  set(i,k,get(i,k)/scale);
	  double temp = get(i,k);
	  s+=temp*temp;
	}
	f=get(i,l);
	g=-SIGN(::sqrt(s),f);
	h=f*g-s;
	set(i,l,f-g);
	for(k=l;k<n;k++)rv1[k]=get(i,k)/h;
	for(j=l;j<m;j++){
	  for(s=0.0,k=l;k<n;k++)s+=get(j,k)*get(i,k);
	  for(k=l;k<n;k++)set(j,k,get(j,k)+s*rv1[k]);
	}
	for(k=l;k<n;k++)set(i,k,get(i,k)*scale);
      }
    }
    anorm=MAX(anorm,(fabs((w).get(i,0))+fabs(rv1[i])));
  }
  for(i=n-1;i>=0;i--){
    if(i<(n-1)){
      if(g){
	for(j=l;j<n;j++)
	  v.set(j,i,(get(i,j)/get(i,l))/g);
	for(j=l;j<n;j++){
	  for(s=0.0,k=l;k<n;k++)s+=get(i,k)*(v).get(k,j);
	  for(k=l;k<n;k++)(v).set(k,j,(v).get(k,j)+s*(v).get(k,i));
	}
      }
      for(j=l;j<n;j++){v.set(i,j,0.0);v.set(j,i,0.0);}
    }
    v.set(i,i,1.0);
    g=rv1[i];
    l=i;
  }
  for(i=MIN(m,n)-1;i>=0;i--){
    l=i+1;
    g=w.get(i,0);
    for(j=l;j<n;j++)set(i,j,0.0);
    if(g){
      g=1.0/g;
      for(j=l;j<n;j++){
	for(s=0.0,k=l;k<m;k++)s+=get(k,i)*get(k,j);
	f=(s/get(i,i))*g;
	for(k=i;k<m;k++)set(k,j,get(k,j)+get(k,i)*f);
      }
      for(j=i;j<m;j++)set(j,i,get(j,i)*g);
    }
    else for(j=i;j<m;j++)set(j,i,0.0);
    set(i,i,get(i,i)+1);
  }
  for(k=n-1;k>=0;k--){
    for(its=1;its<=30;its++){
      flag=1;
      for(l=k;l>=0;l--){
	nm=l-1;//for some dome reason this is okay, because 
	//rv1[0]==0
	if((double)(fabs(rv1[l])+anorm) == anorm){
	  flag=0;
	  break;
	}
	if((double)(fabs(w.get(nm,0))+anorm) ==  anorm)break;
      }
      if(flag){
	c=0.0;
	s=1.0;
	for(i=l;i<=k;i++){
	  f=s*rv1[i];
	  rv1[i]*=c;
	  if((double)(fabs(f)+anorm) == anorm) break;
	  g=w.get(i,0);
	  h=pythag(f,g);
	  w.set(i,0,h);
	  h=1.0/h;
	  c=g*h;
	  s=-f*h;
	  for(j=0;j<m;j++){
	    y=get(j,nm);
	    z=get(j,i);
	    set(j,nm,y*c+z*s);
	    set(j,i,z*c-y*s);
	  }
	}
      }
      z=w.get(k,0);
      if(l==k){
	if(z<0.0){
	  w.set(k,0,-z);
	  for(j=0;j<n;j++)v.set(j,k,-v.get(j,k));
	}
	break;
      }
      if(its==30){printf("no convergence in svd\n");return;}
      x=w.get(l,0);
      nm=k-1;
      y=w.get(nm,0);
      g=rv1[nm];
      h=rv1[k];
      f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
      g=pythag(f,1.0);
      f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
      c=s=1.0;
      for(j=l;j<=nm;j++){ //here
	i=j+1;
	g=rv1[i];
	y=w.get(i,0);
	h=s*g;
	g=c*g;
	z=pythag(f,h);
	rv1[j]=z;
	c=f/z;
	s=h/z;
	f=x*c+g*s;
	g=g*c-x*s;
	h=y*s;
	y*=c;
	for(jj=0;jj<n;jj++){
	  x=v.get(jj,j);
	  z=v.get(jj,i);
	  v.set(jj,j,x*c+z*s);
	  v.set(jj,i,z*c-x*s);
	}
	z=pythag(f,h);
	w.set(j,0,z);
	if(z){
	  z=1.0/z;
	  c=f*z;
	  s=h*z;
	}
	f=c*g+s*y;
	x=c*y-s*g;
	for(jj=0;jj<m;jj++){
	  y=get(jj,j);
	  z=get(jj,i);
	  set(jj,j,y*c+z*s);
	  set(jj,i,z*c-y*s);
	}
      }
      rv1[l]=0.0;
      rv1[k]=f;
      w.set(k,0,x);
    }
  }
  double * temp=a.data;
  a.data=data;
  data=temp;

  for(i=0;i<w.m-1;i++){
    int big=i;
    for(j=i+1;j<w.m;j++){
      if(w.get(big,0)<w.get(j,0)){
	big=j;
      }
    }
    double temp = w.get(big,0);
    w.set(big,0,w.get(i,0));
    w.set(i,0,temp);
    for(j=0;j<v.n;j++){
      temp=v.get(j,big);
      v.set(j,big,v.get(j,i));
      v.set(j,i,temp);
    }
    for(j=0;j<a.m;j++){
      temp=a.get(j,big);
      a.set(j,big,a.get(j,i));
      a.set(j,i,temp);
    }
      
  }
  //printf("done svd cond:%30.18f\n",(w.get(0,0))/(w.get(w.m-1,0)));
  delete rv1;
}
//Lapack qr decomposition
//returns 1 on success
int Matrix::Lqr(Matrix & Q,Matrix & R){
  Matrix T=this->transpose();//Make column major for lapack
  int info=0;
  Matrix tau(MIN(m,n),1);
  Matrix work(n,1);
  dgeqr2_(m,n,T.data,m,tau.data,work.data,info);
  if(info<0){
    printf("argument %d to dgeqr2 was bad\n",-info);
    return 0;
  }
  T=T.transpose();//Undo the column major for me
  Q=Matrix::eye(m,m);
  R=Matrix(m,n);
  R.setAll(0);
  for(int i=0;i<MIN(m,n);i++)
    for(int j=i;j<n;j++)
      R(i,j)=T(i,j);
  //Unpack Q, its a product of reflectors (according to lapack)
  Matrix v(Q.m,1);
  for(int i=0;i<MIN(m,n);i++){
    for(int j=0;j<i;j++)
      v[j]=0;
    v[i]=1;
    for(int j=i+1;j<Q.m;j++)
      v[j]=T(j,i);
    Matrix H=Matrix::eye(Q.m,Q.m);
    H-=(v*v.transpose()*tau[i]);
    Q=Q*H;
  }
  return info==0;
}
double Matrix::pythag(double a,double b){
  double absa,absb;
  absa=fabs(a);
  absb=fabs(b);
  if(absa>absb){
    double temp = absb/absa;
    return absa*::sqrt(1.0+temp*temp);
  }
  else{
    double temp =  absa/absb;
    return (absb==0.0?0.0:absb*::sqrt(1.0+temp*temp));
  }
}
    
int Matrix::QR(Matrix &Qt,Matrix &A){
  int j=0;
  int i=0;
  int k=0;
      
  A = copy();

  Matrix v = Matrix(A.m,1);
  double alpha;
  double beta;
  double sigma;
  Matrix Q = Matrix::eye(m);
  Matrix a;
    
  while(k<A.n){
    i=0;
    alpha=0;
    v.setAll(0);
    i=k;
    while(i<A.m){
      alpha+=pow(A.get(i,k),2);
      v.set(i,0,A.get(i,k));
      i++;
    }
	
    alpha = pow(alpha,0.5);
    if(v(k,0)>=0)
      alpha = -1*alpha;
	
    v(k,0)=v(k,0)-1*alpha;
	
    i=k;
    beta=v.dot(v);
	
    if(beta==0){
      k++;
      continue;
    }
	
    j=k;
    while(j<A.n){
      a=A.getColumn(j);
      sigma=v.dot(a);
      i=0;
      while(i<A.m){
	A.set(i,j,A(i,j)-((2*sigma)*v(i,0))/beta);
	i++;
      }
      j++;
    }
    j=0;
    while(j<Q.n){
      a=Q.getColumn(j);
      sigma=v.dot(a);
      i=0;
      while(i<A.m){
	Q(i,j)=Q(i,j)-((2*sigma)*v(i,0))/beta;
	i++;
      }
      j++;
    }
    k++;
  }
  Qt = Q.transpose();
  return 1;
}

double Matrix::det(){
  if(m!=n){
    printf("can you get det of non-square matrix\n");
    return 0;
  }
  if(m==1||n==1)return data[0];
  //expand about top row
  double ret=0;
  int sign=1;
  Matrix subMatrix(m-1,n-1);
    
  for(int j=0;j<n;j++){
    for(int i=1;i<m;i++){
      int ki=0;
      for(int k=0;k<n;k++){
	if(k==j)
	  continue;
	subMatrix(i-1,ki)=get(i,k);
	ki++;
      }
    }
    ret+=get(0,j)*sign*subMatrix.det();
    sign*=-1;
  }
  return ret;
}

Matrix Matrix::inverse() const {
  if(m!=n){
    printf("trying to invert non-square matrix\n");
    return Matrix();
  }
  int i, j, k;
    
  Matrix cop = copy();
  Matrix inv = Matrix::eye(m);
  for(j=0;j<m;j++){
    double largest=0;
    int largei=0;
    for(i=j;i<m;i++){
      if(fabs(cop(i,j))>largest){
	//printf("setting %f %f\n",largest,cop->get(i,j));
	largei=i;
	largest=fabs(cop(i,j));
      }
    }
    // printf("largest: %f\n",largest);
    //pivot
    for(k=0;k<m;k++){
      double temp=cop(j,k);
      cop(j,k)=cop(largei,k);
      cop(largei,k)=temp;
	
      temp=inv(j,k);
      inv(j,k)=inv(largei,k);
      inv(largei,k)=temp;
    }
      
    double ival=1.0/cop(j,j);
    for(k=0;k<m;k++){
      cop(j,k)=cop(j,k)*ival;
      inv(j,k)=inv(j,k)*ival;
      //data[j+4*k]*=ival;
      //inv[j+4*k]*=ival;
    }
    for(i=0;i<m;i++){
      if(i!=j){
	double scale=cop(i,j);
	for(k=0;k<m;k++){
	  cop(i,k)=cop(i,k)-scale*cop(j,k);
	  inv(i,k)=inv(i,k)-scale*inv(j,k);
	}
      }
    }
  }
  return inv;
}


void Matrix::print(){
  for(int i=0;i<m;i++){
    for(int j=0;j<n;j++){
      printf("%f ",(*this)(i,j));
    }
    printf("\n");
  }
}




#ifdef TEST_MATRIX
int main()
{
  Matrix A(2,2);
  A(0,0)=2;
  A(1,1)=1;
  A.print();
  
  Matrix B=A,C;
  B=A;
  B=A;
  printf("before\n");
  A.printMatlab("A");
  B.printMatlab("B");
  B=B+A;
  A.printMatlab("A");
  B.printMatlab("B");
  B=A*B+A*B+B.transpose()*A.transpose();
  B.printMatlab("B");

  Matrix D(2,2);
  D(0,0)=2;
  D(1,0)=1;
  D(0,1)=1;
  D(1,1)=1;
  printf("About to do svd\n");
  D.svd(A,B,C);
  A.print();
  B.print();
  C.print();
  D.det();
  printf("done svd\n");

  Matrix R=Matrix::random(3,3);
  
  R.printMatlab("R");
  printf("Determinant %f\n",R.det());

}

#endif
