#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#ifdef __PY_PLUS_PLUS
//Should expose sstream, iostream 
#include <sstream>
#include <iostream>
#include <boost/python/tuple.hpp>
#include <boost/python/extract.hpp>
using boost::python::tuple;
using boost::python::extract;
#endif

#include "dgesvd.h"
#include "dgesdd.h"
#include "dgelss.h"
#include "dgglse.h"
#include "dgeqr2.h"
#include "lmdif.h" //from minpack
#include "lmder.h" //from minpack

#ifndef MAX
#define MAX(a,b)((a)>(b)?(a):(b))
#endif

#ifndef MIN
#define MIN(a,b)((a)<(b)?(a):(b))
#endif

#define SIGN(a,b)((b)>=0 ? fabs(a):-fabs(a))

//#define BOUNDS_CHECK


#ifndef MATRIX_H
#define MATRIX_H

class Matrix{
 public:
  int m;
  int n;
  int * refcount;
  static int acount;
  static int gcount;
  double * data;
  Matrix(){
    m=0;
    n=0;
    data=0;
    refcount=0;
    
    gcount++;
  }
  Matrix(const Matrix & mat){
    data=mat.data;
    m=mat.m;
    n=mat.n;
    refcount=mat.refcount;
    (*refcount)++;
    
    gcount++;
  }
  Matrix(int _m,int _n){
    m=_m;
    n=_n;
    data = new double[m*n];
    refcount=new int;
    *refcount=1;

    acount++;
    gcount++;
  }
  ~Matrix(){
    destroy();
  }

  void destroy(){
    gcount--;

    if(refcount){
      (*refcount)--;
#ifdef MAT_DEBUG
      printf("destructing, refcount is %d\n",*refcount);
#endif
      if(*refcount==0){
	acount--;
	delete [] data;
	delete refcount;
	data=0;
	refcount=0;
#ifdef MAT_DEBUG
	printf("\treclaiming space %d,%d\n",acount,gcount);
#endif
      }
    }
  }
#ifndef __PY_PLUS_PLUS
  void printMatlab(char * name,FILE * file=stdout,const char * fmt="%8.6f ") const{
    fprintf(file,"%s = [...\n",name);
    for(int i=0;i<m;i++){
      for(int j=0;j<n;j++){
	fprintf(file,fmt,get(i,j));
      }
      if(i!=m-1)
	fprintf(file,"\n");
    }
    fprintf(file,"];\n");
  }
#endif
  Matrix rod2matrix();
  Matrix matrix2rod();

  //Implmentation from the matlab rot2yxz
  Matrix rot2zxy();
  Matrix rot2xyz();
  void print();

  void setAll(double d){
    int all=m*n;
    for(int i=0;i<all;i++){
      data[i]=d;
    }
  }
  
  void dotmult(Matrix & other){
    for(int i=0;i<m*n;i++){
      data[i]*=other.data[i];
    }
  }
  void operator=(const Matrix & other){
    destroy();
    m=other.m;
    n=other.n;
    data=other.data;
    refcount=other.refcount;
    (*refcount)++;

    gcount++;
  }
  void operator*=(double d){
    int all=m*n;
    for(int i=0;i<all;i++){
      data[i]*=d;
    }
  }
  double dot(Matrix & other)
  {
    double res=0;
    if(other.m!=m || other.n!=n){
      printf("incompatible matrix dot product\n");
    }
    for(int i=0;i<m;i++){
      for(int j=0;j<n;j++){
	res+=(*this)(i,j)*other(i,j);
      }
    }
    return res;
  }
  double minInRow(int i) const{
    double min=get(i,0);
    for(int j=1;j<n;j++){
      if(get(i,j)<min)min=get(i,j);
    }
    return min;
  }
  double minInCol(int j) const{
    double min=get(0,j);
    for(int i=1;i<m;i++){
      if(get(i,j)<min)min=get(i,j);
    }
    return min;
  }
  double maxInRow(int i) const{
    double max=get(i,0);
    for(int j=1;j<n;j++){
      if(get(i,j)>max)max=get(i,j);
    }
    return max;
  }
  double maxInCol(int j) const{
    double max=get(0,j);
    for(int i=1;i<m;i++){
      if(get(i,j)>max)max=get(i,j);
    }
    return max;
  }
  void setSubmatrix(int i0,int j0,const Matrix & subm){
    for(int i=0;i<subm.m;i++){
      for(int j=0;j<subm.n;j++){
	set(i+i0,j+j0,subm.get(i,j));
      }
    }
  }
  Matrix submatrix(int i0,int j0,int nrows,int ncols) const{
    Matrix subm(nrows,ncols);
    for(int i=0;i<nrows;i++){
      for(int j=0;j<ncols;j++){
	subm(i,j)=get(i0+i,j0+j);
      }
    }
    return subm;
  }
  Matrix getRow(int i) const{
    Matrix row(1,n);
    for(int j=0;j<n;j++){
      row[j]=get(i,j);
    }
    return row;
  }
  
  Matrix getColumn(int j) const{
    Matrix col(m,1);
    for(int i=0;i<m;i++){
      col[i]=get(i,j);
    }
    return col;
  }
  Matrix getCols(const int * which,int ncols){
    Matrix cols(m,ncols);
    for(int i=0;i<m;i++){
      for(int j=0;j<ncols;j++){
	cols(i,j)=get(i,which[j]);
      }
    }
    return cols;
  }

  Matrix operator+(const Matrix & other){
    if(m!=other.m||n!=other.n){
      printf("incompatible matrix addition\n");
    }
    Matrix res(m,n);
    int top=m*n;
    for(int i=0;i<top;i++){
      res[i]=data[i]+other.get(i);
    }
    return res;
  }
  Matrix operator-(const Matrix & other){
    if(m!=other.m||n!=other.n){
      printf("incompatible matrix subtraction\n");
    }
    Matrix res(m,n);
    int top=m*n;
    for(int i=0;i<top;i++){
      res[i]=data[i]-other.get(i);
    }
    return res;
  }
  void operator-=(const Matrix & other){
    if(m!=other.m||n!=other.n){
      printf("incompatible matrix addition\n");
    }
    int top=m*n;
    for(int i=0;i<top;i++){
      data[i]-=other.get(i);
    }
  }
  void operator+=(const Matrix & other){
    if(m!=other.m||n!=other.n){
      printf("incompatible matrix addition\n");
    }
    int top=m*n;
    for(int i=0;i<top;i++){
      data[i]+=other.get(i);
    }
  }

  void operator+=(double d){
    int all=m*n;
    for(int i=0;i<all;i++){
      data[i]+=d;
    }
  }
  
  void set(int i,int j,double val){
#ifdef BOUNDS_CHECK
    if(i>=m || i<0 || j>=n || j<0){
      printf("set: [%d,%d] not in [%d,%d]\n",i,j,m,n);
      int x=0;
      x=1/x;
      exit(1);
    }
#endif
    data[i*n+j]=val;
  }
  double get(int i) const{
#ifdef BOUNDS_CHECK
    if(i>=m*n){
      printf("index out of bounds %d >= %dx%d=%d\n",m,n,m*n);
      int x=0;
      x=1/x;
      exit(1);
 
    }
#endif
    return data[i];
  }
  double get(int i,int j) const{
#ifdef BOUNDS_CHECK
    if(i>=m || i<0 || j>=n || j<0){
      printf("get: [%d,%d] not in [%d,%d]\n",i,j,m,n);
      int x=0;
      x=1/x;
      exit(1);
    }
#endif
    return data[i*n+j];
  }
  
  double & operator[](int i){
#ifdef BOUNDS_CHECK
    if(i>=m*n){
      printf("index out of bounds\n");
      int x=0;
      x=1/x;
      exit(1);
    }
#endif
    return data[i];
  }
  double & operator()(int i,int j){
#ifdef BOUNDS_CHECK
    if(i>=m || i<0 || j>=n || j<0){
      printf("get: [%d,%d] not in [%d,%d]\n",i,j,m,n);
      int x=0;
      x=1/x;
      exit(1);
    }
#endif
    return data[i*n+j];
  }
  
  int eig(Matrix & eigVectors,Matrix & eigenValues,int toget=-1);

  void Lsvd(Matrix & a,Matrix & w,Matrix & v);
  void LsvdLarge(Matrix & a,Matrix & w,Matrix & v);

  //numerical recipes in c
  void svd(Matrix &a,Matrix &w,Matrix &v);
  int Lqr(Matrix & Q,Matrix & R);
  double pythag(double a,double b);
  int QR(Matrix &Qt,Matrix &A);

  Matrix operator*(double d){
    int i=0;
    int top=m*n;
    Matrix res(m,n);
    for(i=0;i<top;i++){
      res[i]=data[i]*d;
    }
    return res;
  }
    
  Matrix operator*(const Matrix & other) const{
    int i=0;
    int j=0;
    int k=0;
    if(n!=other.m){
      printf("invalid sizes for matrix multiplication\n");
      return Matrix(0,0);
    }
    
    Matrix ret(m,other.n);
    
    for(i=0;i<ret.m;i++){
      for(j=0;j<ret.n;j++){
	double sum=0;
	for(k=0;k<n;k++){
	  sum+=(*this).get(i,k)*(other.get(k,j));
	}
	ret(i,j)=sum;
      }
    }
    return ret;
  }
  static Matrix random(int m,int n){
    Matrix ret(m,n);
    int top=m*n;
    for(int i=0;i<top;i++){
      ret.data[i]=((double)rand())/((double)RAND_MAX);
    }
    return ret;
  }

  static Matrix eye(int m,int n){
    Matrix ret(m,n);
    int min=MIN(m,n);
    ret.setAll(0);
    for(int i=0;i<min;i++){
      ret(i,i)=1;
    }
    return ret;
  }
  static Matrix eye(int size){
    Matrix ret(size,size);
    ret.setAll(0);
    for(int i=0;i<size;i++){
      ret(i,i)=1;
    }
    return ret;
  } 
  void sub(Matrix * mat){
    if(m!=mat->m || n!=mat->n){
      printf("cannot subtract matrices of different dims\n");
      return;
    }
    for(int i=0;i<m;i++){
      for(int j=0;j<n;j++){
	set(i,j,get(i,j)-mat->get(i,j));
      }
    }
  }
  Matrix transpose(){
    Matrix ret(n,m);
    for(int i=0;i<m;i++){
      for(int j=0;j<n;j++){
	ret(j,i)=get(i,j);
      }
    }
    return ret;
  }
  Matrix copy() const{
      Matrix c(m,n);
      memcpy(c.data,data,sizeof(double)*m*n);
      return c;
    }
  Matrix sqrt(){
    Matrix res(m,n);
    for(int i=0;i<m;i++){
      for(int j=0;j<n;j++){
	res(i,j)=::sqrt(operator()(i,j));
      }
    }
    return res;
  }
  static Matrix rot2D(double theta){
    Matrix rot=Matrix::eye(3);
    double c=cos(theta);
    double s=sin(theta);
    rot(0,0)=c;
    rot(0,1)=-s;
    rot(1,0)=s;
    rot(1,1)=c;
    return rot;
  }
  static Matrix trans2D(double x,double y){
    Matrix trans=Matrix::eye(3);
    trans(0,2)=x;
    trans(1,2)=y;
    return trans;
  }
  static Matrix rotx(double theta){
    Matrix rot=Matrix::eye(4);
    double c=cos(theta);
    double s=sin(theta);
    rot(1,1)=c;
    rot(1,2)=-s;
    rot(2,1)=s;
    rot(2,2)=c;
    return rot;
  }
  static Matrix roty(double theta){
    Matrix rot=Matrix::eye(4);
    double c=cos(theta);
    double s=sin(theta);
    rot(0,0)=c;
    rot(0,2)=s;
    rot(2,0)=-s;
    rot(2,2)=c;
    return rot;
  }
  static Matrix rotz(double theta){
    Matrix rot=Matrix::eye(4);
    double c=cos(theta);
    double s=sin(theta);
    rot(0,0)=c;
    rot(0,1)=-s;
    rot(1,0)=s;
    rot(1,1)=c;
    return rot;
  }
  static Matrix scale(double sx,double sy,double sz){
    Matrix sc=Matrix::eye(4,4);
    sc(0,0)=sx;
    sc(1,1)=sy;
    sc(2,2)=sz;
    return sc;
  }
  static Matrix trans(double x,double y,double z){
    Matrix trans=Matrix::eye(4);
    trans(0,3)=x;
    trans(1,3)=y;
    trans(2,3)=z;
    return trans;
  }
  /***********************************************************
   * Use minpack to do nonlinear least squares using leven-marq.
   * \param func the function to minimize (ie. must subtract rhs to make
   *       zeros.  if iflag is 1 set fvec (don't touch fjac), if iflag=2
   *       then update fjac (don't touch fvec).
   * \param x the initial parameter vector which you wish to optmiize
   * \param fvec the output of the functions
   * \param ftol,xtol,gtol tolerances for terminating
   * \param maxfev maximum number of calls to func
   * \param epsfcn the epsilon to use for calculating jacobian
   * \param diag an optional scaling parameter, if 0 then minpack picks scales
   * \param factor not sure, see minpack documentation
   * \param nprint not sure, see minpack docs.
   * \returns the solution in x, and the integer output from the call
   * to minpack. 0 if bad params, and other numbers for reason for 
   * termination
   *
   ***********************************************************/
  static  int  lsqnonlin(void (* func)(const int & m,
				       const int & n,
				       double * x,
				       double * fvec,
				       double * fjac,
				       const int & ldfjac,
				       int & iflag),
		 Matrix & x,Matrix & fvec,
		 double ftol=10e-7,double xtol=10e-7,double gtol=10e-7,
		 int maxfev=10000000,double * diag=0,
		 double factor=100,int nprint=0)
    {
      int mode=1;
      int n=x.m*x.n;      //Number of variables
      int m=fvec.m*fvec.n;//Number of functions
      int info=0;
      int nfev=0;
      int njev=0;
      Matrix fjac(m,n);
      int ldfjac=m;
      int * ipvt=new int[m*n];
      Matrix qtf(n,1);//output containing q'*fvec
      Matrix wa1(n,1);//work arrays
      Matrix wa2(n,1);
      Matrix wa3(n,1);
      Matrix wa4(m,1);//work array of length m
      
      // Allocate the diagonal scale factors if the user 
      // did not provide, and set the mode accordingly
      if(diag)mode=2;
      else diag=new double[n];
      
      lmder_(func,m,n,x.data,fvec.data,fjac.data,ldfjac,ftol,xtol,
	     gtol,maxfev,diag,mode,
	     factor,nprint,info,nfev,njev,
	     ipvt,qtf.data,wa1.data,wa2.data,wa3.data,wa4.data);
      
      if(mode==1)delete [] diag;
      delete [] ipvt;
      return info;
    }   
  /***********************************************************
   * Use minpack to do nonlinear least squares using leven-marq.
   * \param func the function to minimize (ie. must subtract rhs to make
   *       zeros
   * \param x the initial parameter vector which you wish to optmiize
   * \param fvec the output of the functions
   * \param ftol,xtol,gtol tolerances for terminating
   * \param maxfev maximum number of calls to func
   * \param epsfcn the epsilon to use for calculating jacobian
   * \param diag an optional scaling parameter, if 0 then minpack picks scales
   * \param factor not sure, see minpack documentation
   * \param nprint not sure, see minpack docs.
   * \returns the solution in x, and the integer output from the call
   * to minpack. 0 if bad params, and other numbers for reason for 
   * termination
   *
   ***********************************************************/
  static  int  lsqnonlin(void (* func)(const int & m,
			       const int & n,
			       double * x,
			       double * fvec,
			       int & iflag),
		 Matrix & x,Matrix & fvec,
		 double ftol=10e-7,double xtol=10e-7,double gtol=10e-7,
		 int maxfev=10000000,double epsfcn=10e-8,double * diag=0,
		 double factor=100,int nprint=0)
    {
      int mode=1;
      int n=x.m*x.n;      //Number of variables
      int m=fvec.m*fvec.n;//Number of functions
      int info=0;
      int nfev=0;
      Matrix fjac(m,n);
      int ldfjac=m;
      int * ipvt=new int[m*n];
      Matrix qtf(n,1);//output containing q'*fvec
      Matrix wa1(n,1);//work arrays
      Matrix wa2(n,1);
      Matrix wa3(n,1);
      Matrix wa4(m,1);//work array of length m
      
      // Allocate the diagonal scale factors if the user 
      // did not provide, and set the mode accordingly
      if(diag)mode=2;
      else diag=new double[n];
      
      lmdif_(func,m,n,x.data,fvec.data,ftol,xtol,
	     gtol,maxfev,epsfcn,diag,mode,
	     factor,nprint,info,nfev,fjac.data,ldfjac,
	     ipvt,qtf.data,wa1.data,wa2.data,wa3.data,wa4.data);
      
      if(mode==1)delete [] diag;
      delete [] ipvt;
      return info;
    }   
  static Matrix linLeastSq(Matrix &A,Matrix &b){
    Matrix U ,w, V;
    printf("doing svd\n");
    A.svd(U,w,V);
    printf("done svd\n");
    Matrix W(V.m,V.m);
    Matrix Ut=U.transpose();
 
    W.setAll(0);
    for(int i=0;i<W.m;i++){
      W(i,i)=1.0/w[i];
    }

    Matrix result=V*W*Ut*b;

        
    return result;
  }
  static Matrix LlinLeastSq(Matrix &A,Matrix &b,Matrix & B,Matrix & d){
    Matrix At=A.transpose();
    Matrix Bt=B.transpose();
    Matrix bt=b.transpose();//Just do so data isn't clobbered
    Matrix dt=d.transpose();//dido
    Matrix X(A.n,1);
    int lwork=-1;
    Matrix work(1,1);
    int info=0;
    dgglse_(A.m,A.n,B.m,At.data,A.m,Bt.data,
	    B.m,bt.data,dt.data,X.data,work.data,lwork,info);
    lwork=(int)(work[0]);
    work=Matrix(lwork,1);
    dgglse_(A.m,A.n,B.m,At.data,A.m,Bt.data,
	    B.m,bt.data,dt.data,X.data,work.data,lwork,info);
    return X;
  }
  static Matrix LlinLeastSq(Matrix &A,Matrix &b){
    Matrix s(MIN(A.m,A.n),1);
    int rank;
    int info;
    int lwork=2*(3*MIN(A.m,A.n)+MAX(2*MIN(A.m,A.n),MAX(MAX(A.m,A.n),b.n )));
    Matrix work(lwork,1);
    Matrix At=A.transpose();
    Matrix bt=b.transpose();
    dgelss_(A.m,A.n,b.n,At.data,A.m,bt.data,b.m,s.data,-1,rank,work.data,lwork,info);
    //dgelsx_(A.m,A.n,b.
    Matrix x(A.n,1);
    for(int i=0;i<x.m;i++){
      x[i]=bt[i];
    }
    return x;
  }
  double det();
  Matrix inverse() const;

#ifndef __PY_PLUS_PLUS
  int save(FILE * file){
    int dims[2];
    int wrote=0;
    dims[0]=m;
    dims[1]=n;
    if(2==fwrite(dims,sizeof(int),2,file)){
      wrote=fwrite(data,sizeof(double),dims[0]*dims[1],file);
      if(wrote==dims[0]*dims[1])return 1;
    }
    return 0;
  }
  int load(FILE * file){
    int dims[2];
    int red=0;
    if(2==fread(dims,sizeof(int),2,file)){
      operator=(Matrix(dims[0],dims[1]));
      red=fread(data,sizeof(double),dims[0]*dims[1],file);
      if(red==dims[0]*dims[1])return 1;
    }
    operator=(Matrix(0,0));
    return 0;
  }
#endif
  double meanAll(){
    double mean=0;
    for(int i=0;i<m*n;i++){
      mean+=data[i];
    }
    return mean/(m*n);
  }

#ifdef __PY_PLUS_PLUS
  void print(std::ostream & ost) const{
    for(int i=0;i<m;i++){
      for(int j=0;j<n;j++){
	ost << get(i,j);
	if(j<(n-1))ost << ",";
      }
      ost << std::endl;
    }
  }
  operator std::string () const{
    std::stringstream ss;
    print(ss);
    return ss.str();
  }
  void python_set_arr(int i,double d){
    operator[](i)=d;
  }
  void python_set(const tuple & t,double d){
    operator()(extract<int>(t[0]),extract<int>(t[1]))=d;
  }
  double & operator[](const tuple & t){
    return operator()(extract<int>(t[0]),extract<int>(t[1]));
  }
#endif
};

typedef Matrix Matrixd;

#endif
