#include "image.h"
#include "matrix.h"
#include <GL/glut.h>
extern "C"{
  #include <jpeglib.h>
}
template <class T>
Image<T>::Image(int _w,int _h,int _nchannels)
{
  data=0;
  refcount=0;
  width=_w;
  height=_h;
  nchannels=_nchannels;
  if(_w!=0&&_h!=0){
    refcount = new int[1];
    *refcount=1;
    data = new T[width*height*nchannels];
    //printf("allocing %p (%dx%dx%d)\n",data,width,height,nchannels);
  }
  else{
    //printf("null image\n");
  }
}

template <class T>
Image<T>::Image(const Image<T> & image)
{
  width=image.width;
  height=image.height;
  nchannels=image.nchannels;
  refcount=image.refcount;
  (*refcount)++;
  data=image.data;
}

template <class T>
void Image<T>::release(){
  if(refcount){
    (*refcount)--;
    if(*refcount==0){
      //printf("deleting %p (%dx%dx%d)\n",data,width,height,nchannels);
      delete refcount;
      delete [] data;
      data=0;
      refcount=0;
    }
  }
}

template <class T>
Image<T>::~Image(){
  release();
  width=0;
  height=0;
  nchannels=0;
}

template <class T>
Image<T> Image<T>::map(const Image<unsigned int> & indices){
  Image<T> image(w,h,nchannels);
  if(indices.w!=w || indices.h!=h){
    printf("Image::map inconsistent sizes for map %dx%d!=%dx%d\n",
	   w,h,indices.w,indices.h);
    return image;
  }
  for(int i=0;i<width*height;i++){
    for(int j=0;j<nchannels;j++){
      /*if(indices[i]>=width*height){
	printf("bad index in map %d>=%dx%d=%d\n",
	    indices[i],width*height,width,height);
	    }
	    else if(indices[i]<0){
	    printf("bad index in map %d<0\n",
		   indices[i]);
		   }*/
      image[nchannels*i+j]=(data[nchannels*indices[i]+j]);
    }
  }
  return image;
}

template <class T>
void Image<T>::checkSize(int resize,int * newWidth,int * newHeight){
  int w=0,h=0;
  while((1<<w)<width)
    w=w+1;
  while((1<<h)<height)
    h=h+1;
  
  if(newWidth){
    *newWidth=(1<<w);
    *newHeight=(1<<h);
  }
  if((1<<w)!=width ||
     (1<<h)!=height){
    if(resize){
      int format=channelToGL();
      
      Image<T> im2(1<<w,1<<h,nchannels);
      
      glPixelStorei(GL_UNPACK_ALIGNMENT,1);
      gluScaleImage(format,
		    width,height,
		    typeToGL(),
		    getDataPtr(),
		    1<<w,
		    1<<h,
		    typeToGL(),
		    im2.getDataPtr());
      //printf("resizing\n");
      operator=(im2);
    }    
  }
}

template <class T>
void Image<T>::initTexture(int mode){
  checkSize();
	
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,mode);
  
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  glTexImage2D(GL_TEXTURE_2D,0,channelToGL(),width,height,
	       0,channelToGL(),
	       typeToGL(),data);
}

template <class T>
void Image<T>::initTextureNoResize(double * sizex,double * sizey,
			 int mode){
  int neww,newh;
  checkSize(0,&neww,&newh);
  *sizex=(double)width/(double)neww;
  *sizey=(double)height/(double)newh;
  
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,mode);
  
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  
  Image<T> newIm(neww,newh,nchannels);
  glTexImage2D(GL_TEXTURE_2D,0,channelToGL(),neww,newh,
	       0,channelToGL(),
	       typeToGL(),newIm.data);
  glTexSubImage2D(GL_TEXTURE_2D,0,
		  0,0,width,height,
		  channelToGL(),typeToGL(),data);
}

int Image<unsigned char>::visit(Image<unsigned char> & vis,int x,int y,int on){
  int cnt=0;
  int * xs=new int[width*height];
  int * ys=new int[width*height];
  int head=0;
  int tail=0;
  xs[tail]=x;
  ys[tail++]=y;

      while(head!=tail){
	x=xs[head];
	y=ys[head];
	if(!vis(x,y)){
	  cnt++;
	  vis(x,y)=1;
	
	  if(x>0 && operator()(x-1,y,0)==on && !vis(x-1,y)){
	    xs[tail]=x-1;
	    ys[tail]=y;
	    tail++;
	    if(tail>width*height)
	      tail=0;
	  }
	  if(y>0 && operator()(x,y-1,0)==on && !vis(x,y-1)){
	    xs[tail]=x;
	    ys[tail]=y-1;
	    tail++;
	    if(tail>width*height)
	      tail=0;
	  }
	  if(x<width-1 && operator()(x+1,y,0)==on && !vis(x+1,y)){
	    xs[tail]=x+1;
	    ys[tail]=y;
	    tail++;
	    if(tail>width*height)
	      tail=0;
	  }
	  if(y<height-1 && operator()(x,y+1,0)==on && !vis(x,y+1)){
	    xs[tail]=x;
	    ys[tail]=y+1;
	    tail++;
	    if(tail>width*height)
	      tail=0;
	  }      
	}
	head++;
	if(head>width*height)
	  head=0;
	
      }
      delete xs;
      delete ys;
      return cnt;
    }

void Image<unsigned char>
::fill(int x,int y,int from,int to,int depth,int & howmany){
      if(depth>width)return;
      if(operator()(x,y,0)==from){
	operator()(x,y,0)=to;
	howmany--;
	if(howmany<=0)return;
	if(x>0){
	  fill(x-1,y,from,to,depth+1,howmany);
	}
	if(y>0){
	  fill(x,y-1,from,to,depth+1,howmany);
	}
	if(x<width-1){
	  fill(x+1,y,from,to,depth+1,howmany);
	}
	if(y<height-1){
	  fill(x,y+1,from,to,depth+1,howmany);
	}
      }
    }

void Image<unsigned char>::clean(int size,int on){
  Image<unsigned char> visited(width,height,1);
  visited=0;
  printf("cleaning....................\n");
  
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      if(operator()(x,y,0)==on && !visited(x,y)){
	int thissize=visit(visited,x,y,on);
	
	if(thissize<size){
	  fill(x,y,(int)operator()(x,y,0),!on,0,thissize);
	}
      }
    }
  }
}

template <class T>
void Image<T>::spur(){
      for(int y=1;y<height-1;y++){
	for(int x=1;x<width-1;x++){
	  int up=(int)operator()(x,y+1,0);
	  int down=(int)operator()(x,y-1,0);
	  int left=(int)operator()(x-1,y,0);
	  int right=(int)operator()(x+1,y,0);
	  if(left && right){
	    operator()(x,y,0)=1;
	  }
	  if(!left && !right){
	    operator()(x,y,0)=0;
	  }
	  if(up && down){
	    operator()(x,y,0)=1;
	  }
	  if(!up && !down){
	    operator()(x,y,0)=0;
	  }

	}
      }
    }

template <class T>
Image<T> Image<T>::getChannel(int which){
  if(nchannels==0)return Image<T>(0,0,0);
  Image<T> res(width,height,1);
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      res(x,y)
	=operator()(x,y,which);
    }
  }
  return res;
}
template <class T>
void Image<T>::setToChannel(int which){
  if(nchannels==0)return;
  
  for(int y=0;y<height;y++){
    for(int x=0;x<width;x++){
      for(int k=0;k<nchannels;k++){
	operator()(x,y,k)
	  =operator()(x,y,which);
      }
    }
  }
}

template <class T>
void Image<T>::operatorEquals(const Image & other){
  if(&other!=this){
    release();
    width=other.width;
    height=other.height;
    nchannels=other.nchannels;
    refcount=other.refcount;
    (*refcount)++;
    data=other.data;
  }
}

//overloaded operator= for each type
void Imagef::operator=(const Image8 & image){
    operatorEquals(Image<float>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(float)(image[i])/255.0f;
    }
}
void Imagef::operator=(const Imagef & image){
    operatorEquals(image);
}
void Imagef::operator=(const Image16 & image){
    operatorEquals(Image<float>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(float)(image[i])/65536.0f;
    }
}
void Imagef::operator=(const Image32 & image){
  operatorEquals(Image<float>(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(float)(image[i])/(float)(1<<31);
  }
}

void Image8::operator=(const Image8 & image){
    operatorEquals(image);
}
void Image8::operator=(const Imagef & image){
    operator=(Image<unsigned char>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(unsigned char)(image[i]*255.0f);
    }
}
void Image8::operator=(const Image16 & image){
    operator=(Image<unsigned char>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(unsigned char)(((double)image[i]*255.0)/65535.0);
    }
}

void Image8::operator=(const Image32 & image){
  operator=(Image<unsigned char>(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned char)(((double)image[i]*255.0)/(double)(1<<31));
  }
}

void Image16::operator=(const Image16 & image){
  operatorEquals(image);
}
void Image16::operator=(const Image8 & image){
    operator=(Image<unsigned short>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
      data[i]=(unsigned short)(image[i]<<8);
    }
}
void Image16::operator=(const Imagef & image){
    operator=(Image<unsigned short>(image.width,image.height,image.nchannels));
    for(int i=0;i<image.width*image.height*image.nchannels;i++){
	data[i]=(unsigned short)(image[i]*65535.0f);
    }
}
void Image16::operator=(const Image32 & image){
  operator=(Image<unsigned short>(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned short)(image[i]/65536);
  }
}

void Image32::operator=(const Image32 & image){
  operatorEquals(image);
}
void Image32::operator=(const Image8 & image){
  operator=(Image32(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned int)(image[i]<<24);
  }
}
void Image32::operator=(const Imagef & image){
  operator=(Image32(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned int)(image[i]*(1<<31));
  }
}
void Image32::operator=(const Image16 & image){  
  operator=(Image32(image.width,image.height,image.nchannels));
  for(int i=0;i<image.width*image.height*image.nchannels;i++){
    data[i]=(unsigned int)(image[i]<<16);
  }
}

template <class T>
int Image<T>::readTGA(char * filename){
  FILE * f = fopen(filename,"rb");
  if(!f)return 0;
  
  unsigned char bpp;  
  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  char tga_header_gray[12] = {0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
  char file_header[12];

  fread(file_header,sizeof(file_header),1,f);

  if (memcmp(file_header, tga_header_rgb, 12) &&
      memcmp(file_header, tga_header_gray, 12)) {
    fclose(f);
    printf("can not read TGA\n");
    return 0;
  }
  Image8 im;

  unsigned short w;
  unsigned short h;
  
  fread(&w,sizeof(w),1,f);
  fread(&h,sizeof(h),1,f);
  fread(&bpp,sizeof(bpp),1,f);
  bpp/=8;
  fgetc(f);
  im=Image8(w,h,bpp);
  int nread;
  if(w*h*bpp!=(nread=fread(im.data,1,w*h*bpp,f))){
    printf("nread is %d not %d\n",nread,w*h*bpp);
    return 0;
  }

  if(bpp==3 || bpp==4) // flip red and blue
    for(int i=0;i<w*h*bpp;i+=bpp)
    {
      char tmp=im.data[i];
      im.data[i]=im.data[i+2];
      im.data[i+2]=tmp;
    }
  operator=(im);
  //return im;
}

template <class T>
int Image<T>::readJPG(char * filename){
  FILE * ifile = fopen(filename,"rb");
  if(!ifile)return 0;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPLE * jsp;
  
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, ifile);
  
  jpeg_read_header(&cinfo, TRUE);
  jpeg_start_decompress(&cinfo); 

  Image8 image(cinfo.output_width,cinfo.output_height,cinfo.output_components);
  
  jsp = (JSAMPLE *) new char[cinfo.output_width*cinfo.output_components];
  
  int i=0;
  while(i<cinfo.output_height){
    int ret = jpeg_read_scanlines(&cinfo,&jsp,1);
    if(ret==0){
      fprintf(stderr,"Error reading scanlines\n");
      return 0;
    }
    memcpy(image.data+i*image.w*image.nchannels,jsp,image.w*image.nchannels);
    i++;
  }

  delete jsp;
  
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo); 
  fclose(ifile);

  operator=(image);
}

template <class T>
int Image<T>::readPPM(char * filename){
  FILE * f = fopen(filename,"rb");
  int width,height,number,max,i=0;
  int nread=0;
  Image<unsigned char> image;
  if(!f){
    printf("Error opening %s\n",filename);
    return 0;
  }
  char buffer[512];
  if(NULL==fgets(buffer,512,f)){
    fclose(f);
    return 0;
  }
  
  if(buffer[0]!='P'){
    fclose(f);
    return 0;
  }
  number=buffer[1]-'0';
  
  while(1){
    if(NULL==fgets(buffer,512,f)){
      fclose(f);
      return 0;
    }
    if(buffer[0]!='#')
      break;
  }
  
  sscanf(buffer,"%d %d",&width,&height);
  if(NULL==fgets(buffer,512,f)){
    fclose(f);
    return 0;
  }
  //P1 is the bitmap, no max
  if(number!=1){
    sscanf(buffer,"%d",&max);
    if(max>65535){
      fclose(f);
      return 0;
    }
  }
  
  if(max>255){
    printf("ppm with max>255 not handled\n");
    exit(-1);
  }
  
  printf("reading image P%d, max=%d,width=%d,height=%d\n",
	 number,max,width,height);
  
  switch(number){
  case 1:
    printf("not handled yet\n");
    fclose(f);
    return 0;
    break;
  case 2:
    image=Image<unsigned char>(width,height,1);  
    for(i=0;i<width*height;i++){
      int n;
      
      if(!fscanf(f,"%d",&n)){
	fclose(f);
	return 0;
      }
      if(i==0)
	printf("[%d]\n",n);
      image[i]=n;
    }
    break;
  case 3:
    image=Image<unsigned char>(width,height,3);
    for(i=0;i<width*height*3;i++){
      int n;
      if(!fscanf(f,"%d",&n)){
	fclose(f);
	return 0;
      }
      image[i]=n;
    }
    break;
		
  case 5:
    //This assumes same endian in machine for max>255
    image=Image<unsigned char>(width,height,1);
    if(((unsigned int)width*height)!=
       fread((unsigned char *)image.getDataPtr(),(max>255)+1,width*height,f)){
      fclose(f);
      return 0;
    }
    break;
  case 6:
    image=Image<unsigned char>(width,height,3);
    //This assumes same endian in machine for max>255
		
    if(width*height*3!=
       (nread=fread(image.getDataPtr(),(max>255)+1,width*height*3,f))){
      fclose(f);
      printf("unexpected EOF %d!=%d\n",nread,width*height*3);
      return 0;
    }
    break;
  default: 
    printf("unknown magic number %d",number);
    fclose(f);
    return 0;
  }
	
  operator=(image);
	
  fclose(f);
  return 1;
}

//template <class T>
int Image<unsigned char>::writeTGA(char * filename){
  FILE * f = fopen(filename,"wb");
  Image8 im;
  im=*this;
  if(!f)return 0;
  unsigned short w=width;
  unsigned short h=height;
  unsigned char bpp=nchannels;
  char tga_header_rgb[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  char tga_header_gray[12] = {0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0}; 
  if(bpp==3 || bpp==4)fwrite(tga_header_rgb,sizeof(tga_header_rgb),1,f);
  else fwrite(tga_header_gray,sizeof(tga_header_gray),1,f);
  fwrite(&w,2,1,f);
  fwrite(&h,2,1,f);
  fputc(bpp*8,f);
  if(bpp==4)fputc(32|8,f);
  else fputc(32,f); //flipping

  if(bpp==4)
  for(int i=0;i<w*h*bpp;i+=bpp)
  {
    fputc(im.data[i+2],f);
    fputc(im.data[i+1],f);
    fputc(im.data[i],f);
    fputc(im.data[i+3],f);
  }
  else if(bpp==3)
  for(int i=0;i<w*h*bpp;i+=bpp)
  {
    fputc(im.data[i+2],f);
    fputc(im.data[i+1],f);
    fputc(im.data[i],f);
  }

  else
    fwrite(im.data,w*h*bpp,1,f);
  fclose(f);
  return 18+w*h*bpp;
}

/*
int Image<unsigned char>::writeJPG(char * filename){
{
  FILE * ofile=fopen(filename,"wb");
  if(!ofile)return 0;
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPLE * jsp;

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, ofile);
  
  if(nchannels==1)
    cinfo.input_components=JCS_GRAYSCALE;
  else {
    printf("unable to write color jpegs\n");
    return 0;
  }
  
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo,30,1); 
  jpeg_start_compress(&cinfo,TRUE);
    
  jsp = (JSAMPLE *)new char[cinfo.image_width];

  int i=0;
  while(cinfo.next_scanline<cinfo.image_height){
    
    int nread=fread(jsp,1,cinfo.image_width,ifile);
    int ret = jpeg_write_scanlines(&cinfo,&jsp,1);
    if(ret==0){
      fprintf(stderr,"Error writing scanlines\n");
      return(0);
      break;
    }
  }
  delete jsp;
  
  fflush(ifile);
  jpeg_finish_compress(&cinfo);
  fclose(ofile);
  jpeg_destroy_compress(&cinfo); 
  fclose(ifile);
  return 1;
}
 */
template <class T>
int Image<T>::writePPM(char * filename){
  Image8 im;
  im=*this;
  ::writePPM((unsigned char *)im.data,im.w,im.h,filename,im.nchannels);
  return 1;
}

void writePPM(unsigned char * image,int w,int h,char * filename,int nchannels)
{
    FILE * file = fopen(filename,"wb");
    int i=0;

    if(nchannels==3){
	fprintf(file,"P6\n%d %d\n%d\n",w,
		h,
		255);
	for(i=0;i<h;i++){
	    fwrite(image+(h-i-1)*w*3,1,w*3,file);
	}
    }
    else if(nchannels==1){
	fprintf(file,"P5\n%d %d\n%d\n",w,
		h,
		255);
	for(i=0;i<h;i++){
	    fwrite(image+(h-i-1)*w,1,w,file);
	}
    }
    fclose(file);    
}

template <class T>
int Image<T>::read(char * filename){
  if((strstr(filename,".ppm")==
      filename+strlen(filename)-4) ||
     (strstr(filename,".pgm")==
      filename+strlen(filename)-4)){
    return readPPM(filename);
  }
  else if((strstr(filename,".tga")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".TGA")==
	   filename+strlen(filename)-4)){
    return readTGA(filename);
  }
  else if((strstr(filename,".jpg")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".JPG")==
	   filename+strlen(filename)-4)){
    return readJPG(filename);
  }
  return 0;
}

int Image<unsigned char>::save(char * filename){
  if((strstr(filename,".ppm")==
      filename+strlen(filename)-4) ||
     (strstr(filename,".pgm")==
      filename+strlen(filename)-4)){
    return writePPM(filename);
  }
  else if((strstr(filename,".tga")==
	   filename+strlen(filename)-4) ||
	  (strstr(filename,".TGA")==
	   filename+strlen(filename)-4)){
    return writeTGA(filename);
  }
  return 0;
}

/*Image8 colorSeg(Image8 & image,Image8 & color)
{
    Imagef imf;
    imf=image;
    Imagef cof;
    cof=color;
    return colorSeg(imf,cof);
}*/
template <class T>
Image8 colorSeg(Image<T> & inputim,Image<T> & inputcolor)
{
  Imagef image;
  Imagef color;
  image.operator=(inputim);
  color.operator=(inputcolor);

  Matrix imflat(3,image.getWidth()*image.getHeight());
  Matrix sample(3,color.getWidth()*color.getHeight());
  for(int y=0;y<color.getHeight();y++){
    for(int x=0;x<color.getWidth();x++){
      sample(0,y*color.getWidth()+x)=color(x,y,0);
      sample(1,y*color.getWidth()+x)=color(x,y,1);
      sample(2,y*color.getWidth()+x)=color(x,y,2);
    }
  }

  for(int y=0;y<image.getHeight();y++){
    for(int x=0;x<image.getWidth();x++){
      imflat(0,y*image.getWidth()+x)=image(x,y,0);
      imflat(1,y*image.getWidth()+x)=image(x,y,1);
      imflat(2,y*image.getWidth()+x)=image(x,y,2);
    }
  }

  Matrix mean(3,1);
  mean.setAll(0);
  for(int j=0;j<sample.n;j++){
    mean[0]+=sample(0,j);
    mean[1]+=sample(1,j);
    mean[2]+=sample(2,j);
  }
  mean[0]/=sample.n;
  mean[1]/=sample.n;
  mean[2]/=sample.n;
  
  for(int j=0;j<sample.n;j++){
    sample(0,j)-=mean[0];
    sample(1,j)-=mean[1];
    sample(2,j)-=mean[2];
  }
  for(int j=0;j<imflat.n;j++){
    imflat(0,j)-=mean[0];
    imflat(1,j)-=mean[1];
    imflat(2,j)-=mean[2];
  }

  Matrix covar=mean*(mean.transpose());
  Matrix evec,eval,d;
  covar.svd(evec,eval,d);

  Matrix imtrans=evec.transpose()*imflat;
  Matrix sampletrans=evec.transpose()*sample;
  Matrix stdv(3,1);
  Matrix mean2(3,1);
  stdv.setAll(0);
  for(int j=0;j<sampletrans.n;j++){
    double dx=sampletrans(0,j);
    double dy=sampletrans(1,j);
    double dz=sampletrans(2,j);
    stdv[0]+=(dx*dx);
    stdv[1]+=(dy*dy);
    stdv[2]+=(dz*dz);
  }
  stdv*=(1.0/sampletrans.n);
  double sigma=sqrt(stdv[0]+stdv[1]+stdv[2]);
  Image8 mask(image.getWidth(),image.getHeight(),1);

  for(int y=0;y<image.getHeight();y++){
    for(int x=0;x<image.getWidth();x++){
      double dx=
	imtrans(0,y*image.getWidth()+x);
      double dy=
	imtrans(1,y*image.getWidth()+x);
      double dz=
	imtrans(2,y*image.getWidth()+x);
      double len=sqrt(dx*dx+dy*dy+dz*dz);
      if(len>3.0*sigma){
	  mask(x,y,0)=255;
      }
      else{
	  mask(x,y,0)=0;
      }
    }
  }
  return mask;
}


#ifdef IMAGE_MAIN

double sx,sy;
int hasdisplayed=0;
int ww=640;
int wh=480;
int bdown=0;
GLuint tex;
Image8 image;
char ** imageNames=0;
int nimages=0;
int imagei=0;

void imageChanged();

void resize(int x,int y)
{
  ww=x;
  wh=y;
  glViewport(0,0,ww,wh);
}

void mouse(GLint button,GLint state,int x,int y)
{

}
void motion(int x,int y)
{
  if(bdown){
  }
}

void keys(unsigned char c,int x,int y)
{
  switch(c){
  case '+':
    ww*=2;
    wh*=2;
    glutReshapeWindow(ww,wh);
    break;
  case '-':
    ww/=2;
    wh/=2;
    glutReshapeWindow(ww,wh);
    break;
  case ' ':
    if(hasdisplayed){
      imagei++;
      imageChanged();
      hasdisplayed=0;
    }
    break;
  }
  glutPostRedisplay();
}


void drawChecks()
{
  int vp[4];
  glGetIntegerv(GL_VIEWPORT,vp);

  GLuint checks;
  glGenTextures(1,&checks);
  glViewport(0,0,32,32);
  glBegin(GL_QUADS);
  //draw some checks
  for(int i=0;i<2;i++){
    for(int j=0;j<2;j++){
      glColor3f(0.6,0.6,0.6);
      if((i+j)%2==0)
	glColor3f(0.7,0.7,0.7);
      glVertex3f(i/(float)2.0f,j/(float)2.0f,0);
      glVertex3f((i+1)/(float)2.0f,j/(float)2.0f,0);
      glVertex3f((i+1)/(float)2.0f,(j+1)/(float)2.0f,0);
      glVertex3f(i/(float)2.0f,(j+1)/(float)2.0f,0);
    }
  }
  glEnd();
  glBindTexture(GL_TEXTURE_2D,checks);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
  glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB,
		   0,0,32,32,0);
 
  glViewport(vp[0],vp[1],vp[2],vp[3]);
  
  float tcx=(sx)*image.w/16.0f;
  float tcy=(sx)*image.h/16.0f;
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,checks);
  glBegin(GL_QUADS);
  glTexCoord2f(0,0);  glVertex2f(0,0);
  glTexCoord2f(tcx,0); glVertex2f(1.0,0);
  glTexCoord2f(tcx,tcy);glVertex2f(1.0,1.0);
  glTexCoord2f(0,tcy); glVertex2f(0,1.0);
  glEnd();
  glDeleteTextures(1,&checks);
}

void display()
{
  glClearColor(0.0,0.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  
  hasdisplayed=1;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,1,1,0,-1,1);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  drawChecks();
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tex);
  
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glColor3f(1.0,1.0,1.0);
  glBegin(GL_QUADS);
   glTexCoord2f(0,0);  glVertex2f(0,0);
   glTexCoord2f(sx,0); glVertex2f(1.0,0);
   glTexCoord2f(sx,sy);glVertex2f(1.0,1.0);
   glTexCoord2f(0,sy); glVertex2f(0,1.0);
  glEnd();
  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);
  glutSwapBuffers();

}

void imageChanged()
{
  if(imagei>=nimages)imagei=0;
  image.read(imageNames[imagei]);
  glBindTexture(GL_TEXTURE_2D,tex);
  
  image.initTextureNoResize(&sx,&sy,GL_REPLACE);
  ww=image.w;
  wh=image.h;
  glutReshapeWindow(ww,wh);
  glutSetWindowTitle(imageNames[imagei]);
  glViewport(0,0,ww,wh);
}
void init()
{
  glGenTextures(1,&tex);
  imageChanged();
}

int main(int ac,char ** av){
  /*Image<unsigned char> a(2,2,1);
  Image<unsigned char> b(2,2,1);
  Image<unsigned char> c(2,2,1);
  a=1;
  a(1,1)=2;
  a.print("%d ");
  b=a+2;
  b.print("%d ");
  c=a&b;
  c.print("%d ");
  c.initTexture();
  printf("done addition\n");
  */
  if(ac==0){
    return 0;
  }
  else{
    imagei=0;
    nimages=ac-1;
    imageNames = new char *[nimages];
    for(int i=0;i<nimages;i++){
      imageNames[i]=new char[strlen(av[i+1])+1];
      strcpy(imageNames[i],av[i+1]);
    }
  }
  glutInit(&ac,av);
  glutInitDisplayMode(GLUT_RGB|GLUT_DEPTH|GLUT_DOUBLE);
  glutInitWindowSize(ww,wh);
  glutInitWindowPosition(0,0);
  glutCreateWindow("win");
  glutDisplayFunc(display);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keys);
  init();
  glutMainLoop();
  
  return 0;
}
#endif

template class Image<float>;
template class Image<unsigned char>;
template class Image<unsigned short>;
template class Image<unsigned int>;
template Image8 colorSeg(Image<unsigned char> & im,
			 Image <unsigned char> & col);
