#include <stdio.h>
#include <stdlib.h>
#include <wx/wx.h>
#include <ext/hash_map>
#include "chart.h"
#include "vec.h"
#include "plugin_defines.h"

using namespace __gnu_cxx;

extern char currentProj[];

struct VecHasher{
  size_t operator()(const Vec3f & v) const{
    return (size_t)(317*(v.x+1)+317*317*(v.y+1)+317*317*317*(v.z+1));
  }
  int operator()(const Vec3f & v1,const Vec3f & v2) const{
    return (v1.x==v2.x)&&(v1.y==v2.y)&&(v1.z==v2.z);
  }
};

extern "C" const char *  getName(){
  return "Inject";
}
extern "C" const char *  getInfo(){
  return "Perform injective mapping";
}
extern "C" int getType(){
  return GEOMPLUGIN_OPERATION_TYPE;
}
//This function doesn't do anything...its all done by the requires
extern "C" int apply(Chart ** charts,int * ncharts){
  if(*ncharts==0)return 0;

  Chart c0;
  Chart c1;
 
  char fname[1024];
  sprintf(fname,"%s/geometry_orig.obj",currentProj);
  c0.loadobj(fname);
  sprintf(fname,"%s/geometry.obj",currentProj);
  c1.loadobj(fname);

  hash_map<Vec3f,int,VecHasher> cur_to_orig;

  if(c0.nvert!=c1.nvert || !c0.nvert){
    wxMessageBox("Geometry files have different # of verts","Failed");
    return 0;
  }

  int dup=0;
  for(int i=0;i<c1.nvert;i++){
    if(cur_to_orig.count(c1.vert[i]))dup++;
    else cur_to_orig[c1.vert[i]]=i;
  }
  if(dup){
    char msg[1024];
    sprintf(msg,"There are %d duplicats, this may not work",dup);
    wxMessageBox(msg,"Failed");
  }
  int errs=0;
  for(int i=0;i<*ncharts;i++){
    for(int j=0;j<(*charts)[i].nvert;j++){
      if(!cur_to_orig.count((*charts)[i].vert[j]))errs++;
      else (*charts)[i].vert[j]=c0.vert[cur_to_orig[(*charts)[i].vert[j]]];
    }
    (*charts)[i].initNormals();
  }
  if(errs){
    char msg[1024];
    sprintf(msg,"There were %d errors doing mapping",errs);
    
    wxMessageBox(msg,"Errors");
    return 0;
  }
  
  return 0;
}
extern "C" const void * getPanel(void * v){
  return 0;
}
