#ifndef DEF_MESH
#define DEF_MESH
#include "vec.h"
#include "matrix.h"
#include "image.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <vector>
#include "voxbin.h"
#include <ext/hash_map>

using namespace __gnu_cxx;
using std::vector;

#define TS_USE_TC_FOR_VERT    0x1
#define TS_USE_VERT_FOR_TC    0x2
#define TS_DEPTH_MODE         0x4
#define TS_NORM_TO_ALPHA_MODE 0x8



/** HalfEdge data structure
 */
class HalfEdge
{
 public:
  unsigned int face;
  unsigned int vert;
  HalfEdge * next;
  HalfEdge * opp;
  ~HalfEdge(){
    next=0;
    opp=0;
    face=0;
    vert=0;
  }
};

/** Face data structure
    Useful for adding extra data to a face
 */
class Face{
 public:
  Face(){
    edge=0;
    flags=0;
  }
  HalfEdge * edge;
  Vec2f tc[3];
  Vec3f col[3];
  int flags;
};

#define DM_CUR_VERT_BUF 0x1
#define DM_DER_VERT_BUF 0x2
#define DM_ALL_VERT_BUFS 0x3

/** Def(ormable)Mesh class
 * This is the deformable mesh
 */
class DefMesh
{
 private:
  GLuint tex[2];//The second textuer is used for visibility
  int required_texw;
  int required_texh;
 public:
  //stats
  VoxBin po;  
  int triangleTexWidth;
  int desiredTriangleWidth;
  static const double normalDerivativeScale=0.8;
  int lastnconn;
  int maxvalence;
  /**The first entry is used for the current data,
     the second is used for : verts-sc*normal,
     and the third is used for: verts+sc*normal
  GLuint vertBuf[3];
  GLuint normBuf[3];
  GLuint tcBuf;
  */
  Vec3f * vert;
  Vec3f * norm;
  Face * tri;
  int ntri;
  int nvert;
  int maxtri;
  int maxvert;
  int useTexNumber;
  double sigma;
  Vec2f binrangex,binrangey,binrangez;
  
  DefMesh(double sigma_=0.05);
  ~DefMesh();
  void ensureTri(int n);
  void ensureVert(int n);
  void clear();
  void checkConsistency(char * str);
  void fixNeighboringVert(HalfEdge * he);
  void fixNeighboringVert(int face);
  void allocFace(HalfEdge ** he1,HalfEdge ** he2,HalfEdge ** he3);
  HalfEdge * allocHalfEdge();
  int recursiveDelete(HalfEdge * he,int & cnt,int & fixed);
  void recursiveDeleteCnt(HalfEdge * he,hash_map<int,char> & vis,int label=1);
  void freeHalfEdge(HalfEdge * he);
  int fixAdj();
  void initNormals();
  int  loadobj(char * filename,int center=1);
  int  writeobj(char * filename);
  void drawme(int flags);//send down triangles with tex and norms, no modification
  void draw(int drawtex=0,int drawCol=0);
  void drawTC();
  void move();
  void genSphere(double rad);
  void smooth(float param=0.2,vector <int> * onlydo=0);
  int  lechaudRemesh();
  void lechaudRemeshHelper(int face);
  void makeVirtualNeighbors(HalfEdge * he,int removeFromOctree);
  void axialTform(HalfEdge * he1,HalfEdge * he2);
  void removeUnusedTri();
  void ranges(Vec3f & min,Vec3f & max);
  void scale(Vec3f & sc);
  void translate(Vec3f & sc);
};

#endif
