#include "pqt.h"

int (* distanceHandler)(const Vec3f & nodePoint,void * nodeData,
			 const Vec3f & pt,void * queryData,
			 double rad)=0;

void PONode::bruteDistanceQuery(const Vec3f & pt,double rad,void *queryData){
    if(!valid)return;
    double d=(point-pt).len();
    //checked++;
    if(d<rad){
      if(distanceHandler){
	if(!nUserData){
	   distanceHandler(point,0,pt,queryData,rad);
	}
       	for(int i=0;i<nUserData;i++){
	  distanceHandler(point,userData[i],pt,queryData,rad);
	}
      }
      //printf("%f %f %f == %f<%f\n",point.x,point.y,point.z,d,rad);
      //found++;
    }
    if(children){
      for(int i=0;i<8;i++){
	children[i].bruteDistanceQuery(pt,rad,queryData);
      }
    }
  }
//a return value > 0 means do not process anymore
  int PONode::distanceQuery(const Vec3f & pt,double rad,int & found,void * queryData){
    if(!valid)return 0;
    double d=(point-pt).len();
    //checked++;
    if(d<rad){
      found++;
      if(distanceHandler){
	if(!nUserData){
	  int ret=0;
	  distanceHandler(point,0,pt,queryData,rad);
	  if(ret)return 1;
	}
	for(int i=0;i<nUserData;i++){
	  int ret=distanceHandler(point,userData[i],pt,queryData,rad);
	  if(ret)return 1;
	}
      }
      //printf("%f %f %f == %f<%f\n",point.x,point.y,point.z,d,rad,queryData);
      
    }
    if(children){
      float dz=pt.z-point.z;
      float dy=pt.y-point.y;
      float dx=pt.x-point.x;

      if(dz-rad<=0){
	if(dy-rad<=0){
	  if(dx-rad<=0){
	    int ret=children[0].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }if(dx+rad>0){
	    int ret=children[1].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	}
	if(dy+rad>0){
	  if(dx-rad<=0){
	    int ret=children[2].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	  if(dx+rad>0){
	    int ret=children[3].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	}      
      }
      if(dz+rad>0){
	if(dy-rad<=0){
	  if(dx-rad<=0){
	    int ret=children[4].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	  if(dx+rad>0){
	    int ret=children[5].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	}
	if(dy+rad>0){
	  if(dx-rad<=0){
	    int ret=children[6].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	  if(dx+rad>0){
	    int ret=children[7].distanceQuery(pt,rad,found,queryData);
	    if(ret)return 1;
	  }
	}  
      }
    }
    return 0;
  }

void PointOctree::setDistanceHandler(int (* dhandler)(const Vec3f & nodePoint,
						       void * nodeData,
						       const Vec3f & pt,
						       void * queryData,
						       double rad))
{
  distanceHandler=dhandler;
}

#if defined POINTOCTREE_MAIN
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

int ww=256;
int wh=256;
int mpos[2];
int bdown=0;


Vec3f crot(0,0,0);
Vec3f cpos(0,0,-2);
void display()
{
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  glPushMatrix();
  glTranslatef(cpos.x,cpos.y,cpos.z);
  glRotatef(crot.x,1,0,0);
  glRotatef(crot.y,0,1,0);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);


  glPopMatrix();
  glutSwapBuffers();
}

void mouse(GLint button,GLint state,int x,int y)
{
  bdown=(state==GLUT_DOWN)?button+1:0;
  mpos[0]=x;
  mpos[1]=y;
  glutPostRedisplay();
}

void keys(unsigned char c,int x,int y)
{
  switch(c){
  case 27:exit(0);
  }
  glutPostRedisplay();
}

void motion(int x,int y)
{
  double dx=x-mpos[0];
  double dy=y-mpos[1];
  if(bdown==1){
    crot.x+=dy*0.1;
    crot.y+=dx*0.1;
  }
  if(bdown==2){
    cpos.x+=dx*0.01;
    cpos.z-=dy*0.01;
  }
  mpos[0]=x;
  mpos[1]=y;
  glutPostRedisplay();
}

void init()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(50,(double)ww/(double)wh,1,1000);
  glViewport(0,0,ww,wh);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}



int main(int ac,char * av[])
{
  glutInit(&ac,av);
  glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE |GLUT_DEPTH);

  glutInitWindowSize(ww,wh);
  glutCreateWindow("window");
  glutDisplayFunc(display);
  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  glutKeyboardFunc(keys);

  init();

  for(int k=0;k<1000;k++){
    PointOctree po;
    for(int i=0;i<2000;i++){
      double x,y,z;
      x=((double)(rand()%10000))/10000.0;
      y=((double)(rand()%10000))/10000.0;
      z=((double)(rand()%10000))/10000.0;
      po.insert(Vec3f(x,y,z));
  }

  for(int i=0;i<2000;i++){
    double x,y,z;
    x=((double)(rand()%10000))/10000.0;
    y=((double)(rand()%10000))/10000.0;
    z=((double)(rand()%10000))/10000.0;
    po.distanceQuery(Vec3f(x,y,z),0.0001);
  }
  printf("%d\n",k);
  }
  glutMainLoop();
  return 0;
}


#endif
