#ifndef POINT_QUADTREE_H
#define POINT_QUADTREE_H

#include "matrix.h"
#include "vec.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

class PONode{
  void ** userData;
  int nUserData;
  int maxUserData;

  Vec3f point;
  PONode * children;
  unsigned char valid;
 public:
  PONode(){
    children=0;
    valid=0;
    nUserData=0;
    maxUserData=0;
    userData=0;
  }
  PONode(Vec3f & pt,void * _userData=0){
    point=pt;
    children=0;
    valid=1;
    maxUserData=0;
    nUserData=0;
    ensureMaxUserData(6);
    userData[nUserData++]=_userData;
  }
  void ensureMaxUserData(int n){
    if(n>=maxUserData){
      int newMax=(n>maxUserData+6)?n:(maxUserData+6);
      void ** newud=new void*[newMax];
      memcpy(newud,userData,sizeof(void *)*maxUserData);
      delete [] userData;
      maxUserData=newMax;
      userData=newud;
    }
  }
  ~PONode(){
    if(children){
      delete [] children;
    }
    maxUserData=0;
    nUserData=0;
    if(userData)delete [] userData;
  }
  void insert(const Vec3f & pt,void * _userData=0){
    if(children==0){
      children = new PONode[8];
    }
    if(!valid){
      //inserted++;
      point=pt;
      valid=1;
      nUserData=0;
      ensureMaxUserData(6);
      userData[nUserData++]=_userData;
      return;
    }
    if((pt-point).len()==0){
      ensureMaxUserData(nUserData+1);
      userData[nUserData++]=_userData;
      return;
    }
    int index=0;
    index=4*(pt.z>point.z)+2*(pt.y>point.y)+(pt.x>point.x);
    children[index].insert(pt,_userData);
  }
  void removeUserData(const Vec3f & pt){
    if(!valid)return;
    if((pt-point).len()<10e-9){
      nUserData=0;
    }
    if(children){
      int index=0;
      index=4*(pt.z>point.z)+2*(pt.y>point.y)+(pt.x>point.x);
      children[index].removeUserData(pt);
    }
  }
  void bruteDistanceQuery(const Vec3f & pt,double rad,void * queryData=0);
  int distanceQuery(const Vec3f & pt,double rad,int & found,void * queryData=0);
		       
  void print(int depth){
    if(valid){
      
      printf("%3.2f %3.2f %3.2f ",point.x,point.y,point.z);
      printf("\n");
      for(int i=0;i<8;i++){
	for(int j=0;j<depth;j++)putchar(' ');
	printf("%d :",i);children[i].print(depth+1);
	printf("\n");
      }
    }
  }
};

class PointOctree
{
  PONode * root;
 public:
  PointOctree(){
    root =new PONode();
  }
  ~PointOctree(){
    delete root;
    root=0;
  }
  void insert(const Vec3f & pt,void * ud=0){
    root->insert(pt,ud);
  }
  void removeAll(){
    delete root;
    root=new PONode();
  }
  //it seems that a remove would require a complete remaking of the
  //tree...instead we will just enullinate the userData
  void removeUserData(const Vec3f & pt){
    root->removeUserData(pt);
  }
  void setDistanceHandler(int (* dhandler)(const Vec3f & nodePoint,
					    void * nodeData,
					    const Vec3f & pt,void * queryData,
					    double rad));
  int distanceQuery(const Vec3f & pt,double rad,void * queryData=0){
    //checked=0;
    //found=0;
    int found=0;
    root->distanceQuery(pt,rad,found,queryData);
    return found;
    //printf("Eff  : checked/inserted = %d/%d, found=%d\n",checked,inserted,found);
    /*int foundeff=found;
    checked=0;
    found=0;
    root->bruteDistanceQuery(pt,rad);
    printf("Brute: checked/inserted = %d/%d, found=%d\n",checked,inserted,found);

    if(found!=foundeff){
      printf("found=%d!=%d=foundeff\n",found,foundeff);
      exit(0);
    }
    */
  }
  void print(){
    root->print(0);
  }
};

#endif
