#include <stdio.h>
#include <stdlib.h>
#include "chart.h"
#include "defmesh.h"
#include "plugin_defines.h"

#include <wx/progdlg.h>

extern "C" const char *  getName(){
  return "Remesh";
}
extern "C" const char *  getInfo(){
  return "Remesh the geometry";
}
extern "C" int getType(){
  return GEOMPLUGIN_OPERATION_TYPE;
}
//This function doesn't do anything...its all done by the requires
extern "C" int apply(Chart ** charts,int * ncharts){
  if(*ncharts>1)
    return GEOMPLUGIN_RESTORE_BIT | GEOMPLUGIN_APPLY_AGAIN_BIT;
  else if(*ncharts==1){
    const int maxits=20;
    wxBeginBusyCursor();
    wxProgressDialog   progress("Remeshing Geometry","",maxits,NULL,wxPD_ELAPSED_TIME|wxPD_APP_MODAL|wxPD_AUTO_HIDE);
    progress.Show(TRUE);

    printf("saving object\n");
    char tmpdir[1024]="/tmp/";
    char name[1024];
#ifdef WIN32
    GetTempPath(sizeof(tmpdir),tmpdir);
#endif
    snprintf(name,1024,"%s/_obj_remesh_temp_obj_",tmpdir);
    (*charts)[0].saveobj(name);
    printf("done saving object\n");
    DefMesh dm;
    printf("loading object\n");
    if(!dm.loadobj(name,0))return 0;
    printf("done loading object\n");
    double maxdist=0;
    for(int i=0;i<dm.ntri;i++){
      HalfEdge * he=dm.tri[i].edge;
      if(!he)continue;
      for(int k=0;k<3;k++,he=he->next){
	Vec3f dist=(dm.vert[he->vert]-dm.vert[he->next->vert]);
	maxdist=MAX(maxdist,dist.len());
      }
    }
    printf("maxdist :%f\n",maxdist);
    dm.sigma=maxdist/2.5;
    int lastntri=dm.ntri;
    int lastnvert=dm.nvert;
    for(int i=0;i<maxits;i++){
      wxString mesg;
      mesg.Printf("Iteration %d/%d",i+1,maxits);
      progress.Update(i,mesg);

      printf("remeshing %d, (nvert:%d,ntri:%d)\n",i,dm.nvert,dm.ntri);
      dm.lechaudRemesh();
      //make the assumption that if the triangles and the nvert stayed
      //the same, then the remeshing must be complete.
      if(dm.ntri==lastntri &&dm.nvert==lastnvert)break;
      lastntri=dm.ntri;
      lastnvert=dm.nvert;
    }
    printf("saving from defmesh\n");
    dm.writeobj(name);
    Chart c;
    c.loadobj(name);
    (*charts)[0]=c;

    wxEndBusyCursor();
  }
  return 0;
}

extern "C" const void * getPanel(void * v){
  return 0;
}
