#include <stdio.h>
#include <stdlib.h>
#include "chart.h"
#include "plugin_defines.h"
#include <wx/wx.h>
#include <wx/spinctrl.h>
#include <wx/progdlg.h>
#include <wx/process.h>

class MyProcess : public wxProcess{
public:
  wxInputStream * inputStream;
  int curProgress;
  wxString curMessage;
  int term;
  MyProcess() : wxProcess(wxPROCESS_REDIRECT){
    inputStream=0;
    term=0;
  }
  void read(wxProgressDialog & progress){
    inputStream = GetInputStream();
    printf("input stream is %p\n",inputStream);
    if(inputStream){
      wxString str;
      while(!inputStream->Eof()){
	char c=inputStream->GetC();
	if(!inputStream->LastRead())break;
	if(c=='\n'||c=='\r'){
	  const char * cstr=str.ToAscii();
	  
	  if(strstr(cstr,"message:")==cstr){
	    char msg[2048];
	    curMessage=str.SubString(8,str.size());
	  }
	  else if(strstr(cstr,"progress:")==cstr){
	    int prog=0;
	    if(sscanf(cstr,"progress:%d",&prog))curProgress=prog;
	    printf("curProgress is %d\n",curProgress);
	  }
	  progress.Update(curProgress,curMessage);
	  str=wxT("");
	}
	else
	  str+=c;
      }
    }
  }
  void wait(){
    while(!term)
      wxYield();
  }
  void OnTerminate(int pid, int status){
    printf("terminated %d %d\n",pid, status);
    term = true;
  }
};

wxPanel * panel;
wxSpinCtrl * nrays;
extern "C" const char *  getName(){
  return "MI(n)";
}
extern "C" const char *  getInfo(){
  return "Compute geometry from SFS using MI data structure";
}
extern "C" int getType(){
  return GEOMPLUGIN_CREATE_TYPE;
}
//This function doesn't do anything...its all done by the requires
extern "C" int create(char * dir,char * imfile,
		      int nimages,
		      char * geomFile,
		      float sx,float sy,float sz,
		      float tx,float ty,float tz){
  if(!dir && !imfile)return GEOMPLUGIN_NEEDS_BOUNDING_BOX_BIT;
  char cmdline[2048];
  printf("calling ray\n");
  snprintf(cmdline,2048,"%s --dir \"%s\" --file %s --nrays %d --nimages %d --sx %8.4f --sy %8.4f --sz %8.4f --tx %8.4f --ty %8.4f --tz %8.4f --ofile \"%s\"",
	   "ray",
	   dir,imfile,nrays->GetValue(),
	   nimages,
	   sx,sy,sz,
	   tx,ty,tz,geomFile);
  printf("cmdline %s\n",cmdline);
  wxBeginBusyCursor();
  wxProgressDialog   progress(wxT("Computing Geometry"),wxT(""),nimages,
			      NULL,wxPD_ELAPSED_TIME|wxPD_REMAINING_TIME|wxPD_APP_MODAL);
  progress.Show(TRUE);

  MyProcess myProcess;
  wxExecute(wxString::FromAscii(cmdline), wxEXEC_ASYNC, &myProcess);
  myProcess.read(progress);
  myProcess.wait();

  progress.Show(FALSE);
  wxEndBusyCursor();

  /*wxString curMessage;
  int curProgress=0;
  
  wxProcess * process=wxProcess::Open(cmdline);
  if(process){
    process->Redirect();
    wxInputStream * inputStream=process->GetInputStream();
    if(inputStream){
      wxString str;
      while(!inputStream->Eof()){
	char c=inputStream->GetC();
	if(!inputStream->LastRead())break;
	if(c=='\n'||c=='\r'){
	  const char * cstr=str.c_str();
	  
	  if(strstr(cstr,"message:")==cstr){
	    char msg[2048];
	    curMessage=str.SubString(8,str.size());
	  }
	  else if(strstr(cstr,"progress:")==cstr){
	    int prog=0;
	    if(sscanf(cstr,"progress:%d",&prog))curProgress=prog;
	    printf("curProgress is %d\n",curProgress);
	  }
	  progress.Update(curProgress,curMessage);
	  str="";
	}
	else
	  str+=c;
      }
    }
  }
  progress.Show(FALSE);
  wxEndBusyCursor();
  */
  
    /*
#ifdef WIN32
  FILE * file=_popen(cmdline,"rt");
#else
  FILE * file=popen(cmdline,"r");
#endif

  wxString str;
  int c;
  while(EOF!=(c=fgetc(file))){
    if(c=='\n'||c=='\r'){
      const char * cstr=str.c_str();
      
      if(strstr(cstr,"message:")==cstr){
	char msg[2048];
	if(sscanf(cstr,"message:%s",msg))curMessage=msg;
      }
      else if(strstr(cstr,"progress:")==cstr){
	int prog=0;
	if(sscanf(cstr,"progress:%d",&prog))curProgress=prog;
      }
      progress.Update(curProgress,curMessage);
      str="";
    }
    else
      str+=c;
  }
  progress.Show(FALSE);
  wxEndBusyCursor();

#ifdef WIN32
  _pclose(file);
#else
  pclose(file);
#endif
    */
  printf("Done with %s plugin\n",getName());
  return 0;
}
extern "C" const void * getPanel(wxWindow * parent){
  panel=new wxPanel(parent);
  wxBoxSizer * sizer 
    = new wxStaticBoxSizer(new wxStaticBox(panel,-1,wxT("Options")),
			     wxVERTICAL);
  wxBoxSizer * row = new wxBoxSizer(wxHORIZONTAL);
  nrays =new wxSpinCtrl(panel,-1);
  nrays->SetRange(0,512);
  nrays->SetValue(32);

  row->Add(new wxStaticText(panel,-1,wxT("# rays")),1);
  row->Add(nrays,1,wxALIGN_RIGHT|wxGROW|wxEXPAND|wxEXPAND|wxRIGHT,4);
  sizer->Add(row,0,wxGROW|wxEXPAND|wxEXPAND|wxRIGHT,2);
  sizer->Layout();
  row->Layout();
  panel->SetToolTip(wxString::FromAscii(getInfo()));
  panel->SetSizer(sizer);
  return panel;
}
