#include <stdio.h>
#include <stdlib.h>
#include "chart.h"
#include "matrix.h"
#include "texgen.h"
#include "plugin_defines.h"

extern "C" const char *  getName(){
  return "Spherical";
}
extern "C" const int  getType(){
  return TEXPLUGIN_MAP_TYPE;
}
extern "C" const char *  getInfo(){
  return "Spherical texture coordinates";
}
//Get the texture coordinates
extern "C" int apply(Chart ** charts,int * ncharts){
  if(*ncharts==1){
    Chart newChart;
    //Need to finish this eventually
    Vec3f avg(0,0,0);
    double miny=100000;
    double maxy=-100000;
    //Get the average position
    for(int j=0;j<(*charts)[0].nvert;j++){
      avg+=(*charts)[0].vert[j];
      miny=MIN(miny,(*charts)[0].vert[j][1]);
      maxy=MAX(maxy,(*charts)[0].vert[j][1]);
    }
    avg*=(1.0/(*charts)[0].nvert);

    //Compute the texture coordinates 
    for(int j=0;j<(*charts)[0].nvert;j++){
      Vec3f c=(*charts)[0].vert[j]-avg;
      double azi=(atan2(c.z,c.x)+M_PI);
      (*charts)[0].tc[j][0]=(azi)/(2*M_PI);
      (*charts)[0].tc[j][1]= atan2(sqrt(c.x*c.x+c.z*c.z),c.y);//cyl --> (*charts)[0].vert[j].y/(maxy-miny);
      printf("%f %f\n",(*charts)[0].tc[j][0],(*charts)[0].tc[j][1]);
    }

    //Find the place to split, overmap maps a vertex to a
    //duplicated vertex that wraps around
    int extravert=0;
    int * overMap = new int[(*charts)[0].nvert];
    memset(overMap,-1,sizeof(int)*(*charts)[0].nvert);
    Chart * c=(*charts);
    for(int j=0;j<(*charts)[0].npoly;j++){
      for(int k=0;k<3;k++){
	int first=c->poly[j][k];
	int second=c->poly[j][(k+1)%3];
	
	double ang1=(atan2(c->vert[first][2]-avg.z,c->vert[first][0]-avg.x))+M_PI;
	double ang2=(atan2(c->vert[second][2]-avg.z,c->vert[second][0]-avg.x))+M_PI;
	double diff=ang2-ang1;
	if(fabs(diff)>M_PI){
	  if(ang1<ang2){
	    if(overMap[first]==-1){
	      overMap[first]=extravert++;
	    }
	  }
	  else if(overMap[second]==-1)
	    overMap[second]=extravert++;
	}
      }
    }

    //initialize the new chart and copy over the vertices
    newChart.init(c->npoly,c->nvert+extravert);
    for(int j=0;j<(*charts)[0].nvert;j++){
      newChart.vert[j]=(*charts)[0].vert[j];
      newChart.tc[j]=(*charts)[0].tc[j];
      newChart.norm[j]=(*charts)[0].norm[j];
      if(overMap[j]>=0){
	newChart.vert[overMap[j]+(*charts)[0].nvert]=newChart.vert[j];
	newChart.tc[overMap[j]+(*charts)[0].nvert]=newChart.tc[j];
	newChart.norm[overMap[j]+(*charts)[0].nvert]=newChart.norm[j];
      }
    }

    //set the triangles, making sure to map the edges that went over
    // to the duplicated triangles
    for(int j=0;j<(*charts)[0].npoly;j++){
      for(int k=0;k<3;k++)
	newChart.poly[j][k]=(*charts)[0].poly[j][k];
      
      //fix up the overs
      for(int k=0;k<3;k++){
	int first=c->poly[j][k];
	int second=c->poly[j][(k+1)%3];
	
	double ang1=(atan2(c->vert[first][2]-avg.z,c->vert[first][0]-avg.x))+M_PI;
	double ang2=(atan2(c->vert[second][2]-avg.z,c->vert[second][0]-avg.x))+M_PI;
	double diff=ang2-ang1;
	if(fabs(diff)>M_PI){
	  if(ang1<ang2){
	    newChart.poly[j][k]=(*charts)[0].nvert+overMap[first];
	  }
	  else{
	    newChart.poly[j][(k+1)%3]=(*charts)[0].nvert+overMap[second];
	  }
	}
      }
    }
    //scale down to texture size
    double maxx=1.0;
    for(int j=(*charts)[0].nvert;j<(newChart.nvert);j++){
      newChart.tc[j].x+=1;
      maxx=MAX(newChart.tc[j].x,maxx);
    }
    for(int j=0;j<(newChart.nvert);j++){
      newChart.tc[j].x/=maxx;
    }
    delete [] overMap;

    //Copy to the results
    (*charts)[0]=newChart;
    newChart.release();

    printf("need %d extra\n",extravert);
    printf("%d %d\n",(*charts)[0].nvert);
    printf("done delete\n");
  }
  else printf("only works on one unsplit object\n");
  return 0;
}

extern "C" const void * getPanel(void * v){
  return 0;
}
