#include "vsa.h"
using namespace std;
using namespace ::__gnu_cxx;


/** Get the triangle area and normal for the given tri.
 */
float VSA::triangleAreaAndNormal(int tri,Vec3f & norm)
{
  //use heron's formula for area
  int v0=tris[3*tri  ];
  int v1=tris[3*tri+1];
  int v2=tris[3*tri+2];
  Vec3f vec1,vec2;
  double a=(vec1=(vert[v1]-vert[v0])).len();
  double b=(vec2=(vert[v2]-vert[v1])).len();
  double c=(vert[v0]-vert[v2]).len();
  double s=(a+b+c)/2.0f;
  norm=(vec1.cross(vec2));
  /*if(norm.dot(dm->norm[he1->vert])<0 &&
     norm.dot(dm->norm[he2->vert])<0 ){
    printf("normal is pointing in the wrong direction %f\n",norm.dot(dm->norm[he1->vert]));
    }*/
  norm.normalize();
  return sqrt(s*(1-a)*(1-b)*(1-c));
}

/** flood_fill using the given heap, initialized with the
 *  seeds.
 */
void VSA::flood_fill(vector<TriError> & heap){
  printf("starting flood fill %d\n",heap.size());
  make_heap(heap.begin(),heap.end());
  memset(tri2proxy,-1,sizeof(int)*ntris);
  for(int i=0;i<proxies.size();i++){
    proxies[i]->triangles.clear();
  }
  while(!heap.empty()){
    //an inner loop removes everything off the heap,
    //and then stuff that has tri2proxy<0 is added to the heap
    //and we go again.
    while(!heap.empty()){
      pop_heap(heap.begin(),heap.end());
      TriError cur=heap[heap.size()-1];
      heap.pop_back();

      int tri=cur.tri;
      int proxy_id=cur.proxy;
      Proxy & proxy=*(proxies[proxy_id]);
    
      if(tri2proxy[tri]>=0){
	/*float err=error_func(*proxies[tri2proxy[tri]],tri);
	  if(cur.error<err){
	  printf("problem with heap %f < %f\n",cur.error,err);
	  }*/
	continue;
      }
      tri2proxy[tri]=proxy_id;
      proxy.triangles.push_back(tri);

      //add the current triangles neighbors
      for(int k=0;k<3;k++){
	int neighbor=opp[3*tri+k];
	if(neighbor>=0 && tri2proxy[neighbor]==-1){
	  heap.push_back(TriError(neighbor,proxy_id,
				  (this->*error_func)(proxy,neighbor)));
	  push_heap(heap.begin(),heap.end());
	}
      }
    }
    //some triangles may not be connected, we need to add them
    //one by one until they all have a proxy.
    //The choice is random, but really these are unconnected triangles
    //and they should probably have their own proxy...
    for(int i=0;i<ntris;i++){
      if(tri2proxy[i]<0){
	int proxy_id=rand()%nproxies;
	heap.push_back(TriError(i,proxy_id,
				(this->*error_func)(*proxies[proxy_id],i)));
	break;
      }
    }
    if(!heap.empty())printf("unconnected triangles, added one\n");
  }
  /*
  for(int i=0;i<proxies.size();i++){
    for(int j=i+1;j<proxies.size();j++){
      if(proxies[i]->n.dot(proxies[j]->n)==1){
	printf("normals for %d and %d are equal %f,%f,%f\n",i,j,
	       proxies[i]->n.x,proxies[i]->n.y,proxies[i]->n.z);
      }
    }
  }
  
  for(int i=0;i<proxies.size();i++){
    printf("proxies[%d]->triangles.size()=%d\n",i,proxies[i]->triangles.size());
    for(int j=0;j<proxies[i]->triangles.size();j++){
      if(tri2proxy[proxies[i]->triangles[j]]!=i){
	printf("tri2proxy or triangles are wrong for triangle %d\n",proxies[i]->triangles[j]);
      }
    }
  }
  */
}

/** Consider all possible adjacent regions, and merge the regions that
 *  result in the least amount of error.
 */
void VSA::remove_region()
{
  printf("removing region\n");
  hash_map<int,int> * adj = new hash_map<int,int>[nproxies];
  float bestError=10e10;
  int besta=0;
  int bestb=0;
  for(int i=0;i<ntris;i++){

    for(int k=0;k<3;k++){
      int opposite=opp[3*i+k];
      if(opposite<0)continue;
      if(tri2proxy[i]!=tri2proxy[opposite]){
	adj[tri2proxy[i]][tri2proxy[opposite]]=1;
	adj[tri2proxy[opposite]][tri2proxy[i]]=1;
      }
    }
  }
  printf("got adj\n");
  for(int i=0;i<nproxies;i++){
    hash_map<int,int>::iterator finger=adj[i].begin();
    hash_map<int,int>::iterator end=adj[i].end();
    while(finger!=end){
      int other=(finger++)->first;
    
      //Only check a pair once
      if(other>i){
	//Try combining the pair
	//Should be relaced with a merge_proxies_func
        double w1=proxies[i]->total_weight;
	double w2=proxies[other]->total_weight;
	Vec3f n1=proxies[i]->n;
	Vec3f n2=proxies[other]->n;
	Vec3f n=(n1*w1+n2*w2);
	n.normalize();
	
	float thisError=0;
      	proxy_triangles::iterator finger=proxies[i]->triangles.begin();
	proxy_triangles::iterator end=proxies[i]->triangles.end();
	while(finger!=end){
	  int tri=*(finger++);
	  thisError+=(this->*error_func)(*proxies[i],tri);
	}
	finger=proxies[other]->triangles.begin();
	end=proxies[other]->triangles.end();
	while(finger!=end){
	  int tri=*(finger++);
	  thisError+=(this->*error_func)(*proxies[other],tri);
	}
	if(thisError<bestError){
	  bestError=thisError;
	  besta=i;
	  bestb=other;
	}
      }
    }
  }

  printf("best %d %d\n",besta,bestb);
  
  //add all of bestb's triangles to besta
  proxy_triangles::iterator finger=proxies[bestb]->triangles.begin();
  proxy_triangles::iterator end=proxies[bestb]->triangles.end();
  while(finger!=end){
    int tri=*(finger++);
    proxies[besta]->add(tri);
  }
  proxies[bestb]->clear();
  
  //Now move best b to the end of the proxies list
  //by swapping (that is, if it isn't alreay at the end)
  if(bestb!=proxies.size()-1){
    int last_id=proxies.size()-1;
    Proxy * last_proxy=proxies[last_id];
    proxies[last_id]=proxies[bestb];
    proxies[bestb]=last_proxy;
    
    finger=last_proxy->triangles.begin();
    end=last_proxy->triangles.end();
    while(finger!=end){
      int tri=*(finger++);
      if(tri2proxy[tri]!=last_id){
	printf("error moving the triangles from the last proxy to bestb\ntri2proxy[%d]==%d,should be %d (bestb==%d)",
	       tri,tri2proxy[tri],last_id,bestb);
      }
      tri2proxy[tri]=bestb;
    }
  }
 
  //Finally delete the proxy and remove the pointer
  //from the proxies list
  delete proxies[proxies.size()-1];
  proxies.pop_back();
  nproxies--;

  delete [] adj;
}

/** Find the worst approximated region and split it into two.
 */
void VSA::add_region()
{
  int worst_proxy=0;
  float worst_error=-10e10;
  int worst_triangle=0;
  for(int proxy_id=0;proxy_id<proxies.size();proxy_id++){
    proxy_triangles & triangles=(proxies[proxy_id]->triangles);
    Proxy & proxy=*(proxies[proxy_id]);
    float proxy_error=-10e10;
    int proxy_worst_triangle=0;
    proxy_triangles::iterator finger=triangles.begin();
    proxy_triangles::iterator last=triangles.end();

    while(finger!=last){
      int tri=*finger;
      float this_error=(this->*error_func)(proxy,tri);
      if(this_error>proxy_error){
	proxy_error=this_error;
	proxy_worst_triangle=tri;
      }
      finger++;
    }

    if(proxy_error>worst_error){
      worst_error=proxy_error;
      worst_triangle=proxy_worst_triangle;
      worst_proxy=proxy_id;
    }
  }
  printf("worst proxy id is %d\n",worst_proxy);
  //Remove the triangle from the old proxy
  proxies[worst_proxy]->remove(worst_triangle);

  nproxies++;
  Vec3f normal;
  triangleAreaAndNormal(worst_triangle,normal);
  Vec3f barycenter=(vert[tris[worst_triangle*3+0]]+
		    vert[tris[worst_triangle*3+1]]+
		    vert[tris[worst_triangle*3+2]])*(1.0/3.0);
  Proxy * proxy=new Proxy(normal,barycenter);
  proxy->add(worst_triangle);
  proxies.push_back(proxy);
}

/** Perform an iteration of the VSA algorithm.
 *  Given the current regions, find the best fitting proxy.
 *  Then add the best triangle and its associated error
 *  to a global heap and perform the flood fill operation
 */
void VSA::iteration()
{
  vector<TriError> heap;
  //fit new proxy for each region
  for(int proxy_id=0;proxy_id<proxies.size();proxy_id++){
    proxy_triangles & triangles=(proxies[proxy_id]->triangles);
    Vec3f norm(0,0,0);
    Vec3f center(0,0,0);
    double totalWeight=0;
    proxy_triangles::iterator finger=triangles.begin();
    proxy_triangles::iterator end=triangles.end();
    while(finger!=end){
      int tri=*finger;
      Vec3f thisNorm;
      double a=triangleAreaAndNormal(tri,thisNorm);
      thisNorm*=a;
      norm+=thisNorm;
      Vec3f thisCenter=(vert[tris[3*tri+0]]+
			vert[tris[3*tri+1]]+
			vert[tris[3*tri+2]])*(1.0/3.0);
      thisCenter*=a;
      totalWeight+=a;
      center+=thisCenter;
      finger++;
    }
    center/=totalWeight;
    norm.normalize();
    proxies[proxy_id]->x=center;
    proxies[proxy_id]->n=norm;
    proxies[proxy_id]->total_weight=totalWeight;
    //Find best triangle
    int besttri=0;
    float besterror=10e10;
    finger=triangles.begin();
    while(finger!=end){
      int tri=*(finger++);
      double thiserror=(this->*error_func)(*(proxies[proxy_id]),tri);
      
      if(thiserror<besterror){
	besterror=thiserror;
	besttri=tri;
      }
    }
    heap.push_back(TriError(besttri,proxy_id,(this->*error_func)(*(proxies[proxy_id]),besttri)));
  }
  //printf("iteration heap contains %d elements\n",heap.size());
  // for(int i=0;i<heap.size();i++){
  //  printf("[tri:%d,proxy:%d,error:%f]\n",heap[i].tri,heap[i].proxy,heap[i].error);
  //}

  flood_fill(heap);
}

/** Initialize the proxies by randomly selecting 
    triangles and fitting a proxy, then floodfilling.
 */
void VSA::initProxies()
{
  printf("initing proxies, ntri=%d\n",ntris);
  if(tri2proxy==0)
    tri2proxy=new int[ntris];
    
  if(nproxies>ntris){
    printf("WARNING: nproxies>ntri\n");
    nproxies=ntris/2;
  }

  for(int i=0;i<proxies.size();i++)
    delete proxies[i];
  proxies.clear();
  memset(tri2proxy,-1,sizeof(int)*ntris);
  for(int i=0;i<nproxies;i++){
    while(1){
      int p=rand()%ntris;
      if(tri2proxy[p]==-1){
	int v0=tris[3*p+0];
	int v1=tris[3*p+1];
	int v2=tris[3*p+2];

	Vec3f  vec1=vert[v1]-vert[v0];
	Vec3f  vec2=vert[v2]-vert[v1];
	Vec3f  n=(vec1.cross(vec2));
	n.normalize();
	if(n.dot(Vec3f(0,-1,0))!=1){
	  Vec3f x=(vert[v0]+vert[v1]+vert[v2])*(1.0/3.0);
	  Proxy * proxy=new Proxy(n,x);
	  proxy->add(p);
	  proxies.push_back(proxy);
	  tri2proxy[p]=i;
	  break;
	}
      }
    }
  }
  //Flood fill
  
  //initialize the heap
  vector<TriError> heap;
  for(int i=0;i<proxies.size();i++){
    int tri=*(proxies[i]->triangles.begin());
    float err=(this->*error_func)(*proxies[i],tri);
    heap.push_back(TriError(tri,i,(this->*error_func)(*proxies[i],tri)));
  }

  flood_fill(heap);
}

