#ifndef VSA_H
#define VSA_H

#include <getopt.h>
#include <vector>
#include <ext/hash_map>
#include <list>
#include <algorithm>
#include <vec.h>


typedef std::list<int> proxy_triangles;

class Proxy{
public:
  Vec3f n;
  Vec3f x;
  double total_weight;
  proxy_triangles triangles;
  Proxy(const Vec3f & n=Vec3f(1,0,0),Vec3f x=Vec3f(0,0,0)){
    this->n=n;
    this->x=x;
    this->triangles.clear();
    total_weight=1;
  }
  void clear(){
    triangles.clear();
  }
  void remove(int tri){
    triangles.remove(tri);
  }
  void add(int tri){
    triangles.push_back(tri);
  }
};

class TriError{
public:
  int tri;
  int proxy;
  float error;
  TriError(int t=0,int p=0,float e=10e10){
    tri=t;
    proxy=p;
    error=e;
  }
  int operator<(const TriError & other) const{
    return error>=other.error;
  }
};


class VSA{
public:
  int nproxies;
  float (VSA:: * error_func)(Proxy & p,int tri);
  std::vector<Proxy *> proxies;
  int * tri2proxy;

  int * tris;
  Vec3f * vert;
  int * opp;
  int ntris;

  VSA(Vec3f * vert=0,
      int * tris=0,
      int * opp=0,
      int ntris=0,
      int nproxies=40){
    printf("VSA(ntris:%d,%d)\n",ntris,nproxies);
   
    tri2proxy=0;
    this->nproxies=nproxies;

    this->tris=0;
    this->opp=0;
    this->ntris=0;
    
    init(vert,tris,opp,ntris);
  }
  float L21_error_func(Proxy & p,int tri)
  {
    Vec3f norm;
    double area=triangleAreaAndNormal(tri,norm);
    double ndiff=(norm-p.n).len();
    ndiff*=ndiff*area;
    return ndiff;
  }
  void init(Vec3f * vert,int * tris,int * opp,int ntris){
    release();
    error_func=&VSA::L21_error_func;
    
    this->vert=vert;
    this->tris=tris;
    this->opp=opp;
    this->ntris=ntris;

    tri2proxy=new int[ntris];
    memset(tri2proxy,-1,sizeof(int)*ntris);
    initProxies();
  }
  /*
  void init(DefMesh * defmesh){
    printf("in vsa init dm->ntri: %d\n",defmesh->ntri);
    release();
    error_func=&VSA::L21_error_func;
    proxies.clear();
    if(defmesh){
      dm=defmesh;
      tri2proxy = new int[dm->ntri];
      memset(tri2proxy,-1,sizeof(int)*dm->ntri);
      this->dm=dm;
      initProxies();
    }
    }*/
  void release(){
    for(int i=0;i<(int)proxies.size();i++)
      delete proxies[i];
    proxies.clear();
    delete [] tri2proxy;
    tri2proxy=0;


    ntris=0;
    tris=0;
    opp=0;
  }
  ~VSA(){
    release();
  }
  void find_planes(){
    for(int its=0;its<2;its++){
      for(int i=0;i<5;i++){
	iteration();
      }
      remove_region();
      iteration();
      add_region();
      iteration();
    }
  }
  void add_region();
  void remove_region();
  void iteration();
  void flood_fill(std::vector<TriError> & heap);
  void initProxies();
  float triangleAreaAndNormal(int tri,Vec3f & norm);
};

#endif
