#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include "chart.h"
#include <wx/wx.h>
#include <wx/spinctrl.h>
#include "texgen.h"
//#include "defmesh.h"
#include "vsa.h"
#include "plugin_defines.h"

#include <wx/progdlg.h>

#define PLUGIN_PROGRESS

using namespace std;
using namespace __gnu_cxx;

wxPanel * panel=0;
wxSpinCtrl * nfaces=0;
extern "C" const char *  getName(){
  return "VSA Split Charts !!";
}
extern "C" const int  getType(){
  return TEXPLUGIN_SPLIT_TYPE;
}
extern "C" const char *  getInfo(){
  return "Split... the charts";
}
extern "C" int apply(Chart ** charts,int * ncharts){
  if(*charts && *ncharts==1){
#ifdef PLUGIN_PROGRESS    
    wxBeginBusyCursor();
    wxProgressDialog   progress(wxT("Splitting Progress"),wxT(""),nfaces->GetValue()+1,NULL,
				wxPD_ELAPSED_TIME|wxPD_REMAINING_TIME|wxPD_APP_MODAL|wxPD_AUTO_HIDE);
    progress.Show(TRUE);
#endif
    int newncharts=0;

    /*
    printf("saving the object\n");
    char tmpdir[1024]="/tmp/";
    char name[1024];
#ifdef WIN32
    GetTempPath(sizeof(tmpdir),tmpdir);
#endif
    //wxMessageBox("Saving the object","Saving");
    snprintf(name,1024,"%s/_obj_remesh_temp_obj_",tmpdir);
    (*charts)[0].saveobj(name);
    DefMesh dm;
    dm.loadobj(name,0);
    */
    //printf("loading the object %d\n",dm.ntri);
    
    //wxMessageBox("Loading the object","Loading");
#ifdef PLUGIN_PROGRESS
    progress.Update(1,wxT("Getting Proxies"));
#endif
    int * tris,* opp,ntris;
    vector<Vec3f> vert;
    
    Chart * c=(charts[0]);
    ntris=c->npoly;
    tris=new int[ntris*3];
    opp=new int[ntris*3];
    int * mapsto=new int[c->nvert];
    memset(mapsto,0xFF,sizeof(int)*c->nvert);
    hash_map<int,int> edge2face;
    for(int i=0;i<c->nvert;i++){
      for(int j=0;j<vert.size();j++){
	double d=(vert[j]-c->vert[i]).len();
	if(d<1e-8){
	  mapsto[i]=j;
	  break;
	}
      }
      if(mapsto[i]<0){
	mapsto[i]=vert.size();
	vert.push_back(c->vert[i]);
      }
    }
    int nvert=vert.size();
    for(int i=0;i<c->npoly;i++){
      for(int j=0;j<3;j++){
	tris[3*i+j]=mapsto[c->poly[i][j]];
      }
      for(int j=0;j<3;j++){
	int edge=tris[3*i+j]*nvert+tris[3*i+(j+1)%3];
	edge2face[edge]=i;
      }
    }
    int haveopp=0;
    for(int i=0;i<c->npoly;i++){
      for(int j=0;j<3;j++){
	int opp_edge=tris[3*i+(j+1)%3]*nvert+tris[3*i+j];
	if(edge2face.count(opp_edge)){
	  opp[3*i+j]=edge2face[opp_edge];
	  haveopp++;
	}
	else opp[3*i+j]=-1;
      }
    }
    printf("found %d/%d opps\n",haveopp,3*ntris);
    delete [] mapsto;
    
    Vec3f * vcopy=new Vec3f[vert.size()];
    
    for(int i=0;i<vert.size();i++)vcopy[i]=vert[i];
    VSA vsa(vcopy,tris,opp,ntris,nfaces->GetValue());
    vsa.find_planes();
    
    delete [] vcopy;

    Chart * newCharts=new Chart[vsa.nproxies];
    delete [] (*charts);
    *ncharts=vsa.nproxies;
    *charts=newCharts;
    
    int * vmapsto=new int[vert.size()];
    for(int i=0;i<vsa.nproxies;i++){
#ifdef PLUGIN_PROGRESS
      wxString mesg;
      mesg.Printf(wxT("Mapping Chart %d/%d"),i+1,vsa.nproxies);
      progress.Update(i+1,mesg);
#endif

      
      int uniqueVert=0;
      memset(vmapsto,-1,sizeof(int)*vert.size());
      proxy_triangles::iterator finger=vsa.proxies[i]->triangles.begin();
      proxy_triangles::iterator end=vsa.proxies[i]->triangles.end();
      for(;finger!=end;finger++){
	int poly=*finger;
	for(int k=0;k<3;k++){
	  int vind=tris[3*poly+k];
	  if(vmapsto[vind]==-1){
	    vmapsto[vind]=uniqueVert++;
	  }
	}
      }
      (*charts)[i].init(vsa.proxies[i]->triangles.size(),uniqueVert);
      printf("mapping chart %d, tris:%d, vert:%d\n",i,vsa.proxies[i]->triangles.size(),uniqueVert);
      finger=vsa.proxies[i]->triangles.begin();
      int ntri=0;
      for(;finger!=end;finger++){
	int poly=*finger;
	for(int k=0;k<3;k++){
	  int vind=tris[3*poly+k];
	  (*charts)[i].poly[ntri][k]=vmapsto[vind];
	  (*charts)[i].vert[vmapsto[vind]]=vert[vind];
	  //(*charts)[i].norm[vmapsto[vind]]=norm[vind];
	}
	ntri++;
      }
      (*charts)[i].initNormals();
    }  
    delete [] tris;
    delete [] opp;
    delete [] vmapsto;
#ifdef PLUGIN_PROGRESS
    //progress.Show(FALSE);
    wxEndBusyCursor();
#endif
  }
  return 0;
}

extern "C" const void * getPanel(wxWindow * parent){
  if(!panel){
    panel=new wxPanel(parent);
    wxSizer * vsizer = new wxBoxSizer(wxVERTICAL);
    wxSizer * sizer = new wxBoxSizer(wxHORIZONTAL);
    sizer->Add(new wxStaticText(panel,-1,wxT("# planes")));
    sizer->Add(nfaces = new wxSpinCtrl(panel,-1),1,wxGROW);
    nfaces->SetRange(4,500);
    nfaces->SetValue(20);
    vsizer->Add(sizer,0);
    panel->SetSizer(vsizer);
  }
  return panel;
}
