#include "savedlg.h"
#include "glcanvas.h"

IMPLEMENT_DYNAMIC_CLASS(SaveDialog, wxDialog)
  BEGIN_EVENT_TABLE(SaveDialog, wxDialog)
  EVT_BUTTON(ID_Save,SaveDialog::OnSave)
  EVT_BUTTON(ID_Close,SaveDialog::OnClose)
  EVT_SPINCTRL(1,SaveDialog::OnSpin)
  EVT_RADIOBUTTON(-1,SaveDialog::OnRadio)
  //EVT_SPIN_DOWN(1,SaveDialog::OnSpinDown)
END_EVENT_TABLE()

extern GLCanvas * glCanvas;
extern wxListBox * imageList;
extern wxListBox * calibImageList;
extern wxListBox * segImageList;


SaveDialog::SaveDialog(wxWindow * parent,wxWindowID id):wxDialog(parent,id,wxT("Segment"),
								 wxPoint(-1,-1),
								 wxSize(350,240)){
  wxBoxSizer * top = new wxBoxSizer(wxHORIZONTAL);
  wxBoxSizer * bottom = new wxStaticBoxSizer(
					     new wxStaticBox(this,-1,wxT("Lookup Tables")),
					     wxHORIZONTAL);
 
  wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);
  wxBoxSizer * row;
  wxBoxSizer * tsgroup = 
    new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Texture size")),wxVERTICAL);
  wxBoxSizer * nvgroup = 
    new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("# Vec")),wxVERTICAL);
  
  save=new wxButton(this,ID_Save,wxT("Save"));
  cancel=new wxButton(this,ID_Close,wxT("Cancel"));

  int maxnvec=GetSegImageCount();
  maxnvec=maxnvec-maxnvec%4;

  for(int i=0;i<3;i++){
    nvec[i]=new wxSpinCtrl(this,1);
    nvec[i]->SetRange(4,maxnvec);
    nvec[i]->SetValue(maxnvec);
  }
  for(int i=0;i<3;i++){
    texw[i]=new wxSpinCtrl(this,1);
  }
  for(int i=0;i<3;i++){
    texh[i]=new wxSpinCtrl(this,1);
    int mx=(i)?256:1024;
    
    texw[i]->SetRange(16,mx);
    texh[i]->SetRange(16,mx);

    texw[i]->SetValue(256);
    texh[i]->SetValue(256);
  }
  
  char * names[3]={"Y","U","V"};
  for(int i=0;i<3;i++){
    row = new wxBoxSizer(wxHORIZONTAL);
    row->Add(new wxStaticText(this,-1,wxString::FromAscii(names[i])));
    row->Add(texw[i],0,wxGROW|wxEXPAND);
    row->Add(texh[i],0,wxGROW|wxEXPAND);
    tsgroup->Add(row);
  }
  top->Add(tsgroup,2,wxGROW|wxEXPAND);

  //Number of Vectors 
  for(int i=0;i<3;i++){
    row = new wxBoxSizer(wxHORIZONTAL);
    row->Add(new wxStaticText(this,-1,wxString::FromAscii(names[i])));
    row->Add(nvec[i],0,wxGROW|wxEXPAND);
    nvgroup->Add(row);
  }
  top->Add(nvgroup,1,wxGROW|wxEXPAND);

    
  wxBoxSizer * lsgroup = 
    new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Lut Size")),wxHORIZONTAL);
  
  lsgroup->Add(lutw=new wxTextCtrl(this,-1,wxT("32"),wxDefaultPosition,wxSize(40,20)));
  lsgroup->Add(luth=new wxTextCtrl(this,-1,wxT("32"),wxDefaultPosition,wxSize(40,20)));
 
  cubegroup=
    new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Cube Map Lut")),wxVERTICAL);
  cubegroup->Add(cubebut=new wxRadioButton(this,-1,wxT("Use Cube Map lut")));
  cubegroup->Add(cubemethod=new wxComboBox(this,-1));
  cubemethod->Append(wxT("Nearest"));
  cubemethod->Append(wxT("Global"));
  cubemethod->Append(wxT("Triangle"));
  cubemethod->SetSelection(2);
  
  origgroup=
    new wxStaticBoxSizer(new wxStaticBox(this,-1,wxT("Original Lut")),wxVERTICAL);
  origgroup->Add(origbut=new wxRadioButton(this,-1,wxT("Use original lut")));
  origgroup->Add(origmethod=new wxComboBox(this,-1));
  origmethod->Append(wxT("Linear"));
  origmethod->Append(wxT("Triangle"));
  origmethod->Append(wxT("Cubic"));
  origmethod->Append(wxT("V4"));

  origmethod->SetSelection(0);
  //origgroup->Add(force1dim=new wxCheckBox(this,-1,"Force 1 Dim."));

  cubebut->SetValue(0);
  origbut->SetValue(1);
  cubegroup->GetStaticBox()->Disable();
  cubemethod->Disable();
  origmethod->Enable();
  origgroup->GetStaticBox()->Enable();
  
  /*
  cubebut->SetValue(1);
  origmethod->Disable();
  //force1dim->Disable();
  origgroup->GetStaticBox()->Disable();
  */

  bottom->Add(lsgroup,0,wxGROW|wxEXPAND);
  bottom->Add(cubegroup,0,wxGROW|wxEXPAND);
  bottom->Add(origgroup,0,wxGROW|wxEXPAND);
  
  row = new wxBoxSizer(wxHORIZONTAL);
  row->Add(save);
  row->Add(cancel);

  all->Add(top,3,wxGROW|wxEXPAND);
  all->Add(bottom);
  all->Add(row,0,wxEXPAND);
 
  SetSizer(all);
  Layout();
#if wxUSE_TOOLTIPS

#endif 	
}

/** This is called whenever the number in 'any' of the spin boxes has changed.
 * Annoyingly, all below is necessary to avoid recursive calls
 * when ensuring that the U and V components will fit into the
 * Y texture.  Also, the number of eigenvectors are clamped to be 
 * divisible by 4 in this function.
*/
void SaveDialog::OnSpin(wxSpinEvent & event)
{
  int i=0;
  int wasnvec=0;
  int changed=1;
  int newval=0;
  int wasw=0;
  wxSpinCtrl * obj = (wxSpinCtrl *)event.GetEventObject();
  printf("%d %d %d\n",((wxSpinCtrl *)obj)->GetValue(),
	 event.GetPosition(),obj->GetMax());
  event.Veto();
  for(i=0;i<3;i++){
    if(nvec[i]==obj){
      int val=obj->GetValue();
      if(obj->GetValue()%4==1){
	newval=obj->GetValue()+3;
	obj->SetValue(newval);
      }
      else{
	newval=4*(obj->GetValue()/4);
	obj->SetValue(newval);
      }
      wasnvec=1;
      break;
    }
    else if(texw[i]==obj){
      int val=obj->GetValue();
      int num=16;
      wasw=1;
      while(num<val){
	num<<=1;
      }
      int diff=num-val;
      if(diff==1){
	newval=num>>1;
	obj->SetValue(newval);
      }
      else if(diff!=0){
	newval=num;
	obj->SetValue(newval);
      }
      else changed=0;
      break;
    }
    else if(texh[i]==obj){
      int val=obj->GetValue();
      int num=16;
      wasw=0;
      while(num<val){
	num<<=1;
      }
      int diff=num-val;
      if(diff==1){
	newval=num>>1;
	obj->SetValue(newval);
      }
      else if(diff!=0){
	newval=num;
	obj->SetValue(newval);
      }
      else changed=0;
      break;
    }
  }
  if(i==0 && !wasnvec && changed){
    for(int i=1;i<3;i++){
      printf("setting range %d %d\n",16,newval);
      if(wasw){
	int v=texw[i]->GetValue();
	texw[i]->SetRange(16,newval);
	texw[i]->SetValue(v);
      }
      else{
	int v=texh[i]->GetValue();
	texh[i]->SetRange(16,newval);
	texh[i]->SetValue(v);
      }
    }
  }
}

void SaveDialog::OnRadio(wxCommandEvent & event)
{
  if(event.GetEventObject()==cubebut){
    origgroup->GetStaticBox()->Disable();
    origmethod->Disable();
    //force1dim->Disable();
    
    cubegroup->GetStaticBox()->Enable();
    cubemethod->Enable();
  }
  else{
    cubegroup->GetStaticBox()->Disable();
    cubemethod->Disable();
    
    origmethod->Enable();
    //force1dim->Enable();
    origgroup->GetStaticBox()->Enable();
  }
}
