#ifndef SAVE_DLG_H
#define SAVE_DLG_H

#include <wx/wx.h>
#include <wx/notebook.h>
#include <wx/spinbutt.h>
#include <wx/spinctrl.h>
#include "guitypes.h"
#include "image.h"
#include "dynatex.h"

/**Class allowing for selection of parameters for model export.
 *  The save dialog allows the user to select the number
 * of eigenvectors, texture size, lookup table size, and lookup
 * table method for an exported model.
 */
class SaveDialog : public wxDialog{
    wxButton * save;
    wxButton * cancel;
    wxSpinCtrl * nvec[3];
    wxSpinCtrl * texw[3];
    wxSpinCtrl * texh[3];
    wxTextCtrl * lutw,*luth;
    wxRadioButton * cubebut,* origbut;
    wxComboBox * cubemethod,* origmethod;
    wxCheckBox * force1dim;
    wxStaticBoxSizer * origgroup, * cubegroup;
 public:
    DECLARE_DYNAMIC_CLASS(SaveDialog)

    SaveDialog(wxWindow * parent=0,wxWindowID id=-1);
    ~SaveDialog(){
      printf("Destroying save dialog\n");
    }
    void OnSpin(wxSpinEvent & event);
    void OnRadio(wxCommandEvent & event);
    void OnSave(wxCommandEvent & event){
	EndModal(wxID_OK);
    }
    void OnClose(wxCommandEvent & event){
	EndModal(FALSE);
    }
    ///Obtain the parameters from the dialog.
    void GetParams(int * nreig,
		   int * tw,int * th,
		   int & lutw,int & luth,
		   int & flags){
      flags=0;
      for(int i=0;i<3;i++){
	char str[1024];
	snprintf(str,1024,"%d %d %d",
		 nvec[i]->GetValue(),
		 texw[i]->GetValue(),
		 texh[i]->GetValue());
	sscanf(str,"%d %d %d",nreig+i,tw+i,th+i);
      }
      sscanf((const char*)this->lutw->GetLineText(0).ToAscii(),"%d",&lutw);
      sscanf((const char*)this->luth->GetLineText(0).ToAscii(),"%d",&luth);
      if(cubebut->GetValue()){
	wxString str=cubemethod->GetStringSelection();
	flags=LUT_CUBE;
	if(str==wxT("Nearest")){
	  flags|=LUT_NEAREST;
	}
	else if(str==wxT("Global")){
	  flags|=LUT_GLOBAL;
	}
	else if(str==wxT("Triangle")){
	  flags|=LUT_TRIANGLE;
	}
	else{
	  printf("invalid name in cubemethod combo\n");
	  exit(0);
	}
      }
      else{
	wxString str=origmethod->GetStringSelection();
	if(str==wxT("Nearest")){
	  flags|=LUT_NEAREST;
	}
	else if(str==wxT("Linear")){
	  flags|=LUT_LINEAR;
	  printf("FIXME: forcing lut height to 2\n");
	  luth=2;
	}
	else if(str==wxT("Global")){
	  flags|=LUT_GLOBAL;
	}
	else if(str==wxT("Triangle")){
	  flags|=LUT_TRIANGLE;
	}
	else if(str==wxT("Cubic")){
	  flags|=LUT_CUBIC;
	}
	else if(str==wxT("V4")){
	  flags|=LUT_V4;
	}
	else{
	  printf("invalid name in orig method combo\n");
	  exit(0);
	}
      }
      printf("flags (cube:%d,glob:%d,near:%d,tri:%d,lin:%d,cubic:%d,v4;%d)\n",
	     (flags&LUT_CUBE)>0,
	     (flags&LUT_GLOBAL)>0,
	     (flags&LUT_NEAREST)>0,
	     (flags&LUT_TRIANGLE)>0,
	     (flags&LUT_CUBIC)>0,
	     (flags&LUT_V4)>0);
	
    }
 protected:
    DECLARE_EVENT_TABLE();
};
#endif
