#include <wx/wx.h>
#include <wx/glcanvas.h>
#include <wx/spinctrl.h>
#include <wx/notebook.h>
#include <stdio.h>
#include <stdlib.h>
#include "Timer.h"



class OptionsFrame : public wxDialog{
  enum {OPT_GRAY=wxID_HIGHEST,
	OPT_RED,OPT_GREEN,OPT_BLUE,
	OPT_SMALL,
	OPT_LARGE,OPT_SIZE};
  wxRadioButton * hardwareRB;
  wxRadioButton * softwareRB;
  wxCheckBox * mmxRB;
 public:
  OptionsFrame(wxWindow * parent):wxDialog(parent,-1,"Options",
					   wxPoint(-1,-1),
					   wxSize(230,180)){
    wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer * top = new wxBoxSizer(wxHORIZONTAL);
    
    wxStaticBox * renderModeBox = new wxStaticBox(this,
                                                  -1,
                                                  "Rendering Mode");
    wxStaticBox * generalBox = new wxStaticBox(this,
					       -1,
					       "General");
    wxBoxSizer * renderMode = new wxStaticBoxSizer(renderModeBox,
						   wxVERTICAL);
    wxBoxSizer * quality = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer * general = new wxStaticBoxSizer(generalBox,
						wxVERTICAL);
    
    hardwareRB = new wxRadioButton(this,OPT_HW,"Hardware",
				   wxPoint(-1,-1));
    softwareRB = new wxRadioButton(this,OPT_SW,"Software",
				   wxPoint(-1,-1));
    mmxRB = new wxCheckBox(this,OPT_MMX,"MMX",
			   wxPoint(-1,-1));
    
    wxCheckBox * freezeTexRB = new wxCheckBox(this,OPT_FREEZE,
					      "Freeze Texture",
					      wxPoint(-1,-1));
    wxCheckBox * staticTexRB = new wxCheckBox(this,OPT_STATIC,
					      "Static Texture",
					      wxPoint(-1,-1));
        char strings[3][7]={"High","Medium","Low"};
        wxStaticBox * basisRatioBox = new wxStaticBox(this,
                                                      -1,
                                                      "Basis Ratio");
        wxStaticBox * textureResBox = new wxStaticBox(this,
                                                      -1,
                                                      "Texture Res.");
        wxBoxSizer * basisRatioSizer = new wxStaticBoxSizer(basisRatioBox,
                                                            wxVERTICAL);
        wxBoxSizer * textureResSizer = new wxStaticBoxSizer(textureResBox,
                                                            wxVERTICAL);

        wxSlider * basisRatioSlider = new wxSlider(this,OPT_NBASIS,100,0,100,
                                                   wxPoint(-1,-1),
                                                   wxSize(-1,-1));
                                                                 
        wxComboBox * textureResCombo = new wxComboBox(this,OPT_TEXRES,
                                                      strings[0],
                                                      wxPoint(-1,-1),
                                                      wxSize(-1,-1),0,0,
                                                      wxCB_READONLY );
        textureResCombo->Append(strings[0]);
        textureResCombo->Append(strings[1]);
        textureResCombo->Append(strings[2]);
 
        renderMode->Add(hardwareRB,1,wxGROW,0);
        renderMode->Add(softwareRB,1,wxEXPAND,0);
        renderMode->Add(mmxRB,1,wxLEFT|wxEXPAND|wxALIGN_RIGHT,10);

        basisRatioSizer->Add(basisRatioSlider,1,wxGROW);
        textureResSizer->Add(textureResCombo,1);
 quality->Add(basisRatioSizer,1,wxGROW);
        quality->Add(textureResSizer,1,wxGROW);

        general->Add(freezeTexRB,1);
        general->Add(staticTexRB,1);

        top->Add(renderMode,1,wxEXPAND);
        top->Add(quality,1,wxEXPAND);

        all->Add(top,1,wxEXPAND);
        all->Add(general,1,wxEXPAND);

        SetSizer(all);
        SetAutoLayout(TRUE);
        Layout();

#if wxUSE_TOOLTIPS
        basisRatioSlider->SetToolTip(_T("Percentage of basis vectors to render 
ith"));
        textureResCombo->SetToolTip(_T("Set the dynamic texture resolution"));
        staticTexRB->SetToolTip(_T("Toggle static texture"));
        freezeTexRB->SetToolTip(_T("Freeze the dynamic texture"));
        softwareRB->SetToolTip(_T("Render dynamic texture using sofware"));
        hardwareRB->SetToolTip(_T("Render dynamic texture using hardware"));
        mmxRB->SetToolTip(_T("Toggle MMX acceleration in software mode"));

#endif 

    }
    void OnClose(wxCloseEvent * event){
        printf("closing options frame\n");
        Show(FALSE);
    }
    ~OptionsFrame(){
        printf("Deleting options frame\n");
    }
    void NBasis(wxScrollEvent & event){
        printf("Slider event\n");
    }
    void OnHW(wxCommandEvent & event){
        printf("hw\n");
        mmxRB->Disable();
    }
void OnSW(wxCommandEvent & event){
        printf("sw\n");
        mmxRB->Enable();
    }
    void OnMMX(wxCommandEvent & event){

        printf("mmx\n");
    }
    void OnTexRes(wxCommandEvent & event){
        printf("on texture res\n");
    }
    void OnStatic(wxCommandEvent & event){
        printf("static\n");
    }
    void OnFreeze(wxCommandEvent & event){
        printf("freeze\n");
    }
    
    protected:
    DECLARE_EVENT_TABLE();
};
BEGIN_EVENT_TABLE(OptionsFrame, wxDialog)
    EVT_COMMAND_SCROLL_THUMBTRACK(OPT_NBASIS,OptionsFrame::NBasis) 
    EVT_CLOSE(OptionsFrame::OnClose)
    EVT_RADIOBUTTON(OPT_HW,OptionsFrame::OnHW)
    EVT_RADIOBUTTON(OPT_SW,OptionsFrame::OnSW)
    EVT_CHECKBOX(OPT_MMX,OptionsFrame::OnMMX)
    EVT_CHECKBOX(OPT_STATIC,OptionsFrame::OnStatic)
    EVT_CHECKBOX(OPT_FREEZE,OptionsFrame::OnFreeze)
    EVT_COMBOBOX(OPT_TEXRES,OptionsFrame::OnTexRes)
END_EVENT_TABLE()




class MyApp : public wxApp{
    MyFrame * frame;
public:
    virtual bool OnInit(){
        frame = new OptionsFrame();
        SetTopWindow(frame);
        return true;
    }
    MyApp(){

    }
    
    void OnClose(wxCloseEvent & event){
        printf("on close\n");
    }
protected:
    //  DECLARE_EVENT_TABLE();
};

IMPLEMENT_APP(MyApp)


