#include <wx/wx.h>
#include <wx/listctrl.h>
#include <wx/checklst.h>
#include <wx/event.h>
#include <wx/glcanvas.h>
#include <sys/time.h>
#include <wx/splitter.h>
#include <wx/spinbutt.h>
#include <wx/spinctrl.h>
#include "image.h"
#include "Matrix.h"
#include "newseg.h"
#include <libraw1394/raw1394.h>
#include <libdc1394/dc1394_control.h>
//#define DMA
#define YUV2RGB(y, u, v, r, g, b)\
  r = y + ((v*1436) >>10);\
  g = y - ((u*352 + v*731) >> 10);\
  b = y + ((u*1814) >> 10);\
  r = r < 0 ? 0 : r;\
  g = g < 0 ? 0 : g;\
  b = b < 0 ? 0 : b;\
  r = r > 255 ? 255 : r;\
  g = g > 255 ? 255 : g;\
  b = b > 255 ? 255 : b

void uyvy2rgb (unsigned char *src, unsigned char *dest, int NumPixels) {
  register int i = (NumPixels << 1)-1;
  register int j = NumPixels + ( NumPixels << 1 ) -1;
  register int y0, y1, u, v;
  register int r, g, b;
  
  while (i > 0)
  {
    y1 = (unsigned char) src[i--];
    v  = (unsigned char) src[i--] - 128;
    y0 = (unsigned char) src[i--];
    u  = (unsigned char) src[i--] - 128;
    YUV2RGB (y1, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
    YUV2RGB (y0, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
  }
}

void yuv411torgb (unsigned char *src, unsigned char *dest, int NumPixels) {
    register int i = (NumPixels+(NumPixels>>1))-1;
    register int j = ( NumPixels*3) -1;
    register int y0,y1,y2,y3, u, v;
    register int r, g, b;
    
    while (i > 0)
    {
	y3=(unsigned char) src[i--];
	y2=(unsigned char) src[i--];
	v =(unsigned char) src[i--] - 128;
	y1=(unsigned char) src[i--];
	y0=(unsigned char) src[i--];
	u =(unsigned char) src[i--] - 128;
	
	YUV2RGB (y3, u, v, r, g, b);
	dest[j--] = b;
	dest[j--] = g;
	dest[j--] = r;
	YUV2RGB (y2, u, v, r, g, b);
	dest[j--] = b;
	dest[j--] = g;
	dest[j--] = r;
	YUV2RGB (y1, u, v, r, g, b);
	dest[j--] = b;
	dest[j--] = g;
	dest[j--] = r;
	YUV2RGB (y0, u, v, r, g, b);
	dest[j--] = b;
	dest[j--] = g;
	dest[j--] = r;
    }
}

void yuv444torgb (unsigned char *src, unsigned char *dest, int NumPixels) {
  register int i = (NumPixels*3)-1;
  register int j = (NumPixels*3)-1;
  register int y, u, v;
  register int r, g, b;
  
  while (i > 0)
  {
      
      v  = (unsigned char) src[i--] - 128;
      y  = (unsigned char) src[i--];
      u = (unsigned char) src[i--] - 128;
           
      YUV2RGB (y, u, v, r, g, b);
      dest[j--] = b;
      dest[j--] = g;
      dest[j--] = r;
  }
}


enum {ID_About,ID_Quit,ID_ImageList,ID_CalibImageList,
      ID_Calibrate,ID_Remove,ID_RemoveAll,ID_Play,ID_Save,ID_Load,ID_Capture,
      ID_Grab,ID_Close,ID_OK,ID_Mode,ID_Format,ID_Camera};

enum {HAS_EXTRINSIC=1,HAS_INTRINSIC=2,
      HAS_DISTORTION=4,HAS_IMAGE=8};

enum {RGB,YUV444,YUV422,YUV411,MONO,MONO16};
int nchannels[]={3,3,3,3,1,1};
int bytesPerChannel[]={1,1,1,1,1,2};

void (* func[6])(unsigned char *,unsigned char *,int)={0,yuv444torgb,uyvy2rgb,yuv411torgb,0,0};
char fmt2str[][7]={"RGB","YUV444","YUV422","YUV411","MONO","MONO16"};

int format0modes[][4]={{160,120,YUV444,64},
		      {320,240,YUV422,65},
		      {640,480,YUV411,66},
		      {640,480,YUV422,67},
		      {640,480,RGB,68},
		      {640,480,MONO,69},
		      {640,480,MONO16,70}};

int format1modes[][4]={{800,600,YUV422,96},
		      {800,600,RGB,97},
		      {800,600,MONO,98},
		      {1024,768,YUV422,99},
		      {1024,768,RGB,100},
		      {1024,768,MONO,101},
		      {800,600,MONO16,102},
		      {1024,768,MONO16,103}};

int format2modes[][4]={{1280,960,YUV422,128},
		      {1280,960,RGB,129},
		      {1280,960,MONO,130},
		      {1600,1200,YUV422,131},
		      {1600,1200,RGB,132},
		      {1600,1200,MONO,133},
		       {1280,960,MONO16,134},
		      {1600,1200,MONO16,135}};




class ImageData{
public:
    Image8 image;
    int flags;
    Matrix extrinsic;
    Matrix intrinsic;
    Matrix distortion;
    CalibInfo * calibInfo;
    ImageData(){
	double fx=363.0017;//413.1795329561925;
	double fy=363.5357;//413.3604660994160;
	double px=152.360809036324;
	double py=113.112125563360;

	distortion=Matrix(5,1);
	intrinsic=Matrix::eye(3);
	extrinsic=Matrix::eye(4);
	distortion.setAll(0);

	distortion[0]=-0.40743818922655;
	distortion[1]=0.53393818146118;
	distortion[2]=0.00036239062064;
	distortion[3]=-0.00380086091997;
	distortion[4]=0;
	
	intrinsic(0,0)=fx;
	intrinsic(1,1)=fy;
	intrinsic(0,2)=px;
	intrinsic(1,2)=py;
	intrinsic.print();
	flags=0;
	calibInfo=0;
    }
    ~ImageData(){
	delete calibInfo;
    }
};

class MyGLCanvas: public wxGLCanvas{
    CalibInfo * calibInfo;
    GLuint tex;
    int imagew;
    int imageh;
    double xsize;
    double ysize;
public:
    MyGLCanvas(wxWindow * parent):wxGLCanvas(parent,-1){
	tex=0;
	imagew=1;
	imageh=1;
	calibInfo=0;
	xsize=ysize=1;
    }
    void UpdateTexture(ImageData & imageData){
	if(tex){
	    printf("binding texture %p %p %p\n",
		   &imageData,&(imageData.image),imageData.image.data);
	    glEnable(GL_TEXTURE_2D);
	    glBindTexture(GL_TEXTURE_2D,tex);
	    imagew=imageData.image.width;
	    imageh=imageData.image.height;
	    imageData.image.initTextureNoResize(&xsize,&ysize);
	    Refresh();
	    
	}
	printf("deleting calibInfo\n");
	if(calibInfo){
	    delete calibInfo;
	    calibInfo=0;
	}
	printf("done deleete\n");
	if(imageData.calibInfo){
	    calibInfo=imageData.calibInfo->copy();
	    printf("done copy\n");
	}
    }
    void OnPaint(wxPaintEvent & event){
	static float rot=0;
	rot+=1;
	wxPaintDC dc(this);
	glClear(GL_COLOR_BUFFER_BIT);
#ifndef __WXMOTIF__
	if (!GetContext()) return;
#endif
	SetCurrent();

	if(tex==0){
	    glEnable(GL_TEXTURE_2D);
	    glGenTextures(1,&tex);
	    glDisable(GL_TEXTURE_2D);
	}

	int w,h;
	GetSize(&w,&h);
	glViewport(0,0,w,h);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,1,1,0,-100,100);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,tex);
	glColor3f(1.0,1.0,1.0);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);glVertex2f(0,0);
	glTexCoord2f(xsize,0);glVertex2f(1,0);
	glTexCoord2f(xsize,ysize);glVertex2f(1,1);
	glTexCoord2f(0,ysize);glVertex2f(0,1);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	if(calibInfo){
	    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
	    glOrtho(0,imagew,imageh,0,-100,100);
	    glMatrixMode(GL_MODELVIEW);
	    glLoadIdentity();
	    drawCalibInfo();
	}


	SwapBuffers();
    }
    void drawCalibInfo(){
	glColor3f(1.0,0.0,0.0);
	for(int i=0;i<calibInfo->nlines;i++){
	    int biggest=calibInfo->sizes[calibInfo->lines[i][0]];
	    if(calibInfo->sizes[calibInfo->lines[i][1]]>biggest)
		biggest=calibInfo->sizes[calibInfo->lines[i][1]];
	    if(calibInfo->sizes[calibInfo->lines[i][2]]>biggest)
		biggest=calibInfo->sizes[calibInfo->lines[i][2]];
	    if(calibInfo->sizes[calibInfo->lines[i][3]]>biggest)
		biggest=calibInfo->sizes[calibInfo->lines[i][3]];
	    int middle=(int)(0.75*biggest);
	    
	    for(int j=0;j<4;j++){
		if(calibInfo->sizes[calibInfo->lines[i][j]]>middle){
		    glColor3f(0.2,0.2,0.6);
		    glPointSize(3);
		}
		else{
		    glColor3f(0.8,1.0,0.8);
		    glPointSize(3);
		}
		glBegin(GL_POINTS);
		glVertex2d(calibInfo->centers[calibInfo->lines[i][j]],
			   calibInfo->centers[calibInfo->lines[i][j]+calibInfo->ncenters]);
		glEnd();
		
	    }
	    glLineWidth(2);
	    glColor3f(0.6,0.6,0.6);
	    glBegin(GL_LINES);
	    glVertex2d(calibInfo->centers[calibInfo->lines[i][0]],
		       calibInfo->centers[calibInfo->lines[i][0]+calibInfo->ncenters]);
	    glVertex2d(calibInfo->centers[calibInfo->lines[i][3]],
		       calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters]);
	    glEnd();
	    
	    if(!calibInfo->labels){
		drawString(calibInfo->centers[calibInfo->lines[i][3]],
			   calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters],
			   "    %d\n",i);
	    }
	    else{
		if(calibInfo->labels[i]!=calibInfo->labels2[i])
		    glColor3f(0.1,1.0,0.1);
		drawString(calibInfo->centers[calibInfo->lines[i][3]],
			   calibInfo->centers[calibInfo->lines[i][3]+calibInfo->ncenters],
			   "   %d,%d,%d",i,calibInfo->labels[i],calibInfo->labels2[i]);
	    }
/*
	    printf("%d:%d,%d:%d,%d:%d,%d:%d\n",
		   calibInfo->lines[i][0],calibInfo->sizes[calibInfo->lines[i][0]],
		   calibInfo->lines[i][1],calibInfo->sizes[calibInfo->lines[i][1]],
		   calibInfo->lines[i][2],calibInfo->sizes[calibInfo->lines[i][2]],
		   calibInfo->lines[i][3],calibInfo->sizes[calibInfo->lines[i][3]]);*/
	}
	
	glBegin(GL_POINTS);
	glColor3f(0.0,0.0,1.0);
	glVertex2f(calibInfo->center[0],calibInfo->center[1]);
	glEnd();
    }
    void OnEraseBackground(wxEraseEvent & event){
    }
    void OnSize(wxSizeEvent & event){
    }
    void OnEnterWindow(wxMouseEvent & event){
    }
    void OnKeyDown(wxKeyEvent & event){
	printf("canvas on key down %d %c\n",event.GetKeyCode(),event.GetKeyCode());
    }
    void OnKeyUp(wxKeyEvent & event){
	printf("canvas on key up\n");
    }
protected:
    DECLARE_EVENT_TABLE();
};
MyGLCanvas * glCanvas;
wxListBox * imageList;
wxListBox * calibImageList;

BEGIN_EVENT_TABLE(MyGLCanvas, wxGLCanvas)
    EVT_KEY_DOWN(MyGLCanvas::OnKeyDown )
    EVT_KEY_UP( MyGLCanvas::OnKeyUp )
    EVT_SIZE(MyGLCanvas::OnSize)
    EVT_PAINT(MyGLCanvas::OnPaint)
    EVT_ERASE_BACKGROUND(MyGLCanvas::OnEraseBackground)
    EVT_ENTER_WINDOW( MyGLCanvas::OnEnterWindow )
END_EVENT_TABLE()


class CalibrateDialog : public wxDialog{
    enum {OPT_NBASIS=wxID_HIGHEST,
	  OPT_HW,OPT_SW,OPT_MMX,
          OPT_SMALL,OPT_BIG,OPT_GRAY};
    

    wxCheckBox * gray;
    wxSlider * green, * blue, * red;
    wxSpinCtrl * small, * big;
public:
    CalibrateDialog(wxWindow * parent,wxWindowID id):wxDialog(parent,id,"Options",
							      wxPoint(-1,-1),
							      wxSize(230,200)){
	wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);

	wxStaticBox * threshBox = new wxStaticBox(this,
						  -1,
						  "Threshold");
	wxStaticBox * generalBox = new wxStaticBox(this,
						  -1,
						  "Sizes");
	wxBoxSizer * threshBoxSizer = new wxStaticBoxSizer(threshBox,
							   wxVERTICAL);
	wxBoxSizer * buttonSizer = new wxBoxSizer(wxHORIZONTAL);

	wxBoxSizer * general = new wxStaticBoxSizer(generalBox,
						    wxVERTICAL);

	gray = new wxCheckBox(this,OPT_GRAY,"Gray",
				 wxPoint(-1,-1));
	red = new wxSlider(this,OPT_NBASIS,160,0,255,
			   wxPoint(-1,-1),
			   wxSize(-1,-1));
	green = new wxSlider(this,OPT_NBASIS,120,0,255,
			   wxPoint(-1,-1),
			   wxSize(-1,-1));
	blue = new wxSlider(this,OPT_NBASIS,160,0,255,
			   wxPoint(-1,-1),
			   wxSize(-1,-1));

	big = new wxSpinCtrl(this,OPT_BIG,"1000",wxDefaultPosition,
			     wxDefaultSize,
			     wxSP_ARROW_KEYS,0,10000,1000);
	small = new wxSpinCtrl(this,OPT_SMALL,"15",wxDefaultPosition,
			       wxDefaultSize,
			       wxSP_ARROW_KEYS,0,10000,15);

	wxBoxSizer * temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1," "),0,wxALIGN_CENTER);
	temp->Add(gray,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW);

	temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1,"R"));
	temp->Add(red,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

	temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1,"G"));
	temp->Add(green,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

	temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1,"B"));
	temp->Add(blue,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

	wxFlexGridSizer * grid = new  wxFlexGridSizer(1,2,2,2);

	grid->Add(new wxStaticText(this,-1,"Small"),0,wxALIGN_RIGHT);
	grid->Add(small,0,wxALIGN_RIGHT);
	
	grid->Add(new wxStaticText(this,-1,"Big "),0,wxALIGN_RIGHT);
	grid->Add(big,0,wxALIGN_RIGHT);

	general->Add(grid,1,wxGROW);

	buttonSizer->Add(new wxButton(this,ID_OK,"OK"),1,wxSHAPED|wxALIGN_CENTRE);
	buttonSizer->Add(new wxButton(this,ID_Close,"Close"),1,wxSHAPED|wxALIGN_CENTRE);

	all->Add(threshBoxSizer,3,wxGROW);
	all->Add(general,2,wxGROW);
	all->Add(buttonSizer,1,wxGROW|wxALIGN_CENTRE);

	SetSizer(all);
	SetAutoLayout(TRUE);
	Layout();
#if wxUSE_TOOLTIPS

#endif 
    }
    void GetThresh(int * thresh){
	thresh[0]=red->GetValue();
	thresh[1]=green->GetValue();
	thresh[2]=blue->GetValue();
    }
    int GetSmall(){
	return small->GetValue();
    }
    int GetBig(){
	return big->GetValue();
    }
    void NBasis(wxScrollEvent & event){
	wxObject * obj = event.GetEventObject();
	if(gray->GetValue()){
	    int value = ((wxScrollBar *)obj)->GetValue();
	    red->SetValue(value);
	    green->SetValue(value);
	    blue->SetValue(value);
	}
	int sel=imageList->GetSelection();
	if(sel>=0){
	    ImageData * imd=(ImageData *)imageList->GetClientData(sel);
	    if(imd){
		ImageData thresh;
	  
		double val[3]={red->GetValue(),
			       green->GetValue(),
			       blue->GetValue()};
		thresh.image=imd->image.threshLess(val)*255;
		thresh.intrinsic=imd->intrinsic.copy();
		thresh.extrinsic=imd->extrinsic.copy();
		thresh.distortion=imd->distortion.copy();
		
		double residual=calibrateImage(imd->image,
					       red->GetValue(),
					       green->GetValue(),
					       blue->GetValue(),
					       small->GetValue(),
					       big->GetValue(),
					       thresh.distortion,
					       thresh.intrinsic,
					       thresh.extrinsic);
		thresh.calibInfo = new CalibInfo();
		calibrateGetInfo(*(thresh.calibInfo));
		calibrateRelease();
		glCanvas->UpdateTexture(thresh);
	    }
	}
	glCanvas->Refresh();
    }
    void OnGray(wxCommandEvent & event){
	if(gray->GetValue()){
	    int value = red->GetValue()+green->GetValue()+blue->GetValue();
	    red->SetValue(value/3);
	    green->SetValue(value/3);
	    blue->SetValue(value/3);
	    glCanvas->Refresh();
	}
    }
    void OnSpinSmall(wxSpinEvent & event){
	glCanvas->Refresh();
    }
    void OnSpinBig(wxSpinEvent & event){
	glCanvas->Refresh();
    }
    void OnOkay(wxCommandEvent & event){
	EndModal(wxID_OK);
    }
    void OnClose(wxCommandEvent & event){
	EndModal(wxID_CANCEL);
    }
protected:
    DECLARE_EVENT_TABLE();
};
BEGIN_EVENT_TABLE(CalibrateDialog, wxDialog)
    EVT_BUTTON(ID_OK,CalibrateDialog::OnOK)
    EVT_BUTTON(ID_Close,CalibrateDialog::OnClose)
    EVT_SPIN(OPT_SMALL,CalibrateDialog::OnSpinSmall)
    EVT_SPIN(OPT_BIG,CalibrateDialog::OnSpinBig)
    EVT_COMMAND_SCROLL_THUMBTRACK(OPT_NBASIS,CalibrateDialog::NBasis)
    EVT_CLOSE(CalibrateDialog::OnClose)
    EVT_CHECKBOX(OPT_GRAY,CalibrateDialog::OnGray)
END_EVENT_TABLE()

class CaptureDialog:public wxDialog{
    int * modeInfo;
    raw1394handle_t handle;
    int numCameras; 
    nodeid_t* camera_nodes;
    dc1394_camerainfo *cameras;
    dc1394_cameracapture *captures;
    dc1394_miscinfo *misc_infos;
    wxComboBox * cameraCombo,* modeCombo,* formatCombo;
    wxSpinCtrl * nimages,* every;
    int started;
    int capturing;
    int captureCount;

    
    
public:
    CaptureDialog(wxWindow * parent,wxWindowID id):
	wxDialog(parent,id,"Capture Options",wxDefaultPosition,
		 wxSize(310,160)){
		
	cameraCombo=new wxComboBox(this,ID_Camera,"Camera 0");
	modeCombo=new wxComboBox(this,ID_Mode);
	formatCombo=new wxComboBox(this,ID_Format);
	nimages=new wxSpinCtrl(this,-1);
	every=new wxSpinCtrl(this,-1);
	wxButton * grab=
	    new wxButton(this,ID_Grab,"Grab");
	wxButton * close=
	    new wxButton(this,ID_Close,"Close");

	nimages->SetRange(0,512);
	nimages->SetValue(64);

	every->SetRange(0,512);
	every->SetValue(30);

	wxFlexGridSizer * sizer = new wxFlexGridSizer(6,2,4,4);

	sizer->Add(new wxStaticText(this,-1,"Camera",wxDefaultPosition,
				    wxDefaultSize,wxALIGN_RIGHT),
		   0,wxGROW|wxALIGN_RIGHT);
	sizer->Add(cameraCombo,0,wxEXPAND);
	sizer->Add(new wxStaticText(this,-1,"Format",wxDefaultPosition,
				    wxDefaultSize,wxALIGN_RIGHT),
		   0,wxGROW|wxALIGN_RIGHT);
	sizer->Add(formatCombo,0,wxEXPAND);
	sizer->Add(new wxStaticText(this,-1,"Mode ",wxDefaultPosition,
				    wxDefaultSize,wxALIGN_RIGHT),
		   0,wxGROW|wxALIGN_RIGHT);
	sizer->Add(modeCombo,0,wxEXPAND);
	sizer->Add(new wxStaticText(this,-1,"#Images",wxDefaultPosition,
				    wxDefaultSize,wxALIGN_RIGHT),
		   0,wxGROW|wxALIGN_RIGHT);
	sizer->Add(nimages,0,wxEXPAND);
	sizer->Add(new wxStaticText(this,-1,"Every",wxDefaultPosition,
				    wxDefaultSize,wxALIGN_RIGHT)
		   ,0,wxGROW|wxALIGN_RIGHT);
	sizer->Add(every,0,wxEXPAND);
	sizer->Add(grab,0,wxGROW);
	sizer->Add(close,0,wxGROW);
	sizer->SetMinSize(300,150);
	sizer->AddGrowableCol(1);
	SetSizer(sizer);
	//SetAutoLayout(TRUE);
	//Validate();
	started=0;
	capturing=0;
	captureCount=0;
	Init();
	
    }
    void initCapture(){
	int which = cameraCombo->GetSelection();
	unsigned int format;
	unsigned int mode;
	if(which==-1){
	    return;
	}
	if(started){
	    printf("Stopping transmission\n");
	    dc1394_stop_iso_transmission(handle,camera_nodes[which]);
	    dc1394_release_camera(handle, &captures[which]);
#ifdef DMA
	    dc1394_dma_release_camera(handle, &captures[which]);
#endif
	    started=0;
	}
	if(formatCombo->GetSelection()<0){
	    return;
	}
	wxString str =formatCombo->GetStringSelection();
	sscanf((const char *)str.ToAscii(),"Format %d",&format);
	format+=FORMAT_MIN;
	
	str =modeCombo->GetStringSelection();
	sscanf((const char *)str.ToAscii(),"Mode %d",&mode);

	printf("initing capture %d %d\n",format,mode);
	switch(format-FORMAT_MIN){
	    case 0:
		modeInfo=format0modes[mode-MODE_FORMAT0_MIN];
		break;
	}
	
#ifdef DMA
	if (dc1394_dma_setup_capture(handle, 
				     camera_nodes[which],  
				     which,
				     format,
				     mode,
				     SPEED_400,
				     FRAMERATE_15,
				     8, // number of dma buffers
				     0, // drop frames
				     "/dev/video1394", 
				     &captures[which])
	      != DC1394_SUCCESS){
	    printf("Setting Camera %d failed\n", which);
	    exit(1);
	  }
#else
	  if (dc1394_setup_capture(handle, 
				   camera_nodes[which],  
				   which,
				   format,
				   mode,
				   SPEED_400,
				   FRAMERATE_15,
				   &captures[which])
	      != DC1394_SUCCESS){
	    printf("Setting Camera %d failed\n", which);
	    exit(1);
	  }
#endif
	  if(dc1394_start_iso_transmission(handle,camera_nodes[which])
	     != DC1394_SUCCESS)
	  {
	      fprintf(stderr, "Could not start ISO transmission\n");
	      exit(1);
	  }
	  started=1;
	  printf("init capture complete\n");
    }
    void UpdateModes(){
	quadlet_t modes;
	int format;
	int camera=cameraCombo->GetSelection();
	modeCombo->Clear();
	if(camera==-1)return;
       
	wxString fmt =formatCombo->GetStringSelection();
	sscanf((const char *)fmt.ToAscii(),"Format %d",&format);
	printf("updating modes, camera=%d, format=%d\n",camera,format);
	if(DC1394_SUCCESS!=
	   dc1394_query_supported_modes(handle,
					camera_nodes[camera],
					FORMAT_MIN+format,
					&modes)){
	    
	    printf("error querying modes for camera %d\n",camera);
	}
	else{
	    switch(format){
		case 0:
		    for(int k=0;k<NUM_FORMAT0_MODES;k++){
			if(modes & 1<<(32-k-1)){
			    printf("%d supported\n",
				   MODE_FORMAT0_MIN+
				   k);
			    wxString name;
			    name.sprintf("Mode %d:%dx%d-%s",
					 format0modes[k][3],
					 format0modes[k][0],
					 format0modes[k][1],
					 fmt2str[format0modes[k][2]]);
			    modeCombo->Append(name);
			}
		    }
		    break;
	    }
	}
    }
    void UpdateFormat(){
	int i=cameraCombo->GetSelection();
	formatCombo->Clear();
	if(i==-1)return;
	
	quadlet_t value;
	if(DC1394_SUCCESS!=dc1394_query_supported_formats(handle, camera_nodes[i],
							  &value)){
	    printf("error querying formats for camera %d\n",i);
	}
	
	for(int j=0;j<7;j++){
	    if(value & 1<<(32-j-1)){
		wxString fmt;
		fmt.sprintf("Format %d",j);
		formatCombo->Append(fmt);
	    }
	}
	formatCombo->SetSelection(0);
	UpdateModes();
    }
    void Init(){ 
	handle = dc1394_create_handle(0);
	if (handle==NULL) {
	  printf("did you insmod the drivers?\n");
	  exit(1);
	}
	camera_nodes = dc1394_get_camera_nodes(handle, &numCameras, 1); 
	cameras=new dc1394_camerainfo[numCameras];
	captures=new dc1394_cameracapture[numCameras];
	misc_infos=new dc1394_miscinfo[numCameras];
	
	for(int i=0;i<numCameras; i++){
	  printf("node %d\n",camera_nodes[i]); 
	  
	  dc1394_get_camera_info(handle, camera_nodes[i], &cameras[i]);
	  dc1394_get_camera_misc_info(cameras[i].handle,cameras[i].id,
				      &misc_infos[i]);
	  dc1394_print_camera_info(&cameras[i]); 
	  
	  wxString str;
	  str.sprintf("%s",cameras[i].vendor);
	  cameraCombo->Append(str);
	  if(i==0)cameraCombo->SetSelection(0);

	  UpdateFormat();
	}
	initCapture();
    }
    void OnMode(wxCommandEvent & event){
	initCapture();
    }
    void OnFormat(wxCommandEvent & event){
	UpdateModes();
	initCapture();
    }
    void OnCamera(wxCommandEvent & event){
	UpdateFormat();
	initCapture();
    }
    void GetImageFromCamera(){
#ifdef DMA
	if(dc1394_dma_single_capture(captures) != DC1394_SUCCESS){
	    fprintf(stderr, "Could not capture multi frames\n");
	    exit(1);
	}
	dc1394_dma_done_with_buffer(&captures[0]);	
#else
	
	if(dc1394_single_capture(handle,captures) != DC1394_SUCCESS){
	    fprintf(stderr, "Could not capture multi frames\n");
	    exit(1);
	}
#endif
	printf("got capture %d %d\n",modeInfo[0],modeInfo[1]);
	Image8 im(modeInfo[0],modeInfo[1],nchannels[modeInfo[2]]);
	if(bytesPerChannel[modeInfo[2]]==2){
	    Image16 im16(modeInfo[0],modeInfo[1],nchannels[modeInfo[2]]);
	    if(func[modeInfo[2]]){
		printf("Unable to handle 16bit images with more than 1 channel\n");
	    }
	    else{
		memcpy(im16.data,captures[0].capture_buffer,modeInfo[0]*modeInfo[1]*
		       nchannels[modeInfo[2]]*bytesPerChannel[modeInfo[2]]);
	    }
	    im=im16;
	}
	else{
	    if(func[modeInfo[2]]){
		func[modeInfo[2]]((unsigned char*)captures[0].capture_buffer,
				  im.data, modeInfo[0]*modeInfo[1]);
	    }
	    else{
		memcpy(im.data,captures[0].capture_buffer,modeInfo[0]*modeInfo[1]*
		       nchannels[modeInfo[2]]*bytesPerChannel[modeInfo[2]]);
	    }
	}
	//set the gl windows texture
	{
	    ImageData imd;
	    imd.image=im;
	    glCanvas->UpdateTexture(imd);
	    glCanvas->Refresh();
	}

	if(capturing){
	    if(captureCount<nimages->GetValue()){
		ImageData * imd = new ImageData();
		imd->image=im;
		wxString name;
		name.sprintf("cap-%04d",captureCount);
		imageList->Append(name,imd);
		captureCount++;
	    }
	    else{
		capturing=0;
		captureCount=0;
	    }
	}
    }
    void OnGrab(wxCommandEvent & event){
	printf("about to capture\n");
	capturing=1;
	captureCount=0;
    }
    void OnClose(wxCommandEvent & event){
	if(started){
	    dc1394_stop_iso_transmission(handle,camera_nodes[0]);
	    dc1394_release_camera(handle, &captures[0]);
#ifdef DMA
	    dc1394_dma_release_camera(handle, &captures[0]);
#endif
	    started=0;
	}
	delete camera_nodes;
	delete cameras;
	delete captures;
	delete misc_infos;
	raw1394_destroy_handle(handle);

	EndModal(wxID_CANCEL);
    }
    void OnIdle(wxIdleEvent & event){
	if(started){
	    GetImageFromCamera();
	    event.RequestMore(TRUE);
	}
    }
    DECLARE_EVENT_TABLE();
};
BEGIN_EVENT_TABLE(CaptureDialog, wxDialog)
    EVT_BUTTON(ID_Close, CaptureDialog::OnClose)
    EVT_BUTTON(ID_Grab, CaptureDialog::OnGrab)
    EVT_COMBOBOX(ID_Mode,CaptureDialog::OnMode)
    EVT_COMBOBOX(ID_Format,CaptureDialog::OnFormat)
    EVT_COMBOBOX(ID_Camera,CaptureDialog::OnCamera)
    EVT_IDLE(CaptureDialog::OnIdle)
END_EVENT_TABLE()







class ImageList:public wxListBox{
    wxMenu * imageListMenu;
    int index;

    //current calibration parameters
    int thresh[3];
    int small;
    int big;

    enum {IDLE=0,CALIBRATING,PLAYING} state;
 public:
    DECLARE_DYNAMIC_CLASS(ImageList)
    ImageList():wxListBox(){
    }
    ImageList(wxWindow * parent):
	wxListBox(parent,ID_ImageList,wxDefaultPosition,
		  wxDefaultSize,0,0,wxLB_EXTENDED){
	printf("initing ImageList\n");
	imageListMenu = new wxMenu();
	imageListMenu->Append(ID_Remove,"Remove Selected","Delete Selected",wxITEM_NORMAL);
	imageListMenu->Append(ID_RemoveAll,"Remove All","Delete Selected",wxITEM_NORMAL);
	imageListMenu->AppendSeparator();
	imageListMenu->Append(ID_Play,"Play","Play",wxITEM_NORMAL);
	imageListMenu->AppendSeparator();
	imageListMenu->Append(ID_Load,"Load","Load Images");
	imageListMenu->Append(ID_Capture,"Capture","Capture Images");
	imageListMenu->AppendSeparator();
	imageListMenu->Append(ID_Save,"Save","Save Selected",wxITEM_NORMAL);
	imageListMenu->Append(ID_Calibrate,"Calibrate","Calibrate",wxITEM_NORMAL);
	state=IDLE;
    }
    void OnIdle(wxIdleEvent & event){
	if(state==PLAYING){
	    if(index>=GetCount()){
		state=IDLE;
		index=0;
	    }
	    else{
		static struct timeval tv;
		static int called=0;
		if(index==1&&!called){
		    gettimeofday(&tv,NULL);
		    called=1;
		}
		static struct timeval now;
		gettimeofday(&now,NULL);
	
		unsigned long long span = (now.tv_sec-tv.tv_sec)*1000+
		    (now.tv_usec-tv.tv_usec)/1000;
		
		if(span>10){
		    wxArrayInt selections;
		    GetSelections(selections);
		    for(int i=selections.GetCount()-1;i>=0;i--){
			Deselect(selections[i]);
		    }
		    gettimeofday(&tv,NULL);
		    SetSelection(index);
		    OnListSelected(wxCommandEvent());

		    index++;
		    called=0;
		}
		event.RequestMore(TRUE);
	    }
	}
	else if(state==CALIBRATING){
	    if(index>=GetCount()){
		state=IDLE;
		index=0;
	    }
	    else{
		wxArrayInt selections;
		GetSelections(selections);
		if(selections.GetCount()){
		    if(calibImageList->GetString(0)==
		       "[Calibrated Images]"){
			calibImageList->Clear();
		    }
		    printf("calibrating %d\n",selections.GetCount());
		    
		    for(int i=0;i<selections.GetCount();i++){
			calibImageList->Append(GetString(selections[i]),
					       GetClientData(selections[i]));
			if(GetClientData(selections[i])){
			    ImageData * imd = (ImageData *) 
				GetClientData(selections[i]);
			    double residual=calibrateImage(imd->image,
							   thresh[0],
							   thresh[1],
							   thresh[2],
							   small,big,
							   imd->distortion,
							   imd->intrinsic,
							   imd->extrinsic);
			    if(imd->calibInfo){
				delete imd->calibInfo;
			    }
			    imd->calibInfo = new CalibInfo();
			    calibrateGetInfo(*(imd->calibInfo));
			    calibrateRelease();

			    printf("residual %f\n",residual);
			}
		    }
		    for(int i=selections.GetCount()-1;i>=0;i--){
			printf("deleting selection %d\n",i);
			Delete(selections[i]);
		    }
		    printf("done calibrating\n");
		    
		    
		    
		}
		index++;
	    }
	}
	else if(GetCount()==0){
	    Append("[Add Images Here]");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    event.RequestMore(FALSE);
	}
	else{
	    event.RequestMore(FALSE);
	}
    }
    void OnPlay(wxMouseEvent & event){
	for(int i=0;i<GetCount();i++){
	    Deselect(i);
	}
	Select(0);
	state=PLAYING;
	index=1;
	Refresh();
    }
    void OnCalibrate(wxMouseEvent & event){
	CalibrateDialog * cd=new CalibrateDialog(NULL,0);
	if(cd->ShowModal()==wxID_OK){
	    index=0;
	    state=CALIBRATING;

	    cd->GetThresh(thresh);
	    big=cd->GetBig();
	    small=cd->GetSmall();
	    printf("got thresh %d %d %d\n",thresh[0],thresh[1],thresh[2]);
	    printf("got small %d and big %d\n",small,big);
	}
	cd->Destroy();
    }
    void OnCapture(wxMouseEvent & event){
	CaptureDialog * cd =new CaptureDialog(NULL,0);
	cd->ShowModal();
	cd->Destroy();
    }

    void OnLoad(wxMouseEvent & event){
	wxFileDialog fd(GetParent(),"Select file to load", "","","*.pgm|*.ppm",wxOPEN|wxMULTIPLE);
	if(wxID_OK==fd.ShowModal()){
	    wxArrayString names;
	    wxArrayString paths;
	    fd.GetFilenames(names);
	    fd.GetPaths(paths);
	    if(GetString(0)=="[Add Images Here]"){
		Clear();
	    }
	    for(int i=0;i<names.GetCount();i++){
		ImageData * imd = new ImageData();
		imd->image.readPPM((const char *)paths[i].ToAscii());
		Append(names[i],imd);
	    }
	}
    }
    
    void OnSave(wxMouseEvent & event){
	wxFileDialog fd(GetParent(),"Select file to load", "","","*.*",wxSAVE);
	if(wxID_OK==fd.ShowModal()){
	    
	}
    }
    void OnListSelected(const wxCommandEvent & event){
	int i=GetSelection();
	
	//So texture doesn't get updated for multi selections
	wxArrayInt sel;
	GetSelections(sel);
	int n=sel.GetCount();
	if(n>1)return;

	if(i>=0 && GetClientData(i)){
	    glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
	}
    }

    void OnRightDown(wxMouseEvent & event){
	PopupMenu(imageListMenu,(wxPoint(event.GetX(),event.GetY())));
    }
    void OnRemove(wxCommandEvent & event){
	wxArrayInt selections;
	GetSelections(selections);
	for(int i=selections.GetCount()-1;i>=0;i--){
	    Delete(selections[i]);
	}
    }
    void OnRemoveAll(wxCommandEvent & event){
	Clear();
    }
    DECLARE_EVENT_TABLE();
};

IMPLEMENT_DYNAMIC_CLASS(ImageList,wxListBox)

BEGIN_EVENT_TABLE(ImageList, wxListBox)
    EVT_RIGHT_DOWN(ImageList::OnRightDown)
    EVT_MENU    (ID_Remove,ImageList::OnRemove)
    EVT_MENU    (ID_RemoveAll,ImageList::OnRemoveAll)
    EVT_MENU    (ID_Load,ImageList::OnLoad)
    EVT_MENU    (ID_Play,ImageList::OnPlay)
    EVT_MENU    (ID_Save,ImageList::OnSave)
    EVT_MENU    (ID_Capture,ImageList::OnCapture)
    EVT_MENU    (ID_Calibrate,ImageList::OnCalibrate)
    EVT_LISTBOX(-1, ImageList::OnListSelected)
    EVT_IDLE (ImageList::OnIdle)
END_EVENT_TABLE()


class CalibImageList:public wxListBox{
    wxMenu * menu;
    int playing;
 public:
    DECLARE_DYNAMIC_CLASS(CalibImageList)
    CalibImageList():wxListBox(){
    }
    CalibImageList(wxWindow * parent):
	wxListBox(parent,ID_CalibImageList,wxDefaultPosition,
		  wxDefaultSize,0,0,wxLB_EXTENDED){
	menu = new wxMenu();
	menu->Append(ID_Remove,"Remove Selected","Delete Selected",wxITEM_NORMAL);
	menu->Append(ID_RemoveAll,"Remove All","Delete Selected",wxITEM_NORMAL);
	menu->AppendSeparator();
	menu->Append(ID_Play,"Play","Play",wxITEM_NORMAL);
	menu->AppendSeparator();
	menu->Append(ID_Load,"Load","Load Images");
	menu->AppendSeparator();
	menu->Append(ID_Save,"Save","Save Selected",wxITEM_NORMAL);

	playing=0;
    }
    void OnIdle(wxIdleEvent & event){
	if(playing){
	    if(playing>=GetCount()){
		playing=0;
	    }
	    else{
		static struct timeval tv;
		static int called=0;
		if(playing==1&&!called){
		    gettimeofday(&tv,NULL);
		    called=1;
		}
		static struct timeval now;
		gettimeofday(&now,NULL);
	
		unsigned long long span = (now.tv_sec-tv.tv_sec)*1000+
		    (now.tv_usec-tv.tv_usec)/1000;
		
		if(span>10){
		    wxArrayInt selections;
		    GetSelections(selections);
		    for(int i=selections.GetCount()-1;i>=0;i--){
			Deselect(selections[i]);
		    }
		    gettimeofday(&tv,NULL);
		    SetSelection(playing);
		    OnListSelected(wxCommandEvent());

		    playing++;
		    called=0;
		}
		event.RequestMore(TRUE);
	    }
	}
	else if(GetCount()==0){
	    Append("[Calibrated Images]");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    Append(" ");
	    event.RequestMore(FALSE);
	}
	else{
	    event.RequestMore(FALSE);
	}
    }
    void OnPlay(wxMouseEvent & event){
	for(int i=0;i<GetCount();i++){
	    Deselect(i);
	}
	Select(0);
	playing=1;
	Refresh();
    }
    void OnLoad(wxMouseEvent & event){
	wxFileDialog fd(GetParent(),"Select file to load", "","","*.pgm|*.ppm",wxOPEN|wxMULTIPLE);
	if(wxID_OK==fd.ShowModal()){
	    wxArrayString names;
	    wxArrayString paths;
	    fd.GetFilenames(names);
	    fd.GetPaths(paths);
	    if(GetString(0)=="Calibrated Images"){
		Clear();
	    }
	    for(int i=0;i<names.GetCount();i++){
		ImageData * imd = new ImageData();
		imd->image.readPPM((const char *)paths[i].ToAscii());
		Append(names[i],imd);
	    }
	}
    }
    
    void OnSave(wxMouseEvent & event){
	wxFileDialog fd(GetParent(),"Select file to load", "","","*.*",wxSAVE);
	if(wxID_OK==fd.ShowModal()){
	    
	}
    }
    
    void OnListSelected(const wxCommandEvent & event){
	int i=GetSelection();

	//So texture doesn't get updated for multi selections
	wxArrayInt sel;
	GetSelections(sel);
	int n=sel.GetCount();
	if(n>1)return;

	if(i>=0 && GetClientData(i))
	    glCanvas->UpdateTexture(*((ImageData*)GetClientData(i)));
    }

    void OnRightDown(wxMouseEvent & event){
	PopupMenu(menu,(wxPoint(event.GetX(),event.GetY())));
    }
    void SendToImageList(int i){
	if(imageList->GetString(0)=="[Add Images Here]"){
	    imageList->Clear();
	}
	imageList->Append(GetString(i),
			  GetClientData(i));
    }
    void OnRemove(wxCommandEvent & event){
	wxArrayInt selections;
	GetSelections(selections);
	for(int i=0;i<selections.GetCount();i++){
	    SendToImageList(selections[i]);
	    
	}
	for(int i=selections.GetCount()-1;i>=0;i--){
	    Delete(selections[i]);
	}

    }
    void OnRemoveAll(wxCommandEvent & event){
	for(int i=0;i<GetCount();i++){
	    SendToImageList(i);
	}
	Clear();
    }
    DECLARE_EVENT_TABLE();
};
IMPLEMENT_DYNAMIC_CLASS(CalibImageList,wxListBox)

BEGIN_EVENT_TABLE(CalibImageList, wxListBox)
    EVT_RIGHT_DOWN(CalibImageList::OnRightDown)
    EVT_MENU    (ID_Remove,CalibImageList::OnRemove)
    EVT_MENU    (ID_RemoveAll,CalibImageList::OnRemoveAll)
    EVT_MENU    (ID_Load,CalibImageList::OnLoad)
    EVT_MENU    (ID_Play,CalibImageList::OnPlay)
    EVT_MENU    (ID_Save,CalibImageList::OnSave)
    EVT_LISTBOX(-1, CalibImageList::OnListSelected)
    EVT_IDLE (CalibImageList::OnIdle)
END_EVENT_TABLE()


class MyFrame: public wxFrame
{
public:
    MyFrame::MyFrame(const wxString& title,
		     const wxPoint& pos, const wxSize& size)
       : wxFrame((wxFrame *)NULL, -1, title, pos, size){
	wxMenu *menuFile = new wxMenu;
	menuFile->Append( ID_About, "&About..." );
	menuFile->AppendSeparator();
	menuFile->Append( ID_Quit, "E&xit" );
	wxMenuBar *menuBar = new wxMenuBar;
	menuBar->Append( menuFile, "&File" );
	SetMenuBar( menuBar );
	CreateStatusBar();
	SetStatusText( "Welcome to wxWindows!" );

	
	wxSplitterWindow * vertSplit = new wxSplitterWindow(this,-1);
	wxSplitterWindow * split = new wxSplitterWindow(vertSplit,-1);
	
      	imageList = new ImageList(split);
	calibImageList = new CalibImageList(split);
	glCanvas = new MyGLCanvas(vertSplit);

	wxBoxSizer * boxsizer = new wxBoxSizer(wxHORIZONTAL);
	
	split->SetMinimumPaneSize(50);
        vertSplit->SetMinimumPaneSize(50);
	vertSplit->SetSashPosition(1000,TRUE);

	split->SplitHorizontally(imageList,calibImageList);
	vertSplit->SplitVertically(split,glCanvas);

	wxRect rect=split->GetRect();
	//split->SetMinimumPaneSize(100);
	
	
	

	//boxsizer->Add(split,1,wxGROW);
	//boxsizer->Add(glCanvas,1,wxALIGN_CENTER|wxGROW);
	boxsizer->Add(vertSplit,1,wxGROW);
	SetSizer(boxsizer);

	split->Validate();
	vertSplit->Validate();
	Validate();
    }
    void OnDelete(wxCommandEvent & event){
	long item = -1;
	while(1)
	{
	
	}
    }
    void OnExit(wxCommandEvent& event){
    }
    void OnAbout(wxCommandEvent & event){
    }
    void OnSize(wxSizeEvent& event){
    }
    void OnButton3(wxMouseEvent & event){
	printf("button 3\n");
    }
    
    
    DECLARE_EVENT_TABLE();
};
BEGIN_EVENT_TABLE(MyFrame, wxFrame)
  EVT_MENU    (ID_About,  MyFrame::OnAbout)
  EVT_MENU    (ID_Quit,   MyFrame::OnExit)
      // EVT_SIZE    (           MyFrame::OnSize)
     
END_EVENT_TABLE()


class MyApp: public wxApp
{
public:
    MyFrame * frame;
    bool OnInit(){
	frame = new MyFrame("Frame",wxDefaultPosition,wxSize(640,480));
	frame->Show( TRUE );
	SetTopWindow( frame );
	return TRUE;
    }
};

IMPLEMENT_APP(MyApp)
