#ifndef NEWSEG_H
#define NEWSEG_H
#ifdef USING_GLEE
#include <GLee.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <GL/gl.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include "matrix.h"
#include "vec.h"
#include "image.h"

class CalibInfo{
 public:
    double * centers;
    int * sizes;
    int * labels;
    int ** lines;
    double center[2];
    int ncenters;
    int nlines;
    double residual;
    CalibInfo(){
	ncenters=0;
	nlines=0;
	centers=0;
	sizes=0;
	labels=0;
	lines=0;
	center[0]=0;
	center[1]=0;
    }
    CalibInfo * copy(){
	CalibInfo * ci = new CalibInfo();

	ci->ncenters=ncenters;
	ci->nlines =nlines;
	ci->center[0]=center[0];
	ci->center[1]=center[1];
	
	ci->centers = new double[ncenters*2];
	memcpy(ci->centers,centers,2*sizeof(double)*ci->ncenters);
	
	ci->sizes = new int[ncenters];
	memcpy(ci->sizes,sizes,sizeof(int)*ci->ncenters);
	
	ci->labels = new int[nlines];
	memcpy(ci->labels,labels,sizeof(int)*ci->nlines);
	
	ci->lines = new int *[nlines];
	for(int i=0;i<ci->nlines;i++){
	    ci->lines[i]=new int[4];
	    memcpy(ci->lines[i],lines[i],sizeof(int)*4);
	}
	ci->residual=residual;
	return  ci;
    }
    void release(){
      for(int i=0;i<nlines;i++){
	delete lines[i];
      }
      delete centers;
      delete sizes;
      delete labels;
      delete lines;
      
      ncenters=0;
      nlines=0;
      
      centers=0;
      sizes=0;
      labels=0;
      lines=0;
      center[0]=0;
      center[1]=0;
    }
    ~CalibInfo(){
	release();
    }
};

void drawString(double x,double y,char * fmt,...);
int calibrateGetInfo(CalibInfo & ci);
void calibrateRelease();
double calibrateImage(Image8 & orig,int rt,int gt,
		      int bt,int smallSize,
		      int large,int useautothresh,
		      Matrix & kc,Matrix & intrinsic,Matrix & extrinsic,
		      int fitFocalLength=0);
void groupToPoint(double * centers,int * sizes,int ncenters,
		  int ** lines,int nlines,int * labels,int * ignore,
		  Matrix &worldp,Matrix &imagep);
Matrix  projectPoints(Matrix & kc,Matrix &intrinsic,
		      Matrix &rot,Matrix &trans,Matrix &points);
Matrix projectPoints(Matrix & kc,Matrix &intrinsic,Matrix &params,
		     Matrix &worldp);
#endif
