#define GL_GLEXT_PROTOTYPES
#include <wx/wx.h>
#include <wx/glcanvas.h>
#include <wx/spinctrl.h>
#include <wx/notebook.h>
#include <stdio.h>
#include <stdlib.h>
#include "image.h"
#include "matrix.h"
#include <libraw1394/raw1394.h>
#include <libdc1394/dc1394_control.h>
#include <GL/glut.h>
#include <GL/gl.h>

Matrix getGLProj(double near,double far)
{
  Matrix proj=Matrix::eye(4,4);

  /*proj(0,0)=intr(0,0);
  proj(1,1)=intr(1,1);
  //screw the skew
  proj(0,3)=intr(0,2);
  proj(1,3)=intr(1,2);*/

  proj(2,2)=(far+near)/(near-far);
  proj(2,3)=2*far*near/(near-far);
  proj(3,2)=-1;//copy -z into w
  proj(3,3)=0;
  return proj;
}


/**
  The modelview should be scaled by -1,1,-1 before
  applying the extrinsics.  This should not be done in
  here, or else lighting etc will be wrong!
 */
void applyGLIntrProj(Matrix & intr,
		     int imw,int imh,
		     double near,double far)
{
  Matrix proj44=getGLProj(near,far); //use the zplane distance as approx
  glTranslatef(-1,-1,0);
  glScalef(2.0/(double)(imw),2.0/(double)(imh),1);
  glTranslatef(intr(0,2),intr(1,2),0);
  glScalef(-intr(0,0),-intr(1,1),1);
  glMultTransposeMatrixd(proj44.data);
  //glScalef(1,1,-1);
}



void displayHelper();
void calibInit();

double calibrateImage(Image8 & orig,
		      int rt,int gt,
		      int bt,int small,
		      int large,int useautothresh,
		      Matrix & kc,Matrix & intrinsic,
		      Matrix & extrinsic);
void calibrateRelease();

extern int ww,wh;
Image8 orig;
Matrix extrinsic;
Matrix kc;
Matrix intrinsic;

nodeid_t* camera_nodes;
dc1394_camerainfo *cameras;
dc1394_cameracapture *captures;
dc1394_miscinfo *misc_infos;
unsigned char c[640*480*3];
int numCameras; 
raw1394handle_t handle;


wxCheckBox * gray;
wxSlider * red;
wxSlider * green;
wxSlider * blue;
wxSpinCtrl * big;
wxSpinCtrl * small;
wxCheckBox * freeze;
wxFrame * frame;


#define NUM_TIMES 50
#define WIDTH 320
#define HEIGHT 240

#define USE_CAMERA
//#define DMA_CAP

#define YUV2RGB(y, u, v, r, g, b)\
  r = y + ((v*1436) >>10);\
  g = y - ((u*352 + v*731) >> 10);\
  b = y + ((u*1814) >> 10);\
  r = r < 0 ? 0 : r;\
  g = g < 0 ? 0 : g;\
  b = b < 0 ? 0 : b;\
  r = r > 255 ? 255 : r;\
  g = g > 255 ? 255 : g;\
  b = b > 255 ? 255 : b

void uyvy2rgb (unsigned char *src, unsigned char *dest, int NumPixels) {
  register int i = (NumPixels << 1)-1;
  register int j = NumPixels + ( NumPixels << 1 ) -1;
  register int y0, y1, u, v;
  register int r, g, b;
  
  while (i > 0)
  {
    y1 = (unsigned char) src[i--];
    v  = (unsigned char) src[i--] - 128;
    y0 = (unsigned char) src[i--];
    u  = (unsigned char) src[i--] - 128;
    YUV2RGB (y1, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
    YUV2RGB (y0, u, v, r, g, b);
    dest[j--] = b;
    dest[j--] = g;
    dest[j--] = r;
  }
}

class MyGLCanvas: public wxGLCanvas{
public:
    MyGLCanvas(wxWindow * parent):wxGLCanvas(parent,-1){
    }
    void OnPaint(wxPaintEvent & event){
	wxPaintDC dc(this);

#ifndef __WXMOTIF__
	if (!GetContext()) return;
#endif
	SetCurrent();

	GetSize(&ww,&wh);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0,0,ww,wh);
	glOrtho(0,orig.w*2,orig.h,0,-1,100000);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
#ifdef USE_CAMERA
	if(!freeze->GetValue()){
#ifndef DMA_CAP
	  if(dc1394_single_capture(handle,captures) != DC1394_SUCCESS){
	    fprintf(stderr, "Could not capture multi frames\n");
	    exit(1);
	  }
#else
	  if(dc1394_dma_multi_capture(captures, numCameras) != DC1394_SUCCESS){
	    fprintf(stderr, "Could not capture multi frames\n");
	    exit(1);
	    }
	    dc1394_dma_done_with_buffer(&captures[0]);	  
	    
#endif
	}
	uyvy2rgb((unsigned char*)captures[0].capture_buffer,
		 orig.data, WIDTH*HEIGHT);
#endif
	double res=calibrateImage(orig,red->GetValue(),
				  green->GetValue(),
				  blue->GetValue(),
				  small->GetValue(),big->GetValue(),
				  1,
				  kc,intrinsic,extrinsic);
	displayHelper();
	calibrateRelease();

	Matrix intri=Matrix::eye(4);
	Matrix proje=Matrix::eye(4);

	intri(0,0)=363;
	intri(1,1)=363;
	intri(0,3)=152;
	intri(1,3)=113;
	
	proje(3,2)=1.0;
	proje(3,3)=0;

	
	Matrix allmat=intri*proje*extrinsic;
	allmat=allmat.transpose();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0,0,ww/2,wh);
	
	applyGLIntrProj(intrinsic,
			320,240,0.001,1000);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(-1,1,-1);
	glMultTransposeMatrixd(extrinsic.data);

	//Old stuff
	//glOrtho(0,320,240,0,-1000,10000);
	//glMatrixMode(GL_MODELVIEW);
	//glLoadIdentity();
	//glMultMatrixd(allmat.data);

	glLineWidth(2);
	glBegin(GL_LINES);
  
	glColor3f(0,1,0);
	glVertex3i(0,0,0);
	glVertex3i(0,100,0);
	
	glColor3f(1,0,0);
	glVertex3i(0,0,0);
	glVertex3i(100,0,0);
	
	glColor3f(0,0,1);
	glVertex3i(0,0,0);
	glVertex3i(0,0,100);
	glEnd();
	glLineWidth(1);
	
	glColor3f(0.4,0.4,0.4);
	
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glutSolidTeapot(1);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Matrix pt(3,1);
	pt[0]=extrinsic(0,3);
	pt[1]=extrinsic(1,3);
	pt[2]=extrinsic(2,3);

	pt*=(1.0/pt[2]);
	pt=intrinsic*pt;
	pt.printMatlab("pt");
	glColor3f(0,0,1);
	glPointSize(4);
	glBegin(GL_POINTS);
	glVertex2f(pt[0],pt[1]);
	glEnd();
	glPointSize(1);
	
	SwapBuffers();
	
	char text[256];
	sprintf(text,"Res:%f",res);
	frame->SetStatusText(text);
    }
    void OnEraseBackground(wxEraseEvent & event){
    }
    void OnSize(wxSizeEvent & event){
    }
    void OnEnterWindow(wxMouseEvent & event){
    }
    void OnKeyDown(wxKeyEvent & event){
	printf("canvas on key down %d %c\n",event.GetKeyCode(),event.GetKeyCode());
    }
    void OnKeyUp(wxKeyEvent & event){
	printf("canvas on key up\n");
    }
protected:
    DECLARE_EVENT_TABLE();
};

MyGLCanvas * glCanvas;

BEGIN_EVENT_TABLE(MyGLCanvas, wxGLCanvas)
    EVT_KEY_DOWN(MyGLCanvas::OnKeyDown )
    EVT_KEY_UP( MyGLCanvas::OnKeyUp )
    EVT_SIZE(MyGLCanvas::OnSize)
    EVT_PAINT(MyGLCanvas::OnPaint)
    EVT_ERASE_BACKGROUND(MyGLCanvas::OnEraseBackground)
    EVT_ENTER_WINDOW( MyGLCanvas::OnEnterWindow )
END_EVENT_TABLE()

class OptionsFrame : public wxDialog{
    enum {OPT_NBASIS=wxID_HIGHEST,
	  OPT_HW,OPT_SW,OPT_MMX,OPT_FREEZE,
          OPT_SMALL,OPT_BIG,OPT_GRAY};
public:
    OptionsFrame(wxWindow * parent):wxDialog(parent,-1,"Options",
					    wxPoint(-1,-1),
					    wxSize(230,180)){
	wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);

	wxStaticBox * threshBox = new wxStaticBox(this,
						  -1,
						  "Threshold");
	wxStaticBox * generalBox = new wxStaticBox(this,
						  -1,
						  "Sizes");
	wxBoxSizer * threshBoxSizer = new wxStaticBoxSizer(threshBox,
							   wxVERTICAL);
	wxBoxSizer * general = new wxStaticBoxSizer(generalBox,
						    wxVERTICAL);

	gray = new wxCheckBox(this,OPT_GRAY,"Gray",
				 wxPoint(-1,-1));
	red = new wxSlider(this,OPT_NBASIS,160,0,255,
			   wxPoint(-1,-1),
			   wxSize(-1,-1));
	green = new wxSlider(this,OPT_NBASIS,120,0,255,
			   wxPoint(-1,-1),
			   wxSize(-1,-1));
	blue = new wxSlider(this,OPT_NBASIS,160,0,255,
			   wxPoint(-1,-1),
			   wxSize(-1,-1));

	big = new wxSpinCtrl(this,OPT_BIG,"1000",wxDefaultPosition,
			     wxDefaultSize,
			     wxSP_ARROW_KEYS,0,10000,1000);
	small = new wxSpinCtrl(this,OPT_SMALL,"15",wxDefaultPosition,
			       wxDefaultSize,
			       wxSP_ARROW_KEYS,0,10000,15);
	freeze = new wxCheckBox(this,OPT_FREEZE,"Freeze",
			    wxPoint(-1,-1));
	wxBoxSizer * temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1," "),0,wxALIGN_CENTER);
	temp->Add(gray,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW);

	temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1,"R"));
	temp->Add(red,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

	temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1,"G"));
	temp->Add(green,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

	temp = new wxBoxSizer(wxHORIZONTAL);
	temp->Add(new wxStaticText(this,-1,"B"));
	temp->Add(blue,1,wxGROW);
	threshBoxSizer->Add(temp,1,wxGROW|wxALIGN_CENTER);

	wxFlexGridSizer * grid = new  wxFlexGridSizer(2,2,2,2);

	grid->Add(new wxStaticText(this,-1,"Small"),0,wxALIGN_RIGHT);
	grid->Add(small,0,wxALIGN_RIGHT);
	
	grid->Add(new wxStaticText(this,-1,"Big "),0,wxALIGN_RIGHT);
	grid->Add(big,0,wxALIGN_RIGHT);

	general->Add(grid,1,wxGROW);
	general->Add(freeze,1,wxGROW);

	all->Add(threshBoxSizer,1,wxGROW);
	all->Add(general,1,wxGROW);

	SetSizer(all);
	SetAutoLayout(TRUE);
	Layout();

#if wxUSE_TOOLTIPS

#endif 

    }
    void OnClose(wxCloseEvent * event){
	printf("closing options frame\n");
	Show(FALSE);
    }
    ~OptionsFrame(){
	printf("Deleting options frame\n");
    }
  void NBasis(wxScrollEvent & event){
    wxObject * obj = event.GetEventObject();
    if(gray->GetValue()){
      int value = ((wxScrollBar *)obj)->GetValue();
      red->SetValue(value);
      green->SetValue(value);
      blue->SetValue(value);
    }
    glCanvas->Refresh();
  }
  void OnGray(wxCommandEvent & event){
    if(gray->GetValue()){
      int value = red->GetValue()+green->GetValue()+blue->GetValue();
      red->SetValue(value/3);
      green->SetValue(value/3);
      blue->SetValue(value/3);
      glCanvas->Refresh();
    }
  }
  void OnSpinSmall(wxSpinEvent & event){
    glCanvas->Refresh();
  }
  void OnSpinBig(wxSpinEvent & event){
    glCanvas->Refresh();
  }
    protected:
    DECLARE_EVENT_TABLE();
};
BEGIN_EVENT_TABLE(OptionsFrame, wxDialog)
    EVT_SPIN(OPT_SMALL,OptionsFrame::OnSpinSmall)
    EVT_SPIN(OPT_BIG,OptionsFrame::OnSpinBig)
    EVT_COMMAND_SCROLL_THUMBTRACK(OPT_NBASIS,OptionsFrame::NBasis)
    EVT_CLOSE(OptionsFrame::OnClose)
    EVT_CHECKBOX(OPT_GRAY,OptionsFrame::OnGray)
END_EVENT_TABLE()

class MyFrame : public wxFrame{
    OptionsFrame * options;
public:
  
    enum {ID_EXIT=wxID_HIGHEST,
	  ID_OPEN,
	  ID_SAVE,
	  ID_CAMERA,
	  ID_EDITOR,
	  ID_OPTIONS,
	  ID_INFO,
	  ID_TUTORIAL,
	  ID_ABOUT,
	  ID_HELP};
protected:
    DECLARE_EVENT_TABLE();
public:
    void open(wxCommandEvent & event){
	SetStatusText("Open");
	wxFileDialog * fileDialog = new wxFileDialog(this,"Open");
	if(fileDialog->ShowModal()==wxID_OK){
	    SetStatusText(fileDialog->GetFilename(),0);
	}
	
	fileDialog->Destroy();
    }
    void OnOptions(wxCommandEvent & event){
	options->Show(TRUE);
    }
    void OnExit(wxCommandEvent & event){
	printf("Exiting\n");
	Close();
    }
    void OnClose(wxCommandEvent & event){
	printf("Closing\n");
	Close();
    }
    MyFrame():wxFrame(NULL,-1,"ImaBaRender"){
      wxMenu * menu=new wxMenu();
	wxMenuBar * menuBar=new wxMenuBar();
	menu->Append(ID_OPEN,"&Open");
	menu->Append(ID_SAVE,"&Save");
	menu->AppendSeparator();
	menu->Append(ID_EXIT,"E&xit");
	menuBar->Append(menu,"&File");

	SetMenuBar(menuBar);
	CreateStatusBar();
	SetStatusText("");
	Show(TRUE);
	glCanvas = new MyGLCanvas(this);
	
	options = new OptionsFrame(this);
	options->Show();
#if wxUSE_TOOLTIPS

#endif

    }
    void OnKeyDown(wxKeyEvent & ke){
	printf("keydown \n");
    }
    void OnKeyUp(wxKeyEvent& ke){
	printf("key up\n");
    }
    void OnIdle(wxIdleEvent & idle){
      glCanvas->Refresh();
    }
};

class MyApp : public wxApp{
public:
    virtual bool OnInit(){
	frame = new MyFrame();
	SetTopWindow(frame);

	orig=Image8(320,240,3);
	//calibInit();

	extrinsic=Matrix(4,4);
	intrinsic=Matrix(3,3);
	
	kc=Matrix(5,1);
	kc.setAll(0);
	kc[0]=-0.4;
	kc[1]= 0.534;

	intrinsic=Matrix::eye(3,3);
	intrinsic(0,0)=320;
	intrinsic(1,1)=240;
	intrinsic(0,2)=160;
	intrinsic(1,2)=120;
	

	char s[256];
	int i;
	
#ifdef USE_CAMERA
	handle = dc1394_create_handle(0);
	if (handle==NULL) {
	  printf("did you insmod the drivers?\n");
	  exit(1);
	}
	
	camera_nodes = dc1394_get_camera_nodes(handle, &numCameras, 1); 
	dc1394_set_iris(handle,camera_nodes[0],1);
	printf("numCameras: %d\n",numCameras);
	
	if(numCameras==0)
	{
	  exit(0);
	}
	
	cameras=new dc1394_camerainfo[numCameras];
	captures=new dc1394_cameracapture[numCameras];
	misc_infos=new dc1394_miscinfo[numCameras];
	
	for(i=0;i<numCameras; i++){
	  printf("node %d\n",camera_nodes[i]); 
	  dc1394_get_camera_info(handle, camera_nodes[i], &cameras[i]);
	  dc1394_get_camera_misc_info(cameras[i].handle,cameras[i].id,
				      &misc_infos[i]);
#ifdef DMA_CAP
	  if (dc1394_dma_setup_capture(handle, 
				       camera_nodes[i],  
				       i,
				       FORMAT_VGA_NONCOMPRESSED,
				       MODE_320x240_YUV422,
				       SPEED_400,
				       FRAMERATE_30,
				       8, // number of dma buffers
				       0, // drop frames
				       "/dev/video1394", 
				       &captures[i])
	      != DC1394_SUCCESS){
	    printf("Setting Camera %d failed\n", i);
	    exit(1);
	  }
#else
	  if (dc1394_setup_capture(handle, 
				   camera_nodes[i],  
				   i,
				   FORMAT_VGA_NONCOMPRESSED,
				   MODE_320x240_YUV422,
				   SPEED_400,
				   FRAMERATE_30,
				   &captures[i])
	      != DC1394_SUCCESS){
	    printf("Setting Camera %d failed\n", i);
	    exit(1);
	  }
#endif
	  //dc1394_stop_iso_transmission(handle,camera_nodes[i]);
	  if(dc1394_start_iso_transmission(handle,camera_nodes[i]) != DC1394_SUCCESS)
	  {
	    fprintf(stderr, "Could not start ISO transmission\n");
	    exit(1);
	  }
	}
#endif 
	/*
	for(i=0;i<numCameras; i++){
	  dc1394_stop_iso_transmission(handle,camera_nodes[i]);
	  dc1394_dma_release_camera(handle, &captures[i]);
	}
	
	raw1394_destroy_handle(handle);
	*/
	printf("done init\n");
	return true;
    }
    MyApp(){
		
    }
    
    void OnClose(wxCloseEvent & event){
	printf("on close\n");
    }
protected:
    //  DECLARE_EVENT_TABLE();
};

BEGIN_EVENT_TABLE(MyFrame,wxFrame)
    EVT_KEY_DOWN(MyFrame::OnKeyDown )
    EVT_KEY_UP( MyFrame::OnKeyUp )
    EVT_MENU(ID_EXIT,MyFrame::OnExit)
    EVT_MENU(ID_OPEN,MyFrame::open)
    EVT_IDLE(MyFrame::OnIdle)
END_EVENT_TABLE()

IMPLEMENT_APP(MyApp)
