#ifndef SEGMENT_DLG_H
#define SEGMENT_DLG_H

#include <wx/wx.h>
#include <wx/notebook.h>
#include <wx/spinbutt.h>
#include <wx/spinctrl.h>
#include "guitypes.h"
#include "image.h"
#include "glcanvas.h"

/**Dialog for controlling the segmentation of images.
   Two methods of segmentation are supported: color and background subtraction.
*/
class SegmentDialog : public wxDialog,GLMouseListener{
    wxSpinCtrl * nregions;
    wxButton * load;
    wxStaticText * fileName;
    wxString path;
    wxSlider * thresh;
    wxListBox * regions;
    wxButton * add;
    wxButton * remove;
    wxNotebook * nb;
    wxPoint mouseDown;
    wxPoint mouseUp;
    wxPoint mouseCur;
    int glWidth;
    int glHeight;

    int segmentIndex;
    int segmenting;

    Image8 bgimage;
public:
    DECLARE_DYNAMIC_CLASS(SegmentDialog)
      SegmentDialog(wxWindow * parent=0,wxWindowID id=-1):wxDialog(parent,id,wxT("Segment"),
							      wxPoint(-1,-1),
							      wxSize(230,200)){
	wxBoxSizer * all = new wxBoxSizer(wxVERTICAL);
	
	nb = new wxNotebook(this,-1);
	wxPanel * subPanel = new wxPanel(nb,-1);
	wxPanel * colPanel = new wxPanel(nb,-1);

	wxBoxSizer * subSizer = new wxBoxSizer(wxVERTICAL);
	wxBoxSizer * colSizer = new wxBoxSizer(wxVERTICAL);
	wxBoxSizer * buttonSizer = new wxBoxSizer(wxHORIZONTAL);
	wxBoxSizer * horSizer;
	wxBoxSizer * verSizer;
	//Initialize the background subtraction segmentor
	fileName = new wxStaticText(subPanel,-1,wxT("No File Selected"),wxDefaultPosition,
				    wxDefaultSize,wxALIGN_CENTRE|wxALIGN_CENTRE_VERTICAL);
	path=wxT("");
	load = new wxButton(subPanel,ID_Load,wxT("Load"));
	thresh = new wxSlider(subPanel,ID_Thresh,20,0,255,
			      wxPoint(-1,-1),
			      wxSize(-1,-1));
	horSizer = new wxStaticBoxSizer(new wxStaticBox(subPanel,-1,wxT("Select Background Image"))
					,wxHORIZONTAL);
	horSizer->Add(load,0,wxLEFT|wxRIGHT|wxALIGN_CENTRE|wxALIGN_CENTRE_VERTICAL,2);
	horSizer->Add(fileName,1,wxGROW|wxLEFT|wxRIGHT|wxALIGN_CENTRE_VERTICAL,2);
	subSizer->Add(horSizer,1,wxGROW);

	horSizer = new wxStaticBoxSizer(new wxStaticBox(subPanel,-1,wxT("Threshold"))
					,wxHORIZONTAL);
	horSizer->Add(thresh,1,wxGROW|wxALIGN_CENTRE_VERTICAL);
	subSizer->Add(horSizer,1,wxGROW);
	subPanel->SetSizer(subSizer);
	//End of the background subtraction init

	//Init the color 
	

	horSizer = new wxBoxSizer(wxHORIZONTAL);
	verSizer = new wxBoxSizer(wxVERTICAL);
	add = new wxButton(colPanel,ID_Add,wxT("Add"));
	remove = new wxButton(colPanel,ID_Remove,wxT("Remove"));
	
	verSizer->Add(add);
	verSizer->Add(remove);

	regions = new wxListBox(colPanel,-1);
	horSizer->Add(regions,1,wxGROW|wxEXPAND);
	horSizer->Add(verSizer,1,wxGROW|wxEXPAND);
	colSizer->Add(horSizer,1,wxGROW|wxEXPAND);
	
	colPanel->SetSizer(colSizer);

	//End init color segmentor

	nb->AddPage(colPanel,wxT("Color Seg."),TRUE);
	nb->AddPage(subPanel,wxT("Image Sub."),FALSE);

	buttonSizer->Add(new wxButton(this,ID_OK,wxT("OK")),1,wxSHAPED|wxALIGN_CENTRE);
	buttonSizer->Add(new wxButton(this,ID_Close,wxT("Close")),1,wxSHAPED|wxALIGN_CENTRE);

#ifdef WX_2_8	
	all->Add(nb,3,wxGROW|wxEXPAND);
#else
	all->Add(new wxNotebookSizer(nb),3,wxGROW|wxEXPAND);
#endif
	all->Add(buttonSizer,1,wxGROW|wxALIGN_CENTRE);

	SetSizer(all);
	Layout();
#if wxUSE_TOOLTIPS

#endif 
	segmenting=0;
	CentreOnScreen();
    }
    void SyncDisplay();
    void OnLoad(wxCommandEvent & event);
    void OnAdd(wxCommandEvent & event);
    void OnRemove(wxCommandEvent & event);
    void OnThresh(wxScrollEvent & event);
    void OnOK(wxCommandEvent & event);
    void OnIdle(wxIdleEvent & event);
    void OnClose(wxCommandEvent & event);
    void OnCloseEvent(wxCloseEvent & event);
    virtual void drawGL(int w,int h);
    Image8 SegmentImage(Image8 & im);
    void RegionsChanged();
    virtual void mouse(wxMouseEvent & event,int GLw,int GLh);
    bool Show(const bool show);
protected:
    DECLARE_EVENT_TABLE();
};


#endif
