#include "mesh.h"
#include "image.h"
#include "globals.h"
using namespace CapguiGlobals;
#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>

/*
  struct's and defines for MD2 File were obtained from:
   http://tfc.duke.free.fr/us/tutorials/models/md2.htm

  these include:
  struct MD2AnimType;
  struct T_MD2AnimStats;
  MD2AnimStats animstats;
  struct T_MD2Triangle;
  struct T_MD2TexCoord;
  struct T_MD2Vertex;
  struct T_MD2Frame;

  which I am sure the author probably took directly from
  the quake 2 source code.

  I claim no ownership over that code.
*/

//http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef struct T_MD2AnimStats
{
    unsigned char first,last,fps;
}MD2AnimStats;

//http://tfc.duke.free.fr/us/tutorials/models/md2.htm
MD2AnimStats animstats[ 21 ] = 
{
    // first, last, fps
    {   0,  39,  9 },   // STAND
    {  40,  45, 10 },   // RUN
    {  46,  53, 10 },   // ATTACK
    {  54,  57,  7 },   // PAIN_A
    {  58,  61,  7 },   // PAIN_B
    {  62,  65,  7 },   // PAIN_C
    {  66,  71,  7 },   // JUMP
    {  72,  83,  7 },   // FLIP
    {  84,  94,  7 },   // SALUTE
    {  95, 111, 10 },   // FALLBACK
    { 112, 122,  7 },   // WAVE
    { 123, 134,  6 },   // POINT
    { 135, 153, 10 },   // CROUCH_STAND
    { 154, 159,  7 },   // CROUCH_WALK
    { 160, 168, 10 },   // CROUCH_ATTACK
    { 196, 172,  7 },   // CROUCH_PAIN
    { 173, 177,  5 },   // CROUCH_DEATH
    { 178, 183,  7 },   // DEATH_FALLBACK
    { 184, 189,  7 },   // DEATH_FALLFORWARD
    { 190, 197,  7 },   // DEATH_FALLBACKSLOW
    { 198, 198,  5 },   // BOOM
};


void Mesh::createList()
{
  GLuint listTemp=0;
  if(list<0){
    listTemp=glGenLists(1);
  }
  else{
    listTemp=list;
  }
  printf("creating list\n");
  list=-1;
  glNewList(listTemp,GL_COMPILE);
  draw(0);
  glEndList();
  list=listTemp;
}

void Mesh::alloc(int _nvert,int _ntri)
{
  if(vert){
    delete [] vert;
  }
  if(tri){
    delete [] tri;
  }
  if(norm)
    delete [] norm;
  nvert=_nvert;
  ntri=_ntri;
  vert=new Vec3f[nvert];
  tri=new Triangle[ntri];
  norm=new Vec3f[nvert];
}

void Mesh::allocTextures(int _ntex)
{
  int i=0;
  if(tex){
    delete [] tex;
  }
  if(tvert){
    for(i=0;i<nvert;i++){
      delete [] tvert[i];
    }
    delete [] tvert;
  }
  ntex=_ntex;
  tex=new GLuint[ntex];
  tvert=new Vec2f *[ntex];
  
  //Allocate the texture coordinates and the
  // teture id
  for(i=0;i<ntex;i++){
    tvert[i]=new Vec2f[nvert];
  }
  glGenTextures(ntex,tex);
}

void Mesh::setTextureImage(Image8& im,int index)
{
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tex[index]);
  im.initTexture();
}

typedef struct NODE
{
  struct NODE * next;
  int edge;
}Node;
/*
class Vec2f{
public:
  double x,y;
  Vec2f(double _x,double_y){
    x=_x;
    y=_y;
  }
  Vec2f operator+(Vec2f o){
    return Vec2f(x+o.x,y+o.y);
  }
  Vec2f operator+(Vec2f & o){
    return Vec2f(x+o.x,y+o.y);
  }
  Vec2f operator-(Vec2f o){
    return Vec2f(x-o.x,y-o.y);
  }
  Vec2f operator-(Vec2f & o){
    return Vec2f(x-o.x,y-o.y);
  }
};
*/

class Line{
public:
  Vec2f p,d;
  double t;
  double s;
  Line(){
    p[0]=p[1]=0;
    d[0]=d[1]=0;
  }
  Line(Vec2f & _p1,Vec2f & _p2){
    p[0]=_p1[0];p[1]=_p1[1];
    d[0]=_p2[0]-_p1[0];
    d[1]=_p2[1]-_p1[1];
  }
  int intersect(Line uv,Vec2f & where){
    s=((uv.p[1]-p[1])*d[0]+(p[0]-uv.p[0])*d[1])/
      (uv.d[0]*d[1]-uv.d[1]*d[0]);
        
    t=(uv.p[0]+s*uv.d[0]-p[0])/d[0];
    if(d[1]>d[0]){
      t=(uv.p[1]+s*uv.d[1]-p[1])/d[1];
    }
    if(s>=0.0 && s<=1.0){
      if(t>=0.0 && t<=1.0){
	if(t<0)t=0;
	if(t>1)t=1;
	where[0]=p[0]+t*d[0];
	where[1]=p[1]+t*d[1];
	return 1;
      }
    }
    return 0;
  }
};

/*
    bottom.intersect(edge,p0);
    top.intersect(edge,p1);
    double thresh=0.1;
    
    double small=0.01;
    double big=0.99;
    
      if(fabs(bottom.t)<thresh){
      silmap[0][((int)y1)*view[2]+((int)(x1))-1]=(big);
      silmap[1][((int)y1)*view[2]+((int)(x1))-1]=small;
      
      silmap[0][(((int)y1)-1)*view[2]+((int)(x1))]=small;
      silmap[1][(((int)y1)-1)*view[2]+((int)(x1))]=big;
      
      silmap[0][(((int)y1)-1)*view[2]+((int)(x1))-1]=(big);
      silmap[1][(((int)y1)-1)*view[2]+((int)(x1))-1]=(big);
      
      silmap[0][(((int)y1))*view[2]+((int)(x1))]=(small);
      silmap[1][(((int)y1))*view[2]+((int)(x1))]=(small);
      }
      else if(fabs(bottom.t-1.0)<thresh){
      silmap[0][((int)y1)*view[2]+((int)(x1))+1]=(small);
      silmap[1][((int)y1)*view[2]+((int)(x1))+1]=small;
      
      silmap[0][(((int)y1)-1)*view[2]+((int)(x1))]=big;
      silmap[1][(((int)y1)-1)*view[2]+((int)(x1))]=big;
      
      silmap[0][(((int)y1)-1)*view[2]+((int)(x1))+1]=(small);
      silmap[1][(((int)y1)-1)*view[2]+((int)(x1))+1]=(big);
      
      silmap[0][(((int)y1))*view[2]+((int)(x1))]=(big);
      silmap[1][(((int)y1))*view[2]+((int)(x1))]=(small);
      }
      
      else if(fabs(top.t)<thresh){
      
      silmap[0][((int)y1)*view[2]+((int)(x1))-1]=(big);
      silmap[1][((int)y1)*view[2]+((int)(x1))-1]=big;
      
      silmap[0][(((int)y1)+1)*view[2]+((int)(x1))]=(small);
      silmap[1][(((int)y1)+1)*view[2]+((int)(x1))]=(small);
      
      silmap[0][(((int)y1)+1)*view[2]+((int)(x1))-1]=(big);
      silmap[1][(((int)y1)+1)*view[2]+((int)(x1))-1]=(small);
      
      silmap[0][(((int)y1))*view[2]+((int)(x1))]=(small);
      silmap[1][(((int)y1))*view[2]+((int)(x1))]=big;
      }
      else if(fabs(top.t-1.0)<thresh){
      silmap[0][((int)y1)*view[2]+((int)(x1))+1]=(small);
      silmap[1][((int)y1)*view[2]+((int)(x1))+1]=big;
      
      silmap[0][(((int)y1)+1)*view[2]+((int)(x1))]=big;
      silmap[1][(((int)y1)+1)*view[2]+((int)(x1))]=(small);
      
      silmap[0][(((int)y1)+1)*view[2]+((int)(x1))+1]=(small);
      silmap[1][(((int)y1)+1)*view[2]+((int)(x1))+1]=(small);
      
      silmap[0][(((int)y1))*view[2]+((int)(x1))]=big;
      silmap[1][(((int)y1))*view[2]+((int)(x1))]=big;
      }*/

void processEdge(float * silmap[4],int * view,
		 Line edge,
		 double x1,double y1,
		 double x2,double y2)
{
  Line ver;
  Line hor;
  
  Vec2f p0,p1;
  Vec2f l,r,t,b;
  double smalls=1;
  double larges=0;
  int which=0;
  int first=1;
  
  y1=(int)y1;
  
  while(y1<=y2){    
    p0[0]=0;
    p0[1]=(int)y1;
    p1[0]=view[2];
    p1[1]=(int)y1;
    hor=Line(p0,p1);
    
    if(hor.intersect(edge,b)){
      silmap[0][((int)b[1])*view[2]+((int)b[0])]=b[0]-((int)b[0]);
      //silmap[1][((int)y1)*view[2]+((int)x1)]=b[1]-((int)y1);
    }
    y1++;
  }
  x1=(int)x1;
  while(x1<=x2){
    p0[0]=((int)x1);
    p0[1]=0;
    p1[0]=((int)x1);
    p1[1]=view[3];
    ver=Line(p0,p1);
    
    if(ver.intersect(edge,l)){
      //silmap[0][((int)y1)*view[2]+((int)x1)]=l[0]-((int)x1);
      silmap[1][((int)l[1])*view[2]+((int)l[0])]=l[1]-((int)l[1]);
    }
    x1++;
  }
}

void scanConvert(float * silmap[4],int * view,
		 double x1,double y1,
		 double x2,double y2)
{
  int first=1;
  if(x1>=0 && x1<view[2] &&
     y1>=0 && y1<view[3] &&
     x2>=0 && x2<view[2] &&
     y2>=0 && y2<view[3]){
    double rise=(y2-y1);
    double run =(x2-x1);
    if(fabs(rise)>fabs(run)){
      double inc=(double)run/(double)rise;
      int incsign = (inc>0)?1:-1;

      if(y1>y2){
	double temp=y1;
	y1=y2;
	y2=temp;
	temp=x1;
	x1=x2;
	x2=temp;
      }

      Vec2f t1,t2;
      t1[0]=x1;
      t1[1]=y1;
      t2[0]=x2;
      t2[1]=y2;
      Line edge=Line(t1,t2);

      if(x1>x2){
	double temp=x1;
	x1=x2;
	x2=temp;
      }
      
      processEdge(silmap,view,
		  edge,
		  x1,y1,
		  x2,y2);
    }
    else{
      double inc=(double)rise/(double)run;
      int incsign = (inc>0)?1:-1;
      if(x1>x2){
	double temp=y1;
	y1=y2;
	y2=temp;
	temp=x1;
	x1=x2;
	x2=temp;
      } 

      Vec2f t1,t2;
      t1[0]=x1;
      t1[1]=y1;
      t2[0]=x2;
      t2[1]=y2;
      Line edge=Line(t1,t2);

      if(y1>y2){
	double temp=y1;
	y1=y2;
	y2=temp;
      }

      processEdge(silmap,view,
		  edge,
		  x1,y1,
		  x2,y2);
      
    }
  }
}

void Mesh::drawSilEdges(Vec3f & lightPos)
{
    int i=0;
    glColor3f(0.0,0.0,0.0);
    glLineWidth(4);
    glBegin(GL_LINES);
    
    for(i=0;i<nedges;i++){
	if(edges[i].opp>=0){
	    Vec3f *n1=&(tri2norm[edges[i].face]);
	    Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
	    Vec3f l1=lightPos-vert[edges[i].vert];
	    
	    double d1=n1->dot(l1);
	    double d2=n2->dot(l1);
	    
	    if(signf(d1)!=signf(d2)){
		glVertex3fv(vert[edges[i].vert].data);
		glVertex3fv(vert[edges[edges[i].next].vert].data);
	    }
	}else{
	    //    glVertex3fv(vert[edges[i].vert].data);
	    //glVertex3fv(vert[edges[edges[i].next].vert].data);
	}
    }
    glEnd();
}

/* Don't have access to the viewport transformation */
void Mesh::drawEdges2(Vec3f & lightPos,float offset,float width)
{
    int view[4];
    int i=0;
    double x,y;
    glDisable(GL_CULL_FACE);
    glGetIntegerv(GL_VIEWPORT,view);

    glMatrixMode(GL_MODELVIEW);
    applyTransform();
/*
    glProgramLocalParameter4dARB(GL_VERTEX_PROGRAM_ARB,0,
				 2.0/view[2]-1.0,0,0,0);
    glProgramLocalParameter4dARB(GL_VERTEX_PROGRAM_ARB,1,
				 0,2.0/view[3]-1.0,0,0);
    glProgramLocalParameter4dARB(GL_VERTEX_PROGRAM_ARB,2,
				 offset,offset,1.0,0);
    //This won't work if the viewport is not at 0,0
    glProgramLocalParameter4dARB(GL_VERTEX_PROGRAM_ARB,3,
				 ((double)view[2])/2.0,((double)view[3])/2.0,0,0);
*/

    
    glMultiTexCoord4f(GL_TEXTURE3,2.0/view[2]-1.0,0.0,0.0,0.0);
    glMultiTexCoord4f(GL_TEXTURE4,0.0,2.0/view[3]-1.0,0.0,0.0);
    glMultiTexCoord4f(GL_TEXTURE5,offset,offset,1.0,0);
    glMultiTexCoord4f(GL_TEXTURE6,
		      ((double)view[2])/2.0,((double)view[3])/2.0,0,0);
    

    glDisable(GL_DEPTH_TEST);

    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE1);
    glDisable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE2);
    glDisable(GL_TEXTURE_2D);

    glDisable(GL_LIGHTING);
    glColor3f(0.0,1.0,1.0);
    glBegin(GL_QUADS);


    for(i=0;i<nedges;i++){
	if(edges[i].opp>=0){
	    Vec3f *n1=&(tri2norm[edges[i].face]);
	    Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
	    Vec3f l1=lightPos-vert[edges[i].vert];
	    
	    double d1=n1->dot(l1);
	    double d2=n2->dot(l1);
	    if(signf(d1)>=0 && signf(d2)<0){
		Vec3f v1=vert[edges[i].vert];
		Vec3f v2=vert[edges[edges[i].next].vert];
		glMultiTexCoord4f(GL_TEXTURE0,v1.data[0],v1.data[1],
				  v1.data[2],1.0);
		glMultiTexCoord4f(GL_TEXTURE1,v2.data[0],v2.data[1],
				  v2.data[2],1.0);

		glMultiTexCoord4f(GL_TEXTURE2,-1.0,-1.0,0.0,0.0);
		glVertex3fv(v1.data);
		glMultiTexCoord4f(GL_TEXTURE2,1.0,1.0,0.0,0.0);
		glVertex3fv(v1.data);
		glMultiTexCoord4f(GL_TEXTURE2,1.0,1.0,0.0,0.0);
		glVertex3fv(v2.data);
		glMultiTexCoord4f(GL_TEXTURE2,-1.0,-1.0,0.0,0.0);
		glVertex3fv(v2.data);
	    }
	}
	else{
	}
    }
    glEnd();
}


void Mesh::drawShadowPolygons(Vec3f lightPos)
{
  int i=0;
  float lightp[4];
  float p1[4];
  float p2[4];
  p1[3]=p2[3]=0;

  lightp[0]=-lightPos[0];
  lightp[1]=-lightPos[1];
  lightp[2]=-lightPos[2];
  lightp[3]=0;

  //lightPos=lightPos*-1;

  applyTransform();

  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_STENCIL_TEST);

  glStencilFunc(GL_ALWAYS,0,0xFF);
  glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
  glDepthMask(GL_FALSE);

  glDepthFunc(GL_LEQUAL);
  glDisable(GL_BLEND);

  glStencilOp(GL_KEEP,GL_KEEP,GL_INCR_WRAP);
  glCullFace(GL_BACK);
  
  float thresh=100;
  for(i=0;i<nedges;i++){
    if(edges[i].opp>=0){
      Vec3f *n1=&(tri2norm[edges[i].face]);
      Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
      Vec3f l1=lightPos-vert[edges[i].vert];

      double d1=n1->dot(l1);
      double d2=n2->dot(l1);

      if((signf(d1)!=signf(d2)) && signf(d1)>0){
	  Vec3f op1 = vert[edges[i].vert]-lightPos;
	  Vec3f op2 = vert[edges[edges[i].next].vert]-lightPos;
	  
	  p1[0]=op1[0];p2[0]=op2[0];
	  p1[1]=op1[1];p2[1]=op2[1];
	  p1[2]=op1[2];p2[2]=op2[2];
       
	  glBegin(GL_QUADS);
	  glVertex3fv(vert[edges[i].vert].data);
	  glVertex3fv(vert[edges[edges[i].next].vert].data);
	  glVertex4fv(p2);
	  glVertex4fv(p1);
	  glEnd();
      }
    }else{
	Vec3f *n1=&(tri2norm[edges[i].face]);
	Vec3f l1=lightPos-vert[edges[i].vert];
	l1.normalize();
	double d1=n1->dot(l1);

	if(d1>=0){
	    Vec3f op1 = vert[edges[i].vert]-lightPos;
	    Vec3f op2 = vert[edges[edges[i].next].vert]-lightPos;
	    
	    p1[0]=op1[0];p2[0]=op2[0];
	    p1[1]=op1[1];p2[1]=op2[1];
	    p1[2]=op1[2];p2[2]=op2[2];
	    
	    glBegin(GL_QUADS);
	    glVertex3fv(vert[edges[i].vert].data);
	    glVertex3fv(vert[edges[edges[i].next].vert].data);
	    glVertex4fv(p2);
	    glVertex4fv(p1);
	    glEnd();
	}
    }
  }

  glStencilOp(GL_KEEP,GL_KEEP,GL_DECR_WRAP);
  glCullFace(GL_FRONT);

  for(i=0;i<nedges;i++){
    if(edges[i].opp>=0){
      Vec3f *n1=&(tri2norm[edges[i].face]);
      Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
      Vec3f l1=lightPos-vert[edges[i].vert];
      
      double d1=n1->dot(l1);
      double d2=n2->dot(l1);

      if((signf(d1)!=signf(d2)) && signf(d1)>0){
	  Vec3f op1 = vert[edges[i].vert]-lightPos;
	  Vec3f op2 = vert[edges[edges[i].next].vert]-lightPos;
	  
	  p1[0]=op1[0];p2[0]=op2[0];
	  p1[1]=op1[1];p2[1]=op2[1];
	  p1[2]=op1[2];p2[2]=op2[2];
	  
	  glBegin(GL_QUADS);
	  glVertex3fv(vert[edges[i].vert].data);
	  glVertex3fv(vert[edges[edges[i].next].vert].data);
	  glVertex4fv(p2);
	  glVertex4fv(p1);
	  glEnd();
      }
    }else{
	Vec3f *n1=&(tri2norm[edges[i].face]);
	Vec3f l1=lightPos-vert[edges[i].vert];
	l1.normalize();
	double d1=n1->dot(l1);
	
	if(d1>=0){
	    Vec3f op1 = vert[edges[i].vert]-lightPos;
	    Vec3f op2 = vert[edges[edges[i].next].vert]-lightPos;
	    
	    p1[0]=op1[0];p2[0]=op2[0];
	    p1[1]=op1[1];p2[1]=op2[1];
	    p1[2]=op1[2];p2[2]=op2[2];
	    
	    glBegin(GL_QUADS);
	    glVertex3fv(vert[edges[i].vert].data);
	    glVertex3fv(vert[edges[edges[i].next].vert].data);
	    glVertex4fv(p2);
	    glVertex4fv(p1);
	    glEnd();
	}
    }
  }
  glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
  glDisable(GL_STENCIL_TEST);
  glCullFace(GL_BACK);
  glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
  glDepthMask(GL_TRUE);

  glDepthFunc(GL_LESS);
}



void Mesh::drawEdges(Vec3f & lightPos,float offset,float width)
{

    int i=0;
    double ipm[16];
    double pm[16];
    double model[16];
    double proj[16];
    int view[4];
    double x1,x2,y1,y2,z1,z2;

    glMatrixMode(GL_MODELVIEW);
    applyTransform();

    glGetDoublev(GL_PROJECTION_MATRIX,proj);
    glGetDoublev(GL_MODELVIEW_MATRIX,model);
    glGetIntegerv(GL_VIEWPORT,view);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMultMatrixd(proj);
    glMultMatrixd(model);
    glFlush();
    glGetDoublev(GL_MODELVIEW_MATRIX,pm);
    invert(ipm,pm);
    glPopMatrix();

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(view[0],view[2],view[1],view[3],0,-1);
    //gluOrtho2D(view[0],view[2],view[1],view[3]);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    
    glEnable(GL_DEPTH_TEST);
    
    glColor3f(0.0,0.0,0.0);
    glBegin(GL_QUADS);
    
    for(i=0;i<nedges;i++){
	if(edges[i].opp>=0){
	    Vec3f *n1=&(tri2norm[edges[i].face]);
	    Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
	    Vec3f l1=lightPos-vert[edges[i].vert];
	    
	    double d1=n1->dot(l1);
	    double d2=n2->dot(l1);
	    
	    if(signf(d1)>=0 && signf(d2)<0){
		myProject(vert[edges[i].vert].data[0],
			  vert[edges[i].vert].data[1],
			  vert[edges[i].vert].data[2],
			  pm,view,&x1,&y1,&z1);
		myProject(vert[edges[edges[i].next].vert].data[0],
			  vert[edges[edges[i].next].vert].data[1],
			  vert[edges[edges[i].next].vert].data[2],
			  pm,view,&x2,&y2,&z2);
		//z1=2*z1-1;
		//z2=2*z2-1;
		x1+=offset;
		y1+=offset;
		x2+=offset;
		y2+=offset;
		double bx1=x1/((double)view[2]);
		double bx2=x2/((double)view[2]);
		double by1=y1/((double)view[3]);
		double by2=y2/((double)view[3]);

		glMultiTexCoord4f(GL_TEXTURE0,x1,y1,1.0,0.0);
		glMultiTexCoord4f(GL_TEXTURE1,x2,y2,1.0,0.0);
		
		Vec3f dir = Vec3f(x2-x1,y2-y1,0);
		dir.normalize();
		dir.mult(width);
		x1-=dir[0];
		y1-=dir[1];
		x2+=dir[0];
		y2+=dir[1];
		dir.mult(3*width);
		dir.rotatez(M_PI/2.0);
		glMultiTexCoord2f(GL_TEXTURE2,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE3,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE4,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE5,bx1,by1);
		glVertex3f(x1-dir[0],y1-dir[1],z1);
		glVertex3f(x1+dir[0],y1+dir[1],z1);
		glMultiTexCoord2f(GL_TEXTURE2,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE3,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE4,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE5,bx2,by2);
		glVertex3f(x2+dir[0],y2+dir[1],z2);
		glVertex3f(x2-dir[0],y2-dir[1],z2);
		//glVertex3fv(vert[edges[i].vert].data);
		//glVertex3fv(vert[edges[edges[i].next].vert].data);

	    }
	}else{
	    myProject(vert[edges[i].vert].data[0],
			  vert[edges[i].vert].data[1],
			  vert[edges[i].vert].data[2],
			  pm,view,&x1,&y1,&z1);
		myProject(vert[edges[edges[i].next].vert].data[0],
			  vert[edges[edges[i].next].vert].data[1],
			  vert[edges[edges[i].next].vert].data[2],
			  pm,view,&x2,&y2,&z2);
		x1+=offset;
		y1+=offset;
		x2+=offset;
		y2+=offset;

		double bx1=x1/((double)view[2]);
		double bx2=x2/((double)view[2]);
		double by1=y1/((double)view[3]);
		double by2=y2/((double)view[3]);
		
		glMultiTexCoord4f(GL_TEXTURE0,x1,y1,1.0,0.0);
		glMultiTexCoord4f(GL_TEXTURE1,x2,y2,1.0,0.0);

		Vec3f dir = Vec3f(x2-x1,y2-y1,0);
		dir.normalize();
		dir.mult(width);
		x1-=dir[0];
		y1-=dir[1];
		x2+=dir[0];
		y2+=dir[1];
		dir.mult(3*width);	
		dir.rotatez(M_PI/2.0);
		//glTexCoord4f(x1,y1);
		glMultiTexCoord2f(GL_TEXTURE2,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE3,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE4,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE5,bx1,by1);
		glVertex3f(x1-dir[0],y1-dir[1],z1);
		glVertex3f(x1+dir[0],y1+dir[1],z1);
		//glTexCoord2f(x2,y2);
		glMultiTexCoord2f(GL_TEXTURE2,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE3,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE4,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE5,bx2,by2);
		glVertex3f(x2+dir[0],y2+dir[1],z2);
		glVertex3f(x2-dir[0],y2-dir[1],z2);
	}
    }
    glEnd();

    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    
    glDepthFunc(GL_LESS);
}

void Mesh::drawEdges3(Vec3f & lightPos,float offset,float width)
{

    int i=0;
    double ipm[16];
    double pm[16];
    double model[16];
    double proj[16];
    int view[4];
    double x1,x2,y1,y2,z1,z2;

    glMatrixMode(GL_MODELVIEW);
    applyTransform();

    glGetDoublev(GL_PROJECTION_MATRIX,proj);
    glGetDoublev(GL_MODELVIEW_MATRIX,model);
    glGetIntegerv(GL_VIEWPORT,view);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMultMatrixd(proj);
    glMultMatrixd(model);
    glFlush();
    glGetDoublev(GL_MODELVIEW_MATRIX,pm);
    invert(ipm,pm);
    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
    
    glColor3f(0.0,0.0,0.0);
    glBegin(GL_QUADS);
    
    for(i=0;i<nedges;i++){
	if(edges[i].opp>=0){
	    Vec3f *n1=&(tri2norm[edges[i].face]);
	    Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
	    Vec3f l1=lightPos-vert[edges[i].vert];
	    
	    double d1=n1->dot(l1);
	    double d2=n2->dot(l1);
	    
	    if(signf(d1)>=0 && signf(d2)<0){
		myProject(vert[edges[i].vert].data[0],
			  vert[edges[i].vert].data[1],
			  vert[edges[i].vert].data[2],
			  pm,view,&x1,&y1,&z1);
		myProject(vert[edges[edges[i].next].vert].data[0],
			  vert[edges[edges[i].next].vert].data[1],
			  vert[edges[edges[i].next].vert].data[2],
			  pm,view,&x2,&y2,&z2);

		x1+=offset;
		y1+=offset;
		x2+=offset;
		y2+=offset;
		double bx1=x1/((double)view[2]);
		double bx2=x2/((double)view[2]);
		double by1=y1/((double)view[3]);
		double by2=y2/((double)view[3]);

		glMultiTexCoord4f(GL_TEXTURE0,x1,y1,1.0,0.0);
		glMultiTexCoord4f(GL_TEXTURE1,x2,y2,1.0,0.0);
		
		Vec3f dir = Vec3f(x2-x1,y2-y1,0);
		dir.normalize();
		dir.mult(width);
		x1-=dir[0];
		y1-=dir[1];
		x2+=dir[0];
		y2+=dir[1];
		dir.mult(3*width);
		dir.rotatez(M_PI/2.0);
		glMultiTexCoord2f(GL_TEXTURE2,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE3,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE4,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE5,bx1,by1);

		double tx,ty,tz;
		myUnProject(x1-dir[0],y1-dir[1],z1,
			    ipm,view,&tx,&ty,&tz);
		glVertex3f(tx,ty,tz);
		myUnProject(x1+dir[0],y1+dir[1],z1,
			    ipm,view,&tx,&ty,&tz);
		glVertex3f(tx,ty,tz);
		glMultiTexCoord2f(GL_TEXTURE2,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE3,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE4,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE5,bx2,by2);
		myUnProject(x2+dir[0],y2+dir[1],z2,
			    ipm,view,&tx,&ty,&tz);
		glVertex3f(tx,ty,tz);
		myUnProject(x2-dir[0],y2-dir[1],z2,
			    ipm,view,&tx,&ty,&tz);
		glVertex3f(tx,ty,tz);
		//glVertex3fv(vert[edges[i].vert].data);
		//glVertex3fv(vert[edges[edges[i].next].vert].data);

	    }
	}else{
	    myProject(vert[edges[i].vert].data[0],
			  vert[edges[i].vert].data[1],
			  vert[edges[i].vert].data[2],
			  pm,view,&x1,&y1,&z1);
		myProject(vert[edges[edges[i].next].vert].data[0],
			  vert[edges[edges[i].next].vert].data[1],
			  vert[edges[edges[i].next].vert].data[2],
			  pm,view,&x2,&y2,&z2);
		x1+=offset;
		y1+=offset;
		x2+=offset;
		y2+=offset;

		double bx1=x1/((double)view[2]);
		double bx2=x2/((double)view[2]);
		double by1=y1/((double)view[3]);
		double by2=y2/((double)view[3]);
		
		glMultiTexCoord4f(GL_TEXTURE0,x1,y1,1.0,0.0);
		glMultiTexCoord4f(GL_TEXTURE1,x2,y2,1.0,0.0);

		Vec3f dir = Vec3f(x2-x1,y2-y1,0);
		dir.normalize();
		dir.mult(width);
		x1-=dir[0];
		y1-=dir[1];
		x2+=dir[0];
		y2+=dir[1];
		dir.mult(3*width);	
		dir.rotatez(M_PI/2.0);
		//glTexCoord4f(x1,y1);
		glMultiTexCoord2f(GL_TEXTURE2,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE3,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE4,bx1,by1);
		glMultiTexCoord2f(GL_TEXTURE5,bx1,by1);
		glVertex3f(x1-dir[0],y1-dir[1],z1);
		glVertex3f(x1+dir[0],y1+dir[1],z1);
		//glTexCoord2f(x2,y2);
		glMultiTexCoord2f(GL_TEXTURE2,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE3,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE4,bx2,by2);
		glMultiTexCoord2f(GL_TEXTURE5,bx2,by2);
		glVertex3f(x2+dir[0],y2+dir[1],z2);
		glVertex3f(x2-dir[0],y2-dir[1],z2);
	}
    }
    glEnd();

    glDepthFunc(GL_LESS);
}


void Mesh::projectEdges(Vec3f & lightPos,double * model,double * proj,
			int * view,
			float * silmap[2])
{
  double pm[16];
  int i;
  double x1,y1,z1,x2,y2,z2;
  int x1i,y1i,x2i,y2i;

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMultMatrixd(proj);
  glMultMatrixd(model);
  glFlush();
  glGetDoublev(GL_MODELVIEW_MATRIX,pm);
  glPopMatrix();

  for(i=0;i<nedges;i++){
    if(edges[i].opp>=0){
      Vec3f *n1=&(tri2norm[edges[i].face]);
      Vec3f *n2=&(tri2norm[edges[edges[i].opp].face]);
      Vec3f l1=lightPos-vert[edges[i].vert];
      
      double d1=n1->dot(l1);
      double d2=n2->dot(l1);

      if(signf(d1)!=signf(d2)){
	myProject(vert[edges[i].vert].data[0],
		  vert[edges[i].vert].data[1],
		  vert[edges[i].vert].data[2],
		  pm,view,&x1,&y1,&z1);
	myProject(vert[edges[edges[i].next].vert].data[0],
		  vert[edges[edges[i].next].vert].data[1],
		  vert[edges[edges[i].next].vert].data[2],
		  pm,view,&x2,&y2,&z2);
	
	scanConvert(silmap,view,x1,y1,x2,y2);
      }
    }else{
     
      myProject(vert[edges[i].vert].data[0],
		vert[edges[i].vert].data[1],
		vert[edges[i].vert].data[2],
		pm,view,&x1,&y1,&z1);
      myProject(vert[edges[edges[i].next].vert].data[0],
		vert[edges[edges[i].next].vert].data[1],
		vert[edges[edges[i].next].vert].data[2],
		pm,view,&x2,&y2,&z2);

      scanConvert(silmap,view,x1,y1,x2,y2);
    }
  }
}

void Mesh::initEdges()
{
  int i=0;
  int j=0;
  int edgei=0;

  

  if(edges){
    delete [] edges;
  }
  nedges=ntri*3;

  Node ** nodes=new Node*[nvert];
  memset(nodes,0,sizeof(Node*)*nvert);

  edges=new Edge[nedges];
  memset(edges,0xFF,sizeof(Edge)*nedges);
  
  Node ** pnodes=new Node*[nvert];
  memset(pnodes,0,sizeof(Node*)*nvert);
  
  for(i=0;i<ntri;i++){
    for(int g=0;g<3;g++){
      Node * n = new Node();
      n->edge=i;
      n->next=pnodes[tri[i][g]];
      pnodes[tri[i][g]]=n;
    }
  }
  printf("inserted\n");

   int equal=0;
  for(i=nvert;i<nvert;i++){
    Vec3f v1=vert[i];
    for(j=i+1;j<nvert;j++){
      Vec3f v2=vert[j];
      Vec3f diff=v1-v2;
      if(diff.len()<0.0000000001){
        Node * finger=pnodes[j];
        Node * last=NULL;
      
        printf("%d==%d\n",i,j);
        while(finger){
          int pi=finger->edge;
          for(int g=0;g<3;g++){
            if(tri[pi][g]==j){
              tri[pi][g]=i;
            }
          }
          last=finger;
          finger=finger->next;
        }
        if(last){
          last->next=pnodes[i];
          pnodes[i]=pnodes[j];
          pnodes[j]=0;
        }
        equal++;
      }
    }
    if(i%1000==0){
      printf("%d/%d\n",i,nvert);
    }
  }
  printf("equal = %d\n",equal);

  for(i=0;i<ntri;i++){
    int basei=edgei;

    for(j=0;j<3;j++){
	Node * n=new Node();
	n->next=nodes[tri[i][j]];
	n->edge=edgei;
	nodes[tri[i][j]]=n;

	edges[edgei].face=i;
	edges[edgei].vert=tri[i][j];
	edges[edgei].next=((j+1)%3)+basei;
	edges[edgei].prev=((j-1)%3)+basei;
	edgei++;
    }
  }

  printf("nedges %d,ntri=%d,ndgei=%d\n",nedges,
         ntri,edgei);
  //Can this be any less efficient.
  //loop through all edges and set the opposite faces correctly
  int found=0;
  int seen=0;
  for(i=0;i<nvert;i++){
    Node * finger=nodes[i];
    
    while(finger){
      int e1v1=i;
      int e1v2=edges[edges[finger->edge].next].vert;
      Node * middleFinger=nodes[e1v2];
      int foundit=0;
      while(middleFinger){
        int e2v1=edges[middleFinger->edge].vert;
	int e2v2=edges[edges[middleFinger->edge].next].vert;
	
        if((e1v1==e2v2) && (e1v2==e2v1) && middleFinger!=finger){
          foundit=1;
          found+=2;
          edges[finger->edge].opp=middleFinger->edge;
          edges[middleFinger->edge].opp=finger->edge;
          break;
        }
        middleFinger=middleFinger->next;
      }
      finger=finger->next;
      seen++;
    }
  }
  printf("seen %d\n",seen);
  for(i=0;i<nvert;i++){
      Node * finger;
      finger=nodes[i];
      Node * pinkyFinger = nodes[i];
      while(finger){
	  pinkyFinger=finger;
	  finger = finger->next;
	  delete pinkyFinger;
      }
  }
  printf("Found %d\n",found);
  delete nodes;

/*
  found=0;
  for(i=0;i<nedges;i++){
    int e1v1=edges[i].vert;
    int e1v2=edges[edges[i].next].vert;
    int foundit=0;
    for(j=0;j<nedges;j++){
      int e2v1=edges[j].vert;
      int e2v2=edges[edges[j].next].vert;
      if((e1v1==e2v2 && e1v2==e2v1)){
        foundit=1;
        edges[i].opp=j;
        edges[j].opp=i;
      }
    }
    if(!foundit){
      
    }
    else{
      found+=2;
      //printf("Found opposite edge\n");
    }
    }*/
  printf("Found %d of %d\n",found,nedges);
  printf("calculating normals\n");
  calculateNormals();
}

void Mesh::calculateNormals()
{
  int i=0;
  int * count = new int[nvert];
  float scale=1;
 
  //assume md2 and reverse normals
  if(nframes){
    printf("reversing frames\n");
      scale=-1;
  }
  //Allocate the normals if they are not already
  
  if(!norm)
      norm = new Vec3f[nvert];
  if(!tri2norm)
      tri2norm = new Vec3f[ntri];
  memset(norm,0,sizeof(Vec3f)*nvert);
  memset(count,0,sizeof(int)*nvert);
  for(i=0;i<ntri;i++){
      Vec3f a = vert[tri[i][1]]-vert[tri[i][0]];
      Vec3f b = vert[tri[i][2]]-vert[tri[i][1]];
      Vec3f n = a.cross(b);
      
      n.mult(scale);
      n.normalize();
      tri2norm[i]=n;

      norm[tri[i][0]].add(n);
      norm[tri[i][1]].add(n);
      norm[tri[i][2]].add(n);
      count[tri[i][0]]++;
      count[tri[i][1]]++;
      count[tri[i][2]]++;
  }
  for(i=0;i<nvert;i++){
    printf("b:%f %f %f\n",norm[i].x,norm[i].y,norm[i].z);
    norm[i].normalize();
    printf("a:%f %f %f\n",norm[i].x,norm[i].y,norm[i].z);
  }
  delete [] count;
}

void Mesh::applyTransform()
{
    glTranslatef(pos[0],pos[1],pos[2]);
    glScalef(scale[0],scale[1],scale[2]);
    glRotatef(rot[2],0,0,1);
    glRotatef(rot[1],0,1,0);
    glRotatef(rot[0],1,0,0);
}

void Mesh::applyITransform()
{
    glRotatef(-rot[0],1,0,0);
    glRotatef(-rot[1],0,1,0);
    glRotatef(-rot[2],0,0,1);
    glScalef(1.0/scale[0],1.0/scale[1],1.0/scale[2]);
    glTranslatef(-pos[0],-pos[1],-pos[2]);
}


void Mesh::draw(int transform)
{
  int i=0;

  glEnable(GL_CULL_FACE);
  if(nframes)
      glFrontFace(GL_CW);
  else
      glFrontFace(GL_CCW);

  if(transform){
      applyTransform();
  }
  
  if(list>=0){
    glCallList(list);
    return;
  }
 
  //No support for more than one set of
  //texture coordinates and one texture
  //currently
  int drw=1;
  if(drw){
  if(tvert && tvert[0]){
      
      //This could be bad news
      if(tex){
	  
		  glActiveTexture(GL_TEXTURE0);
		  glBindTexture(GL_TEXTURE_2D,this->tex[0]);
      }
	  glMaterialfv(GL_FRONT_AND_BACK,
		       GL_AMBIENT_AND_DIFFUSE,surf[tri2surf[i]].color);
	  /*glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,
	    white);*/
		
	  glMaterialfv(GL_FRONT_AND_BACK,
				 GL_SPECULAR,surf[tri2surf[i]].color);
	  glColor4fv(surf[tri2surf[i]].color);

      glBegin(GL_TRIANGLES);
      for(i=0;i<ntri;i++){
	  glNormal3fv(tri2norm[i].data);
	  glMultiTexCoord2fv(GL_TEXTURE0,tvert[0][tri2tvert[i][0]].data);
	  glVertex3fv(vert[tri[i][0]].data);
	  glMultiTexCoord2fv(GL_TEXTURE0,tvert[0][tri2tvert[i][1]].data);
	  glVertex3fv(vert[tri[i][1]].data);
	  glMultiTexCoord2fv(GL_TEXTURE0,tvert[0][tri2tvert[i][2]].data);
	  glVertex3fv(vert[tri[i][2]].data);
      }
      glEnd();
     
      //glActiveTexture(GL_TEXTURE0);
      //glDisable(GL_TEXTURE_2D);
      
  }
  //no texture, just draw the mesh
  else if(0){
      glBegin(GL_TRIANGLES);
      for(i=0;i<ntri;i++){
	  glNormal3fv(norm[tri[i][0]].data);
	  glVertex3fv(vert[tri[i][0]].data);
	  
	  glNormal3fv(norm[tri[i][1]].data);
	  glVertex3fv(vert[tri[i][1]].data);
	  
	  glNormal3fv(norm[tri[i][2]].data);
	  glVertex3fv(vert[tri[i][2]].data);
      }
      glEnd();
  }
  else{
      glBegin(GL_TRIANGLES);
      float white[]={1.0,1.0,1.0,1.0};
	  float grey[]={0.2,0.2,0.2,1.0};
      for(i=0;i<ntri;i++){
	  glMaterialfv(GL_FRONT_AND_BACK,
		       GL_AMBIENT_AND_DIFFUSE,surf[tri2surf[i]].color);
	  /*glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,
	    white);*/
	  if(shinny){

		  glMaterialfv(GL_FRONT_AND_BACK,
		       GL_AMBIENT,white);
		 glMaterialfv(GL_FRONT_AND_BACK,
		       GL_SPECULAR,white);
	  }
	  else
		glMaterialfv(GL_FRONT_AND_BACK,
		       GL_SPECULAR,grey);
	  glColor4fv(surf[tri2surf[i]].color);
	  
	  if(flatShading){    
	      glNormal3fv(tri2norm[i].data);
	      glVertex3fv(vert[tri[i][0]].data);
	      glVertex3fv(vert[tri[i][1]].data);
	      glVertex3fv(vert[tri[i][2]].data);
	  }
	  else{
	      
	      glNormal3fv(norm[tri[i][0]].data);
	      glVertex3fv(vert[tri[i][0]].data);
	      glNormal3fv(norm[tri[i][1]].data);
	      glVertex3fv(vert[tri[i][1]].data);
	      glNormal3fv(norm[tri[i][2]].data);
	      glVertex3fv(vert[tri[i][2]].data);
	  }
      }
      glEnd();
  }
  }
  glFrontFace(GL_CCW);
  //glDisable(GL_CULL_FACE);
}

int tagEqual(char * a,char * b)
{
  return a[0]==b[0]&&a[1]==b[1]
    &&a[2]==b[2]&&a[3]==b[3];
}


void Mesh::genTriangle()
{
  nvert=3;
  vert=new Vec3f[3];
  vert[0][0]=-1.0;
  vert[0][1]=0;
  vert[0][2]=-1.0;

  vert[1][0]= 1.0;
  vert[1][1]=0;
  vert[1][2]=-1.0;

  vert[2][0]=0;
  vert[2][1]=0;
  vert[2][2]=1.0;
  /*
  vert[0].rotatey(M_PI/8.0);
  vert[1].rotatey(M_PI/8.0);
  vert[2].rotatey(M_PI/8.0);
  */
  ntri=1;
  tri = new Triangle[ntri];
  tri2surf = new int[ntri];
  tri[0][0]=0;
  tri[0][1]=1;
  tri[0][2]=2;

  //tri[1][0]=0;
  //tri[1][1]=2;
  //tri[1][2]=1;

  tri2surf[0]=0;
  //tri2surf[1]=1;

  nsurf=1;
  surf = new Surface[1];
  surf[0].name=0;
  surf[0].color[0]=0.1;
  surf[0].color[1]=0.1;
  surf[0].color[2]=0.8;
  surf[0].color[3]=1.0;
  initEdges();
}

void Mesh::genPlane()
{
    float inc=0.1;
    int width=0;
    nvert=0;
    ntri=0;
    for(float i=-1.0;i<(1.0+inc);i+=inc){
	width=0;
	for(float j=-1.0;j<(1.0+inc);j+=inc){
	    nvert++;
	    width++;
	    if(i<1.0 && j<1.0){
		ntri+=2;
	    }
	}
    }
    printf("%d %d\n",nvert,ntri);
    vert=new Vec3f[nvert];
    tvert=new Vec2f *[1];
    tvert[0]=new Vec2f[nvert];
    tri=new Triangle[ntri];
    tri2tvert=new Triangle[ntri];
    int ii=0;
    int jj=0;
    int trii=0;
    for(float i=-1.0;i<(1.0+inc);i+=inc){	
	jj=0;
	for(float j=-1.0;j<(1.0+inc);j+=inc){
	    vert[ii*width+jj]=Vec3f(j,0,i);
	    tvert[0][ii*width+jj][0]=(j+1.0)/2.0;
	    tvert[0][ii*width+jj][1]=(i+1.0)/2.0;
	    if(i<1.0 && j<1.0){
		tri[trii][0]=ii*width+jj;
		tri[trii][1]=(ii+1)*width+jj;
		tri[trii][2]=(ii+1)*width+jj+1;
		tri2tvert[trii][0]=tri[trii][0];
		tri2tvert[trii][1]=tri[trii][1];
		tri2tvert[trii][2]=tri[trii][2];
		trii++;

		tri[trii][0]=ii*width+jj;
		tri[trii][1]=(ii+1)*width+jj+1;
		tri[trii][2]=(ii)*width+jj+1;
		tri2tvert[trii][0]=tri[trii][0];
		tri2tvert[trii][1]=tri[trii][1];
		tri2tvert[trii][2]=tri[trii][2];
		trii++;
	    }
	    jj++;
	}
	ii++;
    }
    printf("inited vert\n");
    nsurf=1;
    surf = new Surface[nsurf];
    surf[0].name=0;
    surf[0].color[0]=0.5;
    surf[0].color[1]=0.4;
    surf[0].color[2]=0.4;
    surf[0].color[3]=1.0;
    
    tri2surf = new int[ntri];
    memset(tri2surf,0,sizeof(int)*ntri);

    initEdges();
}

void Mesh::genSphere(){
  float theta=0;
  float phi=0;
  float density=(M_PI/8);
  int nply=0;
  
  nvert=0;
  for(phi=0;phi<(M_PI+density/2.0);phi+=density){
    for(theta=0;theta<2*M_PI;theta+=density){
      nvert++;
    }
  }
  printf("%d\n",nvert);
  vert=new Vec3f[nvert];
  int vi=0;
  int nphi=0;
  int ntheta=2*nphi;
  for(phi=0;phi<=(M_PI+density/2.0);phi+=density){
    ntheta=0;
    for(theta=0;theta<2*M_PI;theta+=density){
      vert[vi][0]=0.4*sin(phi)*cos(theta);
      vert[vi][1]=0.4*sin(phi)*sin(theta);
      vert[vi][2]=0.4*cos(phi);
      vi++;
      ntheta++;
    }
    nphi++;
  }
  printf("%d,%d %d\n",vi,nvert,nphi*ntheta);
 
  ntri=nvert*3;
  tri = new Triangle[ntri];
  tri2surf = new int[ntri];

  memset(tri,0,sizeof(Triangle)*ntri);
  memset(tri2surf,0,sizeof(int)*ntri);

  printf("allocated mem for sphere %d triangles\n",ntri);
  nply=0;
  int phii=0;
  int thetai=0;
  for(phi=0;phi<(M_PI-density+density/2.0);phi+=density){
    thetai=0;
    for(theta=0;theta<(2*M_PI-density);theta+=density){
	
      tri[nply][0]=phii*ntheta+thetai;
      tri[nply][1]=(phii+1)*ntheta+thetai;
      tri[nply][2]=(phii+1)*ntheta+thetai+1;
      tri2surf[nply]=nply%2;
      nply++;

      tri[nply][0]=phii*ntheta+thetai;
      tri[nply][1]=(phii+1)*ntheta+thetai+1;
      tri[nply][2]=phii*ntheta+thetai+1;
      tri2surf[nply]=nply%2;

      thetai++;
      nply++;
    }
    
    tri[nply][0]=phii*ntheta+thetai;
    tri[nply][1]=(phii+1)*ntheta+thetai;
    tri[nply][2]=(phii+1)*ntheta;
    tri2surf[nply]=nply%2;
    nply++;
    
    tri[nply][0]=phii*ntheta+thetai;
    tri[nply][2]=(phii+1)*ntheta;
    tri[nply][3]=phii*ntheta;
    tri2surf[nply]=nply%2;
    nply++;

    thetai++;
    nply++;
    phii++;
  }

  thetai=0;
  for(theta=0;theta<(2*M_PI-density);theta+=density){
      tri[nply][0]=phii*ntheta+thetai;
      tri[nply][1]=(nphi-1)*(ntheta);
      tri[nply][2]=(nphi-1)*(ntheta);
      tri2surf[nply]=0;
      nply++;
      
      tri[nply][0]=phii*ntheta+thetai;
      tri[nply][1]=(nphi-1)*(ntheta);
      tri[nply][2]=phii*ntheta+thetai+1;
      tri2surf[nply]=0;
      thetai++;
      nply++;
  }
  
  tri[nply][0]=phii*ntheta+thetai;
  tri[nply][1]=(nphi-1)*ntheta+thetai;
  tri[nply][2]=(nphi-1)*ntheta;
  tri2surf[nply]=0;
  nply++;

  tri[nply][0]=phii*ntheta+thetai;
  tri[nply][1]=(nphi-1)*ntheta;
  tri[nply][2]=phii*ntheta;
  tri2surf[nply]=0;
  thetai++;
  nply++;
  phii++;
  
  printf("gettin surfaces %d(%d),%d\n",ntri,nply,nvert);
  //exit(0);
                              
  ntri=nply;
  nsurf=2;
  printf("about to allocate surfaces\n");
  surf = new Surface[2];
  printf("allocated surfaces\n");
  surf[0].color[0]=0.4;
  surf[0].color[1]=0.4;
  surf[0].color[2]=0.0;
  surf[0].color[3]=1.0;
  surf[0].name=0;

  surf[1].color[0]=0.0;
  surf[1].color[1]=0.0;
  surf[1].color[2]=0.4;
  surf[1].color[3]=1.0;
  surf[1].name=0;

  printf("about to initedges\n");
  initEdges();
  /*
  for(int i=0;i<nply;i++){
    Vec3f norm=Vec3f(0,0,0);
    norm=norm+vert[poly[i].vert[0]];
    norm=norm+vert[poly[i].vert[1]];
    norm=norm+vert[poly[i].vert[2]];
    norm=norm+vert[poly[i].vert[3]];
    norm.mult(0.25);
    norm.normalize();
    poly[i].norm=norm;
    }*/

}

//From http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef struct md2header{
    int     id;              //"IPD2"
    int     version;    
    int     texw;         
    int     texh;        
    int     framesize;          // size of one frame in bytes

    int     ntex;        
    int     nvert; 
    int     ntc;   
    int     ntri; 
    int     ngl;   
    int     nframes;
    
    int     ofs_skins;          // offset to skin names (64 bytes each)
    int     ofs_st;             // offset to s-t texture coordinates
    int     ofs_tris;           // offset to triangles
    int     ofs_frames;         // offset to frame data
    int     ofs_glcmds;         // offset to opengl commands
    int     ofs_end;            // offset to end of file
}MD2Header;

//From http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef struct T_MD2TexCoord{
    short s;
    short t;
}MD2TexCoord;

//From http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef struct T_MD2Vertex{
    unsigned char pos[3];
    unsigned char normalIndex;
}MD2Vertex;

//From http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef struct T_MD2Frame{
    float       scale[3];
    float       translate[3];  
    char        name[16];      
    MD2Vertex    vert[1]; 
}MD2Frame;

//From http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef struct T_MD2Triangle{
    short   vert[3];   
    short   tvert[3];
}MD2Triangle;

//Somewhat based on tutorial at:
//http://tfc.duke.free.fr/us/tutorials/models/md2.htm
int Mesh::loadMD2(char * filename,char * skinFile)
{
    unsigned char * buffer;
    int ID;
    int i=0;
    char c[5];
    MD2Header header;
    FILE * file = fopen(filename,"rb");
    if(!file)return 0;

    ID=((int)'I')|(((int)'D')<<8)|(((int)'P')<<16)|(((int)'2')<<24);
    
   
    fread(&header,sizeof(MD2Header),1,file);
    
    if(header.id!=ID){
	printf("%d!=%d\n",ID,header.id);
	return 0;
    }

    nvert=header.nvert;
    ntri=header.ntri;
    ntvert=header.ntc;
    ntex=1;
    nframes=header.nframes;

    printf("Got valid MD2 file %d %d %d\n",header.nvert,ntri,nframes);

    vert = new Vec3f[nvert];
    realVert=new Vec3f[nvert*nframes];
    tri=new Triangle[ntri];
    tri2surf = new int[ntri];
    tri2tvert =new Triangle[ntri];
    nsurf=1;
    surf = new Surface[1];
    surf[0].name=0;
    surf[0].color[0]=1.0;
    surf[0].color[1]=1.0;
    surf[0].color[2]=1.0;
    surf[0].color[3]=1.0;

    memset(tri2surf,0,sizeof(int)*ntri);
    tex = new GLuint[ntex];
    glGenTextures(ntex,tex);
    tvert = new Vec2f *[ntex];
    for(i=0;i<ntex;i++){
	tvert[i]=new Vec2f[ntvert];
    }

    buffer=new unsigned char[ntvert*sizeof(MD2TexCoord)];
    fseek(file,header.ofs_st,SEEK_SET);
    fread(buffer,ntvert,sizeof(MD2TexCoord),file);
    MD2TexCoord * tc=(MD2TexCoord*)buffer;

    for(i=0;i<ntvert;i++){
	tvert[0][i][0]=((float)tc[i].s)/((float)header.texw);
	tvert[0][i][1]=((float)tc[i].t)/((float)header.texh);
    }

    delete [] buffer;

    int lastslash=0;
    buffer=new unsigned char[header.ntex*64];
    fseek(file,header.ofs_skins,SEEK_SET);
    fread(buffer,1,header.ntex*64,file);
   
    for(i=0;buffer[i]!=0 && i<64;i++){
	if(buffer[i]=='/')
	    lastslash=i;
	if(buffer[i]=='.'){
	    buffer[i+1]='p';
	    buffer[i+2]='p';
	    buffer[i+3]='m';
	    break;
	}
    }
   
    char ifilename[256];

    if(skinFile)sprintf(ifilename,skinFile);
    else sprintf(ifilename,"models%s",buffer+lastslash);
    Image8 image;
    image.readPPM((char *)ifilename);
    
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex[0]);
    image.initTexture();
   
    delete [] buffer;


    buffer=new unsigned char[header.framesize*header.nframes];
    
    fseek(file,header.ofs_frames,SEEK_SET);
    int read=fread(buffer,1,header.framesize*header.nframes,file);
    printf("read %d, shoudl have %d\n",read,header.framesize*nframes);
    printf("allocated %d\n",nvert*nframes);

    float miny=100000;

    for(int j=0;j<nframes;j++){
	MD2Frame * frame=(MD2Frame *)(buffer+header.framesize*j);
	
	for(i=0;i<nvert;i++){
	    realVert[i+nvert*j][0]=
		(frame->vert[i].pos[0]*frame->scale[0])+frame->translate[0];
	    realVert[i+nvert*j][1]=
		(frame->vert[i].pos[1]*frame->scale[1])+frame->translate[1];
	    realVert[i+nvert*j][2]=
		(frame->vert[i].pos[2]*frame->scale[2])+frame->translate[2];
	    realVert[i+nvert*j].rotatex(-M_PI/2.0);
	    realVert[i+nvert*j].rotatey(M_PI/2.0);

	    realVert[i+nvert*j][0]*=0.1;
	    realVert[i+nvert*j][1]*=0.1;
	    realVert[i+nvert*j][2]*=0.1;

	    if(realVert[i+nvert*j][1]<miny){
		miny=realVert[i+nvert*j][1];
	    }
	}
    }
    printf("miny %f\n",miny);
    
    for(int j=0;j<nframes;j++){
	for(i=0;i<nvert;i++){
	    realVert[i+nvert*j][1]-=(miny+0.3);
	}
    }

    delete [] buffer;
    MD2Triangle * tbuffer=new MD2Triangle[ntri];
    
    fseek(file,header.ofs_tris,SEEK_SET);
    fread(tbuffer,sizeof(MD2Triangle),ntri,file);
    
    for(i=0;i<ntri;i++){
	tri[i][0]=tbuffer[i].vert[0];
	tri[i][1]=tbuffer[i].vert[1];
	tri[i][2]=tbuffer[i].vert[2];
	tri2tvert[i][0]=tbuffer[i].tvert[0];
	tri2tvert[i][1]=tbuffer[i].tvert[1];
	tri2tvert[i][2]=tbuffer[i].tvert[2];
    }
    //   ntri=10;
    //ntex=0;
    delete [] tbuffer;
    fclose(file);

    initEdges();
    setScale(Vec3f(1.0,1.0,1.0));
    setPos(Vec3f(0.0,0.0,0.0));
    setAnimState(RUN);
    move(0);

    printf("done loading MD2\n");
}

int Mesh::loadLWO(char * filename)
{
  unsigned char sizeBuffer[4];
  char buffer[1024];
  char tag[4];
  int fileSize=0;
  int readSoFar=0;
  unsigned int size;
  FILE * file=fopen(filename,"rb");
  if(!file){printf("error opening %s\n",filename);return 0;}

  if(4!=fread(tag,1,4,file)){
    printf("premature end of file %s\n",filename);return 0;
  }
  printf("loading lwo %s\n",filename);
  fread(sizeBuffer,1,4,file);
  fileSize=sizeBuffer[3]|(sizeBuffer[2]<<8)|(sizeBuffer[1]<<16)|(sizeBuffer[0]<<24);

  if(tagEqual(tag,"FORM")){
    fread(buffer,1,4,file);
    if(!tagEqual(buffer,"LWOB"))
      {printf("invalid form type %s\n",buffer);return 0;}
    printf("buffer :%s,%d\n",buffer,fileSize);
  }else{
    printf("not a valid LWO file\n");return 0;
  }

  while(1){
    int nreadTag=0;
    if(4!=(nreadTag=fread(tag,1,4,file))){
      if(nreadTag==0)
	break;
      printf("premature end of file %s\n",filename);return 0;
    }
    if(4!=fread(sizeBuffer,1,4,file)){
      break;
    }
    size=sizeBuffer[3]|(sizeBuffer[2]<<8)|
      (sizeBuffer[1]<<16)|(sizeBuffer[0]<<24);
  
    if(tagEqual(tag,"PNTS")){
      nvert=size/12;
      vert=new Vec3f[nvert];
      for(unsigned i=0;i<size/12;i++){
	unsigned char fbuf[12];
	if(12!=fread(fbuf,1,12,file)){
	  printf("premature end of file %s\n",filename);return 0;
	  return 0;
	}
	unsigned int d0=fbuf[3]|(fbuf[2]<<8)|(fbuf[1]<<16)|(fbuf[0]<<24);
	unsigned int d1=fbuf[7]|(fbuf[6]<<8)|(fbuf[5]<<16)|(fbuf[4]<<24);
	unsigned int d2=fbuf[11]|(fbuf[10]<<8)|(fbuf[9]<<16)|(fbuf[8]<<24);
	memcpy(vert[i].data,&d0,4);
	memcpy(vert[i].data+1,&d1,4);
	memcpy(vert[i].data+2,&d2,4);
      }
    }
    else if(tagEqual(tag,"SRFS")){
      int last=0;
      char * buf = new char[size];
      int maxsurf=10;

      surf=new Surface[maxsurf];
 
      if(size!=fread(buf,1,size,file)){
	printf("Premature end of file SRFS\n");
	return 0;
      }
      for(unsigned int i=0;i<size;){
	if(buf[i]==0){
	  if(nsurf>=maxsurf){
	    Surface * back=surf;
	    maxsurf+=10;
	    surf=new Surface[maxsurf];
	    memcpy(surf,back,nsurf*sizeof(Surface));
	    delete back;
	  }
	  surf[nsurf].name=new char[strlen(buf+last)+1];
	  surf[nsurf].color[0]=0.5;
	  surf[nsurf].color[1]=0.5;
	  surf[nsurf].color[2]=0.5;
	  surf[nsurf].color[3]=1.0;
	  strcpy(surf[nsurf].name,buf+last);
	  nsurf++;

	  last=i+1;
	  if(!(i&0x1)){
	    i++;
	    last++;
	  }
	}
	i++;
      }
      delete buf;
    }
    else if(tagEqual(tag,"PCHS")){
      unsigned nread=0;
      unsigned char * buf;
      buf = new unsigned char[size];
      while(nread<size){
	int vertices;
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file %s\n","POLS");
	  return 0;
	}
	vertices=buf[1+nread]|(buf[nread]<<8);	
	nread+=2;

	for(int j=0;j<vertices;j++){
	  if(2!=fread(buf+nread,1,2,file)){
	    printf("premature end of file\n");
	    return 0;
	  }
	  nread+=2;
	}
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file\n");
	  return 0;
	}
	nread+=2;
      }
      delete []buf;
      printf("PCHS not handled correctly yet\n");
    }
    else if(tagEqual(tag,"CRVS")){
      unsigned nread=0;
      unsigned char * buf;
      buf = new unsigned char[size];
      while(nread<size){
	int vertices;
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file %s\n","CRVS");
	  return 0;
	}
	vertices=buf[1+nread]|(buf[nread]<<8);	
	nread+=2;

	for(int j=0;j<vertices;j++){
	  if(2!=fread(buf+nread,1,2,file)){
	    printf("premature end of file\n");
	    return 0;
	  }
	  nread+=2;
	}
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file\n");
	  return 0;
	}
	nread+=2;
	//curve flags
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file\n");
	  return 0;
	}
	nread+=2;
      }
      delete []buf;
      printf("CRVS not handled correctly yet\n");
    }
    else if(tagEqual(tag,"POLS")){
      unsigned int nread=0;
      unsigned char * buf;
      buf = new unsigned char[size];
      ntri=0;
      while(nread<size){
	int vertices;
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file %s\n","POLS");
	  return 0;
	}
	vertices=buf[1+nread]|(buf[nread]<<8);	
	nread+=2;

	for(int j=0;j<vertices;j++){
	  if(2!=fread(buf+nread,1,2,file)){
	    printf("premature end of file\n");
	    return 0;
	  }
	  nread+=2;
	}
	if(2!=fread(buf+nread,1,2,file)){
	  printf("premature end of file\n");
	  return 0;
	}
	nread+=2;

	for(int j=0;j<vertices-2;j++){
	    ntri++;
	}

	
      }
      tri=new Triangle[ntri];
      tri2surf=new int[ntri];
      //tri2norm = new Vec3f[ntri];

      int pi=0;
      nread=0;
      while(nread<size){
	int vertices;
	int * vertbuf;
	vertices=buf[1+nread]|(buf[nread]<<8);	
	nread+=2;

	vertbuf=new int[vertices];
	for(int j=0;j<vertices;j++){
	  vertbuf[j]=buf[1+nread]|(buf[nread]<<8);
	  if(vertbuf[j]>=nvert){
	      printf("error triangle:%d %d %d\n",pi,nvert,vertices);
	  }
	  nread+=2;
	}
	
	int srf=buf[nread+1]|(buf[nread]<<8);
	nread+=2;
	if(srf==-1){
	  printf("Detail polygons not handled correctly\n");
	  vertices=buf[1+nread]|(buf[nread]<<8);
	  nread+=2;
	  nread+=vertices;
	}
	
	for(int j=0;j<vertices-2;j++){
	    tri[pi][0]=vertbuf[0];
	    tri[pi][1]=vertbuf[j+1];
	    tri[pi][2]=vertbuf[j+2];
	    tri2surf[pi]=srf%nsurf;;
		//printf("Surface %d\n",srf);
	    pi++;
	}
	delete vertbuf;
      }
      delete buf;
    }
    else if(tagEqual(tag,"SURF")){
      unsigned int i=0;
      unsigned short size2;
      int surfi=0;
      while(1){
		  printf("Got surf tag\n");
	fread(buffer+i,1,1,file);
	if(buffer[i]==0 && (i&0x1==0)){
	  i+=2;
	  break;
	}
	else if(buffer[i]==0 && (i&0x1==1)){
	  i++;
	  break;
	}
	i++;
      }
      
      for(surfi=0;surfi<nsurf;surfi++){
	if(strcmp(buffer,surf[surfi].name)==0){
	  break;
	}
      }
      if(surfi==nsurf){
	printf("Did not find surface\n");
      }
      unsigned char * temp = new unsigned char[size];
      fread(temp,1,size-i,file);
      size=size-i;

      for(i=0;i<size;){
	char * tag2=(char *)(temp+i);
	i+=4;
	size2=temp[i+1]|(temp[i+0]<<8);
	i+=2;
	if(tagEqual(tag2,"COLR")){
	  printf("Got COLR\n");
	  surf[surfi].color[0]=((float)temp[i])/255.0f;
	  surf[surfi].color[1]=((float)temp[i+1])/255.0f;
	  surf[surfi].color[2]=((float)temp[i+2])/255.0f;
	  surf[surfi].color[3]=((float)temp[i+3])/255.0f;  
	  i+=4;
	}
	else if(tagEqual(tag2,"DIFF")){
	  printf("Got DIFF\n");
	  surf[surfi].color[0]=((float)temp[i])/255.0f;
	  surf[surfi].color[1]=((float)temp[i+1])/255.0f;
	  surf[surfi].color[2]=((float)temp[i+2])/255.0f;
	  surf[surfi].color[3]=((float)temp[i+3])/255.0f;  
	  i+=4;
	}
	else{
	  tag2[4]=0;
	  printf("Unknown SURF tubChunk %s, size=%d\n",tag2,size2);
	  i+=size;
	}
      }

      delete temp;
    }
    else {
      char * temp=(char*)malloc(size);
      if(size!=fread(temp,1,size,file)){
	printf("premature end of file %s\n",filename);return 0;
      }
      tag[4]=0;
      printf("Unknown tag %s, %d\n",tag,size);
      
    }
    //else printf("%s\n",tag);
  }
  printf("done loading lwo %s, nsurf=%d, ntri=%d",filename,nsurf,ntri);
  for(int k=0;k<nsurf;k++){
	  printf("%f %f %f %f\n",surf[k].color[0],
			surf[k].color[1],
			surf[k].color[2],
			surf[k].color[3]);
  }
  initEdges();
}

void Mesh::setAnimState(MD2AnimType t)
{
    if(t>=0 && t<MAX_ANIMATIONS){
	animState=t;
	animFrame=animstats[animState].first;
    }
    if(animState==JUMP){
	jumpApex=0;
    }
}

void Mesh::control(int bdown,
		   unsigned char * kd,
		   float dt,
		   float dx,
		   float dy,
		   int * mpos,
		   float lastx,
		   float lasty,
		   int freeze,
		   float uiaccel)
{
    static double rot0speed=0;
    static double rot1speed=0;
    static double sspeed=0;
    static double uspeed=0;
    Vec3f dir(0,0,-1);

    if(freeze)return;

    dx=mpos[0]-lastx;
    dy=mpos[1]-lasty;
    
    rot0speed-=SIGNF(rot0speed)*dt*0.5;
    rot1speed-=SIGNF(rot1speed)*dt*0.5;

    if(bdown==1){
	
	rot0speed+=dt*dy;
	rot1speed+=dt*dx;

	if(rot0speed>1)
	    rot0speed=1;
	if(rot0speed<-1)
	    rot0speed=-1;
	if(rot1speed>1)
	    rot1speed=1;
	if(rot1speed<-1)
	    rot1speed=-1;
	
	//rot[0]-=0.2*uiaccel*rot0speed;
	rot[1]-=0.2*uiaccel*rot1speed;
	//if(rot[0]>=80)rot[0]=80;
	//if(rot[0]<=-80)rot[0]=-80;
    }

    if(kd['h']){
	if(animState!=WAVE){
	    prevAnimState=animState;
	    setAnimState(WAVE);
	}
    }

    if(kd['w']||kd['W']){
	speed+=2*dt;
    }
    else if(kd['S']||kd['s']){
	speed-=2*dt;
    }
    else{
	speed-=SIGNF(speed)*(0.5*dt);
    }

    if(kd['a']||kd['A']){
	sspeed+=1*dt;
    }
    else if(kd['d']||kd['D']){
	sspeed-=1*dt;
    }
    else sspeed-=SIGNF(sspeed)*(0.5*dt);

    if(kd['F']){
	uspeed-=1*dt;	
	if(animState!=FLIP){
	    prevAnimState=animState;
	    setAnimState(FLIP);
	}
    }

    if(kd['z']){
	uspeed-=1*dt;	
	if(animState!=CROUCH_WALK){
	    prevAnimState=animState;
	    setAnimState(CROUCH_WALK);
	}
    }
    

    if(kd[' ']){
	uspeed-=1*dt;	
	if(animState!=JUMP){
	    prevAnimState=animState;
	    setAnimState(JUMP);
	}	
    }
    else if(kd['z']||kd['Z']){
	uspeed+=1*dt;
    }
    else uspeed-=SIGNF(uspeed)*(0.5*dt);

    uspeed=0;
    
    if(speed>2.0)speed=2.0;
    if(speed<-2) speed=-2.0;

    if(sspeed>1.0)sspeed=1.0;
    if(sspeed<-1)sspeed=-1.0;

    if(uspeed>1.0)uspeed=1.0;
    if(uspeed<-1)uspeed=-1.0;

    dir.rotatey(rot[1]*M_PI/180.0);
    dir.mult(uiaccel*speed*dt);
    pos=pos+dir;

    dir=Vec3f(-1,0,0);
    dir.rotatey(rot[1]*M_PI/180.0);
    dir.mult(uiaccel*sspeed*dt);
    pos=pos+dir;

    /*
    dir=Vec3f(0,1,0);
    dir.mult(uiaccel*uspeed*dt);
    pos=pos+dir;
    */
    if(fabs(speed)<=0.1){
	//if(animState!=STAND && animState!=JUMP && animState!=WAVE)
	if(animState==RUN)  
	    setAnimState(STAND);
    }
    else{
	if(animState==STAND)
	    setAnimState(RUN);
    }

}

void Mesh::move(float dt)
{
    if(nframes){
	int i=0;
	int next;
	time+=dt;
	float spf=1.0/((float)animstats[animState].fps);
	float elapsed=time-lastUpdated;

	while(elapsed>spf){
	    animFrame=animFrame+1;
	    if(animFrame>animstats[animState].last){
		//animState=(MD2AnimType)(animState+1);
		//if(animState>MAX_ANIMATIONS)animState=(MD2AnimType)0;
		if(prevAnimState!=MAX_ANIMATIONS){
		    setAnimState(prevAnimState);
		    prevAnimState=MAX_ANIMATIONS;
		}
	
		animFrame=animstats[animState].first;
	    }
	    elapsed-=spf;
	    lastUpdated=time;
	}

	if(animState==JUMP){
	    float mid=(animstats[animState].first+
		     animstats[animState].last)/2.0;
	    if(animFrame<mid){
		jumpApex+=dt*(mid-animFrame);
	    }
	    else if(animFrame>=mid+1){
		speed=0;
	    }
	}

	next=animFrame+1;
	if(next>animstats[animState].last){
	    if(prevAnimState!=MAX_ANIMATIONS){
		next=animstats[prevAnimState].first;
	    }
	    else
		next=animstats[animState].first;
	}
	elapsed/=spf;
	double melapsed=(1.0-elapsed);
	for(i=0;i<nvert;i++){
	    vert[i].lerp(melapsed,realVert[animFrame*nvert+i],
			 realVert[next*nvert+i]);
	    //vert[i]=vert[i]+elapsed*realVert[animFrame*nvert+i];
	}
	calculateNormals();
    
	float gravity=1.5;
	jumpApex-=gravity*dt;
	pos[1]+=jumpApex;
	if(pos[1]<0)pos[1]=0;
	
    }
}

void Mesh::test2()
{
  // char fbase[]="/usr/foisy2/submit/c610/data/ps_";
  char fbase[]="../startup/ps_";
  char fname[255];

  int i=0;
  int j=0;
  int ti=0;
  int res=64;
  float hmw;
  float hmh;

  sprintf(fname,"%sheight_0.pgm",fbase);
  Image8 height;
  height.readPPM(fname);
  printf("done height(%s).\n",fname);
  sprintf(fname,"%stexture_1.ppm",fbase);
  Image8 texture;
  texture.readPPM(fname);
  printf("done texture(%s).\n",fname);

  alloc(res*res,res*res*2);
  allocTextures(1);
  setTextureImage(texture,0);

  hmh=height.getHeight()-1;
  hmw=height.getWidth()-1;
  
  /*FIXME
  for(i=0;i<res;i++){
    for(j=0;j<res;j++){
      vert[i*res+j].data[0]=(float)j/((float)res)-0.5;
      vert[i*res+j].data[1]=(height->get((int)(hmw*((float)j)/((float)res)),
					 (int)(hmh*((float)i)/((float)res)),
					 0)/65536.0)/10;
      vert[i*res+j].data[2]=(float)i/((float)res)-0.5;
      vert[i*res+j].mult(20);
      
      tvert[0][i*res+j][0]=(float)j/((float)res);
      tvert[0][i*res+j][1]=(float)i/((float)res);
	
      if(i<(res-1) && j<(res-1)){
	tri[ti][0]=i*(res)+j;
	tri[ti][1]=i*(res)+j+1;
	tri[ti++][2]=(i+1)*(res)+j+1;
	tri[ti][0]=i*res+j;
	tri[ti][1]=(i+1)*res+j+1;
	tri[ti++][2]=(i+1)*res+j;
      }
    }
    }
  printf("%d/%d\n",ti,ntri);
  //calculateNormals();
  printf("done calculating normals.\n");
  */
  /*  createList();
  printf("done creating list\n");

  delete [] vert;
  delete [] tvert[0];
  delete [] tvert;
  delete [] tri;
  nvert=0;
  tvert=0;
  tri=0;
  ntri=0;
  vert=0;
  */
}
