#ifdef WIN32
#include <GL/GLee.h>
#endif
//#define GL_GLEXT_PROTOTYPES
#include "vec.h"
#include "image.h"
#include <stdio.h>
#include <GL/gl.h>

#ifndef MESH_H
#define MESH_H

/*
  struct's and defines for MD2 File were obtained from:
   http://tfc.duke.free.fr/us/tutorials/models/md2.htm

  these include:
  struct MD2AnimType;
  struct T_MD2AnimStats;
  MD2AnimStats animstats;
  struct T_MD2Triangle;
  struct T_MD2TexCoord;
  struct T_MD2Vertex;
  struct T_MD2Frame;
  
  which I am sure the author probably took directly from
  the quake 2 source code.

  I claim no ownership over that code.
*/

typedef struct SURFACE{
  char * name;
  float color[4];
}Surface;

typedef struct EDGE{
  int next;
  int opp;
  int prev;
  int vert;
  int face; //index into polys
}Edge;

typedef int Triangle[3];
typedef int Quad[4];

//http://tfc.duke.free.fr/us/tutorials/models/md2.htm
typedef enum {
    STAND,
    RUN,
    ATTACK,
    PAIN_A,
    PAIN_B,
    PAIN_C,
    JUMP,
    FLIP,
    SALUTE,
    FALLBACK,
    WAVE,
    ANIM_POINT,
    CROUCH_STAND,
    CROUCH_WALK,
    CROUCH_ATTACK,
    CROUCH_PAIN,
    CROUCH_DEATH, 
    DEATH_FALLBACK,
    DEATH_FALLFORWARD,
    DEATH_FALLBACKSLOW,
    BOOM,
    MAX_ANIMATIONS
} MD2AnimType;

class Mesh
{
  // protected:
 public:
    MD2AnimType animState;
    int animFrame;
    MD2AnimType prevAnimState;
    float jumpApex;
    double speed;

    float lastUpdated;
    float time;

    int nframes;
    int flatShading;
    int nvert;
    int ntri;
    Vec3f * vert;
    Vec3f * realVert;
    Vec3f * norm;
	int shinny;
    
    Edge * edges;
    int nedges;
    
    int nsurf;
    Surface * surf;
    
    Triangle * tri;
   
    int ntex;
    Vec2f ** tvert;
    int ntvert;

    GLint list;

    Vec3f pos;
    Vec3f rot;
    Vec3f scale;
    
    int * tri2surf;
    Triangle * tri2tvert;
    Vec3f * tri2norm;
 public:
    GLuint * tex;
  Mesh():pos(0,0,0),rot(0,0,0){
      nframes=0;
      
      flatShading=0;
      ntri=0;
      tri=0;
      nvert=0;
      tex=0;
      vert=0;
      realVert=0;
      ntex=0;
      tvert=0;
      list=-1;
      norm=0;
      nsurf=0;
      surf=0;
      edges=0;
      tri2surf=0;
      tri2norm=0;
      tri2tvert=0;
      scale=Vec3f(1.0,1.0,1.0);

      time=0;
      lastUpdated=0;
      speed=0;
      prevAnimState=MAX_ANIMATIONS;
      setAnimState(STAND);
	  jumpApex=0;
      
	  shinny=0;

	  move(0);

  }
  void setShinny(int s){
	  shinny=s;
  }
  int getNumTri(){
      return ntri;
  }
  Vec3f getRot(){
      return rot;
  }
  Vec3f getPos(){
      return pos;
  }
  void setScale(Vec3f sc){
      scale=sc;
  }
  void setPos(Vec3f p){
      pos=p;
  }
  void setRot(Vec3f r){
      rot=r;
  }
  void test()
    {
      alloc(5,2);
      vert[0]=Vec3f(0.0,0,0);
      vert[1]=Vec3f(1.0,0,0);
      vert[2]=Vec3f(1.0,1.0,0);
      vert[3]=Vec3f(0,1.0,0);
      vert[4]=Vec3f(0.5,1.0,0);
      /*tri[0][0]=0;
      tri[0][1]=1;
      tri[0][2]=2;
      tri[1][0]=0;
      tri[1][1]=4;
      tri[1][2]=3;*/
    }
  virtual void test2();
  void setAllSurf(float * col){
	  for(int i=0;i<nsurf;i++){
		  surf[i].color[0]=col[0];
		  surf[i].color[1]=col[1];
		  surf[i].color[2]=col[2];
		  surf[i].color[3]=col[3];
	  }
  }
  void setSurf(int i,float * col){
	  surf[i].color[0]=col[0];
	  surf[i].color[1]=col[1];
	  surf[i].color[2]=col[2];
	  surf[i].color[3]=col[3];
  }

  virtual void alloc(int _nvert,int _ntri);
  virtual void allocTextures(int _ntex);
  virtual void setTextureImage(Image8 & texim,int which);
  virtual GLint getTexture(int which){
    if(which<ntex){
      return tex[which];
    }
    return 0;
  }
  void drawShadowPolygons(Vec3f lightPos);
  void genTriangle();
  void genSphere();
  void genPlane();
  void initEdges();
  void drawSilEdges(Vec3f & lightPos);
  void drawEdges3(Vec3f & lightPos,float offset=-0.5,float width=0.5);
  void drawEdges2(Vec3f & lightPos,float offset=-0.5,float width=0.5);
  void drawEdges(Vec3f & lightPos,float offset=-0.5,float width=0.5);
  void projectEdges(Vec3f & lightPos,double * model,double * proj,int * view,
		    float * silmap[4]);
  int loadLWO(char * filename);
  int loadMD2(char * filename,char * skinFile);
  virtual void createList();
  virtual void draw(int transform=1);
  virtual void calculateNormals();
  void applyTransform();
  void applyITransform();
  void setFlatShading(int f){flatShading=f;}
  void setAnimState(MD2AnimType s);
  void move(float dt);

  void control(int bdown,
	       unsigned char * kd,
	       float dt,
	       float dx,
	       float dy,
	       int * mpos,
	       float lastx,
	       float lasty,
	       int freeze,
	       float uiaccel);

  virtual ~Mesh(){
    int i=0;

    //Fixme only one texture so far
    if(tvert){
	printf("deleting tvert %d\n",tvert);
	delete [] tvert[0];
	delete [] tvert;
    }

    if(ntex){
	printf("deleting textures %d\n",ntex);
	glDeleteTextures(ntex,tex);
	delete [] tex;
    }
    if(list>=0){
	printf("deleting list\n");
	glDeleteLists(list,1);
    }
    list=-1;
    delete [] tri;
    
    if(surf){
	for(i=0;i<nsurf;i++){
	    delete surf[i].name;
	}
	delete [] surf;
    }
    if(vert){
      delete [] vert;
      delete [] realVert;
    }
    if(norm)
	delete [] norm;
    delete [] tri2norm;
    delete [] tri2surf;
    delete [] tri2tvert;

    delete [] edges;
    
    tri2surf=0;
    tri2norm=0;
    nsurf=0;
    ntri=0;
    norm=0;
    ntex=0;
    vert=0;
    tvert=0;
    
    printf("Destroying mesh\n");
  }  
};

#endif



