#ifdef WIN32
#include <GLee.h>
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glu.h>
#include "ray.h"
#include "image.h"
#include "heap.h"
#include "mc.h"
#include <stdarg.h>
#include <getopt.h>
#include <unistd.h>
#include "globals.h"
#include "shortestpath.h"
#include "matrix.h"
//#include <pthread.h>

#include "clb.h"

#include "chart.h"
#include "simplify.h"

#include "texgen.h"
#include "dynatex.h"

#include "cshep2.h"
#include "tshep2.h"
#include "dsyevx.h"
#include "dgemm.h"
#include "dgesv.h"
#include "dsysv.h"

using namespace CapguiGlobals;

extern "C" void idbvip_(int & MD,int & NCP,int & NDP,float * XD,float * YD,float * ZD,int & NIP,float *XI,float * YI,float * ZI,int * IWK,float * WK);

extern int face[6][3];

int texw[3]={256,256,256};
int texh[3]={256,256,256};
int debug=0;

//Get rid of these
extern Vec3d * campts;
extern _3ints * camtri;
extern int ncamtri;
extern GLuint cubeTex[6];
//End get rid of these


#define NRAYS 64
#define MIN_DIST 1.0/(1000.0*(double)NRAYS)

//#define MERGE_PRINT
#define EPS 10e-8



void drawString(int x,int y,char * fmt,...)
{
  char str[2048];
  va_list ap;
  
  va_start(ap,fmt);
  vsnprintf(str,2048,fmt,ap);
  va_end(ap);

  char * c=str;
  glRasterPos2i(x,y);
  while(*c){
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10,*c);
    c++;
  }
}

unsigned char bad[500000];

int dc(const void * _a,const void * _b)
{
  double * a = (double *)_a;
  double * b = (double *)_b;
  if(fabs(*a)<fabs(*b))return -1;
  if(fabs(*a)>fabs(*b))return 1;
  return 0;
}

class Line{
public:
  double p[2],d[2];
  double t;
  double s;
  double ca,sa;
  double angle;
  int angleset;
  Line(){
    p[0]=p[1]=0;
    d[0]=d[1]=0;
    angleset=0;
  }
  Line(float * _p1,float * _p2){
    p[0]=_p1[0];p[1]=_p1[1];
    d[0]=_p2[0]-_p1[0];
    d[1]=_p2[1]-_p1[1];
    angleset=0;
  }
  Line(double * _p1,double * _p2){
    p[0]=_p1[0];p[1]=_p1[1];
    d[0]=_p2[0]-_p1[0];
    d[1]=_p2[1]-_p1[1];
    angleset=0;;
  }
  int intersect(Line &uv,double * where){
    s=((uv.p[1]-p[1])*d[0]+(p[0]-uv.p[0])*d[1])/
      (uv.d[0]*d[1]-uv.d[1]*d[0]);
        
    t=(uv.p[0]+s*uv.d[0]-p[0])/d[0];
    if(d[1]>d[0]){
      t=(uv.p[1]+s*uv.d[1]-p[1])/d[1];
    }
    if(s>=0 && s<=1.0){
      if(t>=0 && t<=1.0){
	if(t<0)t=0;
	if(t>1)t=1;
	where[0]=p[0]+t*d[0];
	where[1]=p[1]+t*d[1];
	return 1;
      }
    }
    return 0;
  }
    
  int intersect2(Line &uv,double * where){
    s=((uv.p[1]-p[1])*d[0]+(p[0]-uv.p[0])*d[1])/
      (uv.d[0]*d[1]-uv.d[1]*d[0]);

    if(s<0.0 | s>1.0)return 0;

    t=(uv.p[0]+s*uv.d[0]-p[0])/d[0];
    if(fabs(d[1])>fabs(d[0])){
      t=(uv.p[1]+s*uv.d[1]-p[1])/d[1];
    }
    if(t>=0.0 && t<=1.0){
	*where=s;
	
	s-=0.1;
	double pt[2];
	double ptd[2];
	pt[0]=uv.p[0]+s*uv.d[0];
	pt[1]=uv.p[1]+s*uv.d[1];
	
	ptd[0]=pt[0]-p[0];
	ptd[1]=pt[1]-p[1];
	
	if(!angleset){
	    if(fabs(d[0])<10e-10 && 0){
	      //printf("vertical\n");
		if(d[1]>0){
		    angle=M_PI/2.0;
		}
		else
		    angle=1.5*M_PI;
	    }
	    else{
		angle=atan2(d[1],d[0]);
	    }
	    angleset=1;
	    ca=cos(-angle);
	    sa=sin(-angle);
	    
	}
	pt[0]=ca*ptd[0]-sa*ptd[1];
	pt[1]=sa*ptd[0]+ca*ptd[1];
	
	if(pt[1]>0){
	    (*where)*=-1;
	}
	return 1;
    }
    return 0;
  }
};


class SilData{
public:
    int w;
    int h;

    int nlines;
    int maxlines;
    Line * data;

    SilData(int w,int h){
	maxlines=1024;
	nlines=0;
	data = new Line[maxlines];
    }
    ~SilData(){
	delete data;
	maxlines=nlines=0;
    }
    void insert(double * pt1,double * pt2){
	if(nlines>=maxlines){
	    Line * newdata = new Line[maxlines+256];
	    memcpy(newdata,data,sizeof(Line)*nlines);
	    maxlines+=256;
	    delete data;
	    data=newdata;
	}
       
	data[nlines]=Line(pt1,pt2);
	nlines++;
	
    }
    int intersectAll(Line & line,double ** isects){
	int i=0;
	int nintersections=0;
	int which=0;

	*isects = new double[64];
	for(i=0;i<nlines;i++){
	    double where;
	    if(data[i].intersect2(line,&where)){
		(*isects)[which]=where;
		which++;
	    }
	}
	nintersections=which;
	qsort(*isects,nintersections,sizeof(double),dc);
	for(int i=0;i<nintersections-1;i++){
	    if(fabs((*isects)[i+1]-(*isects)[i])<0.00001){
		//printf("found dup\n");
		for(int j=i+1;j<nintersections;j++){
		    (*isects)[j-1]=(*isects)[j];
		}
		nintersections--;
	    }
	}
	return nintersections;
    }
    void removeAll(){
	nlines=0;
    }
};
SilData silData(256,256);

void Ray::remove(double t)
{
  int i=0;
  while(i<nseg){
    if(t>=enter[i] && t<=exit[i]){
      if(fabs(t-enter[i])<fabs(t-exit[i])){
	enter[i]=t+2*EPS;
      }
      else{
	exit[i]=t-2*EPS;
      }
      return;
    }
    i++;
  }
}

void Ray::include(double t)
{
    int i=0;
    while(i<nseg){
	if(t>=enter[i] && t<=exit[i]){
	    return;
	}
	if(t<enter[i] && (i==0)){
	  enter[i]=t;
#ifdef PRINT_DEBUG
	  printf("less than first\n");
#endif
	    return;
	}
	else if(t<enter[i]){
	    if(fabs(t-enter[i])<fabs(t-exit[i-1])){
		enter[i]=t;
	    }
	    else
		exit[i-1]=t;
	    return;
	}
	i++;
    }
    if(i==nseg && nseg){
#ifdef PRINT_DEBUG
      printf("greater than last\n");
#endif
	exit[nseg-1]=t;
    }
}

int Ray::getBoundary(double t1,double t2,int t1on,int t2on,Vec3f & output)
{
    int set=0;
    int i=0;
    if(t1>t2){
	double temp=t1;
	t1=t2;
	t2=temp;
	
	int t=t1on;
	t1on=t2on;
	t2on=t;
    }
    while(i<nseg){
	if(t1<=enter[i] && enter[i]<=t2){
	    output = dir*enter[i]+start;
	    set++;
	}
	if(t1<=exit[i] && exit[i]<=t2){
	    output = dir*exit[i]+start;
	    set++;
	}
	i++;
    }

    if(set==0){
	if(nseg&&0){
#ifdef PRINT_DEBUG
	    printf("could not find [%f,%f,%d,%d] in ",t1,t2,t1on,t2on);
	    print();printf("\n");
#endif
	}
      	/*if(t1on)
	    output=dir*t1+start;
	else if(t2on)
	    output=dir*t2+start;
	    else*/
	    output=dir*0.5*(t1+t2)+start;
	return 0;
    }
    return 1;
}

int Ray::inside(double t)
{
    int i=0;
    while(i<nseg){
	if(enter[i]<=t && t<=(exit[i]+EPS))//FIXME: <=?
	    return 1;
	i++;
    }
    return 0;
}
void Ray::draw(){
    int i=0;
    glBegin(GL_LINES);
    if(flag){
	Vec3f p1=start;
	Vec3f p2=dir+start;
	if(flag){
	    glColor3f(0.0,1.0,0.0);
	}
	else{
	    glColor3f(1.0,1.0,1.0);
	}
	glVertex3fv(p1.getData());
	if(flag){
	    glColor3f(0.0,1.0,0.0);
	}
	else{
	    glColor3f(0.2,0.2,0.1);
	}
	glVertex3fv(p2.getData());
    }
    else{
	while(i<nseg){
	    Vec3f p1=dir*enter[i]+start;
	    Vec3f p2=dir*exit[i]+start;
	    glColor3f(1.0,1.0,1.0);
	    glVertex3fv(p1.getData());
	    glColor3f(0.5,0.5,0.5);
	    glVertex3fv(p2.getData());
	    i++;
	}
    }
    glEnd();
}
void Ray::insert(double start,double stop)
{
    if(nseg>=maxseg){
	double * newenter = new double[maxseg+4];
	double * newexit = new double[maxseg+4];
	memcpy(newenter,enter,nseg*sizeof(double));
	memcpy(newexit,exit,nseg*sizeof(double));
	maxseg=maxseg+4;
	delete enter;
	delete exit;
	enter=newenter;
	exit=newexit;
    }
    if(start>stop){
	double temp=start;
	start=stop;
	stop=temp;
    }
    if(start<0)start=0;
    if(stop>1)stop=1;
    if(start==stop){
#ifdef PRINT_DEBUG
      printf("start == stop == %f\n",start);
      print();printf("\n");
#endif
	flag=1;
	return;
    }
    int i=0;
    for(i=0;i<nseg;i++){
	if(start<enter[i]){
	    break;
	}
    }
    for(int j=nseg;j>i;j--){
	enter[j]=enter[j-1];
	exit[j]=exit[j-1];
    }
    
    enter[i]=start;
    exit[i]=stop;
    nseg++;
}

void Ray::makeConsistent()
{
    for(int i=0;i<nseg;i++){
	if(enter[i]>exit[i]){
	    printf("enter(%f) > exit(%f)\n",enter[i],exit[i]);
	}
	else if(i<(nseg-1)&& exit[i]>enter[i+1]){
	    printf("bad order\n");
	    print();
	}
    }
}
void Ray::print()
{
    if(nseg==0)printf("0");
    for(int i=0;i<nseg;i++){
	printf("[%f %f]",enter[i],exit[i]);
    }
}
void Ray::merge(Ray & o){
    int i1=0;
    int i2=0;

    int beforecount=0;
    for(double d=0;d<=1.0;d+=0.001){
	if(inside(d) && o.inside(d)){
	    beforecount++;
	}
    }

#ifdef MERGE_PRINT
    int prnt=0;
    if(nseg+o.nseg>2){
    printf("merging : ");
    print();
    printf(" with :");
    o.print();
    printf("\n");
    prnt=1;
    }
#endif
    while(i1<nseg && i2<o.nseg){
	while(exit[i1]<o.enter[i2]){
	    for(int j=i1;j<nseg;j++){
		enter[j]=enter[j+1];
		exit[j]=exit[j+1];
	    }
	    nseg--;
#ifdef MERGE_PRINT
	    if(prnt)printf("deleting (%d/%d)\n",i1,nseg);
#endif
	    if(i1>=nseg)break;
	}
	if(i1>=nseg)break;
	if(enter[i1]<o.enter[i2]){
	    enter[i1]=o.enter[i2];
	}
	else if(enter[i1]>o.exit[i2]){
#ifdef MERGE_PRINT
	    if(prnt)printf("no intersection %d,%d\n",i1,i2);
#endif
	    i2++;
	    continue;
	}
	if(exit[i1]>o.exit[i2]){
	    if(i2<(o.nseg-1) && exit[i1]>o.enter[i2+1]){
		//need more storage
#ifdef MERGE_PRINT
		if(prnt)printf("getting extra storage\n");
#endif
		if(nseg>=maxseg){
		    double * newenter = new double[maxseg+4];
		    double * newexit = new double[maxseg+4];
		    memcpy(newenter,enter,nseg*sizeof(double));
		    memcpy(newexit,exit,nseg*sizeof(double));
		    maxseg=maxseg+4;
		    delete enter;
		    delete exit;
		    enter=newenter;
		    exit=newexit;
		}
		for(int j=nseg;j>i1;j--){
		    enter[j]=enter[j-1];
		    exit[j]=exit[j-1];
		}
		enter[i1+1]=o.exit[i2];
		exit[i1+1]=exit[i1];
		nseg++;
	    }
	    exit[i1]=o.exit[i2];
	    i2++;
	    i1++;
	}
	else
	    i1++;
    }
    nseg=i1;
#ifdef MERGE_PRINT
    if(prnt){
    printf("Result:");
    print();
    printf("\n");
    }
#endif

    int aftercount=0;
    for(double d=0;d<=1.0;d+=0.001){
	if(inside(d)){
	    aftercount++;
	}
    }
    //printf("before %d %d\n",beforecount,aftercount);
    if(beforecount!=aftercount){
	printf("intersection not good!\n");
	::exit(0);
    }

}

void RaySet::draw(){
    for(int i=0;i<w*h;i++){
	rays[i].draw();
    }
}

void printMatrix(double * mat,char * str=0){
    if(str)printf("%s:\n",str);
    for(int i=0;i<4;i++){
	for(int j=0;j<4;j++){
	    printf("%f ",mat[i+j*4]);
	}
	printf("\n");
    }
}

int MI::getBoundaryPoint(Vec3f & pt1,Vec3f & pt2,int p1on,int p2on,Vec3f & output)
{
    if(pt1[2]!=pt2[2]){
	int j=(int)(pt1[0]);
	int i=(int)(pt1[1]);
	return raysets[0].rays[i*(raysets[0].w)+j].getBoundary(pt1[2]/(d-1),pt2[2]/(d-1),p1on,p2on,output);
    }
    else if(pt1[1]!=pt2[1]){
	int j=(int)(pt1[0]);
	int i=(int)(pt1[2]);
	return raysets[1].rays[i*(raysets[1].w)+j].getBoundary(pt1[1]/(h-1),pt2[1]/(h-1),p1on,p2on,output);
    }
    else if(pt1[0]!=pt2[0]){
	int j=(int)(pt1[1]);
	int i=(int)(pt1[2]);
	return raysets[2].rays[i*(raysets[2].w)+j].getBoundary(pt1[0]/(w-1),pt2[0]/(w-1),p1on,p2on,output);
    }
}

int MI::getBoundaryPoint2(Vec3f & pt1,Vec3f & pt2,int p1on,int p2on,Vec3f & output)
{
    if(pt1[2]!=pt2[2]){
	int j=(int)(pt1[0]);
	int i=(int)(pt1[1]);
	int upsucc=0,dnsucc=0,rsucc=0,lsucc=0;
	Vec3f up,dn,l,r;
	if(i<h-1){
	    upsucc=raysets[0].rays[(i+1)*(raysets[0].w)+j].getBoundary(pt1[2]/(d-1)-1.0/(d-1),
								       pt2[2]/(d-1)+1.0/(d-1),
								       p1on,p2on,up);
	}
	if(i>0){
	    dnsucc=raysets[0].rays[(i-1)*(raysets[0].w)+j].getBoundary(pt1[2]/(d-1)-1.0/(d-1),
								       pt2[2]/(d-1)+1.0/(d-1),
								       p1on,p2on,dn);
	}

	if(j<w-1){
	    rsucc=raysets[0].rays[(i)*(raysets[0].w)+j+1].getBoundary(pt1[2]/(d-1),pt2[2]/(d-1),p1on,p2on,r);
	}
	if(j>0){
	    lsucc=raysets[0].rays[(i)*(raysets[0].w)+j-1].getBoundary(pt1[2]/(d-1),pt2[2]/(d-1),p1on,p2on,l);
	}

	if(dnsucc&&upsucc){
#ifdef PRINT_DEBUG
	    printf("its okay\n");
#endif
	    output=dn*0.5;
	    output=up*0.5+output;
	    return 1;
	}
	else if(lsucc&&rsucc){
#ifdef PRINT_DEBUG
	    printf("its okay\n");
#endif
	    output=l*0.5;
	    output=r*0.5+output;
	    return 1;
	}
	return 0;
    }
    else if(pt1[1]!=pt2[1]){
	int j=(int)(pt1[0]);
	int i=(int)(pt1[2]);

	int upsucc=0,dnsucc=0,rsucc=0,lsucc=0;
	Vec3f up,dn,l,r;
	if(i<d-1){
	    upsucc=raysets[1].rays[(i+1)*(raysets[1].w)+j].getBoundary(pt1[1]/(h-1),pt2[1]/(h-1),p1on,p2on,up);
	}
	if(i>0){
	    dnsucc=raysets[1].rays[(i-1)*(raysets[1].w)+j].getBoundary(pt1[1]/(h-1),pt2[1]/(h-1),p1on,p2on,dn);
	}
	if(j<w-1){
	    rsucc=raysets[1].rays[(i)*(raysets[1].w)+j+1].getBoundary(pt1[1]/(h-1),pt2[1]/(h-1),p1on,p2on,r);
	}
	if(j>0){
	    lsucc=raysets[1].rays[(i)*(raysets[1].w)+j-1].getBoundary(pt1[1]/(h-1),pt2[1]/(h-1),p1on,p2on,l);
	}

	if(dnsucc&&upsucc){
#ifdef PRINT_DEBUG
	    printf("its okay\n");
#endif
	    output=dn*0.5;
	    output=up*0.5+output;
	    return 1;
	}
	else if(lsucc&&rsucc){
#ifdef PRINT_DEBUG
	    printf("its okay\n");
#endif
	    output=l*0.5;
	    output=r*0.5+output;
	    return 1;
	}
	return 0;
    }
    else if(pt1[0]!=pt2[0]){
	int j=(int)(pt1[1]);
	int i=(int)(pt1[2]);

	int upsucc=0,dnsucc=0,rsucc=0,lsucc=0;
	Vec3f up,dn,l,r;

	if(i<d-1){
	    upsucc=raysets[2].rays[(i+1)*(raysets[2].w)+j].getBoundary(pt1[0]/(w-1),pt2[0]/(w-1),p1on,p2on,up);
	}
	if(i>0){
	    dnsucc=raysets[2].rays[(i-1)*(raysets[2].w)+j].getBoundary(pt1[0]/(w-1),pt2[0]/(w-1),p1on,p2on,dn);
	}

	if(j<h-1){
	    rsucc=raysets[2].rays[(i)*(raysets[2].w)+j+1].getBoundary(pt1[0]/(w-1),pt2[0]/(w-1),p1on,p2on,r);
	}
	if(j>0){
	    lsucc=raysets[2].rays[(i)*(raysets[2].w)+j-1].getBoundary(pt1[0]/(w-1),pt2[0]/(w-1),p1on,p2on,l);
	}

	if(dnsucc&&upsucc){
#ifdef PRINT_DEBUG
	    printf("its okay\n");
#endif
	    output=dn*0.5;
	    output=up*0.5+output;
	    return 1;
	}
	else if(lsucc&&rsucc){
#ifdef PRINT_DEBUG
	    printf("its okay\n");
#endif
	    output=l*0.5;
	    output=r*0.5+output;
	    return 1;
	}
	return 0;
    }
}

void MI::draw(int d1,int d2,int d3){
    glPushMatrix();
    //glScalef(scale[0],scale[1],scale[2]);
    //glTranslatef(trans[0],trans[1],trans[2]);
    glTranslatef(-0.5,-0.5,-0.5);
    if(d1)
	raysets[0].draw();
    if(d2)
	raysets[1].draw();
    if(d3)
	raysets[2].draw();

    glPopMatrix();
}

void MI::draw(Image8 &image,double * intrinsic,double * projection,double * extrinsic,
	      int d1,int d2,int d3){

  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  //glOrtho(0,256,0,256,-100,100);
  double mat[16]={2.0/320.0,0,0,0,  0,2.0/240.0,0.0,0.0, 0.0,0.0,0.0,0.0, -1.0,-1.0,0.0,1.0}; 
  glMultMatrixd(mat);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  
  glMultMatrixd(intrinsic);
  glMultMatrixd(projection);
  glMultMatrixd(extrinsic);
  glTranslatef(trans[0],trans[1],trans[2]);
  glScalef(scale[0],scale[1],scale[2]);
  
  
  printErr("pts\n");
  
  if(d1)
    raysets[0].draw();
  if(d2)
    raysets[1].draw();
  if(d3)
    raysets[2].draw();

  glPopMatrix();
  
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  
}

int RaySet::inside(int j,int i,double d){
    return rays[i*w+j].inside(d);
/*
    if(plane==0){
	return rays[y*w+x].inside((double)z/(double)(d-1)); 
    }
    else if(plane==1){
	return rays[z*w+x].inside((double)y/(double)(h-1)); 
    }
    else if(plane==2){
	return rays[z*w+y].inside((double)x/(double)(w-1)); 
	}*/
}

void RaySet::shave(Image8 & image){
    double temp[16];
    double mat[16];
    glGetDoublev(GL_MODELVIEW_MATRIX,mat);
#ifdef PRINT_DEBUG
    for(int j=0;j<4;j++){
	for(int i=0;i<4;i++){
	    printf("%f ",mat[i+4*j]);
	}
	printf("\n");
    }
    printf("there are %d lines\n",silData.nlines);
#endif
    for(int i=0;i<w*h;i++){
	if(rays[i].nseg==0)
	    continue;
	Ray * ray = rays+i;
	Vec3f p1=ray->start;
	Vec3f p2=ray->dir+ray->start;
	Ray other;

	/*
	temp[0]=p1[0];
	temp[1]=p1[1];
	temp[2]=p1[2];
	temp[3]=1;
	
	temp[4]=p2[0];
	temp[5]=p2[1];
	temp[6]=p2[2];
	temp[7]=1;
	
	glPushMatrix();
	glMultMatrixd(temp);
	glFlush();
	glGetDoublev(GL_MODELVIEW_MATRIX,temp);
	glPopMatrix();
	*/
	temp[8] =p1[0];
	temp[9] =p1[1];
	temp[10]=p1[2];
	temp[11]=1;
	
	temp[12]=p2[0];
	temp[13]=p2[1];
	temp[14]=p2[2];
	temp[15]=1;
	
	multPoint(temp,mat,temp+8);
	multPoint(temp+4,mat,temp+12);

	Vec3f proj1(temp[0]/temp[2],temp[1]/temp[2],1.0);
	Vec3f proj2(temp[4]/temp[6],temp[5]/temp[6],1.0);
	//printf("%f %f\n",proj1.x,proj1.y);

	double dx=fabs(proj2[0]-proj1[0]);
	double dy=fabs(proj2[1]-proj1[1]);

	Line line(proj1.getData(),proj2.getData());
	
	double start,stop;
	double * intersections;
	int nintersects=silData.intersectAll(line,&intersections);

	double x1,y1,x2,y2;
	int gotstart=0;
	int gotstop=0;
	int usey=fabs(line.d[1])>fabs(line.d[0]);
	int stopped=0;
	start=0;
	other.nseg=0;

	for(int n=0;n<nintersects;n++){
	  double x1,y1,x2,y2;
	  if(intersections[n]>0){
	    //set this to the point
	    double pt[2];
	    pt[0]=line.p[0]+fabs(intersections[n])*line.d[0];
	    pt[1]=line.p[1]+fabs(intersections[n])*line.d[1];
	    gotstart=1;
	    if(usey)
		start=(temp[1]-pt[1]*temp[2])/
		    ((temp[6]-temp[2])*pt[1]-(temp[5]-temp[1]));
	    else
		start=(temp[0]-pt[0]*temp[2])/
		    ((temp[6]-temp[2])*pt[0]-(temp[4]-temp[0]));
	  }
	  else{
	    double pt[2];
            pt[0]=line.p[0]+fabs(intersections[n])*line.d[0]; 
            pt[1]=line.p[1]+fabs(intersections[n])*line.d[1];
	   
	    if(usey)
		stop=(temp[1]-pt[1]*temp[2])/
		    ((temp[6]-temp[2])*pt[1]-(temp[5]-temp[1]));
	    else
		stop=(temp[0]-pt[0]*temp[2])/
		    ((temp[6]-temp[2])*pt[0]-(temp[4]-temp[0]));
	    gotstart=0;
	    stopped++;
	    other.insert(start,stop);
	  }
	}
	if(gotstart){
	    stop=1;
	    other.insert(start,stop);
	}
	else if(nintersects==0){
	    double pt[2];
	    int in=0;
	    int tot=0;
	    if((int)proj1[0]>=0 && (int)proj1[1]>=0 &&
	       (int)proj1[0]<image.getWidth() && 
	       (int)proj1[1]<image.getHeight()){
	      in+=image((int)proj1[0],(int)proj1[1],0)>0;
	      tot++;
	    }
	    if((int)proj2[0]>=0 && (int)proj2[1]>=0 &&
	       (int)proj2[0]<image.getWidth() && 
	       (int)proj2[1]<image.getHeight()){
	      in+=image((int)proj2[0],(int)proj2[1],0)>0;
	      tot++;
	    }
	    pt[0]=line.p[0]+0.5*line.d[0]; 
	    pt[1]=line.p[1]+0.5*line.d[1]; 
	    if((int)pt[0]>=0 && (int)pt[1]>=0 &&
	       (int)pt[0]<image.getWidth() && 
	       (int)pt[1]<image.getHeight()){
		in+=(image((int)pt[0],(int)pt[1],0)>0);
	    }
	    if(tot==0){
	    }
	    else if(in==tot||in>=2){
		other.insert(0,1);
	    }
	}
	delete intersections;
/*
	glPushMatrix();
	glLoadIdentity();
	glBegin(GL_LINES);
	glVertex2fv(proj1.data);
	glVertex2fv(proj2.data);
	glEnd();
	glPopMatrix();
*/
/*
	if(dy>dx){
	    //FIXME: always assumes that the full object is in each
	    //image
	    double x=proj1[0];
	    double y=proj1[1];
	    double ex=proj2[0];
	    double ey=proj2[1];

	    if(proj1[1]>proj2[1]){
		y=proj2[1];
		ey=proj1[1];
		x=proj2[0];
		ex=proj1[0];
	    }
	    
	    int inside=0;
	    int inrange=0;
	    
	    double start=0;
	    double stop=0;
	    double ty;

	    dx=(ex-x)/(ey-y);
	    
	    double extra = y-((int)y);
	    y-=extra;
	    x-=extra*dx;

	    inside=image((int)x,(int)y,0)>0;
	    
	    while(y<ey){
		int ix,iy;
		ix=(int)x;
		iy=(int)y;
		inrange=0;

		if(ix>=0 && ix<image.getWidth()
		   && iy>=0 && iy<image.getHeight()){
		  double pt[2];
		  inrange=1;
		  if(silData.intersect(line,ix,iy,pt)){
		    if(ix==(int)pt[0] && iy==(int)pt[1]){
		      ty=(temp[1]-pt[1]*temp[2])/
			((temp[6]-temp[2])*pt[1]-(temp[5]-temp[1]));
		      if(!inside){
			inside=1;
			start=ty;
		      }
		      else{
			inside=0;
			stop=ty;
			Vec3f diff=proj2;
			diff.sub(proj1);
			Vec3f point=diff*((start+stop)/2.0)+proj1;
			if(!image((int)point[0],(int)point[1],0)){
			  printf("error\n");
			}
			printf("%f %f\n",start,stop);
			other.insert(start,stop);
		      }
		    }
		  }
		}
		x+=dx;
		y+=1;
	    }
	    if(inrange && inside){
	      ty=(temp[1]-y*temp[2])/
			((temp[6]-temp[2])*y-(temp[5]-temp[1]));
	      other.insert(start,ty);
	    }
	}
	else if(dy<=dx){
	    //FIXME: always assumes that the full object is in each
	    //image
	    double x=proj1[0];
	    double y=proj1[1];
	    double ex=proj2[0];
	    double ey=proj2[1];

	    if(proj1[0]>proj2[0]){
		x=proj2[0];
		y=proj2[1];
		ex=proj1[0];
		ey=proj1[1];
	    }

	    int inside=0;
	    int inrange=0;
	    double start=0;
	    double stop=0;
	    double tx;
	    
	    dy=(ey-y)/(ex-x);

	    double extra = x-((int)x);
	    x-=extra;
	    y-=extra*dy;

	    inside=image((int)x,(int)y,0);

	    while(x<ex){
		int ix,iy;
		ix=(int)x;
		iy=(int)y;
		inrange=0;
		if(ix>=0 && ix<image.getWidth()
		   && iy>=0 && iy<image.getHeight()){
		  double pt[2];
		  inrange=1;
		  
		  if(silData.intersect(line,ix,iy,pt)){
		    if(ix==(int)pt[0] && iy==(int)pt[1]){
		      tx=(temp[0]-pt[0]*temp[2])/
			((temp[6]-temp[2])*pt[0]-(temp[4]-temp[0]));
		      
		      if(!inside){
			inside=1;
			start=tx;
		      }
		      else{
			inside=0;
			stop=tx;

			Vec3f diff=proj2;
			diff.sub(proj1);
			Vec3f point=diff*((start+stop)/2.0)+proj1;
			if(!image((int)point[0],(int)point[1],0)){
			  printf("error\n");
			}

			other.insert(start,stop);
		      }
		    }
		  }
		}
		x+=1;
		y+=dy;
	    }
	    if(inrange && inside){
	      tx=(temp[0]-x*temp[2])/
			((temp[6]-temp[2])*x-(temp[4]-temp[0]));

	      

		other.insert(start,tx);
	    }
	    }*/
	//other.makeConsistent();
	ray->merge(other);
	/*ray->nseg=other.nseg;
	if(ray->nseg){
	    ray->enter[0]=other.enter[0];
	    ray->exit[0]=other.exit[0];
	    }*/
    }
}
int gawd;
int MI::inside(int x,int y,int z)
{
    int in0=raysets[0].inside(x,y,z/((double)(d-1)));
    int in1=raysets[1].inside(x,z,y/((double)(h-1)));
    int in2=raysets[2].inside(y,z,x/((double)(w-1)));
    int in=in0+in1+in2;

    if(in!=0 && in!=3){
      /*
      printf("inconsistent check %f %f %f %d\n",
	     x/(double)(NRAYS-1),y/(double)(NRAYS-1),z/(double)(NRAYS-1),in);
      printf("\t\t(%f-%f),(%f-%f),(%f-%f)",
	     (double)z/(double)(NRAYS-1),
	     (double)(z+1)/(double)(NRAYS-1),
	     (double)y/(double)(NRAYS-1),
	     (double)(y+1)/(double)(NRAYS-1),
	     (double)x/(double)(NRAYS-1),
	     (double)(x+1)/(double)(NRAYS-1));
      printf("\t\t(%f)",(double)1/(double)(NRAYS-1));
      printf("\t\t %d %d %d\n",
	     in0,in1,in2);
      if(in==1){
	raysets[0].rays[y*NRAYS+x].remove((double)z/(double)(NRAYS-1));
	raysets[1].rays[z*NRAYS+x].remove((double)y/(double)(NRAYS-1));
	raysets[2].rays[z*NRAYS+y].remove((double)x/(double)(NRAYS-1));
      }
      else if(in==2){
	raysets[0].rays[y*NRAYS+x].include((double)z/(double)(NRAYS-1));
	raysets[1].rays[z*NRAYS+x].include((double)y/(double)(NRAYS-1));
	raysets[2].rays[z*NRAYS+y].include((double)x/(double)(NRAYS-1));
      }
     
      raysets[0].rays[y*w+x].flag=1;
      raysets[1].rays[z*w+x].flag=1;
      raysets[2].rays[z*h+y].flag=1;

      raysets[0].rays[y*w+x].print();printf("\n");
      raysets[1].rays[z*w+x].print();printf("\n");
      raysets[2].rays[z*h+y].print();printf("\n");
     
      printf("\t\t %d %d %d\n",raysets[0].inside(x,y,z),
	     raysets[1].inside(x,y,z),
	     raysets[2].inside(x,y,z));
      
      printf("\n");
      printf("-----\n");*/
    }
    else{
	
    }
    gawd=in;
    return in>1;
}

int current=0;


#ifdef RAY_MAIN
void shaveIt();
void keys(unsigned char c,int x,int y);
Image8 * images;
int curw;//current image w
int curh;//current image h
GLuint tex;
MI * mi;

int wireframe=0;
int bdown=0;
int mpos[2];
int drawnormals=0;
int draw=0;
int draw1=1;
int draw2=1;
int draw3=1;
int drawtex=0;
int drawMarchingCubes=0;

GLuint win;

Chart * charts=0;
int ncharts=0;

int ci=0;
Vec3f cpos=Vec3f(0,0,-5);


unsigned char * texImages=0;
unsigned char * occluded=0;
double extrinsic[16];
double intrinsic[16];
double projection[16];
int nimages;
char outfile[1024]="geom.obj";
char imfile[1024];
char extfile[1024];
double * extrinsics;
double * intrinsics;

void motion(int x,int y){
    float dx=mpos[0]-x;
    float dy=mpos[1]-y;
    if(bdown==1){
	double matrix[16];
	glGetDoublev(GL_MODELVIEW_MATRIX,matrix);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(dx*0.1,0.0,1.0,0.0);
	glRotatef(dy*0.1,1.0,0.0,0.0);
	glMultMatrixd(matrix);
    }
    if(bdown==2){
	cpos[0]+=-dx*0.001;
	cpos[1]+=+dy*0.001;
    }
    if(bdown==3){
	cpos[2]+=dy*0.1;
    }
    mpos[0]=x;
    mpos[1]=y;
}

void mouse(GLint button,GLint state,int x,int y){
    
    if(state==GLUT_DOWN){
	bdown=button+1;
    }
    else bdown=0;
    
    mpos[0]=x;
    mpos[1]=y;
}

typedef float linept[2];

int isSilEdge(Image8 & image,int j,int i)
{
    int w=image.getWidth();
    int h=image.getHeight();
    int issil=0;
    int on=0;
    if(j>=w || i>=h || j<=0 || i<=0){
	return 0;
    }
    on+=image(j,i)>0;
    on+=image(j-1,i)>0;
    on+=image(j-1,i-1)>0;
    on+=image(j,i-1)>0;
    return (on>0 && on<4);
}

void follow(Image8 & image,short * vis,int x,int y,int nshapes,
	    linept ** line,int * nline,int * maxnline)
{
  int start=((*nline)==0);
  int dirs=0;
  int w=image.getWidth()+1;
  int h=image.getHeight()+1;
  vis[y*w+x]=nshapes*40+1;
  
  if((*nline)>=(*maxnline)){
    linept * lineBack=*line;
    *line = new linept[*maxnline+32];
    memcpy(*line,lineBack,sizeof(linept)*(*maxnline));
    (*maxnline)=(*maxnline)+32;
    delete *lineBack;
  }
  (*line)[*nline][0]=x-0.5;//subtract half a pixel to place in image coords
  (*line)[*nline][1]=y-0.5;
  (*nline)+=1;
  
  if((image(x-1,y)>0||
      image(x-1,y-1)>0)
     && isSilEdge(image,x-1,y) && vis[(y)*w+x-1]==-1){
    follow(image,vis,x-1,y,nshapes,line,nline,maxnline);
    if(start){
      start=0;
      for(int i=0;i<(*nline)/2;i++){
	double a0=(*line)[i][0];
	double a1=(*line)[i][1];
	(*line)[i][0]=(*line)[(*nline)-i-1][0];
	(*line)[i][1]=(*line)[(*nline)-i-1][1];
	(*line)[(*nline)-i-1][0]=a0;	
	(*line)[(*nline)-i-1][1]=a1;
      }
    }
    else
      return;
  }
  if((image(x,y-1)>0||
      image(x-1,y-1)>0)
     && isSilEdge(image,x,y-1) && vis[(y-1)*w+x]==-1){
    follow(image,vis,x,y-1,nshapes,line,nline,maxnline);
    if(start){
      start=0;
      for(int i=0;i<(*nline)/2;i++){
	double a0=(*line)[i][0];
	double a1=(*line)[i][1];
	(*line)[i][0]=(*line)[(*nline)-i-1][0];
	(*line)[i][1]=(*line)[(*nline)-i-1][1];
	(*line)[(*nline)-i-1][0]=a0;	
	(*line)[(*nline)-i-1][1]=a1;
      }
    }
    else return;
  }
  if((image(x,y)>0||
      image(x,y-1)>0)
     && isSilEdge(image,x+1,y) && vis[(y)*w+x+1]==-1){
    follow(image,vis,x+1,y,nshapes,line,nline,maxnline);
    if(start){
      start=0;
      for(int i=0;i<(*nline)/2;i++){
	double a0=(*line)[i][0];
	double a1=(*line)[i][1];
	(*line)[i][0]=(*line)[(*nline)-i-1][0];
	(*line)[i][1]=(*line)[(*nline)-i-1][1];
	(*line)[(*nline)-i-1][0]=a0;	
	(*line)[(*nline)-i-1][1]=a1;
      }
    }
    else return;
  }
  if((image(x,y)>0||
      image(x-1,y)>0)
     && isSilEdge(image,x,y+1) && vis[(y+1)*w+x]==-1){
    follow(image,vis,x,y+1,nshapes,line,nline,maxnline);
    if(start){
      start=0;
      for(int i=0;i<(*nline)/2;i++){
	double a0=(*line)[i][0];
	double a1=(*line)[i][1];
	(*line)[i][0]=(*line)[(*nline)-i-1][0];
	(*line)[i][1]=(*line)[(*nline)-i-1][1];
	(*line)[(*nline)-i-1][0]=a0;	
	(*line)[(*nline)-i-1][1]=a1;
      }
    }
    else return;
  }
}

double getError(char * toRemove,linept ** shapes,int * shapeSizes,int i,int j)
{
    
    int prev=0;
    int next=0;
    for(int jj=1;jj<shapeSizes[i];jj++){
	prev=(j-jj+shapeSizes[i])%shapeSizes[i];//Assumes it is a loop...
	if(!toRemove[prev])break;
    }
    for(int jj=1;jj<shapeSizes[i];jj++){
	next=(j+jj)%shapeSizes[i];//Assumes it is a loop...
	if(!toRemove[next])break;
    }
 
    Vec3f before(shapes[i][prev][0],
		shapes[i][prev][1],
		0);
    Vec3f after(shapes[i][next][0],
	       shapes[i][next][1],
	       0);
    Vec3f pt(shapes[i][j][0],
	    shapes[i][j][1],
	    0);
    
    after.sub(before);
    after.normalize();
    pt.sub(before);
    double ptlen=pt.normalize();
    double dp=pt.dot(after);
    after.mult(dp);
    after.add(before);
    pt=Vec3f(shapes[i][j][0],
	    shapes[i][j][1],
	    0);	    
    pt.sub(after);
    return pt.normalize();
}

char ** refineShape(Image8 & image,linept ** shapes,
		   int * shapeSizes,int nshapes)
{
    char ** toRemove = new char *[nshapes] ;
    
    for(int i=0;i<nshapes;i++){
      int closed=(fabs(shapes[i][0][0]-shapes[i][shapeSizes[i]-1][0])+
		  fabs(shapes[i][0][1]-shapes[i][shapeSizes[i]-1][1])<=1);
	Heap heap(shapeSizes[i]);
	toRemove[i] = new char[shapeSizes[i]];
	memset(toRemove[i],0,shapeSizes[i]);
	
	for(int j=1;j<shapeSizes[i]-1;j++){
	    //printf("Get error %d/%d %d/%d\n",i,nshapes,j,shapeSizes[i]);
	    heap.insert(j,getError(toRemove[i],shapes,shapeSizes,i,j));
	}
	//Detect if shape is loop and add end points
	if(closed){
	  heap.insert(0,getError(toRemove[i],shapes,shapeSizes,i,0));
	  heap.insert(shapeSizes[i]-1,getError(toRemove[i],shapes,shapeSizes,i,shapeSizes[i]-1));
#ifdef PRINT_DEBUG
	  printf("shape %d is closed (%d poitns)\n",i,shapeSizes[i]);
#endif
	}
	else{
#ifdef PRINT_DEBUG
	  printf("shape %d is not closed\n",i);
#endif
	}
	double f=0;
	int nremove=7*shapeSizes[i]/8;
	int removed=0;
	while(removed<nremove && f<3){
	    int which=heap.remove(&f);
	    if(which==-1){
#ifdef PRINT_DEBUG
		printf("no more left to remove\n");
#endif
		break;
	    }
	    toRemove[i][which]=1;

	    for(int jj=1;jj<shapeSizes[i];jj++){
		int ind=(jj-which)%shapeSizes[i];
		if(ind<0)ind+=shapeSizes[i];
		//don't need to update end points if its not closed
		if(!closed && (ind==0||ind==shapeSizes[i]-1))continue;
		if(!toRemove[i][ind]){
		    heap.update(ind,
				getError(toRemove[i],shapes,shapeSizes,i,ind));
		    break;
		}
	    }
	    for(int jj=1;jj<shapeSizes[i];jj++){
		int ind=(jj+which)%shapeSizes[i];
		//don't need to updated end points if its not closed
		if(!closed && (ind==0||ind==shapeSizes[i]-1))continue;
		if(!toRemove[i][ind]){
		    heap.update(ind,
				getError(toRemove[i],shapes,shapeSizes,i,ind));
		    break;
		}
	    }
	    //printf("%f(%d)\n",f,removed);
	    removed++;
	}
    }
    return toRemove;
}



int findSilhouette(Image8 & image,linept *** _shapes=0,
		   int ** _shapeSizes=0)
{
    static int called=0;
    short * vis = new short[(image.getWidth()+1)*(image.getHeight()+1)];
    int w=image.getWidth()+1;
    int h=image.getHeight()+1;
    
    linept ** shapes;
    int * shapeSizes;
    int maxShapes=4;
    int nshapes=0;

    shapes = new linept *[maxShapes];
    shapeSizes = new int[maxShapes];
    image=image>0;
    memset(vis,-1,sizeof(short)*w*h);
    for(int i=0;i<h;i++){
	for(int j=0;j<w;j++){
	    if(vis[i*w+j]==-1 && isSilEdge(image,j,i)){
#ifdef PRINT_DEBUG
		printf("%d %d\n",j,i);
#endif
		if(nshapes>=maxShapes){
		    linept ** newShapes = new linept *[maxShapes+8];
		    int * newSizes = new int[maxShapes+8];
		    
		    memcpy(newShapes,shapes,sizeof(linept *)*(maxShapes));
		    memcpy(newSizes,shapeSizes,sizeof(int)*(maxShapes));
		    maxShapes=maxShapes+8;
		    
		    delete shapes;
		    delete shapeSizes;
		    
		    shapes=newShapes;
		    shapeSizes=newSizes;
		}
		int maxnline=0;
		shapeSizes[nshapes]=0;
		shapes[nshapes]=0;

		follow(image,vis,j,i,nshapes,
		       shapes+nshapes,shapeSizes+nshapes,&maxnline);
		if(shapeSizes[nshapes]<6){
		  //delete shapes[nshapes];
#ifdef PRINT_DEBUG
		  printf("Shape of size %d found (%d)\n",shapeSizes[nshapes]
			 ,nshapes);
#endif
		}
		for(int k=1;k<shapeSizes[nshapes];k++){
		  double dist=fabs(shapes[nshapes][k][0]-shapes[nshapes][k-1][0])
		    +fabs(shapes[nshapes][k][1]-shapes[nshapes][k-1][1]);
		  if(dist>1.1){
		    printf("not continuous line\n");
		    for(int i=0;i<shapeSizes[nshapes];i++){
		      printf("%f %f\n",shapes[nshapes][i][0],
			     shapes[nshapes][i][1]);
		    }
		    exit(0);
		  }
		}
		//else{
#ifdef PRINT_DEBUG
		    printf("%d/%d\n",shapeSizes[nshapes],maxnline);
#endif
		    nshapes++;
		    //}
	    }
	}
    }
#ifdef PRINT_DEBUG
    printf("found %d shapes\n",nshapes);
    //char fname[256];
    //sprintf(fname,"sil-%d.pgm",called);
    //Image::writePPM((unsigned char *)image->getDataPtr(),
    //	    w,h,fname,1);
    printf("%d shapes \n",nshapes);
#endif
    Image8 im2(w,h,1);
    for(int i=0;i<h;i++){
	for(int j=0;j<w;j++){
	    im2(j,i)=vis[i*w+j];
	}
    }
    
    //im2.writeTGA("sil.tga");
    called++;

    delete vis;

    if(_shapes){
	*_shapes=shapes;
	*_shapeSizes=shapeSizes;
    }
    else{
	for(int i=0;i<nshapes;i++){
	    delete shapes[i];
	}
	delete shapes;
	delete shapeSizes;
    }
    return nshapes;
}



void MI::shave(Image8 & image,
	       double *intrinsic,
	       double * projection,
	       double * extrinsic)
{
  linept ** shapes;
  int * shapeSizes;
  int nshapes=findSilhouette(image,&shapes,&shapeSizes);
  for(int i=0;i<nshapes;i++){
    int j=0;
    int votescorrect=0;
    int votesincorrect=0;
    for(j=0;j<shapeSizes[i]-1;j++){
      double p1[2];
      double p2[2];
      
      p1[0]=shapes[i][j+1][0]-shapes[i][j][0];
      p1[1]=shapes[i][j+1][1]-shapes[i][j][1];
      
      p2[0]=-p1[1];
      p2[1]= p1[0];
      
      double pleft[2];
      double pright[2];
      
      pleft[0]=p2[0]+((double)shapes[i][j][0]+shapes[i][j+1][0])/2.0;
      pleft[1]=p2[1]+((double)shapes[i][j][1]+shapes[i][j+1][1])/2.0;
      pright[0]=-p2[0]+((double)shapes[i][j][0]+shapes[i][j+1][0])/2.0;
      pright[1]=-p2[1]+((double)shapes[i][j][1]+shapes[i][j+1][1])/2.0;
      
      int left=image((int)pleft[0],(int)pleft[1],0)>0;
      int right=image((int)pright[0],(int)pright[1],0)>0;
      
      if(left==1 && right==0){
	votescorrect++;
      }
      else if(left==0 && right==1){
	votesincorrect++;
      }
      else{
	
      }
    }
    if(votesincorrect>votescorrect){
#ifdef PRINT_DEBUG
      printf("oriented wrong %d %d\n",votescorrect,votesincorrect);
#endif
      for(int j=0;j<shapeSizes[i]/2;j++){
	  linept temp;
	  memcpy(temp,shapes[i][j],sizeof(linept));
	  memcpy(shapes[i][j],shapes[i][shapeSizes[i]-j-1],sizeof(linept));
	  memcpy(shapes[i][shapeSizes[i]-j-1],temp,sizeof(linept));
	}
    }
  }
#ifdef PRINT_DEBUG
  printf("refining shape\n");
#endif 
  char ** toRemove=refineShape(image,shapes,shapeSizes,nshapes);
  silData.removeAll();
  for(int i=0;i<nshapes;i++){
    int closed=(fabs(shapes[i][0][0]-shapes[i][shapeSizes[i]-1][0])+
		fabs(shapes[i][0][1]-shapes[i][shapeSizes[i]-1][1])<=1);
    //memset(toRemove[i],0,sizeof(char)*shapeSizes[i]);
    for(int j=0;j<shapeSizes[i];j++){
      if(!toRemove[i][j]){
	double p1[2];
	double p2[2];
	p1[0]=shapes[i][j][0];
	p1[1]=shapes[i][j][1];
	
	int k=1;
	if(closed){
	  for(;k<shapeSizes[i];k++){
	    if(!toRemove[i][(j+k)%shapeSizes[i]]){
	      break;
	    }
	  }
	  if(k<shapeSizes[i]){
	    p2[0]=shapes[i][(j+k)%shapeSizes[i]][0];
	    p2[1]=shapes[i][(j+k)%shapeSizes[i]][1];
	    
	    silData.insert(p1,p2);
	  }
	}
	//Don't allow the edges to wrap around..
	else{
	  for(k=j+1;k<shapeSizes[i];k++){
	    if(!toRemove[i][k]){
	      break;
	    }
	  }
	  if(k<shapeSizes[i]){
	    p2[0]=shapes[i][k][0];
	    p2[1]=shapes[i][k][1];
	    
	    silData.insert(p1,p2);
	  } 
	}
	
      }
    }
    delete shapes[i];
  }
#ifdef PRINT_DEBUG
  printf("done refinement\n");
#endif
  delete shapes;
  delete shapeSizes;
  delete toRemove;
#ifdef PRINT_DEBUG
  printf("done refinement, and delete\n");
#endif
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  
  glMultMatrixd(intrinsic);
  glMultMatrixd(projection);
  glMultMatrixd(extrinsic);
  
  glTranslatef(trans[0],trans[1],trans[2]);
  glScalef(scale[0],scale[1],scale[2]);
  
  //printf("shave 0:\n");
  raysets[0].shave(image);
  //printf("shave 1:\n");
  raysets[1].shave(image);
  //printf("shave 2:\n");
  raysets[2].shave(image);

  glPopMatrix();
}

/*
void setupTexture(int t)
{
    float mat[]={0.5,0,0,0,
		 0,0.5,0,0,
		 0,0,0.5,0,
		 0.5,0.5,0.5,1.0};
    float splane[4],tplane[4],rplane[4],qplane[4];

    glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
    glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
    glTexGeni(GL_R,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
    glTexGeni(GL_Q,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMultMatrixf(mat);
    glOrtho(0,256,0,256,-10,1);
    glMultMatrixd(intrinsic);
    glMultMatrixd(projection);
    glMultMatrixd(extrinsics+t*16);
    //glRotatef(angle*180/M_PI,0,1,0);
    
    glTranslatef(mi->trans[0],mi->trans[1],mi->trans[2]);
    glScalef(mi->scale[0],mi->scale[1],mi->scale[2]);

    glGetFloatv(GL_MODELVIEW_MATRIX,mat);
    glPopMatrix();

    splane[0]=mat[0];splane[1]=mat[4];splane[2]=mat[8];splane[3]=mat[12];
    tplane[0]=mat[1];tplane[1]=mat[5];tplane[2]=mat[9];tplane[3]=mat[13];
    rplane[0]=mat[2];rplane[1]=mat[6];rplane[2]=mat[10];rplane[3]=mat[14];
    qplane[0]=mat[3];qplane[1]=mat[7];qplane[2]=mat[11];qplane[3]=mat[15];

    glTexGenfv(GL_S,GL_EYE_PLANE,splane);
    glTexGenfv(GL_T,GL_EYE_PLANE,tplane);
    glTexGenfv(GL_R,GL_EYE_PLANE,rplane);
    glTexGenfv(GL_Q,GL_EYE_PLANE,qplane);
    
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,tex[t]);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);
    glEnable(GL_TEXTURE_GEN_Q);
}
*/


void display();

void displayCameraViews()
{
  static GLuint depthTex=0;
  static int i=0;
  static int count=0;
  if(i==0){
    if(texImages!=0)delete texImages;;
    texImages=new unsigned char[256*256*3*nimages];
    if(occluded!=0)delete occluded;
    occluded = new unsigned char[texw[0]*texh[0]*nimages];
  }
  if(i>=nimages){
    glutDisplayFunc(display);
    glutIdleFunc(display);
    return;
  }
  getTextures(intrinsics,projection,extrinsics,
	      images,
	      charts,ncharts,
	      mi->scale,mi->trans,
	      texImages,occluded,texw[0],texh[0],i,i+1);
  glutSwapBuffers();
  i++;
}

void display(){
    glClearColor(0.45,0.45,0.7,1.0);
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    
    glPushMatrix();

    double matrix[16];
    glGetDoublev(GL_MODELVIEW_MATRIX,matrix);
    glLoadIdentity();
    glTranslatef(cpos[0],cpos[1],cpos[2]);
    glMultMatrixd(matrix);

    glDisable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glDisable(GL_CULL_FACE);
    glPolygonMode(GL_FRONT,GL_FILL);
    glPolygonMode(GL_BACK,GL_FILL);
    //marchingCubesTest(0,0);
    
    if(!drawMarchingCubes)
    {
      int w=mi->w;
      int h=mi->h;
      int d=mi->d;
	mi->draw(draw1,draw2,draw3);
	/*
	glPushMatrix();
	glTranslatef(-0.5,-0.5,-0.5);
	glColor3f(1.0,0.0,0.0);
	glPointSize(3);
	glBegin(GL_POINTS);
	for(int k=0;k<d;k++){
	  for(int i=0;i<h;i++){
	    for(int j=0;j<w;j++){
	      gawd=0;
	      mi->inside(j,i,k);
	      if(gawd==3){
		glColor3f(0.0,0.0,1.0);
		glVertex3f(j/(double)(w-1),i/(double)(h-1),k/(double)(d-1));
	      }
	      if(gawd!=0 && gawd!=3){
		glColor3f(1.0,0.0,0.0);
		glVertex3f(j/(double)(w-1),i/(double)(h-1),k/(double)(d-1));
	      }
	    }
	  }
	}
	glEnd();
	glPopMatrix();*/
    }
    if(drawMarchingCubes){
      glPushMatrix();
      //glScalef(mi->scale.x,mi->scale.y,mi->scale.z);
      //glTranslatef(mi->trans.x,mi->trans.y,mi->trans.z);
      glTranslatef(-0.5,-0.5,-0.5);
            
      glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,1);
      glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,1);
      glEnable(GL_LIGHTING);
      glEnable(GL_LIGHT0);
      /* glEnable(GL_LIGHT1);
	 glEnable(GL_LIGHT2);
	 glEnable(GL_LIGHT3);
	 glEnable(GL_LIGHT4);
	 float pos0[]={0,10,0,1};
	 float pos1[]={0,-10,0,1};
	 float pos2[]={0,0,10,1};
	 float pos3[]={0,0,-10,1};
	 float pos4[]={10,0,0,1};
	 float pos5[]={-10,0,0,1};
	 glLightfv(GL_LIGHT0,GL_POSITION,pos0);
    glLightfv(GL_LIGHT1,GL_POSITION,pos1);
    glLightfv(GL_LIGHT2,GL_POSITION,pos2);
    glLightfv(GL_LIGHT3,GL_POSITION,pos3);
    glLightfv(GL_LIGHT4,GL_POSITION,pos4);
    */
      glColor3f(0.9,0.9,0.9);
      glDisable(GL_BLEND);
    
      
      if(drawtex){
	printf("texture drawing disabled...\n");
	/*
	float mat2[16];
	glGetFloatv(GL_MODELVIEW_MATRIX,mat2);
	double angle=atan2(mat2[2],mat2[0]);
	angle+=M_PI;
	//angle*=-1;
	//angle+=M_PI;
      	if(angle<0)angle+=2*M_PI;
	int t=(int)(nimages*angle/(2*M_PI));
	printf("angle %f\n",angle);
	printf("t %d\n",t);
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,tex[t]);

	float val=(nimages*angle/(2.0*M_PI))-(float)t;
	val=1-val;
	float half[]={val,val,val,val};
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE_EXT);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_EXT, GL_INTERPOLATE);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_RGB_EXT, GL_TEXTURE0);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_RGB_EXT, GL_TEXTURE1);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_RGB_EXT, GL_CONSTANT_EXT);
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, half);
	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,tex[(t+1)%nimages]);
	*/
	/* Projective texture setup
	  float mat2[16];
	  glGetFloatv(GL_MODELVIEW_MATRIX,mat2);
	  double angle=atan2(-mat2[2],mat2[0]);
	  //angle+=M_PI;
	  //angle*=-1;
	  //angle+=M_PI;
	  if(angle<0)angle+=2*M_PI;
	  int t=(int)(nimages*angle/(2*M_PI));
	  glActiveTexture(GL_TEXTURE0);
	  setupTexture(t);
	  
	  float val=(nimages*angle/(2.0*M_PI))-(float)t;
	  //val=0;
	  val=1-val;
	  float half[]={val,val,val,val};
	  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE_EXT);
	  glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB_EXT, GL_INTERPOLATE);
	  glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_RGB_EXT, GL_TEXTURE0);
	  glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_RGB_EXT, GL_TEXTURE1);
	  glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_RGB_EXT, GL_CONSTANT_EXT);
	  glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, half);
	  //printf("val: %f\n",val);
	  glActiveTexture(GL_TEXTURE1);
	  setupTexture((t+1)%nimages);*/
      }
      
      //glDisable(GL_LIGHTING);
      //glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      if(wireframe){
	glPolygonMode(GL_FRONT,GL_LINE);
      }
      if(ncharts && charts){
	for(int j=0;j<ncharts;j++){
	  float diff[4];
	  diff[0]=(float)j/(ncharts);
	  diff[1]=1-diff[0];
	  diff[2]=0.5;
	  diff[3]=0.5;
	  glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,diff);
	  charts[j].draw(drawtex);
	}
      }
      if(wireframe){
	glPolygonMode(GL_FRONT,GL_FILL);
      }

      glActiveTexture(GL_TEXTURE0);
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
      glDisable(GL_TEXTURE_GEN_R);
      glDisable(GL_TEXTURE_GEN_Q);
      glDisable(GL_TEXTURE_2D);
      glActiveTexture(GL_TEXTURE1);
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
      glDisable(GL_TEXTURE_GEN_R);
      glDisable(GL_TEXTURE_GEN_Q);
      glDisable(GL_TEXTURE_2D);

      glDisable(GL_LIGHTING);
      if(drawnormals && ncharts){
	  glBegin(GL_LINES);
	  for(int i=0;i<charts[0].npoly;i++){
	      Vec3f v2=charts[0].norm[charts[0].poly[i][0]];
	      v2.mult(0.1);
	      v2.add(charts[0].vert[charts[0].poly[i][0]]);
	      glVertex3fv(charts[0].vert[charts[0].poly[i][0]].getData());
	      glVertex3fv(v2.getData());
	      
	  }
	  glEnd();
      }

      if(ncharts){
	glColor3f(1.0,0.0,1.0);
	glPointSize(4);
	glBegin(GL_POINTS);
	for(int i=0;i<charts[0].nvert;i++){
	  if(bad[i]){
	    glVertex3fv(charts[0].vert[i].getData());
	  }
	}
	glEnd();
      }

      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      //delete poly;
      glPopMatrix();


      if(camtri){
	glClear(GL_COLOR_BUFFER_BIT);
	glBegin(GL_TRIANGLES);
	for(int i=0;i<ncamtri;i++){
	  glColor3f((float)i/(float)ncamtri,
		    (float)i/(float)ncamtri,
		    (float)i/(float)ncamtri);
	  glVertex3dv(campts[camtri[i][0]].getData());
	  glVertex3dv(campts[camtri[i][1]].getData());
	  glVertex3dv(campts[camtri[i][2]].getData());
	}
	glEnd();
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	 	
	for(int k=0;k<6;k++){
	  int ind=0;
	  int ind0=0;
	  int ind1=0;
	  if(k<2){
	    ind=0;
	    ind0=1;
	    ind1=2;
	  }
	  else if(k<4){
	    ind=1;
	    ind0=0;
	    ind1=2;
	  }
	  else if(k<6){
	    ind=2;
	    ind0=0;
	    ind1=1;
	  }
	  float pts[4][3];
	  pts[0][ind]=face[k][ind];
	  pts[0][ind0]=-1;
	  pts[0][ind1]=-1;

	  pts[1][ind]=face[k][ind];
	  pts[1][ind0]=1;
	  pts[1][ind1]=-1;

	  pts[2][ind]=face[k][ind];
	  pts[2][ind0]=1;
	  pts[2][ind1]=1;

	  pts[3][ind]=face[k][ind];
	  pts[3][ind0]=-1;
	  pts[3][ind1]=1;
	    	    

	  glColor4f(1.0,1.0,1.0,1.0);
	  glBindTexture(GL_TEXTURE_2D,cubeTex[k]);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0.0,0.0);glVertex3fv(pts[0]);
	  glTexCoord2f(1.0,0.0);glVertex3fv(pts[1]);
	  glTexCoord2f(1.0,1.0);glVertex3fv(pts[2]);
	  glTexCoord2f(0.0,1.0);glVertex3fv(pts[3]);
	  glEnd();
	}
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glPopMatrix();
	glutSwapBuffers();
	return;
      }
    }
    glPopMatrix();

    
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,curw,0,curh,-100,100);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glDisable(GL_LIGHTING);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    
    glColor4f(1.0,1.0,1.0,0.2+0.7*(draw));
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    //glBindTexture(GL_TEXTURE_2D,tex[(ci+nimages-1)%nimages]);
    glBindTexture(GL_TEXTURE_2D,tex);
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    glBegin(GL_QUADS);
    glTexCoord2f(0,0);glVertex3f(0,0,0);
    glTexCoord2f(0,1.0);glVertex3f(0,curh,0);
    glTexCoord2f(1.0,1.0);glVertex3f(curw,curh,0);
    glTexCoord2f(1.0,0);glVertex3f(curw,0,0);
    glEnd();
    
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glColor4f(1.0,1.0,1.0,1.0);


    if(draw){
      glPushMatrix();
      mi->draw(images[ci],intrinsic,projection,extrinsic,draw1,draw2,draw3);
      glPopMatrix();
    }
    
    //shaveIt();
    
    
    
    //printf("glerror %d\n",glGetError());
    
    glColor3f(0.0,0.0,1.0);
    for(int i=0;i<silData.nlines;i++){
	glBegin(GL_LINES);
	glVertex2dv(silData.data[i].p);
	double o[2];
	
	o[0]=silData.data[i].p[0]+silData.data[i].d[0];
	o[1]=silData.data[i].p[1]+silData.data[i].d[1];
	glVertex2dv(o);
	glEnd();
    }

    /*
    glDisable(GL_BLEND);
    for(int i=0;i<nshapes;i++){
	glColor3f((double)i/(double)nshapes,1.0-(double)i/(double)nshapes,1.0);
	glBegin(GL_LINE_LOOP);
	glVertex2iv(shapes[i][0]);
	
	for(int j=0;j<shapeSizes[i];j++){
	    if(!toRemove[i][j]){
		glVertex2iv(shapes[i][j]);
	    }
	}

	glVertex2iv(shapes[i][shapeSizes[i]-1]);
	glEnd();
	
	glPointSize(4);
	glColor3f(0.0,0.0,1.0);
	glBegin(GL_POINTS);

	for(int j=0;j<shapeSizes[i];j++){
	    if(!toRemove[i][j]){
		glVertex2iv(shapes[i][j]);
	    }
	}
	glColor3f(1.0,1.0,1.0);
	glEnd();

	delete toRemove[i];
	delete shapes[i];
    }
    delete shapes;
    delete shapeSizes;
    delete toRemove;
    */

    glClear(GL_DEPTH_BUFFER_BIT);
    glPushMatrix();
    glLoadIdentity();
    glColor3f(1.0,1.0,1.0);
    if(ncharts){
      int y=130;
      glColor3f(0.0,0.0,0.0);
      for(int i=ncharts-1;i>=0;i--){
	double avg,min,max,wa,ta;
	drawString(10,y,"tri: %d vert:%d ratios:%f",charts[i].npoly,
		   charts[i].nvert,charts[i].areaRatios(avg,min,max,wa,ta));
	y+=9;
      }
    }
    glPopMatrix();


    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

    //Draw TC

    /*
    int wid=(int)(sqrt(ncharts)+0.5);
    int hi=((double)ncharts/(double)wid+0.5);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,wid,0,hi,-1,1);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glDisable(GL_LIGHTING);
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    glColor3f(1.0,1.0,1.0);

    glTranslatef(wid,-1,0);
    if(ncharts){
      for(int i=0;i<ncharts;i++){
	if(i%wid==0){
	  glTranslatef(-wid,1,0);
	}
	charts[i].drawTC();
	glTranslatef(1,0,0);
      }
    }
    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_LIGHTING);
    glPolygonMode(GL_FRONT,GL_FILL);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    //End Draw TC
    */

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,1,0,1,-1,1);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glDisable(GL_LIGHTING);
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    glColor3f(1.0,1.0,1.0);

    if(ncharts){
      for(int i=0;i<ncharts;i++){
	charts[i].drawTC();
      }
    }
    glColor3f(1.0,1.0,1.0);

    glEnable(GL_DEPTH_TEST);
    glPolygonMode(GL_FRONT,GL_FILL);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    

    glutSwapBuffers();
}

void * shaveIt_t(void *){
  shaveIt();
}

void shaveIt(){
    static int called=0;
    static int done=0;
    if(ci>=nimages){
      done=1;
      ci=ci%nimages;
      return;
    }
    if(done)return;
/*
    memset(intrinsic,0,sizeof(double)*16);
    memset(projection,0,sizeof(double)*16);
    intrinsic[0]=1200;
    intrinsic[5]=1200;
    intrinsic[10]=1;
    intrinsic[15]=1;
    intrinsic[12]=128;
    intrinsic[13]=128;//128;
    
    projection[0]=1;
    projection[5]=1;
    projection[10]=-1;
    projection[11]=-1;
*/
   
    //for(int i=0;i<16;i++)
    if(ci<nimages)
    {
      /*glPushMatrix();
	glLoadIdentity();
	glTranslatef(0,0,2.5);
	glRotatef(-ci*360.0/nimages,0,1,0);
	glGetDoublev(GL_MODELVIEW_MATRIX,extrinsic);
	glPopMatrix();*/
	memcpy(extrinsic,extrinsics+16*ci,sizeof(double)*16);
	memcpy(intrinsic,intrinsics+16*ci,sizeof(double)*16);
	if(called){
	    Image8 mask;
	    curw=images[ci].width;
	    curh=images[ci].height;
	    mask=images[ci].getChannel(3);
	    mi->shave(mask,intrinsic,projection,extrinsic);

	    if(debug){
	      glEnable(GL_TEXTURE_2D);
	      glBindTexture(GL_TEXTURE_2D,tex);
	      Image8 other=images[ci].subimage(0,0,curw,curh);
	      other.initTexture();
	      glDisable(GL_TEXTURE_2D);
	    }
	    ci++;
	}
    }
    called++;
}




void * test(void * v)
{
  for(int i=0;i<10000000;i++){
    double v = sqrt(v);
    if(i%10000==0)
      printf("%d %f\n",i,v);
  }
}

void doMarchingCubes()
{
    memset(bad,0,sizeof(char)*100000);
    if(ncharts){
      ncharts=0;
      delete [] charts;
    }
    charts = new Chart[1];
    memset(charts,0,sizeof(Chart));
    ncharts=1;
    charts[0].npoly=marchingCubes(*mi,&charts[0].poly,
				  &charts[0].vert,
				  &charts[0].nvert,
				  &charts[0].norm,MIN_DIST);
    drawMarchingCubes=1;
    charts[0].initNormals();
}

void writeGeometry()
{
  Matrix mat(0,0);
  saveobj(outfile,charts,1,mat);
}

void keys(unsigned char c,int x,int y){
  static int which=0;
  
  switch(c){
  case 'W':
    writeGeometry();
    exit(0);
    break;
  case 'w':
    wireframe=!wireframe;
    break;
  case 'n':
    drawnormals=!drawnormals;
    break;
  case 'm':
    doMarchingCubes();
    break;
  case 'd':
    {
      int nreig[3]={nimages,nimages,nimages};
      initDynaTex("/usr/scratch/birkbeck/sfs/data0",
		  charts,ncharts,mi->scale,mi->trans,
		  extrinsics,
		  texImages,occluded,texw[0],texh[0],
		  nimages,nreig,texw,texh);
      break;
    }
  case 'p':
    packCharts(charts,ncharts,256,256);
	    break;
	    
	case 'C':
	    for(int i=0;i<ncharts;i++)
		getTexCoordsLSCM(charts[i]);
	    break;
	case 'c':
	    for(int i=0;i<ncharts;i++)
		getTexCoordsMDS(charts[i]);
	    //getTexCoordsMDS(g_poly,g_npoly,g_vert,g_nvert);
	    break;
	case 't':
	    drawtex=!drawtex;
	    break;
	case 'i':{
	    break;
	}
	case '1':
	    draw1=!draw1;
	    break;
	case '2':
	    draw2=!draw2;
	    break;
	case '3':
	    draw3=!draw3;
	    break;
	case '0':
	    draw=!draw;
	    break;
	case 'S':
	    //pthread_t t1;
	    //pthread_create(&t1,NULL,&split,NULL);
	    //printf("thread exited\n");
	    //pthread_join(t1,NULL);
	{
	    int newncharts=0;
	    Chart * newcharts=split(charts,newncharts);
	    delete [] charts;
	    charts=newcharts;
	    ncharts=newncharts;
	    printf("split created %d\n",ncharts);
	    break;
	}
	case 'g':
	  printf("about to simplify\n");
	  simplify(charts[0],60);
	  
	  //simplify(&charts[0].poly,&charts[0].npoly,&charts[0].vert,
	  //     &charts[0].nvert,&charts[0].norm);
	    break;
	case '[':
	    glutSetWindow(win);
	    glutDisplayFunc(displayCameraViews);
	    glutIdleFunc(displayCameraViews);
	    break;
	case ']':
	    glutSetWindow(win);
	    glutIdleFunc(display);
	    glutDisplayFunc(display);
	    break;
	case '+':
	    ci=(ci+1)%ncharts;
	    break;
	case 's':
	    drawMarchingCubes=0;
	    shaveIt();
	    break;
    }
    glutPostRedisplay();
    which++;
}

void init(){
  images = new Image8[nimages];
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45,(double)1.0,0.1,1000);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  intrinsics = new double[16*nimages];
  extrinsics = new double[16*nimages];
  
  /*OLD loading file FILE * file = fopen(extfile,"rb");
    if(!file){
    printf("using default angles\n");
    glPushMatrix();
    for(int i=0;i<nimages;i++){
    glLoadIdentity();
    glTranslatef(0,0,3);
    glRotatef(-i*360.0/(double)nimages,0,1,0);
    
    glGetDoublev(GL_MODELVIEW_MATRIX,extrinsics+16*i);
      }
      glPopMatrix();
      }
      else{*/
  
  char fname[1024];
  
  for(int i=0;i<nimages;i++){
    FILE * f;
    Matrix intr;
    Matrix extr;
    Matrix proj;
    Matrix dist;
    CalibInfo ci;
    
    snprintf(fname,1024,extfile,i);
    f=fopen(fname,"rb");
#ifdef PRINT_DEBUG
    printf("fname %s %d\n",fname,f);
#endif
    if(!f){
      printf("cannot load camera parameters for image %d\n",i);
    }
    else{
      double * extrinsic = extrinsics+16*i;
      double * intrinsic = intrinsics+16*i;
      readCLB(f,intr,extr,proj,dist,ci);
      for(int k=0;k<4;k++){
	for(int j=0;j<4;j++){
	  extrinsic[k+j*4]=extr(k,j);
	}
      }
      memset(intrinsic,0,sizeof(double)*16);
      intrinsic[0]=intr(0,0);
      intrinsic[5]=intr(1,1);
      intrinsic[10]=1;
      intrinsic[12]=intr(0,2);
      intrinsic[13]=intr(1,2);
      intrinsic[15]=1;
      
      memset(projection,0,sizeof(double)*16);
      projection[0]=1;
      projection[5]=1;
      projection[10]=1;
      projection[11]=1;
#ifdef PRINT_DEBUG
      extr.printMatlab("extr");
      intr.printMatlab("intr:");
#endif
      fclose(f);
    }
  }
  
  glEnable(GL_TEXTURE_2D);
  glGenTextures(1,&tex);
  glDisable(GL_TEXTURE_2D);
  
  for(int i=0;i<nimages;i++){
    char fname[1024];
    snprintf(fname,1024,imfile,i);
    images[i].read(fname);
  }
  glEnable(GL_DEPTH_TEST);
}

void fuck(){
  int np=4;
  srand(-1);
  Matrix xd=Matrix::random(np,1);
  Matrix yd=Matrix::random(np,1);
  Matrix zd(np,1);
  Matrix xi(12,12);
  Matrix yi(12,12);

  for(int i=0;i<np;i++){
    zd[i]=exp(-((xd[i]-0.5)*(xd[i]-0.5) +
		(yd[i]-0.5)*(yd[i]-0.5)));
  }

  for(int i=0;i<xi.m;i++){
    for(int j=0;j<xi.n;j++){
      xi(i,j)=(double)j/(double)xi.n;
      yi(i,j)=(double)i/(double)xi.m;
    }
  }
  glShadeModel(GL_SMOOTH);
  Matrix zi=griddata(xd,yd,zd,xi,yi,LUT_LINEAR);
  xd.printMatlab("xd");
  yd.printMatlab("yd");
  zd.printMatlab("zd");
  xi.printMatlab("xi");
  yi.printMatlab("yi");
  zi.printMatlab("zi");
  
  glutSwapBuffers();
}

void processCommandLine(int ac,char * av[])
{
    static struct option long_opt[] = {
                   {"sx", 1, 0, 0},
                   {"sy", 1, 0, 0},
                   {"sz", 1, 0, 0},
                   {"tx", 1, 0, 0},
                   {"ty", 1, 0, 0},
		   {"tz", 1, 0, 0},
                   {"dir", 1, 0, 0},
		   {"file", 1, 0, 0},
		   {"ofile",1,0,0},
		   {"nrays", 1, 0, 0}, 
		   {"nimages", 1, 0, 0},
                   {0, 0, 0, 0}
               };
    Vec3f sc(1,1,1);
    Vec3f tr(-0.5,-0.5,-0.5);
    char dir[1024]=".";
    char file[256]="image%.4.tga";
    int nrays=32;
    nimages=0;
    while(1){
	int opt_index=0;
	char c=getopt_long(ac,av,"",long_opt,&opt_index);
	printf("%d %d\n",c,opt_index);
	if(c==-1)break;
	switch(opt_index){
	    case 0:
	    case 1:
	    case 2:
		sscanf(optarg,"%f",&(sc[opt_index]));
		break;
	    case 3:
	    case 4:
	    case 5:
		sscanf(optarg,"%f",&(tr[opt_index-3]));
		break;	
	    case 6:
		snprintf(dir,1024,"%s",optarg);
		break;
	    case 7:
		snprintf(file,256,"%s",optarg);
		break;
	    case 8:
		snprintf(outfile,1024,"%s",optarg);
		break;
	    case 9:
		sscanf(optarg,"%d",&nrays);
		break;	
	    case 10:
		sscanf(optarg,"%d",&nimages);
		break;
	}
    }
    printf("%s/%s nrays=%d\n",dir,file,nrays);
    printf("Scale %f %f %f\n",sc.x,sc.y,sc.z);
    printf("Trans %f %f %f\n",tr.x,tr.y,tr.z);

    mi = new MI(nrays,nrays,nrays);
    mi->setTrans(tr);
    mi->setScale(sc);

    snprintf(extfile,1024,"%s/%s.clb",dir,file);
    snprintf(imfile,1024,"%s/%s",dir,file);
    printf("extfile %s\nimfile %s\n",extfile,imfile);
}

#define NNORM 10000000
int main(int ac,char * av[]){
  processCommandLine(ac,av);

  printf("calling init\n");
  glutInit(&ac,av);
  printf("done glutInit\n");
  glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_STENCIL|GLUT_RGB);
  printf("initing isplay mode\n");

  /*Matrix A=Matrix::random(4,4);
  Matrix a,b;
  A.eig(a,b);
  A.printMatlab("A");
  a.print();
  b.print();
  return 0;*/
  /*Vec3f axis;
  double angle;
  Vec3f * norms = new Vec3f[NNORM];
  int i=0;
  for(i=0;i<NNORM;i++){
    norms[i]=Vec3f(((double)(rand()%2048))/2048.0-0.5,
		  ((double)(rand()%2048))/2048.0-0.5,
		  1);
    norms[i].normalize();
    //norms[i].print();
  }
  cone(norms,NNORM,axis,angle);
  for(i=0;i<NNORM;i++){
    if(!incone(axis,angle,norms[i])){
      break;
    }
  }
  if(i==NNORM){
    printf("all %d normals within cone\n",i);
  }
  else
    printf("some normals not in cone\n");
  axis.print();

  
  printf("angle %f\n",angle);
  return 0;
  */ 
  glutInitWindowSize(512,512);      
  win=glutCreateWindow("window");
  
  if(debug){
      glutMouseFunc(mouse);
      glutMotionFunc(motion);
      glutKeyboardFunc(keys);
      glutDisplayFunc(display);
      glutIdleFunc(display);
    
    //glutDisplayFunc(fuck);
    //glutIdleFunc(fuck);

    //sprintf(extfile,"out.calib");
    //strcpy(imfile,"out-%02d.ppm");
    //nimages=32;

    //Matrix m = Matrix::random(512*512,128);
    //pca(m,96);
    //return 0;
       
      init();
      printf("going into main loop\n");
      glEnable(GL_NORMALIZE);
      glutMainLoop();
    }
    else{
      init();
      for(int i=0;i<nimages;i++){
	printf("message: shaving %d\n",i);
	printf("progress: %d\n",i);
	fflush(stdout);
	shaveIt();
      }
      doMarchingCubes();
      writeGeometry();
    }
    return 0;
}
#endif




