#include <stdio.h>
#include <stdlib.h>
//#include "Vec3.h"
#include "matrix.h"
#include "vec.h"
#include "image.h"
#include "mc.h"
#ifndef RAY_H
#define RAY_H

class Ray{
 public:
  int flag;
    Vec3f start;
    Vec3f dir;
    double * enter;
    double * exit;
    int nseg;
    int maxseg;
    ~Ray(){
	delete enter;
	delete exit;
	maxseg=0;
	nseg=0;
	    }
    Ray(){
	enter=0;
	exit=0;
	nseg=0;
	maxseg=0;
	flag=0;
    }
    void init(Vec3f & s,Vec3f & e){
	start=s;
	dir=e;
	dir.sub(start);

	nseg=1;
	maxseg=1;

	enter=new double[maxseg];
	exit=new double[maxseg];
	enter[0]=0;
	exit[0]=1;
    }
    int  inside(double t);
    void makeConsistent();
    void insert(double start,double stop);
    void merge(Ray & o);
    void draw();
    void print();
    int getBoundary(double start,double stop,int t1on,int t2on,Vec3f & output);
    int getBoundary(double start,double stop,Vec3f & output);
    void include(double t);
    void remove(double t);
    
};


class RaySet{
 public:
    int w;
    int h;
    Ray * rays;
    int plane;
    RaySet(int w=0,int h=0,int _plane=0){
	rays=0;
	this->w=w;
	this->h=h;
	rays = new Ray[w*h];
	plane=_plane;
	printf("initializing plane %d\n",plane);
	for(int i=0;i<h;i++){
	  for(int j=0;j<w;j++){
	    Vec3f start;
	    Vec3f end;
		switch(plane){
		case 0:
		  start=Vec3f((double)j/(w-1),(double)i/(h-1),0.0);
		  end=Vec3f((double)j/(w-1),(double)i/(h-1),1.0);
		  break;
		case 1:
		  start=Vec3f((double)j/(w-1),0.0,(double)i/(h-1));
		  end=Vec3f((double)j/(w-1),1.0,(double)i/(h-1));
		  break;
		case 2:
		  start=Vec3f(0.0,(double)j/(w-1),(double)i/(h-1));
		  end=Vec3f(1.0,(double)j/(w-1),(double)i/(h-1));
		  break;
		}
		rays[i*w+j].init(start,end);
	  }
	}
	printf("done initializing plane %d\n",plane);
    }
    void draw();
    void shave(Image8 & image);
    void draw(Image8 & image,double * ,double *,double *);
    int inside(int x,int y,double z);
};

class MI: public Cubable
{
 public:
  Vec3f scale;
  Vec3f trans;
  int w,h,d;

  RaySet raysets[3];
  MI(int _w,int _h,int _d){
    w=_w;
    h=_h;
    d=_d;
    raysets[0]=RaySet(w,h,0);
    raysets[1]=RaySet(w,d,1);
    raysets[2]=RaySet(h,d,2);
    scale=Vec3f(1 ,1 ,1 );
    trans=Vec3f(-0.5,-0.5,-0.5);
  }
  int getWidth(){return w;}
  int getHeight(){return h;}
  int getDepth(){return d;}
  int getBoundaryPoint(Vec3f & pt1,Vec3f & pt2,int p1on,int p2on,Vec3f & output);
  int getBoundaryPoint2(Vec3f & pt1,Vec3f & pt2,int p1on,int p2on,Vec3f & output);
  void draw(int ,int ,int );
  void draw(Image8 &,double *,double *,double *,int ,int ,int);
  int inside(int,int,int);
  void shave(Image8 &,double *,double *,double *);
  void setTrans(Vec3f & tr){
      trans=tr;
  }
  void setScale(Vec3f & sc){
      scale=sc;
  }
};
#endif
