#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//format information at 
//http://duke.usask.ca/~macphed/soft/fig/FORMAT3.2.txt
#define XFIG_VERSION_STRING "#FIG 3.2"

const char orient2str[2][10]={"Landscape","Portrait"};
const char justification2str[][16]={"Center","Flush Left"};
const char units2str[][16]={"Metric","Inches"};
const char papersize2str[][16]={"Letter", "Legal", "Ledger", "Tabloid",
				       "A", "B", "C", "D", "E",
				       "A4",   "A3", "A2", "A1", "A0","B5"};
const char multiple_page2str[][16] = {"Single","Multiple"};

class xfigheader{
  enum {LANDSCAPE,PORTRAIT} orient;
  enum {CENTER,FLUSH_LEFT} justification;
  enum {METRIC,INCHES} units;
  enum {LETTER,LEGAL,LEDGER,TABLOID,A,B,C,D,E,A4,A3,A2,A1,A0,B5} papersize;
  double magnification;
  enum {SINGLE,MULTIPLE} multiple_page;
  
  int   transparent;//(color number for transparent color for GIF
  //export. -2=None, -1=background, 0-31 for
  //standard colors or 32- for user colors)
  int	resolution;
  const static int      coord_system=2;//upper left is origin
 public:
  xfigheader(){
    orient=LANDSCAPE;
    justification=CENTER;
    units=METRIC;
    papersize=LETTER;
    magnification=100;
    multiple_page=SINGLE;

    transparent=-2;    
    resolution=1200;
  }
  void print(FILE * file=stdout){
    fprintf(file,"%s\n%s\n%s\n%s\n%s\n%g\n%s\n%d\n%d %d\n",
	    XFIG_VERSION_STRING,
	    orient2str[orient],
	    justification2str[justification],
	    units2str[units],
	    papersize2str[papersize],
	    magnification,
	    multiple_page2str[multiple_page],
	    transparent,
	    resolution,
	    coord_system);
  }
}header;




class circle{
  const static int object_code=1;	
  const static int	sub_type=3;//		(1: ellipse defined by radii
  //	                         2: ellipse defined by diameters
  //				 3: circle defined by radius
  //				 4: circle defined by diameter)
  const static int	line_style=0;		//(enumeration type)
  const static int	thickness=1;		//(1/80 inch)
  const static int	pen_color=0;            //(enumeration type, pen color)
  const static int	fill_color=0;		//(enumeration type, fill color)
  const	static int	depth=50;               //(enumeration type)
  const static int	pen_style=0;            //(pen style, not used)
  const static int	area_fill=20;           //(enumeration type, -1 = no fill)
  const static float	style_val=0;		//(1/80 inch)
  const static int	direction=1;		//(always 1)
  const static float	angle=0;		//(radians, the angle of the x-axis)
  
  int	center_x, center_y;	//(Fig units)
  int	radius_x, radius_y;	//(Fig units)
  int	start_x, start_y;	//(Fig units; the 1st point entered)
  int	end_x, end_y;		//(Fig units; the last point entered)
 public:
  circle(int x=0,int y=0,int radius=100)
  {
    center_x=x;
    center_y=y;
    radius_x=radius;
    radius_y=radius;
  }
  void print(FILE * file=stdout){
    fprintf(file,"%d %d %d %d %d %d %d %d %d %g %d %g %d %d %d %d %d %d %d %d\n",
	    object_code,sub_type,line_style,thickness,pen_color,fill_color,depth,
	    pen_style,area_fill,style_val,direction,angle,center_x,center_y,
	    radius_x,radius_y,start_x,start_y,end_x,end_y);
  }
}c;


class line{
  const static int   object_code=2;	
  const static int	sub_type=1;
  const static int	line_style=2;	//dotted?
  const static int	thickness=1;	//(1/80 inch)
  const static int	pen_color=0;    //(enumeration type, pen color)
  const static int	fill_color=7;	//(enumeration type, fill color)
  const	static int	depth=50;       //(enumeration type)
  const static int	pen_style=0;    //(pen style, not used)
  const static int	area_fill=-1;   //(enumeration type, -1 = no fill)
  const static float	style_val=6.0;	//(1/80 inch) distance between dots
  const static int	join_style=0;
  const static int	cap_style=0;
  const static int	radius=-1;		
  const static int	forward_arrow=0;	
  const static int	backward_arrow=0;	
  const static int	npoints=2;
  int sx,sy,ex,ey;
 public:
  line(int sx_,int sy_,int ex_,int ey_)
  {
    sx=sx_;
    sy=sy_;
    ex=ex_;
    ey=ey_;
  }
  void print(FILE * file=stdout){
    fprintf(file,"%d %d %d %d %d %d %d %d %d %g %d %d %d %d %d %d\n    %d %d %d %d\n",
	    object_code,sub_type,line_style,thickness,
	    pen_color,fill_color,depth,
	    pen_style,area_fill,style_val,
	    join_style,cap_style,radius,forward_arrow,backward_arrow,npoints,
	    sx,sy,ex,ey);
  }
};


int main(int ac, char * av[])
{
  fprintf(stderr," usage ./a.out w h small_radius (this message always appears)");
  fprintf(stderr,"defaults: w=7500,h=7500,sr=0.031578947 (proportional to w and h)\n");
  header.print();
  float w=7500;
  float h=7500;
  //The radius of the small and large dots.
  //big radius (br) should produce circle with at least twice as 
  //much area. so 2*pi*br^2=4*pi*sr^2 --> br^2=2*sr^2--> br>=sqrt(2)*sr
  double sr = 0.031578947;
  double br;
  
  if(ac>1)
    sscanf(av[1],"%f",&w);
  if(ac>2)
    sscanf(av[2],"%f",&h);
  if(ac>3)
    sscanf(av[3],"%lf",&sr);
  
  br = pow(2,0.55)*sr;

  float minwh=(w<h)?w:h;
  float biggest_radius=minwh/2-br*minwh/2;//given the width and height of the page
  float radii_ratio[]={18.0/30.0,22.0/30.0,26.0/30.0,1.0};
  //Counter clockwise inside to outside
  int pattern [][4] = {//{1,1,1,1},
			{0,0,0,1},
			{0,0,1,0},
			{0,0,1,1},
			{0,1,0,0},
			{0,1,0,1},
			{0,1,1,0},
			{0,1,1,1},
			{1,0,0,0},
			{1,0,0,1},
			{1,0,1,0},
			{1,0,1,1},
			{1,1,0,0},
			{1,1,0,1},
			{1,1,1,0}};
  
  for(int i=0;i<14;i++){
    double angle=2*M_PI*((double)i)/14.0;
    double c=cos(angle);
    double s=sin(angle);
   
    for(int j=0;j<4;j++){
      double r=pattern[i][j]?br:sr;
      circle circ((int)(biggest_radius*radii_ratio[j]*c+w/2.0),
		  (int)(biggest_radius*radii_ratio[j]*s+h/2.0),
		  (int)(biggest_radius*r));
      circ.print();
    }
  }

  //draw some diagonal lines.
  if(w>7500||h>7500){
    int xs=-w;
    int everyx=5500;
    int ys=500;
    int everyy=5500;
    for(int x=xs;x<2*w;x+=everyx){
      int sx=x;
      int ex=x+w;
      int sy=0;
      int ey=h;
      if(x<w){
	if(sx<0){
	  sy=-sx;
	  sx=0;
	}
	if(ex>w){
	  ey-=(ex-w);
	  ex=w;
	}
	line l(sx,sy,ex,ey);
	l.print();
      }
      if(x>=0){
	sx=x;
	sy=0;
	ex=x-w;
	ey=h;
	if(sx>w){
	  sy=(sx-w);
	  sx=w;
	}
	if(ex<0){
	  ey-=(-ex);
	  ex=0;
	}
	line l(sx,sy,ex,ey);
	l.print();
      }
    }
  }
  return 0;
}
