#include "image.h"
#include "matrix.h"
#include "clb.h"
#include <GL/glut.h>
Matrix P1,P2,P3,H1,H2,H3;
Matrix F1,F2,F3;
GLuint tex[3];
int ww=640;
int wh=480;
Imagef imo[3];

void reshape(int x,int y)
{
  ww=x;
  wh=y;
  glViewport(0,0,ww,wh);
}

void tex_quad(float x,float y,
	      float w,float h)
{
  glBegin(GL_QUADS);
  glTexCoord2f(0,0);glVertex2f(x,y);
  glTexCoord2f(1,0);glVertex2f(x+w,y);
  glTexCoord2f(1,1);glVertex2f(x+w,y+h);
  glTexCoord2f(0,1);glVertex2f(x,y+h);
  glEnd();
}


void display()
{
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,1,0,1,-1,1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();


  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D,tex[0]);
  tex_quad(0,0,0.5,0.5);
  
  Matrix pts(3,4),pts1,pts2,pts3;
  pts.setAll(1);
  pts(0,0)=0;pts(1,0)=0;
  pts(0,1)=imo[0].w;pts(1,1)=0;
  pts(0,2)=imo[0].w;pts(1,2)=imo[0].h;
  pts(0,3)=0;pts(1,3)=imo[0].h;

  pts1=H1*pts;
  pts2=H2*pts;
  pts3=H3*pts;
  for(int j=0;j<pts.n;j++){
    pts1(0,j)/=pts1(2,j);
    pts1(1,j)/=pts1(2,j);
    pts2(0,j)/=pts2(2,j);
    pts2(1,j)/=pts2(2,j);
    pts3(0,j)/=pts3(2,j);
    pts3(1,j)/=pts3(2,j);
  }
  double miny1=pts1.minInRow(1);
  double maxy1=pts1.maxInRow(1);
  double miny2=pts2.minInRow(1);
  double maxy2=pts2.maxInRow(1);

  double minx1=pts1.minInRow(0);
  double maxx1=pts1.maxInRow(0);
  double minx3=pts3.minInRow(0);
  double maxx3=pts3.maxInRow(0);
  
  miny2=0.5*(miny2-miny1)/(maxy1-miny1);
  maxy2=0.5*(maxy2-miny1)/(maxy1-miny1);
  printf("%f %f\n",miny1,maxy1);
  printf("%f %f\n",miny2,maxy2);
  glBindTexture(GL_TEXTURE_2D,tex[1]);
  tex_quad(0.5,miny2,0.5,maxy2-miny2);

  minx3=0.5*(minx3-minx1)/(maxx1-minx1);
  maxx3=0.5*(maxx3-minx1)/(maxx1-minx1);
  printf("%f %f\n",miny1,maxy1);
  printf("%f %f\n",miny2,maxy2);

  glBindTexture(GL_TEXTURE_2D,tex[2]);
  tex_quad(minx3,0.5,maxx3-minx3,0.5);
  
  glDisable(GL_TEXTURE_2D);

  glColor3f(1.0,1.0,1.0);
  glBegin(GL_LINES);
  glVertex2f(0 , 0.1);glVertex2f(1.0,0.1);
  glVertex2f(0 , 0.4);glVertex2f(1.0,0.4);
  
  glVertex2f(0.1,0);glVertex2f(0.1,1.0);
  glVertex2f(0.4,0);glVertex2f(0.4,1.0);
  glVertex2f(0.1,0);glVertex2f(0.1,1.0);
  glVertex2f(0.25,0);glVertex2f(0.25,1.0);
  glEnd();
  
  glutSwapBuffers();
}


Matrix rank2()
{
  Matrix U,S,V;
  Matrix F=Matrix::random(3,3);
  F.Lsvd(U,S,V);
  Matrix s=Matrix::eye(3,3);
  s(0,0)=S[0];
  s(1,1)=S[1];
  s(2,2)=0;
  F=U*s*V.transpose();
  F.Lsvd(U,S,V);
  S.printMatlab("S");
  return F;
}

void findBounds(Imagef & im1,Matrix & H1,
		double & minx,double & maxx,
		double & miny,double & maxy)
{
  Matrix pts(3,4);
  pts.setAll(1);
  pts(0,0)=0;pts(1,0)=0;
  pts(0,1)=im1.w;pts(1,1)=0;
  pts(0,2)=im1.w;pts(1,2)=im1.h;
  pts(0,3)=0;pts(1,3)=im1.h; 

  pts=H1*pts;
  for(int j=0;j<pts.n;j++){
    pts(0,j)/=pts(2,j);
    pts(1,j)/=pts(2,j);
  }

  minx=pts.minInRow(0);
  miny=pts.minInRow(1);
  maxx=pts.maxInRow(0);
  maxy=pts.maxInRow(1);
}

void tri_rectify(int,int,Matrix &,Matrix &,Matrix &,Matrix &,Matrix &,Matrix &);
Matrix fundamentalMatrix(Matrix & P1,Matrix & P2);
int main(int ac,char * av[])
{
  Matrix intr=Matrix::eye(3,4);
 
  P1=intr*Matrix::trans(0,0,-10);
  P2=intr*Matrix::trans(1,0,-10);
  P3=intr*Matrix::trans(0,1,-10);
  F1=fundamentalMatrix(P1,P2);
  F2=fundamentalMatrix(P1,P3);
  F3=fundamentalMatrix(P2,P3);
  tri_rectify(128,128,F1,F2,F3,H1,H2,H3);

  if(ac>=5){
    Matrix P[3];
    Imagef im[3];
    for(int i=0;i<3;i++){
      char name[2048];
      sprintf(name,"%s",av[1]);
      sprintf(name,name,atoi(av[2+i]));
      printf("[%s]\n",name);
      im[i].read(name);
      printf("%d %d\n",im[i].w,im[i].h);
      strcat(name,".clb");
      printf("[%s]\n",name);
      FILE * f=fopen(name,"rb");
      if(!f){
	printf("can not open %s\n",name);
	exit(1);
      }
      Matrix intr(3,3);
      Matrix extr(4,4);
      Matrix pbad(3,4);
      Matrix dist(5,1);
      Matrix p=Matrix::eye(3,4);
      readCLB(f,intr.data,extr.data,pbad.data,dist.data);fclose(f);
      P[i]=intr*p*extr;
    }
    
    F1=fundamentalMatrix(P[0],P[1]);
    F2=fundamentalMatrix(P[0],P[2]);
    F3=fundamentalMatrix(P[1],P[2]);
    tri_rectify(im[0].w,im[0].h,F1,F2,F3,H1,H2,H3);

    Matrix P0h=H1*P[0];
    Matrix P1h=H2*P[1];
    Matrix P2h=H3*P[2];
    
    F1=fundamentalMatrix(P0h,P1h);
    F1.printMatlab("F1");
    
    F2=fundamentalMatrix(P0h,P2h);
    F2.printMatlab("F2");
    
    F3=fundamentalMatrix(P1h,P2h);
    F3.printMatlab("F3");
    

    double minx,maxx,miny,maxy;
    findBounds(im[0],H1,minx,maxx,miny,maxy);
    
    imo[0]=applyHomography(im[0],H1,minx,maxx,miny,maxy);
    imo[0].writePNG("imo-0.png");
    findBounds(im[1],H2,minx,maxx,miny,maxy);
    imo[1]=applyHomography(im[1],H2,minx,maxx,miny,maxy);
    imo[1].writePNG("imo-1.png");
    findBounds(im[2],H3,minx,maxx,miny,maxy);
    imo[2]=applyHomography(im[2],H3,minx,maxx,miny,maxy);
    imo[2].writePNG("imo-2.png");

   
    glutInit(&ac,av);
    glutInitDisplayMode(GLUT_ALPHA|GLUT_RGBA|GLUT_DOUBLE);
    glutInitWindowSize(ww,wh);
    glutCreateWindow("win");
    glutDisplayFunc(display);

     printf("done writing\n");
    glGenTextures(3,tex);
    for(int i=0;i<3;i++){
      printf("image %d initing texture %d\n",i,tex[i]);
      glBindTexture(GL_TEXTURE_2D,tex[i]);
      Image8 im;
      im=imo[i];
      im.initTexture();
    }
    printf("initing glut\n");

    glutMainLoop();
  }
}
