#ifdef USING_GLEE
#include "../GLee.h"
#endif
#include "dynatex.h"
#include <cshep2.h>
#include <tshep2.h>
#include <dsyevx.h>
#include <dgemm.h>
#include <dgesv.h>
#include <dsysv.h>

#include <ext/hash_map>
using namespace ::__gnu_cxx;
using namespace CapguiGlobals;
//old gcc
//#include <hash_map>

/* \brief THe windows dll needed to call GLeeInit, so
   this was a quick solution.
   
   Need to call glee_loader.init() before using any
   GL extensions.
*/
class MyPreloader {
public:
  bool success;
  MyPreloader(){
    success = false;
    init();
  }
  bool init(){
#ifdef USING_GLEE
    if(success)return success;
    success = GLeeInit();
    if(!success)
      printf("my preloader, initting glee failed: %d\n", success);
    return success;
#else
    return true;
#endif
  }
}glee_loader;

#ifdef WIN32
void mkstemp(char * c)
{
  char * c2;
  int attempt=0;
  int i=0;

  c2 = (char *)malloc(strlen(c)+1);

  while(attempt<10){
    strcpy(c2,c);
    while(i<strlen(c2)){
      if(c2[i]=='X'){
	c2[i]=(char)(rand()%26)+('A');
	if(rand()%2){
	  c2[i]=tolower(c2[i]);
	}
      }
      i++;
    }
    FILE * file=fopen(c2,"r");
    if(!file){
      break;
    }
    fclose(file);
    attempt++;
  }
  if(attempt<10){
    printf("Success for mkstemp with %s on attempt %d\n",c2,attempt);
    strcpy(c,c2);
  }
  free(c2);
}
#endif

int face[6][3]={{1,0,0},
		{-1,0,0},
		{0,1,0},
		{0,-1,0},
		{0,0,1},
		{0,0,-1}};

//Get rid of these
Vec3d * campts=0;
_3ints * camtri=0;
int ncamtri=0;
GLuint cubeTex[6];

void ppt(double * res,double * mat,Vec3f & v)
{
  double pt[4];
  pt[0]=v[0];
  pt[1]=v[1];
  pt[2]=v[2];
  pt[3]=1;
  multPoint(res,mat,pt);
  res[0]/=res[3];
  res[1]/=res[3];
}


void drawBorderNoTex(Chart & chart,int * border,int nborder,int max)
{
  for(int li=max;li>1;li--){
    //chart.drawTC();
    glLineWidth(li);
    glPointSize(li);
    glBegin(GL_POINTS);
    for(int i=0;i<nborder;i++){
      int v1=border[2*i];
      int v2=border[2*i+1];
      glVertex2fv(chart.tc[v1].data);
      glVertex2fv(chart.tc[v2].data);
    }
    glEnd();
    
    glBegin(GL_LINES);
    for(int i=0;i<nborder;i++){
      int v1=border[2*i];
      int v2=border[2*i+1];
      glVertex2fv(chart.tc[v1].data);
      glVertex2fv(chart.tc[v2].data);
    }
    glEnd();
  }
  glLineWidth(1);
  glPointSize(1);
}

void drawBorder(Vec3d * projPoints,
		Chart & chart,int * border,int nborder,int max)
{
  for(int li=max;li>1;li--){
    //chart.drawTC();
    glLineWidth(li);
    glPointSize(li);
    glBegin(GL_POINTS);
    for(int i=0;i<nborder;i++){
      int v1=border[2*i];
      int v2=border[2*i+1];
      glTexCoord2dv(projPoints[v1].data);
      glVertex3f(chart.tc[v1].x,
		 chart.tc[v1].y,
		 projPoints[v1].z);
      glTexCoord2dv(projPoints[v2].data);
      glVertex3f(chart.tc[v2].x,
		 chart.tc[v2].y,
		 projPoints[v2].z);
    }
    glEnd();
    
    glBegin(GL_LINES);
    for(int i=0;i<nborder;i++){
      int v1=border[2*i];
      int v2=border[2*i+1];
      glTexCoord2dv(projPoints[v1].data);
      glVertex3f(chart.tc[v1].x,
		 chart.tc[v1].y,
		 projPoints[v1].z);
      glTexCoord2dv(projPoints[v2].data);
      glVertex3f(chart.tc[v2].x,
		 chart.tc[v2].y,
		 projPoints[v2].z);
    }
    glEnd();
  }
  glLineWidth(1);
  glPointSize(1);
}

Image32 findNearest(Image8 & im,int off)
{
  Imagef distance(im.w,im.h,1);
  Image32 closest(im.w,im.h,1);
  distance=(double)im.w*im.h;
  closest=0;
  int di=0;
  int * queue = new int[im.w*im.h];
  int * inq = new int[im.w*im.h];
  int ninq=0;
  
  int * queue2 = new int[im.w*im.h];
  int * inq2 = new int[im.w*im.h];
  int ninq2=0;
  float root2=sqrt(2);
  int eightneigh=1;
  memset(queue,0,sizeof(int)*im.w*im.h);
  memset(inq,0xFF,sizeof(int)*im.w*im.h);
  memset(queue2,0,sizeof(int)*im.w*im.h);
  memset(inq2,0xFF,sizeof(int)*im.w*im.h);

  for(int i=0;i<im.h;i++){
    for(int j=0;j<im.w;j++){
      if(im(j,i)!=off){
	int ioff[]={0,0,1,-1,1,1,-1,-1};
	int joff[]={-1,1,0,0,-1,1,-1,1};
	float ds[]={1,1,1,1,root2,root2,root2,root2};
	for(int k=0;k<8;k++){
	  int i2=i+ioff[k];
	  int j2=j+joff[k];
	  if(i2>=0 && j2>=0 &&
	     i2<im.h && j2<im.w){
	    int index2=i2*im.w+j2;
	    if(ds[k]<distance(j2,i2)){
	      if(inq2[index2]==-1){
		inq2[index2]=ninq2;
		queue2[ninq2++]=index2;
	      }
	      closest(j2,i2)=i*im.w+j;
	      distance(j2,i2)=ds[k];
	    }
	  }
	  closest(j,i)=i*im.w+j;
	  distance(j,i)=0;
	}
      }
    }
  }
  di=1;
  while(ninq2){
    ninq=ninq2;
    ninq2=0;

    int * qback=queue;
    queue=queue2;    
    queue2=qback;
   
    int * inqback=inq;
    inq=inq2;
    inq2=inqback;

    ninq2=0;
    memset(inq2,0xFF,sizeof(int)*im.w*im.h);//this prbly isn't necessary but to be safe
    
    while(ninq){
      int index=queue[--ninq];
      int i=index/im.w;
      int j=index-im.w*i;
      int ioff[]={0,0,1,-1,1,1,-1,-1};
      int joff[]={-1,1,0,0,-1,1,-1,1};
      float ds[]={1,1,1,1,root2,root2,root2,root2};

      for(int k=0;k<8;k++){
	int i2=i+ioff[k];
	int j2=j+joff[k];
	if(i2>=0 && j2>=0 &&
	   i2<im.h && j2<im.w){
	  int index2=i2*im.w+j2;
	  if((distance(j,i)+ds[k])<distance(j2,i2) /*&& inq2[index2]==-1*/){
	    if(inq2[index2]==-1){
	      inq2[index2]=ninq2;
	      queue2[ninq2++]=index2;
	    }
	    closest(j2,i2)=closest(j,i);
	    distance(j2,i2)=ds[k]+distance(j,i);
	  }
	}
      }
      inq[index]=0;
    }
    printf("Distance %d, %d\n",di,ninq2);
    di++;
  }
  
  float maxdist=0;
  for(int i=0;i<im.w*im.h;i++){
    if(distance[i]>maxdist && distance[i]<im.w*im.h)maxdist=distance[i];
  }
  printf("maxdist %f\n",maxdist);
  Image8 dscale(im.w,im.h,1);
  for(int i=0;i<im.h;i++){
    for(int j=0;j<im.w;j++){
      dscale(j,i)=(unsigned char)(255.0f*(float)distance(j,i)/(float)maxdist);
    }
  }//dscale.writePPM("dscale.ppm");
  
  int nbad=0;
  for(int i=0;i<im.w*im.h;i++){
    if(distance[i]>=im.w*im.h){
      nbad++;
      closest[i]=0;
    }
  }
  if(nbad)
    printf("FIXME: nbad = %d (index out of bounds for map)\n",nbad);
  
  //return dscale;
  //return distance;
  
  return closest;
}

/*
  Create an openGL projection matrix with the given focal length
  and principle point.  This assumes rendering to an image/viewport
  of size imw,imh and depth range minz,maxz.

  If you write out the equations (assuming viewport of ww,wh):
  x-coord:
  (A11*X/Z+A12*Y/Z+A13*Z/Z+A14+1)*ww/2=a11*(X/Z)+a12*(Y/Z)+a13(Z/Z)
  Assuming A14 is zero, then you get:
  A11=a11*2/ww
  A12=a12*2/ww
  A13=a13*2/ww-1
  (similarily for y)

  The above equation ensures that a point (X,Y,Z) projects to the
  same coordinate using openGL as the typical K*eye(3,4) matrix would.
  
  If viewport is different, say (ww/2)*sc, then it should be scaled properly.
*/
void myProjection(double fx,double fy,
		  double px,double py,
		  int imw,int imh,
		  double minz,double maxz)
{
  if(!glee_loader.init())return;
  
  double mat[16];
  memset(mat,0,sizeof(mat));

  mat[0]=fx*2/imw;
  
  //Aspect
  //mat[4]=intrinsics[4]*2/imw;
  mat[5]=fy*2/imh;

  mat[8]=px*2/imw-1;
  mat[9]=py*2/imh-1;
  mat[10]=-(minz+maxz)/(minz-maxz);;
  mat[11]=1;
  
  mat[14]= 2*maxz*minz/(minz-maxz);
  
  glMultMatrixd(mat);
}
#include "../timer.h"
void getTexturesHW(double * intrinsics,
		   double * projection,
		   double * extrinsics,
		   Image8 * images,
		   Chart * charts,
		   int ncharts,
		   Vec3f & scale,
		   Vec3f & trans,
		   unsigned char * texImages,
		   unsigned char * occluded,
		   int texw,int texh,
		   int start,int end,
		   int nimages)
{
  if(!glee_loader.init())return;

  //Save the GL state
  glPushAttrib(GL_CURRENT_BIT|GL_ENABLE_BIT|GL_DEPTH_BUFFER_BIT);

  Timer totalTime;
  totalTime.start();

  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);

  /*
    Allocate some textures and display lists on processing of the first image
    so that we can save some time later on 
  */
  static GLuint camTex=0;
  static GLuint depthTex=0;
  static GLuint list1=0,list2=0,list2lines=0;
  if(start==0){
    //initialize the depth texture
    glGenTextures(1,&camTex);
    glGenTextures(1,&depthTex);
    glBindTexture(GL_TEXTURE_2D,depthTex);
    //Need to init the tex params for the depth tex (camTex gets inited by image.initTexture later)
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

    list1=glGenLists(1);
    list2=glGenLists(1);
    list2lines=glGenLists(1);
  }
  glColor3f(0,0,0);

  for(int cami=start;cami<end;cami++){
    printErr("Begin of loop\n");
    glBindTexture(GL_TEXTURE_2D,camTex);
    double scx=800.0/1024.0,scy=600.0/1024.0;
    images[cami].initTextureNoResize(&scx,&scy);
    
    //get depth range for this camera using the bounding box
    double minz=10e10;
    double maxz=-10e10;
    for(int x=0;x<=1;x++){
      for(int y=0;y<=1;y++){
	for(int z=0;z<=1;z++){
	  Vec3f pt(x,y,z);
	  pt=Vec3f(pt.x*scale.x,pt.y*scale.y,pt.z*scale.z);
	  pt+=trans;
	  double thisz=extrinsics[cami*16+2]*pt.x+
	    extrinsics[cami*16+6]*pt.y+
	    extrinsics[cami*16+10]*pt.z+
	    extrinsics[cami*16+14];
	  minz=MIN(minz,thisz);
	  maxz=MAX(maxz,thisz);
	}
      }
    }
    //printf("depth range %f %f\n",minz,maxz);

    //Setup the projection to be the same as the image
    //NOTE: this will fail if size is bigger than the window size!!!
    glViewport(0,0,images[cami].w,images[cami].h);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    myProjection(intrinsics[cami*16],
		 intrinsics[cami*16+5],
		 intrinsics[cami*16+12],
		 intrinsics[cami*16+13],
		 images[cami].w,images[cami].h,minz,maxz);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixd(extrinsics+cami*16);
    glTranslatef(trans.x,trans.y,trans.z);
    glScalef(scale.x,scale.y,scale.z);

    //compute the depth of the mesh in this camera and put it in a texture
    glClearColor(0,0,0,0);
    glClear(GL_DEPTH_BUFFER_BIT);
    glColorMask(0,0,0,0);
    glDepthMask(1);
    glEnable(GL_DEPTH_TEST);

    int vp[4];
    double proj[16];
    double model[16];
    glGetDoublev(GL_PROJECTION_MATRIX,proj);
    glGetDoublev(GL_MODELVIEW_MATRIX,model);
    glGetIntegerv(GL_VIEWPORT,vp);

    //NOTE: there seems to be a bug in the NVIDIA drivers.  When
    // the depth texture is 24 bits and the depth buffer is 24 bits
    // the offset should be 1.  But this is not the case for some reason.
    // The code below assumes 24bit depth buffer and 8 bit texture (hence
    // the offset is 1*(256*256)).  I am not using 24-bit textures because
    // the offset still had to be this large and that didn't seem right.  Now
    // with 8bit texture and 24bit depth buffer it at least makes sense to have
    // such a large offset.
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(0.0,32512.0f);

    //do the drawing here (matrices have been setup)
    if(cami==0){
      glNewList(list1,GL_COMPILE_AND_EXECUTE);
      glBegin(GL_TRIANGLES);
      for(int ci=0;ci<ncharts;ci++){
	for(int j=0;j<charts[ci].npoly;j++){
	  glVertex3fv(charts[ci].vert[charts[ci].poly[j][0]].data);
	  glVertex3fv(charts[ci].vert[charts[ci].poly[j][1]].data);
	  glVertex3fv(charts[ci].vert[charts[ci].poly[j][2]].data);
	}
      }
      glEnd();
      glEndList();
    }
    //call display list for every frame but the first
    else glCallList(list1);
    
    glPolygonOffset(0,0);
    glDisable(GL_POLYGON_OFFSET_FILL);
    glColorMask(1,1,1,1);
    
    int dtw=(int)round(((double)images[cami].w)/scx);
    int dth=(int)round(((double)images[cami].h)/scy);
    //int dtw=1;
    //int dth=1;
    //while(dtw<images[cami].width)dtw*=2;
    //while(dth<images[cami].height)dth*=2;

    //this won't work if images are different sizes
    glBindTexture(GL_TEXTURE_2D,depthTex);

    //Get the depth and put in a texture
    if(cami==0){
      int depthbits;
      glGetIntegerv(GL_DEPTH_BITS,&depthbits);
      printf("depthbits:%d\n",depthbits);
      glCopyTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT,0,0,dtw,dth,0);
      /*
      if(depthbits==8)
	glCopyTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT,0,0,dtw,dth,0);
      else if(depthbits==16)
	glCopyTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT16,0,0,dtw,dth,0);
      else
	glCopyTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT24,0,0,dtw,dth,0);
      */
    }
    else
      glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,images[cami].w,
			  images[cami].h);
    printErr("after getting depth texture\n");
    //Image8 d(images[cami].w,images[cami].h,1);
    //glReadPixels(0,0,d.w,d.h,GL_DEPTH_COMPONENT,GL_UNSIGNED_BYTE,d.data);
    //char fname[1024];
    //sprintf(fname,"/tmp/d%d.png",cami);
    //d.writePNG(fname);

    //draw the texture space warping the current image as the texture
    //and comparing the depth in the current image to the depth-texture
    //for visibility

    glScissor(0,0,texw,texh); //only draw in the texture area
    glEnable(GL_SCISSOR_TEST);
    glDisable(GL_DEPTH_TEST);

    glClear(GL_COLOR_BUFFER_BIT);

    //setup the projection to be orthographic 
    glViewport(0,0,texw,texh); 
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,1,0,1,-1,1);//-minz,-maxz);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    //activate texture 0 with the color image as the texture
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,camTex);

    //activate texture 1 with the depth texture
    glActiveTexture(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,depthTex);
 	
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_COMPARE_MODE_ARB,
		    GL_COMPARE_R_TO_TEXTURE);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
    glTexParameteri(GL_TEXTURE_2D,GL_DEPTH_TEXTURE_MODE_ARB,GL_LUMINANCE);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    //setup projective texture coordinate matrix for tex0 and tex1
    glActiveTexture(GL_TEXTURE0);
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();
    double center[16]={scx*0.5,0,0,0,
		       0,scy*0.5,0,0,
		       0,0,0.5,0,
		       scx*0.5,scy*0.5,0.5,1};
    glMultMatrixd(center);
    glMultMatrixd(proj);
    glMultMatrixd(model);

    glActiveTexture(GL_TEXTURE1);
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();
    glMultMatrixd(center);
    glMultMatrixd(proj);
    glMultMatrixd(model);
    
    glClear(GL_DEPTH_BUFFER_BIT);
    glColor3f(1,1,1);

    //draw the texture coordinates as verts with original verts as texture coordinates
    if(cami==0){
      glLineWidth(2);
      glPointSize(2);

      //Draw the border lines and points to aid crack problems
      glEnable(GL_DEPTH_TEST);
      glNewList(list2lines,GL_COMPILE_AND_EXECUTE);
      for(int ci=0;ci<ncharts;ci++){
	int * border=0;
	int nborder=charts[ci].getBorder(&border);
	glBegin(GL_POINTS);
	for(int i=0;i<nborder;i++){
	  int i1=border[2*i];
	  int i2=border[2*i+1];
	  Vec2f & tc0=charts[ci].tc[i1];
	  Vec2f & tc1=charts[ci].tc[i2];

	  Vec3f & v0=charts[ci].vert[i1];
	  Vec3f & v1=charts[ci].vert[i2];
	  	  
	  glMultiTexCoord3fv(GL_TEXTURE1,v0.data);
	  glTexCoord3fv(v0.data);
	  glVertex3f(tc0.x,tc0.y,0.5);

	  glMultiTexCoord3fv(GL_TEXTURE1,v1.data);
	  glTexCoord3fv(v1.data);
	  glVertex3f(tc1.x,tc1.y,0.5);
	}
	glEnd();
	glBegin(GL_LINES);
	for(int i=0;i<nborder;i++){
	  int i1=border[2*i];
	  int i2=border[2*i+1];
	  Vec2f & tc0=charts[ci].tc[i1];
	  Vec2f & tc1=charts[ci].tc[i2];

	  Vec3f & v0=charts[ci].vert[i1];
	  Vec3f & v1=charts[ci].vert[i2];
	  	  
	  glMultiTexCoord3fv(GL_TEXTURE1,v0.data);
	  glTexCoord3fv(v0.data);
	  glVertex3f(tc0.x,tc0.y,0.5);

	  glMultiTexCoord3fv(GL_TEXTURE1,v1.data);
	  glTexCoord3fv(v1.data);
	  glVertex3f(tc1.x,tc1.y,0.5);
	}
	glEnd();
	delete [] border;
      }
      glEndList();

      //draw several widths
      for(int ps=4;ps<=6;ps+=2){
	glLineWidth(ps);
	glPointSize(ps);
	glCallList(list2lines);
      }
      glLineWidth(1);
      glPointSize(1);

      //Now draw all the triangles
      glEnable(GL_DEPTH_TEST);
      glNewList(list2,GL_COMPILE_AND_EXECUTE);
            
      glBegin(GL_TRIANGLES);
      for(int ci=0;ci<ncharts;ci++){
	for(int j=0;j<charts[ci].npoly;j++){
	  Vec2f & tc0=charts[ci].tc[charts[ci].poly[j][0]];
	  Vec2f & tc1=charts[ci].tc[charts[ci].poly[j][1]];
	  Vec2f & tc2=charts[ci].tc[charts[ci].poly[j][2]];
	  
	  Vec3f & v0=charts[ci].vert[charts[ci].poly[j][0]];
	  Vec3f & v1=charts[ci].vert[charts[ci].poly[j][1]];
	  Vec3f & v2=charts[ci].vert[charts[ci].poly[j][2]];

	  glMultiTexCoord3fv(GL_TEXTURE1,v0.data);
	  glTexCoord3fv(v0.data);
	  glVertex3f(tc0.x,tc0.y,0.5);
	  glMultiTexCoord3fv(GL_TEXTURE1,v1.data);
	  glTexCoord3fv(v1.data);
	  glVertex3f(tc1.x,tc1.y,0.5);
	  glMultiTexCoord3fv(GL_TEXTURE1,v2.data);
	  glTexCoord3fv(v2.data);
	  glVertex3f(tc2.x,tc2.y,0.5);
	}
      }
      glEnd();
      glEndList();
      
      glDisable(GL_DEPTH_TEST);
      glCallList(list2);
    }
    //Do the drawing with the display lists on cami!=0
    else{
      glEnable(GL_DEPTH_TEST);
      
      for(int ps=2;ps<=6;ps+=2){
	glLineWidth(ps);
	glPointSize(ps);
	glCallList(list2lines);
      }
      glPointSize(1);
      glLineWidth(1);

      glDisable(GL_DEPTH_TEST);
      glCallList(list2);
    }
    //POp the texture matrix stacks
    glActiveTexture(GL_TEXTURE1);
    glMatrixMode(GL_TEXTURE);
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);
    
    glActiveTexture(GL_TEXTURE0);
    glMatrixMode(GL_TEXTURE);
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);

    glDisable(GL_SCISSOR_TEST);

    //Image8 texture(texw,texh,3);
    //glReadPixels(0,0,texw,texh,GL_RGB,
    //		 GL_UNSIGNED_BYTE,texture.data);
    //texture=texture.map(nearest);
    //memcpy(texImages+curi*texw*texh*3,texture.data,texw*texh*3);
    Image8 im(texw,texh,4);

    //Should probably use alpha to set "occluded", but alpha_bits seems to
    //be 0 with wxWidgets all the time, so occluded is 1 when color==(0,0,0)
    int alphabits=0;
    glGetIntegerv(GL_ALPHA_BITS,&alphabits);

    //Read back the warped texture
    glReadPixels(0,0,texw,texh,im.channelToGL(),im.typeToGL(),im.data);
    int occ=0;
    for(int y=0;y<im.h;y++){
      for(int x=0;x<im.w;x++){
	for(int k=0;k<3;k++)
	  texImages[cami*texw*texh*3+y*texw*3+x*3+k]=im(x,y,k);
	occluded[cami*texw*texh+y*texw+x]=
	  (im(x,y,0)==0 && im(x,y,1)==0 && im(x,y,2)==0)?255:0;
	occ+=occluded[cami*texw*texh+y*texw+x];
      }
    }
    //printf("occluded is %d\n",occ);
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
  }
  //When we are all done then delete all the stuff we allocated
  if(end==nimages){
    glDeleteTextures(1,&camTex);
    glDeleteTextures(1,&depthTex);
    glDeleteLists(list1,1);
    glDeleteLists(list2,1);
    glDeleteLists(list2lines,1);
    camTex=depthTex=0;
  }
  //printf("total time is %f\n",totalTime.stop());
  glPopAttrib();
}

//NOTE: texw and texh are the dimensions of rgb images in texImages
//      and should be the max texture size over the # of channels
void getTextures(double * intrinsics,
		 double * projection,
		 double * extrinsics,
		 Image8 * images,
		 Chart * charts,
		 int ncharts,
		 Vec3f & scale,
		 Vec3f & trans,
		 unsigned char * texImages,
		 unsigned char * occluded,
		 int texw,int texh,
		 int start,int end)
{
  if(!glee_loader.init())return;

  static Image32 nearest;
  double neard= 0;
  double fard=  1.1;
  int vp[4];
  
  glClearColor(0.0,0.0,0.0,0.0);
  glGetIntegerv(GL_VIEWPORT,vp);
  printf("Get textures %d to %d (%d %d)\n",start,end,texw,texh);

  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  GLuint camTex=0;
  GLuint depthTex=0;

  //initialize the depth texture
  glEnable(GL_TEXTURE_2D);
  glGenTextures(1,&camTex);
  glGenTextures(1,&depthTex);
  glDisable(GL_TEXTURE_2D);

  if(start==0){
    glViewport(0,0,texw,texh);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,1,0,1,-1,1);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
   
    glDisable(GL_LIGHTING);
    glColor3f(1.0,1.0,1.0);
    /*
    for(int i=0;i<ncharts;i++){
      int * border=0;
      int nborder=charts[i].getBorder(&border);
      drawBorderNoTex(charts[i],border,nborder,10);
      if(border && nborder)delete [] border;
    }
    */

    glBegin(GL_TRIANGLES);
    for(int i=0;i<ncharts;i++)
      for(int j=0;j<charts[i].npoly;j++){	
	glVertex2fv(charts[i].tc[charts[i].poly[j][0]].data);
	glVertex2fv(charts[i].tc[charts[i].poly[j][1]].data);
	glVertex2fv(charts[i].tc[charts[i].poly[j][2]].data);
      }
    glEnd();

    glFlush();
    Image8 red(texw,texh,1);
    glReadPixels(0,0,texw,texh,GL_RED,red.typeToGL(),red.data);
    //red.writePPM("red.ppm");
    nearest=findNearest(red,0);
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
  }
  
  for(int curi=start;curi<end;curi++){
    int curw=images[curi].width;
    int curh=images[curi].height;
   
    glBindTexture(GL_TEXTURE_2D,camTex);
    {
      Image8 im8;
      im8=images[curi].subimage(0,0,images[curi].width,
				images[curi].height);
      im8.initTexture();
    }

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glViewport(0,0,texw,texh);
    glOrtho(0,curw-1,0,curh-1,neard,fard);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    
    glDisable(GL_CULL_FACE);
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    
    double mat[16];
    glLoadIdentity();
    glMultMatrixd(intrinsics+curi*16);
    glMultMatrixd(projection);
    glMultMatrixd(extrinsics+curi*16);
    
    glTranslatef(trans.x,trans.y,trans.z);
    glScalef(scale.x,scale.y,scale.z);
        
    glGetDoublev(GL_MODELVIEW_MATRIX,mat);
    glLoadIdentity();

    if(ncharts){
      double pt[4],res[4];
      pt[3]=1;
      glDisable(GL_LIGHTING);
      glDisable(GL_BLEND);
      
      glDisable(GL_TEXTURE_2D);

      Vec3d ** projPoints=new Vec3d*[ncharts];
      
      int nearSet=0;
      int farSet=0;
      for(int i=0;i<ncharts;i++){
	projPoints[i]=new Vec3d[charts[i].nvert];
	
	glColor3f(i/(double)(ncharts),1.0-(i/(double)ncharts),0.5);
	for(int j=0;j<charts[i].npoly;j++){
	  double pt0[4];
	  double pt1[4];
	  double pt2[4];
	  
	  ppt(pt0,mat,charts[i].vert[charts[i].poly[j][0]]);
	  if(!nearSet || pt0[2]<neard){
	    nearSet=1;
	    neard=pt0[2];
	  }
	  if(!farSet || pt0[2]>fard){
	    farSet=1;
	    fard=pt0[2];
	  }
	  ppt(pt1,mat,charts[i].vert[charts[i].poly[j][1]]);
	  if(pt1[2]<neard)neard=pt1[2];
	  if(pt1[2]>fard)fard=pt1[2];
	  ppt(pt2,mat,charts[i].vert[charts[i].poly[j][2]]);
	  if(pt2[2]<neard)neard=pt2[2];
	  if(pt2[2]>fard)fard=pt2[2];
	  
	  projPoints[i][charts[i].poly[j][0]]=Vec3d(pt0[0],pt0[1],pt0[2]);
	  projPoints[i][charts[i].poly[j][1]]=Vec3d(pt1[0],pt1[1],pt1[2]);
	  projPoints[i][charts[i].poly[j][2]]=Vec3d(pt2[0],pt2[1],pt2[2]);
	}
      }
      neard*=-1;
      fard*=-1;
      printf("neard %f, fard %f\n",neard,fard);
      printErr("projected points\n");

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0,curw-1,0,curh-1,neard,fard);//here was texw texh
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      
      glEnable(GL_DEPTH_TEST);
      glBegin(GL_TRIANGLES);
      for(int i=0;i<ncharts;i++){
	for(int j=0;j<charts[i].npoly;j++){

	  float col;
	  col=(projPoints[i][charts[i].poly[j][0]].z-(-neard))/(-fard-(-neard));
	  glColor3f(col,col,col);
	  //glColor3f(1.0,0.0,0.0);
	  glVertex3dv(projPoints[i][charts[i].poly[j][0]].data);
	  col=(projPoints[i][charts[i].poly[j][1]].z-(-neard))/(-fard-(-neard));
	  glColor3f(col,col,col);
	  //glColor3f(1.0,0.0,0.0);
	  glVertex3dv(projPoints[i][charts[i].poly[j][1]].data);
	  col=(projPoints[i][charts[i].poly[j][2]].z-(-neard))/(-fard-(-neard));
	  glColor3f(col,col,col);
	  //glColor3f(1.0,0.0,0.0);
	  glVertex3dv(projPoints[i][charts[i].poly[j][2]].data);
	}
      }
      glEnd();

      Image8 depthImage(texw,texh,1);
      glReadPixels(0,0,texw,texh,
		   GL_RED,depthImage.typeToGL(),depthImage.data);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,depthTex);
      depthImage.initTexture();

      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
      glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);    
      printErr("after getting depth\n");
            
      float vp[]={0.5,0,0,0,
		  0,0.5,0,0,
		  0,0,0.5,0,
		  0.5,0.5,0.5,1.0};
      
      glMatrixMode(GL_TEXTURE);
      glPushMatrix();
      glLoadIdentity();
      glMultMatrixf(vp);
      glOrtho(0,curw-1,0,curh-1,neard,fard);//fixme:Image width and height
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0,1.0,0,1.0,-1,1);
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      
      glClearColor(1.0,1.0,1.0,0.0);
      glClear(GL_COLOR_BUFFER_BIT);
      glDisable(GL_DEPTH_TEST);

      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,depthTex);

      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glBegin(GL_TRIANGLES);
      for(int i=0;i<ncharts;i++){
	glColor3f(1.0,1.0,1.0);
	for(int j=0;j<charts[i].npoly;j++){
	  glTexCoord2dv(projPoints[i][charts[i].poly[j][0]].data);
	  glVertex2f(charts[i].tc[charts[i].poly[j][0]].x,
		     charts[i].tc[charts[i].poly[j][0]].y);
	  glTexCoord2dv(projPoints[i][charts[i].poly[j][1]].data);
	  glVertex2f(charts[i].tc[charts[i].poly[j][1]].x,
		     charts[i].tc[charts[i].poly[j][1]].y);
	  glTexCoord2dv(projPoints[i][charts[i].poly[j][2]].data);
	  glVertex2f(charts[i].tc[charts[i].poly[j][2]].x,
		     charts[i].tc[charts[i].poly[j][2]].y);
	}
      }
      glEnd();
      glDisable(GL_TEXTURE_2D);
      printErr("after drawing warped depth\n");
      
      Image8 warpedDepth(texw,texh,1);
      glRasterPos2i(0,0);
      glReadPixels(0,0,texw,texh,
		   GL_RED,warpedDepth.typeToGL(),warpedDepth.data);
      //Image32 nearest;
      //nearest=findNearest(warpedDepth,255);
      //warpedDepth.writeTGA("warpedDepth.tga");
      //m.writePPM("dist.ppm");
      
      warpedDepth=warpedDepth.map(nearest);
      //warpedDepth.writePPM("warpedDepth.ppm");
      printErr("after initing warpedDepth\n");
      
      glClearStencil(0);
      glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
      glRasterPos3f(0,0,0);
      //glDrawPixels(texw,texh,GL_DEPTH_COMPONENT,GL_FLOAT,warpedDepth.data);

      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,camTex);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0,1,0,1,neard,fard);
      glMatrixMode(GL_MODELVIEW);

      glDisable(GL_DEPTH_TEST);
      glDisable(GL_TEXTURE_2D);//new
      glBegin(GL_TRIANGLES);
      for(int i=0;i<ncharts;i++){
	glColor3f(1.0,1.0,1.0);
	for(int j=0;j<charts[i].npoly;j++){
	  float col;
	  col=(projPoints[i][charts[i].poly[j][0]].z-(-neard))/(-fard-(-neard));
	  glColor3f(col,col,col);
	  //glTexCoord2dv(projPoints[i][charts[i].poly[j][0]].data);
	  glVertex3f(charts[i].tc[charts[i].poly[j][0]].x,
		     charts[i].tc[charts[i].poly[j][0]].y,
		     projPoints[i][charts[i].poly[j][0]].z);
	  col=(projPoints[i][charts[i].poly[j][1]].z-(-neard))/(-fard-(-neard));
	  //glTexCoord2dv(projPoints[i][charts[i].poly[j][1]].data);
	  glColor3f(col,col,col);
	  glVertex3f(charts[i].tc[charts[i].poly[j][1]].x,
		     charts[i].tc[charts[i].poly[j][1]].y,
		     projPoints[i][charts[i].poly[j][1]].z);
	  col=(projPoints[i][charts[i].poly[j][2]].z-(-neard))/(-fard-(-neard));
	  //glTexCoord2dv(projPoints[i][charts[i].poly[j][2]].data);
	  glColor3f(col,col,col);
	  glVertex3f(charts[i].tc[charts[i].poly[j][2]].x,
		     charts[i].tc[charts[i].poly[j][2]].y,
		     projPoints[i][charts[i].poly[j][2]].z);
	}
      }

      glEnd();
      glDisable(GL_STENCIL_TEST);
      printErr("after drawing texture\n");

      Image8 texdepth(texw,texh,1);
      glReadPixels(0,0,texw,texh,GL_RED,
		   texdepth.typeToGL(),texdepth.data);
      texdepth=texdepth.map(nearest);
     
      Image8 diff=(texdepth>(warpedDepth+10));
      //diff.clean(3,0);
      //diff.clean(3,1);
      diff*=255;
      //images[curi].writePPM("theimage.ppm");
      //diff.writePPM("occluded.ppm");
      memcpy(occluded+curi*texw*texh,diff.data,texw*texh);

      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,camTex);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
      glDisable(GL_DEPTH_TEST);
      
      for(int i=0;i<ncharts;i++){
	int * border=0;
	int nborder=charts[i].getBorder(&border);
	drawBorder(projPoints[i],charts[i],border,nborder,7);
	if(border && nborder)delete [] border;
      }
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glBegin(GL_TRIANGLES);
      for(int i=0;i<ncharts;i++){
	glColor3f(1.0,1.0,1.0);
	for(int j=0;j<charts[i].npoly;j++){
	  glTexCoord2dv(projPoints[i][charts[i].poly[j][0]].data);
	  glVertex3f(charts[i].tc[charts[i].poly[j][0]].x,
		     charts[i].tc[charts[i].poly[j][0]].y,
		     projPoints[i][charts[i].poly[j][0]].z);
	  glTexCoord2dv(projPoints[i][charts[i].poly[j][1]].data);
	  glVertex3f(charts[i].tc[charts[i].poly[j][1]].x,
		     charts[i].tc[charts[i].poly[j][1]].y,
		     projPoints[i][charts[i].poly[j][1]].z);
	  glTexCoord2dv(projPoints[i][charts[i].poly[j][2]].data);
	  glVertex3f(charts[i].tc[charts[i].poly[j][2]].x,
		     charts[i].tc[charts[i].poly[j][2]].y,
		     projPoints[i][charts[i].poly[j][2]].z);
	}
      }
      glEnd();

      Image8 texture(texw,texh,3);
      glReadPixels(0,0,texw,texh,GL_RGB,
		   GL_UNSIGNED_BYTE,texture.data);
      //texture=texture.map(nearest);
      memcpy(texImages+curi*texw*texh*3,texture.data,texw*texh*3);
      //texture.writePPM("texture.ppm");
      glMatrixMode(GL_TEXTURE);
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
      printErr("popped matrix\n");
      
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,camTex);
      glViewport(0,texh,texw,texh);
      
      glDisable(GL_DEPTH_TEST);
      glBegin(GL_QUADS);
      glTexCoord2f(0,0);glVertex2f(0,0);
      glTexCoord2f(0,1.0);glVertex2f(0,1);
      glTexCoord2f(1.0,1.0);glVertex2f(1,1);
      glTexCoord2f(1.0,0);glVertex2f(1,0);
      glEnd();
      glEnable(GL_DEPTH_TEST);

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0,curw-1,0,curh-1,neard,fard);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      printErr("after setting ortho\n");

      glEnable(GL_DEPTH_TEST);
      
      glEnable(GL_CULL_FACE);
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      glDisable(GL_TEXTURE_2D);

      glBegin(GL_TRIANGLES);
      for(int i=0;i<ncharts;i++){
	glColor3f(1.0,1.0,1.0);
	for(int j=0;j<charts[i].npoly;j++){
	  glVertex2dv(projPoints[i][charts[i].poly[j][0]].data);
	  glVertex2dv(projPoints[i][charts[i].poly[j][1]].data);
	  glVertex2dv(projPoints[i][charts[i].poly[j][2]].data);
	}
      }
      glEnd();
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glDisable(GL_CULL_FACE);

      glActiveTexture(GL_TEXTURE0);
      glDisable(GL_TEXTURE_GEN_S);
      glDisable(GL_TEXTURE_GEN_T);
      glDisable(GL_TEXTURE_GEN_R);
      glDisable(GL_TEXTURE_GEN_Q);
      glDisable(GL_TEXTURE_2D);

      for(int i=0;i<ncharts;i++){
	delete [] projPoints[i];
      }
      delete [] projPoints;
    }

    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }
  glViewport(vp[0],vp[1],vp[2],vp[3]);
  glDeleteTextures(1,&camTex);
}



void initDynaTexYUV(char * path,
		    Chart * charts,
		    int ncharts,Vec3f & scale,Vec3f & trans,
		    double * extrinsics,Matrix & X,
		    Matrix & Y,Matrix & U,Matrix & V,
		    int * nreigs,int * widths,int * heights,
		    int lutw,int luth,int flags)
{
  if(!glee_loader.init())return;

  int neigy=nreigs[0];
  int neigu=nreigs[1];
  int neigv=nreigs[2];

  if(neigy<0)
    neigy=31;
  if(neigu<0)
    neigu=15;
  if(neigv<0)
    neigv=15;
  if(neigy>=Y.n)neigy=Y.n-1;
  if(neigu>=U.n)neigu=U.n-1;
  if(neigv>=V.n)neigv=V.n-1;
  
  neigy=(neigy+1)-(neigy+1)%4-1;
  neigu=(neigu+1)-(neigu+1)%4-1;
  neigv=(neigv+1)-(neigv+1)%4-1;

  printf("in init dyna tex %d %d %d\n",neigy,neigu,neigv);
  
  Matrix meanY,meanU,meanV,YY,YU,YV;
  Matrix EigY,EigU,EigV;
  Matrix * gridY, * gridU, * gridV;

  printf("doing pca %d %d keeping %d\n",Y.m,Y.n,neigy);
  printf("doing pca %d %d keeping %d\n",U.m,U.n,neigu);
  printf("doing pca %d %d keeping %d\n",V.m,V.n,neigv);
  
  pca(Y,neigy,meanY,EigY,YY);
  pca(U,neigu,meanU,EigU,YU);
  pca(V,neigv,meanV,EigV,YV);

  printf("done pca's\n");

  int cols[]={0,1};
  int gridres[]={lutw,luth};

  Matrix X12=X.getCols(cols,2);
  
  char basepath[1024];
  char fname[1024];
  sprintf(basepath,path);
  //sprintf(basepath,"/usr/scratch/birkbeck/sfs/data0");
  //sprintf(basepath,"./data0");

  if(!(flags&LUT_CUBE)){
    printf("building grids\n");
    gridY=buildgrid(gridres,X12,YY,flags);
    gridU=buildgrid(gridres,X12,YU,flags);
    gridV=buildgrid(gridres,X12,YV,flags);
    printf("done building grids\n");

    //remove the cube grids if any
    sprintf(fname,"%s/%s",basepath,"cubeGridY");remove(fname);
    sprintf(fname,"%s/%s",basepath,"cubeGridU");remove(fname);
    sprintf(fname,"%s/%s",basepath,"cubeGridV");remove(fname);

    sprintf(fname,"%s/%s",basepath,"gridY");
    gridwrite(fname,gridY,neigy);
    
    sprintf(fname,"%s/%s",basepath,"gridU");
    gridwrite(fname,gridU,neigu);
    
    sprintf(fname,"%s/%s",basepath,"gridV");
    gridwrite(fname,gridV,neigv);
  }
  else{
    //here
    int dim=lutw;
    if(luth<lutw)dim=luth;
    Vec3d * cameraCenters=getCameraCenters(extrinsics,X,scale,trans);
    double * coeffs = getCubeCoeffs(cameraCenters,X.m,dim,flags);
    Matrix *cubeGridY=buildCubeGrid(coeffs,YY,dim);
    Matrix *cubeGridU=buildCubeGrid(coeffs,YU,dim);
    Matrix *cubeGridV=buildCubeGrid(coeffs,YV,dim);
  
    /*
      for(int face=0;face<6;face++){
      printf("face%d=[...\n",face);
      for(int y=0;y<dim;y++){
      for(int x=0;x<dim;x++){
      printf("%f ",coeffs[face*dim*dim*X.m+
      y*dim*X.m+
      x*X.m]);
      }
      printf("\n");
      }
      printf("]\n");
      }*/
    delete coeffs;
    delete cameraCenters;

    sprintf(fname,"%s/%s",basepath,"cubeGridY");
    gridwriteCube(fname,cubeGridY,neigy);
    
    sprintf(fname,"%s/%s",basepath,"cubeGridU");
    gridwriteCube(fname,cubeGridU,neigu);
    
    sprintf(fname,"%s/%s",basepath,"cubeGridV");
    gridwriteCube(fname,cubeGridV,neigv);
  }
  
  sprintf(fname,"%s/%s",basepath,"geometry.obj");

  //X(:,[3 1 2])*180/pi);
  Matrix X2(X.m,X.n);
  for(int i=0;i<X.m;i++){
    X2(i,0)=X(i,2);
    X2(i,1)=X(i,0);
    X2(i,2)=X(i,1);
  }
  if(flags&LUT_CUBE){
    Matrix angles;
    //Object is always upside down...this
    //is just a temporary fix
    /*for(int i=0;i<ncharts;i++){
      for(int j=0;j<charts[i].nvert;j++){
	charts[i].vert[j].rotatex(M_PI);
      }
      }*/
    //FIXME: this will be broken, since the save no longer
    //just flips about x..so the cube coefficients need to be generated
    //differently...(is this fixed?)
    saveobj(fname,charts,ncharts,angles,
	    Vec3f(1,-1,1)*scale,
	    Vec3f(-0.5,-0.5,-0.5),flags);
    /*for(int i=0;i<ncharts;i++){
      for(int j=0;j<charts[i].nvert;j++){
	charts[i].vert[j].rotatex(M_PI);
      }
      }*/
  }
  else{
    saveobj(fname,
	    charts,ncharts,X2,
	    Vec3f(1,-1,1)*scale,
	    Vec3f(-0.5,-0.5,-0.5),flags);
  }
  sprintf(fname,"%s/%s",basepath,"dataY");
  eigwrite(fname,meanY,EigY,widths[0],heights[0]);

  sprintf(fname,"%s/%s",basepath,"dataU");
  eigwrite(fname,meanU,EigU,widths[1],heights[1]);

  sprintf(fname,"%s/%s",basepath,"dataV");
  eigwrite(fname,meanV,EigV,widths[2],heights[2]);
}
  
void initDynaTexRGB(char * path,
		    Chart * charts,
		    int ncharts,Vec3f & scale,Vec3f & trans,
		    double * extrinsics,
		    Matrix & X,Matrix & reds,
		    Matrix & greens,
		    Matrix & blues,
		    int * neig,int * widths,int * heights,
		    int lutw,int luth,int flags)
{
  if(!glee_loader.init())return;

  printf("in initDynaTexRGB\n");
  printf("before conversion to yuv\n");
  printf("red: %d %d\n",reds.m,reds.n);
  printf("green: %d %d\n",greens.m,greens.n);
  printf("blue: %d %d\n",blues.m,blues.n);
  rgb2yuv(reds,greens,blues);
  
  resize(greens,widths[0],heights[0],widths[1],heights[1]);
  resize(blues,widths[0],heights[0],widths[2],heights[1]);
  //resize(reds,greens,blues);

  printf("converted to yuv\n");
  printf("red: %d %d\n",reds.m,reds.n);
  printf("green: %d %d\n",greens.m,greens.n);
  printf("blue: %d %d\n",blues.m,blues.n);
  initDynaTexYUV(path,charts,ncharts,scale,trans,
		 extrinsics,X,reds,greens,blues,neig,widths,heights,lutw,luth,flags);
}

//frees images and occluded when it is done with them
void initDynaTex(char * path,Chart * charts,
		 int ncharts,Vec3f & scale,Vec3f & trans,
		 double * extrinsics,
		 unsigned char *images,
		 unsigned char * occluded,
		 int w,
		 int h,
		 int nimages,int * nreigs,
		 int * widths,
		 int * heights,
		 int lutw,int luth,int flags)
{
  if(!glee_loader.init())return;

  printf("in initDynaTex %d %d %d\n",w,h,nimages);
  Matrix X(nimages,3);//euler parametrs for each image
  Matrix rs(w*h,nimages);
  Matrix gs(w*h,nimages);
  Matrix bs(w*h,nimages);
  printf("allocated memory for X,rs,gs,bs\n");
  //nest so the following images get deleted before continuing
  {
    Matrix meanr(w*h,1);
    Matrix meang(w*h,1);
    Matrix meanb(w*h,1);
    Matrix occmeanr(w*h,1);
    Matrix occmeang(w*h,1);
    Matrix occmeanb(w*h,1);
    printf("alloced memory for mean[rgb] occmean[rgb], viscnt, occcnt\n");

    int * viscnt=new int[w*h];
    int * occcnt=new int[w*h];
    printf("alloced memory for viscnt, occnt\n");

    
    for(int i=0;i<nimages;i++){
      Matrix ext(4,4);
      memcpy(ext.data,extrinsics+i*16,sizeof(double)*16);
      ext=ext.transpose();//extrinsics are column major, we need row major
      
      /*if(flags&LUT_LINEAR){
	Matrix eul=tr2yx(ext);
	//Leave the x-rotation so there is something for
	//the mean
	X(i,0)=i/100.0 ;//eul[0]*180.0/M_PI;
	X(i,0)=atan2(ext(2,2),ext(2,0))*180.0/M_PI;
	X(i,2)=eul[2]*180.0/M_PI;
	}
	else{*/
      //After this call, each row of the matrix X contains
      //The rotation angles y,x,z to be interpreted as
      // p'=t*rot(x)*rot(y)*p
      // the z rotation (the third column) is zero
      Matrix eul=tr2yx(ext);
      X(i,0)=-1*eul[0]*180.0/M_PI+180;
      //it seems that the following  should not be multiplied by -1
      //but it was done to be the same as the house
      //This is possibly due to the default rotation axis in the
      //renderer being about -y
      X(i,1)=-1*(eul[1]*180.0/M_PI+180); 
      X(i,2)=eul[2]*180.0/M_PI;
    }
    meanr.setAll(0);
    meang.setAll(0);
    meanb.setAll(0);
    occmeanr.setAll(0);
    occmeanb.setAll(0);
    occmeang.setAll(0);
    memset(viscnt,0,sizeof(int)*w*h);
    memset(occcnt,0,sizeof(int)*w*h);
    
    printf("building double matrix of images\n");
    for(int i=0;i<h;i++){
      for(int j=0;j<w;j++){
	int ind=i+j*h;//columns first
	for(int k=0;k<nimages;k++){
	  rs(ind,k)=(double)(images[k*w*h*3+ind*3  ]/255.0);
	  gs(ind,k)=(double)(images[k*w*h*3+ind*3+1]/255.0);
	  bs(ind,k)=(double)(images[k*w*h*3+ind*3+2]/255.0);
	  if(!occluded[k*w*h+ind]){
	    meanr(ind,0)+=rs(ind,k);
	    meang(ind,0)+=gs(ind,k);
	    meanb(ind,0)+=bs(ind,k);
	    viscnt[ind]++;
	  }
	  else{
	    occmeanr(ind,0)+=rs(ind,k);
	    occmeang(ind,0)+=gs(ind,k);
	    occmeanb(ind,0)+=bs(ind,k);
	    occcnt[ind]++;
	  }
	}
      }
    }
    
    for(int i=0;i<w*h;i++){
      if(viscnt[i]){
	meanr(i,0)/=viscnt[i];
	meang(i,0)/=viscnt[i];
	meanb(i,0)/=viscnt[i];
	for(int j=0;j<nimages;j++){
	  if(occluded[j*w*h+i])
	  {
	    rs(i,j)=meanr(i,0);
	    gs(i,j)=meang(i,0);
	    bs(i,j)=meanb(i,0);
	    //gs(i,j)=1;
	    //rs(i,j)=1;
	    //bs(i,j)=0;
	  }
	  //Show the occluded parts as black
	  /*
	    rs(i,j)*=(occluded[j*w*h+i]==0);//255.0;
	    gs(i,j)*=(occluded[j*w*h+i]==0);//255.0;
	    bs(i,j)*=(occluded[j*w*h+i]==0);//255.0;*/
	}
      }
      else{
	//These pixels were never visible in any image
	occmeanr(i,0)/=occcnt[i];
	occmeang(i,0)/=occcnt[i];
	occmeanb(i,0)/=occcnt[i];
	for(int j=0;j<nimages;j++){
	  rs(i,j)=occmeanr(i,0);
	  gs(i,j)=occmeang(i,0);
	  bs(i,j)=occmeanb(i,0);
	  /*gs(i,j)=0;
	    rs(i,j)=0;
	    bs(i,j)=1;*/
	  
	}
      }
    }
    delete [] viscnt;
    delete [] occcnt;
  }
  //Compute static texture
  {
    printf("gettint the static texture\n");
    Imagef staticTex(w,h,3);
    staticTex=0;
    int values[3][nimages];
    for(int i=0;i<h;i++){
      for(int j=0;j<w;j++){
	int nvals=0;
	for(int imi=0;imi<nimages;imi++){
	  if(!occluded[imi*w*h+i*w+j]){
	    for(int k=0;k<3;k++){
	      int val=images[imi*3*w*h+i*3*w+j*3+k];
	      values[k][nvals]=val;
	    }
	    nvals++;
	  }
	}
	  
	for(int ii=0;ii<nvals-1;ii++){
	  int smallest=ii;
	  int smallestLen=values[0][ii]*values[0][ii]+
	    values[1][ii]*values[1][ii]+
	    values[2][ii]*values[2][ii];
	  for(int jj=ii+1;jj<nvals;jj++){
	    int thisLen=values[0][jj]*values[0][jj]+
	      values[1][jj]*values[1][jj]+
	      values[2][jj]*values[2][jj];
	    if(thisLen<smallestLen){
	      smallest=jj;
	      smallestLen=thisLen;
	    }
	  }
	  for(int k=0;k<3;k++){
	    int back=values[k][smallest];
	    values[k][smallest]=values[k][ii];
	    values[k][ii]=back;
	  }
	  //printf("%3.0f ",values[ii]);
	}
	//its just the mean now.
	float mean[3];
	for(int k=0;k<3;k++){
	  mean[k]=0;
	  for(int h=0;h<nvals;h++){
	    mean[k]+=values[k][h];
	  }
	  if(nvals)
	    mean[k]/=nvals;
	  staticTex(j,i,k)=mean[k];
	}
       
	//int median=nvals/2;
	//for(int k=0;k<3;k++){
	//staticTex(j,i,k)=values[k][median];
	//}
      }
    }
    
    char fname[256];
    staticTex=staticTex*(1.0/255.0);
    sprintf(fname,"%s/static_texture.ppm",path);
    staticTex.writePPM(fname);
  }
  //End compute static texture


  //these are not needed anymore
  delete []images;
  delete []occluded;

  /*  Image8 * image=new Image8(3);
      image->alloc(w,h,3);
      for(int i=0;i<w*h;i++){
      image->set(i*3,(int)(meanr(i,0)*255));
      image->set(i*3+1,(int)(meang(i,0)*255));
      image->set(i*3+2,(int)(meanb(i,0)*255));
      }
      Image::writePPM((unsigned char *)image->getDataPtr(),w,h,"mean.ppm",3);
  */
  
  printf("done with double matrix\n");
  //int nreigs[3]={-1,-1,-1};
  //int widths[3]={texw[0],texw[1],texw[2]};
  //int heights[3]={texh[0],texh[1],texh[2]};
  initDynaTexRGB(path,charts,ncharts,scale,trans,
		 extrinsics,X,rs,gs,bs,nreigs,widths,heights,lutw,luth,flags);
  //X.printMatlab("X");
}

Matrix * buildgrid(int * gridres,
		   const Matrix & X,
		   const Matrix & Y,int flags)
{
  int nreig=Y.m;
  double max_X[2];
  double min_X[2];
  
  max_X[0]=X.maxInCol(0);
  max_X[1]=X.maxInCol(1);
  min_X[0]=X.minInCol(0);
  min_X[1]=X.minInCol(1);

  /*  
  static int called=0;
  if(!called){
    X.printMatlab("X");
    Y.printMatlab("Y");
  }
  else if(called==1){
    X.printMatlab("XU");
    Y.printMatlab("YU");
  }
  else{
    X.printMatlab("XV");
    Y.printMatlab("YV");
  }
  called++;
  */

  Matrix * igrid=new Matrix[nreig];
  Matrix gridx(gridres[0],gridres[1]);
  Matrix gridy(gridres[0],gridres[1]);
  
  //Y.printMatlab("Y");
  //X.printMatlab("X");
  
  printf("angles %f - %f\n",min_X[0],max_X[0]);
  printf("angles %f - %f\n",min_X[1],max_X[1]);
  
  for(int i=0;i<gridres[0];i++){
    for(int j=0;j<gridres[1];j++){
      gridy(i,j)=((double)i/((double)gridres[0]-1))*(max_X[0]-min_X[0])+min_X[0];
      gridx(i,j)=((double)j/((double)gridres[1]-1))*(max_X[1]-min_X[1])+min_X[1];
    }
  }
  printf("inited grid\n");
  
  //gridy.printMatlab("gy");
  //gridx.printMatlab("gx");
  
  //Detect one dimensional lookup table...
  if(fabs(max_X[1]-min_X[1])<0.1 ||flags&LUT_LINEAR){
    printf("one dimensional grid %f\n",fabs(max_X[1]-min_X[1]),
	   (flags&LUT_LINEAR)>0);
    min_X[0]=0;
    max_X[0]=359.9;

    Matrix t1=X.getColumn(0);
    for(int k=0;k<nreig;k++){
      int i=0;
      Matrix yi=Y.getRow(k);
      igrid[k]=Matrix(gridres[0],gridres[1]);

      int thesmallest=0;
      int thebiggest=0;
      
      for(int j=1;j<t1.m;j++){
	if(t1[j]<t1[thesmallest]) thesmallest=j;
	if(t1[j]>t1[thebiggest]) thebiggest=j;
      }

      for(i=0;i<gridres[0];i++){
	double val=((double)i/((double)gridres[0]-1))*(max_X[0]-min_X[0])+min_X[0];
	int smaller=-1;
	int bigger=-1;
	for(int j=0;j<t1.m;j++){
	  if(t1[j]<val && (smaller==-1 || t1[j]>t1[smaller])) smaller=j;
	  if(t1[j]>val && (bigger==-1  || t1[j]<t1[bigger])) bigger=j;
	}
	if(bigger==-1)bigger=thesmallest;
	if(smaller==-1)smaller=thebiggest;
	double len=t1[bigger]-t1[smaller];
	double t=val-t1[smaller];
	t=t/len;
	igrid[k](i,0)=(1-t)*yi[smaller]+t*yi[bigger];
	for(int j=1;j<gridres[1];j++){
	  igrid[k](i,j)=igrid[k](i,0);
	}
      } 
    }
  }
  else if(fabs(max_X[0]-min_X[0])<0.1){//Assume that the linear means in the Y rotation
    printf("one dimensional grid\n");
    Matrix t1=X.getColumn(1);
    X.printMatlab("X");
    for(int k=0;k<nreig;k++){
      int i=0;
      Matrix yi=Y.getRow(k);
      igrid[k]=Matrix(gridres[0],gridres[1]);
      printf("resolution %d %d\n",gridres[0],gridres[1]);
      for(i=0;i<gridres[0];i++){
	double val=((double)i/((double)gridres[0]-1))*(max_X[1]-min_X[1])+min_X[1];
	int smaller=-1;
	int bigger=-1;
	for(int j=0;j<t1.m;j++){
	  if(t1[j]<val && (smaller==-1 || t1[j]>t1[smaller])){
	    smaller=j;
	  }
	  if(t1[j]>val && (bigger==-1 || t1[j]<t1[bigger])){
	    bigger=j;
	  }
	}
	//printf("smaller %d, bigger %d\n",smaller,bigger);
	double len=t1[bigger]-t1[smaller];
	double t=val-t1[smaller];
	t=t/len;
	igrid[k](i,0)=(1-t)*yi[smaller]+t*yi[bigger];
	for(int j=1;j<gridres[1];j++){
	  igrid[k](i,j)=igrid[k](i,0);
	}
      } 
    }
  }
  else{
    for(int i=0;i<nreig;i++){
      Matrix t1=X.getColumn(1);
      Matrix t2=X.getColumn(0);
      Matrix yi=Y.getRow(i);
      printf("calling griddata %d\n",i);
      igrid[i]=griddata(t1,t2,yi,
			gridx,gridy,flags,i==0);
    }
  }
  return igrid;
}
  
void pca(Matrix & mat,int nreig ,Matrix & mean,Matrix & eigIm,Matrix & Y)
{  
  //mat.printMatlab("myInput");
  //covar is the in/out param to the first mult
  Matrix covar(mat.n,mat.n);
  mean=Matrix(mat.m,1);
  covar.setAll(0);
  mean.setAll(0);

  if(nreig>mat.n)nreig=mat.n;
  printf("%%pca: mat(%d,%d) nreig %d\n",mat.m,mat.n,nreig);
  printf("%%pca: getting the mean texture\n");
  //Get the mean texture in mean
  for(int i=0;i<mat.m;i++){
    for(int j=0;j<mat.n;j++){
      mean[i]+=mat.get(i,j);
    }
    mean[i]/=mat.n;
  }
  printf("%%pca: subtracting the mean texture out\n");
  //Subtract the mean texture
  for(int i=0;i<mat.m;i++){
    for(int j=0;j<mat.n;j++){
      mat(i,j)-=mean[i];
    }
  }
  
  printf("pca: performing the covariance multiplication\n");
  //usually X'*X, but the first transpose is not 
  //necessary since our matrices are row major,
  //and LINPACKS are column major
  dgemm_('N','T',mat.n,mat.n,mat.m,
	 1.0,mat.data,mat.n,mat.data,mat.n,
	 0.0,covar.data,mat.n);
  
  //covar = mat.transpose()*mat;
  //covar.printMatlab("covar");

  //printf("pca: initializing eigenmatrices\n");
  //printf("pca: size(mat)=%d,%d size(covar)=%d,%d\n",
  // mat.m,mat.n,covar.m,covar.n);
  Matrix vectors(nreig,mat.n);
  
  //nest so matrices get deleted automatically
  {
    int lwork=mat.n*mat.n;
    if(mat.n<8)lwork=128;
    Matrix wr(mat.n,1);
    Matrix wi(mat.n,1);
    Matrix vr(mat.n,mat.n);
    Matrix work(lwork,1);
    int info,* ifail = new int[mat.n];
    int found;
    int * iwork = new int[5*mat.n]; 
    Matrix values(nreig,1);
    printf("pca: calling dsyevx\n");
    
    dsyevx_('V','I','U',mat.n,covar.data,mat.n,
	    0,0,mat.n-nreig+1,mat.n,0.0,found,values.data,
	    vectors.data,mat.n,work.data,lwork,
	    iwork,ifail,info);
    printf("done dsyevx\n");
    values.printMatlab("eigenValues");
    //vectors.printMatlab("vectors");
    delete ifail;
    
    //values.printMatlab("values");
    //first nreig rows of vectors'
    //so the first important colums of vectors
  }
  printf("pca: getting Y\n");
  Y=Matrix(mat.n,nreig);
  for(int i=0;i<mat.n;i++){
    for(int j=0;j<nreig;j++){
      Y(i,j)=vectors(nreig-j-1,i);
    }
  }
  //OLD
  //Matrix eigIm=mat*Y;
  //Y=Y.transpose();

  //NEW
  printf("pca: final call to dgemm\n");
  eigIm=Matrix(Y.n,mat.m);
  dgemm_('T','T',mat.m,Y.n,mat.n,
	 1.0,mat.data,mat.n,Y.data,Y.n,
	 0.0,eigIm.data,eigIm.n);
  
  eigIm=eigIm.transpose();//this is necessary cuz result was column major
  Y=Y.transpose();        //this was done before anyhow

  for(int k=0;k<nreig;k++){
    double max=fabs(eigIm(0,k));
    for(int i=1;i<eigIm.m;i++){
      double v=fabs(eigIm(i,k));
      if(v>max)max=v;
    }
    for(int i=0;i<eigIm.m;i++){
      eigIm(i,k)/=max;
    }
    for(int j=0;j<Y.n;j++){
      Y(k,j)*=max;
    }
  }
  /*
  Y.printMatlab("myY");
  mean.printMatlab("myIMean");
  eigIm.printMatlab("myEigIm");*/
}


Matrix tr2yx(Matrix & R)
{
  Matrix a(3,1);
  
  a[0]=atan2(-R(2,0),R(2,2));
  a[1]=atan2(R(2,1),R(1,1));
  a[2]=0.0;
  return a;
}

Matrix tr2rpy(Matrix & m)
{
  double eps=2e-16;
  Matrix rpy(3,1);
  rpy.setAll(0);

  if(fabs(m(0,0)) < eps && fabs(m(1,0))<eps){
    rpy[0]=0;
    rpy[1]=atan2(-m(2,0),m(0,0));
    rpy[2]=atan2(-m(1,2),m(1,1));
  }
  else{
    rpy[0]=atan2(m(1,0),m(0,0));
    double sp=sin(rpy[0]);
    double cp=cos(rpy[0]);
    rpy[1]=atan2(-m(2,0),cp*m(0,0)+sp*m(1,0));
    rpy[2]=atan2(sp*m(0,2)-cp*m(1,2),cp*m(1,1)-sp*m(0,1));
  }
  return rpy;
}

Matrix tr2eul(Matrix & m)
{
  double eps=2e-16;
  Matrix euler(3,1);
  euler.setAll(0);
  
  if(fabs(m(1,2)) > eps &&
     fabs(m(0,2)) > eps){
    euler[0]=atan2(m(1,2),m(0,2));
    double sp=sin(euler[0]);
    double cp=cos(euler[0]);
    euler[1]=atan2(cp*m(0,2)+sp*m(1,2),m(2,2));
    euler[2]=atan2(-sp*m(0,0)+cp*m(1,0),-sp*m(0,1)+cp*m(1,1));
  }
  else{
    euler[0]=0;
    euler[1]=atan2(m(0,2),m(2,2));
    euler[2]=atan2(m(1,0),m(1,1));
  }
  return euler;
}

void zeroOutsideTriangulation(Matrix & xd,Matrix & yd,Matrix & zd,Matrix & xi,
			      Matrix & yi,Matrix & zi,int dolinear=0,int uselast=0)
{
  if(!glee_loader.init())return;

  static int called=0;
  static double xrange[2];
  static double yrange[2];
  static char tempfile[256];
  static char cmd[1024];

  if(!uselast){
    xrange[0]=10e10;
    xrange[1]=-10e10;
    yrange[0]=10e10;
    yrange[1]=-10e10;
    if(called)
      remove(tempfile);

    sprintf(tempfile,"%s/tmpfileXXXXXX",P_tmpdir);
    mkstemp(tempfile);
  //sprintf(cmd,"/usr/vis/research/qhull-2003.1/src/qdelaunay o QJ> %s ",tempfile);
    sprintf(cmd,"qdelaunay o QJ> %s ",tempfile);
    printf("cmd %s\n",cmd);
  
    FILE * p = popen(cmd,"w");
    if(!p)return;
    fprintf(p,"%d %d\n",2,xd.m);
  
    for(int i=0;i<xd.m;i++){
      fprintf(p,"%f %f\n",xd[i],yd[i]);
      if(xd[i]>xrange[1])xrange[1]=xd[i];
      if(yd[i]>yrange[1])yrange[1]=yd[i];
      if(xd[i]<xrange[0])xrange[0]=xd[i];
      if(yd[i]<yrange[0])yrange[0]=yd[i];
    }
    fclose(p);
  }
  else{
    printf("using most recent triangulation\n");
  }
  FILE * f=fopen(tempfile,"r");
  int ints[4];
  int ntri=0;
  int nvert=0;
  _3ints * tri;
  
  fscanf(f,"%d",ints);
  fscanf(f,"%d %d %d\n",ints+0,ints+1,ints+2);
  nvert=ints[0];
  ntri=ints[1];
  
  tri = new _3ints[ntri];
  
  for(int i=0;i<nvert;i++){
    float f1,f2,f3;
    fscanf(f,"%f %f %f",&f1,&f2,&f3);
  }
  
  for(int i=0;i<ntri;i++){
    int ngon;
    fscanf(f,"%d %d %d %d\n",&ngon,tri[i],tri[i]+1,tri[i]+2);
  }
  int viewport[4];
  glGetIntegerv(GL_VIEWPORT,viewport);
  glViewport(0,0,xi.m,xi.n);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(xrange[0],xrange[1],yrange[0],yrange[1],-1000,1000);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  
  glDisable(GL_LIGHTING);
  glActiveTexture(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  glActiveTexture(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  
  glClearColor(0.0,1.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
  
  double zrange[2]={10e10,-10e10};
  for(int i=0;i<zd.m*zd.n;i++){
    if(zd[i]>zrange[1])zrange[1]=zd[i];
    if(zd[i]<zrange[0])zrange[0]=zd[i];
  }
  
  printf("x:%f %f,y:%f %f,z:%f %f\n",
	 xrange[0],xrange[1],yrange[0],yrange[1],
	 zrange[0],zrange[1]);

  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  glBegin(GL_TRIANGLES);
  for(int i=0;i<ntri;i++){
    int ngon;
    float c=0;
    //glColor3f(1.0,1.0,1.0);
    c=(zd[tri[i][0]]-zrange[0])/(zrange[1]-zrange[0]);
    glColor3f(c,0,0);
    glVertex2d(xd[tri[i][0]],yd[tri[i][0]]);

    c=(zd[tri[i][1]]-zrange[0])/(zrange[1]-zrange[0]);
    glColor3f(c,0,0);
    glVertex2d(xd[tri[i][1]],yd[tri[i][1]]);
    
    c=(zd[tri[i][2]]-zrange[0])/(zrange[1]-zrange[0]);
    glColor3f(c,0,0);
    glVertex2d(xd[tri[i][2]],yd[tri[i][2]]);
  }
  glEnd();
  
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  
  glRasterPos2i(0,0);
  unsigned char * st = new unsigned char[xi.m*xi.n*20];

  if(dolinear){
    float * interp = new float[xi.m*xi.n*20];
    glReadPixels(0,0,xi.m,xi.n,
		 GL_RED,GL_FLOAT,interp);
    for(int i=0;i<xi.m*xi.n;i++){
      zi[i]=(double)(interp[i])*(zrange[1]-zrange[0])+zrange[0];
    }
    delete interp;
  }
  
  glReadPixels(0,0,xi.m,xi.n,
	       GL_GREEN,GL_UNSIGNED_BYTE,st);
  for(int i=0;i<xi.m*xi.n;i++){
    if(st[i])zi[i]=0;
  }
  delete st;
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);
  printErr("after");

  fclose(f);
  called++;
}

Matrix griddata(Matrix & xd,Matrix & yd,Matrix & zd,
		Matrix & xi,Matrix & yi,int flags,int first)
{
  if(flags&LUT_TRIANGLE){
    Matrix zi(xi.m,xi.n);
    zeroOutsideTriangulation(xd,yd,zd,xi,yi,zi,1,!first);
    return zi;
  }
  /*if(strcmp(type,"linear")==0){
    printf("got linear\n");
    Matrix zi(xi.m,xi.n);
    zi.setAll(0);

    char tempfile[256];
    char cmd[1024];
    sprintf(tempfile,"%s/tmpfileXXXXXX",P_tmpdir);
    mkstemp(tempfile);
    sprintf(cmd,"/usr/vis/research/qhull-2003.1/src/qdelaunay o > %s ",tempfile);
    printf("cmd %s\n",cmd);
    
    FILE * p = popen(cmd,"w");
    if(p){
      fprintf(p,"%d %d\n",2,xd.m);

      for(int i=0;i<xd.m;i++){
	fprintf(p,"%f %f\n",yd[i],xd[i]);
      }
      fclose(p);

      
      FILE * f=fopen(tempfile,"r");
      int ints[4];
      int ntri=0;
      int nvert=0;
      _3ints * tri;

      fscanf(f,"%d",ints);
      fscanf(f,"%d %d %d\n",ints+0,ints+1,ints+2);
      nvert=ints[0];
      ntri=ints[1];

      tri = new _3ints[ntri];

      for(int i=0;i<nvert;i++){
	float f1,f2,f3;
	fscanf(f,"%f %f %f",&f1,&f2,&f3);
      }

      for(int i=0;i<ntri;i++){
	int ngon;
	fscanf(f,"%d %d %d %d\n",&ngon,tri[i],tri[i]+1,tri[i]+2);
      }
      if(0){
      xd.printMatlab("xd");
      yd.printMatlab("yd");
      zd.printMatlab("zd");
      xi.printMatlab("xi");
      yi.printMatlab("yi");
      }
      for(int i=0;i<xi.m;i++){
	for(int j=0;j<xi.n;j++){
	  int k=0;
	  Vec3f p(xi(i,j),yi(i,j),0);
	  
	  for(k=0;k<ntri;k++){
	    Vec3f a((float)xd[tri[k][0]],(float)yd[tri[k][0]],0);
	    Vec3f b((float)xd[tri[k][1]],(float)yd[tri[k][1]],0);
	    Vec3f c((float)xd[tri[k][2]],(float)yd[tri[k][2]],0);
	    

	    Vec3f ab = b-a;
	    Vec3f bc = c-b;
	    Vec3f ca = a-c;
	    Vec3f ap = p-a;
	    Vec3f bp = p-b;
	    Vec3f cp = p-c;

	    Vec3f v1=ab.cross(ap);
	    Vec3f v2=bc.cross(bp);
	    Vec3f v3=ca.cross(cp);
	    	    
	    if(v1.dot(v2)*v1.dot(v3)>0){
	      zi(i,j)=1;
	      break;
	    }
	  }
	  if(k==ntri){
	    zi(i,j)=0;
	  }
	  
	}
      }

      fclose(f);
      remove(tempfile);
    }
    else{
      printf("error opening pipe\n");
    }    
    return zi;
  }
*/
  //
  if(flags&LUT_V4){
    /*
    xd.printMatlab("xd");
    yd.printMatlab("yd");
    zd.printMatlab("zd");
    xi.printMatlab("xi");
    yi.printMatlab("yi");
    */

    Matrix zi(xi.m,xi.n);
    int npoints=xd.m*xd.n;
    Matrix dist(npoints,npoints);
    Matrix g(npoints,npoints);
    Matrix w;
    for(int i=0;i<npoints;i++){
      for(int j=0;j<npoints;j++){
	double dx=xd[i]-xd[j];
	double dy=yd[i]-yd[j];
	dist(i,j)=sqrt(dx*dx+dy*dy);
	if(i==j)dist(i,j)=1;
	g(i,j)=dist(i,j)*dist(i,j)*(log(dist(i,j))-1);
	if(i==j)g(i,j)=0;
      }
    }
    zd.m*=zd.n;
    zd.n=1;

    w=Matrix(npoints,1);
    
    {
      Matrix gt=g.transpose();
      int pivot[npoints];
      int info;
      Matrix work(npoints*npoints,1);
      memcpy(w.data,zd.data,sizeof(double)*zd.m*zd.n);
      
      g.printMatlab("g");
      //dgesv_(npoints,1,gt.data,npoints,
      //     pivot,w.data,npoints,info);
      dsysv_('U',npoints,1,gt.data,npoints,
	     pivot,w.data,npoints,work.data,work.m,info);
      
      printf("info=%d;\n",info);

      //w.printMatlab("w");
      //zd.printMatlab("zd");
    }
    //w=Matrix::linLeastSq(g,zd);

    

    g=Matrix(npoints,1);
    for(int i=0;i<xi.m;i++){
      for(int j=0;j<xi.n;j++){
	Matrix d(npoints,1);
	double  px=xi(i,j);
	double  py=yi(i,j);
	for(int k=0;k<npoints;k++){
	  double dx=px-xd[k];
	  double dy=py-yd[k];
	  d[k]=sqrt(dx*dx+dy*dy);
	  if(d[k]==0){
	    d[k]=1;
	    g[k]=0;
	  }
	  else
	    g[k]=d[k]*d[k]*(log(d[k])-1);
	}
	zi(i,j)=g.dot(w);
      }
    }
    //    zi.printMatlab("zi");
    zeroOutsideTriangulation(xd,yd,zd,xi,yi,zi,0,!first);
    return zi;
  }
  else if(flags&LUT_CUBIC){
    printf("got cubic\n");
    Matrix zi(xi.m,xi.n);
    
    int n,nc,nw,nr,*lcell,*lnext,ier;
    double xmin,ymin,dx,dy,rmax,*rw,*a;
    double sx,sy;
    n=xd.m;
    nc=17;
    if(nc<9)nc=9;
    if(nc>40)nc=40;
    if(nc>n-1)nc=n-1;
    nw=(int)(1.5*nc);
    if(nw>40)nw=40;
    if(nw>n-1)nw=n-1;
    if(nw<1)nw=1;
    nr=(int)sqrt(n/3);
    if(nr<1)nr=1;
    
    lcell=new int[nr*nr*10];
    lnext =new int[n];
    rw= new double [n*10];
    a = new double[10*n];
    
    printf("calling cshep2\n");
    cshep2_(n,xd.data,yd.data,zd.data,
	    nc,nw,nr,lcell,lnext,xmin,ymin,dx,dy,rmax,rw,a,ier);
    
    printf("called cshep2 %d\n",ier);
    
    //printf("xmin %f, ymin %f, dx: %f, dy %f\n",
    // xmin,ymin,dx,dy);
    
    if(ier){
      printf("error %d\n",ier);
      printf("xmin %f, ymin %f, dx: %f, dy %f\n",
	     xmin,ymin,dx,dy);

      //xi.print();
      //yi.print();


    }
    else{
      for(int i=0;i<xi.m;i++){
	for(int j=0;j<xi.n;j++){
	  double px=(double)xi(i,j);
	  double py=(double)yi(i,j);
	  zi(i,j)=
	    cs2val_(px,py,n,xd.data,yd.data,zd.data,
		    nr,lcell,lnext,xmin,ymin,dx,dy,rmax,rw,a);
	}
      }
    }
    delete rw;
    delete a;
    delete lnext;
    delete lcell;
    zeroOutsideTriangulation(xd,yd,zd,xi,yi,zi,0,!first);
    return zi;
  }
  printf("got nothing\n");
}


class TriVec3f{
public:
  Vec3f data[3];
  TriVec3f(const Vec3f & a_,const Vec3f & b_,const Vec3f & c_){
    data[0]=a_;
    data[1]=b_;
    data[2]=c_;
  }
  int operator==(const TriVec3f & o)const{
    int eq=0;
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
	if(data[i]==o.data[j]){
	  eq++;
	  break;
	}
      }
    }
    return eq==3;
  }
};

struct TriVec3fEq{
  int operator()(const TriVec3f & a,
		 const TriVec3f & b){
    return a==b;
  }
};

struct TriVec3fHash{
  size_t operator()(const TriVec3f & tri) const{
    return (int)(65537*(tri.data[0].x+tri.data[0].y+tri.data[0].z+
			tri.data[1].x+tri.data[1].y+tri.data[1].z+
			tri.data[2].x+tri.data[2].y+tri.data[2].z));
  }
};

//NOTE: here that translation gets applied before any scaling
//i.e, the translation is used to place the origin at the
//     center of the unit cube. 
void saveobj(char * filename,
	     Chart * charts, int ncharts,Matrix & X,const Vec3f & scale,
	     const Vec3f & trans,int flags)
{
  FILE * file = fopen(filename,"w");

  for(int i=0;i<ncharts;i++){
    for(int j=0;j<charts[i].nvert;j++){
      fprintf(file,"v %f %f %f\n",
	      scale.x*(charts[i].vert[j].x+trans.x),
	      scale.y*(charts[i].vert[j].y+trans.y),
	      scale.z*(charts[i].vert[j].z+trans.z));
    }
  }

  if(ncharts && charts[0].tc){
      for(int i=0;i<ncharts;i++){
	  for(int j=0;j<charts[i].nvert;j++){
	      fprintf(file,"vt %f %f\n",
		      charts[i].tc[j].x,
		      charts[i].tc[j].y);
	  }
      }
  }


  //This actually only needs an edge that appears in two
  //or more charts (this would work for clipped triangles also)
  hash_map<TriVec3f,int,TriVec3fHash,TriVec3fEq> writtenTris;
  
  int dups=0;
  for(int i=0;i<ncharts;i++){
    for(int j=0;j<charts[i].npoly;j++){
      TriVec3f tri(charts[i].vert[charts[i].poly[j][0]],
		   charts[i].vert[charts[i].poly[j][1]],
		   charts[i].vert[charts[i].poly[j][2]]);
      if(writtenTris.find(tri)!=writtenTris.end()){
	dups++;
      }
      else
	writtenTris[tri]=1;
    }
  }


  //new
  //This needs to be more sophisticated...
  //currently bordering triangles are removed regardless
  
  int pi=1;
  if(dups>ncharts){
    printf("detected duplicates...removing the chart border\n");
    for(int i=0;i<ncharts;i++){
      int * onborder=new int[charts[i].nvert];
      int * border;
      int nborder;
      memset(onborder,0,sizeof(int)*charts[i].nvert);
      nborder=charts[i].getBorder(&border);
      for(int j=0;j<2*nborder;j++){
	onborder[border[j]]=1;
      }
      delete border;
      
      for(int j=0;j<charts[i].npoly;j++){
	TriVec3f tri(charts[i].vert[charts[i].poly[j][0]],
		     charts[i].vert[charts[i].poly[j][1]],
		     charts[i].vert[charts[i].poly[j][2]]);
	int nonb=onborder[charts[i].poly[j][0]]+
	  onborder[charts[i].poly[j][1]]+
	  onborder[charts[i].poly[j][2]];
	if(nonb>=2){
	  //its on the border..don't bother with it
	}
	else{
	  fprintf(file,"f %d %d %d\n",
		  charts[i].poly[j][0]+pi,
		  charts[i].poly[j][1]+pi,
		  charts[i].poly[j][2]+pi);
	}
      }
      delete onborder;
      pi+=charts[i].nvert;
    }
    printf("got %d dups\n",dups);
  }
  else{
    //old
    for(int i=0;i<ncharts;i++){
      for(int j=0;j<charts[i].npoly;j++){
	fprintf(file,"f %d %d %d\n",
		charts[i].poly[j][0]+pi,
		charts[i].poly[j][1]+pi,
		charts[i].poly[j][2]+pi);
      }
      pi+=charts[i].nvert;
    }
  }
  if(X.m >0 && X.n>0){
      double means[3];
      double mins[3];
      double maxs[3];
      mins[0]=X.minInCol(0);
      mins[1]=X.minInCol(1);
      mins[2]=X.minInCol(2);
      maxs[0]=X.maxInCol(0);
      maxs[1]=X.maxInCol(1);
      maxs[2]=X.maxInCol(2);
      
      means[0]=0;
      means[1]=0;
      means[2]=0;
      for(int i=0;i<X.m;i++){
	  means[0]+=X(i,0);
	  means[1]+=X(i,1);
	  means[2]+=X(i,2);
      }
      means[0]/=X.m;
      means[1]/=X.m;
      means[2]/=X.m;
      
      if(mins[2]==maxs[2]){
	  mins[2]-=1;
      }
      //FIXME: force lookup table to be 360 degrees
      //Wow, potentially a bad idea if the LINEAR table is in the 0 direction
      //but not a 360 degree view (Neil Nov 2005)
      if(flags&LUT_LINEAR){
	mins[0]=0;
	maxs[0]=359.99;
      }
      printf("FIXME: outputting 0 for mean x rotation\n");
      fprintf(file,"mean_x %f %f %f\n",means[0],means[1],means[2]*0);
      fprintf(file,"min_x %f %f %f\n",mins[0],mins[1],mins[2]);
      fprintf(file,"max_x %f %f %f\n",maxs[0],maxs[1],maxs[2]);
  }
  fclose(file);
}

void eigwrite(char * fname,Matrix & mean,Matrix & eig,int w,int h)
{
  printf("wrigint eig %d %d %d to %s\n",w,h,eig.n,fname);
  int nreig=eig.n+1;
  FILE * file = fopen(fname,"wb");
  fwrite(&h,sizeof(int),1,file);
  fwrite(&w,sizeof(int),1,file);
  fwrite(&nreig,sizeof(int),1,file);
  unsigned char * buf = new unsigned char[w*h*nreig];
  unsigned char * buf2 = new unsigned char[w*h*4];

  for(int i=0;i<w*h;i++){
    int temp = (int)((127*mean[i])+128);
    if(temp<0)temp=0;
    if(temp>255)temp=255;
    unsigned char c=(unsigned char)temp;
    buf[i]=c;
  }

  static int called=0;
  
  char str[256];
  sprintf(str,"image%d.pgm",called);
  //writePPM(buf,w,h,str,1);
  called++;

  for(int k=0;k<eig.n;k++){
    for(int i=0;i<w*h;i++){
      int temp = (int)((127*eig(i,k))+128);
      if(temp<0)temp=0;
      if(temp>255)temp=255;
      unsigned char c=(unsigned char)temp;
      buf[(k+1)*w*h+i]=c;
    }
  }
  for(int k=0;k<nreig/4;k++){
    for(int i=0;i<w*h;i++){
      buf2[4*i]=buf[(k*4)*w*h+i];
      buf2[4*i+1]=buf[(k*4+1)*w*h+i];
      buf2[4*i+2]=buf[(k*4+2)*w*h+i];
      buf2[4*i+3]=buf[(k*4+3)*w*h+i];
    }
    fwrite(buf2,1,w*h*4,file);
  }

    
  fclose(file);
  delete buf;
}

void gridwriteCube(char * fname,Matrix * grid,int ngrid)
{
  FILE * file = fopen(fname,"wb");
  printf("writing cube grid %d %d %d\n",grid[0].n,
	 grid[0].m,ngrid);
  fwrite(&(grid[0].n),sizeof(int),1,file);
  fwrite(&(grid[0].m),sizeof(int),1,file);
  fwrite(&(ngrid),sizeof(int),1,file);

 
  for(int k=0;k<ngrid;k++){
    //write out the grid in column major
    for(int i=0;i<grid[0].m*grid[0].n;i++){
      float f=grid[k][i];
      fwrite(&f,sizeof(float),1,file);
    }
  }
  fclose(file);
}

void gridwrite(char * fname,Matrix * grid,int ngrid)
{
  FILE * file = fopen(fname,"wb");
  printf("writing grid %d %d %d\n",grid[0].m,
	 grid[0].n,ngrid);
  fwrite(&(grid[0].m),sizeof(int),1,file);
  fwrite(&(grid[0].n),sizeof(int),1,file);
  fwrite(&(ngrid),sizeof(int),1,file);

  for(int k=0;k<ngrid;k++){
    //write out the grid in column major
    for(int j=0;j<grid[k].n;j++){
      for(int i=0;i<grid[k].m;i++){
	float f=grid[k](i,j);
	fwrite(&f,sizeof(float),1,file);
      }
    }
  }
  fclose(file);
}

void pca(Matrix & mat,int nreig ,Matrix & mean,Matrix & eigIm,Matrix & Y);


void rgb2yuv(Matrix & red,Matrix & green,Matrix & blue)
{
  
  for(int i=0;i<red.m;i++){
    for(int j=0;j<red.n;j++){
      double Y=  0.299*red(i,j)+0.587 *green(i,j)+ 0.114 *blue(i,j);
      double U= -0.169*red(i,j)-0.3316*green(i,j)+ 0.500 *blue(i,j);
      double V=  0.500*red(i,j)-0.4186*green(i,j)- 0.0813*blue(i,j);

      
      red(i,j)=Y;
      green(i,j)=U;
      blue(i,j)=V;
    }
  }
}

void resize(Matrix & mat,int & fromw,int & fromh,int & w,int & h)
{
  if(fromw==w && fromh==h){
    return;
  }
  printf("in resize from %d %d to %d %d\n",fromw,fromh,w,h);
  if(w>fromw || h>fromh){
    w=fromw;
    h=fromh;
    printf("can not upsample\n");
    return;
  }
  int scaleh=fromh/h;
  int scalew=fromw/w;

  printf("scale is %d %d\n",scalew,scaleh);

  Matrix newMat(w*h,mat.n);
  
  for(int k=0;k<mat.n;k++){
    for(int i=0;i<h;i++){
      for(int j=0;j<w;j++){
	double sum=0;
	int cnt=0;
	for(int ii=i*scaleh;ii<(i+1)*scaleh;ii++){
	  for(int jj=j*scalew;jj<(j+1)*scalew;jj++){
	    sum+=mat(ii+jj*fromh,k);
	    cnt++;
	  }
	}
	newMat(i+j*h,k)=sum/cnt;
      }
    }
  }
  mat=newMat;
}

//These are the camera centers in the objects coordinate system
Vec3d * getCameraCenters(double * extrinsics,Matrix & X,Vec3f & scale,
			 Vec3f & trans)
{
  Vec3d objCenter;//object center in world coordinates
  Vec3d * cameraCenters = new Vec3d[X.m];
  objCenter.x=scale.x*0.5+trans.x;
  objCenter.y=scale.y*0.5+trans.y;
  objCenter.z=scale.z*0.5+trans.z;
  //The 0.5 is because the object is in the unit cube,
  //therefore the center of the object is at (0.5,0.5,0.5)
  //so the center in world coords is trans+scale*(0.5,0.5,0.5)

  for(int i=0;i<X.m;i++){
    double extinv[16];
    Vec3d cpos;
    invert(extinv,extrinsics+i*16);
    cpos=Vec3d(-extinv[12],extinv[13],extinv[14]);
    cpos=cpos-objCenter;
    printf("camera:%f,%f,%f object:%f,%f,%f\n",
	   cpos.x,cpos.y,cpos.z,objCenter.x,objCenter.y,objCenter.z);
    int which=0;
    double m=fabs(cpos.x);
    if(fabs(cpos.y)>m){
      which=1;
      m=fabs(cpos.y);
    }
    if(fabs(cpos.z)>m){
      which=2;
      m=fabs(cpos.z);
    }
    cpos.normalize();
    cameraCenters[i]=cpos;

    printf("pt(%d,:)=[%f %f %f];\n",
	   i,cpos.x,cpos.y,cpos.z);
  }
  return cameraCenters;
}

_3ints * qhull3D(Vec3d * points,int npoints,int & ntri)
{
  char tempfile[256];
  char cmd[1024];

  sprintf(tempfile,"%s/tmpfileXXXXXX",P_tmpdir);
  mkstemp(tempfile);
  sprintf(cmd,"qhull o QJ> %s ",tempfile);
  printf("cmd %s\n",cmd);
  
  FILE * p = popen(cmd,"w");
  if(!p)return 0;
  fprintf(p,"%d %d\n",3,npoints);
  
  for(int i=0;i<npoints;i++){
    fprintf(p,"%f %f %f\n",points[i].x,points[i].y,points[i].z);
  }
  fclose(p);
  
  FILE * f=fopen(tempfile,"r");
  int ints[4];
  int nvert=0;
  _3ints * tri;
  
  fscanf(f,"%d",ints);
  fscanf(f,"%d %d %d\n",ints+0,ints+1,ints+2);
  nvert=ints[0];
  ntri=ints[1];
  
  tri = new _3ints[ntri];
  
  //verticies...ignore
  for(int i=0;i<nvert;i++){
    float f1,f2,f3;
    fscanf(f,"%f %f %f",&f1,&f2,&f3);
  }
  
  //triangles
  for(int i=0;i<ntri;i++){
    int ngon;
    fscanf(f,"%d %d %d %d\n",&ngon,tri[i],tri[i]+1,tri[i]+2);
  }

  printf("pts=[\n");
  for(int i=0;i<npoints;i++){
    printf("%f %f %f\n",points[i].x,points[i].y,points[i].z);
  }
  printf("]\n");

  printf("tri=[\n");
  for(int i=0;i<ntri;i++){
    printf("%d %d %d\n",tri[i][0],tri[i][1],tri[i][2]);
  }
  printf("\n");
  return tri;
}

double triangleArea(Vec3d & v1,Vec3d & v2,Vec3d & v3)
{
  double a=(v1-v2).len();
  double b=(v2-v3).len();
  double c=(v3-v1).len();
  double s=(a+b+c)/2.0;
  return sqrt(s*(s-a)*(s-b)*(s-c));
}

double * getCubeCoeffs(Vec3d * cameraCenters,int ncams,int dim,int flags){
  int ntri=0;
  _3ints * tri=qhull3D(cameraCenters,ncams,ntri);
  Vec3d * normals=new Vec3d[ntri];
  double * ds=new double[ntri];
  if(tri&&ntri){
    for(int i=0;i<ntri;i++){
      Vec3d a=cameraCenters[tri[i][1]]-cameraCenters[tri[i][0]];
      Vec3d b=cameraCenters[tri[i][2]]-cameraCenters[tri[i][0]];
      normals[i] = a.cross(b);
      normals[i].normalize();

      if(normals[i].dot(cameraCenters[tri[i][1]])<0){
	normals[i].mult(-1);
      }
      ds[i]=-normals[i].dot(cameraCenters[tri[i][0]]);

      /*printf("proj: %f %f %f\n",
	     normals[i].dot(cameraCenters[tri[i][0]])+ds[i],
	     normals[i].dot(cameraCenters[tri[i][1]])+ds[i],
	     normals[i].dot(cameraCenters[tri[i][2]])+ds[i]);
      */
    }
  }
  Image8 image(dim,dim,3);

  glGenTextures(6,cubeTex);

  printf("got normals\n");
  printf("getting coeffs for %d cameras\n",ncams);
  fflush(stdout);
  double * coeffs=new double[6*dim*dim*ncams];
  for(int k=0;k<6;k++){
    int ind1=-1;
    int ind2=-1;
    int other=-1;
    for(int i=0;i<3;i++){
      if(face[k][i])other=i;
      else if(!face[k][i] && ind1<0){
	ind1=i;
      }
      else if(!face[k][i] && ind1>=0){
	ind2=i;
      }
    }
    printf("%d %d %d\n",ind1,ind2,other);
    for(int y=0;y<dim;y++){
      for(int x=0;x<dim;x++){
	double xpt=2.0*(double)x/(double)(dim-1) - 1.0;
	double ypt=2.0*(double)y/(double)(dim-1) - 1.0;
	Vec3d point(0,0,0);
	Matrix c;
	int mini=0;
	double sum=0;
	double maxd=0;
	double * data = point.getData();
	data[ind1]=xpt;
	data[ind2]=ypt;
	data[other]=face[k][other];
	data[0]*=-1;
	data[1]*=-1;
	//data[1]=-1*data[1]+1;
	point.normalize();
	c=Matrix(ncams,1);
	c.setAll(0);

	if(flags&LUT_NEAREST){
	  int closestpt=0;
	  double dist=10e10;
	  for(int i=0;i<ncams;i++){
	    Vec3d diff=cameraCenters[i]-point;
	    double thisdist=diff.len();
	    if(thisdist<dist){
	      closestpt=i;
	      dist=thisdist;
	    }
	  }
	  for(int i=0;i<ncams;i++){
	    coeffs[k*dim*dim*ncams+
		   y*dim*ncams+
		   x*ncams+i]=(closestpt==i);
	  }
	}
	else if(flags&LUT_GLOBAL){
	  for(int i=0;i<c.m;i++){
	    double d=point.dot(cameraCenters[i]);
	    if(d<0.866){
	      d=0;
	    }
	    else{
	      d=1.0/(1.0-d+0.01);
	    }
	    c[i]=d;
	    //if(c[i]<c[mini]){
	    //mini=i;
	    //}
	    if(d>maxd){
	      maxd=d;
	    }
	    sum+=d;
	  }
	  for(int i=0;i<c.m;i++){
	    c[i]/=sum;
	    //if(i==mini)c[i]=1.0;
	    //else c[i]=0;
	    coeffs[k*dim*dim*ncams+
		   y*dim*ncams+
		   x*ncams+i]=c[i];
	  }
	}
	else{
	  double maxd=-10;
	  int closesttri=-1;
	  Vec3d pgood;
	  for(int i=0;i<ntri;i++){
	    double t = -(ds[i]) / (normals[i].dot(point));
	    if(t>0){
	      Vec3d p=point*t;
	      
	      Vec3d ab=(cameraCenters[tri[i][1]]-cameraCenters[tri[i][0]])
		.cross(normals[i]);
	      Vec3d bc=(cameraCenters[tri[i][2]]-cameraCenters[tri[i][1]])
		.cross(normals[i]);
	      Vec3d ca=(cameraCenters[tri[i][0]]-cameraCenters[tri[i][2]])
		.cross(normals[i]);
	      
	      double ab_d=-ab.dot(cameraCenters[tri[i][0]]);
	      double bc_d=-bc.dot(cameraCenters[tri[i][1]]);
	      double ca_d=-ca.dot(cameraCenters[tri[i][2]]);
	      
	      if(ab.dot(p)+ab_d>=0.0 &&
		 bc.dot(p)+bc_d>=0.0 &&
		 ca.dot(p)+ca_d>=0.0 &&
		 (closesttri==-1)){
		closesttri=i;
		pgood=p;
	      }
	    }
	  }

	  if(closesttri==-1){
	    closesttri=ntri-1;
	    
	    int closestpt=0;
	    double dist=10e10;
	    for(int i=0;i<ncams;i++){
	      Vec3d diff=cameraCenters[i]-point;
	      double thisdist=diff.len();
	      if(thisdist<dist){
		closestpt=i;
		dist=thisdist;
	      }
	    }
	    for(int i=0;i<ncams;i++){
	      coeffs[k*dim*dim*ncams+
		     y*dim*ncams+
		     x*ncams+i]=(closestpt==i);
	    }
	    continue;	    
	    
	    /*
	    for(int i=0;i<ntri;i++){
	      double d = normals[i].dot(point)+ds[i];
	      Vec3d diff=normals[i]*d;
	      Vec3d p=point-diff;
	      
	      if((closesttri==-1||(fabs(d)<mind))){
		closesttri=i;
		mind=fabs(d);
	      }
	      }*/
	  }
	  
	  //clear coeffs
	  for(int i=0;i<ncams;i++){
	    coeffs[k*dim*dim*ncams+
		   y*dim*ncams+
		   x*ncams+i]=0;
      	  }
	  
	  //get coeffs
	  for(int i=0;i<3;i++){
	    double d=point.dot(cameraCenters[tri[closesttri][i]]);
	    if(d<0.7){
	      d=0;
	    }
	    else{
	      d=1.0/(1.0-d+0.01);
	    }
	    c[tri[closesttri][i]]=d;
	    if(d>maxd){
	      maxd=d;
	    }
	    sum+=d;
	  }

	  /*
	    //Barycentric coords
	  double area=triangleArea(cameraCenters[tri[closesttri][0]],
				   cameraCenters[tri[closesttri][1]],
				   cameraCenters[tri[closesttri][2]]);
	  double col;
	  double c0,c1,c2;
	  c0=triangleArea(pgood,
			  cameraCenters[tri[closesttri][1]],
			  cameraCenters[tri[closesttri][2]])/area;
	  c1=triangleArea(cameraCenters[tri[closesttri][0]],
			  pgood,
			  cameraCenters[tri[closesttri][2]])/area;
	  c2=triangleArea(cameraCenters[tri[closesttri][0]],
			  cameraCenters[tri[closesttri][1]],
			  pgood)/area;
	  coeffs[k*dim*dim*ncams+
		 y*dim*ncams+
		 x*ncams+tri[closesttri][0]]=c0;
	  coeffs[k*dim*dim*ncams+
		 y*dim*ncams+
		 x*ncams+tri[closesttri][1]]=c1;
	  coeffs[k*dim*dim*ncams+
		 y*dim*ncams+
		 x*ncams+tri[closesttri][2]]=c2;
	  col=c0*tri[closesttri][0]+c1*tri[closesttri][1]+c2*tri[closesttri][2]
	  col=255*(col/ncams);
	  */

	  Vec3d e1=cameraCenters[tri[closesttri][1]]-
	      cameraCenters[tri[closesttri][0]];
	  Vec3d e2=cameraCenters[tri[closesttri][2]]-
	      cameraCenters[tri[closesttri][0]];
	  double e1len=e1.normalize();
	  double e2len=e2.normalize();
	  Vec3d norm=e1.cross(e2);
	  Vec3d pintri=pgood-cameraCenters[tri[closesttri][0]];
	  double e1c=(pintri.dot(e1)-pintri.dot(e2)*(e1.dot(e2)))/
	      (1-(e1.dot(e2)*e1.dot(e2)));
	  double e2c=(pintri-e1*e1c).len();
	  e1c/=e1len;
	  e2c/=e2len;
	  e1*=e1len;
	  e2*=e2len;
	  //printf("%f %f %f\n",(pintri-(e1*e1c+e2*e2c)).len(),e1c,e2c);

	  
	  double alpha1=acos(cameraCenters[tri[closesttri][0]].dot
			    (cameraCenters[tri[closesttri][1]]));
	  double c0=sin((1-e1c)*alpha1)/sin(alpha1);
	  double c1=sin(e1c*alpha1)/sin(alpha1);
	  Vec3d pon1=cameraCenters[tri[closesttri][0]]*c0+
	      cameraCenters[tri[closesttri][1]]*c1;
	  
	  double alpha2=acos(pon1.dot
			     (cameraCenters[tri[closesttri][2]]));
	  double pon1c=sin((1-e2c)*alpha2)/sin(alpha2);
	  double c2=sin(e2c*alpha2)/sin(alpha2);
	  
	  c0*=pon1c;
	  c1*=pon1c;
	  
	  coeffs[k*dim*dim*ncams+
		 y*dim*ncams+
		 x*ncams+tri[closesttri][0]]=c0;
	  coeffs[k*dim*dim*ncams+
		 y*dim*ncams+
		 x*ncams+tri[closesttri][1]]=c1;
	  coeffs[k*dim*dim*ncams+
		 y*dim*ncams+
		 x*ncams+tri[closesttri][2]]=c2;
	  double col;
	  col=c0*tri[closesttri][0]+
	      c1*tri[closesttri][1]+c2*tri[closesttri][2];
	      col=255*(col/ncams);

	  image(x,y,0)=(int) col;
	  image(x,y,1)=(int) col;
	  image(x,y,2)=(int) col;
	}
      }
    }
    glBindTexture(GL_TEXTURE_2D,cubeTex[k]);
    image.initTexture();
  }

  campts = new Vec3d[ncams];
  memcpy(campts,cameraCenters,sizeof(Vec3d)*ncams);
  camtri=new _3ints[ntri];
  memcpy(camtri,tri,sizeof(_3ints)*ntri);
  ncamtri=ntri;
  
  delete normals;
  delete tri;

  printf("returning the coefficients\n");
  return coeffs;
}

Matrix * buildCubeGrid(double * coeffs,Matrix & Y,int dim)
{
  Matrix * grid = new Matrix[Y.m];

  printf("building cube grid (Y is %d,%d=ncams) of dimension %d %d\n",
	 Y.m,Y.n,dim,dim);
  
  for(int rowi=0;rowi<Y.m;rowi++){
    Matrix row = Y.getRow(rowi);
    grid[rowi]=Matrix(6*dim,dim);
    
    for(int k=0;k<6;k++){
      for(int y=0;y<dim;y++){
	for(int x=0;x<dim;x++){
	  double val=0;
	  for(int ci=0;ci<Y.n;ci++){
	    val+=coeffs[k*dim*dim*Y.n+y*dim*Y.n+x*Y.n+ci]*row[ci];
	  }
	  grid[rowi](k*dim+y,x)=val;
	}
      }
    }
  }
  return grid;
}
