#ifndef DYNATEX_H
#define DYNATEX_H

#include <matrix.h>
#include <chart.h>
#include <image.h>
#include <globals.h>

#define LUT_CUBE        1
#define LUT_LINEAR      2
#define LUT_TRIANGLE    4
#define LUT_NEAREST     8
#define LUT_GLOBAL      16
#define LUT_V4          32
#define LUT_CUBIC       64

void myProjection(double fx,double fy,
		  double px,double py,
		  int imw,int imh,
		  double minz,double maxz);

void getTexturesHW(double * intrinsics,
		 double * projection,
		 double * extrinsics,
		 Image8 * images,
		 Chart * charts,
		 int ncharts,
		 Vec3f & scale,
		 Vec3f & trans,
		 unsigned char * texImages,
		 unsigned char * occluded,
		 int texw,int texh,
		 int start,int end,int nimages);

void getTextures(double * intrinsics,
		 double * projection,
		 double * extrinsics,
		 Image8 * images,
		 Chart * charts,
		 int ncharts,
		 Vec3f & scale,
		 Vec3f & trans,
		 unsigned char * texImages,
		 unsigned char * occluded,
		 int texw,int texh,
		 int start,int end);

void initDynaTexYUV(char * ,Chart *,int,Vec3f & scale,Vec3f & trans,
		    double * extrinsics,Matrix & X,
		    Matrix & Y,Matrix & U,Matrix & V,
		    int * nreigs,int * widths,int * heights,
		    int lutw=64,int luth=64,int flags=LUT_CUBE|LUT_TRIANGLE);

void initDynaTexRGB(char * ,Chart *,int,Vec3f & scale,Vec3f & trans,
		    double * extrinsics,
		    Matrix & X,Matrix & reds,
		    Matrix & greens,
		    Matrix & blues,
		    int * neig,int * widths,int * heights,
		    int lutw=64,int luth=64,int flags=LUT_CUBE|LUT_TRIANGLE);

void initDynaTex(char * ,Chart *,int,Vec3f & scale,Vec3f & trans,
		 double * extrinsics,
		 unsigned char *images,
		 unsigned char * occluded,
		 int w,
		 int h,
		 int nimages,
		 int * nreig,
		 int * widths,
		 int * heights,
		 int lutw=64,int luth=64,int flags=LUT_CUBE|LUT_TRIANGLE);
		 

Matrix tr2yx(Matrix & R);

Matrix griddata(Matrix & xd,Matrix & yd,Matrix & zd,
		Matrix & xi,Matrix & yi,int flags,int first=1);

void rgb2yuv(Matrix & red,Matrix & green,Matrix & blue);
void resize(Matrix & mat,int & fromw,int & fromh,int & w,int & h);

void eigwrite(char * fname,Matrix & mean,Matrix & eig,int w,int h);
void gridwriteCube(char * fname,Matrix * grid,int ngrid);
void gridwrite(char * fname,Matrix * grid,int ngrid);

Matrix * buildCubeGrid(double * coeffs,Matrix & Y,int dim);
Matrix * buildgrid(int * gridres,
		   const Matrix & X,
		   const Matrix & Y,int flags);

Vec3d * getCameraCenters(double * extrinsics,Matrix & X,
			 Vec3f & scale,Vec3f & trans);
double * getCubeCoeffs(Vec3d * cameraCenters,int ncams,int dim,int flags);
void pca(Matrix & mat,int nreig ,Matrix & mean,Matrix & eigIm,Matrix & Y);

//this should be elsewhere
void saveobj(char * filename,
	     Chart * charts, int ncharts,Matrix & X,
	     const Vec3f & scale=Vec3f(1,1,1),
	     const Vec3f & trans=Vec3f(0.0,0.0,0.0),
  	     int flags=0);
#endif
