#include "heap.h"
#include "shortestpath.h"

int nvert;
double * curd;
int * prev;
Heap * heap=NULL;

Graph * allocGraph(int nvert)
{
  Graph * g = (Graph * )malloc(sizeof(Graph));
  g->edgeList = (EdgeNode **)malloc(sizeof(EdgeList*)*nvert);
  memset(g->edgeList,0,sizeof(EdgeList*)*nvert);
  g->nvert=nvert;
  return g;
}

void graphAddEdge(Graph * g,int from,int to,double cost)
{
  if(g){
    EdgeNode * newNode;
    EdgeNode * finger = g->edgeList[from];
    int dup=0;
    while(finger){
      if(finger->to==to){
	dup=1;
	break;
      }
      finger = finger->next;
    }
    if(!dup){
      newNode = (EdgeNode *)malloc(sizeof(EdgeNode));
      newNode->next = g->edgeList[from];
      g->edgeList[from]=newNode;
      newNode->to = to;
      newNode->dist = cost;
    }
    dup=0;
    finger = g->edgeList[to];
    while(finger){
      if(finger->to==from){
	dup=1;
	break;
      }
      finger = finger->next;
    }
    if(!dup){
      newNode = (EdgeNode *)malloc(sizeof(EdgeNode));
      newNode->next = g->edgeList[to];
      g->edgeList[to]=newNode;
      newNode->to = from;
      newNode->dist = cost;
    }
  }
}

void printGraph(Graph * g)
{
  if(g){
    int i=0;
    while(i<g->nvert){
      EdgeNode * finger = g->edgeList[i];
      EdgeNode * back;
      printf("%d :",i);
      while(finger){
	printf("[%d,%f] ",finger->to,finger->dist);
	finger = finger->next;
      }
      putchar(10);
      i++;
    }
  }
}

void freeGraph(Graph * g)
{
  if(g){
    int i=0;
    while(i<g->nvert){
      EdgeNode * finger = g->edgeList[i];
      EdgeNode * back;
      while(finger){
	back = finger;
	finger = finger->next;
	free(back);
	//free(back);
      }
      i++;
    }
    free(g->edgeList);
    free(g);
  }
}

/*
	Should be called before any calls to dij
*/
void allocSingleSource(Graph * g)
{
  nvert = g->nvert;
  heap = new Heap(g->nvert);
  curd = (double *)malloc(sizeof(double)*g->nvert);
  prev = (int *)malloc(sizeof(int)*g->nvert);
}

void initSingleSource(int source,int nv)
{
  int i=0;
  heap->removeAll();
  memset(curd,0,sizeof(double)*nv);
  
  curd[source]=0;
  while(i<nv){
    prev[i]=-1;
    if(i!=source)
      curd[i]=10e8;
    heap->insert(i,curd[i]);
    i++;
  }
}

/*
	Should be called after the last call to dij
*/
void freeSingleSource()
{
  delete heap;
  free(curd);
  free(prev);
}

/*
	Should call allocSingleSource before, and
	freeSingleSource after last call
*/
void dij(Graph * g,int source)
{
  int curv;
  initSingleSource(source,g->nvert);

  while(heap->peek()!=-1){
    EdgeNode * finger;
    double val;
    curv=heap->remove(&val);
    finger = g->edgeList[curv];
    while(finger){
      int to = finger->to;
      if(curd[to]> (curd[curv]+finger->dist)){
	curd[to]=curd[curv]+finger->dist;
	heap->update(to,curd[to]);
	prev[to]=curv;
      }
      finger = finger->next;
    }
  }
#ifdef DIJ_PRINT
  for(int i=0;i<g->nvert;i++){
    printf("[%d,%f] ",prev[i],curd[i]); 
  }
#endif
}

/*
  void getPred.
  User resposible to free memory
*/
int * getPred()
{
  int * ret = (int *)malloc(sizeof(int)*nvert);
  memcpy(ret,prev,sizeof(int)*nvert);
  return ret;
}

int * getPredParam(int * where)
{
  memcpy(where,prev,sizeof(int)*nvert);
  return where;
}

/*
  void getPred.
  User resposible to free memory
*/
double * getDistances()
{
  double * ret = (double *)malloc(sizeof(double)*nvert);
  memcpy(ret,curd,sizeof(double)*nvert);
  return ret;
}

double * getDistancesParam(double * where)
{
  memcpy(where,curd,sizeof(double)*nvert);
  return where;
}


#ifdef SP_MAIN

int main()
{
  Graph * g = allocGraph(6);
  graphAddEdge(g,0,1,10);
  graphAddEdge(g,0,2,1);
  graphAddEdge(g,1,2,1);
  graphAddEdge(g,2,4,3);
  graphAddEdge(g,4,3,5);
  graphAddEdge(g,4,5,6);
  printGraph(g);

  allocSingleSource(g);
	
  dij(g,0);
  for(int i=0;i<g->nvert;i++){
    printf("[%d,%f] ",prev[i],curd[i]); 
  }
  getchar();

  freeGraph(g);
  freeSingleSource();

  return 0;
}

#endif


