#include <GLee.h>
#include <GL/gl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vec.h"
#include "sot.h"

int wwidth=256;
int wheight=256;

typedef double  _3doubles[3];
typedef int  _3ints[3];
typedef unsigned char uchar;

double q=0.75;//This is the lower bound of the stretch function

Chart chart;
Chart chartout;

Image8  image;
double * scaledImage;
int gwidth;
int gheight;

Vec2d * grid;
Vec2d * dispvec;



inline void add2(Vec2d * res,Vec2d * a,Vec2d * b)
{
  res->x=a->x+b->x;
  res->y=a->y+b->y;
}

inline void sub2(Vec2d * res,Vec2d * a,Vec2d * b)
{
  res->x=a->x-b->x;
  res->y=a->y-b->y;
}

inline double len2(Vec2d * a)
{
  return sqrt(a->x*a->x+a->y*a->y);
}
inline void set2(Vec2d * a,Vec2d * b)
{
  a->x=b->x;
  a->y=b->y;
}
inline void mult2(Vec2d * res,Vec2d * a,double f)
{
  res->x=a->x*f;
  res->y=a->y*f;
}

double getSigma(Vec2d * v);
Vec2d * getGridVec(int i,int j);
Vec2d * getGridVec(int i,int j)
{
  if(i<0||j<0)
    printf("error\n");
  if(i>=gheight||j>=gwidth){
    printf("error\n");
  }
  return grid+i*gwidth+j;
}



double eval(Vec2d * vi,Vec2d * vj)
{
  double svi;
  double svj;
  Vec2d diff;
  sub2(&diff,vj,vi);
  
  svi=getSigma(vi);
  svj=getSigma(vj);
  
  return (fabs(svj-svi)/(svi+svj))/len2(&diff);
}

Vec2d * findh(Vec2d * vi,int i,int j)
{
  Vec2d * besth=NULL;
  double maxVal=-100000000;
  double curVal;
  Vec2d * curh=NULL;
  double svi;
  double svj;
    
  //Get neighbors
  if(i>0){
    curh=getGridVec(i-1,j);
    curVal=eval(vi,curh);
    if(curVal>maxVal){
      maxVal=curVal;
      besth=curh;
    }
  }
  if(j>0){
    curh=getGridVec(i,j-1);
    curVal=eval(vi,curh);
    if(curVal>maxVal){
      maxVal=curVal;
      besth=curh;
    }
  }
  if(i<gheight-1){
    curh=getGridVec(i+1,j);
    curVal=eval(vi,curh);
    if(curVal>maxVal){
      maxVal=curVal;
      besth=curh;
    }
  }
  if(j<gwidth-1){
    curh=getGridVec(i,j+1);
    curVal=eval(vi,curh);
    if(curVal>maxVal){
      maxVal=curVal;
      besth=curh;
    }
  }
  return besth;
}

void move()
{
  int i,j;

  for(i=0;i<gheight;i++){
    for(j=0;j<gwidth;j++){
      Vec2d * v;
      Vec2d * dir;
      Vec2d dc;
      v = grid+i*gwidth+j;
      dir = dispvec+i*gwidth+j;
      if(j!=0 && j!=gwidth-1)
	dc.x = dir->x;
      else dc.x=0;
      if(i!=0 && i!=gheight-1)
	dc.y = dir->y;
      else dc.y=0;
      add2(v,v,&dc);
    }
  }
  for(i=1;i<gheight;i++){
    Vec2d * v = grid+i*gwidth;
    Vec2d * vb = grid+(i-1)*gwidth;
    
    if(v->x<vb->x){
      v->x=vb->x+0.00001;
    }
    v = grid+i*gwidth+j;
    vb = grid+(i-1)*gwidth+j;
    if(v->x<vb->x){
      v->x=vb->x;
    }
  }
  for(j=1;j<gwidth;j++){
    Vec2d * v = grid+j;
    Vec2d * vb = grid+j-1;
    
    if(v->x<vb->x){
      v->x=vb->x+0.00001;
    }
    v = grid+j+gwidth*(gheight-1);
    vb = grid+j-1+gwidth*(gheight-1);
    if(v->x<vb->x){
      v->x=vb->x+0.00001;
    }
  }
  //Fix edge points
}

void doIteration()
{
  int i,j;
  float total=0;
  //Compute displacement vectors
  for(i=0;i<gheight;i++){
    for(j=0;j<gwidth;j++){
      Vec2d * vh;
      Vec2d * vi=&grid[i*gheight+j];
      double grad=0;
      double sh=0;
      double si=0;
      double uij=0;
      double ui=0;
      double rhs=0;
      Vec2d sub;
      Vec2d sum;
      Vec2d tvec;
      Vec2d * vj;
      vh=findh(vi,i,j);
      sh=getSigma(vh);
      si=getSigma(vi);
      sub2(&sub,vh,vi);
      
      grad=((sh-si)/(sh+si))/pow(len2(&sub),2);
      mult2(&sub,&sub,grad);
      //Sub is now the gradient term

      //Compute the left sumation

      sum.x=0;
      sum.y=0;
      

       //Get neighbors
      if(i>0)
	ui+=getSigma(getGridVec(i-1,j));
      if(j>0)
	ui+=getSigma(getGridVec(i,j-1));
      if(i<gheight-1)
	ui+=getSigma(getGridVec(i+1,j));
      if(j<gwidth-1)
	ui+=getSigma(getGridVec(i,j+1));
      
      if(i>0){
	vj=getGridVec(i-1,j);
	uij=getSigma(vj)/ui;
	sub2(&tvec,vj,vi);
	rhs+= uij*pow(len2(&tvec),2);
	mult2(&tvec,&tvec,uij);
	add2(&sum,&sum,&tvec);
      }
      if(j>0){
	vj=getGridVec(i,j-1);
	uij=getSigma(vj)/ui;
	sub2(&tvec,vj,vi);
	rhs+= uij*pow(len2(&tvec),2);
	mult2(&tvec,&tvec,uij);
	add2(&sum,&sum,&tvec);
      }
      if(i<gheight-1){
	vj=getGridVec(i+1,j);
	uij=getSigma(vj)/ui;
	sub2(&tvec,vj,vi);
	rhs+= uij*pow(len2(&tvec),2);
	mult2(&tvec,&tvec,uij);
	add2(&sum,&sum,&tvec);
      }
      if(j<gwidth-1){
	vj=getGridVec(i,j+1);
	uij=getSigma(vj)/ui;
	sub2(&tvec,vj,vi);
	rhs+= uij*pow(len2(&tvec),2);
	mult2(&tvec,&tvec,uij);
	add2(&sum,&sum,&tvec);
      }
      //Negate the right hand sum
      //rhs*=-1;
      mult2(&sub,&sub,rhs);
      add2(&sum,&sum,&sub);
      mult2(&sum,&sum,0.5);
      set2(dispvec+i*gwidth+j,&sum);
      total+=len2(&sum);
    }
  }  
  total = total/(gwidth*gheight);
  //printf("total=%f\n",total);
}

void scaleFreqImage()
{
  int i=0;
  int j=0;
  double sum;
  int count;
  int xinc=image.w/gwidth;
  int yinc=image.h/gheight;

  printf("About to scale freq image\n");

  for(i=0;i<gheight;i++){
    for(j=0;j<gwidth;j++){
      int ii;
      int iitop;
      int jj;
      int jjtop;
      count=0;
      sum=0;
      
      ii=i*yinc;
      iitop=ii+yinc;
      jjtop = j*xinc+xinc;
      while(ii<iitop){
	jj=j*xinc;
	
	while(jj<jjtop){
	  unsigned char c = image(jj,ii,0);
	  sum+=c;
	  count++;
	  jj++;
	}
	ii++;
      }
      sum = (sum/count)/255.0f;
      //printf("%f\n",sum);
      scaledImage[i*gwidth+j]=(sum);
    }
  }
  printf("done scaling freq image\n");
}

/* Grow the grid by a magnitude of two in each direction */
void growGrid()
{
  int oldw=gwidth;
  int oldh=gheight;
  double oldyinc;
  double oldxinc;
  double newyinc;
  double newxinc;

  double id=0;
  double jd=0;
  int i=0;
  int j=0;
  
  Vec2d * newdv = (Vec2d *)malloc(sizeof(Vec2d)*gwidth*gheight*4);
  Vec2d * newgrid = (Vec2d *)malloc(sizeof(Vec2d)*gwidth*gheight*4);

  gwidth*=2;
  gheight*=2;
  
  printf("Growing from %d,%d to %d %d\n",oldw,oldh,gwidth,gheight);
  
  oldyinc = 1.0/(double)(oldh-1.0);
  oldxinc = 1.0/(double)(oldw-1.0);

  newxinc = 1.0/(double)(gwidth-1.0);
  newyinc = 1.0/(double)(gheight-1.0);
  
  while(id<=1.0){
    int oldyind;
    int oldxind;
    int xind;
    int yind;
    jd=0;
    j=0;
    oldyind = (int)(id/oldyinc);

    if(oldyind<oldh-1){
      while(jd<=1.0){
	Vec2d * v1,* v2;
	Vec2d * v3,* v4;
	Vec2d temp1;
	Vec2d temp2;
	Vec2d temp3;
	Vec2d temp4;
	double b;
	double a;
	oldxind = (int)(jd/oldxinc);
	v1 = grid+oldw*oldyind+oldxind;
	v2 = grid+oldw*oldyind+oldxind+1;
	v3 = grid+oldw*(oldyind+1)+oldxind+1;
	v4 = grid+oldw*(oldyind+1)+oldxind;
	

	//End of a column, use the v2 - v3 to interpolate
	if(oldxind==oldw-1){

	  b=(id/oldyinc)-(double)oldyind;
	  mult2(&temp1,v1,1.0-b);
	  mult2(&temp2,v4,b);
	  add2(newgrid+i*gwidth+j,&temp1,&temp2);
	}
	else{
	  b=(id/oldyinc)-(double)oldyind;
	  a=(jd/oldxinc)-(double)oldxind;
	  mult2(&temp1,v1,(1-a)*(1-b));
	  mult2(&temp2,v2,(a)*(1-b));
	  mult2(&temp3,v3,(a)*(b));
	  mult2(&temp4,v4,(1-a)*(b));
	  add2(&temp1,&temp1,&temp2);
	  add2(&temp1,&temp1,&temp3);
	  add2(newgrid+i*gwidth+j,&temp1,&temp4);
	}
	temp1 = newgrid[i*gwidth+j];
	jd+=newxinc;
	j++;
      }
    }
    else{
      while(jd<=1.0){
	Vec2d * v1,* v2;
	Vec2d * v3,* v4;
	Vec2d temp1;
	Vec2d temp2;
	Vec2d temp3;
	Vec2d temp4;
	double b;
	double a;
	oldxind = (int)(jd/oldxinc);
	v1 = grid+oldw*oldyind+oldxind;
	v2 = grid+oldw*oldyind+oldxind+1;

	//End of a column, use the v1
	if(oldxind==oldw-1){
	  set2(newgrid+i*gwidth+j,v1);
	}
	else{
 	  a=(jd/oldxinc)-(double)oldxind;
	  mult2(&temp1,v1,(1-a));
	  mult2(&temp2,v2,(a));
	  add2(newgrid+i*gwidth+j,&temp1,&temp2);
	}
	temp1 = newgrid[i*gwidth+j];
	jd+=newxinc;
	j++;
      }
    }
    id+=newyinc;
    i++;
  }
  printf("done loops\n");
  free(grid);
  free(dispvec);
  printf("done freeing\n");
  grid = newgrid;
  dispvec = newdv;

  printf("about to scale image\n");
  scaleFreqImage();
  printf("done growing grid\n");
  //glutPostRedisplay();
}

double getSigma(Vec2d * v)
{
  double ret;
  int x=(int)(v->x*(gwidth-1));
  int y=(int)(v->y*(gheight-1));
  double c1=q;
  if(!v){
    printf("getting null v in sigma\n");
    return 1.0;
  }
  
  if(x!=(gwidth-1) && y!=(gheight-1)){
    double a;
    double b;
    int x1=x+1;
    int y1=y+1;
    a = (v->x*(gwidth-1))-x;
    b = (v->y*(gheight-1))-y;

    ret = (1-a)*(1-b)*scaledImage[y*gheight+x]+
      a*(1-b)*scaledImage[y*gheight+x1]+
      (a)*(b)*scaledImage[y1*gheight+x1]+
      (1-a)*(b)*scaledImage[y1*gheight+x];
    
    ret=ret*(1-c1)+c1;
  }
  else
    ret = scaledImage[y*gheight+x]*(1-c1)+c1;
  //  printf("%f %f\n",x-v->x*(gwidth-1),y-v->y*(gheight-1));
  return ret;
}

void moveTexCoords(Vec2d * grid,int gwidth,int gheight)
{
  int i=0;
  while(i<chart.nvert){
    double a;
    double b;
    Vec2f v;
    int x,y;

    v=chart.tc[i];
    x=(int)(v[0]*(gwidth-1));
    y=(int)(v[1]*(gheight-1));

    if((x+1)<gwidth && (y+1)<gheight){
      Vec2d * v1,* v2,* v3,* v4;
      float xn,yn;

      a = v[0]*(gwidth-1)-(float)x;
      b = v[1]*(gheight-1)-(float)y;
      v1 = grid+y*gwidth+x;
      v2 = grid+y*gwidth+x+1;
      v3 = grid+(y+1)*gwidth+x+1;
      v4 = grid+(y+1)*gwidth+x;
      
      xn = (1.0-a)*(1.0-b)*v1->x + (a)*(1.0-b)*v2->x+
	a*b*v3->x+(1-a)*b*v4->x;
      yn = (1.0-a)*(1.0-b)*v1->y + (a)*(1.0-b)*v2->y+
	a*b*v3->y+(1.0-a)*b*v4->y;
      v[0]=xn;
      v[1]=yn;
    }
    //Interpolate in the x direction only
    else if((x+1)<(gwidth-1)){
      Vec2d * v1,* v2,* v3,* v4;
      float xn,yn;
      
      a = v[0]*(gwidth-1)-(float)x;
      v1 = grid+y*gwidth+x;
      v2 = grid+y*gwidth+x+1;
      
      xn = (1.0-a)*v1->x + (a)*v2->x;
      yn = (1.0-a)*v1->y + (a)*v2->y;
      
      v[0]=xn;
      v[1]=yn;
      
    }
    else if((y+1)<(gheight-1)){
      Vec2d * v1,* v2,* v3,* v4;
      float xn,yn;
      
      b = v[1]*(gheight-1)-(float)y;
      
      v1 = grid+y*gwidth+x;
      v4 = grid+(y+1)*gwidth+x;
      
      xn = (1.0-b)*v1->x + (b)*v4->x;
      yn = (1.0-b)*v1->y + (b)*v4->y;
      
      v[0]=xn;
      v[1]=yn;
    }
    else{
      Vec2d * v1;
      printf("cant move by bi\n");
      v1=grid+y*gwidth+x;
      v[0]=v1->x;
      v[1]=v1->y;
    }
    chartout.tc[i]=v;
    i++;
  }
}

void drawAreas(Chart & m)
{
  float * areas3d;
  float * areas2d;
  float maxratio=-100000;
  float minratio=1000000;
  int i=0;
  
  areas3d = new float[m.npoly];
  areas2d = new float[m.npoly];
  
  while(i<m.npoly){
    Vec3f a,b,c;
    Vec2f a2,b2,c2;
    float la,lb,lc,s;
    a=m.vert[m.poly[i][0]]-m.vert[m.poly[i][1]];
    b=m.vert[m.poly[i][1]]-m.vert[m.poly[i][2]];
    c=m.vert[m.poly[i][0]]-m.vert[m.poly[i][2]];
    la = a.len();
    lb = b.len();
    lc = c.len();
    s=(la+lb+lc)/2;
    areas3d[i]=sqrt(s*(s-la)*(s-lb)*(s-lc));
    
    a2=m.tc[m.poly[i][0]]-m.tc[m.poly[i][1]];
    b2=m.tc[m.poly[i][1]]-m.tc[m.poly[i][2]];
    c2=m.tc[m.poly[i][0]]-m.tc[m.poly[i][2]];
    la = a2.len();
    lb = b2.len();
    lc = c2.len();
    s=(la+lb+lc)/2;
    areas2d[i]=sqrt(s*(s-la)*(s-lb)*(s-lc));
    
    if((areas2d[i]/areas3d[i])<minratio)
      minratio=areas2d[i]/areas3d[i];
    if((areas2d[i]/areas3d[i])>maxratio)
      maxratio=areas2d[i]/areas3d[i];
    i++;
  }
  
  
  glColor3f(1.0,1.0,1.0);
  
  i=0;
  
  /*
    Anything that is small in the texture but big in the real 
    world will have a small ratio, where triangles that are big 
    in the texture but small in the real world will have a large ratio.
    So anything with a large ratio should be mapped to a small number
  */
  glBegin(GL_TRIANGLES);
  while(i<m.npoly){
    float c = 1.0-(areas2d[i]/areas3d[i]-minratio)/(maxratio-minratio)*0.3;
    glColor3f(c,c,c);
    glVertex2fv(m.tc[m.poly[i][0]].data);
    glVertex2fv(m.tc[m.poly[i][1]].data);
    glVertex2fv(m.tc[m.poly[i][2]].data);
    i++;
  }
  glEnd();
  
  printf("ratios from %f to %f\n",minratio,maxratio);
  
  delete areas2d;
  delete areas3d;
}

void computeInverseAndMove()
{
    int i,j;
    int top=gwidth*gheight;
    Vec2d * inverse;
    float * coeff=(float *)malloc(sizeof(float)*gwidth*gheight);
    memset(coeff,0,sizeof(float)*gwidth*gheight);
    inverse = (Vec2d *)malloc(sizeof(Vec2d)*gwidth*gheight);
    memset(inverse,0,sizeof(Vec2d)*gwidth*gheight);
    
    for(i=0;i<gheight;i++){
      for(j=0;j<gwidth;j++){
	Vec2d * cur = inverse+i*gwidth+j;
	Vec2d * closest=NULL;
	float closestDist=20;
	float dist;
	int ci=0;
	int cj=0;
	
	int whichy;
	int whichx;
	Vec2d orig;

	Vec2d * test;
	Vec2d diff;
	int k;
	cur->x=((double)j)/(gwidth-1.0);
	cur->y=((double)i)/(gheight-1.0);

	test =grid+ci*gwidth+cj;
	sub2(&diff,test,cur);
	closest=test;
	closestDist = len2(&diff);

	while(1){
	  if((cj+1)<gwidth){
	    test = grid+ci*gwidth+(cj+1);
	    sub2(&diff,test,cur);
	    dist = len2(&diff);
	    if(dist<closestDist){
	      closestDist=dist;
	      closest = test;
	      cj++;
	      continue;
	    }
	  }
	  if((ci+1)<gheight){
	    test = grid+(ci+1)*gwidth+cj;
	    sub2(&diff,test,cur);
	    dist = len2(&diff);
	    if(dist<closestDist){
	      closestDist=dist;
	      closest = test;
	      ci++;
	      continue;
	    }
	  }
	  if((cj-1)>=0){
	    test = grid+ci*gwidth+(cj-1);
	    sub2(&diff,test,cur);
	    dist = len2(&diff);
	    if(dist<closestDist){
	      closestDist=dist;
	      closest = test;
	      cj--;
	      continue;
	    }
	  }
	  if((ci-1)>=0){
	    test = grid+(ci-1)*gwidth+cj;
	    sub2(&diff,test,cur);
	    dist = len2(&diff);
	    if(dist<closestDist){
	      closestDist=dist;
	      closest = test;
	      ci--;
	      continue;
	    }
	  }
	  break;
	}

	/*
	for(k=0;k<top;k++){
	  Vec2d diff;
	  float dist;
	  sub2(&diff,grid+k,cur);
	  dist = len2(&diff);
	  if(dist<closestDist){
	    closestDist=dist;
	    closest = grid+k;
	  }
	}
	*/
      
	whichy = (closest-grid)/gwidth;
	whichx = (closest-grid)-whichy*gwidth;
	orig.x=((double)whichx)/(gwidth-1.0f);
	orig.y=((double)whichy)/(gheight-1.0f);
	
	sub2(&diff,cur,grid+whichy*gwidth+whichx);
	add2(cur,&orig,&diff);
	printf(">%f %f\n",cur->x,cur->y);
      }
    }
    moveTexCoords(inverse,gwidth,gheight);
    
    printf("about to free\n");

    free(inverse);
    free(coeff);
}


void shutdown()
{
  printf("shutdown\n");
  if(scaledImage){
    free(scaledImage);
  }
  if(grid)
    free(grid);
  if(dispvec)
    free(dispvec);
  printf("half done\n");
  scaledImage=0;
  grid=0;
  dispvec=0;
  chart.release();
  chartout.release();
  //image=Image8(0,0,0);
  printf("done shutdonw\n");
}

void printUsageString()
{
  printf("Call sot with argument one as the texture coordinates, argument 2 as the image used as the stretching function(higher the value, the more the stretching) as gray scale image, must be cast into bytes.\nArgument 3 is an optional value of q, the default is 0.65, this value must be between [0-1], the lower the value the more stretching will occur.\n");
}

void sot(Chart & ch)
{
  ch.saveobj("/tmp/sot.obj");
  remove("/tmp/areas.tga");
  FILE * f=popen("drawareas /tmp/sot.obj","r");
  pclose(f);
  image.readTGA("/tmp/areas.tga");
  //image.writeTGA("areasread.tga");
  sot(ch,image);
}

void sot(Chart & ch,Image8 & im)
{
  int i=0,j=0;
  image=im;
  chart=ch;//eeh, this is sketch
  chartout.init(chart.npoly,chart.nvert);
  
  /* Initialize the variables */
  scaledImage = (double *)malloc(sizeof(double)*image.w*image.h);
  gwidth=2;
  gheight=2;
  grid = (Vec2d *)malloc(sizeof(Vec2d)*gwidth*gheight);
  dispvec = (Vec2d *) malloc(sizeof(Vec2d)*gwidth*gheight);
  memset(dispvec,0,sizeof(Vec2d)*gwidth*gheight);
  for(i=0;i<gheight;i++){
    for(j=0;j<gwidth;j++){
      grid[i*gwidth+j].x=(double)j/(double)(gwidth-1);
      grid[i*gwidth+j].y=(double)i/(double)(gheight-1);
    }
  }
  
  scaleFreqImage();

  /*
    Do the stretching
  */
  {
    int innermax[]={5,100,200,200,200,50};
    int i,j;
    for(j=0;j<6;j++){
      for(i=0;i<innermax[j]; i++){
	doIteration();
	move();
      }
      if(j<5)
	growGrid();
    }
  }
  /*
    Now compute the inverse and move the texture coordinates
  */
  printf("Computing inverse\n");
  computeInverseAndMove();
  printf("shutting down\n");
  
  memcpy(ch.tc,chartout.tc,sizeof(chart.tc[0])*chart.nvert);
  shutdown();
}

#ifdef SOT_MAIN
#include <GL/glut.h>
/*void mexFunction( int nlhs, mxArray *plhs[],        
  int nrhs, const mxArray *prhs[]){ */
int main(int ac,char *av[]){
  char string[255];
  int i=0;
  int j=0;
  int width;
  int height;
  int dims[3];

  chart.loadobj(av[1]);
  image.read(av[2]);
  chartout.init(chart.npoly,chart.nvert);
  memcpy(chartout.vert,chart.vert,sizeof(chart.vert[0])*chart.nvert);
  memcpy(chartout.poly,chart.poly,sizeof(chart.poly[0])*chart.npoly);

  /* Initialize the variables */
  scaledImage = (double *)malloc(sizeof(double)*image.w*image.h);
  gwidth=2;
  gheight=2;
  grid = (Vec2d *)malloc(sizeof(Vec2d)*gwidth*gheight);
  dispvec = (Vec2d *) malloc(sizeof(Vec2d)*gwidth*gheight);
  memset(dispvec,0,sizeof(Vec2d)*gwidth*gheight);
  for(i=0;i<gheight;i++){
    for(j=0;j<gwidth;j++){
      grid[i*gwidth+j].x=(double)j/(double)(gwidth-1);
      grid[i*gwidth+j].y=(double)i/(double)(gheight-1);
    }
  }
  
  scaleFreqImage();

  /*
    Do the stretching
  */
  {
    int innermax[]={5,100,200,200,200,50};
    int i,j;
    for(j=0;j<6;j++){
      for(i=0;i<innermax[j]; i++){
	doIteration();
	move();
      }
      if(j<5)
	growGrid();
    }
  }
  /*
    Now compute the inverse and move the texture coordinates
  */
  printf("Computing inverse\n");
  computeInverseAndMove();
  printf("shutting down\n");
  shutdown();
  return 0;
}

#endif
