#include "texgen.h"

#include "dgeev.h"
#include "dsyevx.h"
#include "dgemm.h"
#include "dsymm.h"
#include "complex.h"
#include "cgesv.h"
#include "shortestpath.h"
#include "image.h"
#include "timer.h"
#define max(a,b) ((a)>(b))?(a):(b)
#define min(a,b) ((a)<(b))?(a):(b)

void scaleCharts(Chart * charts,int ncharts,double scale,
		 int inplace,int increasing)
{
  printf("scaling charts\n");

  double totalArea=0;
  double diameters[ncharts];
  double centroids[ncharts][2];
  memset(centroids,0,sizeof(double)*ncharts*2);

  for(int i=0;i<ncharts;i++){
    double avg,min,max,worldarea,texarea;
    double scale,diam;

    charts[i].areaRatios(avg,min,max,worldarea,texarea);
    scale=sqrt(worldarea/texarea);

    double minx=10e10,miny=10e10;
    double maxx=-10e10,maxy=-10e10;
    
    for(int j=0;j<charts[i].nvert;j++){
      if(charts[i].tc[j].x<minx)minx=charts[i].tc[j].x;
      if(charts[i].tc[j].y<miny)miny=charts[i].tc[j].y;
      maxx=MAX(charts[i].tc[j].x,maxx);
      maxy=MAX(charts[i].tc[j].y,maxy);
      centroids[i][0]+=charts[i].tc[j].x;
      centroids[i][1]+=charts[i].tc[j].y;
    }
    centroids[i][0]=minx;//=charts[i].nvert;
    centroids[i][1]=miny;//=charts[i].nvert;

    diam=0;
    for(int j=0;j<charts[i].nvert;j++){
      charts[i].tc[j].x-=minx;
      charts[i].tc[j].y-=miny;
      charts[i].tc[j]*=(scale);
      if(charts[i].tc[j].x>diam)diam=charts[i].tc[j].x;
      if(charts[i].tc[j].y>diam)diam=charts[i].tc[j].y;
    }
    diameters[i]=diam;

    //Fixed a bug here, May 10/07
    //When the chart has dimensions bigger than 1 the packChartsNew 
    //doesn't work properly
    if(diam>1){
      for(int j=0;j<charts[i].nvert;j++){
	charts[i].tc[j].x/=diam;
	charts[i].tc[j].y/=diam;
      }
      diameters[i]=1;
    }
    charts[i].areaRatios(avg,min,max,worldarea,texarea);
    totalArea+=texarea;

    printf("chart[%d] %f\n",i,(maxx-minx)/(maxy-miny));
  }
  for(int i=0;i<ncharts;i++){
      for(int j=0;j<charts[i].nvert;j++){
	  charts[i].tc[j]*=(scale/sqrt(totalArea));
      }
  }
  if(inplace){
      for(int i=0;i<ncharts;i++){
	  for(int j=0;j<charts[i].nvert;j++){
	      charts[i].tc[j].x+=centroids[i][0];
	      charts[i].tc[j].y+=centroids[i][1];
	  }
      }
  }
  
  if(increasing){
    //Sort the charts in increasing order of diameter
    Chart temp;
    for(int i=0;i<ncharts;i++){
      int largest=i;
      for(int j=i+1;j<ncharts;j++){
	if(diameters[j]<diameters[largest]){
	  largest=j;
	}
      }
      temp=charts[i];
      charts[i]=charts[largest];
      charts[largest]=temp;
    }
  }
  else{
    Chart temp;
    for(int i=0;i<ncharts;i++){
      int largest=i;
      for(int j=i+1;j<ncharts;j++){
	if(diameters[j]>diameters[largest]){
	  largest=j;
	}
      }
      temp=charts[i];
      charts[i]=charts[largest];
      charts[largest]=temp;
      double d=diameters[i];
      diameters[i]=diameters[largest];
      diameters[largest]=d;
    }
  }
}

void lineFrom(Vec2f * mtvert,
	      double * minb,double * maxb,int v1,int v2,int twidth)
{
  double slope;
  Vec2f p1,p2;
  p1=mtvert[v1];
  p2=mtvert[v2];

  if(p1.x>p2.x){
    Vec2f temp;
    temp = p1;
    p1=p2;
    p2=temp;
  }

  slope = (p2.y-p1.y)/(p2.x-p1.x);
#ifdef PD
  printf("%f slope\n",slope);
#endif
  while(p1.x<p2.x){
    if((int)p1.x>=twidth ||
       (int)p1.x<0){
#ifdef PD
      printf("bad data %d\n",(int)p1.x);
#endif
      break;
    }
    if(maxb[(int)p1.x]<p1.y)
      maxb[(int)p1.x]=p1.y;
    if(minb[(int)p1.x]>p1.y)
      minb[(int)p1.x]=p1.y;
    p1.y+=slope;
    p1.x+=1;
  }
}

double finduv(double * horizon,double * minb, double * maxb,
	      double y, double width,double height,int twidth,
	      int theight,int &u,int &v)
{
  int i,j;
  int to=(int)(twidth-width);
  double minarea=2*twidth*theight;
  for(j=0;j<to;j++){
    int top =(int) ( max(theight,horizon[j]+height));

    for(i=(int)(max(horizon[j]-y,0));i<top;i+=4){
      int k=0;
      int intersect=0;
      double above=0;
      double areawasted=0;
      
      while(k<width){
	double minbk_i = minb[k]+i;
	double thistime = minbk_i-horizon[j+k];
	areawasted+=thistime;

	if(minbk_i>=theight){
	  above +=(minbk_i-theight+1);
	}
	if(thistime<0){
	  intersect=1;
	  break;
	}
	k++;
      }
      //printf("above %f\n",above);
      if(!intersect){
	if(areawasted>=0){
	  areawasted=areawasted+1*above;//used to be 10*above
	  if(areawasted<minarea){
	    minarea = areawasted;
	    u=j;
	    v=i;
#ifdef PD
	    printf("setting (%d,%d) %f\n",j,i,areawasted);
#endif
	  }
	}
      }
    }
  }
  return minarea;
}

double finduv(Chart & c,
	      Vec2f * tc,
	      int twidth,int theight,
	      double * horizon,
	      int & u,
	      int & v,
	      double * minb,
	      double * maxb,
	      int width)
{
  int i=0;
  double height=0;
  double y;

  i=0;
  while(i<twidth){
    minb[i]=theight-1;
    maxb[i]=0;
    i++;
  }

  i=0;
  while(i<c.npoly){
    int v1=(int)c.poly[i][0];
    int v2=(int)c.poly[i][1];
    int v3=(int)c.poly[i][2];
    lineFrom(tc,minb,maxb,v1,v2,twidth);
    lineFrom(tc,minb,maxb,v2,v3,twidth);
    lineFrom(tc,minb,maxb,v1,v3,twidth);
    i++;
  }
  y = (int)minb[0];
  i=0;
  width=twidth;
  height=0;
  i=twidth-1;
  while(i>=0){
    if(maxb[i]!=0){
      if(width==twidth)
	width=i+1;
    }
    if(height<(int)(maxb[i]-minb[i])){
      height = (int)(maxb[i]-minb[i]);
    }
    i--;
  }
  if(width==0)
    width=1;
  if(height==0)
    height=1;

  u=-1;
  v=-1;
  double area=finduv(horizon,minb,maxb,y,width,height,twidth,theight,u,v);
  return area;
}


//if chart is -1, then all the charts are packed
void packCharts(Chart * charts,int ncharts,
		int twidth,int theight)
{
  int i=0;
  int orient=4;
  int curx=0;
  int cury=0;
  double horizon[twidth];

  printf("packing charts %d/%d\n",i,ncharts);

  //if(i>=ncharts)i=0;
  //if(i==0){
    printf("scaling and reseting horizon\n");
    scaleCharts(charts,ncharts);
    memset(horizon,0,sizeof(double)*twidth);
    //}
    printf("about to pack charts\n");
  for(i=0;i<ncharts;i++){
    double area;
    double areaGood=2*twidth*theight;
    int u,v;
    int uGood=-1,vGood=-1;
    double minb[twidth];
    double maxb[twidth];
    double minbGood[twidth];
    double maxbGood[twidth];
    int width;

    Vec2f tcoords[charts[i].nvert];
    Vec2f tcoordsGood[charts[i].nvert];

    for(int k=0;k<orient;k++){
      double minx=1000000;
      double miny=1000000;
      for(int j=0;j<charts[i].nvert;j++){
	tcoords[j].x=charts[i].tc[j].x*(twidth-1);
	tcoords[j].y=charts[i].tc[j].y*(theight-1);
	tcoords[j].rotate((double)k*2.0*M_PI/(double)(orient));
	if(tcoords[j].x<minx)
	  minx=tcoords[j].x;
	if(tcoords[j].y<miny)
	  miny=tcoords[j].y;
      }
      //subtract the minimum out
      for(int j=0;j<charts[i].nvert;j++){
	tcoords[j].x-=minx;
	tcoords[j].y-=miny;
      }
      
      area=finduv(charts[i],
		  tcoords,
		  twidth,theight,
		  horizon,
		  u,
		  v,
		  minb,
		  maxb,
		  width);
      if(uGood==-1 && vGood==-1){
	  memcpy(tcoordsGood,tcoords,sizeof(Vec2f)*charts[i].nvert);
	  memcpy(minbGood,minb,sizeof(double)*twidth);
	  memcpy(maxbGood,maxb,sizeof(double)*twidth);
	  uGood=u;
	  vGood=v;
	  areaGood=area;
      }
      if((area<areaGood) && u!=-1 && v!=-1 && u<twidth &&
	 v<theight){
	memcpy(tcoordsGood,tcoords,sizeof(Vec2f)*charts[i].nvert);
	memcpy(minbGood,minb,sizeof(double)*twidth);
	memcpy(maxbGood,maxb,sizeof(double)*twidth);
	uGood=u;
	vGood=v;
	areaGood=area;
      }
    }
    if(uGood==-1 && vGood==-1){
	uGood=vGood=0;
    }
    if(uGood!=-1 && vGood!=-1){
      int width=twidth;
      for(int j=twidth-1;j>=1;j--){
	if(maxbGood[j]!=0){
	  width=j+1;
	  break;
	}
      }
      for(int j=uGood;j<uGood+width;j++){
	horizon[j]=max(maxbGood[j-uGood]+vGood,horizon[j]);
      }
      for(int j=0;j<charts[i].nvert;j++){
	charts[i].tc[j].x=tcoordsGood[j].x+uGood;
	charts[i].tc[j].y=tcoordsGood[j].y+vGood;
      }
      printf("packing %d at %d %d (%d)\n",i,uGood,vGood,width);
    }
    
    printf("returning from %d\n",i);
  }
  if(i==ncharts){
    for(int k=0;k<ncharts;k++){
      double minx=10e10;
      double maxx=-10e10;
      double miny=10e10;
      double maxy=-10e10;
      double meanx=0;
      double meany=0;
      for(int j=0;j<charts[k].nvert;j++){
	if(charts[k].tc[j].y>maxy)
	  maxy=charts[k].tc[j].y;
	if(charts[k].tc[j].x>maxx)
	  maxx=charts[k].tc[j].x;
	if(charts[k].tc[j].y<miny)
	  miny=charts[k].tc[j].y;
	if(charts[k].tc[j].x<minx)
	  minx=charts[k].tc[j].x;
	meanx+=charts[k].tc[j].x;
	meany+=charts[k].tc[j].y;
      }
      meanx/=charts[k].nvert;
      meany/=charts[k].nvert;
      for(int j=0;j<charts[k].nvert;j++){
	double x=charts[k].tc[j].x-meanx;
	double y=charts[k].tc[j].y-meany;
	charts[k].tc[j].x=x*((maxx-minx)-6)/((maxx-minx))+meanx;
	charts[k].tc[j].y=y*((maxy-miny)-6)/((maxy-miny))+meany;
      }
    }
    


    //make fit in the texture size
    double maxy=0;
    double maxx=0;
    for(int k=0;k<ncharts;k++){
      for(int j=0;j<charts[k].nvert;j++){
	if(charts[k].tc[j].y>maxy)
	  maxy=charts[k].tc[j].y;
	if(charts[k].tc[j].x>maxx)
	  maxx=charts[k].tc[j].x;
      }
    }
    printf("%f\n",maxy);
    double scaley=1.0/maxy;
    double scalex=1.0/maxx;
    for(int k=0;k<ncharts;k++){
      for(int j=0;j<charts[k].nvert;j++){
	charts[k].tc[j].y*=scaley;
	charts[k].tc[j].x*=scalex;
      }
    }
  }
  printf("done packing charts\n");
}


//if chart is -1, then all the charts are packed
double * packChartsDebug(Chart * charts,int ncharts,
		     int twidth,int theight,int i)
{
  static int orient;
  static int curx;
  static int cury;
  static double * horizon=0;
  if(i>=ncharts)return 0;;
  if(i==0){
    curx=0;
    cury=0;
    orient=4;
    delete horizon;
    horizon = new double[twidth];
    memset(horizon,0,sizeof(double)*twidth);
    printf("scaling and reseting horizon\n");
    scaleCharts(charts,ncharts);
  }
  printf("packing charts %d/%d\n",i,ncharts);
  printf("about to pack charts\n");
  int top=i+1;
  for(;i<top;i++){
    double area;
    double areaGood=2*twidth*theight;
    int u,v;
    int uGood=-1,vGood=-1;
    double minb[twidth];
    double maxb[twidth];
    double minbGood[twidth];
    double maxbGood[twidth];
    int width;

    Vec2f tcoords[charts[i].nvert];
    Vec2f tcoordsGood[charts[i].nvert];

    for(int k=0;k<orient;k++){
      double minx=1000000;
      double miny=1000000;
      for(int j=0;j<charts[i].nvert;j++){
	tcoords[j].x=charts[i].tc[j].x*(twidth-1);
	tcoords[j].y=charts[i].tc[j].y*(theight-1);
	tcoords[j].rotate((double)k*2.0*M_PI/(double)(orient));
	if(tcoords[j].x<minx)
	  minx=tcoords[j].x;
	if(tcoords[j].y<miny)
	  miny=tcoords[j].y;
      }
      //subtract the minimum out
      for(int j=0;j<charts[i].nvert;j++){
	tcoords[j].x-=minx;
	tcoords[j].y-=miny;
      }
      
      area=finduv(charts[i],
		  tcoords,
		  twidth,theight,
		  horizon,
		  u,
		  v,
		  minb,
		  maxb,
		  width);
      if(uGood==-1 && vGood==-1){
	  memcpy(tcoordsGood,tcoords,sizeof(Vec2f)*charts[i].nvert);
	  memcpy(minbGood,minb,sizeof(double)*twidth);
	  memcpy(maxbGood,maxb,sizeof(double)*twidth);
	  uGood=u;
	  vGood=v;
	  areaGood=area;
      }
      if((area<areaGood) && u!=-1 && v!=-1 && u<twidth &&
	 v<theight){
	memcpy(tcoordsGood,tcoords,sizeof(Vec2f)*charts[i].nvert);
	memcpy(minbGood,minb,sizeof(double)*twidth);
	memcpy(maxbGood,maxb,sizeof(double)*twidth);
	uGood=u;
	vGood=v;
	areaGood=area;
      }
    }
    if(uGood==-1 && vGood==-1){
	uGood=vGood=0;
    }
    if(uGood!=-1 && vGood!=-1){
      int width=twidth;
      for(int j=twidth-1;j>=1;j--){
	if(maxbGood[j]!=0){
	  width=j+1;
	  break;
	}
      }
      for(int j=uGood;j<uGood+width;j++){
	horizon[j]=max(maxbGood[j-uGood]+vGood,horizon[j]);
      }
      for(int j=0;j<charts[i].nvert;j++){
	charts[i].tc[j].x=tcoordsGood[j].x+uGood;
	charts[i].tc[j].y=tcoordsGood[j].y+vGood;
      }
      printf("packing %d at %d %d (%d)\n",i,uGood,vGood,width);
    }
    
    printf("returning from %d\n",i);
  }
  if(i==ncharts){
    for(int k=0;k<ncharts;k++){
      double minx=10e10;
      double maxx=-10e10;
      double miny=10e10;
      double maxy=-10e10;
      double meanx=0;
      double meany=0;
      for(int j=0;j<charts[k].nvert;j++){
	if(charts[k].tc[j].y>maxy)
	  maxy=charts[k].tc[j].y;
	if(charts[k].tc[j].x>maxx)
	  maxx=charts[k].tc[j].x;
	if(charts[k].tc[j].y<miny)
	  miny=charts[k].tc[j].y;
	if(charts[k].tc[j].x<minx)
	  minx=charts[k].tc[j].x;
	meanx+=charts[k].tc[j].x;
	meany+=charts[k].tc[j].y;
      }
      meanx/=charts[k].nvert;
      meany/=charts[k].nvert;
      for(int j=0;j<charts[k].nvert;j++){
	double x=charts[k].tc[j].x-meanx;
	double y=charts[k].tc[j].y-meany;
	charts[k].tc[j].x=x*((maxx-minx)-6)/((maxx-minx))+meanx;
	charts[k].tc[j].y=y*((maxy-miny)-6)/((maxy-miny))+meany;
      }
    }
    
    //make fit in the texture size
    double maxy=0;
    double maxx=0;
    for(int k=0;k<ncharts;k++){
      for(int j=0;j<charts[k].nvert;j++){
	if(charts[k].tc[j].y>maxy)
	  maxy=charts[k].tc[j].y;
	if(charts[k].tc[j].x>maxx)
	  maxx=charts[k].tc[j].x;
      }
    }
    printf("%f\n",maxy);
    double scaley=1.0/maxy;
    double scalex=1.0/maxx;
    for(int k=0;k<ncharts;k++){
      for(int j=0;j<charts[k].nvert;j++){
	charts[k].tc[j].y*=scaley;
	charts[k].tc[j].x*=scalex;
      }
    }
  }
  printf("done packing charts\n");
  return horizon;
}

void packChartsNew(Chart * charts,int ncharts,int texw,int texh)
{
  int i=0;
  Image8 packed;
  Image8 im;
  int * xs=new int[ncharts];
  int * ys=new int[ncharts];
  double scaleDown=0.95;

  memset(xs,0,sizeof(int)*ncharts);
  memset(ys,0,sizeof(int)*ncharts);

  Timer timer;
  timer.start();
  
  //if(i==0){
  packed=Image8(texw,texh,1);
  packed=0;
  scaleCharts(charts,ncharts,1.0,0,0);
  //}
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0,1,0,1,-1,1);
  glViewport(0,0,texw,texh);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glClearColor(0.0,0.0,0.0,0.0);

  for(i=0;i<ncharts;i++){
    printf("top of loop\n");fflush(stdout);//NEW
    if(i==0)timer.start();
  //if(i<ncharts){
    float maxx=-1;
    float maxy=-1;
    printf("about to get size\n");fflush(stdout);
    for(int j=0;j<charts[i].nvert;j++){
      maxx=MAX(charts[i].tc[j].x,maxx);
      maxy=MAX(charts[i].tc[j].y,maxy);
    }
    printf("got size %f %f\n",maxx,maxy);fflush(stdout);
    int w=(maxx*texw);
    int h=(maxy*texh);
    if(w<1)w=1;
    if(h<1)h=1;
    printf("w:%d h:%d\n",w,h);fflush(stdout);//NEW
    im=Image8(w,h,1);
    //    printf("chart %d, max %d %d\n",i,w,h);
    
    glClear(GL_COLOR_BUFFER_BIT);
    glColor3f(1.0,1.0,1.0);
    charts[i].drawTC();
    glPixelStorei(GL_PACK_ALIGNMENT,1);
    glReadPixels(0,0,im.w,im.h,GL_RED,GL_UNSIGNED_BYTE,im.data);
    if(w==1||h==1)im=1;//make sure there is some pixels that are on
    printf("got it\n");fflush(stdout);//NEW
    /*char fname[1024];
    sprintf(fname,"/tmp/chart-%04d.ppm",i);
    im.writePPM(fname);
    sprintf(fname,"/tmp/packed-%04d.ppm",i);
    packed.writePPM(fname);*/

    int y;
    for(y=0;y<texh-im.h;y++){
      int x=0;
      for(x=0;x<texw-im.w;x++){
	int y2,x2;
	int mod=32;
	int smod=0;
	while(mod>=1){
	  for(y2=mod;y2<im.h;y2+=mod){
	    for(x2=mod;x2<im.w;x2+=mod){
	      int ch=im(x2,y2);
	      int pch=packed(x+x2,y+y2,1);
	      if(ch&&pch)break;
	    }
	    if(x2<im.w)break;
	  }
	  if(y2<im.h)break;
	  mod=mod>>1;
	}
	//it was okay
	if(mod==0){
	  for(y2=0;y2<im.h;y2++){
	    for(x2=0;x2<im.w;x2++){
	      packed(x2+x,y2+y,1)|=im(x2,y2);
	    }
	  }
	  for(int j=0;j<charts[i].nvert;j++){
	    charts[i].tc[j].x=(charts[i].tc[j].x*texw+x);
	    charts[i].tc[j].y=(charts[i].tc[j].y*texh+y);
	  }
	  //printf("%d okay at %d %d\n",i,x,y);
	  xs[i]=x;
	  ys[i]=y;
	  break;
	}
      }
      if(x<texw-im.w)break;
    }
    if(y>=texh-im.h){
      printf("COULD NOT FIND %d, RESTARTING %f\n",i,scaleDown);
      packed=0;
      for(int j=0;j<i;j++){
	for(int k=0;k<charts[j].nvert;k++){
	  charts[j].tc[k].x=(scaleDown*(charts[j].tc[k].x-xs[j])/texw);
	  charts[j].tc[k].y=(scaleDown*(charts[j].tc[k].y-ys[j])/texh);
	}
      }
      for(int j=i;j<ncharts;j++){
	for(int k=0;k<charts[j].nvert;k++){
	  charts[j].tc[k].x=scaleDown*(charts[j].tc[k].x);
	  charts[j].tc[k].y=scaleDown*(charts[j].tc[k].y);
	}
      }
      i=-1;
      
      printf("rescaled\n");
    }
    
  }
  printf("Done packing charts, elapsed time for final it %g\n",timer.stop());
  printf("texw %d, texh %d\n",texw,texh);


  if(i==ncharts){
    for(int k=0;k<ncharts;k++){
      double minx=10e10;
      double maxx=-10e10;
      double miny=10e10;
      double maxy=-10e10;
      double meanx=0;
      double meany=0;
      for(int j=0;j<charts[k].nvert;j++){
	if(charts[k].tc[j].y>maxy)
	  maxy=charts[k].tc[j].y;
	if(charts[k].tc[j].x>maxx)
	  maxx=charts[k].tc[j].x;
	if(charts[k].tc[j].y<miny)
	  miny=charts[k].tc[j].y;
	if(charts[k].tc[j].x<minx)
	  minx=charts[k].tc[j].x;
	meanx+=charts[k].tc[j].x;
	meany+=charts[k].tc[j].y;
      }
      meanx/=charts[k].nvert;
      meany/=charts[k].nvert;
      double rangex=maxx-minx;
      double rangey=maxy-miny;
      double scalex=((maxx-minx)-6)/((maxx-minx));
      double scaley=((maxy-miny)-6)/((maxy-miny));
      double scaling=scalex;
      if(scaling<0 || scaley>scalex)
	scaling=scaley;
      
      //This scaling should be uniform, otherwise
      //continually packing will end up with degenerate charts..
      if(rangex>texw/20 && rangey>texh/20 && scaling>0){
	for(int j=0;j<charts[k].nvert;j++){
	  double x=charts[k].tc[j].x-meanx;
	  double y=charts[k].tc[j].y-meany;
	  charts[k].tc[j].x=x*scaling+meanx;
	  charts[k].tc[j].y=y*scaling+meany;
	}
      }
    }		 
   
    //make fit in the texture size
    double maxy=0;
    double maxx=0;
    for(int k=0;k<ncharts;k++){
      for(int j=0;j<charts[k].nvert;j++){
	if(charts[k].tc[j].y>maxy)
	  maxy=charts[k].tc[j].y;
	if(charts[k].tc[j].x>maxx)
	  maxx=charts[k].tc[j].x;
      }
    }
    printf("%f\n",maxy);
    
    double scaley=1.0/maxy;//=MAX(maxx,maxy);
    double scalex=1.0/maxx;//=scaley;
    double scaling=MIN(scalex,scaley);
    //again, this is made to be uniform
    for(int k=0;k<ncharts;k++){
      for(int j=0;j<charts[k].nvert;j++){
	charts[k].tc[j].y*=scaling;
	charts[k].tc[j].x*=scaling;
      }
    }
  }

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  int err=glGetError();
  if(err)printf("error %d\n",err);
  
  delete [] xs;
  delete [] ys;
}

typedef complex<float> complex_f;
#define I_f complex_f(0,1)

void getTexCoordsLSCM(Chart & ch)
{
  if(ch.nvert<=0)return;
  
  complex_f * M =new complex_f[ch.nvert*ch.nvert];
  Graph * g = allocGraph(ch.nvert);
  int * ecount = new int[ch.nvert];
  int * edgecount = new int[ch.nvert*ch.nvert];
  double * dist = new double[ch.nvert*ch.nvert];
  int * pred = new int[ch.nvert*ch.nvert];
  
  printf("ecount %p %d\n",ecount,sizeof(ecount));
  printf("ch.npoly %d,ch.nvert %d\n",ch.npoly,ch.nvert);

  memset(ecount,0,sizeof(int)*ch.nvert);
  memset(edgecount,0,sizeof(int)*ch.nvert*ch.nvert);

  for(int i=0;i<ch.npoly;i++){
    Vec3f pv1,pv2,pv3;
    Vec3f sa,sb,sc;
    double a;
    double b;
    double c;
    double av1,av2,av3;
    int v1=ch.poly[i][0],v2=ch.poly[i][1],v3=ch.poly[i][2];
    graphAddEdge(g,v1,v2,
		 (ch.vert[v1]-ch.vert[v2]).len());
    graphAddEdge(g,v1,v3,
		 (ch.vert[v1]-ch.vert[v3]).len());
    graphAddEdge(g,v2,v3,
		 (ch.vert[v2]-ch.vert[v3]).len());
   
    ecount[v1]++;
    ecount[v2]++;
    ecount[v3]++;
    
    pv1=ch.vert[v1];
    pv2=ch.vert[v2];
    pv3=ch.vert[v3];
    
    sa=pv1-pv2;
    sb=pv1-pv3;
    sc=pv2-pv3;
    
    a=sa.len();
    b=sb.len();
    c=sc.len();

    //get angles at the vertices
    av1=acos((c*c-a*a-b*b)/(-2*a*b));
    av2=acos((b*b-a*a-c*c)/(-2*a*c));
    av3=acos((a*a-b*b-c*c)/(-2*b*c));
    
    M[v1*ch.nvert+v1]=M[v1*ch.nvert+v1]+1.0/tan(av2)+1.0/tan(av3);
    M[v1*ch.nvert+v2]=M[v1*ch.nvert+v2]-1.0/tan(av3)+I_f;
    M[v1*ch.nvert+v3]=M[v1*ch.nvert+v3]-1.0/tan(av2)-I_f;
    
    M[v2*ch.nvert+v2]=M[v2*ch.nvert+v2]+1.0/tan(av3)+1.0/tan(av1);
    M[v2*ch.nvert+v3]=M[v2*ch.nvert+v3]-1.0/tan(av1)+I_f;
    M[v2*ch.nvert+v1]=M[v2*ch.nvert+v1]-1.0/tan(av3)-I_f;
    
    M[v3*ch.nvert+v3]=M[v3*ch.nvert+v3]+1.0/tan(av1)+1.0/tan(av2);
    M[v3*ch.nvert+v1]=M[v3*ch.nvert+v1]-1.0/tan(av2)+I_f;
    M[v3*ch.nvert+v2]=M[v3*ch.nvert+v2]-1.0/tan(av1)-I_f;

    if(v1<v2){
      edgecount[v1*ch.nvert+v2]++;
    }
    else{
      edgecount[v2*ch.nvert+v1]++;
    }
    if(v2<v3){
      edgecount[v2*ch.nvert+v3]++;
    }      
    else{
      edgecount[v3*ch.nvert+v2]++;
    }
    if(v1<v3){
      edgecount[v1*ch.nvert+v3]++;
    }
    else{
      edgecount[v3*ch.nvert+v1]++;
    }
  }
  //Get the distances between vertices
  allocSingleSource(g);
  for(int i=0;i<ch.nvert;i++){
    dij(g,i);
    getDistancesParam(dist+i*ch.nvert);
    getPredParam(pred+i*ch.nvert);
    printf("%5d\r",i);
    if(i%100)fflush(stdout);
  }
  freeSingleSource();
  freeGraph(g);
  
  int pinned[]={-1,-1};
  double biggestDistance=0;

  for(int i=0;i<ch.nvert;i++){
    for(int j=i+1;j<ch.nvert;j++){
      if(dist[i*ch.nvert+j]>biggestDistance){
	biggestDistance=dist[i*ch.nvert+j];
	pinned[0]=i;
	pinned[1]=j;
      }
    }
  }
  printf("pinned %d %d\n",pinned[0],pinned[1]);
  
  complex_f * b=new complex_f[ch.nvert];
  memset(b,0,sizeof(complex_f)*ch.nvert);
  b[pinned[0]]=complex_f(0.5f,0);
  b[pinned[1]]=complex_f(0.5f,1);

  for(int i=0;i<ch.nvert;i++){
    M[pinned[0]*ch.nvert+i]=complex_f(0.0,0.0);
    M[pinned[1]*ch.nvert+i]=complex_f(0.0,0.0);
  }
  M[pinned[0]*ch.nvert+pinned[0]]=1;
  M[pinned[1]*ch.nvert+pinned[1]]=1;

  complex_f * Mt = new complex_f[ch.nvert*ch.nvert];
  for(int i=0;i<ch.nvert;i++){
    for(int j=0;j<ch.nvert;j++){
      Mt[i*ch.nvert+j]=M[j*ch.nvert+i];
    }
  }
  //memcpy(Mt,M,sizeof(complex_f)*ch.nvert);


  //u=M\b...
  //tch.x = real(u)
  //tch.y = imag(u)
  int * pivot=new int[ch.nvert];
  int info;
  complex_f * result =new complex_f[ch.nvert];
  memcpy(result,b,sizeof(complex_f)*ch.nvert);

  cgesv_(ch.nvert,1,(complex<float>*)Mt,ch.nvert,pivot,result,ch.nvert,info);

  Matrix tc(ch.nvert,2);
  for(int i=0;i<ch.nvert;i++){
    tc(i,0)=result[i].real;
    tc(i,1)=result[i].imag;
  }
  printf("setting tc\n");
  ch.setTC(tc);
  printf("done settint tc\n");
  /*
  FILE * file=fopen("testtexgen.m","w");
  
  fprintf(file,"pinned=[%d,%d]\n",pinned[0],pinned[1]);
  fprintf(file,"big=%f\n",biggestDistance);
  fprintf(file,"M=[");
  for(int i=0;i<ch.nvert;i++){
    for(int j=0;j<ch.nvert;j++){
      fprintf(file,"%12.10f+(%12.10fi) ",M[i][j].real,
	      M[i][j].imag);
    }
    fprintf(file,"\n");
  }
  fprintf(file,"];\n");
  
  fprintf(file,"b=[");
  for(int i=0;i<ch.nvert;i++){
    fprintf(file,"%12.10f+(%12.10fi)\n",b[i].real,
	    b[i].imag);
    fprintf(file,"\n");
  }
  fprintf(file,"];\n");

  fprintf(file,"u=[");
  for(int i=0;i<ch.nvert;i++){
    fprintf(file,"%12.10f+(%12.10fi)\n",result[i].real,
	    result[i].imag);
    fprintf(file,"\n");
  }
  fprintf(file,"];\n");

  fprintf(file,"tri=[");
  for(int i=0;i<ch.npoly;i++){
    fprintf(file,"%d %d %d\n",
	    ch.poly[i][0],ch.poly[i][1],ch.poly[i][2]);
    fprintf(file,"\n");
  }
  fprintf(file,"];\n");
  fclose(file);
  */
  delete ecount;
  delete edgecount;
  delete M;
  delete Mt;
  delete b;
  delete result;
  delete pivot;
  delete dist;
  delete pred;
}

void getTexCoordsLSCM_circle(Chart & ch)
{
  if(ch.nvert<=0)return;
  
  complex_f * M =new complex_f[ch.nvert*ch.nvert];
  float * edges = new float[ch.nvert*ch.nvert];
    
  printf("ch.npoly %d,ch.nvert %d\n",ch.npoly,ch.nvert);

  for(int i=0;i<ch.nvert*ch.nvert;i++)
    edges[i]=-1;

  for(int i=0;i<ch.npoly;i++){
    Vec3f pv1,pv2,pv3;
    Vec3f sa,sb,sc;
    double a;
    double b;
    double c;
    double av1,av2,av3;
    int v1=ch.poly[i][0],v2=ch.poly[i][1],v3=ch.poly[i][2];
    
    pv1=ch.vert[v1];
    pv2=ch.vert[v2];
    pv3=ch.vert[v3];
    
    sa=pv1-pv2;
    sb=pv1-pv3;
    sc=pv2-pv3;
    
    a=sa.len();
    b=sb.len();
    c=sc.len();

    //get angles at the vertices
    av1=acos((c*c-a*a-b*b)/(-2*a*b));
    av2=acos((b*b-a*a-c*c)/(-2*a*c));
    av3=acos((a*a-b*b-c*c)/(-2*b*c));
    
    M[v1*ch.nvert+v1]=M[v1*ch.nvert+v1]+1.0/tan(av2)+1.0/tan(av3);
    M[v1*ch.nvert+v2]=M[v1*ch.nvert+v2]-1.0/tan(av3)+I_f;
    M[v1*ch.nvert+v3]=M[v1*ch.nvert+v3]-1.0/tan(av2)-I_f;
    
    M[v2*ch.nvert+v2]=M[v2*ch.nvert+v2]+1.0/tan(av3)+1.0/tan(av1);
    M[v2*ch.nvert+v3]=M[v2*ch.nvert+v3]-1.0/tan(av1)+I_f;
    M[v2*ch.nvert+v1]=M[v2*ch.nvert+v1]-1.0/tan(av3)-I_f;
    
    M[v3*ch.nvert+v3]=M[v3*ch.nvert+v3]+1.0/tan(av1)+1.0/tan(av2);
    M[v3*ch.nvert+v1]=M[v3*ch.nvert+v1]-1.0/tan(av2)+I_f;
    M[v3*ch.nvert+v2]=M[v3*ch.nvert+v2]-1.0/tan(av1)-I_f;

    edges[v1*ch.nvert+v2]=a;
    edges[v2*ch.nvert+v3]=b;
    edges[v3*ch.nvert+v1]=c;
  }
  int pinned[]={-1,-1};
  double biggestDistance=0;
  int * boundary=new int[ch.nvert];
  memset(boundary,-1,sizeof(int)*ch.nvert);
  int nboundary=0;
  int start=-1;
  double boundaryLength=0;
  for(int i=0;i<ch.nvert;i++){
    for(int j=0;j<ch.nvert;j++){
      if(edges[i*ch.nvert+j]>=0 && 
	 edges[j*ch.nvert+i]<0){
	if(boundary[i]>=0){
	  printf("error in boundary, one vertex cannot go to two others\n");
	}
	if(start==-1)start=i;
	boundary[i]=j;
	boundaryLength+=edges[i*ch.nvert+j];
	nboundary++;
      }
    }
  }
  
  complex_f * b=new complex_f[ch.nvert];
  memset(b,0,sizeof(complex_f)*ch.nvert);
  
  int finger=start;
  double lensofar=0;
  for(int cnt=0;cnt<nboundary;cnt++){
    int next=boundary[finger];
    if(next==-1){
      printf("error in boundary, next is -1\n");
      break;
    }
    double theta=2*M_PI*(lensofar/boundaryLength);
    float x=cos(theta);
    float y=sin(theta);
    
    //make it a square
    if(1){
      float fabsx=fabs(x);
      float fabsy=fabs(y);
      float max=(fabsx>fabsy)?fabsx:fabsy;
      x/=max;
      y/=max;
    }
    
    b[finger]=complex_f(x,y);
    for(int i=0;i<ch.nvert;i++){
      M[finger*ch.nvert+i]=complex_f(0,0);
    }
    M[finger*ch.nvert+finger]=1;
    lensofar+=edges[finger*ch.nvert+next];
    finger=boundary[finger];
  }
  if(finger!=start){
    printf("error in boundary, boundary does not loop\n");
  }
  complex_f * Mt = new complex_f[ch.nvert*ch.nvert];
  for(int i=0;i<ch.nvert;i++){
    for(int j=0;j<ch.nvert;j++){
      Mt[i*ch.nvert+j]=M[j*ch.nvert+i];
    }
  }
  //memcpy(Mt,M,sizeof(complex_f)*ch.nvert);
  //u=M\b...
  //tch.x = real(u)
  //tch.y = imag(u)
  int * pivot=new int[ch.nvert];
  int info;
  complex_f * result =new complex_f[ch.nvert];
  memcpy(result,b,sizeof(complex_f)*ch.nvert);

  printf("Boundary contains %d\n",nboundary);
  cgesv_(ch.nvert,1,(complex<float>*)Mt,ch.nvert,pivot,result,ch.nvert,info);
  printf("done tc gen\n");
  Matrix tc(ch.nvert,2);
  for(int i=0;i<ch.nvert;i++){
    tc(i,0)=result[i].real;
    tc(i,1)=result[i].imag;
  }
  printf("setting tc\n");
  ch.setTC(tc);
  printf("done settint tc\n");

  delete [] pivot;
  delete [] result;
  delete [] boundary;
  delete [] edges;
  delete [] M;
  delete [] Mt;
}

void getTexCoordsSVD(Chart & c)
{
  Matrix pts(3,c.nvert);
  for(int i=0;i<c.nvert;i++){
    pts(0,i)=c.vert[i].x;
    pts(1,i)=c.vert[i].y;
    pts(2,i)=c.vert[i].z;
  }
  pts.transpose().printMatlab("pts");
  Matrix covar=pts*pts.transpose();
  Matrix s,v,d;
  covar.svd(s,v,d);
  
  
  s.printMatlab("s");
  d.printMatlab("d");
  v.printMatlab("v");

  Matrix a,b;
  covar.eig(a,b);
  a.printMatlab("a");
  b.printMatlab("b");

  Matrix rot=d.transpose();
  Matrix tc=rot*pts;
  tc.transpose().printMatlab("tc");
  double minx=10e10,miny=10e10;
  double maxx=-10e10,maxy=-10e10;
  for(int i=0;i<c.nvert;i++){
    c.tc[i].x=tc(0,i);
    c.tc[i].y=tc(1,i);

    maxx=MAX(maxx,c.tc[i].x);
    maxy=MAX(maxy,c.tc[i].y);
    minx=MIN(minx,c.tc[i].x);
    miny=MIN(miny,c.tc[i].y);
  }
  for(int i=0;i<c.nvert;i++){
    c.tc[i].x=(c.tc[i].x-minx)/(maxx-minx);
    c.tc[i].y=(c.tc[i].y-miny)/(maxy-miny);
  }
}

//void getTexCoordsMDS(int ** poly,int npoly,
//		     Vec3f * vert,int nvert)
void getTexCoordsMDS(Chart & c)
{
  if(c.nvert<=0)return;
  Matrix M(c.nvert,c.nvert);
  Graph * g = allocGraph(c.nvert);
  int pred[c.nvert];
  for(int i=0;i<c.npoly;i++){
    graphAddEdge(g,c.poly[i][0],c.poly[i][1],
		 (c.vert[c.poly[i][0]]-c.vert[c.poly[i][1]]).len());
    graphAddEdge(g,c.poly[i][0],c.poly[i][2],
		 (c.vert[c.poly[i][0]]-c.vert[c.poly[i][2]]).len());
    graphAddEdge(g,c.poly[i][1],c.poly[i][2],
		 (c.vert[c.poly[i][1]]-c.vert[c.poly[i][2]]).len());
  }

  allocSingleSource(g);
  for(int i=0;i<c.nvert;i++){
    dij(g,i);
    getDistancesParam(M.data+c.nvert*i);
    getPredParam(pred);
    for(int j=0;j<c.nvert;j++){
      if(pred[j]==-1 && j!=i)
	printf("%d can not get to %d\n",i,j);
    }
    if(i%100)fflush(stdout);
  }
  freeSingleSource();
  freeGraph(g);

  Matrix J=Matrix::eye(c.nvert);
  {
    Matrix o(c.nvert,c.nvert);
    o.setAll(1.0/c.nvert);
    J-=o;
  }
  printf("Set J\n");
  M.dotmult(M);
  printf("done dot mult\n");
  
  /*Matrix JM=J*M;printf("got J*M\n");
  Matrix B=JM*J;printf("got JM*J\n");
  B*=(-0.5);printf("got B * -0.5\n");
  */

  Matrix result(c.nvert,c.nvert);
  Matrix B(c.nvert,c.nvert);

  //multiply fast using linpack
  dsymm_('L','U',c.nvert,c.nvert,
	 -0.5,J.data,c.nvert,M.data,c.nvert,0.0,
	 result.data,c.nvert);printf("done 1\n");
  dsymm_('L','U',c.nvert,c.nvert,
	 1.0,result.data,c.nvert,J.data,c.nvert,0.0,
	 B.data,c.nvert);printf("done 2\n");

  Matrix v,e;
  Matrix s,d;

  printf("getting eig\n");
  int lwork=c.nvert*c.nvert;
  if(c.nvert<8)lwork=128;
  double *wr=new double[c.nvert];
  double *wi=new double[c.nvert];
  double *vr=new double[c.nvert*c.nvert];
  double *work=new double[lwork];
  int info,* ifail = new int[c.nvert];
  int found;
  int * iwork = new int[5*c.nvert]; 
  double values[c.nvert];
  Matrix vectors(2,c.nvert);
  Matrix evals(2,2);
  evals.setAll(0);

  dsyevx_('V','I','U',c.nvert,B.data,c.nvert,
	  0,0,c.nvert-2+1,c.nvert,0.0,found,values,
	  vectors.data,c.nvert,work,lwork,
	  iwork,ifail,info);
  
  printf("%f %f\n",values[0],
	 values[1]);
  vectors=vectors.transpose();
  evals(0,0)=sqrt(values[0]);
  evals(1,1)=sqrt(values[1]);
  Matrix texcoords=vectors*evals;

  /*
  vectors.printMatlab("Vectors");
  texcoords.printMatlab("tc");

  printf("vectors\n");
  printf("tri=[\n");
  for(int i=0;i<c.npoly;i++){
    printf("%d %d %d\n",c.poly[i][0],
	   c.poly[i][1],
	   c.poly[i][2]);
  }
  printf("]\n");
  */
  c.setTC(texcoords);

  delete wr;
  delete wi;
  delete vr;
  delete work;
  delete ifail;
  /*
  dgeev_('N','V',nvert,
	 B.data,nvert,wr,wi,NULL,1,
	 vr,nvert,work,nvert*nvert,
	 info);
  */
   
  //B.eig(v,e,2);
  //B.svd(s,v,d);
}

void cone(Vec3f & v1,Vec3f & axis,double & angle);
int incone(Vec3f & axis,double angle,Vec3f & normal);
int cone_partial(Vec3f * normals,int nnorm,Vec3f & axis,double & angle,
		 int * support,int & insupport,int start=0);

int sharetwo(Chart * g,int i,int j)
{
  int share=0;
  if(i>g->npoly || j>g->npoly)printf("%d || %d > %d\n",g->npoly,i,j);
  for(int a=0;a<3;a++){
    for(int b=0;b<3;b++){
      share+=(g->poly[i][a]==g->poly[j][b]);
    }
  }
  return (share>=2);
}


void dfsSet(int i,int ** adj,int * adjcnt,int * colors,int from,int col)
{
  colors[i]=col;
  for(int j=0;j<adjcnt[i];j++){
    if(colors[adj[i][j]]==from){
      dfsSet(adj[i][j],adj,adjcnt,colors,from,col);
    }
  }
}

void getAdjacentPolys(Chart * g,int *** _adj,int ** _adjcnt);

Chart * split(Chart * g,int & ncolors,double coneAngle,int maxpoly)
{
  int i=0;
  int j=0;
  int s=0;//rand()%g->npoly;
  int setsofar=0;

  printf("in split %d,%d \n",g->npoly,g->nvert);

  int ** adj;
  int * adjcnt;
  getAdjacentPolys(g,&adj,&adjcnt);
  Vec3f * faceNorm = new Vec3f[g->npoly];
 
  int * colors;
  colors = new int[g->npoly]; 
  ncolors=1;
  memset(colors,0,sizeof(int)*g->npoly);

  //initialize the face normals
  for(i=0;i<g->npoly;i++){
    Vec3f a=g->vert[g->poly[i][1]]-g->vert[g->poly[i][0]];
    Vec3f b=g->vert[g->poly[i][2]]-g->vert[g->poly[i][0]];
    faceNorm[i]=a.cross(b);
    faceNorm[i].normalize();
  }

  Vec3f axis;
  double angle;
  int support[3];
  int insupport;
  Vec3f * normals = new Vec3f[g->npoly+10];
  int * front = new int[g->npoly+10];
  char * infront = new char[g->npoly];
  int inq=0;
  int nnorm=0;
  int added=0;

  int first=1;
  while(1){
    inq=0;
    memset(front,0,sizeof(int)*g->npoly);
    memset(infront,0,g->npoly);

    i=0;
    if(first){
      first=0;
      i=rand()%g->npoly;
    }
    for(i=0;i<g->npoly;i++){
      if(!colors[i]){
	s=i;
	break;
      }
    }
    if(i==g->npoly){
      printf("done\n");
      break;
    }
    added=1;
    nnorm=0;
    colors[s]=ncolors;
    normals[nnorm++]=faceNorm[s];
    insupport=0;
    cone_partial(normals,1,axis,angle,support,insupport);
    //cone(normals[0],axis,angle);

    //Add all neighbors
    for(j=0;j<adjcnt[s];j++){
      if(!infront[adj[s][j]] && 
	 !colors[adj[s][j]]){
	front[inq++]=adj[s][j];
	infront[adj[s][j]]=1;
      }
    }

    while(inq){
      int inqback=inq;
      int * frontback = new int[inqback];
      memcpy(frontback,front,sizeof(int)*inqback);
      inq=0;
      for(i=0;i<inqback;i++){
	int face=frontback[i];
	double ang=0;
	int okay=0;
	int sup[3];
	int nsup;
	Vec3f ax;
	int set=0;

	normals[nnorm]=faceNorm[face];
	if(incone(axis,angle,normals[nnorm])){
	  okay=1;
	}
	else{
	  set=1;
	  ax=axis;
	  ang=angle;
	  memcpy(sup,support,sizeof(int)*3);
	  nsup=insupport;

	  int res=cone_partial(normals,nnorm+1,ax,
			       ang,sup,nsup,nnorm);
		
	  //cone(normals,nnorm+1,ax,ang);
	  
	  if(res && ang<coneAngle)
	    okay=1;
	}
	if(okay){
	  if(set){
	    axis=ax;
	    angle=ang;
	    memcpy(support,sup,sizeof(int)*3);
	    insupport=nsup;
	  }
	  
	  colors[face]=ncolors;
	  added++;
	  nnorm++;
	  if(added>=maxpoly){
	      //printf("in split, added greater than max poly\n");
	  }

	  //add all neighbors
	  for(j=0;j<adjcnt[face];j++){
	    if(!infront[adj[face][j]] 
	       && !colors[adj[face][j]]){
	      front[inq++]=adj[face][j];
	      infront[adj[face][j]]=1;
	    }
	  }
	}
      }
      delete frontback;
      
      printf("%04d inq\r",inq);
    }
    if(added==1){
      colors[s]=-1;
    }
    else if(added<6){//used to be 10
      printf("added %d < 6, calling dfs\n",added);
      dfsSet(s,adj,adjcnt,colors,ncolors,-1);
    }
    else{
      printf("angle %f\n",angle*180/M_PI);
      printf("added %d, ncolors=%d\n",added,ncolors);
      ncolors++;
    }
    setsofar+=added;
  }

  printf("Added %d/%d\n",added,g->npoly);
  printf("colors: %d\n",ncolors);

  for(j=0;j<g->npoly;j++){
    if(colors[j]==-1){
      for(int k=0;k<adjcnt[j];k++){
	if(colors[adj[j][k]]>0){
	  dfsSet(j,adj,adjcnt,colors,-1,colors[adj[j][k]]);
	  break;
	  //colors[j]=colors[k];
	}
      }
    }
  }

  ncolors--;
  Chart * charts;
  int * vertsexist = new int[g->nvert];
  int * mapping;
  int biggest=0;
  int totalpolys=0;

  for(int i=0;i<g->npoly;i++){
    colors[i]--;
  }
  charts = new Chart[ncolors+1];
  memset(charts,0,sizeof(Chart)*(ncolors+1));

  for(i=0;i<ncolors;i++){
    int uniqueverts=0;
    int uniquepolys=0;
    memset(vertsexist,0xff,sizeof(int)*g->nvert);
    for(int j=0;j<g->npoly;j++){
      if(colors[j]!=i)continue;
      if(vertsexist[(g->poly)[j][0]]==-1){
	vertsexist[g->poly[j][0]]=uniqueverts;
	uniqueverts++;
      }
      if(vertsexist[(g->poly)[j][1]]==-1){
	vertsexist[g->poly[j][1]]=uniqueverts;
	uniqueverts++;
      }
      if(vertsexist[(g->poly)[j][2]]==-1){
	vertsexist[g->poly[j][2]]=uniqueverts;
	uniqueverts++;
      }
      uniquepolys++;
    }
    
    charts[i].init(uniquepolys,uniqueverts);
    uniquepolys=0;
    for(int j=0;j<g->npoly;j++){
      if(colors[j]!=i)continue;
      charts[i].poly[uniquepolys][0]=vertsexist[g->poly[j][0]];
      charts[i].poly[uniquepolys][1]=vertsexist[g->poly[j][1]];
      charts[i].poly[uniquepolys][2]=vertsexist[g->poly[j][2]];

      charts[i].vert[vertsexist[g->poly[j][0]]]=
	g->vert[g->poly[j][0]];
      charts[i].vert[vertsexist[g->poly[j][1]]]=
	g->vert[g->poly[j][1]];
      charts[i].vert[vertsexist[g->poly[j][2]]]=
	g->vert[g->poly[j][2]];

      charts[i].tc[vertsexist[g->poly[j][0]]]=
	g->tc[g->poly[j][0]];
      charts[i].tc[vertsexist[g->poly[j][1]]]=
	g->tc[g->poly[j][1]];
      charts[i].tc[vertsexist[g->poly[j][2]]]=
	g->tc[g->poly[j][2]];

      charts[i].norm[vertsexist[g->poly[j][0]]]=
	g->norm[g->poly[j][0]];
      charts[i].norm[vertsexist[g->poly[j][1]]]=
	g->norm[g->poly[j][1]];
      charts[i].norm[vertsexist[g->poly[j][2]]]=
	g->norm[g->poly[j][2]];
      uniquepolys++;
    }
    totalpolys+=uniquepolys;
    printf("chart[%d] unique verts %d (polys %d/%d)\n",i,uniqueverts,
	  totalpolys,g->npoly);
  }
  printf("done creating charts\n");
  delete vertsexist;
  delete normals;
  delete faceNorm;
  delete front;
  delete infront;
  delete adjcnt;
  delete colors;
  for(i=0;i<g->npoly;i++){
    delete adj[i];
  }
  delete adj;
  return charts;
}

int incone(Vec3f & axis,double angle,Vec3f & normal)
{
  double a=acos(axis.dot(normal)/(axis.len()*normal.len()));
  if(a<=(angle+10e-6))return 1;
  return 0;
}

int planeIntersection(Vec3f & n1,double d1,Vec3f & n2 ,double d2,
		       Vec3f &p,Vec3f &d)//return values
{
  Vec3f parallel=n1.cross(n2);
  if(fabs(parallel[0])+
     fabs(parallel[1])+
     fabs(parallel[2])<10e-10){
    printf("planeIntersection()----------parallel\n");
    return 0;
  }

  double det=n1.dot(n1)*n2.dot(n2)-n1.dot(n2)*n1.dot(n2);
  double c1=(d1*n2.dot(n2)-d2*n1.dot(n2))/det;
  double c2=(d2*n1.dot(n1)-d1*n1.dot(n2))/det;
  //printf("%f \n",det);
  d=n1.cross(n2);
  d.normalize();
  Vec3f a=n1*c1;
  Vec3f b=n2*c2;
  p=a+b;

  /*  
  for(double t=-10;t<10;t+=1){
    Vec3f point=d*t+p;
    printf("%f %f\n",n1.dot(point)-d1,n2.dot(point)-d2);
  }
  printf("line is :\n");
  p.print();
  d.print();*/
  return 1;
}

void cone(Vec3f & v1,Vec3f & axis,double & angle)
{
  axis=v1;
  angle=0;

}

void cone(Vec3f & v1,Vec3f & v2,Vec3f & axis,double & angle)
{
  axis=v1+v2;
  axis.mult(0.5);
  axis.normalize();
  angle=axis.dot(v2)/(axis.len()*v2.len());
  angle=acos(angle);
}

int cone(Vec3f & v1,Vec3f & v2,Vec3f & v3,Vec3f & axis,double & angle)
{
  Vec3f n1=v2-v1;
  Vec3f n2=v3-v2;

  Vec3f p1=n1*0.5+v1;
  Vec3f p2=n2*0.5+v2;

  double d1=n1.dot(p1);
  double d2=n2.dot(p2);
  
  Vec3f p;
  if(!planeIntersection(n1,d1,n2,d2,p,axis)){
    return 0;
  }
  if(acos(axis.dot(v1))>M_PI/2.0){
    axis.mult(-1);
    /*printf("cone()       : negating axis\n");
    printf("%f %f %f\n",v1.x,v1.y,v1.z);
    printf("%f %f %f\n",v2.x,v2.y,v2.z);
    printf("%f %f %f\n",v3.x,v3.y,v3.z);*/
  }
  angle=axis.dot(v1);
  double a2=axis.dot(v2);
  double a3=axis.dot(v3);
  if(a2>angle)angle=a2;
  if(a3>angle)angle=a3;
  angle=acos(angle);
  return 1;
}

int cone_updateSupport(int * support,int & insupport,
		       Vec3f * normals,int i,Vec3f & axis,double & angle)
{

  if(insupport<3){
    support[(insupport)++]=i;
    if((insupport)==2){
      cone(normals[support[0]],
	   normals[support[1]],
	   axis,angle);
    }
    else{
      if(!cone(normals[support[0]],
	       normals[support[1]],
	       normals[support[2]],axis,angle)){
	//printf("parallel ...\n");
      }
    }
  }
  else{
    Vec3f ax[3];
    double ang[3];
    int ok[3];
    memset(ok,0,sizeof(int)*3);
    ok[2]=cone(normals[support[0]],
	       normals[support[1]],
	       normals[i],ax[2],ang[2]);
    if(ok[2] && incone(ax[2],ang[2],normals[support[2]])){
      ok[2]=1;
    }else ok[2]=0;
    ok[1]=cone(normals[support[0]],
	       normals[support[2]],
	       normals[i],ax[1],ang[1]);
    if(ok[1] && incone(ax[1],ang[1],normals[support[1]])){
      ok[1]=1;
    }else ok[1]=0;
    ok[0]=cone(normals[support[1]],
	       normals[support[2]],
	       normals[i],ax[0],ang[0]);
    if(ok[0]&&incone(ax[0],ang[0],normals[support[0]])){
      ok[0]=1;
    }else ok[0]=0;
    int smallest=-1;
    
    for(int k=0;k<3;k++){
      if(ok[k]){
	if(smallest==-1)smallest=k;
	else if(ang[k]<ang[smallest]){
	  smallest=k;
	}
      }
    }
    if(smallest==-1){
      printf("------------\n");
      printf("new normal %f %f %f\n",
	     normals[i].x,normals[i].y,normals[i].z);
	     
      for(int k=0;k<3;k++){
	printf("support(%d) %f %f %f\n",k,
	       normals[support[k]].x,
	       normals[support[k]].y,
	       normals[support[k]].z);
      }
      printf("old cone\n");
      Vec3f oax;
      double oang;
      cone(normals[support[0]],
	   normals[support[1]],
	   normals[support[2]],oax,oang);
      printf("%f %f %f\n",oax.x,oax.y,oax.z);
      printf("angle %f\n",oang);

      printf("unable to find cone\n");
      return 0;
    }
    axis=ax[smallest];
    angle=ang[smallest];
    support[smallest]=i;
  }
  return 1;
}

int cone(Vec3f * normals,int nnorm,Vec3f & axis,double & angle)
{
  int support[3];
  int insupport=0;
  if(!nnorm){
    axis=Vec3f(0,0,0);
    angle=0;
    return 1;
  }
  support[0]=0;
  insupport=1;

  cone(normals[0],axis,angle);
  //printf("got initial cone %f\n",angle);
  for(int i=1;i<nnorm;i++){
    int sup=0;
    for(int j=0;j<insupport;j++){
      if(support[j]==i){
	sup=1;
	break;
      }
    }
    if(!sup){
      if(!incone(axis,angle,normals[i])){
	if(!cone_updateSupport(support,insupport,normals,i,axis,angle))
	  return 0;
	i=-1;
      }
    }
  }
  return 1;
}

int cone_partial(Vec3f * normals,int nnorm,Vec3f & axis,double & angle,
		 int * support,int & insupport,int start)
{
  if(!nnorm){
    axis=Vec3f(0,0,0);
    angle=0;
    insupport=0;
    return 1;
  }
  if(insupport==0){
    insupport=1;
    support[0]=0;
    cone(normals[0],axis,angle);
  }
  //printf("got initial cone %f\n",angle);
  for(int i=start;i<nnorm;i++){
    int sup=0;
    for(int j=0;j<insupport;j++){
      if(support[j]==i){
	sup=1;
	break;
      }
    }
    if(!sup){
      if(!incone(axis,angle,normals[i])){
	if(!cone_updateSupport(support,insupport,normals,i,axis,angle))
	  return 0;
	i=-1;
      }
    }
  }
  return 1;
}

Chart * splitPlane(Chart * c,int & ncharts){
  Vec3f mean;
  
  for(int i=0;i<c->nvert;i++){
    mean=mean+c->vert[i];
  }
  mean*=(1.0/c->nvert);
  printf("mean: %f %f %f\n",mean.x,mean.y,mean.z);

  int * c1map = new int[c->nvert];
  int * c2map = new int[c->nvert];
  int c1nvert=0,c2nvert=0;
  int c1npoly=0,c2npoly=0;
  memset(c1map,0xFF,sizeof(int)*c->nvert);
  memset(c2map,0xFF,sizeof(int)*c->nvert);

  printf("about to do map\n");

  for(int i=0;i<c->npoly;i++){
    int v1=c->poly[i][0];
    int v2=c->poly[i][1];
    int v3=c->poly[i][2];
    
    if(c->vert[v1].x>mean.x ||
       c->vert[v2].x>mean.x ||
       c->vert[v3].x>mean.x ){
      if(c1map[v1]<0)
	c1map[v1]=c1nvert++;
      if(c1map[v2]<0)
	c1map[v2]=c1nvert++;
      if(c1map[v3]<0)
	c1map[v3]=c1nvert++;
      c1npoly++;
    }
    else{
      if(c2map[v1]<0)
	c2map[v1]=c2nvert++;
      if(c2map[v2]<0)
	c2map[v2]=c2nvert++;
      if(c2map[v3]<0)
	c2map[v3]=c2nvert++;
      c2npoly++;
    }
  }
  printf("c1: %d,%d c2:%d,%d\n",c1npoly,c1nvert,c2npoly,c2nvert);
  Chart * newCharts=new Chart[2];
  newCharts[0].init(c1npoly,c1nvert);
  newCharts[1].init(c2npoly,c2nvert);
  c1npoly=0;
  c2npoly=0;
  for(int i=0;i<c->nvert;i++){
    if(c1map[i]>=0){
      newCharts[0].vert[c1map[i]]=c->vert[i];
      newCharts[0].tc[c1map[i]]=c->tc[i];
    }
    if(c2map[i]>=0){
      newCharts[1].vert[c2map[i]]=c->vert[i];
      newCharts[1].tc[c2map[i]]=c->tc[i];
    }
  }
  for(int i=0;i<c->npoly;i++){
    int v1=c->poly[i][0];
    int v2=c->poly[i][1];
    int v3=c->poly[i][2];
    
    if(c->vert[v1].x>mean.x ||
       c->vert[v2].x>mean.x ||
       c->vert[v3].x>mean.x ){
      newCharts[0].poly[c1npoly][0]=c1map[v1];
      newCharts[0].poly[c1npoly][1]=c1map[v2];
      newCharts[0].poly[c1npoly][2]=c1map[v3];
      c1npoly++;
    }
    else{
      newCharts[1].poly[c2npoly][0]=c2map[v1];
      newCharts[1].poly[c2npoly][1]=c2map[v2];
      newCharts[1].poly[c2npoly][2]=c2map[v3];
      c2npoly++;
    }
  }
  printf("c1: %d,%d c2:%d,%d\n",c1npoly,c1nvert,c2npoly,c2nvert);
  printf("done map\n");
  newCharts[0].initNormals();
  newCharts[1].initNormals();
  printf("done initing normals\n");
  ncharts=2;
  delete c1map;
  delete c2map;
  printf("returning\n");
  return newCharts;
}

void getAdjacentPolys(Chart * g,int *** _adj,int ** _adjcnt){
  int ** adj = new int * [g->npoly];
  int * adjcnt = new int[g->npoly];
  short * infcnt = new short[g->nvert];
  int ** inf = new int *[g->nvert]; //in face
  int i=0;
  int j=0;
  memset(infcnt,0,sizeof(short)*g->nvert);
  for(i=0;i<g->npoly;i++){
    infcnt[g->poly[i][0]]++;
    infcnt[g->poly[i][1]]++;
    infcnt[g->poly[i][2]]++;
  }
  for(i=0;i<g->nvert;i++){
    inf[i]=new int[infcnt[i]];
  }
  memset(infcnt,0,sizeof(short)*g->nvert);
  for(i=0;i<g->npoly;i++){
    inf[g->poly[i][0]][infcnt[g->poly[i][0]]++]=i;
    inf[g->poly[i][1]][infcnt[g->poly[i][1]]++]=i;
    inf[g->poly[i][2]][infcnt[g->poly[i][2]]++]=i;
  }
  memset(adjcnt,0,sizeof(int)*g->npoly);
  memset(adj,0,sizeof(int *)*g->npoly);
  //determine memory needed for each face
  for(i=0;i<g->npoly;i++){
    for(int k=0;k<3;k++){
      for(j=0;j<infcnt[g->poly[i][k]];j++){
	int f2=inf[g->poly[i][k]][j];
	if(i!=f2 && sharetwo(g,i,f2)){
	  adjcnt[i]++;
	}
      }
    }
  }
  //allocate memory for adjacency and reset the
  //count to zero
  for(i=0;i<g->npoly;i++){
    adj[i]=new int[adjcnt[i]];
    adjcnt[i]=0;
  }
  //initialize the adjacency list and the face normals
  for(i=0;i<g->npoly;i++){
    for(int k=0;k<3;k++){
      for(j=0;j<infcnt[g->poly[i][k]];j++){
	int f2=inf[g->poly[i][k]][j];
	if(i!=f2 && sharetwo(g,i,f2)){
	  adj[i][adjcnt[i]++]=f2;
	}
      }
    }
  }
  for(i=0;i<g->nvert;i++){
    delete inf[i];
  }
  delete [] inf;
  delete [] infcnt;
  *_adj=adj;
  *_adjcnt=adjcnt;
}

Chart * splitDistance(Chart * c,int & ncharts,int p1,int p2){
  double farthest=0;
  int    farthestPoly[2];
  if(p1>=0 && p2>=0){
    farthestPoly[0]=p1;
    farthestPoly[1]=p2;
  }
  else{
    for(int i=0;i<c->npoly;i++){
      Vec3f ci=(c->vert[c->poly[i][0]]+
		c->vert[c->poly[i][1]]+
		c->vert[c->poly[i][2]])*(1.0/3.0);
      for(int j=i+1;j<c->npoly;j++){
	Vec3f cj=(c->vert[c->poly[j][0]]+
		  c->vert[c->poly[j][1]]+
		  c->vert[c->poly[j][2]])*(1.0/3.0);
	double d=(ci-cj).len();
	
	if(d>farthest){
	  farthest=d;
	  farthestPoly[0]=i;
	  farthestPoly[1]=j;
	}
      }
    }
  }
  printf("farthest poly are %d and %d %f\n",
	 farthestPoly[0],farthestPoly[1],farthest);
  
  int ** adj;
  int * adjcnt;
  printf("getting adjacent polys\n");
  getAdjacentPolys(c,&adj,&adjcnt);
  printf("got adjacent polys\n");

  int * q1 = new int[c->npoly+1];
  int * q2 = new int[c->npoly+1];
  int q1tail=0,q2tail=0;
  int q1head=0,q2head=0;
  char * inq1 = new char[c->npoly];
  char * inq2 = new char[c->npoly];

  int * c1map,* c2map;
  int c1nvert=0;
  int c2nvert=0;
  int c1npoly=0;
  int c2npoly=0;

  Chart * charts=new Chart[2];
  charts[0].init(c->npoly,c->nvert);
  charts[1].init(c->npoly,c->nvert);

  memset(q1,0,sizeof(int)*c->npoly);
  memset(q2,0,sizeof(int)*c->npoly);
  memset(inq1,0,c->npoly);
  memset(inq2,0,c->npoly);

  c1map = new int[c->nvert];
  c2map = new int[c->nvert];
  memset(c1map,-1,sizeof(int)*c->nvert);
  memset(c2map,-1,sizeof(int)*c->nvert);

  q1[q1tail++]=farthestPoly[0];
  q2[q2tail++]=farthestPoly[1];
  inq1[farthestPoly[0]]=1;
  inq2[farthestPoly[1]]=1;
  //notice that as soon as something is flagged in inq[1-2],
  //that it never comes out
  while(q1tail-q1head>0||q2tail-q2head>0){
    printf("inq1: %d, inq2: %d\n",q1tail-q1head,q2tail-q2head);
    if(q1tail-q1head>0){
      int i=q1[q1head++];
      if(c1map[c->poly[i][0]]==-1){
	charts[0].vert[c1nvert]=c->vert[c->poly[i][0]];
	charts[0].tc[c1nvert]=c->tc[c->poly[i][0]];
	c1map[c->poly[i][0]]=c1nvert++;
      }
      if(c1map[c->poly[i][1]]==-1){
	charts[0].vert[c1nvert]=c->vert[c->poly[i][1]];
	charts[0].tc[c1nvert]=c->tc[c->poly[i][1]];
	c1map[c->poly[i][1]]=c1nvert++;
      }
      if(c1map[c->poly[i][2]]==-1){
	charts[0].vert[c1nvert]=c->vert[c->poly[i][2]];
	charts[0].tc[c1nvert]=c->tc[c->poly[i][2]];
	c1map[c->poly[i][2]]=c1nvert++;
      }
      charts[0].poly[c1npoly][0]=c1map[c->poly[i][0]];
      charts[0].poly[c1npoly][1]=c1map[c->poly[i][1]];
      charts[0].poly[c1npoly][2]=c1map[c->poly[i][2]];
      c1npoly++;

      for(int j=0;j<adjcnt[i];j++){
	if(!inq1[adj[i][j]] &&
	   !inq2[adj[i][j]]){
	  q1[q1tail++]=adj[i][j];
	  inq1[adj[i][j]]=1;
	}
      }
    }
    if(q2tail-q2head>0){
      int i=q2[q2head++];
      if(c2map[c->poly[i][0]]==-1){
	charts[1].vert[c2nvert]=c->vert[c->poly[i][0]];
	charts[1].tc[c2nvert]=c->tc[c->poly[i][0]];
	c2map[c->poly[i][0]]=c2nvert++;
      }
      if(c2map[c->poly[i][1]]==-1){
	charts[1].vert[c2nvert]=c->vert[c->poly[i][1]];
	charts[1].tc[c2nvert]=c->tc[c->poly[i][1]];
	c2map[c->poly[i][1]]=c2nvert++;
      }
      if(c2map[c->poly[i][2]]==-1){
	charts[1].vert[c2nvert]=c->vert[c->poly[i][2]];
	charts[1].tc[c2nvert]=c->tc[c->poly[i][2]];
	c2map[c->poly[i][2]]=c2nvert++;
      }
      charts[1].poly[c2npoly][0]=c2map[c->poly[i][0]];
      charts[1].poly[c2npoly][1]=c2map[c->poly[i][1]];
      charts[1].poly[c2npoly][2]=c2map[c->poly[i][2]];
      c2npoly++;

      for(int j=0;j<adjcnt[i];j++){
	if(!inq1[adj[i][j]] &&
	   !inq2[adj[i][j]]){
	  q2[q2tail++]=adj[i][j];
	  inq2[adj[i][j]]=1;
	}
      }
    }
  }
  //delete queues
  delete q1;
  delete q2;
  delete inq1;
  delete inq2;
  
  printf("0: %d %d, 1: %d %d orig:%d %d\n",
	 c1npoly,c1nvert,c2npoly,c2nvert,c->npoly,c->nvert);
  Chart * newCharts=new Chart[2];
  newCharts[0].init(c1npoly,c1nvert);
  newCharts[1].init(c2npoly,c2nvert);
  
  memcpy(newCharts[0].poly,charts[0].poly,sizeof(newCharts[0].poly[0])*
	 newCharts[0].npoly);
  memcpy(newCharts[0].tc,charts[0].tc,sizeof(newCharts[0].tc[0])*
	 newCharts[0].nvert);
  memcpy(newCharts[0].vert,charts[0].vert,sizeof(newCharts[0].vert[0])*
	 newCharts[0].nvert);
  
  memcpy(newCharts[1].poly,charts[1].poly,sizeof(newCharts[1].poly[0])*
	 newCharts[1].npoly);
  memcpy(newCharts[1].tc,charts[1].tc,sizeof(newCharts[1].tc[0])*
	   newCharts[1].nvert);
    memcpy(newCharts[1].vert,charts[1].vert,sizeof(newCharts[1].vert[0])*
	   newCharts[1].nvert);
    
    newCharts[0].initNormals();
    newCharts[1].initNormals();
    
    printf("Split 0:%d %d, 1: %d %d\n",
	   c1npoly,c1nvert,c2npoly,c2nvert);

    delete [] charts;
    delete c1map;
    delete c2map;
    printf("deleting adj\n");
    for(int i=0;i<c->npoly;i++){
      if(adj[i])
	delete adj[i];
    }
    delete [] adj;
    delete [] adjcnt;
    printf("done deleting adj\n");
    
    ncharts=2;
    return newCharts;
}

#ifdef TEXGEN_MAIN

#include <stdarg.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <GL/gl.h>
#include <GL/glut.h>

int ind=0;
int splitting=0;
int packing=0;
enum {CONE,DISTANCE} splitMethod;
char fileBase[1024];
int nfiles=1;
int ncharts=0;
Chart * charts=0;
int ww=256,wh=256;
double  * hor;
void processCommandLine(int ac,char * av[])
{
  static struct option long_opt[] = {
    {"pack", 1, 0, 0},
    {"cone", 1, 0, 0},
    {"distance", 1, 0, 0},
    {"file", 1, 0, 0},
    {"nfiles", 1, 0, 0},
    {0, 0, 0, 0}
  };
  int temp=0;
  splitting=0;
  packing=0;
  while(1){
    int opt_index=0;
    char c=getopt_long_only(ac,av,"",long_opt,&opt_index);
    if(c==-1)break;
    switch(opt_index){
    case 0:
      packing=1;
      break;
    case 1:
      splitting=1;
      splitMethod=CONE;
      break;
    case 2:
      splitting=1;
      splitMethod=DISTANCE;
      break;
    case 3:
      sscanf(optarg,"%s",fileBase);
      break;
    case 4:
      sscanf(optarg,"%d",&nfiles);
      break;
    default:
      sscanf(optarg,"%d",temp);
      break;
    }
  }
  charts = new Chart[nfiles];
  if(nfiles==1){
    charts[0].loadobj(fileBase);
  }
  else{
    for(int i=0;i<nfiles;i++){
      char fname[1024];
      sprintf(fname,fileBase,i);
      charts[i].loadobj(fname);
    }
  }
  ncharts=nfiles;
}

void keys(unsigned char c,int x,int y)
{
  switch(c){
  case 'a':{
    ind=0;
    packChartsNew(charts,ncharts,ww,wh);
    ind=ncharts;
  }break;
  case 'o':{
    ind=0;
    while(ind<ncharts){
      hor=packChartsDebug(charts,ncharts,ww,wh,ind);
      ind++;
    }
  }break;
  case 'p':
    /*packChartsNew(charts,ncharts,ww,wh,ind);
    ind++;
    if(ind>=ncharts)ind=0;*/
    break;
  }
}
void reshape(int x,int y)
{
  ww=x;
  wh=y;
  glViewport(0,0,ww,wh);
}
void mouse(GLint button,GLint state,int x,int y)
{   
}

void display()
{
  glClearColor(0.0,0.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glViewport(0,0,ww,wh);
  glOrtho(-2,ww+2,2,1.1*wh,-1,1);

  glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  glColor3f(1.0,0.0,0.0);
  glBegin(GL_QUADS);
  glVertex2i(0,0);
  glVertex2i(ww,0);
  glVertex2i(ww,wh);
  glVertex2i(0,wh);
  glEnd();


  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(1.0,1.0,1.0,0.5);
  glPushMatrix();

  for(int i=0;i<ind;i++){
    charts[i].drawTC();
  }
  glPopMatrix();
  
  glDisable(GL_BLEND);
  glColor3f(0.0,0.0,1.0);
  if(hor){
    glBegin(GL_LINES);
    for(int i=0;i<ww-1;i++){
      glVertex2d(i,hor[i]);
      glVertex2d(i+1,hor[i+1]);
    }
    glEnd();
  }

  glutSwapBuffers();
  glutPostRedisplay();
}

int main(int ac,char * av[])
{
  processCommandLine(ac,av);
  glutInit(&ac,av);
  glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE);
  glutInitWindowSize(ww,wh);
  glutInitWindowPosition(120,120);
  glutCreateWindow("texture");
  glutDisplayFunc(display);
  glutMouseFunc(mouse);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keys);

  glutMainLoop();  
}
#endif
