#ifndef VEC_H
#define VEC_H
#include <math.h>
template <class T>
class Vec2{
public:
  union{
    struct{
      T x;
      T y;
    };
    struct{
      T data[2];
    };
  };
  Vec2(){
    x=y=0;
  }
  Vec2(T x,T y){
    this->x=x;
    this->y=y;
  }
  void mult(const double o){
    x*=o;
    y*=o;
  }
  void operator+=(const Vec2 &o){
    x+=o.x;
    y+=o.y;
  }
  void operator-=(const Vec2 & o){
    x-=o.x;
    y-=o.y;
  }
  double dot(const Vec2 & o){
    return x*o.x+y*o.y;
  }
  Vec2 operator-(const Vec2& o){
    return Vec2(x-o.x,y-o.y);
  }
  Vec2 operator+(const Vec2& o){
    return Vec2(x+o.x,y+o.y);
  }
  Vec2 operator*(const T & t) const{
    return Vec2(x*t,y*t);
  }
  //friend operator*(const T &t,const Vec2& o);
  
  void operator*=(T t){
    x*=t;
    y*=t;
  }
  void rotate(double d){
    double c=cos(d);
    double s=sin(d);
    double xn=c*x-s*y;
    double yn=s*x+c*y;
    x=xn;
    y=yn;
  }
  double normalize(){
    double l=len();
    x/=l;
    y/=l;
    return l;
  }
  double len(){
    return sqrt(x*x+y*y);
  }
  T & operator[](int i){
    return data[i];
  }
  T * getData(){
    return &x;
  }
};
typedef Vec2<float> Vec2f;
typedef Vec2<double> Vec2d;
template <class T>
class Vec3{
public:
  union{
    struct{
      T x;
      T y;
      T z;
    };
    struct{
      T data[3];
    };
  };
  Vec3(){
    x=y=z=0;
  }
  Vec3(const Vec3 & o){
    x=o.x;
    y=o.y;
    z=o.z;
  }
  Vec3(T _x,T _y,T _z){
    x=_x;
    y=_y;
    z=_z;
  }
  double len() const{
    return sqrt(x*x+y*y+z*z);
  }
  double dot(const Vec3 & other) const{
    return x*other.x+y*other.y+z*other.z;
  }
  void mult(double d){
    x*=d;
    y*=d;
    z*=d;
  }
  void add(Vec3 & other){
    x+=other.x;
    y+=other.y;
    z+=other.z;
  }
  void sub(Vec3 & other){
    x-=other.x;
    y-=other.y;
    z-=other.z;
  }
  double normalize(){
    double length=len();
    x/=length;
    y/=length;
    z/=length;
    return length;
  }
  Vec3 cross(const Vec3 & other){
    Vec3 res;
    res.x=y*other.z-z*other.y;
    res.y=-x*other.z+z*other.x;
    res.z=x*other.y-y*other.x;
    return res;
  }
  T * getData(){
    return &x;
  }
  T & operator[](int i){
    return (&x)[i];
  }
  Vec3 operator*(const Vec3 & o){
    return Vec3(x*o.x,y*o.y,z*o.z);
  }
  Vec3 operator*(const T & t) const{
    return Vec3(x*t,y*t,z*t);
  }
  Vec3 operator+(const T & t){
    return Vec3(x+t,y+t,z+t);
  }
  Vec3 operator-(const T & t) const{
    return Vec3(x-t,y-t,z-t);
  }
  Vec3 operator+(const Vec3 & o)const{
    return Vec3(x+o.x,y+o.y,z+o.z);
  }
  Vec3 operator-(const Vec3 & o)const {
    return Vec3(x-o.x,y-o.y,z-o.z);
  }
  void operator+=(const Vec3 & o){
    x+=o.x;
    y+=o.y;
    z+=o.z;
  }
  void operator-=(const Vec3 & o){
    x-=o.x;
    y-=o.y;
    z-=o.z;
  }
  void operator*=(const T & t){
    x*=t;
    y*=t;
    z*=t;
  }
  void operator/=(const T & t){
    x/=t;
    y/=t;
    z/=t;
  }
  void rotatez(double d){
    double c=cos(d);
    double s=sin(d);
    double xn=c*x-s*y;
    double yn=s*x+c*y;
    x=xn;
    y=yn;
  }
  void rotatex(double d){
    double c=cos(d);
    double s=sin(d);
    double yn=c*y-s*z;
    double zn=s*y+c*z;
    y=yn;
    z=zn;
  }
  void rotatey(double d){
    double c=cos(d);
    double s=sin(d);
    double xn= c*x+s*z;
    double zn=-s*x+c*z;
    x=xn;
    z=zn;
  }
  void lerp(double a,Vec3 p1,Vec3 p2){
    double b=1-a;
    data[0]=a*p1[0]+b*p2[0];
    data[1]=a*p1[1]+b*p2[1];
    data[2]=a*p1[2]+b*p2[2];
  }
  int operator==(const Vec3 & o) const{
    return x==o.x&&y==o.y&&z==o.z;
  }
};

typedef Vec3<double> Vec3d;
typedef Vec3<float> Vec3f;

#endif
